/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (C) 2021-2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"
#include <infiniband/ibutils/output_stream_primitives.h>

extern IBDiagClbck ibDiagClbck;

int IBDiag::BuildEnhancedCC(list_p_fabric_general_err& congestion_control_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    bool all_devices_sup_cc = true;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &congestion_control_errors);

    ProgressBarNodes progress_bar;

    CC_EnhancedCongestionInfo cc_congestion_info;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::CCEnhancedCongestionInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                    " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || p_curr_node->isSpecialNode())
            continue;

        // check that Enhanced Congestion Info supported
        if (!this->capability_module.IsSupportedSMPCapability(p_curr_node,
                EnSMPCapIsVSCCSupported)) {
            all_devices_sup_cc = false;

            congestion_control_errors.push_back(
                new FabricErrNodeNotSupportCap(p_curr_node, "Enhanced Congestion Control is not supported"));

            continue;
        }

        IBPort *p_curr_port = NULL;
        if (p_curr_node->type == IB_SW_NODE) {
            p_curr_port = p_curr_node->getPort(0);
        } else {
            for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
                p_curr_port = p_curr_node->getPort((phys_port_t)i);
                if (p_curr_port && p_curr_port->is_data_worthy())
                    break;
            }
        }
        if (!p_curr_port)
            continue;

        clbck_data.m_data1 = p_curr_node;
        uint16_t lid = p_curr_port->base_lid;

        // 1.1 Enhanced Congestion Info
        progress_bar.push(p_curr_node);
        this->ibis_obj.CCEnhancedInfoGet(lid,
                                         0, // sl
                                         &cc_congestion_info,
                                         &clbck_data);

        if(ibDiagClbck.GetState())
           goto exit;
    } // nodes loop

    // set a warning if some devices not support CC
    if (!all_devices_sup_cc)
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!congestion_control_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildCCSwithGeneralSettings(list_p_fabric_general_err& congestion_control_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    CC_CongestionSwitchGeneralSettings curr_switch_general_settings;

    ProgressBarNodes progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &congestion_control_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::CCSwitchGeneralSettingsGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                    " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // this MADs are only for switches
        if (!p_curr_node->getInSubFabric() || p_curr_node->type != IB_SW_NODE)
            continue;

        // check if Enhanced Congestion Control is supported
        CC_EnhancedCongestionInfo *p_cc_enhanced_info =
                fabric_extended_info.getCCEnhancedCongestionInfo(p_curr_node->createIndex);
        if (!p_cc_enhanced_info) // not supported
            continue;
        if (!p_cc_enhanced_info->ver0) { // no version supported
            congestion_control_errors.push_back(
                new FabricErrNodeNotSupportCap(p_curr_node,
                                               "This device does not support " \
                                               "any version of Congestion Control attributes"));

            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;
        }

        IBPort *p_curr_port0 = p_curr_node->getPort(0); // get switch port 0
        if (!p_curr_port0)
            continue;

        uint16_t lid = p_curr_port0->base_lid;

        // 1.2 Congestion Switch General Settings
        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);
        this->ibis_obj.CCSwitchGeneralSettingsGet(lid,
                                                  0, // sl
                                                  &curr_switch_general_settings,
                                                  &clbck_data);
        if(ibDiagClbck.GetState())
            goto exit;
    } // nodes loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!congestion_control_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::BuildCCSwithConfig(list_p_fabric_general_err& congestion_control_errors,
                               u_int64_t& enabled_SWs)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    CC_CongestionPortProfileSettings curr_port_profile_settings;
    CC_CongestionSLMappingSettings curr_sl_mapping_settings;

    ProgressBarPorts progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &congestion_control_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                    " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // this MADs are only for switches
        if (!p_curr_node->getInSubFabric() || p_curr_node->type != IB_SW_NODE)
            continue;

        // get switch port 0 and base lid
        if (!p_curr_node->getPort(0))
            continue;
        uint16_t lid = p_curr_node->getPort(0)->base_lid;

        // Check if CC is enabled on device
        CC_CongestionSwitchGeneralSettings *p_switch_general_settings =
                this->fabric_extended_info.getCCSwitchGeneralSettings(p_curr_node->createIndex);
        if (!p_switch_general_settings) // No data for device
            continue;
        if (!p_switch_general_settings->en) // CC is not enabled on device
            continue;

        // for Fabric Summary - count enabled Switches
        ++enabled_SWs;

        // go over all ports of this switch:
        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_sw_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_sw_port || !p_curr_sw_port->is_data_worthy())
                continue;

            // skip ANs' ports
            if (p_curr_sw_port->isSpecialPort())
                continue;

            clbck_data.m_data1 = p_curr_sw_port;

            // 1.3 Congestion Port Profile Settings
            clbck_data.m_handle_data_func =
                &forwardClbck<IBDiagClbck, &IBDiagClbck::CCPortProfileSettingsGetClbck>;

            // get OperationalVLs for this port
            struct SMP_PortInfo * portInfo =
              fabric_extended_info.getSMPPortInfo(p_curr_sw_port->createIndex);
            if(!portInfo) {
                this->SetLastError("DB error - can't find SMP port info port=%s",
                                   p_curr_sw_port->getName().c_str());

                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            u_int8_t op_vls = portInfo->OpVLs;
            u_int8_t vl_num = get_operational_vl_num(op_vls);

            // go over all OperationalVLs of this port
            for (u_int8_t vl = 0; vl < vl_num; ++vl) {
                CLEAR_STRUCT(curr_port_profile_settings);
                curr_port_profile_settings.vl_mask = (u_int16_t)(1 << vl);
                clbck_data.m_data2 = (void *)(uintptr_t)vl;

                progress_bar.push(p_curr_sw_port);

                this->ibis_obj.CCPortProfileSettingsGet(lid,
                                                        pi, // port
                                                        0,  // sl
                                                        &curr_port_profile_settings,
                                                        &clbck_data);
                if(ibDiagClbck.GetState())
                    goto exit;
            }

            // 1.4 Congestion SL Mapping Settings
            clbck_data.m_handle_data_func =
                &forwardClbck<IBDiagClbck, &IBDiagClbck::CCSLMappingSettingsGetClbck>;

            progress_bar.push(p_curr_sw_port);
            this->ibis_obj.CCSLMappingSettingsGet(lid,
                                                  pi, // port
                                                  0,  // sl
                                                  &curr_sl_mapping_settings,
                                                  &clbck_data);
            if(ibDiagClbck.GetState())
                goto exit;

        } // ports loop
    } // nodes loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!congestion_control_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildCCHCAGeneralSettings(list_p_fabric_general_err& congestion_control_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    CC_CongestionHCAGeneralSettings curr_hca_general_settings;

    ProgressBarPorts progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &congestion_control_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::CCHCAGeneralSettingsGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;


    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                    " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // this MADs are only for HCAs
        if (!p_curr_node->getInSubFabric() || p_curr_node->type != IB_CA_NODE)
            continue;

        // check if Enhanced Congestion Control is supported
        CC_EnhancedCongestionInfo *p_cc_enhanced_info =
                fabric_extended_info.getCCEnhancedCongestionInfo(p_curr_node->createIndex);
        if (!p_cc_enhanced_info) // not supported
            continue;

        if (!p_cc_enhanced_info->ver0 && !p_cc_enhanced_info->ver1) { // no version supported
            congestion_control_errors.push_back(
                new FabricErrNodeNotSupportCap(p_curr_node,
                    "This device does not support " \
                    "any version of Congestion Control attributes"));

            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;
        }

        if (p_cc_enhanced_info->ver0 && p_cc_enhanced_info->ver1) { // both versions are supported

            FabricErrNodeWrongConfig* p_err = new FabricErrNodeWrongConfig(p_curr_node,
                    "This device support both versions of Congestion Control attributes");

            p_err->SetLevel(EN_FABRIC_ERR_WARNING);
            congestion_control_errors.push_back(p_err);
        }

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            clbck_data.m_data1 = p_curr_port;
            uint16_t lid = p_curr_port->base_lid;

            // 1.5 Congestion HCA General Settings
            progress_bar.push(p_curr_port);
            this->ibis_obj.CCHCAGeneralSettingsGet(lid,
                                                   0, // sl
                                                   &curr_hca_general_settings,
                                                   &clbck_data);
            if(ibDiagClbck.GetState())
                goto exit;
        } // ports loop
    } // nodes loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!congestion_control_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildCCHCAConfig(list_p_fabric_general_err& congestion_control_errors, u_int64_t& enabled_CAs)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    CC_CongestionHCARPParameters    curr_hca_rp_parameters;
    CC_CongestionHCANPParameters    curr_hca_np_parameters;

    // flag for 'Fabric Summary' counter
    bool node_counted = false;

    ProgressBarPorts progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &congestion_control_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                    " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        node_counted = false;

        // this MADs are only for HCAs
        if (!p_curr_node->getInSubFabric() || p_curr_node->type != IB_CA_NODE)
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            clbck_data.m_data1 = p_curr_port;
            uint16_t lid = p_curr_port->base_lid;

            CC_EnhancedCongestionInfo *p_cc_enhanced_info =
                    fabric_extended_info.getCCEnhancedCongestionInfo(p_curr_node->createIndex);
            if (!p_cc_enhanced_info) // not supported
                continue;

            // Check if CC react/notify is enabled on device
            CC_CongestionHCAGeneralSettings *p_hca_general_settings =
                    this->fabric_extended_info.getCCHCAGeneralSettings(p_curr_port->createIndex);
            if (!p_hca_general_settings) // No data for device
                continue;

            // Validation - add warning if not both en_react and en_notify are on
            if (!p_hca_general_settings->en_react || !p_hca_general_settings->en_notify) {

                stringstream sstr;
                sstr << "HCA General Settings: Not both en_react and en_notify are on on port "
                     << p_curr_port->getName()
                     << endl;

                FabricErrPortInvalidValue* p_curr_err =
                        new FabricErrPortInvalidValue(p_curr_port, sstr.str());
                p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);

                congestion_control_errors.push_back(p_curr_err);
            } else {
                if (!node_counted) {
                    ++enabled_CAs;
                    node_counted = true;
                }
            }

            // works only with ver0
            if (p_cc_enhanced_info->ver0 &&
                    p_hca_general_settings->en_react) { // CC react is enabled  on device
                // 1.6 Congestion HCA RP Parameters
                clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::CCHCARPParametersGetClbck>;

                progress_bar.push(p_curr_port);
                this->ibis_obj.CCHCARPParametersGet(lid,
                                                    0, // sl
                                                    &curr_hca_rp_parameters,
                                                    &clbck_data);
                if(ibDiagClbck.GetState())
                    goto exit;
            }

            // works with ver0 and ver1
            if (p_hca_general_settings->en_notify) { // CC notify is enabled  on device
                // 1.7 Congestion HCA NP Parameters
                clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::CCHCANPParametersGetClbck>;

                progress_bar.push(p_curr_port);
                this->ibis_obj.CCHCANPParametersGet(lid,
                                                    0, // sl
                                                    &curr_hca_np_parameters,
                                                    &clbck_data);
                if(ibDiagClbck.GetState())
                    goto exit;
            }
        } // ports loop
    } // nodes loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!congestion_control_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::BuildCCHCAStatisticsQuery(list_p_fabric_general_err& congestion_control_errors,
                                      bool to_clear_congestion_counters)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    CC_CongestionHCAStatisticsQuery curr_hca_statistics_query;

    ProgressBarPorts progress_bar;

    ibDiagClbck.Set(this, &fabric_extended_info, &congestion_control_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::CCHCAStatisticsQueryGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                    " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // this MADs are only for HCAs
        if (!p_curr_node->getInSubFabric() || p_curr_node->type != IB_CA_NODE)
            continue;

        // check if Enhanced Congestion Control is supported
        CC_EnhancedCongestionInfo *p_cc_enhanced_info =
                fabric_extended_info.getCCEnhancedCongestionInfo(p_curr_node->createIndex);
        if (!p_cc_enhanced_info) // not supported
            continue;

        // no version supported and/or HCA Statistics Query is not supported
        if ((!p_cc_enhanced_info->ver0 && !p_cc_enhanced_info->ver1) ||
                !IsSupportedCCCapability(p_cc_enhanced_info->CC_Capability_Mask,
                                         EnCCHCAStatisticsQuerySupported)) {

            congestion_control_errors.push_back(
                new FabricErrNodeNotSupportCap(p_curr_node,
                    "This device does not support " \
                    "Congestion Control HCA Statistics Query MAD"));

            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;
        }

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            // Check if CC react or CC notify are enabled on device
            CC_CongestionHCAGeneralSettings *p_hca_general_settings =
                    this->fabric_extended_info.getCCHCAGeneralSettings(p_curr_port->createIndex);
            if (!p_hca_general_settings) // No data for device
                continue;

            if (!p_hca_general_settings->en_react && !p_hca_general_settings->en_notify)
                continue;

            clbck_data.m_data1 = p_curr_port;
            uint16_t lid = p_curr_port->base_lid;

            // 1.8 Congestion HCA Statistics Query
            progress_bar.push(p_curr_port);
            this->ibis_obj.CCHCAStatisticsQueryGet(lid,
                                                   0, // sl
                                                   &curr_hca_statistics_query,
                                                   &clbck_data,
                                                   to_clear_congestion_counters);
            if(ibDiagClbck.GetState())
                goto exit;

        } // ports loop
    } // nodes loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!congestion_control_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


// CC Algo Config Support build
int IBDiag::Build_CC_HCA_AlgoConfigSup(list_p_fabric_general_err& cc_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarPorts progress_bar;
    CC_CongestionHCAAlgoConfig cc_hca_algo_config;

    ibDiagClbck.Set(this, &fabric_extended_info, &cc_errors);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::CC_HCA_AlgoConfigSupGetClbck>;

    for (set_pnode::iterator nI = this->discovered_fabric.HCAs.begin();
            nI != this->discovered_fabric.HCAs.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in HCAs set");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || p_curr_node->isSpecialNode())
            continue;

        // check if Enhanced Congestion Control is supported
        CC_EnhancedCongestionInfo *p_cc_enhanced_info =
                fabric_extended_info.getCCEnhancedCongestionInfo(p_curr_node->createIndex);

        if (!p_cc_enhanced_info) // not supported
            continue;

        if (!p_cc_enhanced_info->ver1 ||    // version 1 not supported
                !IsSupportedCCCapability(p_cc_enhanced_info->CC_Capability_Mask,
                                         EnCCHCA_AlgoConfigParamCounetrsSupported)) {
            cc_errors.push_back(new FabricErrNodeNotSupportCap(p_curr_node,
                                                               "This device does not support CC HCA " \
                                                               "Algo Config attributes"));
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;
        }

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            CC_CongestionHCAGeneralSettings *p_cc_hca_general_settings =
                fabric_extended_info.getCCHCAGeneralSettings(p_curr_port->createIndex);

            if (!p_cc_hca_general_settings) // not supported
                continue;

            if (!p_cc_hca_general_settings->en_notify || !p_cc_hca_general_settings->en_react) {

                cc_errors.push_back(new FabricErrPortNotSupportCap(p_curr_port,
                        "PPCC is not enabled on this HCA port"));

                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }

            clbck_data.m_data1 = p_curr_port;
            uint16_t lid = p_curr_port->base_lid;
            progress_bar.push(p_curr_port);

            this->ibis_obj.CCHCAAlgoConfigGet(lid,
                                              0,     // algo_slot - ignored
                                              CC_ALGO_ENCAP_TYPE_2,  // encap_type
                                              &cc_hca_algo_config,
                                              &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;

        } // ports
    } // nodes

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!cc_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

// CC Algo Config Support dump
int IBDiag::DumpCC_HCA_AlgoConfigSupCSVTable(CSVOut &csv_out, list_p_fabric_general_err& cc_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (csv_out.DumpStart(SECTION_CC_HCA_ALGO_CONFIG_SUP))
        IBDIAG_RETURN(rc);

    stringstream sstream;
    sstream << "NodeGUID,"    << "PortGUID,"  << "algo_en,"
            << "algo_status," << "trace_en,"  << "counter_en,"
            << "sl_bitmask,"  << "encap_len," << "encap_type";

    for (u_int8_t algo = 0; algo < MAX_CC_ALGO_SLOT; ++algo)
        sstream << "," << "algo_id_" << +algo
                << "," << "algo_major_version_" << +algo
                << "," << "algo_minor_version_" << +algo;

    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.HCAs.begin();
            nI != this->discovered_fabric.HCAs.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config_sup =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigSup(p_curr_port->createIndex);

            if (!p_cc_hca_algo_config_sup)
                continue;

            sstream.str("");

            sstream << PTR(p_curr_node->guid_get())                 << ","
                    << PTR(p_curr_port->guid_get())                 << ","
                    << +p_cc_hca_algo_config_sup->algo_en           << ","
                    << +p_cc_hca_algo_config_sup->algo_status       << ","
                    << +p_cc_hca_algo_config_sup->trace_en          << ","
                    << +p_cc_hca_algo_config_sup->counters_en       << ","
                    << PTR(p_cc_hca_algo_config_sup->sl_bitmask)    << ","
                    << +p_cc_hca_algo_config_sup->encap_len         << ","
                    << +p_cc_hca_algo_config_sup->encap_type        ;

            CC_CongestionHCAAlgoConfigInfo cc_hca_algo_config_info;
            CC_CongestionHCAAlgoConfigInfo_unpack(&cc_hca_algo_config_info,
                    (const u_int8_t*)p_cc_hca_algo_config_sup->encapsulation);

            // make sure that encap_len is aligned to 4 bytes
            if (p_cc_hca_algo_config_sup->encap_len % 4 != 0) {

                stringstream sstr;
                sstr << "HCAAlgoConfig.encap_len ("
                     << +p_cc_hca_algo_config_sup->encap_len
                     << ") is not aligned to 4 bytes on port "
                     << p_curr_port->getName()
                     << endl;

                FabricErrPortInvalidValue* p_curr_err =
                        new FabricErrPortInvalidValue(p_curr_port, sstr.str());
                p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);

                cc_errors.push_back(p_curr_err);
            }

            // make sure that encap_len is less than MAX_CC_ALGO_SLOT (16*4)
            int cur_len = p_cc_hca_algo_config_sup->encap_len / 4;

            if (cur_len > MAX_CC_ALGO_SLOT) {
                cur_len = MAX_CC_ALGO_SLOT;

                stringstream sstr;
                sstr << "HCAAlgoConfig.encap_len ("
                     << +p_cc_hca_algo_config_sup->encap_len
                     << ") is bigger than max ("
                     << MAX_CC_ALGO_SLOT * 4
                     << ") on port "
                     << p_curr_port->getName()
                     << endl;

                FabricErrPortInvalidValue* p_curr_err =
                        new FabricErrPortInvalidValue(p_curr_port, sstr.str());
                p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);

                cc_errors.push_back(p_curr_err);
            }

            for (u_int8_t algo = 0; algo < cur_len; ++algo)
                sstream << "," << +cc_hca_algo_config_info.info[algo].algo_id
                << "," << +cc_hca_algo_config_info.info[algo].algo_major_version
                << "," << +cc_hca_algo_config_info.info[algo].algo_minor_version;

            // fill NAs for not exist algos
            for (u_int8_t algo = 0; algo < MAX_CC_ALGO_SLOT - cur_len; ++algo)
                sstream << ",NA,NA,NA";

            sstream  << endl;

            csv_out.WriteBuf(sstream.str());

        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_ALGO_CONFIG_SUP);

    IBDIAG_RETURN(rc);
}

// CC Algo Config Info build
int IBDiag::Build_CC_HCA_AlgoConfig(list_p_fabric_general_err& cc_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarPorts progress_bar;
    CC_CongestionHCAAlgoConfig cc_hca_algo_config;

    ibDiagClbck.Set(this, &fabric_extended_info, &cc_errors);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::CC_HCA_AlgoConfigGetClbck>;

    for (set_pnode::iterator nI = this->discovered_fabric.HCAs.begin();
            nI != this->discovered_fabric.HCAs.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in HCAs set");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || p_curr_node->isSpecialNode())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config_sup =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigSup(p_curr_port->createIndex);

            if (!p_cc_hca_algo_config_sup)
                continue;

            clbck_data.m_data1 = p_curr_port;
            uint16_t lid = p_curr_port->base_lid;

            CC_CongestionHCAAlgoConfigInfo cc_hca_algo_config_info;
            CC_CongestionHCAAlgoConfigInfo_unpack(&cc_hca_algo_config_info,
                                                  (const u_int8_t*)p_cc_hca_algo_config_sup->encapsulation);

            int cur_len = min(p_cc_hca_algo_config_sup->encap_len / 4, MAX_CC_ALGO_SLOT);

            for (int algo = 0; algo < cur_len; ++algo) {

                // skip if algo_id is 0
                if (!cc_hca_algo_config_info.info[algo].algo_id)
                    continue;

                clbck_data.m_data2 = (void *)(uintptr_t)algo;
                progress_bar.push(p_curr_port);

                this->ibis_obj.CCHCAAlgoConfigGet(lid,
                                                  (u_int8_t)algo,  // algo_slot
                                                  CC_ALGO_ENCAP_TYPE_1,  // encap_type
                                                  &cc_hca_algo_config,
                                                  &clbck_data);

                if (ibDiagClbck.GetState())
                    goto exit;
            } // algo
        } // ports
    } // nodes

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!cc_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

// CC Algo Config Info dump
void IBDiag::DumpCC_HCA_AlgoConfigCSVTable(CSVOut &csv_out,
                                           u_int64_t& enabled_algo_CAs,
                                           u_int64_t& enabled_algo_CA_ports,
                                           u_int64_t& disabled_algo_CA_ports)
{
    IBDIAG_ENTER;

    if (csv_out.DumpStart(SECTION_CC_HCA_ALGO_CONFIG))
        IBDIAG_RETURN_VOID;

    // flags for 'Fabric Summary' counters
    bool node_counted = false;
    bool port_counted = false;
    bool algo_exist   = false;

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," << "algo_slot,"
            << "algo_en," << "algo_status," << "trace_en,"
            << "counter_en," << "sl_bitmask," << "encap_len,"
            << "encap_type," << "algo_info_text"
            << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.HCAs.begin();
            nI != this->discovered_fabric.HCAs.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        node_counted = false;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            port_counted = false;
            algo_exist   = false;

            for (u_int8_t algo = 0; algo < MAX_CC_ALGO_SLOT; ++algo) {

                CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config = this->fabric_extended_info.
                        getCC_HCA_AlgoConfig(p_curr_port->createIndex, algo);

                if (!p_cc_hca_algo_config)
                    continue;

                algo_exist = true;

                sstream.str("");
                sstream << PTR(p_curr_node->guid_get())             << ","
                        << PTR(p_curr_port->guid_get())             << ","
                        << +algo                                    << ","
                        << +p_cc_hca_algo_config->algo_en           << ","
                        << +p_cc_hca_algo_config->algo_status       << ","
                        << +p_cc_hca_algo_config->trace_en          << ","
                        << +p_cc_hca_algo_config->counters_en       << ","
                        << PTR(p_cc_hca_algo_config->sl_bitmask)    << ","
                        << +p_cc_hca_algo_config->encap_len         << ","
                        << +p_cc_hca_algo_config->encap_type        << ","
                        // char encapsulation[177];
                        << "\"" << p_cc_hca_algo_config->encapsulation << "\""
                        << endl;

                csv_out.WriteBuf(sstream.str());

                // Update 'Fabric Summary' counters
                if (p_cc_hca_algo_config->algo_en) {
                    if (!node_counted) {
                        ++enabled_algo_CAs;
                        node_counted = true;
                    }

                    if (!port_counted) {
                        ++enabled_algo_CA_ports;
                        port_counted = true;
                    }
                }

            } // algos

            // if no algo is enabled for this port - count it as disabled
            if (algo_exist && !port_counted)
                ++disabled_algo_CA_ports;

        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_ALGO_CONFIG);

    IBDIAG_RETURN_VOID;
}

// CC Algo Config Params build
int IBDiag::Build_CC_HCA_AlgoConfigParams(list_p_fabric_general_err& cc_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarPorts progress_bar;
    CC_CongestionHCAAlgoConfigParams cc_hca_algo_config_params;

    ibDiagClbck.Set(this, &fabric_extended_info, &cc_errors);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::CC_HCA_AlgoConfigParamsGetClbck>;

    for (set_pnode::iterator nI = this->discovered_fabric.HCAs.begin();
            nI != this->discovered_fabric.HCAs.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in HCAs set");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || p_curr_node->isSpecialNode())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config_sup =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigSup(p_curr_port->createIndex);

            if (!p_cc_hca_algo_config_sup)
                continue;

            clbck_data.m_data1 = p_curr_port;
            uint16_t lid = p_curr_port->base_lid;

            CC_CongestionHCAAlgoConfigInfo cc_hca_algo_config_info;
            CC_CongestionHCAAlgoConfigInfo_unpack(&cc_hca_algo_config_info,
                                                  (const u_int8_t*)p_cc_hca_algo_config_sup->encapsulation);

            int cur_len = min(p_cc_hca_algo_config_sup->encap_len / 4, MAX_CC_ALGO_SLOT);

            for (int algo = 0; algo < cur_len; ++algo) {

                // skip if algo_id is 0
                if (!cc_hca_algo_config_info.info[algo].algo_id)
                    continue;

                clbck_data.m_data2 = (void *)(uintptr_t)algo;
                progress_bar.push(p_curr_port);

                this->ibis_obj.CCHCAAlgoConfigParamGet(lid,
                                                       (u_int8_t)algo,  // algo_slot
                                                       CC_ALGO_ENCAP_TYPE_2,  // encap_type
                                                       &cc_hca_algo_config_params,
                                                       &clbck_data);

                if (ibDiagClbck.GetState())
                    goto exit;
            } // algo
        } // ports
    } // nodes

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!cc_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

// CC Algo Config Params dump
int IBDiag::DumpCC_HCA_AlgoConfigParamsCSVTable(CSVOut &csv_out, list_p_fabric_general_err& cc_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (csv_out.DumpStart(SECTION_CC_HCA_ALGO_CONFIG_PARAMS))
        IBDIAG_RETURN(rc);

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," << "algo_slot,"
            << "sl_bitmask," << "encap_len," << "encap_type";

    for (u_int8_t i = 0; i < MAX_CC_ALGO_ARRAY_LAYOUT; ++i)
        sstream << ",congestion_param_" << +i;

    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.HCAs.begin();
            nI != this->discovered_fabric.HCAs.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            for (u_int8_t algo = 0; algo < MAX_CC_ALGO_SLOT; ++algo) {

                CC_CongestionHCAAlgoConfigParams* p_cc_hca_algo_config_params =
                        this->fabric_extended_info.getCC_HCA_AlgoConfigParams(p_curr_port->createIndex, algo);

                if (!p_cc_hca_algo_config_params)
                    continue;

                sstream.str("");

                sstream << PTR(p_curr_node->guid_get())                    << ","
                        << PTR(p_curr_port->guid_get())                    << ","
                        << +algo                                           << ","
                        << PTR(p_cc_hca_algo_config_params->sl_bitmask)    << ","
                        << +p_cc_hca_algo_config_params->encap_len         << ","
                        << +p_cc_hca_algo_config_params->encap_type        ;

                // make sure that encap_len is aligned to 4 bytes
                if (p_cc_hca_algo_config_params->encap_len % 4 != 0) {

                    stringstream sstr;
                    sstr << "HCAConfigParam.encap_len ("
                         << +p_cc_hca_algo_config_params->encap_len
                         << ") is not aligned to 4 bytes on port "
                         << p_curr_port->getName()
                         << ", algo = "
                         << +algo
                         << endl;

                    FabricErrPortInvalidValue* p_curr_err =
                            new FabricErrPortInvalidValue(p_curr_port, sstr.str());
                    p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);

                    cc_errors.push_back(p_curr_err);
                }

                // make sure that encap_len is less than MAX_CC_ALGO_ARRAY_LAYOUT (44*4)
                int cur_len = p_cc_hca_algo_config_params->encap_len / 4;

                if (cur_len > MAX_CC_ALGO_ARRAY_LAYOUT) {
                    cur_len = MAX_CC_ALGO_ARRAY_LAYOUT;

                    stringstream sstr;
                    sstr << "HCAConfigParam.encap_len ("
                         << +p_cc_hca_algo_config_params->encap_len
                         << ") is bigger than max ("
                         << MAX_CC_ALGO_ARRAY_LAYOUT * 4
                         << ") on port "
                         << p_curr_port->getName()
                         << ", algo = "
                         << +algo
                         << endl;

                    FabricErrPortInvalidValue* p_curr_err =
                            new FabricErrPortInvalidValue(p_curr_port, sstr.str());
                    p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);

                    cc_errors.push_back(p_curr_err);
                }

                for (u_int8_t i = 0; i < cur_len; ++i)
                    sstream << "," << p_cc_hca_algo_config_params->encapsulation[i];

                for (u_int8_t i = 0; i < MAX_CC_ALGO_ARRAY_LAYOUT - cur_len; ++i)
                    sstream << ",NA";

                sstream << endl;

                csv_out.WriteBuf(sstream.str());

            } // algo
        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_ALGO_CONFIG_PARAMS);

    IBDIAG_RETURN(rc);
}

// CC Algo Counters build
int IBDiag::Build_CC_HCA_AlgoCounters(list_p_fabric_general_err& cc_errors,
                                      bool to_clear_congestion_counters)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarPorts progress_bar;
    CC_CongestionHCAAlgoCounters cc_hca_algo_counters;

    ibDiagClbck.Set(this, &fabric_extended_info, &cc_errors);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::CC_HCA_AlgoCountersGetClbck>;

    for (set_pnode::iterator nI = this->discovered_fabric.HCAs.begin();
            nI != this->discovered_fabric.HCAs.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in HCAs set");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || p_curr_node->isSpecialNode())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config_sup =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigSup(p_curr_port->createIndex);

            if (!p_cc_hca_algo_config_sup)
                continue;

            clbck_data.m_data1 = p_curr_port;
            uint16_t lid = p_curr_port->base_lid;

            CC_CongestionHCAAlgoConfigInfo cc_hca_algo_config_info;
            CC_CongestionHCAAlgoConfigInfo_unpack(&cc_hca_algo_config_info,
                                                  (const u_int8_t*)p_cc_hca_algo_config_sup->encapsulation);

            int cur_len = min(p_cc_hca_algo_config_sup->encap_len / 4, MAX_CC_ALGO_SLOT);

            for (int algo = 0; algo < cur_len; ++algo) {

                // skip if algo_id is 0
                if (!cc_hca_algo_config_info.info[algo].algo_id)
                    continue;

                CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config = this->fabric_extended_info.
                        getCC_HCA_AlgoConfig(p_curr_port->createIndex, algo);
                if (!p_cc_hca_algo_config)
                    continue;

                // skip if counters not enable
                if (!p_cc_hca_algo_config->counters_en)
                    continue;

                clbck_data.m_data2 = (void *)(uintptr_t)algo;
                progress_bar.push(p_curr_port);

                if (to_clear_congestion_counters)
                    this->ibis_obj.CCHCAAlgoCountersSet(lid,
                                                        (u_int8_t)algo,  // algo_slot
                                                        CC_ALGO_ENCAP_TYPE_2,  // encap_type
                                                        &cc_hca_algo_counters,
                                                        &clbck_data,
                                                        to_clear_congestion_counters);
                else
                    this->ibis_obj.CCHCAAlgoCountersGet(lid,
                                                        (u_int8_t)algo,  // algo_slot
                                                        CC_ALGO_ENCAP_TYPE_2,  // encap_type
                                                        &cc_hca_algo_counters,
                                                        &clbck_data);

                if (ibDiagClbck.GetState())
                    goto exit;
            } // algo
        } // ports
    } // nodes

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!cc_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

// CC Algo Counters dump
int IBDiag::DumpCC_HCA_AlgoCountersCSVTable(CSVOut &csv_out, list_p_fabric_general_err& cc_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (csv_out.DumpStart(SECTION_CC_HCA_ALGO_COUNTERS))
        IBDIAG_RETURN(rc);

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," << "algo_slot,"
            << "clear," << "sl_bitmask,"
            << "encap_len," << "encap_type";

    for (u_int8_t i = 0; i < MAX_CC_ALGO_ARRAY_LAYOUT; ++i)
        sstream << ",congestion_counter_" << +i;

    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.HCAs.begin();
            nI != this->discovered_fabric.HCAs.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            for (u_int8_t algo = 0; algo < MAX_CC_ALGO_SLOT; ++algo) {

                CC_CongestionHCAAlgoCounters* p_cc_hca_algo_counters =
                        this->fabric_extended_info.getCC_HCA_AlgoCounters(p_curr_port->createIndex, algo);

                if (!p_cc_hca_algo_counters)
                    continue;

                sstream.str("");

                sstream << PTR(p_curr_node->guid_get())               << ","
                        << PTR(p_curr_port->guid_get())               << ","
                        << +algo                                      << ","
                        << +p_cc_hca_algo_counters->clear             << ","
                        << PTR(p_cc_hca_algo_counters->sl_bitmask)    << ","
                        << +p_cc_hca_algo_counters->encap_len         << ","
                        << +p_cc_hca_algo_counters->encap_type        ;

                // make sure that encap_len is aligned to 4 bytes
                if (p_cc_hca_algo_counters->encap_len % 4 != 0) {

                    stringstream sstr;
                    sstr << "HCAAlgoCounters.encap_len ("
                         << +p_cc_hca_algo_counters->encap_len
                         << ") is not aligned to 4 bytes on port "
                         << p_curr_port->getName()
                         << ", algo = "
                         << +algo
                         << endl;

                    FabricErrPortInvalidValue* p_curr_err =
                            new FabricErrPortInvalidValue(p_curr_port, sstr.str());
                    p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);

                    cc_errors.push_back(p_curr_err);
                }

                // make sure that encap_len is less than MAX_CC_ALGO_ARRAY_LAYOUT (44*4)
                int cur_len = p_cc_hca_algo_counters->encap_len / 4;

                if (cur_len > MAX_CC_ALGO_ARRAY_LAYOUT) {
                    cur_len = MAX_CC_ALGO_ARRAY_LAYOUT;

                    stringstream sstr;
                    sstr << "HCAAlgoCounters.encap_len ("
                         << +p_cc_hca_algo_counters->encap_len
                         << ") is bigger than max ("
                         << MAX_CC_ALGO_ARRAY_LAYOUT * 4
                         << ") on port "
                         << p_curr_port->getName()
                         << ", algo = "
                         << +algo
                         << endl;

                    FabricErrPortInvalidValue* p_curr_err =
                            new FabricErrPortInvalidValue(p_curr_port, sstr.str());
                    p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);

                    cc_errors.push_back(p_curr_err);
                }

                for (u_int8_t i = 0; i < cur_len; ++i)
                    sstream << "," << p_cc_hca_algo_counters->encapsulation[i];

                for (u_int8_t i = 0; i < MAX_CC_ALGO_ARRAY_LAYOUT - cur_len; ++i)
                    sstream << ",NA";

                sstream << endl;

                csv_out.WriteBuf(sstream.str());

            } // algo
        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_ALGO_COUNTERS);

    IBDIAG_RETURN(rc);
}

// CC Algo Validations
int IBDiag::CC_HCA_AlgoValidations(list_p_fabric_general_err& cc_errors,
                                   PPCCAlgoDatabase &ppcc_algo_db)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    for (set_pnode::iterator nI = this->discovered_fabric.HCAs.begin();
            nI != this->discovered_fabric.HCAs.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in HCAs set");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || p_curr_node->isSpecialNode())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config_sup =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigSup(p_curr_port->createIndex);
            if (!p_cc_hca_algo_config_sup)
                continue;

            CC_CongestionHCAAlgoConfigInfo cc_hca_algo_config_sup_encap;
            CC_CongestionHCAAlgoConfigInfo_unpack(&cc_hca_algo_config_sup_encap,
                                                  (const u_int8_t*)p_cc_hca_algo_config_sup->encapsulation);

            list <int> counters_en_algos;
            map <u_int8_t, list <int> > algos_per_sl;
            map <u_int8_t, list <int> > params_per_sl;

            int cur_len = min(p_cc_hca_algo_config_sup->encap_len / 4, MAX_CC_ALGO_SLOT);

            for (int algo = 0; algo < cur_len; ++algo) {

                CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config_info =
                        this->fabric_extended_info.getCC_HCA_AlgoConfig(p_curr_port->createIndex, algo);
                if (!p_cc_hca_algo_config_info)
                    continue;

                // 1. counter_en is on for only 1 algo on port
                if (p_cc_hca_algo_config_info->counters_en)
                    counters_en_algos.push_back(algo);

                // 2. algo config SL bit mask validation
                for (u_int8_t sl = 0; sl < IB_NUM_SL; ++sl)
                    if (p_cc_hca_algo_config_info->sl_bitmask & (1 << sl))
                        algos_per_sl[sl].push_back(algo);

                CC_CongestionHCAAlgoConfigParams* p_cc_hca_algo_config_params =
                        this->fabric_extended_info.getCC_HCA_AlgoConfigParams(p_curr_port->createIndex, algo);
                if (!p_cc_hca_algo_config_params)
                    continue;

                // 3. algo config params SL bit mask validation
                for (u_int8_t sl = 0; sl < IB_NUM_SL; ++sl)
                    if (p_cc_hca_algo_config_params->sl_bitmask & (1 << sl))
                        params_per_sl[sl].push_back(algo);

                // 4. Params range validation
                PPCCAlgo* p_algo = ppcc_algo_db.GetAlgo(cc_hca_algo_config_sup_encap.info[algo].algo_id,
                                                        cc_hca_algo_config_sup_encap.info[algo].algo_major_version,
                                                        cc_hca_algo_config_sup_encap.info[algo].algo_minor_version);
                if (!p_algo)
                    continue;

                int cur_len = min(p_cc_hca_algo_config_params->encap_len / 4, MAX_CC_ALGO_ARRAY_LAYOUT);

                for (int param_i = 0; param_i < cur_len; ++param_i) {

                    PPCCParameter* p_param = p_algo->GetParameter(param_i);

                    if (!p_param)
                        continue;

                    uint32_t minVal = 0;
                    uint32_t maxVal = 0;

                    if (p_param->GetMinVal(minVal) &&
                            p_cc_hca_algo_config_params->encapsulation[param_i] < minVal) {
                        stringstream desc;
                        desc << "Param " << param_i
                             << " of Algo " << +algo
                             << " is lower than min value " << minVal
                             << ". value = " << p_cc_hca_algo_config_params->encapsulation[param_i];

                        cc_errors.push_back(new CC_AlgoParamRangeErr(p_curr_port, desc.str()));
                    }

                    if (p_param->GetMaxVal(maxVal) &&
                            p_cc_hca_algo_config_params->encapsulation[param_i] > maxVal) {
                        stringstream desc;
                        desc << "Param " << param_i
                             << " of Algo " << +algo
                             << " is bigger than max value " << maxVal
                             << ". value = " << p_cc_hca_algo_config_params->encapsulation[param_i];

                        cc_errors.push_back(new CC_AlgoParamRangeErr(p_curr_port, desc.str()));
                    }

                } // params
            } // algo

            if (counters_en_algos.size() > 1)
                cc_errors.push_back(new CC_AlgoCounterEnErr(p_curr_port, counters_en_algos));

            for (u_int8_t sl = 0; sl < IB_NUM_SL; ++sl) {
                if (algos_per_sl[sl].size() > 1) // more than 1 algo per sl
                    cc_errors.push_back(new CC_AlgoSLEnErr(p_curr_port, sl, algos_per_sl[sl]));

                if (params_per_sl[sl].size() > 1) // more than 1 algo params per sl
                    cc_errors.push_back(new CC_AlgoParamsSLEnErr(p_curr_port, sl, params_per_sl[sl]));
            }

        } // ports
    } // nodes

    IBDIAG_RETURN(rc);
}

// CC Algo dump to ppcc file
int IBDiag::WriteCC_HCA_AlgoDataToFile(const string& file_name, PPCCAlgoDatabase &ppcc_algo_db)
{
    IBDIAG_ENTER;

    ofstream sout;
    int rc = this->OpenFile("PPCC HCA Algo",
                            file_name,
                            sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->DumpCC_AlgoData(sout, ppcc_algo_db);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


void IBDiag::DumpCC_AlgoData(ofstream &sout, PPCCAlgoDatabase &ppcc_algo_db)
{
    IBDIAG_ENTER;
    for (set_pnode::iterator nI = this->discovered_fabric.HCAs.begin();
            nI != this->discovered_fabric.HCAs.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config_sup =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigSup(p_curr_port->createIndex);
            if (!p_cc_hca_algo_config_sup)
                continue;

            sout << "\n-------------------------------------------------------------------------" << endl;
            sout << "Port="  << +p_curr_port->num
                 << " Lid="  <<  p_curr_port->base_lid
                 << " GUID=" <<  PTR(p_curr_port->guid_get())
                 << " Port Name=" <<  p_curr_port->getName().c_str()
                 << endl;
            sout << "-------------------------------------------------------------------------" << endl;

            sout << "algo_config_sup_encap_len="  << +p_cc_hca_algo_config_sup->encap_len  << endl
                 << "algo_config_sup_encap_type=" << +p_cc_hca_algo_config_sup->encap_type << endl;

            CC_CongestionHCAAlgoConfigInfo cc_hca_algo_config_info;
            CC_CongestionHCAAlgoConfigInfo_unpack(&cc_hca_algo_config_info,
                    (const u_int8_t*)p_cc_hca_algo_config_sup->encapsulation);

            int cur_len = min(p_cc_hca_algo_config_sup->encap_len / 4, MAX_CC_ALGO_SLOT);

            for (int algo = 0; algo < cur_len; ++algo) {

                // skip if algo_id is 0
                if (!cc_hca_algo_config_info.info[algo].algo_id)
                    continue;

                sout << "--------------------------------------------------" << endl;

                sout << "algo_slot=" << algo
                     << ", algo_id=" << +cc_hca_algo_config_info.info[algo].algo_id
                     << ", algo_major_version=" << +cc_hca_algo_config_info.info[algo].algo_major_version
                     << ", algo_minor_version=" << +cc_hca_algo_config_info.info[algo].algo_minor_version << endl;

                PPCCAlgo* p_algo = ppcc_algo_db.GetAlgo(cc_hca_algo_config_info.info[algo].algo_id,
                                                        cc_hca_algo_config_info.info[algo].algo_major_version,
                                                        cc_hca_algo_config_info.info[algo].algo_minor_version);

                sout << "  algo_name=" << (p_algo ? p_algo->GetName() : "NA") << endl;

                CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config = this->fabric_extended_info.
                        getCC_HCA_AlgoConfig(p_curr_port->createIndex, algo);

                if (p_cc_hca_algo_config) {
                    sout << "  algo_en="       << +p_cc_hca_algo_config->algo_en        << endl;
                    sout << "  algo_status="   << +p_cc_hca_algo_config->algo_status    << endl;
                    sout << "  trace_en="      << +p_cc_hca_algo_config->trace_en       << endl;
                    sout << "  counters_en="   << +p_cc_hca_algo_config->counters_en    << endl;
                    sout << "  sl_bitmask="    << PTR(p_cc_hca_algo_config->sl_bitmask) << endl;
                    sout << "  encapsulation=" << p_cc_hca_algo_config->encapsulation   << endl;
                } else {
                    sout << "  algo_en=NA"       << endl;
                    sout << "  algo_status=NA"   << endl;
                    sout << "  trace_en=NA"      << endl;
                    sout << "  counters_en=NA"   << endl;
                    sout << "  sl_bitmask=NA"    << endl;
                    sout << "  encapsulation=NA" << endl;
                }

                // CC_HCA_AlgoConfigParams
                CC_CongestionHCAAlgoConfigParams* p_cc_hca_algo_config_params =
                        this->fabric_extended_info.getCC_HCA_AlgoConfigParams(p_curr_port->createIndex, algo);

                sout << "CC_HCA_AlgoConfigParams:" << endl;

                if (p_cc_hca_algo_config_params) {

                    sout << "  params_sl_bitmask=" << PTR(p_cc_hca_algo_config_params->sl_bitmask) << endl;
                    sout << "  params_encap_len="  << +p_cc_hca_algo_config_params->encap_len  << endl;
                    sout << "  params_encap_type=" << +p_cc_hca_algo_config_params->encap_type << endl;

                    int cur_param_len = min(p_cc_hca_algo_config_params->encap_len / 4,
                                            MAX_CC_ALGO_ARRAY_LAYOUT);

                    for (int param_i = 0; param_i < cur_param_len; ++param_i) {

                        PPCCParameter* p_param = p_algo ? p_algo->GetParameter(param_i) : NULL;

                        sout << "  param index="   << param_i
                             << ", param name="  << (p_param ? p_param->GetName() : "NA")
                             << ", param value=" << p_cc_hca_algo_config_params->encapsulation[param_i]
                             << endl;
                    }

                } else {
                    sout << "  params_sl_bitmask=NA" << endl;
                    sout << "  params_encap_len=NA"  << endl;
                    sout << "  params_encap_type=NA" << endl;
                }

                // CC_HCA_AlgoCounters
                CC_CongestionHCAAlgoCounters* p_cc_hca_algo_counters =
                        this->fabric_extended_info.getCC_HCA_AlgoCounters(p_curr_port->createIndex, algo);

                sout << "CC_HCA_AlgoCounters:" << endl;

                if (p_cc_hca_algo_counters) {

                    sout << "  Counters sl_bitmask=" << PTR(p_cc_hca_algo_counters->sl_bitmask) << endl;
                    sout << "  encap_len=" << +p_cc_hca_algo_counters->encap_len << endl;
                    sout << "  encap_type=" << +p_cc_hca_algo_counters->encap_type << endl;

                    int cur_cntr_len = min(p_cc_hca_algo_counters->encap_len / 4,
                                           MAX_CC_ALGO_ARRAY_LAYOUT);

                    for (int cntr_i = 0; cntr_i < cur_cntr_len; ++cntr_i) {

                        sout << "  counter index="   << cntr_i
                             << ", counter name="  << ((!p_algo || p_algo->GetCounter(cntr_i) == "") ?
                                                       "NA" : p_algo->GetCounter(cntr_i))
                             << ", counter value=" << p_cc_hca_algo_counters->encapsulation[cntr_i]
                             << endl;
                    }

                } else {
                    sout << "  Counters sl_bitmask=NA" << endl;
                    sout << "  encap_len=NA"           << endl;
                    sout << "  encap_type=NA"          << endl;
                }
            } // algo
        } // ports
    } // nodes

    IBDIAG_RETURN_VOID;
}


// CC - Dump functions
void IBDiag::DumpCCEnhancedInfoCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (csv_out.DumpStart(SECTION_CC_ENHANCED_INFO))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID,ver0Supported,ver1Supported,CC_Capability_Mask" << endl;
    csv_out.WriteBuf(sstream.str());

    struct CC_EnhancedCongestionInfo *p_curr_cc_enhanced_info = NULL;

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        p_curr_cc_enhanced_info = this->fabric_extended_info.
                getCCEnhancedCongestionInfo(p_curr_node->createIndex);
        if (!p_curr_cc_enhanced_info)
            continue;

        sstream.str("");

        sstream << PTR(p_curr_node->guid_get())                      << ","
                << +p_curr_cc_enhanced_info->ver0                    << ","
                << +p_curr_cc_enhanced_info->ver1                    << ","
                << PTR(p_curr_cc_enhanced_info->CC_Capability_Mask)  << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_CC_ENHANCED_INFO);
    IBDIAG_RETURN_VOID;
}


void IBDiag::DumpCCSwitchGeneralSettingsCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (csv_out.DumpStart(SECTION_CC_SWITCH_GENERAL_SETTINGS))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID," << "aqs_time," << "aqs_weight,"
            << "en," << "cap_total_buffer_size," << "cap_cc_profile_step_size" << endl;
    csv_out.WriteBuf(sstream.str());

    struct CC_CongestionSwitchGeneralSettings *p_curr_cc_switch_general_settings = NULL;
    char buffer[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        p_curr_cc_switch_general_settings = this->fabric_extended_info.
                getCCSwitchGeneralSettings(p_curr_node->createIndex);
        if (!p_curr_cc_switch_general_settings)
            continue;

        sstream.str("");

        snprintf(buffer, sizeof(buffer), U64H_FMT ",%u,%u,%u,%u,%u",
                p_curr_node->guid_get(),
                p_curr_cc_switch_general_settings->aqs_time,
                p_curr_cc_switch_general_settings->aqs_weight,
                p_curr_cc_switch_general_settings->en,
                p_curr_cc_switch_general_settings->cap_total_buffer_size,
                p_curr_cc_switch_general_settings->cap_cc_profile_step_size);

        sstream << buffer << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_CC_SWITCH_GENERAL_SETTINGS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCPortProfileSettingsCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (csv_out.DumpStart(SECTION_CC_PORT_PROFILE_SETTINGS))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID," << "portNum," << "vl," << "granularity," << "mode,"
            << "profile1_min," << "profile1_max," << "profile1_percent,"
            << "profile2_min," << "profile2_max," << "profile2_percent,"
            << "profile3_min," << "profile3_max," << "profile3_percent"
            << endl;

    csv_out.WriteBuf(sstream.str());
    struct CC_CongestionPortProfileSettings *p_curr_cc_port_profile_settings = NULL;
    char buffer[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            SMP_PortInfo *p_curr_port_info = fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            if (!p_curr_port_info)          // port info not exist
                continue;

            u_int8_t vl_num = get_operational_vl_num(p_curr_port_info->OpVLs);
            // go over all OperationalVLs of this port
            for (u_int8_t vl = 0; vl < vl_num; ++vl) {

                p_curr_cc_port_profile_settings = this->fabric_extended_info.
                        getCCPortProfileSettings(p_curr_port->createIndex, vl);
                if (!p_curr_cc_port_profile_settings)
                    continue;

                sstream.str("");

                snprintf(buffer, sizeof(buffer), U64H_FMT ",%u,%u,%u,"
                         "%u,%u,%u,%u,%u,%u,%u,%u,%u,%u",
                         p_curr_node->guid_get(),
                         p_curr_port->num,
                         vl,
                         p_curr_cc_port_profile_settings->granularity,
                         p_curr_cc_port_profile_settings->mode,
                         p_curr_cc_port_profile_settings->profile1_min,
                         p_curr_cc_port_profile_settings->profile1_max,
                         p_curr_cc_port_profile_settings->profile1_percent,
                         p_curr_cc_port_profile_settings->profile2_min,
                         p_curr_cc_port_profile_settings->profile2_max,
                         p_curr_cc_port_profile_settings->profile2_percent,
                         p_curr_cc_port_profile_settings->profile3_min,
                         p_curr_cc_port_profile_settings->profile3_max,
                         p_curr_cc_port_profile_settings->profile3_percent);

                sstream << buffer << endl;
                csv_out.WriteBuf(sstream.str());

            } // VLs
        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_PORT_PROFILE_SETTINGS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCSLMappingSettingsCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (csv_out.DumpStart(SECTION_CC_SL_MAPPING_SETTINGS))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID," << "portNum,"
            << "sl_profile_0," << "sl_profile_1," << "sl_profile_2," << "sl_profile_3,"
            << "sl_profile_4," << "sl_profile_5," << "sl_profile_6," << "sl_profile_7,"
            << "sl_profile_8," << "sl_profile_9," << "sl_profile_10," << "sl_profile_11,"
            << "sl_profile_12," << "sl_profile_13," << "sl_profile_14," << "sl_profile_15"
            << endl;

    csv_out.WriteBuf(sstream.str());
    struct CC_CongestionSLMappingSettings *p_curr_cc_sl_mapping_settings = NULL;
    char buffer[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            p_curr_cc_sl_mapping_settings = this->fabric_extended_info.
                    getCCSLMappingSettings(p_curr_port->createIndex);
            if (!p_curr_cc_sl_mapping_settings)
                continue;

            sstream.str("");

            snprintf(buffer, sizeof(buffer), U64H_FMT ",%u,"
                     "%u,%u,%u,%u,"
                     "%u,%u,%u,%u,"
                     "%u,%u,%u,%u,"
                     "%u,%u,%u,%u",
                     p_curr_node->guid_get(),
                     p_curr_port->num,
                     p_curr_cc_sl_mapping_settings->sl_profile_0,
                     p_curr_cc_sl_mapping_settings->sl_profile_1,
                     p_curr_cc_sl_mapping_settings->sl_profile_2,
                     p_curr_cc_sl_mapping_settings->sl_profile_3,
                     p_curr_cc_sl_mapping_settings->sl_profile_4,
                     p_curr_cc_sl_mapping_settings->sl_profile_5,
                     p_curr_cc_sl_mapping_settings->sl_profile_6,
                     p_curr_cc_sl_mapping_settings->sl_profile_7,
                     p_curr_cc_sl_mapping_settings->sl_profile_8,
                     p_curr_cc_sl_mapping_settings->sl_profile_9,
                     p_curr_cc_sl_mapping_settings->sl_profile_10,
                     p_curr_cc_sl_mapping_settings->sl_profile_11,
                     p_curr_cc_sl_mapping_settings->sl_profile_12,
                     p_curr_cc_sl_mapping_settings->sl_profile_13,
                     p_curr_cc_sl_mapping_settings->sl_profile_14,
                     p_curr_cc_sl_mapping_settings->sl_profile_15);

            sstream << buffer << endl;
            csv_out.WriteBuf(sstream.str());

        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_SL_MAPPING_SETTINGS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCHCAGeneralSettingsCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_CC_HCA_GENERAL_SETTINGS))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," << "portNum," << "en_react," << "en_notify"
            << endl;

    csv_out.WriteBuf(sstream.str());
    struct CC_CongestionHCAGeneralSettings *p_curr_cc_hca_general_settings = NULL;
    char buffer[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
         IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
         if (!p_curr_node || !p_curr_node->getInSubFabric())
             continue;

         for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
             IBPort *p_curr_port = p_curr_node->getPort(pi);
             //Skip ports that don't exist or are down or have undefined state
             //or are not in the scope of the fabric.
             if (!p_curr_port || !p_curr_port->is_data_worthy())
                 continue;

             p_curr_cc_hca_general_settings = this->fabric_extended_info.
                     getCCHCAGeneralSettings(p_curr_port->createIndex);
             if (!p_curr_cc_hca_general_settings)
                 continue;

             sstream.str("");

             snprintf(buffer, sizeof(buffer),
                      U64H_FMT "," U64H_FMT ",%u,%u,%u",
                      p_curr_node->guid_get(),
                      p_curr_port->guid_get(),
                      p_curr_port->num,
                      p_curr_cc_hca_general_settings->en_react,
                      p_curr_cc_hca_general_settings->en_notify);

             sstream << buffer << endl;
             csv_out.WriteBuf(sstream.str());

         } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_GENERAL_SETTINGS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCHCARPParametersCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_CC_HCA_RP_PARAMETERS))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," << "portNum," << "clamp_tgt_rate_after_time_inc,"
            << "clamp_tgt_rate," << "rpg_time_reset," << "rpg_byte_reset," << "rpg_threshold,"
            << "rpg_max_rate," << "rpg_ai_rate," << "rpg_hai_rate," << "rpg_gd," << "rpg_min_dec_fac,"
            << "rpg_min_rate," << "rate_to_set_on_first_cnp," << "dce_tcp_g," << "dce_tcp_rtt,"
            << "rate_reduce_mionitor_period," << "initial_alpha_value"
            << endl;

    csv_out.WriteBuf(sstream.str());
    struct CC_CongestionHCARPParameters *p_curr_cc_hca_rp_parameters = NULL;
    char buffer[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            p_curr_cc_hca_rp_parameters = this->fabric_extended_info.
                    getCCHCARPParameters(p_curr_port->createIndex);
            if (!p_curr_cc_hca_rp_parameters)
                continue;

            sstream.str("");

            snprintf(buffer, sizeof(buffer), U64H_FMT "," U64H_FMT ",%u,"
                     "%u,%u,%u,%u,"
                     "%u,%u,%u,%u,"
                     "%u,%u,%u,%u,"
                     "%u,%u,%u,%u",
                     p_curr_node->guid_get(),
                     p_curr_port->guid_get(),
                     p_curr_port->num,
                     p_curr_cc_hca_rp_parameters->clamp_tgt_rate_after_time_inc,
                     p_curr_cc_hca_rp_parameters->clamp_tgt_rate,
                     p_curr_cc_hca_rp_parameters->rpg_time_reset,
                     p_curr_cc_hca_rp_parameters->rpg_byte_reset,
                     p_curr_cc_hca_rp_parameters->rpg_threshold,
                     p_curr_cc_hca_rp_parameters->rpg_max_rate,
                     p_curr_cc_hca_rp_parameters->rpg_ai_rate,
                     p_curr_cc_hca_rp_parameters->rpg_hai_rate,
                     p_curr_cc_hca_rp_parameters->rpg_gd,
                     p_curr_cc_hca_rp_parameters->rpg_min_dec_fac,
                     p_curr_cc_hca_rp_parameters->rpg_min_rate,
                     p_curr_cc_hca_rp_parameters->rate_to_set_on_first_cnp,
                     p_curr_cc_hca_rp_parameters->dce_tcp_g,
                     p_curr_cc_hca_rp_parameters->dce_tcp_rtt,
                     p_curr_cc_hca_rp_parameters->rate_reduce_mionitor_period,
                     p_curr_cc_hca_rp_parameters->initial_alpha_value);

            sstream << buffer << endl;
            csv_out.WriteBuf(sstream.str());

        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_RP_PARAMETERS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCHCANPParametersCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_CC_HCA_NP_PARAMETERS))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," << "portNum," << "min_time_between_cnps,"
            << "cnp_sl," <<"cnp_sl_mode"
            << endl;

    csv_out.WriteBuf(sstream.str());
    struct CC_CongestionHCANPParameters *p_curr_cc_hca_np_parameters = NULL;
    char buffer[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            p_curr_cc_hca_np_parameters = this->fabric_extended_info.
                    getCCHCANPParameters(p_curr_port->createIndex);
            if (!p_curr_cc_hca_np_parameters)
                continue;

            sstream.str("");

            snprintf(buffer, sizeof(buffer), U64H_FMT "," U64H_FMT ",%u,%u,%u,%u",
                    p_curr_node->guid_get(),
                    p_curr_port->guid_get(),
                    p_curr_port->num,
                    p_curr_cc_hca_np_parameters->min_time_between_cnps,
                    p_curr_cc_hca_np_parameters->cnp_sl,
                    p_curr_cc_hca_np_parameters->cnp_sl_mode);

            sstream << buffer << endl;
            csv_out.WriteBuf(sstream.str());

        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_NP_PARAMETERS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCHCAStatisticsQuery(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (csv_out.DumpStart(SECTION_CC_HCA_STATISTICS_QUERY))
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," << "portNum," << "clear," << "cnp_ignored,"
            << "cnp_handled," << "marked_packets," << "cnp_sent," << "timestamp,"
            << "accumulators_period"
            << endl;

    csv_out.WriteBuf(sstream.str());
    struct CC_CongestionHCAStatisticsQuery *p_curr_cc_statistics_query = NULL;

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            CC_EnhancedCongestionInfo *p_cc_enhanced_info =
                    fabric_extended_info.getCCEnhancedCongestionInfo(p_curr_node->createIndex);
            if (!p_cc_enhanced_info) // not supported
                continue;

            p_curr_cc_statistics_query = this->fabric_extended_info.
                    getCCHCAStatisticsQuery(p_curr_port->createIndex);
            if (!p_curr_cc_statistics_query)
                continue;

            sstream.str("");

            sstream << PTR(p_curr_node->guid_get())          << ","
                    << PTR(p_curr_port->guid_get())          << ","
                    << +p_curr_port->num                     << ","
                    << +p_curr_cc_statistics_query->clear    << ",";

            if (p_cc_enhanced_info->ver0)
                sstream << p_curr_cc_statistics_query->cnp_ignored    << ","
                        << p_curr_cc_statistics_query->cnp_handled    << ",";
            else
                sstream << "NA,NA,";

            sstream << p_curr_cc_statistics_query->marked_packets       << ","
                    << p_curr_cc_statistics_query->cnp_sent             << ","
                    << p_curr_cc_statistics_query->timestamp            << ","
                    << p_curr_cc_statistics_query->accumulators_period
                    << endl;

            csv_out.WriteBuf(sstream.str());
        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_STATISTICS_QUERY);
    IBDIAG_RETURN_VOID;
}

bool IBDiag::IsSupportedCCCapability(u_int64_t cc_capability_mask, EnCCCapabilityMaskBit bit) {
        return (cc_capability_mask & (1 << bit));
}

int IBDiag::Build_CC_SLVLCntrs(list_p_fabric_general_err &cc_per_slvl_errors,
                               bool is_reset,
                               CountersPerSLVL* cntrs_per_slvl)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &cc_per_slvl_errors,
                    NULL, &capability_module);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = &forwardClbck<IBDiagClbck, &IBDiagClbck::CCPerSLVLGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct PM_PortRcvXmitCntrsSlVl pm_port_rcvxmit_data_slvl;
    CLEAR_STRUCT(pm_port_rcvxmit_data_slvl);

    for (IBNode* p_curr_node : this->discovered_fabric.Switches) {

        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        IBPort* p_curr_port0 = p_curr_node->getPort(0); // get switch port 0
        if (!p_curr_port0)
            continue;

        // todo: check if there is a different cap-bit for IB spec CC MADs
        if (!this->capability_module.IsSupportedSMPCapability(p_curr_node, EnSMPCapIsVSCCSupported))
            continue;

        // go over all ports of this switch:
        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            // skip ANs
            if (p_curr_port->isSpecialPort())
                continue;

            // Send the MAD
            clbck_data.m_data1 = p_curr_port;
            clbck_data.m_data2 = cntrs_per_slvl;
            progress_bar.push(p_curr_port);

            if (!cntrs_per_slvl->IsVSClass()) {
                this->ibis_obj.PMPerSLVLCounters(is_reset,
                                                 p_curr_port0->base_lid,
                                                 pi,
                                                 cntrs_per_slvl->GetAttrId(),
                                                 &pm_port_rcvxmit_data_slvl,
                                                 &clbck_data);
            } else {
                this->ibis_obj.VSPerVLCounters(is_reset,
                                               p_curr_port->base_lid,
                                               pi,
                                               cntrs_per_slvl->GetAttrId(),
                                               &pm_port_rcvxmit_data_slvl,
                                               &clbck_data);
            }
        } // ports

        if (ibDiagClbck.GetState())
           goto exit;

    } // nodes

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!cc_per_slvl_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}
