/*
 * Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag_types.h"
#include "infiniband/ibdm/Fabric.h"
#include <infiniband/ibdiag/ibdiag.h>

// reasons for bad links according to enum IbdiagBadDirectRouteReason_t in ibdiag_types.h
const char* bad_direct_route_reasons [IBDIAG_BAD_DR_MAX] = {
    "None",
    "Node info get failed",
    "Node duplicate GUID check failed",
    "Node description get failed",
    "Node inconsistency data error",
    "Bad Node Info data",
    "Port info get failed",
    "Port invalid lid",
    "Port inconsistency data error"
};


/****************************************************/
/******************private methods*******************/
/****************************************************/


/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckDuplicatedNodeDescription(list_p_fabric_general_err& nodes_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    //check for duplication
    for (map_str_list_pnode::iterator lI = this->discovered_fabric.NodeByDesc.begin();
            lI != this->discovered_fabric.NodeByDesc.end(); ++lI) {

        // skip Aggregation Nodes
        if (GetSpecialCAPortType(lI->second.front()) == IB_SPECIAL_PORT_AN)
            continue;

        //found duplicated node description
        if ((*lI).second.size() > 1) {
            for (list_p_node::iterator nI = (*lI).second.begin(); nI != (*lI).second.end(); ++nI) {
                nodes_errors.push_back(new FabricErrNodeDuplicatedNodeDesc((*nI)));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        }
    }
    //clean map - do not need it anymore  - not free pointers to nodes (them we need)
    this->discovered_fabric.NodeByDesc.clear();
    IBDIAG_RETURN(rc);
}


//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckLinks(list_p_fabric_general_err& links_errors, IBLinksInfo *p_ib_links_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    bool check_faild = false;
    rc = CheckLegacyLinks(links_errors, p_ib_links_info);
    if (rc && rc != IBDIAG_ERR_CODE_CHECK_FAILED)
        return rc;
    else if (rc == IBDIAG_ERR_CODE_CHECK_FAILED)
        check_faild = true;

    rc = CheckAPortLinks(links_errors, p_ib_links_info);
    if (rc && rc != IBDIAG_ERR_CODE_CHECK_FAILED)
        return rc;
    else if (rc == IBDIAG_ERR_CODE_CHECK_FAILED)
        check_faild = true;

    return check_faild ? IBDIAG_ERR_CODE_CHECK_FAILED : IBDIAG_SUCCESS_CODE;
}

//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckLegacyLinks(list_p_fabric_general_err& links_errors, IBLinksInfo *p_ib_links_info)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    string bad_dr_msg, port_name;
    IBNode *p_node;
    phys_port_t port_num;
    IBPort *port;

    //check for bad links
    for (list_p_bad_direct_route::iterator it = this->bad_direct_routes.begin();
            it != this->bad_direct_routes.end(); ++it) {
        bad_dr_msg.clear();
        bad_dr_msg = bad_direct_route_reasons[(*it)->reason] + std::string(" ") + (*it)->message;

        //if it's related to port discovery errors, add port number
        if ((*it)->reason >= IBDIAG_BAD_DR_PORT_FIRST && (*it)->reason <= IBDIAG_BAD_DR_PORT_LAST) {
            std::ostringstream oss;
            oss << (*it)->port_num;
            bad_dr_msg += " (for port " + oss.str() + ")";
        }

        if ((*it)->direct_route->length > 1) {
            bad_dr_msg += " on node connected to ";
            port_name = "NA";
            //let's try to find a previous good node
            direct_route_t new_direct_route;
            CLEAR_STRUCT(new_direct_route);
            new_direct_route =  *((*it)->direct_route);
            new_direct_route.length = (u_int8_t)((*it)->direct_route->length - 1);
            IBNode *p_prev_node = this->GetNodeByDirectRoute(&new_direct_route);
            // previous node exists
            if (p_prev_node) {
                // connected on this port to the problematic node
                port_num = (*it)->direct_route->path.BYTE[(*it)->direct_route->length - 1];
                port = p_prev_node->getPort(port_num);

                if (port)
                    port_name = port->getName();
            }
        } else if ((*it)->direct_route->length == 1) {
            //it failed on root,
            //probably won't get here as we would fail discovery
            //and not continue to other stages
            p_node = this->root_node;
            bad_dr_msg += " on root node";
            //add port name if we can
            if (p_node) {
                port_num = (*it)->direct_route->path.BYTE[(*it)->direct_route->length - 1];
                port = p_node->getPort(port_num);
                if (port)
                    port_name = port->getName();
            }
        }
        bad_dr_msg += port_name;
        bad_dr_msg += " (DR:" + Ibis::ConvertDirPathToStr((*it)->direct_route) + ")";

        links_errors.push_back(new FabricErrDR(bad_dr_msg));
        rc = IBDIAG_ERR_CODE_CHECK_FAILED;
    }

    //check for not active links
    //mark all ports as not checked
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)                   //port not exists
            continue;
        p_curr_port->counter1 = 0;
    }

    //go over all ports
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric() || p_curr_port->p_aport)
            continue;
        IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (!p_remote_port)         //port not connected
            continue;

        // check if both the current port and the remote port are FNM ports
        bool is_internal_fnm_link = false;
        if (p_curr_port->isFNMPort() && p_remote_port->isFNMPort())
            is_internal_fnm_link = true;

        if ((p_curr_port->counter1) || (p_curr_port->p_remotePort->counter1))   //already checked
            continue;
        //mark as checked
        ++p_curr_port->counter1;
        ++p_curr_port->p_remotePort->counter1;

        // Count IB links
        if (p_ib_links_info && !p_curr_port->p_aport && !is_internal_fnm_link) {
            p_ib_links_info->FillIBLinkWidthSpeedIndex(p_curr_port);
        }

        //check link logical state
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "(port=%s logical_state=%u)<-->(port=%s logical_state=%u)\n",
                p_curr_port->getName().c_str(), p_curr_port->get_internal_state(),
                p_remote_port->getName().c_str(), p_remote_port->get_internal_state());
        if ( (p_curr_port->get_internal_state() == IB_PORT_STATE_ACTIVE) &&
                (p_remote_port->get_internal_state() == IB_PORT_STATE_ACTIVE) )           //link is active
            continue;

        if (p_curr_port->get_internal_state() != p_remote_port->get_internal_state())
            links_errors.push_back(new FabricErrLinkLogicalStateWrong(p_curr_port, p_remote_port));
        else if (p_curr_port->get_internal_state() != IB_PORT_STATE_ACTIVE)
            links_errors.push_back(new FabricErrLinkLogicalStateNotActive(p_curr_port, p_remote_port));

        rc = IBDIAG_ERR_CODE_CHECK_FAILED;
    }
    IBDIAG_RETURN(rc);
}

int IBDiag::CheckAPortLinks(list_p_fabric_general_err& links_errors, IBLinksInfo *p_ib_links_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    this->discovered_fabric.unvisitAllAPorts();

    for (const auto& entry : this->discovered_fabric.APortsBySysGuid) {
        if (entry.second.empty())
            continue;

        for (size_t i = 0; i < entry.second.size(); i++) {
            APort *p_curr_aport = entry.second[i]; 
            bool is_remote_errors = false;
            if (!p_curr_aport || p_curr_aport->visited)
                continue;
            if (!p_curr_aport->any_plane_connected())
                continue;
            p_curr_aport->visited = true;
            // check if there are any errors on the remote APorts, and mark all as visited
            for (size_t i = 1; i < p_curr_aport->ports.size(); ++i) {
                if (p_curr_aport->ports[i] && p_curr_aport->ports[i]->p_remotePort
                        && p_curr_aport->ports[i]->p_remotePort->p_aport) {
                    APort *p_curr_remote_aport = p_curr_aport->ports[i]->p_remotePort->p_aport;
                    is_remote_errors |= !this->errors_by_aport[p_curr_aport].empty();
                    p_curr_remote_aport->visited = true;
                }
            }

            APort *p_remote_aport = p_curr_aport->get_remote_aport();
            // if the current APort or any remote APort has errors, report asymmetrical
            if (!is_remote_errors && this->errors_by_aport[p_curr_aport].empty() && p_remote_aport) {
                p_ib_links_info->FillIBAPortLinkWidthSpeedIndex(p_curr_aport);

                // perform checks on the link
                //check link logical state
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "(port=%s logical_state=%u)<-->(port=%s logical_state=%u)\n",
                    p_curr_aport->getName().c_str(), p_curr_aport->get_internal_state(),
                    p_remote_aport->getName().c_str(), p_remote_aport->get_internal_state());
                if ( (p_curr_aport->get_internal_state() == IB_PORT_STATE_ACTIVE) &&
                        (p_remote_aport->get_internal_state() == IB_PORT_STATE_ACTIVE) )           //link is active
                    continue;

                if (p_curr_aport->get_internal_state() != p_remote_aport->get_internal_state())
                    links_errors.push_back(new FabricErrAPortLinkLogicalStateWrong(p_curr_aport, p_remote_aport));
                else if (p_curr_aport->get_internal_state() != IB_PORT_STATE_ACTIVE)
                    links_errors.push_back(new FabricErrAPortLinkLogicalStateNotActive(p_curr_aport, p_remote_aport));
            } else {
                p_ib_links_info->FillAsymmetricalLinks(p_curr_aport);
            }
        }
    }

    IBDIAG_RETURN(rc);
}

//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckLids(list_p_fabric_general_err& lids_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    map_lid_list_p_port known_lids;

    //go over all nodes
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        u_int8_t start_port_idx;
        u_int8_t end_port_idx;
        if (p_curr_node->type == IB_SW_NODE) {
            start_port_idx = 0;
            end_port_idx = 0;   //only query manage port
        } else {   //CA
            start_port_idx = 1;
            end_port_idx = p_curr_node->numPorts;
        }

        // go over all ports of this node, if SW than query only manage port
        // use unsigned int index to prevent overwrap
        for (unsigned int port_num = start_port_idx; port_num <= end_port_idx; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)port_num);
            if (!p_curr_port)                                       //port not exist
                continue;
            //Skip ports that are down or have undefined state or
            //are not in the scope of the fabric.
            if ((port_num != 0) && !p_curr_port->is_data_worthy())
                continue;
            known_lids[p_curr_port->base_lid].push_back(p_curr_port);
        }
    }

    //check for duplicated / zero lids
    for (map_lid_list_p_port::iterator it = known_lids.begin();
            it != known_lids.end(); ++it) {
        if (it->first == 0) {   //check for zero lid
            // Report errors per APorts, or IBPorts without APorts
            list_p_port legacy_ports;
            list_p_aport aports;
            APort::splitIBPortListToAPorts(it->second, aports, legacy_ports);

            for (IBPort* p_port : legacy_ports)
                lids_errors.push_back(new FabricErrPortZeroLid(p_port));

            for (APort* p_aport : aports)
                lids_errors.push_back(new FabricErrAPortZeroLid(p_aport));

            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        } else {                //check for lid duplication
            if (it->second.size() > 1 && !APort::isSameAPort(it->second)) {
                // Report errors per APorts, or IBPorts without APorts
                list_p_port legacy_ports;
                list_p_aport aports;
                APort::splitIBPortListToAPorts(it->second, aports, legacy_ports);

                for (IBPort* p_port : legacy_ports)
                    lids_errors.push_back(new FabricErrPortDuplicatedLid(p_port));

                for (APort* p_aport : aports)
                    lids_errors.push_back(new FabricErrAPortDuplicatedLid(p_aport, it->first));

                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        }
    }

    // check that all APorts have the same LID on all planes
    for (const auto& entry : this->discovered_fabric.APortsBySysGuid) {
        if (entry.second.empty())
            continue;

        for (size_t i = 0; i < entry.second.size(); i++) {
            APort *p_curr_aport = entry.second[i]; 
            if (!p_curr_aport || p_curr_aport->ports.size() < 1)
                continue;

            // make sure only HCAs are checked
            if (!p_curr_aport->ports[1] || !p_curr_aport->ports[1]->p_node || 
                    p_curr_aport->ports[1]->p_node->type != IB_CA_NODE)
                continue;

            ibport_accessor_t<uint16_t> lid_accessor = [](const IBPort *p_port) {
                if (!p_port)
                    return uint16_t(0);
                return p_port->base_lid;
            };

            if (!p_curr_aport->isEqualAttribute(lid_accessor)) {
                lids_errors.push_back(new FabricErrAPortUnequalLID(p_curr_aport));
            }
        }
    }

    known_lids.clear();
    IBDIAG_RETURN(rc);
}


//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED / ERR_CODE_DB_ERR
int IBDiag::CheckLinkWidth(list_p_fabric_general_err& width_errors, string expected_link_width_str)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    IBLinkWidth expected_link_width = IB_UNKNOWN_LINK_WIDTH;
    // this function reports plane errors on the APort level. to make sure we don't report
    // more then one error on each APort, we skip plane ports if they are in an
    // APort that already has errors.
    if (expected_link_width_str != "") {
        expected_link_width = char2width(expected_link_width_str.c_str());
        if (expected_link_width == IB_UNKNOWN_LINK_WIDTH) {
            this->SetLastError("Link width string is unknown");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
        }
    }

    //mark all ports as not checked
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;
        p_curr_port->counter1 = 0;
    }
    this->discovered_fabric.unvisitAllAPorts();

    //check all ports
    char buffer[256];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())           //port not exists
            continue;
        IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (!p_remote_port)         //port not connected
            continue;
        APort *p_curr_aport = p_curr_port->p_aport;
        APort *p_remote_aport = p_remote_port->p_aport;
        if ((p_curr_aport && p_curr_aport->visited) ||
                (p_remote_aport && p_remote_aport->visited))
            continue;
        SMP_PortInfo *p_curr_port_info = this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
        if (!p_curr_port_info) {
            stringstream ss;
            ss << "DB error - found connected port="
               << (p_curr_aport ? p_curr_aport->getName() : p_curr_port->getName())
               << " without SMPPortInfo";

            if (p_curr_aport) {
                width_errors.push_back(
                    new FabricErrAPortInfoFail(p_curr_aport, ss.str().c_str()));
                p_curr_aport->visited = true;
            } else {
                width_errors.push_back(
                    new FabricErrPortInfoFail(p_curr_port->p_node, p_curr_port->num, ss.str().c_str()));
            }

            continue;
        }
        SMP_PortInfo *p_remote_port_info = this->fabric_extended_info.getSMPPortInfo(p_remote_port->createIndex);
        if (!p_remote_port_info) {
            stringstream ss;
            ss << "DB error - found connected port="
               << (p_remote_aport ? p_remote_aport->getName() : p_remote_port->getName())
               << " without SMPPortInfo";

            if (p_remote_aport) {
                width_errors.push_back(
                    new FabricErrAPortInfoFail(p_remote_aport, ss.str().c_str()));
                p_remote_aport->visited = true;
            } else {
                width_errors.push_back(
                    new FabricErrPortInfoFail(p_remote_port->p_node, p_remote_port->num, ss.str().c_str()));
            }

            continue;
        }

        if ((p_curr_port->counter1) || (p_remote_port->counter1))   //already checked
            continue;
        //mark as checked
        ++p_curr_port->counter1;
        ++p_remote_port->counter1;


        /*
         * Step 1 - check same width on both ports
         */
        if (p_curr_port->get_internal_width() != p_remote_port->get_internal_width()) {
            if (p_curr_aport && p_remote_aport) {
                width_errors.push_back(
                    new FabricErrAPortLinkDifferentWidth(p_curr_aport, p_remote_aport));
                p_curr_aport->visited = true;
                p_remote_aport->visited = true;
            } else {
                width_errors.push_back(
                    new FabricErrLinkDifferentWidth(p_curr_port, p_remote_port));
            }
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            continue;
        }


        /*
         * Step 2 - check en <= sup
         */
        bool err_occured = false;

        // Step 2.1 - check port1
        if (!IsValidEnableWidth(p_curr_port_info->LinkWidthEn,
                p_curr_port_info->LinkWidthSup)) {
            snprintf(buffer, sizeof(buffer),
                     "enable_width=\"%s\" but supported_width=\"%s\"",
                     supwidth2char(p_curr_port_info->LinkWidthEn),
                     supwidth2char(p_curr_port_info->LinkWidthSup));

            if (p_curr_aport) {
                width_errors.push_back(
                    new FabricErrAPortWrongConfig(p_curr_aport, buffer));
                p_curr_aport->visited = true;
            } else {
                width_errors.push_back(
                    new FabricErrPortWrongConfig(p_curr_port, buffer));
            }

            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            //continue;     not continue because need to check other port as well
            err_occured = true;
        }

        // Step 2.2 - check port2
        if (!IsValidEnableWidth(p_remote_port_info->LinkWidthEn,
                p_remote_port_info->LinkWidthSup)) {
            snprintf(buffer, sizeof(buffer),
                     "enable_width=\"%s\" but supported_width=\"%s\"",
                     supwidth2char(p_remote_port_info->LinkWidthEn),
                     supwidth2char(p_remote_port_info->LinkWidthSup));

            if (p_remote_aport) {
                width_errors.push_back(
                    new FabricErrAPortWrongConfig(p_remote_aport, buffer));
                p_remote_aport->visited = true;
            } else {
                width_errors.push_back(
                    new FabricErrPortWrongConfig(p_remote_port, buffer));
            }
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            //continue;     not continue because need to check other port as well
            err_occured = true;
        }
        if (err_occured)
            continue;   //not need next check in this case

        /*
         * Step 3 - check expected value
         */

        /* Step 3.1 - calculate final width if no specific value
           for checking is given from user */
        if (expected_link_width_str == "") {
            //get final speed
            expected_link_width = CalcFinalWidth(p_curr_port_info->LinkWidthEn,
                    p_remote_port_info->LinkWidthEn);
            if (expected_link_width == IB_UNKNOWN_LINK_WIDTH) {
                snprintf(buffer, sizeof(buffer),
                         "(enable_width1=\"%s\", enable_width2=\"%s\" therefore final width is unknown)",
                         supwidth2char(p_curr_port_info->LinkWidthEn),
                         supwidth2char(p_remote_port_info->LinkWidthEn));

                if (p_curr_aport && p_remote_aport) {
                    width_errors.push_back(
                        new FabricErrAPortLinkAutonegError(p_curr_aport, p_remote_aport, buffer));
                    p_curr_aport->visited = true;
                    p_remote_aport->visited = true;
                } else {
                    width_errors.push_back(
                        new FabricErrLinkAutonegError(p_curr_port, p_remote_port, buffer));
                }
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                continue;
            }
        }

        // Step 3.2 - check actual link width
        if (p_curr_port->get_internal_width() != expected_link_width) {
            if (expected_link_width_str == "") {
                snprintf(buffer, sizeof(buffer),
                         "(enable_width1=\"%s\", enable_width2=\"%s\" therefore final width should be %s)",
                         supwidth2char(p_curr_port_info->LinkWidthEn),
                         supwidth2char(p_remote_port_info->LinkWidthEn),
                         width2char(expected_link_width));

                p_curr_port->expected_width = expected_link_width;
                p_remote_port->expected_width = expected_link_width;

            } else
                memset(buffer, 0, sizeof(buffer));

            if (p_curr_aport && p_remote_aport) {
                width_errors.push_back(
                        new FabricErrAPortLinkUnexpectedWidth(p_curr_aport, p_remote_aport, buffer));
                p_curr_aport->visited = true;
                p_remote_aport->visited = true;
            } else {
                width_errors.push_back(
                        new FabricErrLinkUnexpectedWidth(p_curr_port, p_remote_port, buffer));
            }

            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }
    IBDIAG_RETURN(rc);
}


//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED / ERR_CODE_DB_ERR
int IBDiag::CheckLinkSpeed(list_p_fabric_general_err& speed_errors, string expected_link_speed_str)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    IBLinkSpeed expected_link_speed = IB_UNKNOWN_LINK_SPEED;
    if (expected_link_speed_str != "") {
        expected_link_speed = char2speed(expected_link_speed_str.c_str());
        if (expected_link_speed == IB_UNKNOWN_LINK_SPEED) {
            this->SetLastError("Link speed string is unknown");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
        }
    }
    // report plane errors on the APort level, do not report APort errors more then once.
    // a plane port will be checked only if no other plane port produced an error
    this->discovered_fabric.unvisitAllAPorts();

    //mark all ports as not checked
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)                   //port not exists
            continue;

        p_curr_port->counter1 = 0;
    }

    //check all ports
    char buffer[256];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;
        IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (!p_remote_port)         //port not connected
            continue;
        APort *p_curr_aport = p_curr_port->p_aport;
        APort *p_remote_aport = p_remote_port->p_aport;
        if ((p_curr_aport && p_curr_aport->visited) ||
                (p_remote_aport && p_remote_aport->visited))
            continue;

        SMP_PortInfo *p_curr_port_info = this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
        if (!p_curr_port_info) {
            stringstream ss;
            ss << "DB error - found connected port="
               << (p_curr_aport ? p_curr_aport->getName() : p_curr_port->getName())
               << " without SMPPortInfo";

            if (p_curr_aport) {
                speed_errors.push_back(
                    new FabricErrAPortInfoFail(p_curr_aport, ss.str().c_str()));
                p_curr_aport->visited = true;
            } else {
                speed_errors.push_back(
                    new FabricErrPortInfoFail(p_curr_port->p_node, p_curr_port->num, ss.str().c_str()));
            }

            continue;
        }
        SMP_PortInfo *p_remote_port_info = this->fabric_extended_info.getSMPPortInfo(p_remote_port->createIndex);
        if (!p_remote_port_info) {
            stringstream ss;
            ss << "DB error - found connected port="
               << (p_remote_aport ? p_remote_aport->getName() : p_remote_port->getName())
               << " without SMPPortInfo";

            if (p_remote_aport) {
                speed_errors.push_back(
                    new FabricErrAPortInfoFail(p_remote_aport, ss.str().c_str()));
                p_remote_aport->visited = true;
            } else {
                speed_errors.push_back(
                    new FabricErrPortInfoFail(p_remote_port->p_node, p_remote_port->num, ss.str().c_str()));
            }

            continue;
        }
        SMP_MlnxExtPortInfo *p_curr_mlnx_ext_port_info =
                this->fabric_extended_info.getSMPMlnxExtPortInfo(p_curr_port->createIndex);
        SMP_MlnxExtPortInfo *p_remote_mlnx_ext_port_info =
                this->fabric_extended_info.getSMPMlnxExtPortInfo(p_remote_port->createIndex);

        if ((p_curr_port->counter1) || (p_remote_port->counter1))   //already checked
            continue;
        //mark as checked
        ++p_curr_port->counter1;
        ++p_remote_port->counter1;

        /*
         * Step 1 - check same speed on both ports
         */
        if (p_curr_port->get_internal_speed() != p_remote_port->get_internal_speed()) {
            if (p_curr_aport && p_remote_aport) {
                speed_errors.push_back(new FabricErrAPortLinkDifferentSpeed(
                            p_curr_aport, p_remote_aport));
                p_curr_aport->visited = true;
                p_remote_aport->visited = true;
            } else {
                speed_errors.push_back(new FabricErrLinkDifferentSpeed(
                            p_curr_port, p_remote_port));
            }
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            continue;
        }

        /*
         * Step 2 - find cap mask of both ports
         */
        u_int32_t cap_mask_curr_port            = 0;
        u_int16_t cap_mask2_curr_port           = 0;
        u_int32_t cap_mask_remote_port          = 0;
        u_int16_t cap_mask2_remote_port         = 0;

        // Step 2.1 - cap mask of current port
        struct SMP_PortInfo *p_tmp_port_info    = NULL;
        IBPort *p_tmp_port                      = NULL;
        if (p_curr_port->p_node->type == IB_SW_NODE) {
            p_tmp_port = p_curr_port->p_node->getPort(0);
            if (!p_tmp_port) {
                this->SetLastError("DB error - can not found manage port for switch=%s\n",
                        p_curr_port->p_node->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            p_tmp_port_info = this->fabric_extended_info.getSMPPortInfo(
                                    p_tmp_port->createIndex);
        } else
            p_tmp_port_info = p_curr_port_info;

        if (!p_tmp_port_info) {
            this->SetLastError("DB error - can not found port info for port =%s\n",
                    (p_curr_port->p_node->type == IB_SW_NODE) ? \
                            p_tmp_port->getName().c_str() : \
                            p_curr_port->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        cap_mask_curr_port = p_tmp_port_info->CapMsk;
        cap_mask2_curr_port = p_tmp_port_info->CapMsk2;

        // Step 2.2 - cap mask of remote port
        p_tmp_port_info = NULL;
        p_tmp_port      = NULL;
        if (p_remote_port->p_node->type == IB_SW_NODE) {
            p_tmp_port = p_remote_port->p_node->getPort(0);
            if (!p_tmp_port) {
                this->SetLastError("DB error - can not found manage port for switch=%s\n",
                        p_remote_port->p_node->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            p_tmp_port_info = this->fabric_extended_info.getSMPPortInfo(
                              p_tmp_port->createIndex);
        } else
            p_tmp_port_info = p_remote_port_info;

        if (!p_tmp_port_info) {
            this->SetLastError("DB error - can not found port info for port=%s\n",
                    (p_remote_port->p_node->type == IB_SW_NODE) ?   \
                            p_tmp_port->getName().c_str() : \
                            p_remote_port->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        cap_mask_remote_port = p_tmp_port_info->CapMsk;
        cap_mask2_remote_port = p_tmp_port_info->CapMsk2;

        /*
         * Step 3 - check en <= sup
         */
        bool err_occured = false;
        u_int32_t curr_port_speed_sup   = 0;
        u_int32_t curr_port_speed_en    = 0;

        //3.1 check current port
        //get total speeds: regular in first byte, extended in second,
        //mellanox in third
        GetTotalSpeeds(p_curr_port_info, p_curr_mlnx_ext_port_info,
                       cap_mask_curr_port, cap_mask2_curr_port,
                       &curr_port_speed_sup, &curr_port_speed_en, NULL);

        if (!IsValidEnableSpeed(curr_port_speed_en, curr_port_speed_sup)) {
            snprintf(buffer, sizeof(buffer),
                     "enable_speed=\"%s\" but supported_speed=\"%s\"",
                     supspeed2char(curr_port_speed_en).c_str(),
                     supspeed2char(curr_port_speed_sup).c_str());

            if (p_curr_aport) {
                speed_errors.push_back(new FabricErrAPortWrongConfig(p_curr_aport, buffer));
                p_curr_aport->visited = true;
            } else {
                speed_errors.push_back(new FabricErrPortWrongConfig(p_curr_port, buffer));
            }
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            //continue;     not continue because need to check other port as well
            err_occured = true;
        }

        //3.2 check the remote port
        u_int32_t remote_port_speed_sup = 0;
        u_int32_t remote_port_speed_en  = 0;

        GetTotalSpeeds(p_remote_port_info, p_remote_mlnx_ext_port_info,
                       cap_mask_remote_port, cap_mask2_remote_port,
                       &remote_port_speed_sup, &remote_port_speed_en, NULL);

        if (!IsValidEnableSpeed(remote_port_speed_en, remote_port_speed_sup)) {
            snprintf(buffer, sizeof(buffer),
                     "enable_speed=\"%s\" but supported_speed=\"%s\"",
                     supspeed2char(remote_port_speed_en).c_str(),
                     supspeed2char(remote_port_speed_sup).c_str());

            if (p_remote_aport) {
                speed_errors.push_back(new FabricErrAPortWrongConfig(p_remote_aport, buffer));
                p_remote_aport->visited = true;
            } else {
                speed_errors.push_back(new FabricErrPortWrongConfig(p_remote_port, buffer));
            }
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            err_occured = true;
        }

        if (err_occured)
            continue;   //not need next check in this case

        /*
         * Step 4 - check expected value
         * we already have enabled speeds for current port and its remote port
         * Step 4.1 Calculate final speed if no specific value
         * for checking is given from user
         */
        if (expected_link_speed_str == "") {
            expected_link_speed = CalcFinalSpeed(curr_port_speed_en,
                                                 remote_port_speed_en);
            if (expected_link_speed == IB_UNKNOWN_LINK_SPEED) {
                snprintf(buffer, sizeof(buffer),
                         "(enable_speed1=\"%s\", enable_speed2=\"%s\" "
                         "therefore final speed is unknown)",
                         supspeed2char(curr_port_speed_en).c_str(),
                         supspeed2char(remote_port_speed_en).c_str());

                if (p_curr_aport && p_remote_aport) {
                    speed_errors.push_back(new FabricErrAPortLinkAutonegError(
                        p_curr_aport, p_remote_aport, buffer));
                    p_curr_aport->visited = true;
                    p_remote_aport->visited = true;
                } else {
                    speed_errors.push_back(new FabricErrLinkAutonegError(p_curr_port, p_remote_port, buffer));
                }
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                continue;
            }
        }

        // Step 4.2 - check actual link speed
        if (p_curr_port->get_internal_speed() != expected_link_speed) {
            if (expected_link_speed_str == "") {
                snprintf(buffer, sizeof(buffer),
                         "(enable_speed1=\"%s\", enable_speed2=\"%s\" therefore final speed should be %s)",
                         supspeed2char(curr_port_speed_en).c_str(),
                         supspeed2char(remote_port_speed_en).c_str(),
                         speed2char(expected_link_speed));

                p_curr_port->expected_speed = expected_link_speed;
                p_remote_port->expected_speed = expected_link_speed;

            } else
                memset(buffer, 0, sizeof(buffer));

            if (p_curr_aport && p_remote_aport) {
                speed_errors.push_back(new FabricErrAPortLinkUnexpectedSpeed(
                    p_curr_aport, p_remote_aport, buffer));
                p_curr_aport->visited = true;
                p_remote_aport->visited = true;
            } else {
                speed_errors.push_back(new FabricErrLinkUnexpectedSpeed(p_curr_port, p_remote_port, buffer));
            }
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }
    IBDIAG_RETURN(rc);
}


void CheckPortHierarchyInfoByTemplateGuid0x03(IBNode *p_node,
                                              IBPort *p_port,
                                              bool is_splitted,
                                              vector<string> &missing_fields,
                                              vector<string> &extra_fields) {
    IBDIAG_ENTER;

    if (p_node->type == IB_SW_NODE) {
        // Check mandatory fields: ASIC, Cage, Port, Split
        if (p_port->p_port_hierarchy_info->m_asic == -1)
            missing_fields.push_back("ASIC");
        if (p_port->p_port_hierarchy_info->m_cage == -1)
            missing_fields.push_back("Cage");
        if (p_port->p_port_hierarchy_info->m_port == -1)
            missing_fields.push_back("Port");
        if (is_splitted && p_port->p_port_hierarchy_info->m_split == -1)
            missing_fields.push_back("Split");

        // Check extra fields: BDF, Type, Slot
        if (p_port->p_port_hierarchy_info->bdf() != -1)
            extra_fields.push_back("BDF");
        if (p_port->p_port_hierarchy_info->m_type != -1)
            extra_fields.push_back("Type");
        if (p_port->p_port_hierarchy_info->m_slot_value != -1)
            extra_fields.push_back("Slot");
    } else if (p_node->type == IB_CA_NODE) {
        // Check mandatory fields: Slot or BDF
        if (p_port->p_port_hierarchy_info->m_slot_type == IB_SLOT_TYPE_PHYS) {
            if (p_port->p_port_hierarchy_info->bdf() != -1)
                extra_fields.push_back("BDF");
        } else {
            if (p_port->p_port_hierarchy_info->bdf() == -1)
                missing_fields.push_back("BDF");
            if (p_port->p_port_hierarchy_info->m_slot_value != -1)
                extra_fields.push_back("Slot (not physical)");
        }

        // Check mandatory fields: Cage, Port, Split
        if (p_port->p_port_hierarchy_info->m_cage == -1)
            missing_fields.push_back("Cage");
        if (p_port->p_port_hierarchy_info->m_port == -1)
            missing_fields.push_back("Port");
        if (is_splitted && p_port->p_port_hierarchy_info->m_split == -1)
            missing_fields.push_back("Split");

        // Check extra fields: ASIC, Type
        if (p_port->p_port_hierarchy_info->m_asic != -1)
            extra_fields.push_back("ASIC");
        if (p_port->p_port_hierarchy_info->m_type != -1)
            extra_fields.push_back("Type");
    }

    IBDIAG_RETURN_VOID;
}

void CheckPortHierarchyInfoByTemplateGuid0x04(IBPort *p_port,
                                              vector<string> &missing_fields,
                                              vector<string> &extra_fields) {
    IBDIAG_ENTER;

    int port_type = p_port->p_port_hierarchy_info->m_port_type;
    if (port_type == -1) {
            missing_fields.push_back("PortType");
            IBDIAG_RETURN_VOID;
    }

    if (port_type == HIERARCHY_INFO_PORT_TYPE_SW) {
        // Mission Port. Necessary fields:
        // Cage, Port, ASIC, Type
        if (p_port->p_port_hierarchy_info->m_cage == -1)
            missing_fields.push_back("Cage");
        if (p_port->p_port_hierarchy_info->m_port == -1)
            missing_fields.push_back("IPIL");
        if (p_port->p_port_hierarchy_info->m_asic == -1)
            missing_fields.push_back("ASIC");
        if (p_port->p_port_hierarchy_info->m_type == -1)
            missing_fields.push_back("Type");
        if (p_port->p_port_hierarchy_info->m_is_cage_manager == -1)
            missing_fields.push_back("IsCageManager");


        // Unnecessary fields:
        // IBPort
       if (p_port->p_port_hierarchy_info->m_ibport != -1)
            extra_fields.push_back("IBPort");

    } else if (port_type == HIERARCHY_INFO_PORT_TYPE_FNM1) {
        // FNM1 Port. Necessary fildes:
        // ASIC, Type
        if (p_port->p_port_hierarchy_info->m_asic == -1)
            missing_fields.push_back("ASIC");
        if (p_port->p_port_hierarchy_info->m_type == -1)
            missing_fields.push_back("Type");
        if (p_port->p_port_hierarchy_info->m_is_cage_manager == -1)
            missing_fields.push_back("IsCageManager");

        // Unnecessary fields:
        // AsicName, IBPort, Cage, Port, Split
        if (p_port->p_port_hierarchy_info->m_asic_name != -1)
            extra_fields.push_back("AsicName");
        if (p_port->p_port_hierarchy_info->m_ibport != -1)
            extra_fields.push_back("IBPort");
        if (p_port->p_port_hierarchy_info->m_cage != -1)
            extra_fields.push_back("Cage");
        if (p_port->p_port_hierarchy_info->m_port != -1)
            extra_fields.push_back("IPIL");
        if (p_port->p_port_hierarchy_info->m_split != -1)
            extra_fields.push_back("Split");

    } else if (port_type == HIERARCHY_INFO_PORT_TYPE_FNM) {
        // FNM Port. Necessary fildes:
        // AsicName, IBPort, ASIC, Type
        if (p_port->p_port_hierarchy_info->m_asic_name == -1)
            missing_fields.push_back("AsicName");
        if (p_port->p_port_hierarchy_info->m_ibport == -1)
            missing_fields.push_back("IBPort");
        if (p_port->p_port_hierarchy_info->m_asic == -1)
            missing_fields.push_back("ASIC");
        if (p_port->p_port_hierarchy_info->m_type == -1)
            missing_fields.push_back("Type");

        // Unnecessary fields:
        // Cage, Port, Split, Plane, NumOfPlanes, APort
        if (p_port->p_port_hierarchy_info->m_cage != -1)
            extra_fields.push_back("Cage");
        if (p_port->p_port_hierarchy_info->m_port != -1)
            extra_fields.push_back("IPIL");
        if (p_port->p_port_hierarchy_info->m_split != -1)
            extra_fields.push_back("Split");
        if (p_port->p_port_hierarchy_info->m_is_cage_manager != -1)
            extra_fields.push_back("IsCageManager");
        if (p_port->p_port_hierarchy_info->m_plane != -1)
            extra_fields.push_back("Plane");
        if (p_port->p_port_hierarchy_info->m_num_of_planes != -1)
            extra_fields.push_back("NumOfPlanes");
        if (p_port->p_port_hierarchy_info->m_aport != -1)
            extra_fields.push_back("APort");
    }

    IBDIAG_RETURN_VOID;
}

void CheckPortHierarchyInfoByTemplateGuid0x05(IBPort *p_port,
                                              vector<string> &missing_fields,
                                              vector<string> &extra_fields) {
    IBDIAG_ENTER;

    int port_type = p_port->p_port_hierarchy_info->m_port_type;
    if (port_type == -1) {
            missing_fields.push_back("PortType");
            IBDIAG_RETURN_VOID;
    }

    if (port_type == HIERARCHY_INFO_PORT_TYPE_HCA) {
        // Necessary fildes:
        // Cage, IPIL, BDF
        if (p_port->p_port_hierarchy_info->m_cage == -1)
            missing_fields.push_back("Cage");
        if (p_port->p_port_hierarchy_info->m_port == -1)
            missing_fields.push_back("IPIL");
        if (p_port->p_port_hierarchy_info->bdf() == -1)
            missing_fields.push_back("BDF");

        // Unnecessary fields:
        // IBPort, NumberOnBaseBoard
        if (p_port->p_port_hierarchy_info->m_ibport != -1)
            extra_fields.push_back("IBPort");
        if (p_port->p_port_hierarchy_info->m_number_on_base_board != -1)
            extra_fields.push_back("NumberOnBaseBoard"); 

    } else if (port_type == HIERARCHY_INFO_PORT_TYPE_GPU) {
        // Necessary fildes:
        // IBPort, NumberOnBaseBoard
        if (p_port->p_port_hierarchy_info->m_ibport == -1)
            missing_fields.push_back("IBPort");
        if (p_port->p_port_hierarchy_info->m_number_on_base_board == -1)
            missing_fields.push_back("NumberOnBaseBoard"); 

        // Unnecessary fields:
        // Cage, IPIL, Split, BDF, Plane, APort, NumOfPlanes
        if (p_port->p_port_hierarchy_info->m_cage != -1)
            extra_fields.push_back("Cage");
        if (p_port->p_port_hierarchy_info->m_port != -1)
            extra_fields.push_back("IPIL");
        if (p_port->p_port_hierarchy_info->m_split != -1)
            extra_fields.push_back("Split");
        if (p_port->p_port_hierarchy_info->bdf() != -1)
            extra_fields.push_back("BDF");
        if (p_port->p_port_hierarchy_info->m_plane != -1)
            extra_fields.push_back("Plane");
        if (p_port->p_port_hierarchy_info->m_aport != -1)
            extra_fields.push_back("APort");
        if (p_port->p_port_hierarchy_info->m_num_of_planes != -1)
            extra_fields.push_back("NumOfPlanes");
    }

    IBDIAG_RETURN_VOID;
}

int IBDiag::CheckPortHierarchyInfo(list_p_fabric_general_err &hierarchy_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_node = nI->second;
        if (!p_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_node->getInSubFabric() || !p_node->should_support_port_hierarchy_info)
            continue;

        bool is_splitted = p_node->isSplitted();

        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {
            IBPort *p_port = p_node->getPort(i);
            if (!p_port || p_port->isSpecialPort())
                continue;

            // 1. Check that all switch and only active CA ports have hierarchy info
            if (!p_port->p_port_hierarchy_info) {
                if (p_node->type == IB_SW_NODE || p_port->get_internal_state() == IB_PORT_STATE_ACTIVE)
                    hierarchy_errors.push_back(new FabricErrPortHierarchyMissing(p_port));

                continue;
            }

            // 2. Cache is_splitted for this node, since p_node->isSplitted() doesn't work for HCAs
            if (p_port->p_port_hierarchy_info->m_split != -1)
                is_splitted = true;
        }

        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {
            IBPort *p_port = p_node->getPort(i);
            if (!p_port || p_port->isSpecialPort() || !p_port->p_port_hierarchy_info)
                continue;

            // 3. Check that correct port hierarchy info fields are present
            vector<string> missing_fields;
            vector<string> extra_fields;
            
            switch (p_port->p_port_hierarchy_info->m_template_guid) {
                case HIERARCHY_TEMPLATE_GUID_PORT_0x03:
                    CheckPortHierarchyInfoByTemplateGuid0x03(p_node, p_port, is_splitted,
                                                             missing_fields, extra_fields);
                    break;
                case HIERARCHY_TEMPLATE_GUID_PORT_0x04:
                    CheckPortHierarchyInfoByTemplateGuid0x04(p_port, missing_fields, extra_fields);
                    break;
                case HIERARCHY_TEMPLATE_GUID_PORT_0x05:
                    CheckPortHierarchyInfoByTemplateGuid0x05(p_port, missing_fields, extra_fields);
                    break;
                default:
                    hierarchy_errors.push_back(new FabricErrHierarchyTemplateMismatch(p_port,
                                                     p_port->p_port_hierarchy_info->m_template_guid, 0));
                    continue;
            }

            if (!missing_fields.empty()) {
                hierarchy_errors.push_back(
                    new FabricErrPortHierarchyMissingFields(p_port, missing_fields));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }

            if (!extra_fields.empty())
                hierarchy_errors.push_back(
                    new FabricErrPortHierarchyExtraFields(p_port, extra_fields));
        }
    }

    IBDIAG_RETURN(rc);
}
