/*
 * Copyright (c) 2024-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string>
#include <cstring>
#include <cctype>
#include <cstdint>
#include <cstdlib>

#include "infiniband/ibdiag/ibdiag_csv_parser.h"
#include "infiniband/ibdiag/ibdiag_csv_records.h"
#include "infiniband/ibdiag/ibdiag.h"

#define ADD_CSV_FIELD(T, name, member)                                                      \
    parse_section_info.push_back(                                                           \
            ParseFieldInfo<T>(name,                                                         \
                    [](T & obj, const char * value)                                         \
                        { return Parse(value, obj.member); } ))


#define ADD_CSV_FIELD_WITH_NA(T, name, member, na_marker, def_value)                        \
    parse_section_info.push_back(                                                           \
            ParseFieldInfo<T>(name,                                                         \
                    [](T & obj, const char * value)                                         \
                        { return Parse(value, obj.member, &obj.na_marker, def_value); } ))

//
// Section NODES
//  - SMP_NodeInfo
//
int NodeRecord::Init(vector < ParseFieldInfo <class NodeRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(NodeRecord, "NodeDesc", m_description);

    #define __ADD__(x)  ADD_CSV_FIELD(NodeRecord, #x, m_data.x)
    __ADD__(NumPorts);
    __ADD__(NodeType);
    __ADD__(ClassVersion);
    __ADD__(BaseVersion);
    __ADD__(SystemImageGUID);
    __ADD__(NodeGUID);
    __ADD__(PortGUID);
    __ADD__(DeviceID);
    __ADD__(PartitionCap);
    __ADD__(revision);
    __ADD__(VendorID);
    __ADD__(LocalPortNum);
    #undef __ADD__

   return 0;
}

//
// Section PORTS
//  - SMP_PortInfo
//
int PortRecord::Init(vector < ParseFieldInfo <class PortRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(PortRecord, "NodeGuid", m_node_guid);
    ADD_CSV_FIELD(PortRecord, "PortGuid", m_port_guid);
    ADD_CSV_FIELD(PortRecord, "PortNum", m_port_num);

    #define __ADD__(x)  ADD_CSV_FIELD(PortRecord, #x, m_data.x)
    __ADD__(MKey);
    __ADD__(GIDPrfx);
    __ADD__(MSMLID);
    __ADD__(LID);
    __ADD__(CapMsk);
    __ADD__(M_KeyLeasePeriod);
    __ADD__(DiagCode);
    __ADD__(LinkWidthActv);
    __ADD__(LinkWidthSup);
    __ADD__(LinkWidthEn);
    __ADD__(LocalPortNum);
    __ADD__(LMC);
    __ADD__(MKeyProtBits);
    __ADD__(LinkDownDefState);
    __ADD__(PortPhyState);
    __ADD__(VLArbHighCap);
    __ADD__(VLHighLimit);
    __ADD__(InitType);
    __ADD__(VLCap);
    __ADD__(MSMSL);
    __ADD__(NMTU);
    __ADD__(FilterRawOutb);
    __ADD__(FilterRawInb);
    __ADD__(PartEnfOutb);
    __ADD__(PartEnfInb);
    __ADD__(OpVLs);
    __ADD__(HoQLife);
    __ADD__(VLStallCnt);
    __ADD__(MTUCap);
    __ADD__(InitTypeReply);
    __ADD__(VLArbLowCap);
    __ADD__(PKeyViolations);
    __ADD__(MKeyViolations);
    __ADD__(SubnTmo);
    __ADD__(ClientReregister);
    __ADD__(MulticastPKeyTrapSuppressionEnabled);
    __ADD__(GUIDCap);
    __ADD__(MaxCreditHint);
    __ADD__(OverrunErrs);
    __ADD__(LocalPhyError);
    __ADD__(RespTimeValue);
    __ADD__(LinkRoundTripLatency);
    __ADD__(PortState);
    #undef __ADD__

    ADD_CSV_FIELD_WITH_NA(PortRecord, "CapMsk2", m_data.CapMsk2, m_has_na, 0);


    parse_section_info.push_back(                                                           
            ParseFieldInfo<PortRecord>("LinkSpeedEn",
                    [](PortRecord & obj, const char * value)
                        {
                            uint32_t v  = 0;
                            bool rc = Parse(value, v);
                            if (rc) {
                                obj.m_data.LinkSpeedEn =     (u_int8_t)(v & 0xff);
                                obj.m_data.LinkSpeedExtEn =  (u_int8_t)((v >> 8) & 0xff);
                                obj.m_data.LinkSpeedExtEn2 = (u_int8_t)((v >> 24) & 0xff);
                            }

                            return rc;
                        } ));

    parse_section_info.push_back(                                                           
            ParseFieldInfo<PortRecord>("LinkSpeedActv",
                    [](PortRecord & obj, const char * value)
                        {
                            uint32_t v  = 0;
                            bool rc = Parse(value, v);
                            if (rc) {
                                obj.m_data.LinkSpeedActv =     (u_int8_t)(v & 0xff);
                                obj.m_data.LinkSpeedExtActv =  (u_int8_t)((v >> 8) & 0xff);
                                obj.m_data.LinkSpeedExtActv2 = (u_int8_t)((v >> 24) & 0xff);
                            }

                            return rc;
                        } ));

    parse_section_info.push_back(                                                           
            ParseFieldInfo<PortRecord>("LinkSpeedSup",
                    [](PortRecord & obj, const char * value)
                        {
                            uint32_t v  = 0;
                            bool rc = Parse(value, v);
                            if (rc) {
                                obj.m_data.LinkSpeedSup =     (u_int8_t)(v & 0xff);
                                obj.m_data.LinkSpeedExtSup =  (u_int8_t)((v >> 8) & 0xff);
                                obj.m_data.LinkSpeedExtSup2 = (u_int8_t)((v >> 24) & 0xff);
                            }

                            return rc;
                        } ));

    // ADD_CSV_FIELD(PortRecord, "FECActv", m_fec_actv);
    // ADD_CSV_FIELD(PortRecord, "RetransActv", m_retrans_actv);

    return 0;
}

//
// Section EXTENDED_NODE_INFO
//  - ib_extended_node_info
//
int ExtendedNodeInfoRecord::Init(vector < ParseFieldInfo <class ExtendedNodeInfoRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(ExtendedNodeInfoRecord, "NodeGuid", m_node_guid);

    #define __ADD__(x)  ADD_CSV_FIELD(ExtendedNodeInfoRecord, #x, m_data.x)
    __ADD__(sl2vl_cap);
    __ADD__(sl2vl_act);
    __ADD__(num_pcie);
    __ADD__(num_oob);
    __ADD__(AnycastLIDTop);
    __ADD__(AnycastLidCap);
    __ADD__(node_type_extended);
    __ADD__(asic_max_planes);
    #undef __ADD__

    return 0;
}

//
// Section EXTENDED_PORT_INFO
//  - SMP_MlnxExtPortInfo
//
int ExtendedPortInfoRecord::Init(vector < ParseFieldInfo <class ExtendedPortInfoRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(ExtendedPortInfoRecord, "NodeGuid", m_node_guid);
    ADD_CSV_FIELD(ExtendedPortInfoRecord, "PortGuid", m_port_guid);
    ADD_CSV_FIELD(ExtendedPortInfoRecord, "PortNum", m_port_num);

    #define __ADD__(x)  ADD_CSV_FIELD(ExtendedPortInfoRecord, #x, m_data.x)
    __ADD__(StateChangeEnable);
    __ADD__(RouterLIDEn);
    __ADD__(SHArPANEn);
    __ADD__(AME);
    __ADD__(LinkSpeedSupported);
    __ADD__(LinkSpeedEnabled);
    __ADD__(LinkSpeedActive);
    __ADD__(ActiveRSFECParity);
    __ADD__(ActiveRSFECData);
    __ADD__(CapabilityMask);
    __ADD__(FECModeActive);
    __ADD__(RetransMode);
    __ADD__(FDR10FECModeSupported);
    __ADD__(FDR10FECModeEnabled);
    __ADD__(FDRFECModeSupported);
    __ADD__(FDRFECModeEnabled);
    __ADD__(EDR20FECModeSupported);
    __ADD__(EDR20FECModeEnabled);
    __ADD__(EDRFECModeSupported);
    __ADD__(EDRFECModeEnabled);
    __ADD__(FDR10RetranSupported);
    __ADD__(FDR10RetranEnabled);
    __ADD__(FDRRetranSupported);
    __ADD__(FDRRetranEnabled);
    __ADD__(EDR20RetranSupported);
    __ADD__(EDR20RetranEnabled);
    __ADD__(EDRRetranSupported);
    __ADD__(EDRRetranEnabled);
    __ADD__(IsSpecialPort);
    __ADD__(SpecialPortCapabilityMask);
    __ADD__(IsFNMPort);
    __ADD__(HDRFECModeSupported);
    __ADD__(HDRFECModeEnabled);
    __ADD__(OOOSLMask);
    __ADD__(AdaptiveTimeoutSLMask);
    __ADD__(NDRFECModeSupported);
    __ADD__(NDRFECModeEnabled);
    #undef __ADD__

    ADD_CSV_FIELD(ExtendedPortInfoRecord, "UnhealthyReason",         m_data.unhealthy_reason);
    ADD_CSV_FIELD(ExtendedPortInfoRecord, "BwUtilization",           m_data.bw_utilization);
    ADD_CSV_FIELD(ExtendedPortInfoRecord, "BwUtilEn",                m_data.bw_util_en);
    ADD_CSV_FIELD(ExtendedPortInfoRecord, "MinBwUtilization",        m_data.min_bw_utilization);

    ADD_CSV_FIELD_WITH_NA(ExtendedPortInfoRecord, "SpecialPortType", m_data.SpecialPortType, m_has_na, 0);

    return 0;
}

//
// Section PORT_INFO_EXTENDED
//  - SMP_PortInfoExtended
//
int PortInfoExtendedRecord::Init(vector < ParseFieldInfo <class PortInfoExtendedRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(PortInfoExtendedRecord, "NodeGuid", m_node_guid);
    ADD_CSV_FIELD(PortInfoExtendedRecord, "PortGuid", m_port_guid);
    ADD_CSV_FIELD(PortInfoExtendedRecord, "PortNum",  m_port_num);

    #define __ADD__(x)  ADD_CSV_FIELD(PortInfoExtendedRecord, #x, m_data.x)
    __ADD__(FECModeActive);
    __ADD__(FDRFECModeSupported);
    __ADD__(FDRFECModeEnabled);
    __ADD__(EDRFECModeSupported);
    __ADD__(EDRFECModeEnabled);
    __ADD__(HDRFECModeSupported);
    __ADD__(HDRFECModeEnabled);
    __ADD__(NDRFECModeSupported);
    __ADD__(NDRFECModeEnabled);
    #undef __ADD__

    ADD_CSV_FIELD(PortInfoExtendedRecord, "CapabilityMask", m_data.CapMsk);

    return 0;
}

//
// Section SWITCHES
//  - SMP_SwitchInfo
//
int SwitchRecord::Init(vector < ParseFieldInfo <class SwitchRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(SwitchRecord, "NodeGUID", m_node_guid);

    #define __ADD__(x)  ADD_CSV_FIELD(SwitchRecord, #x, m_data.x)
    __ADD__(LinearFDBCap);
    __ADD__(RandomFDBCap);
    __ADD__(MCastFDBCap);
    __ADD__(LinearFDBTop);
    __ADD__(DefPort);
    __ADD__(DefMCastPriPort);
    __ADD__(DefMCastNotPriPort);
    __ADD__(LifeTimeValue);
    __ADD__(PortStateChange);
    __ADD__(OptimizedSLVLMapping);
    __ADD__(LidsPerPort);
    __ADD__(PartEnfCap);
    __ADD__(InbEnfCap);
    __ADD__(OutbEnfCap);
    __ADD__(FilterRawInbCap);
    __ADD__(FilterRawOutbCap);
    __ADD__(ENP0);
    __ADD__(MCastFDBTop);
    #undef __ADD__

    return 0;
}

//
// Section PORT_HIERARCHY_INFO
//
int PortHierarchyInfoRecord::Init(vector < ParseFieldInfo <class PortHierarchyInfoRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(PortHierarchyInfoRecord, "NodeGUID",      m_node_guid);
    ADD_CSV_FIELD(PortHierarchyInfoRecord, "PortGUID",      m_port_guid);
    ADD_CSV_FIELD(PortHierarchyInfoRecord, "TemplateGUID",  m_template_guid);
    ADD_CSV_FIELD(PortHierarchyInfoRecord, "PortNum",       m_port_num);

    #define __ADD__(n, x) ADD_CSV_FIELD_WITH_NA(PortHierarchyInfoRecord, n, x, m_has_na, -1)
    __ADD__("Bus",                  m_bus);
    __ADD__("Device",               m_device);
    __ADD__("Function",             m_function);
    __ADD__("Type",                 m_type);
    __ADD__("SlotType",             m_slot_type);
    __ADD__("SlotValue",            m_slot_value);
    __ADD__("IsCageManager",        m_is_cage_manager);
    __ADD__("NumberOnBaseBoard",    m_number_on_base_board);
    __ADD__("APort",                m_aport);
    __ADD__("Plane",                m_plane);
    __ADD__("NumOfPlanes",          m_num_of_planes);
    __ADD__("ASIC",                 m_asic);
    __ADD__("Cage",                 m_cage);
    __ADD__("Port",                 m_port);
    __ADD__("Split",                m_split);
    __ADD__("IBPort",               m_ibport);
    __ADD__("PortType",             m_port_type);
    __ADD__("AsicName",             m_asic_name);
    #undef __ADD__

    return 0;
}

//
// Section PHYSICAL_HIERARCHY_INFO
//
int PhysicalHierarchyInfoRecord::Init(vector < ParseFieldInfo <class PhysicalHierarchyInfoRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(PhysicalHierarchyInfoRecord, "NodeGUID", m_node_guid);

    #define __ADD__(n, x) ADD_CSV_FIELD_WITH_NA(PhysicalHierarchyInfoRecord, n, x, m_has_na, -1)
    __ADD__("CampusSerialNum",  m_campus_serial_num);
    __ADD__("RoomSerialNum",    m_room_serial_num);
    __ADD__("RackSerialNum",    m_rack_serial_num);
    __ADD__("SystemType",       m_system_type);
    __ADD__("SystemTopUNum",    m_system_topu_num);
    __ADD__("BoardType",        m_board_type);
    __ADD__("BoardSlotNum",     m_board_slot_num);
    __ADD__("DeviceSerialNum",  m_device_serial_num);
    #undef __ADD__

    return 0;
};

//
// Section AR_INFO
// - adaptive_routing_info
//
int ARInfoRecord::Init(vector <ParseFieldInfo <class ARInfoRecord> > &parse_section_info)
{
    int rc = 0;

    ADD_CSV_FIELD(ARInfoRecord, "NodeGUID", m_node_guid);

    #define __ADD__(x)  ADD_CSV_FIELD(ARInfoRecord, #x, m_data.x)
    __ADD__(e);
    __ADD__(is_arn_sup);
    __ADD__(is_frn_sup);
    __ADD__(is_fr_sup);
    __ADD__(fr_enabled);
    __ADD__(rn_xmit_enabled);
    __ADD__(is_ar_trials_supported);
    __ADD__(sub_grps_active);
    __ADD__(group_table_copy_sup);
    __ADD__(direction_num_sup);
    __ADD__(is4_mode);
    __ADD__(glb_groups);
    __ADD__(by_sl_cap);
    __ADD__(by_sl_en);
    __ADD__(by_transp_cap);
    __ADD__(dyn_cap_calc_sup);
    __ADD__(group_cap);
    __ADD__(group_top);
    __ADD__(group_table_cap);
    __ADD__(string_width_cap);
    __ADD__(ar_version_cap);
    __ADD__(rn_version_cap);
    __ADD__(sub_grps_supported);
    __ADD__(enable_by_sl_mask);
    __ADD__(by_transport_disable);
    __ADD__(ageing_time_value);
    __ADD__(is_hbf_supported);
    __ADD__(by_sl_hbf_en);
    __ADD__(enable_by_sl_mask_hbf);
    __ADD__(is_whbf_supported);
    __ADD__(whbf_en);
    __ADD__(whbf_granularity);
    __ADD__(is_symmetric_hash_supported);
    __ADD__(is_dceth_hash_supported);
    __ADD__(is_bth_dqp_hash_supported);
    __ADD__(is_pfrn_supported);
    __ADD__(pfrn_enabled);
    #undef __ADD__

    return (rc);
}

//
// Section LINKS
//
int LinkRecord::Init(vector < ParseFieldInfo <class LinkRecord> > &parse_section_info)
{
   ADD_CSV_FIELD(LinkRecord, "PortNum1",  port_num1);
   ADD_CSV_FIELD(LinkRecord, "NodeGuid1", node_guid1);

   ADD_CSV_FIELD(LinkRecord, "PortNum2",  port_num2);
   ADD_CSV_FIELD(LinkRecord, "NodeGuid2", node_guid2);

   return 0;
}

//
// Section GENERAL_INFO_SMP
//
int GeneralInfoSMPRecord::Init(vector < ParseFieldInfo <class GeneralInfoSMPRecord> > &parse_section_info)
{
    #define __ADD__(n, x)  ADD_CSV_FIELD(GeneralInfoSMPRecord, n, x)
    __ADD__("NodeGUID", m_node_guid);

    __ADD__("FWInfo_Extended_Major",    m_fw_info_extended_major);
    __ADD__("FWInfo_Extended_Minor",    m_fw_info_extended_minor);
    __ADD__("FWInfo_Extended_SubMinor", m_fw_info_extended_sub_minor);

    __ADD__("CapabilityMask_0",         m_capability_mask_fields[0]);
    __ADD__("CapabilityMask_1",         m_capability_mask_fields[1]);
    __ADD__("CapabilityMask_2",         m_capability_mask_fields[2]);
    __ADD__("CapabilityMask_3",         m_capability_mask_fields[3]);
    #undef __ADD__

    return 0;
}

//
// Section START_EXTENDED_SWITCH_INFO
//  - SMP_ExtendedSwitchInfo
//
int ExtendedSwitchInfoRecord::Init(vector < ParseFieldInfo <class ExtendedSwitchInfoRecord> > &parse_section_info)
{

    ADD_CSV_FIELD(ExtendedSwitchInfoRecord, "NodeGUID", m_node_guid);

    #define __ADD__(x)  ADD_CSV_FIELD(ExtendedSwitchInfoRecord, #x, m_data.x)
    __ADD__(end_to_end_timescale);
    __ADD__(request_issu);
    __ADD__(turbo_path_cap);
    __ADD__(turbo_path_enable);
    __ADD__(req_delay_cap);
    __ADD__(set_trig_th_cap);
    __ADD__(rst_trig_th_cap);
    __ADD__(req_trig_window_cap);
    __ADD__(req_delay);
    __ADD__(set_trig_th);
    __ADD__(rst_trig_th);
    __ADD__(req_trig_window);
    #undef __ADD__

    return 0;
}

//
// Section NODES_INFO
//  - VendorSpec_GeneralInfo
//
 int GeneralInfoGMPRecord::Init(vector < ParseFieldInfo <class GeneralInfoGMPRecord> > &parse_section_info)
 {
    ADD_CSV_FIELD(GeneralInfoGMPRecord, "NodeGUID", m_node_guid);

    #define __ADD__(x) ADD_CSV_FIELD_WITH_NA(GeneralInfoGMPRecord, "HWInfo_" #x, m_data.HWInfo.x, m_has_na, 0)
    __ADD__(DeviceID);
    __ADD__(DeviceHWRevision);
    __ADD__(technology);
    __ADD__(UpTime);
    #undef __ADD__

    #define __ADD__(x) ADD_CSV_FIELD_WITH_NA(GeneralInfoGMPRecord, "FWInfo_" #x, m_data.FWInfo.x, m_has_na, 0)
    __ADD__(SubMinor);
    __ADD__(Minor);
    __ADD__(Major);
    __ADD__(secure_fw);
    __ADD__(signed_fw);
    __ADD__(debug_fw);
    __ADD__(dev_fw);
    __ADD__(BuildID);
    __ADD__(Year);
    __ADD__(Day);
    __ADD__(Month);
    __ADD__(Hour);
    __ADD__(INI_File_Version);
    __ADD__(Extended_Major);
    __ADD__(Extended_Minor);
    __ADD__(Extended_SubMinor);
    __ADD__(isfu_major);
    __ADD__(sec_boot);
    __ADD__(life_cycle);
    #undef __ADD__

    parse_section_info.push_back(
            ParseFieldInfo<GeneralInfoGMPRecord>("FWInfo_PSID",
                    [](GeneralInfoGMPRecord & obj, const char * value)
                            {
                                 std::string psid;
                                 bool rc = Parse(value, psid);
                                 if (rc) {
                                        if (psid == "N/A") {
                                            obj.m_cap_has_na = true;
                                            return rc;
                                        }
                                        CLEAR_STRUCT(obj.m_data.FWInfo.PSID.PSID);
                                        if (psid == "UNKNOWN")
                                                psid = "";
                                        strncpy((char *)obj.m_data.FWInfo.PSID.PSID, psid.c_str(),
                                                    sizeof(obj.m_data.FWInfo.PSID.PSID) - 1);
                                 }
                                 return rc;
                             } ));

    #define __ADD__(x) ADD_CSV_FIELD_WITH_NA(GeneralInfoGMPRecord, "SWInfo_" #x, m_data.SWInfo.x, m_has_na, 0)
    __ADD__(SubMinor);
    __ADD__(Minor);
    __ADD__(Major);
    #undef __ADD__

    ADD_CSV_FIELD_WITH_NA(GeneralInfoGMPRecord, "CapabilityMask_0", m_data.CapabilityMask.capability0, m_cap_has_na, 0);
    ADD_CSV_FIELD_WITH_NA(GeneralInfoGMPRecord, "CapabilityMask_1", m_data.CapabilityMask.capability1, m_cap_has_na, 0);
    ADD_CSV_FIELD_WITH_NA(GeneralInfoGMPRecord, "CapabilityMask_2", m_data.CapabilityMask.capability2, m_cap_has_na, 0);
    ADD_CSV_FIELD_WITH_NA(GeneralInfoGMPRecord, "CapabilityMask_3", m_data.CapabilityMask.capability3, m_cap_has_na, 0);

    return 0;
 }

//
// Section PM_INFO
//  - PM_PortCounters
//  - PM_PortCountersExtended
//  - PM_PortRcvErrorDetails
//  - PM_PortXmitDiscardDetails
//
int PMInfoRecord::Init(vector < ParseFieldInfo <class PMInfoRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(PMInfoRecord, "NodeGUID", m_node_guid);
    ADD_CSV_FIELD(PMInfoRecord, "PortGUID", m_port_guid);
    ADD_CSV_FIELD(PMInfoRecord, "PortNumber", m_port_num);

    // PM_PortCounters
    #define __ADD__(x) ADD_CSV_FIELD(PMInfoRecord, #x, m_pm_port_counters.x)
    __ADD__(LinkDownedCounter);
    __ADD__(LinkErrorRecoveryCounter);
    __ADD__(SymbolErrorCounter);
    __ADD__(PortRcvRemotePhysicalErrors);
    __ADD__(PortRcvErrors);
    __ADD__(PortXmitDiscards);
    __ADD__(PortRcvSwitchRelayErrors);
    __ADD__(ExcessiveBufferOverrunErrors);
    __ADD__(LocalLinkIntegrityErrors);
    __ADD__(PortRcvConstraintErrors);
    __ADD__(PortXmitConstraintErrors);
    __ADD__(VL15Dropped);
    __ADD__(PortXmitData);
    __ADD__(PortRcvData);
    __ADD__(PortXmitPkts);
    __ADD__(PortRcvPkts);
    __ADD__(PortXmitWait);
    #undef __ADD__

    // PM_PortCountersExtended
    #define __ADD__(x) ADD_CSV_FIELD(PMInfoRecord, #x "Extended", m_pm_port_counters_ext.x)
    __ADD__(PortXmitData);
    __ADD__(PortRcvData);
    __ADD__(PortXmitPkts);
    __ADD__(PortRcvPkts);
    #undef __ADD__

    // PM_PortRcvErrorDetails
    #define __ADD__(x) ADD_CSV_FIELD(PMInfoRecord, #x, m_pm_port_counters_ext.x)
    __ADD__(PortUniCastXmitPkts);
    __ADD__(PortUniCastRcvPkts);
    __ADD__(PortMultiCastXmitPkts);
    __ADD__(PortMultiCastRcvPkts);
    #undef __ADD__

    // PM_PortRcvErrorDetails
    #define __ADD__(x) ADD_CSV_FIELD(PMInfoRecord, #x "Ext", m_pm_port_counters_ext.x)
    __ADD__(SymbolErrorCounter);
    __ADD__(LinkErrorRecoveryCounter);
    __ADD__(LinkDownedCounter);
    __ADD__(PortRcvErrors);
    __ADD__(PortRcvRemotePhysicalErrors);
    __ADD__(PortRcvSwitchRelayErrors);
    __ADD__(PortXmitDiscards);
    __ADD__(PortXmitConstraintErrors);
    __ADD__(PortRcvConstraintErrors);
    __ADD__(LocalLinkIntegrityErrors);
    __ADD__(ExcessiveBufferOverrunErrors);
    __ADD__(VL15Dropped);
    __ADD__(PortXmitWait);
    __ADD__(QP1Dropped);
    #undef __ADD__

    // PM_PortRcvErrorDetails
    #define __ADD__(x) ADD_CSV_FIELD(PMInfoRecord, #x, m_pm_port_rcv_error_details.x)
    __ADD__(PortLocalPhysicalErrors);
    __ADD__(PortMalformedPacketErrors);
    __ADD__(PortBufferOverrunErrors);
    __ADD__(PortDLIDMappingErrors);
    __ADD__(PortVLMappingErrors);
    __ADD__(PortLoopingErrors);
    #undef __ADD__

    // PM_PortXmitDiscardDetails
    #define __ADD__(x) ADD_CSV_FIELD(PMInfoRecord, #x, m_pm_port_xmit_discard_details.x)
    __ADD__(PortInactiveDiscards);
    __ADD__(PortNeighborMTUDiscards);
    __ADD__(PortSwLifetimeLimitDiscards);
    __ADD__(PortSwHOQLifetimeLimitDiscards);
    #undef __ADD__

    return 0;
}

//
// Section HBF_PORT_COUNTERS
//  - port_routing_decision_counters
//
int HBFPortCountersRecord::Init(vector < ParseFieldInfo <class HBFPortCountersRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(HBFPortCountersRecord, "NodeGUID", m_node_guid);
    ADD_CSV_FIELD(HBFPortCountersRecord, "PortGUID", m_port_guid);
    ADD_CSV_FIELD(HBFPortCountersRecord, "PortNumber", m_port_num);

    // port_routing_decision_counters
    #define __ADD__(x) ADD_CSV_FIELD(HBFPortCountersRecord, #x, m_data.x)
    __ADD__(rx_pkt_forwarding_static);
    __ADD__(rx_pkt_forwarding_hbf);
    __ADD__(rx_pkt_forwarding_ar);
    __ADD__(rx_pkt_hbf_fallback_local);
    __ADD__(rx_pkt_hbf_fallback_remote);
    __ADD__(rx_pkt_forwarding_hbf_sg0);
    __ADD__(rx_pkt_forwarding_hbf_sg1);
    __ADD__(rx_pkt_forwarding_hbf_sg2);
    __ADD__(rx_pkt_forwarding_ar_sg0);
    __ADD__(rx_pkt_forwarding_ar_sg1);
    __ADD__(rx_pkt_forwarding_ar_sg2);
    #undef __ADD__

    return 0;
}

//
// Section FAST_RECOVERY_COUNTERS
//  - VS_FastRecoveryCounters
//
int FastRecoveryCountersRecord::Init(vector < ParseFieldInfo <class FastRecoveryCountersRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(FastRecoveryCountersRecord, "NodeGUID", m_node_guid);
    ADD_CSV_FIELD(FastRecoveryCountersRecord, "PortGUID", m_port_guid);
    ADD_CSV_FIELD(FastRecoveryCountersRecord, "PortNumber", m_port_num);

    // VS_FastRecoveryCounters
    #define __ADD__(x) ADD_CSV_FIELD(FastRecoveryCountersRecord, #x, m_data.x)
    __ADD__(trigger);
    __ADD__(counter_overflow);
    __ADD__(num_errors);
    __ADD__(num_warnings);
    __ADD__(num_normals);
    __ADD__(min_value);
    __ADD__(max_value);
    __ADD__(consecutive_normal);
    __ADD__(last_value_0);
    __ADD__(last_value_1);
    __ADD__(last_value_2);
    #undef __ADD__

    return 0;
}

//
// Section CREDIT_WATCHDOG_TIMEOUT_COUNTERS
//  - VS_CreditWatchdogTimeoutCounters
//
int CreditWatchdogTimeoutCountersRecord::Init(vector < ParseFieldInfo <class CreditWatchdogTimeoutCountersRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(CreditWatchdogTimeoutCountersRecord, "NodeGUID", m_node_guid);
    ADD_CSV_FIELD(CreditWatchdogTimeoutCountersRecord, "PortGUID", m_port_guid);
    ADD_CSV_FIELD(CreditWatchdogTimeoutCountersRecord, "PortNumber", m_port_num);

    // VS_CreditWatchdogTimeoutCounters::credit_watchdog_timeout_per_vl[...]
    #define __ADD__(i)                                          \
        ADD_CSV_FIELD(                                          \
                CreditWatchdogTimeoutCountersRecord,            \
                "credit_watchdog_timeout_per_vl_" #i,           \
                m_data.credit_watchdog_timeout_per_vl[i])
    __ADD__(0);
    __ADD__(1);
    __ADD__(2);
    __ADD__(3);
    __ADD__(4);
    __ADD__(5);
    __ADD__(6);
    __ADD__(7);
    #undef __ADD__

    ADD_CSV_FIELD(CreditWatchdogTimeoutCountersRecord,
            "total_port_credit_watchdog_timeout",
            m_data.total_port_credit_watchdog_timeout);

    return 0;
}

//
// Section RN_COUNTERS
//  - port_rn_counters
//
int RNCountersRecord::Init(vector < ParseFieldInfo <class RNCountersRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(RNCountersRecord, "NodeGUID", m_node_guid);
    ADD_CSV_FIELD(RNCountersRecord, "PortGUID", m_port_guid);
    ADD_CSV_FIELD(RNCountersRecord, "PortNumber", m_port_num);

    #define __ADD__(x)  ADD_CSV_FIELD(RNCountersRecord, #x, m_data.x)
    __ADD__(port_rcv_rn_pkt);
    __ADD__(port_xmit_rn_pkt);
    __ADD__(port_rcv_rn_error);
    __ADD__(port_rcv_switch_relay_rn_error);
    __ADD__(port_ar_trials);
    __ADD__(pfrn_received_packet);
    __ADD__(pfrn_received_error);
    __ADD__(pfrn_xmit_packet);
    __ADD__(pfrn_start_packet);
    #undef __ADD__

    return 0;
}

//
// Section PM_PORT_SAMPLES_CONTROL
//  - PM_PortSamplesControl
//
int PMPortSamplesControlRecord::Init(vector < ParseFieldInfo <class PMPortSamplesControlRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(PMPortSamplesControlRecord, "NodeGUID", m_node_guid);
    ADD_CSV_FIELD(PMPortSamplesControlRecord, "PortGUID", m_port_guid);
    ADD_CSV_FIELD(PMPortSamplesControlRecord, "PortNumber", m_port_num);


    #define __ADD__(x)  ADD_CSV_FIELD(PMPortSamplesControlRecord, #x, m_data.x)
    __ADD__(Tick);
    __ADD__(CounterWidth);
    __ADD__(CounterMask0);
    __ADD__(CounterMasks1to9);
    __ADD__(CounterMasks10to14);
    __ADD__(SampleMechanisms);
    __ADD__(SampleStatus);
    __ADD__(VendorMask);
    __ADD__(SampleStart);
    __ADD__(SampleInterval);
    __ADD__(Tag);
    #undef __ADD__

    #define __ADD_COUNTER__(i)  ADD_CSV_FIELD(PMPortSamplesControlRecord, "CounterSelect" #i, m_data.CounterSelect[i])
    __ADD_COUNTER__(0);
    __ADD_COUNTER__(1);
    __ADD_COUNTER__(2);
    __ADD_COUNTER__(3);
    __ADD_COUNTER__(4);
    __ADD_COUNTER__(5);
    __ADD_COUNTER__(6);
    __ADD_COUNTER__(7);
    __ADD_COUNTER__(8);
    __ADD_COUNTER__(9);
    __ADD_COUNTER__(10);
    __ADD_COUNTER__(11);
    __ADD_COUNTER__(12);
    __ADD_COUNTER__(13);
    __ADD_COUNTER__(14);
    #undef __ADD_COUNTER__

    parse_section_info.push_back(
            ParseFieldInfo<PMPortSamplesControlRecord>("OptionMask",
                    [](PMPortSamplesControlRecord & obj, const char * value)
                        {
                            uint64_t v  = 0;
                            bool rc = Parse(value, v);
                            if (rc) {
                                IBDiag::uint64ToPMOptionMask(obj.m_data.PortSampleControlOptionMask, v);
                            }

                            return rc;
                        }));

    return 0;
}

//
// Section CHASSIS_INFO
//  - SMP_ChassisInfo
//
int ChassisInfoRecord::Init(vector < ParseFieldInfo <class ChassisInfoRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(ChassisInfoRecord, "NodeGUID", m_node_guid);

    parse_section_info.push_back(
            ParseFieldInfo<ChassisInfoRecord>("chassis_id",
                    [](ChassisInfoRecord & obj, const char * value)
                            {
                                 std::string id;
                                 bool rc = Parse(value, id);
                                 if (rc) {

                                        CLEAR_STRUCT(obj.m_data.chassis_id);
                                        strncpy((char *)obj.m_data.chassis_id, id.c_str(),
                                                    sizeof(obj.m_data.chassis_id) - 1);
                                 }
                                 return rc;
                             } ));

    return 0;
}
