/*
 * Copyright (c) 2017-2020 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <infiniband/ibdiag/ibdiag_fabric.h>
#include <infiniband/ibdiag/ibdiag_csv_records.h>
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <infiniband/ibutils/output_stream_primitives.h>
#include <infiniband/ibdiag/ibdiag_csv_parser.h>

int IBDiagFabric::UpdateFabric(const string &csv_file)
{
    IBDiag::PrintFileTimestamp(csv_file, "DB CSV");

    int rc = IBDIAG_SUCCESS_CODE;
    CsvFileStream csv_file_stream(csv_file, m_csv_parser);

    SectionParser<NodeRecord>  node_section_parser;
    node_section_parser.Init(SECTION_NODES);
    rc = m_csv_parser.ParseAndHandleSection <NodeRecord, IBDiagFabric>(csv_file_stream,
            node_section_parser, this, &IBDiagFabric::CreateNode);
    if (rc) {
        if(!this->last_error.empty()) ERR_PRINT("%s", this->last_error.c_str());
        ERR_PRINT("Failed to parse " SECTION_NODES " section in csv file\n");
        return rc;
    }

    SectionParser<PortRecord>  port_section_parser;
    port_section_parser.Init(SECTION_PORTS);
    rc = m_csv_parser.ParseAndHandleSection <PortRecord, IBDiagFabric>(csv_file_stream,
            port_section_parser, this, &IBDiagFabric::CreatePort);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_PORTS " section in csv file\n");
        return rc;
    }
    rc = this->CreateDummyPorts();
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_PORTS " section in csv file\n");
        return rc;
    }

    SectionParser<SwitchRecord>  switch_section_parser;
    switch_section_parser.Init(SECTION_SWITCHES);
    rc = m_csv_parser.ParseAndHandleSection <SwitchRecord, IBDiagFabric>(csv_file_stream,
            switch_section_parser, this, &IBDiagFabric::CreateSwitch);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_SWITCHES " section in csv file\n");
        return rc;
    }

    SectionParser<LinkRecord>  link_section_parser;
    link_section_parser.Init(SECTION_LINKS);
    rc = m_csv_parser.ParseAndHandleSection <LinkRecord, IBDiagFabric>(csv_file_stream,
            link_section_parser, this, &IBDiagFabric::CreateLink);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_LINKS " section in csv file\n");
        return rc;
    }

    SectionParser<GeneralInfoGMPRecord> general_info_gmp_section_parser;
    general_info_gmp_section_parser.Init(SECTION_NODES_INFO);
    rc = m_csv_parser.ParseAndHandleSection <GeneralInfoGMPRecord, IBDiagFabric>(csv_file_stream,
            general_info_gmp_section_parser, this, &IBDiagFabric::CreateVSGeneralInfoGMP);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_NODES_INFO " section in csv file\n");
        return rc;
    }

    SectionParser<GeneralInfoSMPRecord> general_info_smp_section_parser;
    general_info_smp_section_parser.Init(SECTION_GENERAL_INFO_SMP);
    rc = m_csv_parser.ParseAndHandleSection <GeneralInfoSMPRecord, IBDiagFabric>(csv_file_stream,
            general_info_smp_section_parser, this, &IBDiagFabric::CreateVSGeneralInfoSMP);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_GENERAL_INFO_SMP " section in csv file\n");
        return rc;
    }

    SectionParser<ExtendedNodeInfoRecord> extended_node_info_section_parser;
    extended_node_info_section_parser.Init(SECTION_EXTENDED_NODE_INFO);
    rc = m_csv_parser.ParseAndHandleSection <ExtendedNodeInfoRecord, IBDiagFabric>(csv_file_stream,
         extended_node_info_section_parser, this, &IBDiagFabric::CreateExtendedNodeInfo);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_EXTENDED_NODE_INFO " section in csv file\n");
        return rc;
    }

    SectionParser<ExtendedPortInfoRecord> extended_port_info_section_parser;
    extended_port_info_section_parser.Init(SECTION_EXTENDED_PORT_INFO);
    rc = m_csv_parser.ParseAndHandleSection <ExtendedPortInfoRecord, IBDiagFabric>(csv_file_stream,
            extended_port_info_section_parser, this, &IBDiagFabric::CreateExtendedPortInfo);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_EXTENDED_PORT_INFO " section in csv file\n");
        return rc;
    }

    SectionParser<ExtendedSwitchInfoRecord> extended_switch_info_section_parser;
    extended_switch_info_section_parser.Init(SECTION_EXTENDED_SWITCH_INFO);
    rc = m_csv_parser.ParseAndHandleSection <ExtendedSwitchInfoRecord, IBDiagFabric>(csv_file_stream,
            extended_switch_info_section_parser, this, &IBDiagFabric::CreateExtendedSwitchInfo);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_EXTENDED_SWITCH_INFO " section in csv file\n");
        return rc;
    }

    SectionParser<PortInfoExtendedRecord> port_info_extended_section_parser;
    port_info_extended_section_parser.Init(SECTION_PORT_INFO_EXTENDED);
    rc = m_csv_parser.ParseAndHandleSection <PortInfoExtendedRecord, IBDiagFabric>(csv_file_stream,
            port_info_extended_section_parser, this, &IBDiagFabric::CreatePortInfoExtended);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_PORT_INFO_EXTENDED " section in csv file\n");
        return rc;
    }

    SectionParser<PortHierarchyInfoRecord> port_hierarchy_info_section_parser;
    port_hierarchy_info_section_parser.Init(SECTION_PORT_HIERARCHY_INFO);
    rc = m_csv_parser.ParseAndHandleSection <PortHierarchyInfoRecord, IBDiagFabric>(csv_file_stream,
            port_hierarchy_info_section_parser, this, &IBDiagFabric::CreatePortHierarchyInfo);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_PORT_HIERARCHY_INFO " section in csv file\n");
        return rc;
    }

    SectionParser<PhysicalHierarchyInfoRecord> physical_hierarchy_info_section_parser;
    physical_hierarchy_info_section_parser.Init(SECTION_PHYSICAL_HIERARCHY_INFO);
    rc = m_csv_parser.ParseAndHandleSection <PhysicalHierarchyInfoRecord, IBDiagFabric>(csv_file_stream,
            physical_hierarchy_info_section_parser, this, &IBDiagFabric::CreatePhysicalHierarchyInfo);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_PHYSICAL_HIERARCHY_INFO " section in csv file\n");
        return rc;
    }

    // AR Info
    SectionParser<ARInfoRecord> ar_info_section_parser;
    ar_info_section_parser.Init(SECTION_AR_INFO);
    rc = m_csv_parser.ParseAndHandleSection <ARInfoRecord, IBDiagFabric>(csv_file_stream,
            ar_info_section_parser, this, &IBDiagFabric::CreateARInfo);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_AR_INFO " section in csv file\n");
        return rc;
    }

    return rc;
}

int IBDiagFabric::CreateNode(const NodeRecord &nodeRecord)
{
    int rc = 0;
    const struct SMP_NodeInfo & x = nodeRecord.m_data;

    IBNode* p_node = discovered_fabric.makeNode((IBNodeType)x.NodeType,
                                                (phys_port_t)x.NumPorts,
                                                x.SystemImageGUID,
                                                x.NodeGUID,
                                                x.VendorID,
                                                x.DeviceID,
                                                x.revision,
                                                nodeRecord.m_description,
                                                true);

    if (!p_node) {
        stringstream ss;
        ss << "Cannot load from file IBNode GUID=" << HEX(x.NodeGUID)
           << ", System GUID=" << HEX(x.NodeGUID)
           << ", ports=" << DEC(x.NumPorts)
           << ", description '" <<  nodeRecord.m_description << "'" << endl;

        last_error = ss.str();

        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }

    //take care of progress bar
    ++nodes_found;
    if (x.NodeType == IB_CA_NODE)
       ++ca_found;
    else
       ++sw_found;
    ports_found += x.NumPorts;

    rc = this->fabric_extended_info.addSMPNodeInfo(p_node, x);

    return rc;
}

int IBDiagFabric::CreatePort(const PortRecord &portRecord)
{
    int rc = 0;

    const struct SMP_PortInfo& port_info = portRecord.m_data;

    IBNode* p_node = discovered_fabric.NodeByGuid[portRecord.m_node_guid];

    static u_int16_t base_lid = 0; // static - to keep the value for all switch ports
    static u_int8_t  lmc = 0;
    static u_int32_t cap_mask = 0;
    static u_int16_t cap_mask2 = 0;

    if (!(p_node->type == IB_SW_NODE && portRecord.m_port_num != 0)) {
        base_lid = port_info.LID;
        lmc = port_info.LMC;
        cap_mask = port_info.CapMsk;
        cap_mask2 = port_info.CapMsk2;
    }

    IBLinkSpeed correct_speed = this->fabric_extended_info.getCorrectSpeed(port_info,
                                                                           cap_mask,
                                                                           cap_mask2);

    IBPort *p_port = discovered_fabric.setNodePort(p_node,
                                                   portRecord.m_port_guid,
                                                   base_lid,
                                                   lmc,
                                                   portRecord.m_port_num,
                                                   (IBLinkWidth)port_info.LinkWidthActv,
                                                   correct_speed,
                                                   (IBPortState)port_info.PortState);

    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);

    p_port->setPortInfoMadWasSent(true);

    rc = this->fabric_extended_info.addSMPPortInfo(p_port, port_info);

    return rc;
}

/*
 * We should create IBPorts and empty Port-Info data MADs also for
 * ports that weren't dump/MADs weren't sent for them.
*/
int IBDiagFabric::CreateDummyPorts()
{
    IBDIAG_ENTER;

    for (map_str_pnode::iterator it_node = this->discovered_fabric.NodeByName.begin();
             it_node != this->discovered_fabric.NodeByName.end(); ++it_node) {
         IBNode *p_curr_node = it_node->second;
         if (!p_curr_node)
             IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);

         struct SMP_NodeInfo *p_node_info =
                               this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
         if (!p_node_info)
             IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);

         lid_t base_lid = 0;
         uint8_t lmc = 0;
         if (p_curr_node->type == IB_SW_NODE) {
             IBPort *p_port_zero = p_curr_node->getPort(0);
             if (!p_port_zero){
                 ERR_PRINT("DB Error - found null port 0 for Node GUID: " U64H_FMT " \n",
                                p_curr_node->guid_get());
                 IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
             }

             base_lid = p_port_zero->base_lid;
             lmc = p_port_zero->lmc;
         }

         for (phys_port_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
              IBPort *p_curr_port = p_curr_node->getPort(port_num);

              if (this->discovered_fabric.defAllPorts) {
                  if (!p_curr_port) {
                      ERR_PRINT("DB Error - found null port for Node GUID: " U64H_FMT " port num: %d\n",
                                p_curr_node->guid_get(), port_num);
                      IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
                  }

                  if (p_curr_port->guid_get())
                      continue;

                  //HCA only sets GUID, LID etc for its Active port
                  if (p_curr_node->type != IB_SW_NODE)
                      continue;
              } else
                  if (p_curr_port)
                      continue;

              struct SMP_PortInfo port_info;
              CLEAR_STRUCT(port_info);
              port_info.LinkSpeedActv = IB_UNKNOWN_LINK_SPEED;
              port_info.LinkWidthActv = IB_UNKNOWN_LINK_WIDTH;
              port_info.PortState = IB_PORT_STATE_DOWN;
              port_info.PortPhyState = IB_PORT_PHYS_STATE_POLLING;

              //this function resets base_lid and lmc for Zero port any time it is called
              //we have to provide correct values
              p_curr_port = this->discovered_fabric.setNodePort(p_curr_node,
                                                             p_node_info->PortGUID,
                                                             base_lid,
                                                             lmc,
                                                             (p_curr_node->type == IB_SW_NODE) ?
                                                               port_num : p_node_info->LocalPortNum,
                                                             (IBLinkWidth)port_info.LinkWidthActv,
                                                             (IBLinkSpeed)port_info.LinkSpeedActv,
                                                             (IBPortState)port_info.PortState);
              if (!p_curr_port)
                  IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
              p_curr_port->setPortInfoMadWasSent(false);

              int rc = this->fabric_extended_info.addSMPPortInfo(p_curr_port, port_info);
              if (rc)
                  IBDIAG_RETURN(rc);
         }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreateSwitch(const SwitchRecord &switchRecord)
{
    int rc = 0;

    IBNode* p_node = discovered_fabric.NodeByGuid[switchRecord.m_node_guid];

    rc = fabric_extended_info.addSMPSwitchInfo(p_node, switchRecord.m_data);

    IBDIAG_RETURN(rc);
}

int IBDiagFabric::CreateLink(const LinkRecord &linkRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node1 = this->discovered_fabric.getNodeByGuid(linkRecord.node_guid1);
    IBNode *p_node2 = this->discovered_fabric.getNodeByGuid(linkRecord.node_guid2);
    if (!(p_node1 && p_node2)) {
        ERR_PRINT("DB error - found null node for Node %d GUID: " U64H_FMT " in csv file, section: "
                  SECTION_LINKS "\n", p_node1 ? 2 : 1,
                  p_node1 ? linkRecord.node_guid2 : linkRecord.node_guid1);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    IBPort *p_port1 = p_node1->getPort(linkRecord.port_num1);
    IBPort *p_port2 = p_node2->getPort(linkRecord.port_num2);
    if (!(p_port1 && p_port2)) {
        ERR_PRINT("DB error - found null port for Node %d GUID: " U64H_FMT " port num: " U32D_FMT
                  " in csv file, section: " SECTION_LINKS "\n", p_port1 ? 2 : 1,
                  p_port1 ? linkRecord.node_guid2 : linkRecord.node_guid1,
                  p_port1 ? linkRecord.port_num2 : linkRecord.port_num1);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    p_port1->p_remotePort = p_port2;
    p_port2->p_remotePort = p_port1;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreateVSGeneralInfoSMP(const GeneralInfoSMPRecord &generalInfoSMPRecord)
{
    IBDIAG_ENTER;

    if (!this->discovered_fabric.getNodeByGuid(generalInfoSMPRecord.m_node_guid)) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_GENERAL_INFO_SMP "\n", generalInfoSMPRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    fw_version_obj_t  fw_info_smp;
    capability_mask_t capability_mask_smp;

    if ((generalInfoSMPRecord.m_fw_info_extended_major != "N/A") &&
        (generalInfoSMPRecord.m_fw_info_extended_minor != "N/A") &&
        (generalInfoSMPRecord.m_fw_info_extended_sub_minor != "N/A")) {
        Parse(generalInfoSMPRecord.m_fw_info_extended_major.c_str(), fw_info_smp.major);
        Parse(generalInfoSMPRecord.m_fw_info_extended_minor.c_str(), fw_info_smp.minor);
        Parse(generalInfoSMPRecord.m_fw_info_extended_sub_minor.c_str(), fw_info_smp.sub_minor);

        this->capability_module.AddSMPFw(generalInfoSMPRecord.m_node_guid, fw_info_smp);
    }

    bool is_capability_mask_available = true;
    for (int capability_field = 0; capability_field < NUM_CAPABILITY_FIELDS; capability_field++) {
         if (generalInfoSMPRecord.m_capability_mask_fields[capability_field] == "N/A") {
             is_capability_mask_available = false;
             break;
         }
         Parse(generalInfoSMPRecord.m_capability_mask_fields[capability_field].c_str(),
               capability_mask_smp.mask[capability_field]);
    }

    if (is_capability_mask_available)
        this->capability_module.AddSMPCapabilityMask(generalInfoSMPRecord.m_node_guid,
                                                     capability_mask_smp);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreateVSGeneralInfoGMP(const GeneralInfoGMPRecord &generalInfoGMPRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node = this->discovered_fabric.getNodeByGuid(generalInfoGMPRecord.m_node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_NODES_INFO "\n", generalInfoGMPRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    capability_mask_t capability_mask_gmp;
    if (!generalInfoGMPRecord.m_cap_has_na) {
        capability_mask_gmp = generalInfoGMPRecord.m_data.CapabilityMask;
        this->capability_module.AddGMPCapabilityMask(generalInfoGMPRecord.m_node_guid,
                                                     capability_mask_gmp);

    }

    fw_version_obj_t  fw_info_gmp;
    if (!generalInfoGMPRecord.m_has_na) {
        p_node->serdes_version = (SerDesVersion) generalInfoGMPRecord.m_data.HWInfo.technology;

        fw_info_gmp = generalInfoGMPRecord.m_data;
        this->capability_module.AddGMPFw(generalInfoGMPRecord.m_node_guid, fw_info_gmp);

        int rc = this->fabric_extended_info.addVSGeneralInfo(p_node, generalInfoGMPRecord.m_data);
        if (rc) {
            ERR_PRINT("Failed to add VSGeneralInfo for node=%s, err=%u",
                        p_node->getName().c_str(), rc);
            IBDIAG_RETURN(rc);
        }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreateExtendedSwitchInfo(const ExtendedSwitchInfoRecord& extendedSwitchInfoRecord)
{
    IBDIAG_ENTER;
    IBNode *p_node = this->discovered_fabric.getNodeByGuid(extendedSwitchInfoRecord.m_node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_EXTENDED_SWITCH_INFO "\n", extendedSwitchInfoRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    int rc = this->fabric_extended_info.addSMP_ExtendedSwitchInfo(p_node, extendedSwitchInfoRecord.m_data);

    if (rc) {
        ERR_PRINT("Failed to store VS Extended Switch Info for node %s, err=%u",
                  p_node->getName().c_str(), rc);
        IBDIAG_RETURN(rc);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    
}

int IBDiagFabric::CreateExtendedNodeInfo(const ExtendedNodeInfoRecord &extendedNodeInfoRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node = this->discovered_fabric.getNodeByGuid(extendedNodeInfoRecord.m_node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_EXTENDED_NODE_INFO "\n", extendedNodeInfoRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    const struct ib_extended_node_info meni = extendedNodeInfoRecord.m_data;
    if (meni.sl2vl_act)
        p_node->setSL2VLAct(meni.sl2vl_act);

    p_node->ext_type = (IBExtNodeType)meni.node_type_extended;

    // Add data to fabric_extended_info
    int rc = this->fabric_extended_info.addSMPExtNodeInfo(p_node, meni);

    if (rc) {
        ERR_PRINT("Failed to store VS Extended Node Info for node %s, err=%u",
                  p_node->getName().c_str(), rc);
        IBDIAG_RETURN(rc);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreateExtendedPortInfo(const ExtendedPortInfoRecord &extendedPortInfoRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node = this->discovered_fabric.getNodeByGuid(extendedPortInfoRecord.m_node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_EXTENDED_PORT_INFO "\n", extendedPortInfoRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    IBPort *p_port = p_node->getPort((phys_port_t)extendedPortInfoRecord.m_port_num);
    if (!p_port) {
        ERR_PRINT("DB error - found null port for Node GUID " U64H_FMT " port num: " U8H_FMT
                  " in csv file, section: " SECTION_EXTENDED_PORT_INFO "\n",
                  extendedPortInfoRecord.m_node_guid, extendedPortInfoRecord.m_port_num);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    if (p_port->guid_get() != extendedPortInfoRecord.m_port_guid) {
        ERR_PRINT("DB error - Mismatch between Port %d GUID " U64H_FMT " in fabric to Port GUID "
                  U64H_FMT " in csv file, section: " SECTION_EXTENDED_PORT_INFO "\n",
                  p_port->num, p_port->guid_get(), extendedPortInfoRecord.m_port_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    const struct SMP_MlnxExtPortInfo& mepi = extendedPortInfoRecord.m_data;

    // Maybe we need to change speed to MLNX speed
    IBLinkSpeed speed = (mepi.LinkSpeedActive ? \
                         mlnxspeed2speed(mepi.LinkSpeedActive) : \
                         p_port->get_common_speed());
    p_port->set_internal_speed(speed);

    // Check if can determine the FEC Mode
    if (EXT_FEC_MODE_CONTROL_AND_REPORT_SUPPORT(mepi.CapabilityMask))
            p_port->set_fec_mode((IBFECMode)mepi.FECModeActive);

    if (mepi.IsSpecialPort)
            p_port->setSpecialPortType((IBSpecialPortType)mepi.SpecialPortType);

    int rc = this->fabric_extended_info.addSMPMlnxExtPortInfo(p_port, mepi);
    if (rc) {
        ERR_PRINT("Failed to store vs extended port info for port %s err=%u\n",
                  p_port->getName().c_str(), rc);
        IBDIAG_RETURN(rc);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreatePortInfoExtended(const PortInfoExtendedRecord &portInfoExtendedRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node = this->discovered_fabric.getNodeByGuid(portInfoExtendedRecord.m_node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_PORT_INFO_EXTENDED "\n", portInfoExtendedRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    IBPort *p_port = p_node->getPort((phys_port_t)portInfoExtendedRecord.m_port_num);
    if (!p_port) {
        ERR_PRINT("DB error - found null port for Node GUID " U64H_FMT " port num: " U8H_FMT
                  " in csv file, section: " SECTION_PORT_INFO_EXTENDED "\n",
                  portInfoExtendedRecord.m_node_guid, portInfoExtendedRecord.m_port_num);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    if (p_port->guid_get() != portInfoExtendedRecord.m_port_guid) {
        ERR_PRINT("DB error - Mismatch between Port %d GUID " U64H_FMT " in fabric to Port GUID "
                  U64H_FMT " in csv file, section: " SECTION_PORT_INFO_EXTENDED "\n",
                  p_port->num, p_port->guid_get(), portInfoExtendedRecord.m_port_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    const struct SMP_PortInfoExtended port_info_extended = portInfoExtendedRecord.m_data;

    if (PORT_INFO_EXT_FEC_MODE_SUPPORT(port_info_extended.CapMsk)) {
            // convert from bit mask to real value
            IBFECMode fec_value = fec_mask2value(port_info_extended.FECModeActive);
            if (fec_value == IB_FEC_NA)
                ERR_PRINT("Wrong FECModeActive value: " U16H_FMT " in csv file, section: "
                          SECTION_PORT_INFO_EXTENDED "\n", port_info_extended.FECModeActive);
            p_port->set_fec_mode(fec_value);
    }

    int rc = this->fabric_extended_info.addSMPPortInfoExtended(p_port, port_info_extended);
    if (rc) {
        ERR_PRINT("Failed to store port info extended for port %s err=%u\n",
                  p_port->getName().c_str(), rc);
        IBDIAG_RETURN(rc);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreatePortHierarchyInfo(const PortHierarchyInfoRecord &portHierarchyInfoRecord)
{
    IBDIAG_ENTER;
    IBNode *p_node = this->discovered_fabric.getNodeByGuid(portHierarchyInfoRecord.m_node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                SECTION_PORT_HIERARCHY_INFO "\n", portHierarchyInfoRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    IBPort *p_port = p_node->getPort((phys_port_t)portHierarchyInfoRecord.m_port_num);
    if (!p_port) {
        ERR_PRINT("DB error - found null port for Node GUID " U64H_FMT " port num: " U8H_FMT
                  " in csv file, section: " SECTION_PORT_HIERARCHY_INFO "\n",
                  portHierarchyInfoRecord.m_node_guid, portHierarchyInfoRecord.m_port_num);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    if (p_port->guid_get() != portHierarchyInfoRecord.m_port_guid) {
        ERR_PRINT("DB error - Mismatch between Port %d GUID " U64H_FMT " in fabric to Port GUID "
                  U64H_FMT " in csv file, section: " SECTION_PORT_HIERARCHY_INFO "\n",
                  p_port->num, p_port->guid_get(), portHierarchyInfoRecord.m_port_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    p_port->p_node->should_support_port_hierarchy_info = true;
    p_port->p_port_hierarchy_info =
            new PortHierarchyInfo(
                    p_node->type,
                    portHierarchyInfoRecord.m_template_guid,
                    portHierarchyInfoRecord.m_port_type,
                    portHierarchyInfoRecord.m_asic_name,
                    portHierarchyInfoRecord.m_ibport,
                    portHierarchyInfoRecord.m_is_cage_manager,
                    portHierarchyInfoRecord.m_number_on_base_board,
                    portHierarchyInfoRecord.m_aport,
                    portHierarchyInfoRecord.m_plane,
                    portHierarchyInfoRecord.m_num_of_planes,
                    {portHierarchyInfoRecord.m_bus, portHierarchyInfoRecord.m_device,
                        portHierarchyInfoRecord.m_function},
                    portHierarchyInfoRecord.m_type,
                    portHierarchyInfoRecord.m_slot_type,
                    portHierarchyInfoRecord.m_slot_value,
                    portHierarchyInfoRecord.m_asic,
                    portHierarchyInfoRecord.m_cage,
                    portHierarchyInfoRecord.m_port,
                    portHierarchyInfoRecord.m_split);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreatePhysicalHierarchyInfo(const PhysicalHierarchyInfoRecord &physicalHierarchyInfoRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node = this->discovered_fabric.getNodeByGuid(physicalHierarchyInfoRecord.m_node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                SECTION_PHYSICAL_HIERARCHY_INFO "\n", physicalHierarchyInfoRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    p_node->p_physical_hierarchy_info =
            new PhysicalHierarchyInfo(
                    physicalHierarchyInfoRecord.m_campus_serial_num,
                    physicalHierarchyInfoRecord.m_room_serial_num,
                    physicalHierarchyInfoRecord.m_rack_serial_num,
                    physicalHierarchyInfoRecord.m_system_type,
                    physicalHierarchyInfoRecord.m_system_topu_num,
                    physicalHierarchyInfoRecord.m_board_type,
                    physicalHierarchyInfoRecord.m_board_slot_num,
                    physicalHierarchyInfoRecord.m_device_serial_num);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// AR Info
int IBDiagFabric::CreateARInfo(const ARInfoRecord &arInfoRecord)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    IBNode* p_node = this->discovered_fabric.getNodeByGuid(arInfoRecord.m_node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_AR_INFO "\n", arInfoRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    const adaptive_routing_info& ar_info = arInfoRecord.m_data;

    rc = fabric_extended_info.addARInfo(p_node, ar_info);

    IBDIAG_RETURN(rc);
}



