/*
 * Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"

extern IBDiagClbck ibDiagClbck;


/****************************************************/
typedef map < u_int16_t, u_int8_t > map_pkey_membership;
typedef vector < struct P_Key_Block_Element > vec_pkey_block_elem;

/****************************************************/
void readPortPartitionTableToMap(IBDMExtendedInfo *extended_info,
        get_pkey_table_func_t get_pkey_tbl,
        u_int32_t createIndex, u_int16_t cap,
        map_pkey_membership &pkeys_to_mem)
{
    IBDIAG_ENTER;
    struct SMP_PKeyTable *p_pkey_tbl;
    uint32_t num_of_blocks;
    int tbl_size;
    u_int16_t pkey;

    num_of_blocks = (cap + IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS - 1) /
            IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;
    tbl_size = IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

    for (u_int32_t i = 0; i < num_of_blocks; ++i) {
        p_pkey_tbl = (extended_info->*get_pkey_tbl)(createIndex, i);
        if (!p_pkey_tbl)
            continue;

        if ((i + 1) * IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS  > cap)
            tbl_size = cap % IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

        for (int j = 0; j < tbl_size; ++j) {
            pkey = p_pkey_tbl->PKey_Entry[j].P_KeyBase;
            if (pkey == 0)
                continue;

            if (p_pkey_tbl->PKey_Entry[j].Membership_Type) {
                pkeys_to_mem.insert(pair<u_int16_t, u_int8_t>(pkey, (u_int8_t)1));
            } else {
                pkeys_to_mem.insert(pair<u_int16_t, u_int8_t>(pkey, (u_int8_t)0));
            }
        }
    }
    IBDIAG_RETURN_VOID;
}

static void readPortPartitionTableToVec(IBDMExtendedInfo *extended_info,
        IBPort * p_port, u_int16_t cap,
        vec_pkey_block_elem &vec_pkeys)
{
    IBDIAG_ENTER;
    struct SMP_PKeyTable *p_pkey_tbl;
    int num_of_blocks;
    int tbl_size;

    vec_pkeys.clear();
    vec_pkeys.resize(cap);
    for (u_int32_t i = 0; i < cap; ++i) {
        vec_pkeys[i].P_KeyBase = 0;
        vec_pkeys[i].Membership_Type = 0;
    }
    num_of_blocks = (cap + IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS - 1) /
            IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;
    tbl_size = IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

    for (int i = 0; i < num_of_blocks; ++i) {
        p_pkey_tbl = extended_info->getSMPPKeyTable(p_port->createIndex, i);
        if (!p_pkey_tbl)
            continue;

        if ((i + 1) * IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS  > cap)
            tbl_size = cap % IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

        for (int j = 0; j < tbl_size; ++j) {
            vec_pkeys[i * IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS + j].P_KeyBase =
                    p_pkey_tbl->PKey_Entry[j].P_KeyBase;
            vec_pkeys[i * IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS + j].Membership_Type =
                    p_pkey_tbl->PKey_Entry[j].Membership_Type;
        }
    }
    IBDIAG_RETURN_VOID;
}


static int comparePartitionKeys(map_pkey_membership &pkeys1,
        map_pkey_membership &pkeys2)
{
    IBDIAG_ENTER;
    map_pkey_membership::iterator it1, it2;
    for (it1 = pkeys1.begin(); it1 != pkeys1.end(); ++it1) {
        it2 = pkeys2.find((*it1).first);
        if (it2 == pkeys2.end() || (*it1) != (*it2))
            IBDIAG_RETURN(1);
    }
    for (it2 = pkeys2.begin(); it2 != pkeys2.end(); ++it2) {
        it1 = pkeys1.find((*it2).first);
        if (it1 == pkeys1.end() || (*it1) != (*it2))
            IBDIAG_RETURN(1);
    }
    IBDIAG_RETURN(0);
}


static string getPartitionKeysStr(map_pkey_membership &pkeys)
{
    IBDIAG_ENTER;
    map_pkey_membership::iterator it;
    string mem_desc, str = "[Pkey-MemType]: ";
    char buffer [1024];
    for (it = pkeys.begin(); it != pkeys.end(); ++it) {
        if (it != pkeys.begin())
            str += ";";
        switch ((*it).second) {
        case 0:
            mem_desc = "limited";
            break;
        case 1:
            mem_desc = "Full";
            break;
        default:
            snprintf(buffer, sizeof(buffer), "Unknown(" U8H_FMT ")", (*it).second);
            mem_desc.assign(buffer);
            break;
        }
        snprintf(buffer, sizeof(buffer), U8H_FMT"-%s", (*it).first, mem_desc.c_str());
        str += buffer;
    }
    IBDIAG_RETURN(str);
}


/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::BuildPKeyMap(map_pkey_p_port_membership& pkey_to_p_port_mem)
{ 
    IBDIAG_ENTER;
    u_int16_t partition_cap;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        SMP_NodeInfo *p_node_info = fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_node_info)
            continue;
        partition_cap = p_node_info->PartitionCap;

        unsigned int start_port = 1;
        unsigned int end_port = p_curr_node->numPorts;
        if (p_curr_node->type == IB_SW_NODE) {
            start_port = 0;
            end_port = 0;
        }

        // use unsigned int index to prevent overwrap
        for (unsigned int i = start_port; i <= end_port; ++i) {
            IBPort * p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            // don't skip switch port 0
            if (!p_curr_port || (i && !p_curr_port->is_data_worthy()))
                continue;

            map_pkey_membership pkey_to_mem;
            readPortPartitionTableToMap(&this->fabric_extended_info,
                    &IBDMExtendedInfo::getSMPPKeyTable,
                    p_curr_port->createIndex, partition_cap, pkey_to_mem);

            // add this port pkeys to map of pkey->port->membership
            for (map_pkey_membership::iterator it_pm = pkey_to_mem.begin();
                    it_pm != pkey_to_mem.end();
                    ++it_pm) {
                u_int16_t pkey = it_pm->first;
                u_int8_t mem = it_pm->second;

                map_p_port_membership p_port_to_mem;
                map_pkey_p_port_membership::iterator it_ppm = pkey_to_p_port_mem.find(pkey);
                if (it_ppm == pkey_to_p_port_mem.end()) {
                    p_port_to_mem.clear();
                    pkey_to_p_port_mem.insert(pair < u_int16_t, map_p_port_membership > (pkey, p_port_to_mem));
                }

                it_ppm = pkey_to_p_port_mem.find(pkey);
                map_p_port_membership::iterator it_pm1 = it_ppm->second.find(p_curr_port);
                if (it_pm1 == it_ppm->second.end()) {
                    it_ppm->second.insert(pair< IBPort*, u_int8_t >(p_curr_port, mem));
                }
            }
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpPartitionKeys(ofstream &sout,
                               const map_pkey_p_port_membership& pkey_to_p_port_mem)
{
    IBDIAG_ENTER;
    char buffer[2096];

    // finally dump all
    for (const auto& pkey_mem_map_pair : pkey_to_p_port_mem) {
        uint16_t curr_pkey = pkey_mem_map_pair.first;
        const map_p_port_membership& member_ports = pkey_mem_map_pair.second;
        long unsigned int printed_members = 0;
        this->discovered_fabric.unvisitAllAPorts();

        // count the number of members that will be printed (only one plane port per APort).
        for (const auto& port_mem_type_pair : member_ports) {
            const IBPort* curr_port = port_mem_type_pair.first;
            if (!curr_port || 
                   (curr_port->p_aport && curr_port->p_aport->visited))
                continue;
            if (curr_port->p_aport)
                curr_port->p_aport->visited = true;
            ++printed_members;
        }

        snprintf(buffer, sizeof(buffer), "GROUP PKey:" U16H_FMT " Hosts:%lu",
            curr_pkey,
            printed_members);
        sout << buffer << endl;

        // clear the visited bit before printing
        this->discovered_fabric.unvisitAllAPorts();

        for (const auto& port_mem_type_pair : member_ports) {
            string membership_str = "Limited";
            string port_name;
            const IBPort* curr_port = port_mem_type_pair.first;
            u_int8_t mem_type = port_mem_type_pair.second;
            if (mem_type)
                membership_str = "Full";
            
            if (!curr_port || 
                   (curr_port->p_aport && curr_port->p_aport->visited))
                continue;
            
            if (curr_port->p_aport) {
                curr_port->p_aport->visited = true;
                port_name = curr_port->p_aport->getName();
            } else
                port_name = curr_port->getName();

            snprintf(buffer, sizeof(buffer), "%s %s lid=" U16H_FMT " guid=" U64H_FMT " dev=%d",
                membership_str.c_str(),
                port_name.c_str(),
                curr_port->base_lid,
                curr_port->guid_get(),
                curr_port->p_node->devId);
            sout << "  " << buffer << endl;
        }
        sout << endl;
    }

    IBDIAG_RETURN_VOID;
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildPartitionKeysDB(list_p_fabric_general_err &pkey_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    u_int16_t partition_cap;
    u_int16_t partition_enf_cap = 0;
    int num_of_blocks_end_port = 0;
    int num_of_blocks_ext_port = 0;
    struct SMP_PKeyTable pkey_table;
    ProgressBarPorts progress_bar;

    CLEAR_STRUCT(pkey_table);

    ibDiagClbck.Set(this, &fabric_extended_info, &pkey_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPPkeyTableGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;


    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        p_curr_node->appData1.val = 0;

        SMP_NodeInfo *p_node_info =
                this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_node_info)
            continue;
        partition_cap = p_node_info->PartitionCap;
        phys_port_t start_port = 1;

        num_of_blocks_end_port =
                (partition_cap + IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS - 1) /
                IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

        if (p_curr_node->type == IB_SW_NODE) {
            SMP_SwitchInfo *p_switch_info =
                    this->fabric_extended_info.getSMPSwitchInfo(p_curr_node->createIndex);
            if (!p_switch_info)
                continue;
            partition_enf_cap = p_switch_info->PartEnfCap;
            start_port = 0;

            num_of_blocks_ext_port =
                    (partition_enf_cap + IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS - 1) /
                    IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;
        }

        for (phys_port_t i = start_port; i <= p_curr_node->numPorts; ++i) {
            phys_port_t curr_port = (phys_port_t)i;
            IBPort *p_curr_port = p_curr_node->getPort(curr_port);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            // don't skip switch port 0
            if (!p_curr_port || (i && !p_curr_port->is_data_worthy()))
                continue;

            u_int16_t num_of_blocks = (u_int16_t) ((i && p_curr_node->type == IB_SW_NODE) ? // external switch ports
                    num_of_blocks_ext_port : num_of_blocks_end_port);

            //get direct route for this port
            direct_route_t *p_curr_direct_route = this->GetDR(p_curr_port);
            if (!p_curr_direct_route) {
                this->SetLastError("DB error - can't find direct route to node=%s, port=%u",
                                   p_curr_node->getName().c_str(), (u_int8_t)p_curr_port->num);
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            for (u_int16_t j = 0; j < num_of_blocks; ++j) {
                clbck_data.m_data1 = p_curr_port;
                clbck_data.m_data2 = (void *)(uintptr_t)j;
                progress_bar.push(p_curr_port);
                this->ibis_obj.SMPPKeyTableGetByDirect(p_curr_direct_route,
                        curr_port, j, &pkey_table, &clbck_data);

                if (ibDiagClbck.GetState())
                    goto exit;
                if (p_curr_node->appData1.val != 0)
                    goto next_node;
            }    //block loop
        } //port loop
next_node:
        continue;
    }  //node loop
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!pkey_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::CheckAPortsPKeys(list_p_fabric_general_err& pkey_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    for (auto& sys_guid_p_aport :this->discovered_fabric.APortsBySysGuid) {
        if (CheckPlanarizedSystemPort0Pkeys(pkey_errors, sys_guid_p_aport.first) < 0)
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        for (APort* p_aport : sys_guid_p_aport.second) {
            if (!p_aport)   
                continue;
            if (CheckPortsPKeys(pkey_errors, p_aport->ports, p_aport->getName()) < 0)
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::CheckPortsPKeys(list_p_fabric_general_err& pkey_errors, const vec_pport& ports,
                    const string& ports_name)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    u_int16_t partition_cap;
    // map: pkey -> (membership, how many of the ports are members)
    map< u_int16_t, pair<uint8_t, uint32_t>> pkey_member_planes;

    // for comparing PartEnfInb and PartEnfOutb in between ports
    bool found_first_port_info = false, confErrReported = false;
    uint8_t firstPartEnfInb = 0;
    uint8_t firstPartEnfOutb = 0;

    // go over all the ports, add each pkey to pkey_member_planes,
    // check that all ports have the same PartEnfInb and PartEnfOutb
    for (const IBPort* p_curr_port : ports) {
        map_pkey_membership pkey_to_mem;
        if (!p_curr_port)
            continue;
        IBNode* p_curr_node = p_curr_port->p_node;
        if (!p_curr_node || !p_curr_port->is_data_worthy())
            continue;

        struct SMP_PortInfo * p_port_info =
            this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
        if (!p_port_info)
            continue;

        if (!found_first_port_info) {
            firstPartEnfInb = p_port_info->PartEnfInb;
            firstPartEnfOutb = p_port_info->PartEnfOutb;
            found_first_port_info = true;
        } else {
            // check that plane's configuration is the same as previous planes
            if (!confErrReported &&
                    (firstPartEnfInb != p_port_info->PartEnfInb ||
                    firstPartEnfOutb != p_port_info->PartEnfOutb)) {
                confErrReported = true;
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                pkey_errors.push_back(new APortWrongPKeyConf(ports_name));
            }
        }

        SMP_NodeInfo *p_node_info = fabric_extended_info.getSMPNodeInfo(
            p_curr_port->p_node->createIndex);
        if (!p_node_info)
            continue;

        partition_cap = p_node_info->PartitionCap;
        readPortPartitionTableToMap(&this->fabric_extended_info,
                &IBDMExtendedInfo::getSMPPKeyTable,
                p_curr_port->createIndex, partition_cap, pkey_to_mem);

        // increment counters for pkeys
        for (const auto& pkey_mem : pkey_to_mem) {
            if (!pkey_member_planes.count(pkey_mem.first))
                // add a new pkey to the membership list
                pkey_member_planes[pkey_mem.first] = make_pair(pkey_mem.second, 1);
            else {
                uint8_t prev_memshp = pkey_member_planes[pkey_mem.first].first;
                // check that this pkey's membership is the same as other planes
                if (prev_memshp != pkey_mem.second) {
                    rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                    pkey_errors.push_back(
                        new APortWrongPKeyMembership(ports_name, pkey_mem.first,
                                                        prev_memshp, pkey_mem.second));
                }
                ++pkey_member_planes[pkey_mem.first].second;
            }
        }
    }

    for (const auto& pkey_membership_count : pkey_member_planes) {
        u_int16_t pkey = pkey_membership_count.first;
        u_int32_t plane_count = pkey_membership_count.second.second;
        if (plane_count != ports.size() - 1) {
            // not all planes are members in this pkey
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            pkey_errors.push_back(
                new APortPlanesMissingPkey(ports_name, pkey));
        }
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckPlanarizedSystemPort0Pkeys(list_p_fabric_general_err& pkey_errors,
                    uint64_t system_guid)
{
    IBDIAG_ENTER;
    IBSystem *p_curr_system = this->discovered_fabric.getSystemByGuid(system_guid);
    vec_pport system_ports0;

    if (!p_curr_system)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    if (!p_curr_system->isMultiNodeSystem())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    // First entry of a ports vector must be null
    system_ports0.push_back(nullptr);
    for (auto& node_name_pair : p_curr_system->NodeByName) {
        IBNode *p_node = node_name_pair.second;
        if (!p_node || p_node->type != IB_SW_NODE) {
            continue;
        }
        
        IBPort *p_port0 = p_node->getPort(0);
        if (!p_port0) {
            continue;
        }

        system_ports0.push_back(p_port0);
    }

    stringstream ss;
    ss << p_curr_system->name << "/port0";

    IBDIAG_RETURN(CheckPortsPKeys(pkey_errors, system_ports0, ss.str()));
}

//Returns: SUCCESS_CODE / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckPartitionKeys(list_p_fabric_general_err& pkey_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    map_pkey_membership node_pkeys_map;
    map_pkey_membership switch_pkeys_map;

    //mark all ports as not checked
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)                   //port not exists
            continue;
        p_curr_port->counter1 = 0;
    }

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        // not checking switch-to-switch pkeys - TODO
        if (p_curr_node->type == IB_SW_NODE)
            continue;

        struct SMP_NodeInfo *p_curr_node_info = NULL;
        struct SMP_SwitchInfo *p_remote_switch_info = NULL;
        struct SMP_PortInfo *p_remote_port_info = NULL;

        for (unsigned int i = 1; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)
                continue;

            IBPort *p_remote_port = p_curr_port->p_remotePort;
            if (!p_remote_port)
                continue;

            if ((p_curr_port->counter1) || (p_curr_port->p_remotePort->counter1))   //already checked
                continue;
            //mark as checked
            ++p_curr_port->counter1;
            ++p_curr_port->p_remotePort->counter1;

            IBNode *p_remote_switch = p_remote_port->p_node;
            if (!p_remote_switch || p_remote_switch->type != IB_SW_NODE)
                continue;

            p_curr_node_info =
                    this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
            if (!p_curr_node_info)
                continue;
            p_remote_switch_info =
                    this->fabric_extended_info.getSMPSwitchInfo(p_remote_switch->createIndex);
            if (!p_remote_switch_info)
                continue;
            p_remote_port_info =
                    this->fabric_extended_info.getSMPPortInfo(p_remote_port->createIndex);
            if (!p_remote_port_info)
                continue;

            if (p_remote_switch->type == IB_SW_NODE &&
                    !(p_remote_port_info->PartEnfInb || p_remote_port_info->PartEnfOutb)) {
                continue;
            }

            u_int16_t node_partition_cap = p_curr_node_info->PartitionCap;
            u_int16_t switch_partition_cap = p_remote_switch_info->PartEnfCap;
            if (!switch_partition_cap)
                continue;

            node_pkeys_map.clear();
            readPortPartitionTableToMap(&this->fabric_extended_info,
                    &IBDMExtendedInfo::getSMPPKeyTable,
                    p_curr_port->createIndex, node_partition_cap, node_pkeys_map);

            // add all vports pkeys of this port to node_pkeys_map
            map_vportnum_vport::iterator VPortsIt = p_curr_port->VPorts.begin();

            for ( ; VPortsIt != p_curr_port->VPorts.end(); ++VPortsIt) {
                IBVPort *cur_vport = VPortsIt->second;

                if (!cur_vport) {
                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                               "-E- Unexpected NULL VPort in the VPorts map for Port GUID: "
                               U64H_FMT ", Port num: " U32D_FMT "\n",
                               p_curr_port->guid_get(), p_curr_port->num);
                    continue;
                }

                if (!cur_vport->m_p_vnode) {
                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                               "-E- Unexpected NULL VNode for VPort GUID: " U64H_FMT
                               ", VPort num: " U32D_FMT "\n",
                               cur_vport->guid_get(), cur_vport->getVPortNum());
                    continue;
                }

                SMP_VNodeInfo *cur_vnode_info =
                        this->fabric_extended_info.getSMPVNodeInfo(cur_vport->m_p_vnode->createIndex);
                if (!cur_vnode_info) {
                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                               "-E- Unexpected NULL VNode info for VNode GUID: " U64H_FMT "\n",
                               cur_vport->m_p_vnode->guid_get());
                    continue;
                }

                readPortPartitionTableToMap(&this->fabric_extended_info,
                        &IBDMExtendedInfo::getSMPVPortPKeyTable,
                        cur_vport->createIndex, cur_vnode_info->vpartition_cap, node_pkeys_map);
            }

            switch_pkeys_map.clear();
            readPortPartitionTableToMap(&this->fabric_extended_info,
                    &IBDMExtendedInfo::getSMPPKeyTable,
                    p_remote_port->createIndex, switch_partition_cap, switch_pkeys_map);
            if (comparePartitionKeys(switch_pkeys_map, node_pkeys_map)) {
                string pkey_str1 = getPartitionKeysStr(node_pkeys_map);
                string pkey_str2 = getPartitionKeysStr(switch_pkeys_map);

                pkey_errors.push_back(
                    new FabricErrPKeyMismatch(p_curr_port, p_remote_port, pkey_str1, pkey_str2));

                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        }
    }
    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WritePKeyFile(const string &file_name,
                          const map_pkey_p_port_membership &pkey_to_p_port_mem)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
    ofstream sout;
    int rc = this->OpenFile("Partition keys",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    IBFabric::GetSwitchLabelPortNumExplanation(sout);
    this->DumpPartitionKeys(sout, pkey_to_p_port_mem);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpPartitionKeysCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    char buffer[2096];
    u_int16_t partition_cap;
    u_int16_t partition_enf_cap = 0;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_PKEY))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,LocalPortNum,BlockNum,Index,PKey,Membership" << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        SMP_NodeInfo *p_node_info = fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_node_info)
            continue;
        partition_cap = p_node_info->PartitionCap;

        SMP_SwitchInfo *p_sw_info = fabric_extended_info.getSMPSwitchInfo(p_curr_node->createIndex);
        if (p_sw_info)
            partition_enf_cap = p_sw_info->PartEnfCap;

        unsigned int start_port = p_curr_node->type == IB_SW_NODE ? 0 : 1;

        for (unsigned int i = start_port; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            // don't skip switch port 0
            if (!p_curr_port || (i && !p_curr_port->is_data_worthy()))
                continue;

            vec_pkey_block_elem vec_pkeys;
            if (i && p_curr_node->type == IB_SW_NODE) { // external switch port
                readPortPartitionTableToVec(&this->fabric_extended_info,
                        p_curr_port, partition_enf_cap, vec_pkeys);
            } else {
                readPortPartitionTableToVec(&this->fabric_extended_info,
                        p_curr_port, partition_cap, vec_pkeys);
            }

            sstream.str("");
            for (u_int32_t idx = 0; idx < vec_pkeys.size(); ++idx) {
                u_int8_t membership = vec_pkeys[idx].Membership_Type;
                u_int16_t pkey = vec_pkeys[idx].P_KeyBase;
                if (!pkey)
                    continue;

                snprintf(buffer, sizeof(buffer), U64H_FMT ","	// node guid
                         U64H_FMT "," 	// port guid
                         U32D_FMT "," 	// port num
                         U32D_FMT "," 	// block num
                         U32D_FMT "," 	// index
                         U16H_FMT ",%d", // pkey, membership
                         p_curr_node->guid_get(),
                         p_curr_port->guid_get(),
                         i,
                         idx / 32,
                         idx % 32,
                         pkey,
                         membership);
                sstream << buffer << endl;
            }
            csv_out.WriteBuf(sstream.str());
        }
    }

    csv_out.DumpEnd(SECTION_PKEY);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

