/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (C) 2021-2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <infiniband/ibutils/output_stream_primitives.h>
#include <arpa/inet.h>

extern IBDiagClbck ibDiagClbck;

int AddTreeIDToQPNList(map_qpn_to_treeid &sharp_an_qpns_to_treeid,
                       u_int32_t qpn,
                       u_int16_t tree_index)
{
    IBDIAG_ENTER;

    map_qpn_to_treeid::iterator it = sharp_an_qpns_to_treeid.find(qpn);
    if (it == sharp_an_qpns_to_treeid.end())
        sharp_an_qpns_to_treeid.insert(pair<u_int32_t, u_int16_t >(qpn, tree_index));
    else
        IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/********SharpTreeEdge***********/
SharpTreeEdge::SharpTreeEdge(u_int32_t qp_num, u_int8_t qp_idx):
                             m_remote_tree_node(NULL), m_qpn(qp_num),
                             m_child_idx(qp_idx), m_qpc_config()
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}

/********SharpTreeNode***********/
SharpTreeNode::SharpTreeNode(SharpAggNode *aggNode,
                             u_int16_t treeid,
                             AM_TreeConfig &tree_config):
                             m_tree_id(treeid),
                             m_child_idx(0),
                             m_tree_config(tree_config),
                             m_agg_node(aggNode),
                             m_parent(NULL)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}

int SharpTreeNode::AddSharpTreeEdge(SharpTreeEdge *p_sharp_tree_eage,
                                    u_int8_t db_index)
{
    IBDIAG_ENTER;

    if (this->m_children.empty() || ((u_int8_t)this->m_children.size() < db_index + 1))
        this->m_children.resize(db_index + 1, NULL);

    if (!this->m_children[db_index]) {
        this->m_children[db_index] = p_sharp_tree_eage;
    }
    //TODO add error for inserting in bad child index

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

SharpTreeEdge * SharpTreeNode::GetSharpTreeEdge(u_int8_t db_index)
{
    IBDIAG_ENTER;
    //not exists
    if ((u_int8_t)m_children.size() < db_index + 1)
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(m_children[db_index]);
}

SharpTree* SharpTreeNode::GetRoot()
{
    IBDIAG_ENTER;

    SharpTreeEdge* cur_parent = m_parent;
    SharpTreeNode* cur_node = this;

    while (true) {

        if (!cur_parent)
            IBDIAG_RETURN(cur_node->GetSharpAggNode()->GetTreeRoot(m_tree_id));

        cur_node = cur_parent->GetRemoteTreeNode();

        if (!cur_node) // node is disconnected
            IBDIAG_RETURN(NULL);

        cur_parent = cur_node->GetSharpParentTreeEdge();
    }

    IBDIAG_RETURN(NULL);
}


/********SharpTree***********/
SharpTree::SharpTree(SharpTreeNode *root):
                     m_root(root), m_max_radix(0)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}

/********SharpAggNode***********/
SharpAggNode::SharpAggNode(IBPort * port):
                           m_port(port), m_an_info(), class_version(CLASS_VERSION_UNKNOWN)
{
    IBDIAG_ENTER;

    CLEAR_STRUCT(m_an_act_jobs);

    IBDIAG_RETURN_VOID;
}

SharpAggNode::~SharpAggNode()
{
    IBDIAG_ENTER;
    for (vector_p_sharp_tree_node::iterator it = this->m_trees.begin();
         it != this->m_trees.end();
         it++) {
       SharpTreeNode * p_sharp_tree_node = (*it);
       delete p_sharp_tree_node;
    }
    this->m_trees.clear();

    IBDIAG_RETURN_VOID;
}

bool SharpAggNode::IsPerfCounterSupported(counters_mode mode, sharp_pm_counters_bit bit)
{
    IBDIAG_ENTER;

    u_int32_t mask;
    switch (mode) {
        case CLU:
            mask = this->m_an_info.perf_clu_mask;
            break;
        case HBA:
            mask = this->m_an_info.perf_hba_mask;
            break;
        case AGGREGATED:
            mask = this->m_an_info.perf_clu_mask | this->m_an_info.perf_hba_mask;
            break;
        case NONE_MODE:
        default:
            mask = NONE_MODE_PERF_CNTR_MASK;
    }

    IBDIAG_RETURN(mask & (1 << (u_int32_t)bit) ? true : false);
}

void SharpAggNode::SetANInfo(struct AM_ANInfo *an_info)
{
    IBDIAG_ENTER;

    this->m_an_info = *an_info;
    this->m_trees.resize(this->m_an_info.tree_table_size, NULL);

    class_version = an_info->active_class_version ?
            an_info->active_class_version : (u_int8_t)DEFAULT_AM_CLASS_VERSION;

    IBDIAG_RETURN_VOID;
}

int SharpAggNode::AddSharpTreeNode(SharpTreeNode *p_sharp_tree_node,
                                   u_int16_t tree_index)
{
    IBDIAG_ENTER;

    if (this->m_trees.empty() || ((u_int16_t)this->m_trees.size() < tree_index + 1))
        this->m_trees.resize(tree_index + 1, NULL);

    if (!this->m_trees[tree_index]) {
        this->m_trees[tree_index] = p_sharp_tree_node;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

SharpTreeNode * SharpAggNode::GetSharpTreeNode(u_int16_t tree_index)
{
    IBDIAG_ENTER;
    //not exists
    if ((u_int16_t)this->m_trees.size() < tree_index + 1)
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->m_trees[tree_index]);
}

int SharpAggNode::AddTreeRoot(u_int16_t tree_id, SharpTreeNode *p_sharp_tree_node)
{
    IBDIAG_ENTER;

    if (this->m_roots.empty() || ((u_int16_t)this->m_roots.size() < tree_id + 1))
        this->m_roots.resize(tree_id + 1, NULL);

    if (!this->m_roots[tree_id]) {
        //make SharpTreeRoot
        SharpTree * p_sharp_tree = new SharpTree(p_sharp_tree_node);
        this->m_roots[tree_id] = p_sharp_tree;
    } else
        IBDIAG_RETURN(1);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

SharpTree* SharpAggNode::GetTreeRoot(u_int16_t tree_id)
{
    IBDIAG_ENTER;

    // not exists
    if ((u_int16_t)m_roots.size() < tree_id + 1)
        IBDIAG_RETURN(NULL);

    IBDIAG_RETURN(m_roots[tree_id]);
}


/********SharpMngr***********/
SharpMngr::SharpMngr(IBDiag * ibdiag, u_int8_t ver, bool ad_hoc):
                     m_fabric_max_trees_idx(0), m_ibdiag(ibdiag), version(ver), is_ad_hoc(ad_hoc)
{
    IBDIAG_ENTER;
    ibDiagClbck.SetSharpMngr(this);
    IBDIAG_RETURN_VOID;
}

SharpMngr::~SharpMngr()
{
    //clear sharp AN list
    list_sharp_an::iterator an_iter = m_sharp_an.begin();
    for ( ; an_iter != m_sharp_an.end(); an_iter++) {
        delete *an_iter;
    }

    // clear class port info map
    map_lid_to_class_ports_info::iterator cpi_iter = m_lid_to_class_port_info.begin();
    for ( ; cpi_iter!= m_lid_to_class_port_info.end(); cpi_iter++) {
        delete cpi_iter->second;
    }
}

int  SharpMngr::DiscoverSharpAggNodes(
        list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    struct IB_ClassPortInfo class_port_info;
    CLEAR_STRUCT(class_port_info);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SharpMngrClassPortInfoClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = m_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != m_ibdiag->GetDiscoverFabricPtr()->NodeByName.end();
            ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node) {
            m_ibdiag->SetLastError("DB error - found null node in NodeByName map"   \
                                   " for node = %s", nI->first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        // MAD should be sent only for Aggregation nodes
        if (!p_node->getInSubFabric() || p_node->getSpecialNodeType() != IB_SPECIAL_PORT_AN)
            continue;

        // go over all base lids of this node:
        for (u_int32_t i = 1; i < (u_int8_t)(p_node->numPorts + 1); ++i) {
            IBPort *p_port = p_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_port || (i != 0 && !p_port->is_data_worthy()))
                continue;

            clbck_data.m_data1 = p_node;
            clbck_data.m_data2 = p_port;
            progress_bar.push(p_port);
            m_ibdiag->GetIbisPtr()->AMClassPortInfoGet(p_port->base_lid,
                                                       DEFAULT_SL,
                                                       p_port->am_key.GetKey(),
                                                       &class_port_info,
                                                       &clbck_data);
            break;
        }
        if(ibDiagClbck.GetState())
            goto exit;
    }
exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("DiscoverSharpAggNodes Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "DiscoverSharpAggNodes Failed. \n");
    }

    IBDIAG_RETURN(rc);
}

int SharpMngr::BuildTreeConfigDB(
        list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    struct AM_TreeConfig tree_config;
    CLEAR_STRUCT(tree_config);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
                &forwardClbck<IBDiagClbck, &IBDiagClbck::SharpMngrTreeConfigClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;
    clbck_data.m_data4 = (void *)(uintptr_t)is_ad_hoc;

    // Go over all ANs
    for (list_sharp_an::iterator nI = this->m_sharp_an.begin(); nI != this->m_sharp_an.end(); ++nI) {

        SharpAggNode *p_sharp_aggnode = *nI;
        if (!p_sharp_aggnode) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        clbck_data.m_data1 = p_sharp_aggnode;
        // for every TreeID (till max of this AN) send TreeConfig
        for (u_int16_t tree_id = 0; tree_id < p_sharp_aggnode->GetANInfo().tree_table_size; tree_id++) {
            clbck_data.m_data2 = (void *)(uintptr_t)tree_id;
            tree_config.tree_id = tree_id;
            tree_config.num_of_children = MAX_CHILD_IDX_IN_TREE_CONFIG_MAD;

            progress_bar.push(p_sharp_aggnode->GetIBPort());
            m_ibdiag->GetIbisPtr()->AMTreeConfigGet(p_sharp_aggnode->GetIBPort()->base_lid,
                                                    DEFAULT_SL,
                                                    p_sharp_aggnode->GetIBPort()->am_key.GetKey(),
                                                    p_sharp_aggnode->getClassVersion(),
                                                    &tree_config,
                                                    &clbck_data);

            if(ibDiagClbck.GetState())
                goto exit;
        }
        if(ibDiagClbck.GetState())
           goto exit;
    }
exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("BuildSHARPAggMngrTreeConfig Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildSHARPAggMngrTreeConfig Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (ibDiagClbck.GetNumErrors() && !sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int  SharpMngr::BuildQPCConfigDB(
        list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    struct AM_QPCConfig qpc_config;
    CLEAR_STRUCT(qpc_config);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SharpMngrQPCConfigClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (u_int16_t tree_id = 0; tree_id <= this->m_fabric_max_trees_idx; tree_id++) {
        for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
             nI != this->m_sharp_an.end();
             ++nI) {

            SharpAggNode *p_sharp_agg_node = *nI;
            if (!p_sharp_agg_node) {
                m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            IBNode * p_node = p_sharp_agg_node->GetIBPort()->p_node;
            SharpTreeNode * p_sharp_tree_node = p_sharp_agg_node->GetSharpTreeNode(tree_id);
            if (!p_sharp_tree_node)
                continue;

            if (p_node->type == IB_SW_NODE)
                continue;

            for (u_int32_t i = 1; i < (u_int8_t)(p_node->numPorts + 1); ++i) {
                IBPort *p_port = p_node->getPort((phys_port_t)i);
                //Skip ports that don't exist or are down or have undefined state
                //or are not in the scope of the fabric.
                if (!p_port || (i != 0 && !p_port->is_data_worthy()))
                    continue;

                clbck_data.m_data1 = p_sharp_agg_node;

                SharpTreeEdge * p_edge_to_parent = p_sharp_tree_node->GetSharpParentTreeEdge();
                if (p_edge_to_parent) {
                    clbck_data.m_data2 = p_edge_to_parent;
                    qpc_config.qpn = p_edge_to_parent->GetQpn();

                    progress_bar.push(p_port);
                    m_ibdiag->GetIbisPtr()->AMQPCConfigGet(p_port->base_lid,
                                                           DEFAULT_SL,
                                                           p_port->am_key.GetKey(),
                                                           p_sharp_agg_node->getClassVersion(),
                                                           &qpc_config,
                                                           &clbck_data);

                    if(ibDiagClbck.GetState())
                        goto exit;
                }

                for (u_int8_t j = 0; j < p_sharp_tree_node->GetChildrenSize(); j++) {

                    SharpTreeEdge * p_sharp_tree_edge = p_sharp_tree_node->GetSharpTreeEdge(j);
                    if (!p_sharp_tree_edge) {
                        m_ibdiag->SetLastError("DB error - found null SharpTreeEdge in trees DB"    \
                                               " for child index = %d", j);
                        rc = IBDIAG_ERR_CODE_DB_ERR;
                        goto exit;
                    }

                    clbck_data.m_data2 = p_sharp_tree_edge;
                    qpc_config.qpn = p_sharp_tree_edge->GetQpn();

                    progress_bar.push(p_port);
                    m_ibdiag->GetIbisPtr()->AMQPCConfigGet(p_port->base_lid,
                                                           DEFAULT_SL,
                                                           p_port->am_key.GetKey(),
                                                           p_sharp_agg_node->getClassVersion(),
                                                           &qpc_config,
                                                           &clbck_data);

                    if(ibDiagClbck.GetState())
                        goto exit;
                }
                break;
            }
            if(ibDiagClbck.GetState())
                goto exit;
        }
    } // trees loop

exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("BuildQPCConfigDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildQPCConfigDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (ibDiagClbck.GetNumErrors() && !sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int SharpMngr::BuildSharpConfigurationDB(
        list_p_fabric_general_err &sharp_discovery_errors)
{

    if (!m_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    ibDiagClbck.Set(m_ibdiag, m_ibdiag->GetIBDMExtendedInfoPtr(), &sharp_discovery_errors);

    INFO_PRINT("Build SHARPAggMngrClassPortInfo\n");
    rc = DiscoverSharpAggNodes(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    if (rc) {
        ERR_PRINT("Failed to build AM Nodes DB.\n");
        IBDIAG_RETURN(rc);
    }

    INFO_PRINT("Discovered %u Aggregation Nodes.\n",
                       (int)m_sharp_supported_nodes.size());

    for (vector_ibnodes::iterator nI = this->m_sharp_supported_nodes.begin();
        nI != this->m_sharp_supported_nodes.end();
        ++nI) {
        IBNode *p_node = *nI;
        if (!p_node) {
            m_ibdiag->SetLastError("DB error - found null node in SharpSupportedNodes vector");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            IBDIAG_RETURN(rc);
        }
        // ibnodes is only HCAs
        for (u_int32_t i = 1; i < (u_int8_t)(p_node->numPorts + 1); ++i) {
            IBPort *p_port = p_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_port || (i != 0 && !p_port->is_data_worthy()))
                continue;
            SharpAggNode *p_sharp_agg_node = new SharpAggNode(p_port);

            this->m_sharp_an.push_back(p_sharp_agg_node);
            m_lid_to_sharp_agg_node.insert(pair<lid_t, SharpAggNode *>(p_port->base_lid,
                                                             p_sharp_agg_node));
            break;
        }
    }

    INFO_PRINT("Build SHARPANInfo\n");
    rc = BuildANInfoDB(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    if (rc) {
        ERR_PRINT("Failed to build AMInfo DB.\n");
        IBDIAG_RETURN(rc);
    }

    RemoveANsNotInVersion();

    INFO_PRINT("Build SHARPANActiveJobs\n");
    rc = BuildANActiveJobsDB(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    if (rc)
        ERR_PRINT("Failed to build ANActiveJobs DB.\n");

    ibDiagClbck.ResetState();

    INFO_PRINT("Build TreeConfigDB\n");
    rc = BuildTreeConfigDB(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    if (rc)
        ERR_PRINT("Failed to build AM TreeConfig DB.\n");

    ibDiagClbck.ResetState();

    INFO_PRINT("Build SHARPAggMngrQPCConfig\n");
    rc = BuildQPCConfigDB(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    if (rc)
        ERR_PRINT("Failed to build AM QPCConfig DB.\n");

    ibDiagClbck.ResetState();

    IBDIAG_RETURN(rc);
}

int  SharpMngr::BuildANInfoDB(
        list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    struct AM_ANInfo an_info;
    CLEAR_STRUCT(an_info);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SharpMngrANInfoClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        //This cannot be NULL as it is locally created and inserted.
        SharpAggNode *p_sharp_aggnode = *nI;

        IBPort *p_port = p_sharp_aggnode->GetIBPort();
        clbck_data.m_data1 = p_sharp_aggnode;

        progress_bar.push(p_port);
        m_ibdiag->GetIbisPtr()->AMANInfoGet(p_port->base_lid,
                                            DEFAULT_SL,
                                            p_port->am_key.GetKey(),
                                            m_lid_to_class_port_info[p_port->base_lid]->ClassVersion,
                                            &an_info,
                                            &clbck_data);

        rc = ibDiagClbck.GetState();
        if (rc)
            break;
    }

    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("BuildANInfoDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildANInfoDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (ibDiagClbck.GetNumErrors() && !sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

void SharpMngr::RemoveANsNotInVersion()
{
    IBDIAG_ENTER;

    // no specific version - use all ANs
    if (version == 0)
        IBDIAG_RETURN_VOID;

    list_sharp_an ans_to_remove;

    // collect all irrelevant ANs
    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
             nI != this->m_sharp_an.end();
             ++nI) {
        SharpAggNode *p_sharp_aggnode = *nI;

        if (p_sharp_aggnode->getClassVersion() != version) {
            ans_to_remove.push_back(p_sharp_aggnode);
        }
    }

    // remove all irrelevant ANs
    SCREEN_PRINT("\n");
    for (list_sharp_an::iterator nI = ans_to_remove.begin();
            nI != ans_to_remove.end();
            ++nI) {

        SharpAggNode *p_removed_an = *nI;

        INFO_PRINT("AN \'%s\' is ignored, not in specified version(%u). "
                "GUID=" U64H_FMT ", LID=%u, version=%u.\n",
                p_removed_an->GetIBPort()->p_node->description.c_str(),
                version,
                p_removed_an->GetIBPort()->guid_get(),
                p_removed_an->GetIBPort()->base_lid,
                p_removed_an->getClassVersion());

        lid_t an_lid = p_removed_an->GetIBPort()->base_lid;

        m_sharp_an.remove(p_removed_an);
        m_lid_to_sharp_agg_node.erase(an_lid);

        map_lid_to_class_ports_info::iterator it = m_lid_to_class_port_info.find(an_lid);
        delete it->second;
        m_lid_to_class_port_info.erase(it);

        delete p_removed_an;
    }

    if (!ans_to_remove.empty()) {
        INFO_PRINT("%lu ANs are ignored.\n", ans_to_remove.size());
    }

    IBDIAG_RETURN_VOID;
}

int  SharpMngr::BuildANActiveJobsDB(list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    struct AM_ANActiveJobs an_jobs;
    CLEAR_STRUCT(an_jobs);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SharpMngrANActiveJobsClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        //This cannot be NULL as it is locally created and inserted.
        SharpAggNode *p_sharp_aggnode = *nI;

        IBPort *p_port = p_sharp_aggnode->GetIBPort();
        clbck_data.m_data1 = p_sharp_aggnode;

        progress_bar.push(p_port);
        rc = m_ibdiag->GetIbisPtr()->AMANActiveJobsGet(p_port->base_lid,
                                                       DEFAULT_SL,
                                                       p_port->am_key.GetKey(),
                                                       p_sharp_aggnode->getClassVersion(),
                                                       &an_jobs,
                                                       &clbck_data);

        if (ibDiagClbck.GetState())
            goto exit;
    }

exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("BuildANActiveJobsDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildANActiveJobsDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (ibDiagClbck.GetNumErrors() && !sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

static int msbOffset(u_int32_t num)
{
    int msb = 0;
    while (num) {
        num >>= 1;
        msb++;
    }

    return msb;
}

int SharpMngr::VerifyTrapsLids(list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    // All ANs report to the same AM (trapLID is the same on all nodes.)
    set < lid_t > trap_lids;
    map_lid_to_class_ports_info::iterator cpi_iter = m_lid_to_class_port_info.begin();
    for ( ; cpi_iter!= m_lid_to_class_port_info.end(); cpi_iter++) {
        trap_lids.insert(cpi_iter->second->TrapLID);
    }

    if (trap_lids.size() > 1)
        sharp_discovery_errors.push_back(
                new SharpErrClassPortInfo("Not all ANs report traps to the same AM LID"));

    // If all ANs have same trapLID make sure that this is the LID of the AM (SM).
    sm_info_obj_t *cur_sm_master = NULL;
    for (list_p_sm_info_obj::iterator it =
            m_ibdiag->GetIBDMExtendedInfoPtr()->getSMPSMInfoListRef().begin();
            it != m_ibdiag->GetIBDMExtendedInfoPtr()->getSMPSMInfoListRef().end();
            ++it) {
        if ((*it)->smp_sm_info.SmState == IBIS_IB_SM_STATE_MASTER) {
            cur_sm_master = *it;
        }
    }

    if (trap_lids.size() == 1 &&
            cur_sm_master &&
            cur_sm_master->p_port->base_lid != *(trap_lids.begin())) {
        SharpErrClassPortInfo *p_curr_sharp_err =
                new SharpErrClassPortInfo("ANs don't report traps to master SM LID");

        p_curr_sharp_err->SetLevel(EN_FABRIC_ERR_WARNING);
        sharp_discovery_errors.push_back(p_curr_sharp_err);
    }

    IBDIAG_RETURN(rc);
}

int SharpMngr::VerifyVersions(list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    set < int > an_versions;
    set < int > sharp_versions;

    map_lid_to_sharpagg_node::iterator an_iter = m_lid_to_sharp_agg_node.begin();
    for ( ; an_iter!= m_lid_to_sharp_agg_node.end(); ++an_iter) {

        // all AN.class_ver <= class_port_info.ver
        if (an_iter->second->GetANInfo().active_class_version >
            m_lid_to_class_port_info[an_iter->first]->ClassVersion ) {

            sharp_discovery_errors.push_back(
                    new SharpErrInvalidActiveVer(an_iter->second->GetIBPort()->p_node));
        }

        // AN.active_class_version = X  --->  AN.active_sharp_version = X
        // control path and data path should be at the same version.
        int class_ver = an_iter->second->getClassVersion();
        int sharp_ver = an_iter->second->GetANInfo().active_sharp_version_bit_mask ?
                msbOffset(an_iter->second->GetANInfo().active_sharp_version_bit_mask) :
                DEFAULT_AM_CLASS_VERSION;

        if (class_ver != sharp_ver)
            sharp_discovery_errors.push_back(
                new SharpErrDiffVerMgmtAndSharp(an_iter->second->GetIBPort()->p_node,
                                                class_ver, sharp_ver));

        an_versions.insert(class_ver);
        sharp_versions.insert(sharp_ver);
    }

    // All ANs are at the same class version (control path)
    if (an_versions.size() > 1) {
        SharpErrVersions *p_curr_sharp_err =
                new SharpErrVersions("active_class_version is not the same on all ANs");

        p_curr_sharp_err->SetLevel(EN_FABRIC_ERR_WARNING);
        sharp_discovery_errors.push_back(p_curr_sharp_err);
    }

    // All ANs are at the same sharp version (data path)
    if (sharp_versions.size() > 1) {
        SharpErrVersions *p_curr_sharp_err =
                new SharpErrVersions("active_sharp_version is not the same on all ANs");

        p_curr_sharp_err->SetLevel(EN_FABRIC_ERR_WARNING);
        sharp_discovery_errors.push_back(p_curr_sharp_err);
    }

    IBDIAG_RETURN(rc);
}

int  SharpMngr::BuildPerformanceCountersDB(
        list_p_fabric_general_err &sharp_discovery_errors, bool per_port)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    struct AM_PerformanceCounters perf_cntr;
    CLEAR_STRUCT(perf_cntr);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_aggnode = *nI;
        if (!p_sharp_aggnode) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        IBPort *p_port = p_sharp_aggnode->GetIBPort();
        clbck_data.m_data1 = p_sharp_aggnode;

        const AM_ANInfo &am_an_info = p_sharp_aggnode->GetANInfo();
        if (per_port) {
            if (!am_an_info.ext_perf_counters_supported) {
                WARN_PRINT("FW version doesn't support extended Sharp Performance Counters"
                           " for Aggregation Node: " U64H_FMT "\n", p_port->p_node->guid_get());
                continue;
            }

            clbck_data.m_handle_data_func =
                      &forwardClbck<IBDiagClbck, &IBDiagClbck::SharpMngrHBAPerfCountersClbck>;

            if ((!p_port->p_remotePort) || (!p_port->p_remotePort->p_node))
                continue;
            IBNode *p_node = p_port->p_remotePort->p_node;

            for (u_int32_t i = 1; i <= p_node->numPorts; ++i) {
                 IBPort *p_curr_port = p_node->getPort((phys_port_t)i);
                 if (!p_curr_port)                       //port not exist
                     continue;

                 //Skip ports that are down or have undefined state
                 //or are not in the scope of the fabric.
                 if (!p_curr_port->is_data_worthy())
                     continue;

                 if (p_curr_port->isSpecialPort())
                     continue;

                 clbck_data.m_data2 = (void *)(uintptr_t)HBA;
                 clbck_data.m_data3 = p_curr_port;

                 progress_bar.push(p_port);
                 m_ibdiag->GetIbisPtr()->AMPerformanceCountersGet(p_port->base_lid,
                                                                  DEFAULT_SL,
                                                                  p_port->am_key.GetKey(),
                                                                  p_sharp_aggnode->getClassVersion(),
                                                                  HBA,
                                                                  &perf_cntr,
                                                                  &clbck_data,
                                                                  p_curr_port->num);

                 if(ibDiagClbck.GetState())
                    goto exit;
            }
        } else {
           clbck_data.m_handle_data_func =
                      &forwardClbck<IBDiagClbck, &IBDiagClbck::SharpMngrPerfCountersClbck>;

           progress_bar.push(p_port);
           // Backward compatibility check for old fw version
           if (am_an_info.ext_perf_counters_supported) {
               clbck_data.m_data2 = (void *)(uintptr_t)AGGREGATED;
               m_ibdiag->GetIbisPtr()->AMPerformanceCountersGet(p_port->base_lid,
                                                                DEFAULT_SL,
                                                                p_port->am_key.GetKey(),
                                                                p_sharp_aggnode->getClassVersion(),
                                                                AGGREGATED,
                                                                &perf_cntr,
                                                                &clbck_data);
           } else {
              clbck_data.m_data2 = (void *)(uintptr_t)NONE_MODE;
              m_ibdiag->GetIbisPtr()->AMPerformanceCountersGet(p_port->base_lid,
                                                               DEFAULT_SL,
                                                               p_port->am_key.GetKey(),
                                                               p_sharp_aggnode->getClassVersion(),
                                                               &perf_cntr,
                                                               &clbck_data);
           }

            if(ibDiagClbck.GetState())
                goto exit;
        }
    }

exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("BuildPerformanceCountersDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildPerformanceCountersDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (ibDiagClbck.GetNumErrors() && !sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int SharpMngr::ResetPerformanceCounters(list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    struct AM_PerformanceCounters perf_cntr;
    CLEAR_STRUCT(perf_cntr);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SharpMngrResetPerfCountersClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_aggnode = *nI;
        if (!p_sharp_aggnode) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        IBPort *p_port = p_sharp_aggnode->GetIBPort();
        clbck_data.m_data1 = p_sharp_aggnode;
        perf_cntr.counter_select = ALL_COUNTERS_SELECT;
        perf_cntr.counter_select2 = ALL_COUNTERS_SELECT;


        const AM_ANInfo &am_an_info = p_sharp_aggnode->GetANInfo();
        progress_bar.push(p_port);
        // Backward compatibility check for old fw version
        if (am_an_info.ext_perf_counters_supported)
            m_ibdiag->GetIbisPtr()->AMPerformanceCountersSet(p_port->base_lid,
                                                             DEFAULT_SL,
                                                             p_port->am_key.GetKey(),
                                                             p_sharp_aggnode->getClassVersion(),
                                                             AGGREGATED,
                                                             &perf_cntr,
                                                             &clbck_data);
        else
            m_ibdiag->GetIbisPtr()->AMPerformanceCountersSet(p_port->base_lid,
                                                             DEFAULT_SL,
                                                             p_port->am_key.GetKey(),
                                                             p_sharp_aggnode->getClassVersion(),
                                                             &perf_cntr,
                                                             &clbck_data);

        if(ibDiagClbck.GetState())
            goto exit;
    }

exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("ResetPerformanceCounters Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "ResetPerformanceCounters Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (!sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int SharpMngr::AddRootID(u_int16_t root_id)
{
    IBDIAG_ENTER;

    if (m_root_ids.find(root_id) == m_root_ids.end())
        m_root_ids.insert(root_id);
    else
        IBDIAG_RETURN(1);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SharpMngr::UpdateMaxRadixOnRoots()
{
    IBDIAG_ENTER;

    for (list_sharp_an::iterator an_it = this->m_sharp_an.begin();
             an_it != this->m_sharp_an.end(); ++an_it) {

        SharpAggNode *p_agg_node = *an_it;
        if (!p_agg_node)
            continue;

        // go over all Trees of this AN
        for (u_int16_t tree_idx = 0; tree_idx < p_agg_node->GetTreesSize(); tree_idx++) {

            SharpTreeNode* p_tree_node = p_agg_node->GetSharpTreeNode(tree_idx);
            if (!p_tree_node)
                continue;

            SharpTree* p_root = p_tree_node->GetRoot();
            if (p_root && p_root->GetMaxRadix() < p_tree_node->GetChildrenSize())
                p_root->SetMaxRadix(p_tree_node->GetChildrenSize());
        } // Trees
    } // ANs

    IBDIAG_RETURN_VOID;
}


int SharpMngr::ConnectTreeEdges(list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    // go over all ANs
    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
            nI != this->m_sharp_an.end();
            ++nI) {

        SharpAggNode *p_sharp_agg_node = *nI;
        if (!p_sharp_agg_node) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            IBDIAG_RETURN(rc);
        }

        IBNode* p_node = p_sharp_agg_node->GetIBPort()->p_node;

        // go over all Trees of this AN
        for (u_int16_t tree_idx = 0; tree_idx < p_sharp_agg_node->GetTreesSize(); tree_idx++) {

            SharpTreeNode* p_sharp_tree_node = p_sharp_agg_node->GetSharpTreeNode(tree_idx);
            if (!p_sharp_tree_node)
                continue;

            // go over all Edges of this Tree
            for (u_int8_t db_idx = 0; db_idx < p_sharp_tree_node->GetChildrenSize(); db_idx++) {

                SharpTreeEdge* p_sharp_tree_edge = p_sharp_tree_node->GetSharpTreeEdge(db_idx);
                if (!p_sharp_tree_edge)
                    continue;

                // get the remote AN for this Edge
                map_lid_to_sharpagg_node::iterator it =
                    this->m_lid_to_sharp_agg_node.find(p_sharp_tree_edge->GetQPCConfig().rlid);

                // remote AN not in map
                if (it == m_lid_to_sharp_agg_node.end()) {
                   IBPort *p_port = m_ibdiag->GetDiscoverFabricPtr()->
                       getPortByLid(p_sharp_tree_edge->GetQPCConfig().rlid);

                   if (!p_port ||
                           !p_port->p_node ||
                           p_port->p_node->type == IB_SW_NODE) {
                       SharpErrEdgeNodeNotFound* p_curr_fabric_node_err =
                           new SharpErrEdgeNodeNotFound(
                                   p_node, p_sharp_tree_edge->GetQPCConfig().rlid);

                       p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                       sharp_discovery_errors.push_back(p_curr_fabric_node_err);
                   }
                   continue;
                }

                SharpAggNode* remote_p_sharp_agg_node = it->second;
                if (!remote_p_sharp_agg_node) {
                    m_ibdiag->SetLastError("DB error - found null SharpAggNode"    \
                            " for lid = %d", p_sharp_tree_edge->GetQPCConfig().rlid);
                    rc = IBDIAG_ERR_CODE_DB_ERR;
                    IBDIAG_RETURN(rc);
                }

                // get the remote SharpTreeNode for this tree_idx of this AN
                SharpTreeNode* remote_p_sharp_tree_node =
                    remote_p_sharp_agg_node->GetSharpTreeNode(tree_idx);

                if (!remote_p_sharp_tree_node) {
                    sharp_discovery_errors.push_back(
                            new SharpErrTreeNodeNotFound(
                                    remote_p_sharp_agg_node->GetIBPort()->p_node, tree_idx));
                    continue;
                }

                p_sharp_tree_edge->SetRemoteTreeNode(remote_p_sharp_tree_node);
                remote_p_sharp_tree_node->SetChildIdx(p_sharp_tree_edge->GetChildIdx());

                if (!remote_p_sharp_tree_node->GetSharpParentTreeEdge()) {
                    sharp_discovery_errors.push_back(
                            new SharpErrParentTreeEdgeNotFound(
                                    remote_p_sharp_agg_node->GetIBPort()->p_node,
                                    p_sharp_tree_edge->GetQPCConfig().rlid,
                                    tree_idx));
                    continue;
                }

                if ((remote_p_sharp_tree_node->GetSharpParentTreeEdge()->GetQPCConfig().rqpn !=
                         p_sharp_tree_edge->GetQPCConfig().qpn) ||
                    (remote_p_sharp_tree_node->GetSharpParentTreeEdge()->GetQPCConfig().qpn !=
                         p_sharp_tree_edge->GetQPCConfig().rqpn)) {
                    sharp_discovery_errors.push_back(
                            new SharpErrMismatchParentChildQPNumber(
                                   remote_p_sharp_agg_node->GetIBPort()->p_node,
                                   p_sharp_agg_node->GetIBPort()->base_lid,
                                   p_sharp_tree_edge->GetQPCConfig().qpn,
                                   p_sharp_tree_edge->GetQPCConfig().rqpn,
                                   remote_p_sharp_agg_node->GetIBPort()->base_lid,
                                   remote_p_sharp_tree_node->GetSharpParentTreeEdge()->GetQPCConfig().qpn,
                                   remote_p_sharp_tree_node->GetSharpParentTreeEdge()->GetQPCConfig().rqpn,
                                   tree_idx));
                    continue;
                }

                if (remote_p_sharp_tree_node->GetSharpParentTreeEdge()->GetQPCConfig().rlid !=
                         p_sharp_agg_node->GetIBPort()->base_lid) {

                    sharp_discovery_errors.push_back(
                            new SharpErrMismatchParentChildQPConfig(
                                remote_p_sharp_agg_node->GetIBPort()->p_node,
                                p_sharp_tree_edge->GetQPCConfig().rlid,
                                p_sharp_agg_node->GetIBPort()->base_lid,
                                remote_p_sharp_tree_node->GetSharpParentTreeEdge()->GetQPCConfig().rlid,
                                tree_idx));
                    continue;
                }

                remote_p_sharp_tree_node->GetSharpParentTreeEdge()->
                       SetRemoteTreeNode(p_sharp_tree_node);

                //TODO add warning on remote SharpTreeNode not found for T

            }// Edges
        }// Trees
    }// ANs

    // after all edges are connected - update max radix on roots.
    this->UpdateMaxRadixOnRoots();

    IBDIAG_RETURN(rc);
}

void SharpMngr::ReportRemoteNodeDoesntExistErr(list_p_fabric_general_err &sharp_discovery_errors,
                                               int &rc, SharpAggNode *p_sharp_agg_node)
{
    sharp_discovery_errors.push_back(
            new SharpErrRemoteNodeDoesntExist(p_sharp_agg_node->GetIBPort()->p_node));
}

void SharpMngr::ReportQPCPortNotZeroErr(list_p_fabric_general_err &sharp_discovery_errors,
                                        int &rc, SharpAggNode *p_sharp_agg_node,
                                        u_int8_t qpc_port, u_int8_t an_port_sel_supported,
                                        SharpAggNode *p_remote_sharp_agg_node,
                                        u_int8_t remote_qpc_port,
                                        u_int8_t remote_an_port_sel_supported)
{
    sharp_discovery_errors.push_back(
            new SharpErrQPCPortNotZero(
                    p_sharp_agg_node->GetIBPort()->p_node, qpc_port,
                           an_port_sel_supported,
                           p_remote_sharp_agg_node->GetIBPort()->p_node, remote_qpc_port,
                           remote_an_port_sel_supported));
}

void SharpMngr::CheckQPCPortsAreValid(list_p_fabric_general_err &sharp_discovery_errors, int &rc,
                                      SharpAggNode *p_sharp_agg_node,
                                      SharpTreeEdge *p_sharp_tree_edge)
{
    SharpAggNode *p_remote_sharp_agg_node =
        p_sharp_tree_edge->GetRemoteTreeNode()->GetSharpAggNode();

    if (!p_remote_sharp_agg_node)
        return;

    u_int8_t an_port_sel_supported = p_sharp_agg_node->GetANInfo().qp_to_port_select_supported;
    u_int8_t remote_an_port_sel_supported =
        p_remote_sharp_agg_node->GetANInfo().qp_to_port_select_supported;

    u_int8_t qpc_port = p_sharp_tree_edge->GetQPCConfig().port;
    u_int8_t remote_qpc_port = p_sharp_tree_edge->GetRemoteTreeNode()->
        GetSharpParentTreeEdge()->GetQPCConfig().port;

    // If both ANs support QP port selection and both QPC ports are not zero, check the link exists
    if (an_port_sel_supported &&
            remote_an_port_sel_supported &&
            qpc_port &&
            remote_qpc_port) {
        IBNode *p_sw = p_sharp_agg_node->GetIBPort()->get_remote_node();
        if (!p_sw)
            this->ReportRemoteNodeDoesntExistErr(sharp_discovery_errors, rc, p_sharp_agg_node);
        IBNode *p_remote_sw = p_remote_sharp_agg_node->GetIBPort()->get_remote_node();
        if (!p_remote_sw)
            this->ReportRemoteNodeDoesntExistErr(sharp_discovery_errors, rc,
                                                 p_remote_sharp_agg_node);

        if (p_sw && p_remote_sw) {
            IBPort *p_sw_port = p_sw->getPort(qpc_port);
            IBPort *p_remote_sw_port = p_remote_sw->getPort(remote_qpc_port);

            bool check = p_sw_port && p_remote_sw_port &&
                         p_sw_port->p_remotePort == p_remote_sw_port &&
                         p_remote_sw_port->p_remotePort == p_sw_port;

            if (!check)
                sharp_discovery_errors.push_back(
                        new SharpErrQPCPortsNotConnected(
                                p_sharp_agg_node->GetIBPort()->p_node,
                                     qpc_port,
                                     p_remote_sharp_agg_node->GetIBPort()->p_node,
                                     remote_qpc_port));
        }
    } else { // Othewise, both QPC ports should be 0
        if (qpc_port)
            this->ReportQPCPortNotZeroErr(sharp_discovery_errors, rc, p_sharp_agg_node, qpc_port,
                                          an_port_sel_supported, p_remote_sharp_agg_node,
                                          remote_qpc_port, remote_an_port_sel_supported);
        if (remote_qpc_port)
            this->ReportQPCPortNotZeroErr(sharp_discovery_errors, rc, p_remote_sharp_agg_node,
                                          remote_qpc_port, remote_an_port_sel_supported,
                                          p_sharp_agg_node, qpc_port, an_port_sel_supported);
    }
}

/*
 1. checking disconnected node
 2. checking duplicate qpn in sharp node
 3. checking rqpn is valid

TODO: check loops in the tree for nodes on the same tree_id
 */
int SharpMngr::CheckSharpTrees(list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    // checking disconnected nodes on the tree that are not root
    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_agg_node = *nI;
        if (!p_sharp_agg_node) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        IBNode * p_node = p_sharp_agg_node->GetIBPort()->p_node;
        map_qpn_to_treeid sharp_an_qpns_to_treeid;

        for (u_int16_t tree_idx = 0; tree_idx < p_sharp_agg_node->GetTreesSize(); tree_idx++) {

            SharpTreeNode * p_sharp_tree_node = p_sharp_agg_node->GetSharpTreeNode(tree_idx);
            if (!p_sharp_tree_node)
                continue;
            SharpTreeEdge * p_parent_tree_edge =
                p_sharp_tree_node->GetSharpParentTreeEdge();

            //if not root and remote parent is null
            if (p_parent_tree_edge &&
                    p_parent_tree_edge->GetQpn() &&
                    !p_parent_tree_edge->GetRemoteTreeNode()) {

                SharpErrDisconnectedTreeNode *p_curr_fabric_node_err =
                    new SharpErrDisconnectedTreeNode(p_node,
                                                     tree_idx,
                                                     p_parent_tree_edge->GetQpn(),
                                                     p_parent_tree_edge->GetQPCConfig().rlid);

                p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                sharp_discovery_errors.push_back(p_curr_fabric_node_err);
            }

            // Check QP state
            if (p_parent_tree_edge &&
                p_parent_tree_edge->GetQPCConfig().state != QP_STATE_ACTIVE) {
                SharpErrQPNotActive *p_curr_fabric_node_err =
                    new SharpErrQPNotActive(
                            p_node, p_parent_tree_edge->GetQpn(),
                            p_parent_tree_edge->GetQPCConfig().state);

                p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                sharp_discovery_errors.push_back(p_curr_fabric_node_err);
            }

            for (u_int8_t child_idx = 0;
                 child_idx < p_sharp_tree_node->GetChildrenSize();
                 child_idx++) {

                SharpTreeEdge * p_sharp_tree_edge =
                    p_sharp_tree_node->GetSharpTreeEdge(child_idx);
                if (!p_sharp_tree_edge)
                    continue;

                // Check there are no duplicate qpns
                if (AddTreeIDToQPNList(sharp_an_qpns_to_treeid,
                                       p_sharp_tree_edge->GetQpn(),
                                       tree_idx))
                    sharp_discovery_errors.push_back(
                            new SharpErrDuplicatedQPNForAggNode(
                                p_node, tree_idx,
                                sharp_an_qpns_to_treeid[p_sharp_tree_edge->GetQpn()],
                                p_sharp_tree_edge->GetQpn()));

                // Check QP state
                if (p_sharp_tree_edge->GetQPCConfig().state != QP_STATE_ACTIVE)
                    sharp_discovery_errors.push_back(
                            new SharpErrQPNotActive(
                                p_node, p_sharp_tree_edge->GetQpn(),
                                p_sharp_tree_edge->GetQPCConfig().state));

                if (p_sharp_tree_edge->GetRemoteTreeNode() &&
                    p_sharp_tree_edge->GetRemoteTreeNode()->GetSharpParentTreeEdge()) {

                    // Check qpn and rqpn are the same for the opposite edges
                    u_int32_t rqpn = p_sharp_tree_edge->GetQPCConfig().rqpn;
                    u_int32_t peer_qpn =
                        p_sharp_tree_edge->GetRemoteTreeNode()->GetSharpParentTreeEdge()->GetQpn();
                    if (rqpn != peer_qpn)
                        sharp_discovery_errors.push_back(
                                new SharpErrRQPNotValid(p_node, rqpn, peer_qpn));

                    u_int32_t qpn = p_sharp_tree_edge->GetQpn();
                    u_int32_t peer_rqpn = p_sharp_tree_edge->
                        GetRemoteTreeNode()->GetSharpParentTreeEdge()->GetQPCConfig().rqpn;
                    if (peer_rqpn != qpn)
                        sharp_discovery_errors.push_back(
                                new SharpErrRQPNotValid(p_node, qpn, peer_rqpn));

                    // Check physical link exists between switches aggregation nodes are connected to
                    this->CheckQPCPortsAreValid(sharp_discovery_errors, rc, p_sharp_agg_node,
                                                p_sharp_tree_edge);
                }
            }
        }
    }

exit:
    IBDIAG_RETURN(rc);
}

void SharpTreeNode::DumpTree(int indent_level, ofstream &sout)
{
    IBDIAG_ENTER;
    char buffer[256];
    string indent_str = "";
    int radix, parent_qpn = 0, remote_parent_qpn = 0;

    if (!this->m_agg_node || !this->m_agg_node->GetIBPort() || !this->m_agg_node->GetIBPort()->p_node)
        //TODO add error
        IBDIAG_RETURN_VOID;

    for (int indent = 0; indent < indent_level; indent++)
        indent_str.append("\t");

    IBPort *p_curr_port = this->m_agg_node->GetIBPort();
    IBNode *p_curr_node = p_curr_port->p_node;
    IBNode *p_remote_sw = p_curr_port->p_remotePort->p_node;
    sout << indent_str;
    snprintf(buffer, sizeof(buffer),
             "(" U32D_FMT "), AN: node guid: " U64H_FMT ", \"" STR_FMT "\", lid:" U32D_FMT
             ", port guid:" U64H_FMT
             ", switch guid: " U64H_FMT ", \"" STR_FMT "\""
             ", Child index:" U32D_FMT,
             indent_level,
             p_curr_node->guid_get(),
             p_curr_node->description.c_str(),
             this->m_agg_node->GetIBPort()->base_lid,
             p_curr_port->guid_get(),
             p_remote_sw->guid_get(),
             p_remote_sw->description.c_str(),
             this->m_child_idx);
    sout << buffer;

    //TODO no need for p_parent_sharp_tree_edge pointer just use this->m_parent
    SharpTreeEdge * p_parent_sharp_tree_edge = this->m_parent;
    if (this->m_parent) {
        parent_qpn = p_parent_sharp_tree_edge->GetQPCConfig().qpn;
        remote_parent_qpn = p_parent_sharp_tree_edge->GetQPCConfig().rqpn;
    }

    radix = GetChildrenSize();
    snprintf(buffer, sizeof(buffer),
             ", parent QPn:" U32H_FMT ", remote parent QPn:" U32H_FMT
             ", radix:" U32D_FMT,
             parent_qpn,
             remote_parent_qpn,
             radix);
    sout << buffer << endl;

    for (u_int8_t child_idx = 0; child_idx < this->GetChildrenSize(); ++child_idx) {
        SharpTreeEdge * p_sharp_tree_edge = this->GetSharpTreeEdge(child_idx);
        if (!p_sharp_tree_edge)
            continue;

        if (p_sharp_tree_edge->GetRemoteTreeNode())
            p_sharp_tree_edge->GetRemoteTreeNode()->DumpTree(indent_level+1, sout);
    }

    IBDIAG_RETURN_VOID;
}

int SharpMngr::WriteSharpFile(const string &file_name)
{
    IBDIAG_ENTER;

    ofstream sout;
    int rc = m_ibdiag->OpenFile("Sharp Data",
                                file_name, sout,
                                false, //no append
                                true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->SharpMngrDumpAllTrees(sout);
    this->SharpMngrDumpAllQPs(sout);

    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int SharpMngr::DumpSharpANInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    stringstream sstream;
    csv_out.WriteBuf(sstream.str());

    if (csv_out.DumpStart(SECTION_SHARP_AN_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    // Write header
    sstream.str("");
    sstream << "GUID,LID,"
            << "endianness,enable_endianness_per_job,reproducibility_disable,"
            << "enable_reproducibility_per_job,tree_job_default_binding,"
            << "active_class_version,"
            << "an_sat_qp_info_supported,an_semaphore_info_supported,job_key_supported,"
            << "tree_job_binding_supported,port_credit_alloc_supported,hba_vl_counters_supported,"
            << "hba_sl_counters_supported,ext_perf_counters_supported,half_buffer_line_optimization_supported,"
            << "reproducibility_per_job_supported,reproducibility_disable_supported,root_gid_group_join,"
            << "packet_based_credits_supported,endianess_per_job_supported,"
            << "multiple_sver_active_supported,streaming_aggregation_supported,"
            << "tree_table_size,internal_precision_support,sat_rmc_operation_supported,"
            << "llt_sat_groups_size_supported,is_hba_timeout_config_supported,"
            << "is_user_qkey_allowed,is_stochastic_rounding_supported,"
            << "qp_to_port_select_supported,tree_radix,tree_radix_used,"
            << "sharp_version_supported_bit_mask,active_sharp_version_bit_mask,"
            << "group_table_size,gt_mode,max_group_num,"
            << "outstanding_operation_table_size,max_aggregation_payload,"
            << "num_semaphores,streaming_aggregation_outstanding_operation,"
            << "operation_buffer_size,num_of_jobs,max_num_qps,max_llt_qps,"
            << "line_size,worst_case_num_lines,num_lines_chunk_mode,"
            << "perf_clu_mask,perf_hba_mask,perf_hba_split_port_mask,"
            << "qp_perf_hba_mask,qp_perf_clu_mask,data_types_supported,mtu_support,"
            << "max_sat_group_table_size,max_llt_group_table_size,"
            << "hba_timeout_config" << endl;
    csv_out.WriteBuf(sstream.str());

    for (list_sharp_an::iterator an_iter = m_sharp_an.begin(); an_iter != m_sharp_an.end(); ++an_iter) {

        SharpAggNode *p_sharp_aggnode = *an_iter;
        AM_ANInfo cur_an = p_sharp_aggnode->GetANInfo();
        IBNode *p_node = p_sharp_aggnode->GetIBPort()->p_node;

        char buffer[24] = {};
        snprintf(buffer, sizeof(buffer), U64H_FMT, p_node->guid_get());

        sstream.str("");
        sstream << buffer << ","
                << p_sharp_aggnode->GetIBPort()->base_lid << ","
                << +cur_an.endianness << ","
                << +cur_an.enable_endianness_per_job      << ","
                << +cur_an.reproducibility_disable        << ","
                << +cur_an.enable_reproducibility_per_job << ","
                << +cur_an.tree_job_default_binding       << ","
                << +cur_an.active_class_version << ","
                << +cur_an.an_sat_qp_info_supported << ","
                << +cur_an.an_semaphore_info_supported << ","
                << +cur_an.job_key_supported << ","
                << +cur_an.tree_job_binding_supported << ","
                << +cur_an.port_credit_alloc_supported << ","
                << +cur_an.hba_vl_counters_supported << ","
                << +cur_an.hba_sl_counters_supported << ","
                << +cur_an.ext_perf_counters_supported << ","
                << +cur_an.half_buffer_line_optimization_supported << ","
                << +cur_an.reproducibility_per_job_supported << ","
                << +cur_an.reproducibility_disable_supported << ","
                << +cur_an.root_gid_group_join << ","
                << +cur_an.packet_based_credits_supported << ","
                << +cur_an.endianess_per_job_supported << ","
                << +cur_an.multiple_sver_active_supported << ","
                << +cur_an.streaming_aggregation_supported << ","
                << +cur_an.tree_table_size << ","
                << +cur_an.internal_precision_support << ","
                << +cur_an.sat_rmc_operation_supported << ","
                << +cur_an.llt_sat_groups_size_supported << ","
                << +cur_an.is_hba_timeout_config_supported << ","
                << +cur_an.is_user_qkey_allowed << ","
                << +cur_an.is_stochastic_rounding_supported << ","
                << +cur_an.qp_to_port_select_supported << ","
                << +cur_an.tree_radix << ","
                << +cur_an.tree_radix_used << ","
                << "0x" << hex << cur_an.sharp_version_supported_bit_mask << dec << ","
                << "0x" << hex << cur_an.active_sharp_version_bit_mask << dec << ","
                << +cur_an.group_table_size << ","
                << +cur_an.gt_mode << ","
                << +cur_an.max_group_num << ","
                << +cur_an.outstanding_operation_table_size << ","
                << +cur_an.max_aggregation_payload << ","
                << +cur_an.num_semaphores << ","
                << +cur_an.streaming_aggregation_outstanding_operation << ","
                << +cur_an.operation_buffer_size << ","
                << +cur_an.num_of_jobs << ","
                << +cur_an.max_num_qps << ","
                << +cur_an.max_llt_qps << ","
                << +cur_an.line_size << ","
                << +cur_an.worst_case_num_lines << ","
                << +cur_an.num_lines_chunk_mode << ","
                << +cur_an.perf_clu_mask << ","
                << +cur_an.perf_hba_mask << ","
                << +cur_an.perf_hba_split_port_mask << ","
                << +cur_an.qp_perf_hba_mask << ","
                << +cur_an.qp_perf_clu_mask << ","
                << +cur_an.data_types_supported << ","
                << +cur_an.mtu_support << ","
                << +cur_an.max_sat_group_table_size << ","
                << +cur_an.max_llt_group_table_size << ","
                << +cur_an.hba_timeout_config << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_SHARP_AN_INFO);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

string activeJobsToStr(AM_ANActiveJobs &an_jobs)
{
    stringstream sstream;
    const int bits_in_dword = 32;

    for (int dword = 0; dword < AN_ACTIVE_JOBS_DWORDS_NUM; ++dword) {
        u_int32_t cur = an_jobs.active_jobs[dword];
        for (int job_offset = 0; job_offset < bits_in_dword && cur; ++job_offset) {
            if (cur & 1)
                sstream << job_offset + (dword * bits_in_dword) << ',';

            cur >>= 1;
        }
    }

    string res = sstream.str();
    if (res.size() == 0)
        res = "no jobs";

    return res;
}

int SharpMngr::WriteSharpANInfoFile(const string &file_name)
{
    IBDIAG_ENTER;

    ofstream sout;
    int rc = m_ibdiag->OpenFile("Sharp AN info",
                                file_name,
                                sout,
                                false, //no append
                                true); //add_header
    if (rc)
          IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    // write comment about bitset 1 and bitset 2.
    stringstream sstream;
    sout << sstream.str() << endl;

    for (list_sharp_an::iterator an_iter = m_sharp_an.begin(); an_iter != m_sharp_an.end(); ++an_iter) {

        SharpAggNode *p_sharp_aggnode = *an_iter;
        AM_ANInfo cur_an = p_sharp_aggnode->GetANInfo();
        IBNode *p_node = p_sharp_aggnode->GetIBPort()->p_node;
        AM_ANActiveJobs cur_an_jobs = p_sharp_aggnode->GetANActiveJobs();
        IBNode *p_remote_sw = p_sharp_aggnode->GetIBPort()->p_remotePort->p_node;

        char buffer[24] = {};
        snprintf(buffer, sizeof(buffer), U64H_FMT, p_node->guid_get());

        sout << setfill('-') << setw(91) << "" << setfill(' ') << endl
             << "GUID=" << buffer << ", LID=" << p_sharp_aggnode->GetIBPort()->base_lid
             << ", SW_GUID=" << HEX(p_remote_sw->guid_get()) << ", "
             << QUOTED(p_remote_sw->description) << endl
             << setfill('-') << setw(91) << "" << setfill(' ') << endl

             << "endianness = " << +cur_an.endianness << endl
             << "enable_endianness_per_job = " << +cur_an.enable_endianness_per_job << endl
             << "reproducibility_disable = " << +cur_an.reproducibility_disable << endl
             << "enable_reproducibility_per_job = " << +cur_an.enable_reproducibility_per_job << endl
             << "tree_job_default_binding = " << +cur_an.tree_job_default_binding << endl
             << "active_class_version = " << +cur_an.active_class_version << endl
             << "an_sat_qp_info_supported = " << +cur_an.an_sat_qp_info_supported << endl
             << "an_semaphore_info_supported = " << +cur_an.an_semaphore_info_supported << endl
             << "job_key_supported = " << +cur_an.job_key_supported << endl
             << "tree_job_binding_supported = " << +cur_an.tree_job_binding_supported << endl
             << "port_credit_alloc_supported = " << +cur_an.port_credit_alloc_supported << endl
             << "hba_vl_counters_supported = " << +cur_an.hba_vl_counters_supported << endl
             << "hba_sl_counters_supported = " << +cur_an.hba_sl_counters_supported << endl
             << "ext_perf_counters_supported = " << +cur_an.ext_perf_counters_supported << endl
             << "half_buffer_line_optimization_supported = " << +cur_an.half_buffer_line_optimization_supported << endl
             << "reproducibility_per_job_supported = " << +cur_an.reproducibility_per_job_supported << endl
             << "reproducibility_disable_supported = " << +cur_an.reproducibility_disable_supported << endl
             << "root_gid_group_join = " << +cur_an.root_gid_group_join << endl
             << "packet_based_credits_supported = " << +cur_an.packet_based_credits_supported << endl
             << "endianess_per_job_supported = " << +cur_an.endianess_per_job_supported << endl
             << "endianess_per_job_supported = " << +cur_an.multiple_sver_active_supported << endl
             << "streaming_aggregation_supported = " << +cur_an.streaming_aggregation_supported << endl
             << "sat_rmc_operation_supported = " << +cur_an.sat_rmc_operation_supported << endl
             << "llt_sat_groups_size_supported = "   << +cur_an.llt_sat_groups_size_supported << endl
             << "is_hba_timeout_config_supported = " << +cur_an.is_hba_timeout_config_supported << endl
             << "is_user_qkey_allowed = "  << +cur_an.is_user_qkey_allowed << endl
             << "is_stochastic_rounding_supported = " << +cur_an.is_stochastic_rounding_supported << endl
             << "qp_to_port_select_supported = " << +cur_an.qp_to_port_select_supported << endl
             << "tree_table_size = " << +cur_an.tree_table_size << endl
             << "internal_precision_support = " << +cur_an.internal_precision_support << endl
             << "tree_radix = " << +cur_an.tree_radix << endl
             << "tree_radix_used = " << +cur_an.tree_radix_used << endl
             << "sharp_version_supported_bit_mask = "
                 << "0x" << hex << cur_an.sharp_version_supported_bit_mask << dec <<endl
             << "active_sharp_version_bit_mask = "
                 << "0x" << hex << cur_an.active_sharp_version_bit_mask << dec << endl
             << "group_table_size = " << +cur_an.group_table_size << endl
             << "gt_mode = " << +cur_an.gt_mode << endl
             << "max_group_num = " << +cur_an.max_group_num << endl
             << "outstanding_operation_table_size = " << +cur_an.outstanding_operation_table_size << endl
             << "max_aggregation_payload = " << +cur_an.max_aggregation_payload << endl
             << "num_semaphores = " << +cur_an.num_semaphores << endl
             << "streaming_aggregation_outstanding_operation = "
                 << +cur_an.streaming_aggregation_outstanding_operation << endl
             << "operation_buffer_size = " << +cur_an.operation_buffer_size << endl
             << "num_of_jobs = " << +cur_an.num_of_jobs << endl
             << "max_num_qps = " << +cur_an.max_num_qps << endl
             << "max_llt_qps = " << +cur_an.max_llt_qps << endl
             << "line_size = " << +cur_an.line_size << endl
             << "worst_case_num_lines = " << +cur_an.worst_case_num_lines << endl
             << "num_lines_chunk_mode = " << +cur_an.num_lines_chunk_mode << endl
             << "perf_clu_mask = " << +cur_an.perf_clu_mask << endl
             << "perf_hba_mask = " << +cur_an.perf_hba_mask << endl
             << "perf_hba_split_port_mask = " << +cur_an.perf_hba_split_port_mask << endl
             << "qp_perf_hba_mask = " << +cur_an.qp_perf_hba_mask << endl
             << "qp_perf_clu_mask = " << +cur_an.qp_perf_clu_mask << endl
             << "qp_perf_clu_mask = " << +cur_an.qp_perf_clu_mask << endl
             << "data_types_supported = " << +cur_an.data_types_supported << endl
             << "mtu_support = " << +cur_an.mtu_support << endl
             << "max_sat_group_table_size = " << +cur_an.max_sat_group_table_size << endl
             << "max_llt_group_table_size = " << +cur_an.max_llt_group_table_size << endl
             << "hba_timeout_config = " << +cur_an.hba_timeout_config << endl
             // add also the active jobs on this AN
             << "active_jobs = " << activeJobsToStr(cur_an_jobs) << endl
             << endl;
    }

    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SharpMngr::SharpMngrDumpAllTrees(ofstream &sout)
{
    IBDIAG_ENTER;

    for (list_sharp_an::iterator an_it = this->m_sharp_an.begin();
             an_it != this->m_sharp_an.end(); ++an_it) {

        SharpAggNode *p_agg_node = *an_it;
        if (!p_agg_node)
            continue;

        const vector_sharp_root_node& root_nodes = p_agg_node->GetSharpRootNodeVec();

        for (vector_sharp_root_node::size_type tree_idx = 0; tree_idx < root_nodes.size(); ++tree_idx) {

            SharpTree* root_node = root_nodes[tree_idx];

            if (!root_node || !root_node->GetRoot())
                continue;

            sout << "Root LID:"  << p_agg_node->GetIBPort()->base_lid << ", "
                 << "TreeID:"    << tree_idx << ", "
                 << "Max Radix:" << root_node->GetMaxRadix() << ", "
                 << "Type:"      << (root_node->GetRoot()->GetTreeConfig().tree_mode ? "SAT" : "LLT");

            // For SAT trees add the LLT tree
            if (root_node->GetRoot()->GetTreeConfig().tree_mode == 1) {
                sout << ", LLT Tree ID:" << root_node->GetRoot()->GetTreeConfig().small_data_tree_id;
            }

            sout << endl;

            root_node->GetRoot()->DumpTree(0, sout);
            sout << endl;

        } // roots
    } // ANs

    IBDIAG_RETURN_VOID;
}

void SharpMngr::DumpQPC(ofstream &sout, struct AM_QPCConfig *qpconfig)
{
    IBDIAG_ENTER;
    char buffer[512];
    char gid[INET6_ADDRSTRLEN];
    if (!qpconfig)
        IBDIAG_RETURN_VOID;

    snprintf(buffer, sizeof(buffer),
             "QPN:" U32H_FMT ", State:" U32D_FMT ", packet_based_credit_req_en:" U32D_FMT
             ", packet_based_credit_resp_en:" U32D_FMT ", TS:" U32H_FMT
             ", G:" U32D_FMT ", SL:" U32D_FMT ", RLID:" U32D_FMT
             ", Traffic Class:" U32D_FMT ", Hop Limit:" U32D_FMT ", RGID:" STR_FMT
             ", RQ PSN:" U32D_FMT ", SQ PSN:" U32D_FMT ", PKey:" U32H_FMT
             ", RQPN:" U32H_FMT ", QKey:" U32D_FMT ", Port:" U32D_FMT
             ", RNR Mode:" U32D_FMT ", RNR Retry Limit:" U32H_FMT
             ", Timeout Retry Limit:" U32D_FMT ", Local Ack Timeout:" U32D_FMT
             ,
             qpconfig->qpn,
             qpconfig->state,
             qpconfig->packet_based_credit_req_en,
             qpconfig->packet_based_credit_resp_en,
             qpconfig->ts,
             qpconfig->g,
             qpconfig->sl,
             qpconfig->rlid,
             qpconfig->traffic_class,
             qpconfig->hop_limit,
             inet_ntop(AF_INET6, qpconfig->rgid, gid, (unsigned int)sizeof gid),
             qpconfig->rq_psn,
             qpconfig->sq_psn,
             qpconfig->pkey,
             qpconfig->rqpn,
             qpconfig->qkey,
             qpconfig->port,
             qpconfig->rnr_mode,
             qpconfig->rnr_retry_limit,
             qpconfig->timeout_retry_limit,
             qpconfig->local_ack_timeout
             );
    sout << buffer;

    IBDIAG_RETURN_VOID;
}

int SharpMngr::SharpMngrDumpAllQPs(ofstream &sout)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    char buffer[256];

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_aggnode = *nI;
        if (!p_sharp_aggnode) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        IBNode * p_node = p_sharp_aggnode->GetIBPort()->p_node;

        map_qpn_to_qpc_config qpc_map;
        //for every TreeID send QPCConfig
        for (u_int16_t i = 0; i < p_sharp_aggnode->GetTreesSize(); i++) {

            SharpTreeNode * p_sharp_tree_node = p_sharp_aggnode->GetSharpTreeNode(i);
            if (!p_sharp_tree_node)
                continue;

            //TODO no need to use p_parent_sharp_tree_edge just use p_sharp_tree_node->GetSharpParentTreeEdge()
            SharpTreeEdge * p_parent_sharp_tree_edge = p_sharp_tree_node->GetSharpParentTreeEdge();
            if (p_parent_sharp_tree_edge) {

                p_parent_sharp_tree_edge->AddToQpcMap(qpc_map);
            }

            for (u_int8_t j = 0; j < p_sharp_tree_node->GetChildrenSize(); j++) {

                SharpTreeEdge * p_sharp_tree_edge = p_sharp_tree_node->GetSharpTreeEdge(j);
                if (!p_sharp_tree_edge) {
                    m_ibdiag->SetLastError("DB error - found null SharpTreeEdge in trees DB for child index = %d", j);
                    continue;
                }

                p_sharp_tree_edge->AddToQpcMap(qpc_map);
            }
        }

        IBNode *p_remote_sw = p_sharp_aggnode->GetIBPort()->p_remotePort->p_node;
        snprintf(buffer, sizeof(buffer),
                 "AN:\"" STR_FMT "\", lid:" U32D_FMT ", node guid:" U64H_FMT
                 ", switch guid:" U64H_FMT ", \"" STR_FMT "\"",
                 p_node->description.c_str(),
                 p_sharp_aggnode->GetIBPort()->base_lid,
                 p_node->guid_get(),
                 p_remote_sw->guid_get(),
                 p_remote_sw->description.c_str());
        sout << endl << buffer << endl;

        for (map_qpn_to_qpc_config::iterator nI = qpc_map.begin();
             nI != qpc_map.end();
             ++nI) {

            u_int32_t qpn = (*nI).first;
            struct AM_QPCConfig * qpconfig = (*nI).second;
            if (!qpn || !qpconfig)
                continue;
            DumpQPC(sout, qpconfig);
            sout << endl;
        }
    }

exit:
    IBDIAG_RETURN(rc);
}

int SharpMngr::DumpSharpPMCountersToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    stringstream sstream;

    if (csv_out.DumpStart(SECTION_SHARP_PM_COUNTERS))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    sstream.str("");
    sstream << "GUID,LID,AggregationNodeDescription,packet_sent,ack_packet_sent,retry_packet_sent"
            << ",rnr_event,timeout_event,oos_nack_rcv,rnr_nack_rcv,packet_discard_transport"
            << ",packet_discard_sharp,aeth_syndrome_ack_packet,hba_sharp_lookup"
            << ",hba_received_pkts,hba_received_bytes,hba_sent_ack_packets,rcds_sent_packets"
            << ",hba_sent_ack_bytes,rcds_send_bytes,hba_multi_packet_message_dropped_pkts"
            << ",hba_multi_packet_message_dropped_bytes" << endl;
    csv_out.WriteBuf(sstream.str());

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_aggnode = *nI;
        if (!p_sharp_aggnode) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            continue;
        }

        IBNode *p_node = p_sharp_aggnode->GetIBPort()->p_node;
        if (!p_node)
            continue;

        const AggNodePerformanceCounters &agg_node_perf_cntr = p_sharp_aggnode->GetPerfCounters();
        if (!agg_node_perf_cntr.perf_cntr)
            continue;

        sstream.str("");
        sstream << PTR(p_node->guid_get())
                << "," << DEC(p_sharp_aggnode->GetIBPort()->base_lid)
                << "," << QUOTED(p_node->description)
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode, PACKET_SENT_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->packet_sent),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode , ACK_PACKET_SENT_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->ack_packet_sent),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode , RETRY_PACKET_SENT_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->retry_packet_sent),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode , RNR_EVENT_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->rnr_event),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode , TIMEOUT_EVENT_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->timeout_event),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode , OOS_NACK_RCV_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->oos_nack_rcv),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode , RNR_NACK_RCV_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->rnr_nack_rcv),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode ,
                                                    PACKET_DISCARD_TRANSPORT_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->packet_discard_transport),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode ,
                                                    PACKET_DISCARD_SHARP_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->packet_discard_sharp),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode ,
                                                    AETH_SYNDROME_ACK_PACKET_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->aeth_syndrome_ack_packet),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode , HBA_SHARP_LOOKUP_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->hba_sharp_lookup),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode ,
                                                    HBA_RECEIVED_PKTS_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->hba_received_pkts),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode , HBA_RECEIVED_BYTES_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->hba_received_bytes),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode ,
                                                    HBA_SENT_ACK_PACKETS_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->hba_sent_ack_packets),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode ,
                                                    RCDS_SENT_PACKETS_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->rcds_sent_packets),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode ,
                                                    HBA_SENT_ACK_BYTES_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->hba_sent_ack_bytes),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode , RCDS_SENT_BYTES_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->rcds_send_bytes),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode ,
                                                    HBA_MULTI_PACKET_MESSAGE_DROPPED_PKTS_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->
                                    hba_multi_packet_message_dropped_pkts),
                             CSTR("N/A"))
                << "," << IF(p_sharp_aggnode->
                             IsPerfCounterSupported(agg_node_perf_cntr.mode ,
                                                    HBA_MULTI_PACKET_MESSAGE_DROPPED_BYTES_BIT),
                             DEC(agg_node_perf_cntr.perf_cntr->
                                    hba_multi_packet_message_dropped_bytes),
                             CSTR("N/A"))
                << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_SHARP_PM_COUNTERS);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int SharpMngr::DumpSharpPMHBAPortCountersToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    stringstream sstream;

    if (csv_out.DumpStart(SECTION_HBA_PORT_COUNTERS))
            IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    sstream.str("");
    sstream << "PortGUID,PortNum,hba_received_pkts,hba_received_bytes,hba_sent_ack_packets"
            << ",rcds_sent_packets,hba_sent_ack_bytes,rcds_send_bytes"
            << ",hba_multi_packet_message_dropped_pkts,hba_multi_packet_message_dropped_bytes"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_aggnode = *nI;
        if (!p_sharp_aggnode) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            continue;
        }

        const AM_ANInfo &am_an_info = p_sharp_aggnode->GetANInfo();
        if (!am_an_info.ext_perf_counters_supported)
            continue;

        IBNode *p_node = p_sharp_aggnode->GetIBPort()->p_node;
        if (!p_node)
            continue;

        if ((!p_sharp_aggnode->GetIBPort()->p_remotePort) ||
            (!p_sharp_aggnode->GetIBPort()->p_remotePort->p_node))
            continue;

        IBNode *p_remote_node = p_sharp_aggnode->GetIBPort()->p_remotePort->p_node;

        const map_port_number_to_hba_pm_counters &map_hba_pm_counters =
                                                 p_sharp_aggnode->GetHBAPerfCounters();

        for (map_port_number_to_hba_pm_counters::const_iterator it = map_hba_pm_counters.begin();
             it != map_hba_pm_counters.end(); ++it) {
                 IBPort *p_curr_port = p_remote_node->getPort(it->first);
                 if (!p_curr_port)
                      continue;

                 sstream.str("");
                 sstream << PTR(p_curr_port->guid_get())
                         << "," << DEC(it->first)
                         << "," << IF(p_sharp_aggnode->
                                      IsPerfCounterSupported(HBA ,HBA_RECEIVED_PKTS_BIT),
                                      DEC(it->second.hba_received_pkts),
                                      CSTR("N/A"))
                         << "," << IF(p_sharp_aggnode->
                                      IsPerfCounterSupported(HBA ,HBA_RECEIVED_BYTES_BIT),
                                      DEC(it->second.hba_received_bytes),
                                      CSTR("N/A"))
                         << "," << IF(p_sharp_aggnode->
                                      IsPerfCounterSupported(HBA ,HBA_SENT_ACK_PACKETS_BIT),
                                      DEC(it->second.hba_sent_ack_packets),
                                      CSTR("N/A"))
                         << "," << IF(p_sharp_aggnode->
                                      IsPerfCounterSupported(HBA , RCDS_SENT_PACKETS_BIT),
                                      DEC(it->second.rcds_sent_packets),
                                      CSTR("N/A"))
                         << "," << IF(p_sharp_aggnode->
                                      IsPerfCounterSupported(HBA , HBA_SENT_ACK_BYTES_BIT),
                                      DEC(it->second.hba_sent_ack_bytes),
                                      CSTR("N/A"))
                         << "," << IF(p_sharp_aggnode->
                                      IsPerfCounterSupported(HBA , RCDS_SENT_BYTES_BIT),
                                      DEC(it->second.rcds_send_bytes),
                                      CSTR("N/A"))
                         << "," << IF(p_sharp_aggnode->
                                      IsPerfCounterSupported(HBA ,
                                                             HBA_MULTI_PACKET_MESSAGE_DROPPED_PKTS_BIT),
                                      DEC(it->second.hba_multi_packet_message_dropped_pkts),
                                      CSTR("N/A"))
                         << "," << IF(p_sharp_aggnode->
                                      IsPerfCounterSupported(HBA ,
                                                             HBA_MULTI_PACKET_MESSAGE_DROPPED_BYTES_BIT),
                                      DEC(it->second.hba_multi_packet_message_dropped_bytes),
                                      CSTR("N/A"))
                         << endl;

                 csv_out.WriteBuf(sstream.str());
        }
    }

    csv_out.DumpEnd(SECTION_HBA_PORT_COUNTERS);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
