/*
 * Copyright (c) 2024-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include "infiniband/ibdm/Fabric.h"
#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibutils/output_stream_primitives.h"

#include "ibdiag_sim_info_dump.h"

const std::vector<const char*> SimInfoData::SMPCapMaskBits = {
    "IsPrivateLinearForwardingSupported",                  //   0
    "IsAdaptiveRoutingSupported",                          //   1
    "IsAdaptiveRoutingRev1Supported",                      //   2
    "IsRemotePortMirroringSupported",                      //   3
    "IsTemperatureSensingSupported",                       //   4
    "IsConfigSpaceAccessSupported",                        //   5
    "IsCableInfoSupported",                                //   6
    "IsSMPEyeOpenSupported",                               //   7

    "IsLossyVLConfigSupported",                            //   8
    "IsExtendedPortInfoSupported",                         //   9
    "IsAccessRegisterSupported",                           //  10
    "IsInterProcessCommunicationSupported",                //  11
    "IsPortSLToPrivateLFTMapSupported",                    //  12
    "IsExtendedNodeInfoSupported",                         //  13
    "IsSlaveVirtualSwitch",                                //  14
    "IsVirtualizationSupported",                           //  15

    "IsAdvancedRetransAndFECModesSupported",               //  16
    "IsSemaphoreLockSupported",                            //  17
    "IsSHArPAggrNodeSupported",                            //  18
    "IsCableInfoPasswordSupported",                        //  19
    "IsVL2VLPerSWIDSupported",                             //  20
    "IsSpecialPortsMarkingSupported",                      //  21
    "IsVPortGUIDInfoSupported",                            //  22
    "IsForcePLFTSubgroup0Supported",                       //  23

    "IsQoSConfigSLRateLimitSupported",                     //  24
    "IsQoSConfigSLVPortRateLimitSupported",                //  25
    "IsQoSConfigSLAllocBWSupported",                       //  26
    "IsQoSConfigSLVPortAllocBWSupported",                  //  27
    "IsEnhancedTrap128Supported",                          //  28
    "IsDevInfoSupported",                                  //  29
    "IsGlobalOOOSupported",                                //  30
    "IsMADAddlStatusSupported",                            //  31

    "IsQoSConfigSLRateLimitPortSupported",                 //  32
    "IsQoSConfigSLRateLimitVPortSupported",                //  33
    "IsQoSConfigSLAllocBWPortSupported",                   //  34
    "IsQoSConfigSLAllocBWVPortSupported",                  //  35
    "IsQoSConfigHighPriorityVLGroupRateLimitSupported",    //  36
    "IsQoSConfigHighPriorityVLGroupAllocBWSupported",      //  37
    "IsQoSConfigLowPriorityVLGroupRateLimitSupported",     //  38
    "IsQoSConfigLowPriorityVLGroupAllocBWSupported",       //  39

    "IsQoSConfigVLRateLimitSupported",                     //  40
    "IsQoSConfigVLAllocBWSupported",                       //  41
    nullptr,                                               //  42
    "IsCableInfoBankNumberSupported",                      //  43
    "IsDCQCNSupported",                                    //  44
    "IsAdaptiveTimeoutSupported",                          //  45
    "IsClassEnableSupported",                              //  46
    "IsExtendedAddressModeSupported",                      //  47

    "IsRouterLIDSupported",                                //  48
    "IsLinearForwardingTableSplitSupported",               //  49
    "IsDataTypetoSLMappingSupported",                      //  50
    "IsQoSConfigVLSupported",                              //  51
    "IsFastRecoverySupported",                             //  52
    "IsCreditWatchdogSupported",                           //  53
    "IsNVLFabricManagementSupported",                      //  54
    "IsNVLReductionManagementSupported",                   //  55

    "IsRailFilterSupported",                               //  56
    "IsNVLHBFConfigSupported",                             //  57
    "IsDataTypetoVLMappingSupported",                      //  58
    "IsContainAndDrainSupported",                          //  59
    "IsProfilesConfigSupported",                           //  60
    "IsBERConfigSupported",                                //  61
    "IsGlobalOOOV2Supported",                              //  62
    "IsVLBufferWeightSupported",                           //  63

    "IsEndPortPlaneFilterConfigSupported",                 //  64
    "IsEntryPlaneFilterSupported",                         //  65
    "IsBWUtiliztionSupported",                             //  66
    "IsExtendedSwitchInfoSupported",                       //  67
    "IsFabricManagerInfoSupported"                         //  68
    "IsExtPortInfoUnhealthyReasonSupported"                //  69
    "IsMulticastPrivateLinearForwardingSupported"          //  70
    "IsPortSLToMulticastPrivateLFTMapSupported"            //  71

    "IsChassisInfoSupported"                               //  72
    "IsPLRTriggerSupported"                                //  73
    "IsFwISSUSupported"                                    //  74
};

const std::vector<const char*> SimInfoData::GMPCapMaskBits = {
    "IsPortPowerStateSupported",                   //   0
    "IsDeviceSoftResetSupported",                  //   1
    "IsExtGPIOPortAccessSupported",                //   2
    "IsPhyConfigSupported",                        //   3
    "IsVendSpecLedControlSupported",               //   4
    "IsConfigSpaceAccessSupported",                //   5
    "IsPerVLCountersSupported",                    //   6
    "IsPerVLExtendedSupported",                    //   7

    "IsPortLLRStatisticsSupported",                //   8
    "IsCounterGroupsSupported",                    //   9
    "IsEnhancedConfigSpaceAccessSupported",        //  10
    "IsFlowMonitoringSupported",                   //  11
    "IsCongestionNotificationSupported",           //  12
    "IsVirtualSwitchSupported",                    //  13
    "IsPortXmitQLengthSupported",                  //  14
    "IsMulticastForwardingTableSupported",         //  15

    "IsPortXmitWaitVLExtendedSupported",           //  16
    "IsExtI2CPortAccessSupported",                 //  17
    "IsDiagnosticDataSupported",                   //  18
    "IsMaxRetransmissionRateSupported",            //  19
    "IsAccessRegisterGMPSupported",                //  20
    "IsPortHistogramsSupported",                   //  21
    "IsPoolHistogramsSupported",                   //  22
    "IsVPortCountersSupported",                    //  23

    "IsDiagnosticDataIdentificationSupported",     //  24
    "IsEnhPort0DeviceResetSupported",              //  25
    "IsDevInfoSupported",                          //  26
    "IsMADAddlStatusSupported",                    //  27
    "IsCableBurnSupported",                        //  28
    "IsMirroringEventsSupported",                  //  29
    "IsBERMonitoringSupported",                    //  30
    "IsBERMonitoringEventsSupported",              //  31

    "IsDCQCNSupported",                            //  32
    "IsPLRMaxRetransmissionRateSupported",         //  33
    "IsEffectiveCounterSupported",                 //  34
    "IsRawBerPerLaneSupported",                    //  35
    "IsAccessRegisterSupported",                   //  36
    "IsMirroringInfoSupported",                    //  37
    "IsExtendedAddressModeSupported",              //  38
    "IsSwitchNetworkInfoSupported",                //  39

    "CreditWatchdogTimeoutCounters",               //  40
    "FastRecoveryCountersSupported",               //  41
    "IsGeneralResetSupported",                     //  42
    "IsBufferHistogramSupported",                  //  43
    "IsPortsHistogramSupported",                   //  44
    "IsPacketBufferAllocControlSupported"          //  45
};

const std::vector<const char*> SimInfoData::PortInfoCapMaskBits = {
    nullptr,                                       //   0
    "IsSM",                                        //   1
    "IsNoticeSupported",                           //   2
    "IsTrapSupported",                             //   3
    "IsOptionalIPDSupported",                      //   4
    "IsAutomaticMigrationSupported",               //   5
    "IsSLMappingSupported",                        //   6
    "IsMKeyNVRAM",                                 //   7

    "IsPKeyNVRAM",                                 //   8
    "IsLEDInfoSupported",                          //   9
    "IsSMdisabled",                                //  10
    "IsSystemImageGUIDSupported",                  //  11
    "IsPKeySwitchExternalPortTrapSupported",       //  12
    "IsCableInfoSupported",                        //  13
    "IsExtendedSpeedsSupported",                   //  14
    "IsCapabilityMask2Supported",                  //  15

    "IsCommunicationManagementSupported",          //  16
    "IsSNMPTunnelingSupported",                    //  17
    "IsReinitSupported",                           //  18
    "IsDeviceManagementSupported",                 //  19
    "IsVendorClassSupported",                      //  20
    "IsDRNoticeSupported",                         //  21
    "IsCapabilityMaskNoticeSupported",             //  22
    "IsBootManagementSupported",                   //  23

    "IsLinkRoundTripLatencySupported",             //  24
    "IsClientReregistrationSupported",             //  25
    "IsOtherLocalChangesNoticeSupported",          //  26
    "IsLinkSpeedWidthPairsTableSupported",         //  27
    "IsVendorSpecificMadsTableSupported",          //  28
    "IsMulticastPKeyTrapSuppressionSupported",     //  29
    "IsMulticastFDBTopSupported",                  //  30
    "IsHierarchyInfoSupported"                     //  31
};

const std::vector<const char*> SimInfoData::PortInfoCapMask2Bits = {
    "IsSetNodeDescriptionSupported",               //   0
    "IsPortInfoExtendedSupported",                 //   1
    "IsVirtualizationSupported",                   //   2
    "IsSwitchPortStateTableSupported",             //   3
    "IsLinkWidth2xSupported",                      //   4
    "IsLinkSpeedHDRSupported",                     //   5
    "IsMKeyProtectBitsExtSupported",               //   6
    "IsEnhancedTrap128Supported",                  //   7

    "IsPartitionTopSupported",                     //   8
    "IsEnhancedQoSArbiterSupported",               //   9
    "IsLinkSpeedNDRSupported",                     //  10
    "IsExtendedSpeeds2Supported",                  //  11
    "IsLinkSpeedXDRSupported",                     //  12
    nullptr,                                       //  13
    nullptr,                                       //  14
    nullptr                                        //  15
};


int SimInfoDumpPy::Dump(const std::string & filename) {

    IBDIAG_ENTER

    std::ofstream  sout;
    int rc = m_ibdiag.OpenFile("Simultator Info (Python)", filename, sout, false, true);

    if (rc)
       IBDIAG_RETURN(rc);

    for (auto& x : m_ibdiag.GetDiscoverFabricPtr()->NodeByName) {
        IBNode* p_node = x.second;

        if (!p_node)
            continue;

        sout << std::setw(80) << "" << std::endl;

        sout << "Node Type: "   << nodetype2char(p_node->type)
             << (p_node->isSpecialNode() ? "(Special Node)" : "")
             << ", Node GUID: " << PTR(p_node->guid_get()) << std::setfill(' ')
             << ", Ports: "     << +p_node->numPorts
             << std::endl;

        sout << "Name: '"        << p_node->getName()
             << "', Alt: '"      << p_node->getAlternativeName()
             << std::endl;

        sout << std::setw(80) << "" << std::endl;

        this->DumpSMPCapMask(sout, *p_node);
        this->DumpGMPCapMask(sout, *p_node);

        this->DumpPortInfoAllCapMask(sout, *p_node);
    }

    m_ibdiag.CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SimInfoDumpPy::DumpPortInfoAllCapMask(std::ostream & sout, IBNode & node) {
    for (phys_port_t i = node.type == IB_SW_NODE ? 0 : 1; i <= node.numPorts; ++i) {
        const IBPort * p_port = node.getPort(i);

        if(!p_port)
            continue;

        if(i && !p_port->is_data_worthy())
            continue;

        SMP_PortInfo *p_info = m_ibdiag.GetIBDMExtendedInfoPtr()->getSMPPortInfo(p_port->createIndex);

        if (!p_info)
            continue;

        DumpPortInfoCapMask(sout, i, p_info, false);
        DumpPortInfoCapMask(sout, i, p_info, true);
    }
}

void SimInfoDumpPy::DumpPortInfoCapMask(std::ostream & sout, phys_port_t index, SMP_PortInfo *p_info, bool is_cap_mask2) {
    std::stringstream text;
    text << std::setw(12) << "" << "# ---------- Port[" << +index << "] CapMask"
         << (is_cap_mask2 ? "2" : "" ) << " ----------" << std::endl;
    text << std::setw(12) << "" << "caps = node.getPortInfo(index).get_capability_mask"
         << (is_cap_mask2 ? "2" : "" ) << "_bits()" << std::endl;

    const std::vector<const char *> &vec 
        = is_cap_mask2 ? SimInfoData::PortInfoCapMask2Bits : SimInfoData::PortInfoCapMaskBits;

    uint32_t capMask = is_cap_mask2 ? p_info->CapMsk2 : p_info->CapMsk;

    bool is_empty   = true;
    for(size_t i = 0; i < vec.size(); ++i ) {
        const char * name = vec[i];

        if (!name)
            continue;

        if (capMask & (1 << i)) {
            text << std::setw(12) << "" << "caps." << name
                 << std::setw(int(50lu - strlen(name)))
                 << " = 1" << std::endl;

            is_empty = false;
        }
    }

    if (!is_empty)
        sout << text.str();
}

void SimInfoDumpPy::DumpSMPCapMask(std::ostream & sout, IBNode & node) {
    sout << std::setw(8) << "" << "# ---------- SMP Caps ----------" << std::endl;
    for (size_t i = 0; i < SimInfoData::SMPCapMaskBits.size(); ++i) {
        const char * name = SimInfoData::SMPCapMaskBits[i];

        if (!name)
            continue;

        if (m_ibdiag.GetCapabilityModulePtr()->IsSupportedSMPCapability(&node, (u_int8_t)i)) {
            sout << std::setw(8) << "" << "node.getCapMaskSMPBits()." << name
                 << std::setw((int)(50lu - strlen(name))) << " = 1" << std::endl;
        }
    }
}

void SimInfoDumpPy::DumpGMPCapMask(std::ostream & sout, IBNode & node) {
    sout << std::setw(8) << "" << "# ---------- GMP Caps ----------" << std::endl;
    for (size_t i = 0; i < SimInfoData::GMPCapMaskBits.size(); ++i) {
        const char * name = SimInfoData::GMPCapMaskBits[i];

        if (!name)
            continue;

        if (m_ibdiag.GetCapabilityModulePtr()->IsSupportedGMPCapability(&node, (u_int8_t)i)) {
            sout << std::setw(8) << "" << "node.getCapMaskGMPBits()." << name
                 << std::setw((int)(50lu - strlen(name))) << " = 1" << std::endl;
        }
    }
}
