 /*
 * Copyright (c) 2021-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#pragma once

#include <stdio.h>
#include <string>
#include <sstream>

#include "nlohmann/json.hpp"

#include "uphy_db.h"

namespace UPHY
{
    #define UPHY_JSON_REASON_PREFIX "\n\t--> "

    class JsonLoader
    {
        private:
            const std::string   m_filename;

            FILE *              m_fd;
            UPHY::DataSet *     m_dataset;
            std::stringstream   m_errors;

        public:
            const UPHY::DataSet *       dataset() const { return m_dataset;   }
            const std::stringstream &   errors() const  { return m_errors;    }

        public:
            JsonLoader(const std::string & filename);
            ~JsonLoader();

        private:
            void read_json(const nlohmann::json & json);

        private:
            void read_header(const nlohmann::json & json);

        private:
            void read_enums(const nlohmann::json & json);
            void read_enum_labels(UPHY::DataSet::Enumerator * enumerator, const nlohmann::json & json);
            DataSet::Enumerator::width_t read_enum_width(const nlohmann::json & json);

        private:
            void read_registers(const nlohmann::json & json);
            void read_register_fields(UPHY::DataSet::Register * reg, const nlohmann::json & json);

        private:
            void read_register_dumps(const nlohmann::json & json, DataSet::Register::dumps_collection_t & value);
            register_access_flags_t read_register_access(const nlohmann::json & json);
            register_unit_flags_t read_register_units(const nlohmann::json & json);
            register_secure_t read_register_secure(const nlohmann::json & json);

        private:
            template<typename T>
                void read(const nlohmann::json & json, const std::string & name, T & value)
                {
                    try {
                        value = json.at( name );
                    }
                    catch ( nlohmann::json::exception & e ) {
                        throw std::runtime_error("Failed to read '" + name + "'" + UPHY_JSON_REASON_PREFIX + e.what());
                    }
                    catch ( ... ) {
                        throw std::runtime_error("Failed to read '" + name + "', unhandled exception");
                    }
                }

    };
}
