/*
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <fcntl.h>
#include <poll.h>

#include <infiniband/umad.h>
#include <infiniband/umad_types.h>

#include "ibis.h"
#include "ibis_verbs.h"


void Ibis::SetVerbsEnabled(bool verbs_enabled)
{
    #if VERBS_SUPPORTED == 1
        m_verbs_enabled = verbs_enabled;
    #endif
}

uint8_t *Ibis::VerbsGetSendMad(uint64_t i)
{
    #if VERBS_SUPPORTED != 1
        return NULL;
    #else
        if (i == (uint64_t)-1)
            i = m_verbs_port->send_queue.front();
        return m_verbs_port->mad_buff + i * 296;
    #endif
}

uint8_t *Ibis::VerbsGetRecvBuff(uint64_t i)
{
    #if VERBS_SUPPORTED != 1
        return NULL;
    #else
        return m_verbs_port->mad_buff + (m_verbs_port->queue_size + i) * 296;
    #endif
}

uint8_t *Ibis::VerbsGetRecvMad(uint64_t i)
{
    // Verbs always reserves 40 bytes for GRH even if it is not used
    return VerbsGetRecvBuff(i) + 40;
}

int Ibis::VerbsPostReceive(uint64_t i)
{
    #if VERBS_SUPPORTED != 1
        return -1;
    #else
        struct ibv_recv_wr recv_wr, *bad_recv_wr;
        struct ibv_sge recv_sge;

        recv_sge.addr = (uint64_t)VerbsGetRecvBuff(i);
        recv_sge.length = 296;
        recv_sge.lkey = m_verbs_port->mr->lkey;

        recv_wr.wr_id = i;
        recv_wr.next = NULL;
        recv_wr.sg_list = &recv_sge;
        recv_wr.num_sge = 1;

        if (ibv_post_recv(m_verbs_port->qp, &recv_wr, &bad_recv_wr)) {
            SetLastError("Post receive failed");
            return 1;
        }

        return 0;
    #endif
}

/*
 * Fetch all completed send requests and mark corresponded buffers as available (add them to
 * send_queue).
 */
int Ibis::VerbsEmptySendQueue()
{
    #if VERBS_SUPPORTED != 1
        return -1;
    #else
        struct ibv_wc wc = {0};
        int n;

        do {
            n = ibv_poll_cq(m_verbs_port->send_cq, 1, &wc);
            if (n < 0) {
                SetLastError("Failed to read work completions from completion queue");
                return 1;
            }

            if (n) {
                if (wc.status != IBV_WC_SUCCESS) {
                    SetLastError("Send work completion %lu failed, status: %s (%d)",
                                 wc.wr_id, ibv_wc_status_str(wc.status), wc.status);
                    return 1;
                }

                m_verbs_port->send_queue.push(wc.wr_id);
            }
        } while(n);

        return 0;
    #endif
}

int Ibis::VerbsOpenPort(const char *dev_name, int port_num)
{
    #if VERBS_SUPPORTED != 1
        return -1;
    #else
        struct ibv_device **device_list;
        struct ibv_device **device_list_it;
        struct ibv_device *device = NULL;
        struct ibv_port_attr port_attr;
        struct ibv_qp_init_attr qp_init_attr;
        struct ibv_qp_attr qp_attr;
        int flags;

        m_verbs_port->dev_name = dev_name;
        m_verbs_port->port_num = port_num;
        m_verbs_port->queue_size = m_max_gmps_on_wire;
        m_verbs_port->context = NULL;
        m_verbs_port->pd = NULL;
        m_verbs_port->send_cq = NULL;
        m_verbs_port->recv_cq = NULL;
        m_verbs_port->recv_comp_ch = NULL;
        m_verbs_port->qp = NULL;
        m_verbs_port->mad_buff = NULL;
        m_verbs_port->mr = NULL;
        m_verbs_port->ah_map = std::map<uint32_t, struct ibv_ah *>();
        m_verbs_port->send_queue = std::queue<uint64_t>();

        device_list = device_list_it = ibv_get_device_list(NULL);
        if (!device_list) {
            SetLastError("ibv_get_device_list() failed");
            return 1;
        }

        while ((device = *device_list_it++)) {
            if (dev_name == NULL || !strcmp(ibv_get_device_name(device), dev_name)) {
                m_verbs_port->context = ibv_open_device(device);
                break;
            }
        }

        ibv_free_device_list(device_list);

        if (!device) {
            SetLastError("Couldn't find device with name %s", dev_name);
            goto fail;
        }

        if (!m_verbs_port->context) {
            SetLastError("Failed to open device %s", dev_name);
            goto fail;
        }

        if (ibv_query_port(m_verbs_port->context, port_num, &port_attr)) {
            SetLastError("Failed to query port %s:%d parameters", dev_name, port_num);
            goto fail;
        }

        m_verbs_port->pd = ibv_alloc_pd(m_verbs_port->context);
        if (!m_verbs_port->pd) {
            SetLastError("Failed to allocate protection domain for device %s", dev_name);
            goto fail;
        }

        m_verbs_port->send_cq = ibv_create_cq(m_verbs_port->context, m_verbs_port->queue_size,
                                             NULL, NULL, 0);
        if (!m_verbs_port->send_cq) {
            SetLastError("Failed to create send completion queue for device %s", dev_name);
            goto fail;
        }

        m_verbs_port->recv_comp_ch = ibv_create_comp_channel(m_verbs_port->context);
        if (!m_verbs_port->recv_comp_ch) {
            SetLastError("Failed to create receive completion event channel for device %s", dev_name);
            goto fail;
        }

        flags = fcntl(m_verbs_port->recv_comp_ch->fd, F_GETFL);
        if (!(flags & O_NONBLOCK)) {
            if (fcntl(m_verbs_port->recv_comp_ch->fd, F_SETFL, flags | O_NONBLOCK) < 0) {
                SetLastError("Failed to set file descriptor of receive completion event channel to "
                            "non-block for device %s", dev_name);
                goto fail;
            }
        }

        m_verbs_port->recv_cq = ibv_create_cq(m_verbs_port->context, m_verbs_port->queue_size,
                                             NULL, m_verbs_port->recv_comp_ch, 0);
        if (!m_verbs_port->recv_cq) {
            SetLastError("Failed to create receive completion queue for device %s", dev_name);
            goto fail;
        }

        memset(&qp_init_attr, 0, sizeof(qp_init_attr));

        qp_init_attr.qp_context = NULL;
        qp_init_attr.send_cq = m_verbs_port->send_cq;
        qp_init_attr.recv_cq = m_verbs_port->recv_cq;
        qp_init_attr.srq = NULL;
        qp_init_attr.cap.max_send_wr = m_verbs_port->queue_size;
        qp_init_attr.cap.max_recv_wr = m_verbs_port->queue_size;
        qp_init_attr.cap.max_send_sge = 1;
        qp_init_attr.cap.max_recv_sge = 1;
        qp_init_attr.cap.max_inline_data = 128;
        qp_init_attr.qp_type = IBV_QPT_UD;
        qp_init_attr.sq_sig_all = 0;

        m_verbs_port->qp = ibv_create_qp(m_verbs_port->pd, &qp_init_attr);
        if (!m_verbs_port->qp) {
            SetLastError("Failed to create queue pair for device %s", dev_name);
            goto fail;
        }

        memset(&qp_attr, 0, sizeof(qp_attr));

        qp_attr.qp_state = IBV_QPS_INIT;
        qp_attr.pkey_index = 0;
        qp_attr.port_num = port_num;
        qp_attr.qkey = UMAD_QKEY;

        if (ibv_modify_qp(m_verbs_port->qp, &qp_attr,
                          IBV_QP_STATE | IBV_QP_PKEY_INDEX | IBV_QP_PORT | IBV_QP_QKEY)) {
            SetLastError("Failed to initialize queue pair for device %s", dev_name);
            goto fail;
        }

        memset(&qp_attr, 0, sizeof(struct ibv_qp_attr));

        qp_attr.qp_state = IBV_QPS_RTR;

        if (ibv_modify_qp(m_verbs_port->qp, &qp_attr, IBV_QP_STATE)) {
            SetLastError("Failed to set queue pair state to RTR for device %s", dev_name);
            goto fail;
        }

        memset(&qp_attr, 0, sizeof(struct ibv_qp_attr));

        qp_attr.qp_state = IBV_QPS_RTS;
        qp_attr.sq_psn = 0;

        if (ibv_modify_qp(m_verbs_port->qp, &qp_attr, IBV_QP_STATE | IBV_QP_SQ_PSN)) {
            SetLastError("Failed to set queue pair state to RTS for device %s", dev_name);
            goto fail;
        }

        // Allocate single buffer for both send and receive queues. First half of the buffer will
        // be used for send requests, second half - for receive requests.
        // 296 bytes = 256 bytes (MAD) + 40 bytes (GRH).
        // Verbs always reserves 40 bytes for GRH even if it is not used.
        m_verbs_port->mad_buff = (uint8_t *)calloc(2 * m_verbs_port->queue_size, 296);
        if (!m_verbs_port->mad_buff) {
            SetLastError("Failed to allocate MAD buffer for device %s", dev_name);
            goto fail;
        }

        m_verbs_port->mr = ibv_reg_mr(m_verbs_port->pd, m_verbs_port->mad_buff,
                                     2 * m_verbs_port->queue_size * 296, IBV_ACCESS_LOCAL_WRITE);
        if (!m_verbs_port->mr) {
            SetLastError("Failed to register memory region for device %s", dev_name);
            goto fail;
        }

        for (uint64_t i = 0; i < m_verbs_port->queue_size; i++) {
            m_verbs_port->send_queue.push(i);

            if (VerbsPostReceive(i))
                goto fail;
        }

        return 0;

    fail:
        VerbsClosePort();

        return 1;
    #endif
}

void Ibis::VerbsClosePort()
{
    #if VERBS_SUPPORTED == 1
        for (std::map<uint32_t, struct ibv_ah *>::iterator it = m_verbs_port->ah_map.begin();
             it != m_verbs_port->ah_map.end(); ++it)
            ibv_destroy_ah(it->second);

        m_verbs_port->ah_map = std::map<uint32_t, struct ibv_ah *>();

        m_verbs_port->send_queue = std::queue<uint64_t>();

        if (m_verbs_port->mr) {
            ibv_dereg_mr(m_verbs_port->mr);
            m_verbs_port->mr = NULL;
        }

        if (m_verbs_port->mad_buff) {
            free(m_verbs_port->mad_buff);
            m_verbs_port->mad_buff = NULL;
        }

        if (m_verbs_port->qp) {
            ibv_destroy_qp(m_verbs_port->qp);
            m_verbs_port->qp = NULL;
        }

        if (m_verbs_port->send_cq) {
            ibv_destroy_cq(m_verbs_port->send_cq);
            m_verbs_port->send_cq = NULL;
        }

        if (m_verbs_port->recv_cq) {
            ibv_destroy_cq(m_verbs_port->recv_cq);
            m_verbs_port->recv_cq = NULL;
        }

        if (m_verbs_port->recv_comp_ch) {
            ibv_destroy_comp_channel(m_verbs_port->recv_comp_ch);
            m_verbs_port->recv_comp_ch = NULL;
        }

        if (m_verbs_port->pd) {
            ibv_dealloc_pd(m_verbs_port->pd);
            m_verbs_port->pd = NULL;
        }

        if (m_verbs_port->context) {
            ibv_close_device(m_verbs_port->context);
            m_verbs_port->context = NULL;
        }
    #endif
}

int Ibis::VerbsSendMad()
{
    #if VERBS_SUPPORTED != 1
        return -1;
    #else
        struct ibv_ah *ah;
        struct ibv_send_wr send_wr, *bad_send_wr;
        struct ibv_sge send_sge;

        ib_mad_addr_t *mad_addr = umad_get_mad_addr(p_umad_buffer_send);
        uint16_t dlid = be16toh(mad_addr->lid);
        uint8_t sl = mad_addr->sl;
        uint32_t key = ((uint32_t)sl << 16) | dlid;

        VerbsEmptySendQueue();

        if (m_verbs_port->send_queue.empty()) {
            SetLastError("No more free send requests left");
            return 1;
        }

        std::map<uint32_t, ibv_ah *>::iterator it = m_verbs_port->ah_map.find(key);

        if (it == m_verbs_port->ah_map.end()) {
            struct ibv_ah_attr ah_attr;

            memset(&ah_attr, 0, sizeof(ah_attr));

            ah_attr.dlid = dlid;
            ah_attr.sl = mad_addr->sl;
            ah_attr.port_num = m_verbs_port->port_num;

            ah = ibv_create_ah(m_verbs_port->pd, &ah_attr);
            if (!ah) {
                SetLastError("Failed to create address handler for port %s:%d to DLID 0x%04X SL 0x%02X",
                             m_verbs_port->dev_name, m_verbs_port->port_num, dlid, sl);
                return 1;
            }

            m_verbs_port->ah_map.insert(std::pair<uint32_t, struct ibv_ah *>(key, ah));
        } else {
            ah = it->second;
        }

        uint64_t i = m_verbs_port->send_queue.front();
        m_verbs_port->send_queue.pop();

        memcpy(VerbsGetSendMad(i), p_pkt_send, IBIS_IB_MAD_SIZE);

        send_sge.addr = (uint64_t)VerbsGetSendMad(i);
        send_sge.length = IBIS_IB_MAD_SIZE;
        send_sge.lkey = m_verbs_port->mr->lkey;

        send_wr.wr_id = i;
        send_wr.next = NULL;
        send_wr.sg_list = &send_sge;
        send_wr.num_sge = 1;
        send_wr.opcode = IBV_WR_SEND;
        send_wr.send_flags = IBV_SEND_SIGNALED;
        send_wr.imm_data = htonl(m_verbs_port->qp->qp_num);
        send_wr.wr.ud.ah = ah;
        send_wr.wr.ud.remote_qpn = be32toh(mad_addr->qpn);
        send_wr.wr.ud.remote_qkey = be32toh(mad_addr->qkey);

        if (ibv_post_send(m_verbs_port->qp, &send_wr, &bad_send_wr)) {
            SetLastError("Post send failed");
            return 1;
        }

        return 0;
    #endif
}

#ifdef IBIS_VERBS_ACTIVE_POLL
static uint64_t GetTimeStamp()
{
    struct timespec ts;

    Ibis::GetClockTime(ts);

    return ts.tv_sec * 1000000000 + ts.tv_nsec;
}
#endif

int Ibis::VerbsRecvMad(int timeout_ms)
{
    #if VERBS_SUPPORTED != 1
        return -1;
    #else
        /*
         * Define IBIS_VERBS_ACTIVE_POLL if you want to use active polling implementation of
         * VerbsRecvMad (poll for receive until timeout is reached).
         */
        #ifndef IBIS_VERBS_ACTIVE_POLL
            struct ibv_wc wc = {0};
            int n;

            ib_user_mad_t *mad = (ib_user_mad_t *)p_umad_buffer_recv;
            mad->status = EAGAIN;

            VerbsEmptySendQueue();

            n = ibv_poll_cq(m_verbs_port->recv_cq, 1, &wc);

            if (!n && timeout_ms) {
                if (ibv_req_notify_cq(m_verbs_port->recv_cq, 0)) {
                    SetLastError("Failed to request notification for receive completion event");
                    return 1;
                }

                n = ibv_poll_cq(m_verbs_port->recv_cq, 1, &wc);

                if (!n) {
                    struct pollfd fd;
                    struct ibv_cq *cq = NULL;
                    void *cq_context = NULL;

                    fd.fd = m_verbs_port->recv_comp_ch->fd;
                    fd.events = POLLIN;
                    fd.revents = 0;

                    int rc = poll(&fd, 1, timeout_ms);

                    if (rc == -1) {
                        SetLastError("Failed to poll for verbs completion event");
                        return 1;
                    } else if (rc == 0) {
                        SetLastError("Timeout waiting for receive");
                        return 2;
                    }

                    if (ibv_get_cq_event(m_verbs_port->recv_comp_ch, &cq, &cq_context) < 0) {
                        SetLastError("Failed to extract verbs completion event");
                        return 1;
                    }

                    ibv_ack_cq_events(cq, 1);

                    n = ibv_poll_cq(m_verbs_port->recv_cq, 1, &wc);
                }
            }

            if (!n && !timeout_ms)
                return 2;

            if (n <= 0) {
                SetLastError("Failed to read work completion from completion queue");
                return 1;
            }

            memcpy(p_pkt_recv, VerbsGetRecvMad(wc.wr_id), IBIS_IB_MAD_SIZE);

            VerbsPostReceive(wc.wr_id);

            mad->status = (wc.status != IBV_WC_SUCCESS) ? 1 : 0;
            mad->addr.lid = htobe16(wc.slid);
            mad->addr.sl = wc.sl;
            mad->addr.qpn = htobe32(wc.qp_num);

            ClearLastError();

            return 0;
        #else
            struct ibv_wc wc;
            uint64_t start_time, curr_time;

            ib_user_mad_t *mad = (ib_user_mad_t *)p_umad_buffer_recv;
            mad->status = EAGAIN;

            VerbsEmptySendQueue();

            start_time = GetTimeStamp();

            do {
                int n = ibv_poll_cq(m_verbs_port->recv_cq, 1, &wc);
                if (n < 0) {
                    SetLastError("Failed to read work completion from completion queue");
                    return 1;
                }

                if (n) {
                    memcpy(mad, VerbsGetRecvMad(wc.wr_id), IBIS_IB_MAD_SIZE);

                    VerbsPostReceive(wc.wr_id);

                    mad->status = (wc.status != IBV_WC_SUCCESS) ? 1 : 0;
                    mad->addr.lid = htobe16(wc.slid);
                    mad->addr.sl = wc.sl;
                    mad->addr.qpn = htobe32(wc.qp_num);

                    ClearLastError();

                    return 0;
                }

                curr_time = GetTimeStamp();
            } while (curr_time - start_time < (uint64_t)timeout_ms * 1000000);

            SetLastError("Timeout waiting for receive");
            return 2;
        #endif
    #endif
}

int Ibis::VerbsUmadRecvMad(int timeout_ms)
{
    #if VERBS_SUPPORTED != 1
        return -1;
    #else
        /*
         * Define IBIS_VERBS_ACTIVE_POLL if you want to use active polling implementation of
         * VerbsUmadRecvMad (poll for receive until timeout is reached).
         */
        #ifndef IBIS_VERBS_ACTIVE_POLL
            int length = IBIS_IB_MAD_SIZE;
            int recv_agent_id;

            struct ibv_wc wc = {0};
            int n = -1;

            bool umad_post_proc = false;
            bool verbs_post_proc = false;

            ib_user_mad_t *mad = (ib_user_mad_t *)p_umad_buffer_recv;
            mad->status = EAGAIN;

            VerbsEmptySendQueue();

            do {
                recv_agent_id = umad_recv(gmp_port.umad_port_id, p_umad_buffer_recv, &length, 0);
                if (recv_agent_id >= 0) {
                    umad_post_proc = true;
                    break;
                }

                n = ibv_poll_cq(m_verbs_port->recv_cq, 1, &wc);
                if (n) {
                    verbs_post_proc = true;
                    break;
                } else {
                    if (ibv_req_notify_cq(m_verbs_port->recv_cq, 0))
                        break;

                    n = ibv_poll_cq(m_verbs_port->recv_cq, 1, &wc);
                    if (n) {
                        verbs_post_proc = true;
                        break;
                    } else {
                        struct pollfd fds[2];

                        fds[0].fd = gmp_port.umad_port_id;
                        fds[0].events = POLLIN;
                        fds[0].revents = 0;

                        fds[1].fd = m_verbs_port->recv_comp_ch->fd;
                        fds[1].events = POLLIN;
                        fds[1].revents = 0;

                        int rc = poll(fds, 2, timeout_ms);

                        if (rc < 0) {
                            break;
                        } else if (rc == 0) {
                            SetLastError("Timeout waiting for receive");
                            return 2;
                        }

                        if (fds[0].revents & POLLIN) {
                            length = IBIS_IB_MAD_SIZE;
                            recv_agent_id = umad_recv(gmp_port.umad_port_id, p_umad_buffer_recv, &length, 0);
                            umad_post_proc = true;
                            break;
                        }

                        if (fds[1].revents & POLLIN) {
                            struct ibv_cq *cq = NULL;
                            void *cq_context = NULL;

                            if (ibv_get_cq_event(m_verbs_port->recv_comp_ch, &cq, &cq_context) < 0) {
                                SetLastError("Failed to extract verbs completion event");
                                return 1;
                            }

                            ibv_ack_cq_events(cq, 1);

                            n = ibv_poll_cq(m_verbs_port->recv_cq, 1, &wc);
                            verbs_post_proc = true;
                            break;
                        }
                    }
                }
            } while (0);

            if (umad_post_proc && recv_agent_id >= 0) {
                u_int8_t mgmt_class = *(u_int8_t *)(p_pkt_recv + 1); // mad_get_field(p_pkt_recv, 0, IB_MAD_MGMTCLASS_F);
                u_int8_t class_version = *(u_int8_t *)(p_pkt_recv + 2);

                if (CheckValidAgentIdForClass(recv_agent_id, mgmt_class, class_version))
                    return 1;

                return 0;
            }

            if (verbs_post_proc && n > 0) {
                memcpy(p_pkt_recv, VerbsGetRecvMad(wc.wr_id), IBIS_IB_MAD_SIZE);

                VerbsPostReceive(wc.wr_id);

                mad->status = (wc.status != IBV_WC_SUCCESS) ? 1 : 0;
                mad->addr.lid = htobe16(wc.slid);
                mad->addr.sl = wc.sl;
                mad->addr.qpn = htobe32(wc.qp_num);

                ClearLastError();

                return 0;
            }

            SetLastError("Failed to receive mad");
            return 1;
        #else
            int length = IBIS_IB_MAD_SIZE;
            int recv_agent_id = -1;

            uint64_t start_time, curr_time;

            VerbsEmptySendQueue();

            start_time = GetTimeStamp();

            do {
                recv_agent_id = umad_recv(gmp_port.umad_port_id, p_umad_buffer_recv, &length, 0);
                if (recv_agent_id >= 0) {
                    u_int8_t mgmt_class = *(u_int8_t *)(this->p_pkt_recv + 1);
                    u_int8_t class_version = *(u_int8_t *)(this->p_pkt_recv + 2);

                    if (CheckValidAgentIdForClass(recv_agent_id, mgmt_class, class_version))
                        return 1;

                    return 0;
                }

                if (!VerbsRecvMad(0))
                    return 0;

                curr_time = GetTimeStamp();
            } while (curr_time - start_time < (uint64_t)timeout_ms * 1000000);

            SetLastError("Timeout waiting for receive");
            return 2;
        #endif
    #endif
}
