/*
 * Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdarg.h>
#include "infiniband/ibdiag/ibdiag_clbck.h"
#include "infiniband/ibdm/Fabric.h"
#include "infiniband/ibutils/output_stream_primitives.h"

IBDiagClbck ibDiagClbck;

#define IBIS_STATUS(X)   " [status=" << PTR((u_int16_t)(X)) << "]"


void IBDiagClbck::SetLastError(const char *fmt, ...)
{
    IBDIAG_ENTER;
    char buffer[IBDIAG_ERR_SIZE] = {0};
    va_list args;

    va_start(args, fmt);
    vsnprintf(buffer, sizeof(buffer), fmt, args);
    va_end(args);
    m_LastError.assign(buffer);
    IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "-E- %s\n", m_LastError.c_str());
    IBDIAG_RETURN_VOID;
}

const char* IBDiagClbck::GetLastError()
{
    IBDIAG_ENTER;
    if (m_LastError != "")
        IBDIAG_RETURN(m_LastError.c_str());
    IBDIAG_RETURN("Unknown");
}

void IBDiagClbck::PMClassPortInfoClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMClassPortInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct IB_ClassPortInfo *p_class_port_info =
            (struct IB_ClassPortInfo *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addPMClassPortInfo(p_node, *p_class_port_info);
        if (rc) {
            SetLastError("Failed to add PMClassPortInfo for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::PMPortSamplesControlGetClbck(const clbck_data_t &clbck_data,
                                               int rec_status,
                                               void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortSamplesControl." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct PM_PortSamplesControl *p_sample_control =
            (struct PM_PortSamplesControl *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addPMPortSamplesControl(p_port, *p_sample_control);

        if (rc) {
            SetLastError("Failed to add PM_PortSamplesControl for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::PMPortCountersGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortCountersGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct PM_PortCounters *p_port_counters=
            (struct PM_PortCounters *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortCounters(p_port, *p_port_counters);
        if (rc) {
            SetLastError("Failed to add PMPortCounters for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::PMPortCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortCountersClear." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }
}

void IBDiagClbck::PMPortCountersExtendedGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortCountersExtendedGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct PM_PortCountersExtended *p_port_counters=
            (struct PM_PortCountersExtended *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortCountersExtended(p_port, *p_port_counters);
        if (rc) {
            SetLastError("Failed to add PMPortCountersExtended for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::PMPortCountersExtendedClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    IBPort *p_port = (IBPort *)clbck_data.m_data1;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortCountersExtendedClear." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }
}

void IBDiagClbck::PMPortExtendedSpeedsGetClbck(
    const clbck_data_t &clbck_data,
    int rec_status,
    void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortExtendedSpeedsCountersGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct PM_PortExtendedSpeedsCounters *p_port_counters=
            (struct PM_PortExtendedSpeedsCounters *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortExtSpeedsCounters(p_port, *p_port_counters);
        if (rc) {
            SetLastError("Failed to add PMPortExtendedSpeedsCounters for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::PMPortExtendedSpeedsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortExtendedSpeedsCountersClear." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }
}

void IBDiagClbck::PMPortExtendedSpeedsRSFECGetClbck(
    const clbck_data_t &clbck_data,
    int rec_status,
    void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_EXT_SPEEDS_RSFEC_FAILED(p_port->p_node->appData1.val))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS;

        stringstream sstr;
        sstr << "PMPortExtendedSpeedsRSFECCountersGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct PM_PortExtendedSpeedsRSFECCounters *p_port_counters =
            (struct PM_PortExtendedSpeedsRSFECCounters *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortExtSpeedsRSFECCounters(
                                        p_port, *p_port_counters);
        if (rc) {
            SetLastError("Failed to add PMPortExtendedSpeedsRSFECCounters "
                         "for port=%s, err=%s",
                         p_port->getName().c_str(),
                         m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::PMPortExtendedSpeedsRSFECClearClbck(
    const clbck_data_t &clbck_data,
    int rec_status,
    void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if ((rec_status & 0x00ff) &&  //bits [8-15] are class specific - don't need them here
            !IS_SUPPORT_EXT_SPEEDS_RSFEC_FAILED(p_port->p_node->appData1.val)) {

        // mark that we already check for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS;

        stringstream sstr;
        sstr << "PMPortExtendedSpeedsRSFECCountersClear." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }
}

void IBDiagClbck::VSPortLLRStatisticsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_LLR_FAILED(p_port->p_node->appData1.val))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_LLR_COUNTERS;

        stringstream sstr;
        sstr << "VSPortLLRStatisticsGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct VendorSpec_PortLLRStatistics *p_port_counters=
            (struct VendorSpec_PortLLRStatistics *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addVSPortLLRStatistics(p_port, *p_port_counters);

        if (rc) {
            SetLastError("Failed to add VSPortLLRStatistics for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VSPortLLRStatisticsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if ((rec_status & 0x00ff) &&  //bits [8-15] are class specific - don't need them here
            !IS_SUPPORT_LLR_FAILED(p_port->p_node->appData1.val)) {

        // mark that we already check for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_LLR_COUNTERS;

        if ((bool)(intptr_t)(clbck_data.m_data2)) {
            stringstream sstr;
            sstr << "VSPortLLRStatisticsClear." << IBIS_STATUS(rec_status);
            m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::PMPortRcvErrorDetailsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortRcvErrorDetailsGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct PM_PortRcvErrorDetails *p_port_rcv_error_details;
        p_port_rcv_error_details = (struct PM_PortRcvErrorDetails *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortRcvErrorDetails(p_port, *p_port_rcv_error_details);
        if (rc)
        {
            SetLastError("Failed to add PM_PortRcvErrorDetails for port=%s, err=%s",
                p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::PMPortRcvErrorDetailsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortRcvErrorDetailsClear." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }
}

void IBDiagClbck::PMPortXmitDiscardDetailsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortXmitDiscardDetailsGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct PM_PortXmitDiscardDetails *p_port_xmit_discard_details;
        p_port_xmit_discard_details = (struct PM_PortXmitDiscardDetails *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortXmitDiscardDetails(p_port,
                                                                    *p_port_xmit_discard_details);
        if (rc)
        {
            SetLastError("Failed to add PM_PortXmitDiscardDetails for port=%s, err=%s",
            p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::PMPortXmitDiscardDetailsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "PMPortXmitDiscardDetailsClear." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }
}

/*** NVL ***/
void IBDiagClbck::NVLAnycastLIDInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "AnycastLIDInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        u_int32_t block_idx = (u_int32_t) ((size_t) clbck_data.m_data2);

        struct SMP_AnycastLIDInfo *p_data =
                (struct SMP_AnycastLIDInfo*) p_attribute_data;

        int rc = m_pFabricExtendedInfo->addNVLAnycastLIDInfo(p_node,
                block_idx, *p_data);

        if (rc) {
            SetLastError(
                    "Failed to add NVLAnycastLIDInfo for node=%s, block=%d, err=%s",
                    p_node->getName().c_str(), block_idx, m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}


void IBDiagClbck::NVLHBFConfigGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "NVLHBFConfig." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct SMP_NVLHBFConfig *p_data =
            (struct SMP_NVLHBFConfig *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addNVLHBFConfig(p_port, *p_data);

        if (rc) {
            SetLastError("Failed to add NVLHBFConfig for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}


void IBDiagClbck::NVLContainAndDrainInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "ContainAndDrainInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        u_int32_t block_idx = (u_int32_t) ((size_t) clbck_data.m_data2);
        struct SMP_ContainAndDrainInfo *p_data =
            (struct SMP_ContainAndDrainInfo *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addContainAndDrainInfo(p_node, block_idx, *p_data);

        if (rc) {
            SetLastError("Failed to add ContainAndDrainInfo for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}


void IBDiagClbck::NVLContainAndDrainPortStateGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "ContainAndDrainPortStateGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        u_int32_t block_idx = (u_int32_t) ((size_t) clbck_data.m_data2);
        struct SMP_ContainAndDrainPortState *p_data =
            (struct SMP_ContainAndDrainPortState *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addContainAndDrainPortState(p_node, block_idx, *p_data);

        if (rc) {
            SetLastError("Failed to add ContainAndDrainPortState for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}


/*** NVL Reduction ***/
void IBDiagClbck::NVLClassPortInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "NVL IB_ClassPortInfo." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct IB_ClassPortInfo *p_data =
            (struct IB_ClassPortInfo *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addNVLClassPortInfo(p_node, *p_data);

        if (rc) {
            SetLastError("Failed to add NVL ClassPortInfo for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

}

void IBDiagClbck::NVLReductionInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "NVLReductionInfo." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct NVLReductionInfo *p_data =
            (struct NVLReductionInfo *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addNVLReductionInfo(p_node, *p_data);

        if (rc) {
            SetLastError("Failed to add NVLReductionInfo for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

}

void IBDiagClbck::NVLReductionPortInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "NVLReductionPortInfo." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct NVLReductionPortInfo *p_data =
            (struct NVLReductionPortInfo *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addNVLReductionPortInfo(p_port, *p_data);

        if (rc) {
            SetLastError("Failed to add NVLReductionPortInfo for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::NVLReductionRoundingModeGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "NVLReductionRoundingMode." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct NVLReductionRoundingMode *p_data =
            (struct NVLReductionRoundingMode *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addNVLReductionRoundingMode(p_node, *p_data);

        if (rc) {
            SetLastError("Failed to add NVLReductionRoundingMode for port=%s, err=%s",
                         p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMP_LFTSplitGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMP_LinearForwardingTableSplit." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct SMP_LinearForwardingTableSplit *p_data =
            (struct SMP_LinearForwardingTableSplit *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addSMP_LFTSplit(p_node, *p_data);

        if (rc) {
            SetLastError("Failed to add SMP_LinearForwardingTableSplit for port=%s, err=%s",
                         p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::NVLReductionForwardingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "NVLReductionForwardingTable." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct NVLReductionForwardingTable *p_data =
            (struct NVLReductionForwardingTable *)p_attribute_data;

        u_int32_t block_idx = (u_int32_t)((intptr_t)clbck_data.m_data2);

        int rc = m_pFabricExtendedInfo->addNVLReductionForwardingTable(p_node, block_idx, *p_data);

        if (rc) {
            SetLastError("Failed to add ReductionForwardingTable for node=%s, err=%s",
                         p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::NVLPenaltyBoxConfigClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "NVLPenaltyBoxConfig." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        u_int32_t block_idx = (u_int32_t) ((size_t) clbck_data.m_data2);

        struct NVLPenaltyBoxConfig *p_data =
                (struct NVLPenaltyBoxConfig*) p_attribute_data;

        int rc = m_pFabricExtendedInfo->addNVLPenaltyBoxConfig(p_node,
                block_idx, *p_data);

        if (rc) {
            SetLastError(
                    "Failed to add NVLPenaltyBoxConfig for node=%s, block=%d, err=%s",
                    p_node->getName().c_str(), block_idx, m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}


void IBDiagClbck::NVLReductionConfigureMLIDMonitorsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "NVLReductionConfigureMLIDMonitors." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct NVLReductionConfigureMLIDMonitors *p_data =
            (struct NVLReductionConfigureMLIDMonitors *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addNVLReductionConfigureMLIDMonitors(p_port, *p_data);

        if (rc) {
            SetLastError("Failed to add NVLReductionConfigureMLIDMonitors for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::NVLReductionCountersGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "NVLReductionCounters." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        u_int32_t block_idx = (u_int32_t) ((size_t) clbck_data.m_data2);

        struct NVLReductionCounters *p_data =
                (struct NVLReductionCounters*) p_attribute_data;

        int rc = m_pFabricExtendedInfo->addNVLReductionCounters(p_port,
                block_idx, *p_data);

        if (rc) {
            SetLastError(
                    "Failed to add NVLReductionCounters for port=%s, block=%d, err=%s",
                    p_port->getName().c_str(), block_idx, m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}


//Callback to GMP GeneralInfo MAD
void IBDiagClbck::VSGeneralInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_p_capability_module)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    int status = rec_status & 0x00ff; //bits [8-15] are class specific - don't need them here

    if (status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        m_pErrors->push_back(new FabricErrNodeNotSupportCap(p_node,
                "The firmware of this device does not support general info GMP capability"));
    } else if (status) {
        stringstream sstr;
        sstr << "VSGeneralInfo." << IBIS_STATUS(rec_status);

        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));
    } else {
        struct VendorSpec_GeneralInfo *p_general_info=
            (struct VendorSpec_GeneralInfo *)p_attribute_data;
        u_int64_t guid = p_node->guid_get();

        m_ErrorState = m_pFabricExtendedInfo->addVSGeneralInfo(p_node, *p_general_info);
        if (m_ErrorState)
            SetLastError("Failed to add VSGeneralInfo for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        fw_version_obj_t gmp_fw;
        gmp_fw = *p_general_info;
        //update the fw in the cache
        m_ErrorState = m_p_capability_module->AddGMPFw(guid, gmp_fw);
        if (m_ErrorState)
            SetLastError("Failed to add GMP Fw Info for node=%s",
                    p_node->getName().c_str());

        // Update SerDes level on IBNode
        p_node->serdes_version = (SerDesVersion)p_general_info->HWInfo.technology;

        capability_mask_t gmp_mask, smp_mask;

        u_int8_t prefix_len = 0;
        u_int64_t matched_guid = 0;
        query_or_mask_t qmask;
        bool gmp_query_per_matched_guid = false;

        //update the guid 2 gmp mask cache only if gmp mask is unknown
        if (!m_p_capability_module->IsGMPMaskKnown(guid)) {
            if (m_p_capability_module->IsLongestGMPPrefixMatch(guid,
                                                               prefix_len,
                                                               matched_guid,
                                                               qmask) &&
                qmask.to_query == true)
                gmp_query_per_matched_guid = true;

            //take the mask from the MAD, if either guid - 'query'
            //configuration match was found or there is NO fw configured device
            if (gmp_query_per_matched_guid ||
                m_p_capability_module->GetGMPFwConfiguredMask(p_node->vendId,
                                                              p_node->devId,
                                                              gmp_fw,
                                                              gmp_mask) !=
                IBDIAG_SUCCESS_CODE)
                gmp_mask = p_general_info->CapabilityMask;

            //update the guid 2 mask cache
            if (m_p_capability_module->AddGMPCapabilityMask(guid, gmp_mask) != IBDIAG_SUCCESS_CODE)
                m_pErrors->push_back(new FabricErrSmpGmpCapMaskExist(p_node, false, gmp_mask));
        }

        //try to update SMP mask for this node that fw dependant
        if (!m_p_capability_module->IsSMPMaskKnown(guid))
        {
            bool smp_query_per_matched_guid = false;
            if (m_p_capability_module->IsLongestSMPPrefixMatch(guid,
                                                               prefix_len,
                                                               matched_guid,
                                                               qmask) &&
                qmask.to_query == true)
                smp_query_per_matched_guid = true;
            //if no 'guid to query' configuration exists,
            //check for fw configured mask
            if (!smp_query_per_matched_guid &&
                m_p_capability_module->GetSMPFwConfiguredMask(p_node->vendId,
                                                              p_node->devId,
                                                              gmp_fw,
                                                              smp_mask) ==
                IBDIAG_SUCCESS_CODE)
                    //update the SMP guid 2 mask cache
                    if (m_p_capability_module->AddSMPCapabilityMask(guid, smp_mask))
                        m_pErrors->push_back(new FabricErrSmpGmpCapMaskExist(p_node, true, smp_mask));
        }
    }

    IBDIAG_RETURN_VOID;
}


void IBDiagClbck::VSSwitchNetworkInfoClbck(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_switch = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar,
                                                        clbck_data.m_data1);
    if (!VerifyObject(p_switch, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSSwitchNetworkInfo." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_switch, sstr.str()));
        IBDIAG_RETURN_VOID;
    }

    VS_SwitchNetworkInfo *p_switch_network_info = (VS_SwitchNetworkInfo *)p_attribute_data;
    int rc = m_pFabricExtendedInfo->addVSSwitchNetworkInfo(p_switch, *p_switch_network_info);
    if (rc) {
        SetLastError("Failed to add VS_SwitchNetworkInfo for switch=%s, err=%s",
                    p_switch->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        m_ErrorState = rc;
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPGUIDInfoTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        if (p_port->p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        stringstream sstr;
        sstr << "SMPGUIDInfoTableGetByLid." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_port->p_node, sstr.str()));

        p_port->p_node->appData1.val = 1;

    } else {
        struct SMP_GUIDInfo *p_guid_info =
            (struct SMP_GUIDInfo *)p_attribute_data;
        u_int8_t block = (u_int8_t)(intptr_t)clbck_data.m_data2;

        m_ErrorState = m_pFabricExtendedInfo->addSMPGUIDInfo(p_port, *p_guid_info, block);
        if (m_ErrorState) {
            this->SetLastError("Failed to add SMPGuidInfo for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPPkeyTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        if (p_port->p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        stringstream sstr;
        sstr << "SMPPKeyTableGetByDirect." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_port->p_node, sstr.str()));

        p_port->p_node->appData1.val = 1;

    } else {
        struct SMP_PKeyTable *p_pkey_table =
            (struct SMP_PKeyTable *)p_attribute_data;
        u_int8_t block = (u_int8_t)(intptr_t)clbck_data.m_data2;

        m_ErrorState = m_pFabricExtendedInfo->addSMPPKeyTable(p_port, *p_pkey_table, block);
        if (m_ErrorState) {
            this->SetLastError("Failed to add SMPPKeyTable for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPVLArbitrationGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        if (p_port->p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        stringstream sstr;
        sstr << "SMP_VLArbitrationMadGetByDirect." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_port->p_node, sstr.str()));

        p_port->p_node->appData1.val = 1;
        p_port->p_node->skipRoutingChecks = true;

    } else {

        struct SMP_VLArbitrationTable* p_vl_arbitration_table =
                (struct SMP_VLArbitrationTable*)p_attribute_data;

        u_int32_t block = (u_int32_t)(intptr_t)clbck_data.m_data2;

        m_ErrorState = m_pFabricExtendedInfo->addSMPVLArbitrationTable(p_port,
                                                                       *p_vl_arbitration_table,
                                                                       block);

        if (m_ErrorState) {
            this->SetLastError("Failed to add SMP_VLArbitrationTable for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPSMInfoMadGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBPort *p_port = (IBPort *)clbck_data.m_data1;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPSMInfoMadGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct SMP_SMInfo *p_sm_info =
            (struct SMP_SMInfo *)p_attribute_data;

        m_ErrorState = m_pFabricExtendedInfo->addSMPSMInfoObj(p_port, *p_sm_info);
        if (m_ErrorState) {
            this->SetLastError("Failed to add SMInfoObj for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPLinearForwardingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        if (p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        p_node->appData1.val = 1;

        stringstream sstr;
        sstr << "SMPLinearForwardingTable (block="
             << block
             << ")."
             << IBIS_STATUS(rec_status);

        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        //build DB
        struct SMP_LinearForwardingTable *p_linear_forwarding_table =
            (struct SMP_LinearForwardingTable *)p_attribute_data;

        for (int curr_entry_in_block = 0; curr_entry_in_block < IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS; ++curr_entry_in_block) {
            lid_t lid = (lid_t)(block * IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS + curr_entry_in_block);
            p_node->setLFTPortForLid(lid, p_linear_forwarding_table->Port[curr_entry_in_block]);
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPMulticastForwardingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;
    u_int8_t port_group = (u_int8_t)(intptr_t)clbck_data.m_data3;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        if (p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        p_node->appData1.val = 1;

        stringstream sstr;
        sstr << "SMPMulticastForwardingTable (block="
             << block
             << ", group="
             << +port_group
             << ")."
             << IBIS_STATUS(rec_status);

        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        //build DB
        struct SMP_MulticastForwardingTable *p_multicast_forwarding_table =
            (struct SMP_MulticastForwardingTable *)p_attribute_data;

        for (int curr_entry_in_block = 0; curr_entry_in_block < IBIS_IB_MAD_SMP_MFT_NUM_BLOCKS;
                ++curr_entry_in_block) {
            lid_t lid = (lid_t)(0xc000 + block * IBIS_IB_MAD_SMP_MFT_NUM_BLOCKS + curr_entry_in_block);
            if (p_multicast_forwarding_table->PortMask[curr_entry_in_block])
                p_node->setMFTPortForMLid(lid,
                        p_multicast_forwarding_table->PortMask[curr_entry_in_block],
                        port_group);
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPSLToVLMappingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_p_sout)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    u_int8_t in_port = (u_int8_t)(intptr_t)clbck_data.m_data2;
    u_int8_t out_port = (u_int8_t)(intptr_t)clbck_data.m_data3;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        stringstream sstr;
        sstr << "SMPSLToVLMappingTableGetByDirect." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

        p_node->skipRoutingChecks = true;
        p_node->appData1.val = 1;

    } else {

        char buffer[1024];

        struct SMP_SLToVLMappingTable *p_slvl_mapping =
            (struct SMP_SLToVLMappingTable *)p_attribute_data;

        snprintf(buffer, sizeof(buffer), U64H_FMT " %u %u"
                 " 0x%x%x 0x%x%x 0x%x%x 0x%x%x"
                 " 0x%x%x 0x%x%x 0x%x%x 0x%x%x\n",
                 p_node->guid_get(),
                 in_port,
                 out_port,
                 p_slvl_mapping->SL0ToVL, p_slvl_mapping->SL1ToVL,
                 p_slvl_mapping->SL2ToVL, p_slvl_mapping->SL3ToVL,
                 p_slvl_mapping->SL4ToVL, p_slvl_mapping->SL5ToVL,
                 p_slvl_mapping->SL6ToVL, p_slvl_mapping->SL7ToVL,
                 p_slvl_mapping->SL8ToVL, p_slvl_mapping->SL9ToVL,
                 p_slvl_mapping->SL10ToVL, p_slvl_mapping->SL11ToVL,
                 p_slvl_mapping->SL12ToVL, p_slvl_mapping->SL13ToVL,
                 p_slvl_mapping->SL14ToVL, p_slvl_mapping->SL15ToVL);

        (*m_p_sout) << buffer;

        p_node->setSLVL(in_port, out_port, 0, p_slvl_mapping->SL0ToVL);
        p_node->setSLVL(in_port, out_port, 1, p_slvl_mapping->SL1ToVL);
        p_node->setSLVL(in_port, out_port, 2, p_slvl_mapping->SL2ToVL);
        p_node->setSLVL(in_port, out_port, 3, p_slvl_mapping->SL3ToVL);
        p_node->setSLVL(in_port, out_port, 4, p_slvl_mapping->SL4ToVL);
        p_node->setSLVL(in_port, out_port, 5, p_slvl_mapping->SL5ToVL);
        p_node->setSLVL(in_port, out_port, 6, p_slvl_mapping->SL6ToVL);
        p_node->setSLVL(in_port, out_port, 7, p_slvl_mapping->SL7ToVL);
        p_node->setSLVL(in_port, out_port, 8, p_slvl_mapping->SL8ToVL);
        p_node->setSLVL(in_port, out_port, 9, p_slvl_mapping->SL9ToVL);
        p_node->setSLVL(in_port, out_port, 10, p_slvl_mapping->SL10ToVL);
        p_node->setSLVL(in_port, out_port, 11, p_slvl_mapping->SL11ToVL);
        p_node->setSLVL(in_port, out_port, 12, p_slvl_mapping->SL12ToVL);
        p_node->setSLVL(in_port, out_port, 13, p_slvl_mapping->SL13ToVL);
        p_node->setSLVL(in_port, out_port, 14, p_slvl_mapping->SL14ToVL);
        p_node->setSLVL(in_port, out_port, 15, p_slvl_mapping->SL15ToVL);
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPVSGeneralInfoFwInfoGetClbck(const clbck_data_t &clbck_data,
                                                 int rec_status,
                                                 void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_p_capability_module)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    int status = rec_status & 0x00ff; //bits [8-15] are class specific - don't need them here

    if (status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {        //not supported for this device
        m_pErrors->push_back(
                new FabricErrNodeNotSupportCap(p_node,
                    "The firmware of this device "
                    "does not support GeneralInfoSMP MAD (Fw)"));

    } else if (status) {

        stringstream sstr;
        sstr << "SMPVSGeneralInfoFwInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct FWInfo_Block_Element *p_fw_info =
            (struct FWInfo_Block_Element *)p_attribute_data;
        fw_version_obj_t smp_fw;
        smp_fw.major = p_fw_info->Extended_Major;
        smp_fw.minor = p_fw_info->Extended_Minor;
        smp_fw.sub_minor = p_fw_info->Extended_SubMinor;
        m_p_capability_module->AddSMPFw(p_node->guid_get(), smp_fw);
        capability_mask_t mask;
        u_int8_t prefix_len = 0;
        u_int64_t matched_guid = 0;
        query_or_mask_t qmask;
        uint64_t guid = p_node->guid_get();

        //if this node has not mask yet, check if we can get it now
        if (!m_p_capability_module->IsSMPMaskKnown(guid)) {
            //check if there is configuration per GUID that says query
            //this case has precedence over fw configurations
            bool query_per_matched_guid = false;
            qmask.to_query = false;

            if (m_p_capability_module->IsLongestSMPPrefixMatch(
                p_node->guid_get(), prefix_len, matched_guid, qmask) &&
                qmask.to_query == true)
                query_per_matched_guid = true;

            //no match by guid with 'query'
            if (!query_per_matched_guid) {
                //check if this node has configured mask based on fw version
                //(<ven id, dev id> -> <fw, mask>)
                mask.clear();

                //check if it matched by fw configured device then
                if (m_p_capability_module->GetSMPFwConfiguredMask(
                    p_node->vendId, p_node->devId, smp_fw, mask) ==
                    IBDIAG_SUCCESS_CODE)
                    //update the cache
                    if (m_p_capability_module->AddSMPCapabilityMask(
                        p_node->guid_get(), mask) != IBDIAG_SUCCESS_CODE)
                        m_pErrors->push_back(new FabricErrSmpGmpCapMaskExist(p_node, true, mask));
            }
        }

        //try to update gmp mask for this guid
        //update the guid 2 gmp mask cache only if gmp mask is unknown
        if (!m_p_capability_module->IsGMPMaskKnown(guid)) {
            bool gmp_query_per_matched_guid = false;
            qmask.to_query = false;
            if (m_p_capability_module->IsLongestGMPPrefixMatch(guid,
                                                               prefix_len,
                                                               matched_guid,
                                                               qmask) &&
                qmask.to_query == true)
                gmp_query_per_matched_guid = true;

            mask.clear();
            //if no 'guid to query' configuration exists,
            //check for fw configured mask
            if (!gmp_query_per_matched_guid &&
                m_p_capability_module->GetGMPFwConfiguredMask(p_node->vendId,
                                                              p_node->devId,
                                                              smp_fw,
                                                              mask) ==
                IBDIAG_SUCCESS_CODE)
                //update the guid 2 mask cache
                if (m_p_capability_module->AddGMPCapabilityMask(guid, mask) != IBDIAG_SUCCESS_CODE)
                    m_pErrors->push_back(new FabricErrSmpGmpCapMaskExist(p_node, false, mask));
        }

        //check if it's the same fw reported in smp and gmp
        fw_version_obj_t gmp_fw;
        if (m_p_capability_module->GetGMPFw(guid, gmp_fw) == IBDIAG_SUCCESS_CODE && smp_fw != gmp_fw)
            m_pErrors->push_back(new FabricErrSmpGmpFwMismatch(p_node, smp_fw, gmp_fw));
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPVSGeneralInfoCapabilityMaskGetClbck(const clbck_data_t &clbck_data,
                                                    int rec_status,
                                                    void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_p_capability_module)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    int status = rec_status & 0x00ff;  //bits [8-15] are class specific - don't need them here

    if (status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {     //not supported for this device
        m_pErrors->push_back(
                new FabricErrNodeNotSupportCap(
                        p_node,
                        "The firmware of this device "
                        "does not support GeneralInfoSMP MAD (Capability)"));

    } else if (status) {

        stringstream sstr;
        sstr << "SMPVSGeneralInfoCapabilityMaskGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct GeneralInfoCapabilityMask *p_general_info =
            (struct GeneralInfoCapabilityMask *)p_attribute_data;
        capability_mask_t mask;
        mask = *p_general_info;
        m_ErrorState = m_p_capability_module->AddSMPCapabilityMask(p_node->guid_get(), mask);
        if (m_ErrorState)
            SetLastError("Failed to add SMP Capability Mask for node=%s",
                         p_node->getName().c_str());
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::GSIPerSLVLGetClbck(const clbck_data_t &clbck_data,
                                     int rec_status,
                                     void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_pFabricExtendedInfo ||
        !m_p_capability_module)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    int status = rec_status & 0x00ff; //bits [8-15] are class specific - don't need them here

    class CountersPerSLVL * p_cntrs_per_slvl =
        (class CountersPerSLVL *)clbck_data.m_data2;

    if (status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        FabricErrPortNotSupportCap *p_curr_fabric_port_err =
            new FabricErrPortNotSupportCap(p_port,
                                           "This port does not support PM " +
                                           p_cntrs_per_slvl->GetCntrHeader() +
                                           " MAD although capability bit is on");

        p_curr_fabric_port_err->SetLevel(EN_FABRIC_ERR_WARNING);
        m_pErrors->push_back(p_curr_fabric_port_err);

    } else if (status) {

        stringstream sstr;
        sstr << p_cntrs_per_slvl->GetCntrHeader() << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {
        struct PM_PortRcvXmitCntrsSlVl *p_per_slvl_ctrs =
            (PM_PortRcvXmitCntrsSlVl *)p_attribute_data;

        pair_ibport_slvl_cntr_data_t sl_vl_data(p_port, *p_per_slvl_ctrs);

        p_cntrs_per_slvl->m_set_port_data_update.insert(sl_vl_data);

        if (m_ErrorState)
            this->SetLastError("Failed to store pm per slvl counter"
                               " for port %s, err=%s",
                    p_port->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPVSExtendedPortInfoGetClbck(const clbck_data_t &clbck_data,
                                                int rec_status,
                                                void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_pFabricExtendedInfo ||
        !m_p_capability_module)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    int status = rec_status & 0x00ff; //bits [8-15] are class specific - don't need them here

    if (status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        m_pErrors->push_back(
                new FabricErrPortNotSupportCap(p_port,
                       "The firmware of this device "
                       "does not support ExtendedPortInfoSMP MAD"));
    } else if (status) {

        stringstream sstr;
        sstr << "SMPVSExtendedPortInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {
        SMP_MlnxExtPortInfo *p_mlnx_ext_port_info =
            (SMP_MlnxExtPortInfo *)p_attribute_data;

        /*
         * Check the MAD result - maybe we need to change speed to MLNX speed
         */
        if (p_mlnx_ext_port_info->LinkSpeedActive)
            p_port->set_internal_speed(mlnxspeed2speed(p_mlnx_ext_port_info->LinkSpeedActive));

        u_int8_t active_cell_size = m_pIBDiag->GetLLRActiveCellSize();
        //assume the LLR active cell size, that was given by the user
        //for FDR/FDR10 and higher speeds
        if (p_port->get_internal_speed() >= IB_LINK_SPEED_14 && active_cell_size)
            p_mlnx_ext_port_info->RetransMode = active_cell_size;

        //check if can determine the FEC Mode
        if (EXT_FEC_MODE_CONTROL_AND_REPORT_SUPPORT(
                p_mlnx_ext_port_info->CapabilityMask))
            p_port->set_fec_mode((IBFECMode)p_mlnx_ext_port_info->FECModeActive);

        if (p_mlnx_ext_port_info->IsSpecialPort)
            p_port->setSpecialPortType(
                       (IBSpecialPortType)p_mlnx_ext_port_info->SpecialPortType);

        /*
         * Save smp mlnx ext port info attribute in the fabric
         */
        m_ErrorState = m_pFabricExtendedInfo->addSMPMlnxExtPortInfo(
                                                p_port,
                                                *p_mlnx_ext_port_info);
        if (m_ErrorState)
            this->SetLastError("Failed to store vs extended port info"
                               " for port %s, err=%s",
                    p_port->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPExtendedNodeInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPExtendedNodeInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        //build DB
        struct ib_extended_node_info *p_ext_node_info =
            (struct ib_extended_node_info *)p_attribute_data;

        if (p_ext_node_info->sl2vl_act)
            p_node->setSL2VLAct(p_ext_node_info->sl2vl_act);

        p_node->ext_type = (IBExtNodeType)p_ext_node_info->node_type_extended;

        // Add data to fabric_extended_info
        m_ErrorState = m_pFabricExtendedInfo->addSMPExtNodeInfo(p_node, *p_ext_node_info);

        if (m_ErrorState)
            this->SetLastError("Failed to store VS Extended Node Info for node %s, err=%s",
                               p_node->getName().c_str(),
                               m_pFabricExtendedInfo->GetLastError());
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPPLFTInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPPLFTInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        //build DB
        ib_private_lft_info * p_plft_info =
            (struct ib_private_lft_info *)p_attribute_data;

        if (p_plft_info->Active_Mode != 0)
            p_node->setPLFTEnabled();
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPPortSLToPrivateLFTMapGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPPortSLToPrivateLFTMapGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        //build DB
        ib_port_sl_to_private_lft_map * p_plft_map =
            (struct ib_port_sl_to_private_lft_map *)p_attribute_data;

        ib_port_sl_to_plft * pPortSLToPLFT = NULL;

        for (int curr_entry_in_block = 0;
              curr_entry_in_block < IBIS_IB_MAD_PLFT_MAP_PORT_BLOCK_SIZE;
              ++curr_entry_in_block) {
            phys_port_t port =
                (phys_port_t)(block * IBIS_IB_MAD_PLFT_MAP_PORT_BLOCK_SIZE +
                                      curr_entry_in_block);

            if (port > p_node->numPorts)
                break;

            pPortSLToPLFT =
                &p_plft_map->PortSLToPLFT[curr_entry_in_block];

            p_node->setPLFTMapping(port, 0, pPortSLToPLFT->PLFTToPortSL0);
            p_node->setPLFTMapping(port, 1, pPortSLToPLFT->PLFTToPortSL1);
            p_node->setPLFTMapping(port, 2, pPortSLToPLFT->PLFTToPortSL2);
            p_node->setPLFTMapping(port, 3, pPortSLToPLFT->PLFTToPortSL3);
            p_node->setPLFTMapping(port, 4, pPortSLToPLFT->PLFTToPortSL4);
            p_node->setPLFTMapping(port, 5, pPortSLToPLFT->PLFTToPortSL5);
            p_node->setPLFTMapping(port, 6, pPortSLToPLFT->PLFTToPortSL6);
            p_node->setPLFTMapping(port, 7, pPortSLToPLFT->PLFTToPortSL7);
            p_node->setPLFTMapping(port, 8, pPortSLToPLFT->PLFTToPortSL8);
            p_node->setPLFTMapping(port, 9, pPortSLToPLFT->PLFTToPortSL9);
            p_node->setPLFTMapping(port, 10, pPortSLToPLFT->PLFTToPortSL10);
            p_node->setPLFTMapping(port, 11, pPortSLToPLFT->PLFTToPortSL11);
            p_node->setPLFTMapping(port, 12, pPortSLToPLFT->PLFTToPortSL12);
            p_node->setPLFTMapping(port, 13, pPortSLToPLFT->PLFTToPortSL13);
            p_node->setPLFTMapping(port, 14, pPortSLToPLFT->PLFTToPortSL14);
            p_node->setPLFTMapping(port, 15, pPortSLToPLFT->PLFTToPortSL15);
        }

        if (p_node->getMaxPLFT() >= MAX_PLFT_NUM) {
            char buffer[512];
            snprintf(buffer, sizeof(buffer), "SMPARInfoGet unsupported PLFT number=%u", p_node->getMaxPLFT());

            m_pErrors->push_back(new FabricErrNodeWrongConfig(p_node, buffer));

            p_node->setMaxPLFT((u_int8_t)(MAX_PLFT_NUM - 1));
        }

        //printf("guid:"U64H_FMT" MaxPLFT:%d\n",p_node->guid_get(), (int)p_node->getMaxPLFT());

    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPPrivateLFTTopGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    u_int8_t pLFT = (u_int8_t)(intptr_t)clbck_data.m_data2;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPPrivateLFTTopGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        //build DB
        struct ib_private_lft_map * p_plft_map =
            (struct ib_private_lft_map *)p_attribute_data;

        p_node->setLFDBTop(pLFT, p_plft_map->LFT_Top);
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPHBFConfigGetClbck(const clbck_data_t &clbck_data,
                                    int rec_status,
                                    void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPHBFConfigGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
         m_pFabricExtendedInfo->addHBFConfig(p_node, *(hbf_config*)p_attribute_data);
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPpFRNConfigGetClbck(const clbck_data_t &clbck_data,
                                        int rec_status,
                                        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMP_pFRNConfigGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        SMP_pFRNConfig *p_pfrn = (SMP_pFRNConfig *)p_attribute_data;
        m_pFabricExtendedInfo->addpFRNConfig(p_node, *p_pfrn);
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::N2NClassPortInfoGetClbck(const clbck_data_t &clbck_data,
                                           int rec_status,
                                           void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "N2NClassPortInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        IB_ClassPortInfo *p_cpi = (IB_ClassPortInfo*)p_attribute_data;
        m_pFabricExtendedInfo->addN2NClassPortInfo(p_node, *p_cpi);
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::NeighborsInfoGetClbck(const clbck_data_t &clbck_data,
                                        int rec_status,
                                        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (p_node->appData1.val) // already reported
            IBDIAG_RETURN_VOID;

        stringstream sstr;
        sstr << "NeighborsInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

        p_node->appData1.val = 1;

    } else {

        struct NeighborsInfo *p_neighbors_info = (NeighborsInfo*)p_attribute_data;

        u_int32_t block = (u_int32_t)(intptr_t)clbck_data.m_data2;

        for (u_int32_t i = 0; i < NEIGHBORS_IN_BLOCK; ++i) {
            m_pFabricExtendedInfo->addNeighborsRecord(p_node,
                                                      p_neighbors_info->Neighbors[i],
                                                      block * NEIGHBORS_IN_BLOCK + i);
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::N2NKeyInfoGetClbck(const clbck_data_t &clbck_data,
                                     int rec_status,
                                     void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "N2NKeyInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        Class_C_KeyInfo *p_key_info = (Class_C_KeyInfo *)p_attribute_data;
        m_pFabricExtendedInfo->addN2NKeyInfo(p_node, *p_key_info);
    }

    IBDIAG_RETURN_VOID;
}


///
// AR
///
void IBDiagClbck::SMPARInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPARInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        //build DB
        adaptive_routing_info *p_ar_info =
            (struct adaptive_routing_info *)p_attribute_data;

        if (p_ar_info->e) {
            if (p_ar_info->glb_groups != 1)
                m_pErrors->push_back(
                        new FabricErrNodeWrongConfig(
                                p_node, "SMPARInfoGet unsupported non global groups"));
            else if (p_ar_info->is4_mode)
                m_pErrors->push_back(new FabricErrNodeWrongConfig(p_node, "SMPARInfoGet unsupported IS4Mode"));
            else
                m_pFabricExtendedInfo->addARInfo(p_node, *p_ar_info);
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPWeightsHBFConfigGetClbck(const clbck_data_t &clbck_data,
                                            int rec_status,
                                            void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPWHBFConfigGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

        IBDIAG_RETURN_VOID;
    }

    u_int8_t block_index = (u_int8_t)(intptr_t)clbck_data.m_data2;
    AdditionalRoutingData *p_routing_data = (AdditionalRoutingData *)p_node->appData3.ptr;
    if (p_routing_data)
        p_routing_data->AddSubGroupWeights(block_index, *(whbf_config*)p_attribute_data);

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPARGroupTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPARGroupTableGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

        p_node->appData2.val = RETRIEVE_STAGE_REC_DONE;

    } else {
        //build DB
        ib_ar_grp_table *p_ar_group_table =
            (struct ib_ar_grp_table *)p_attribute_data;

        u_int16_t groupBlockNum = (u_int16_t)p_node->appData1.val;

        list_phys_ports portsList;
        bool set = false;
        for (u_int8_t blockEntry = 0;
              blockEntry < IBIS_IB_MAD_SMP_AR_GROUP_TABLE_BLOCK_SIZE ;
              blockEntry++) {
            portsList.clear();
            getPortsList(p_ar_group_table->Group[blockEntry], portsList);
            u_int16_t groupTop = p_node->getARGroupTop();

            //Old ar plugin or old fw do not support group top
            //(top equals 0) for this case the end condition is
            //empty portsList.
            if (!portsList.empty() || groupTop) {
                u_int16_t groupNumber =
                    (u_int16_t)((blockEntry + groupBlockNum *
                     IBIS_IB_MAD_SMP_AR_GROUP_TABLE_BLOCK_SIZE) /
                     (p_node->getARSubGrpsActive()+1));

                if (p_node->isARGroupTopSupported()) {
                    if (groupTop < groupNumber) {
                        break;
                    } else if (groupTop == groupNumber) {
                        p_node->appData2.val = RETRIEVE_STAGE_REC_DONE;
                    }
                }

                p_node->setARPortGroup(groupNumber, portsList);
                set = true;
            }
        }

        if (set) {

            p_node->appData1.val++;

            if (p_node->appData2.val != RETRIEVE_STAGE_REC_DONE) {
                p_node->appData2.val = RETRIEVE_STAGE_SEND;
            }

            AdditionalRoutingData *p_routing_data =
                    (AdditionalRoutingData *)p_node->appData3.ptr;

            if(p_routing_data)
                p_routing_data->AddGroupTable(groupBlockNum, p_ar_group_table);
        } else
            p_node->appData2.val = RETRIEVE_STAGE_REC_DONE;
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::getPortsList(u_int64_t pgSubBlockElement,
                               phys_port_t portOffset,
                               list_phys_ports &portsList)
{
    for (phys_port_t port = 0 ; port < 64 ; ++port) {
        if ( pgSubBlockElement & ((u_int64_t)1 << port) )
            portsList.push_back((phys_port_t)(port+portOffset));
    }
}

void IBDiagClbck::getPortsList(ib_portgroup_block_element &pgBlockElement,
                               list_phys_ports &portsList)
{
    if (pgBlockElement.SubGroup_0)
        getPortsList(pgBlockElement.SubGroup_0,
                      (phys_port_t)0, portsList);
    if (pgBlockElement.SubGroup_1)
        getPortsList(pgBlockElement.SubGroup_1,
                     (phys_port_t)(1*64), portsList);
    if (pgBlockElement.SubGroup_2)
        getPortsList(pgBlockElement.SubGroup_2,
                     (phys_port_t)(2*64), portsList);
    if (pgBlockElement.SubGroup_3)
        getPortsList(pgBlockElement.SubGroup_3,
                     (phys_port_t)(3*64), portsList);
}

void IBDiagClbck::SMPARLinearForwardingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBNode *p_node = (IBNode *)clbck_data.m_data1;
     if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;
    u_int8_t pLFT = (u_int8_t)(intptr_t)clbck_data.m_data3;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        if (p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        p_node->appData1.val = 1;

        stringstream sstr;
        sstr << "SMPARLinearForwardingTableGet (block="
             << block
             << ", pLFT="
             << +pLFT
             << ")."
             << IBIS_STATUS(rec_status);

        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        //build DB
        struct ib_ar_linear_forwarding_table_sx *p_ar_linear_forwarding_table =
            (struct ib_ar_linear_forwarding_table_sx *)p_attribute_data;

        ib_ar_lft_block_element_sx *pLidEntry = NULL;

        for (int curr_entry_in_block = 0;
              curr_entry_in_block < IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX;
              ++curr_entry_in_block) {
            lid_t lid = (lid_t)(block * IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX +
                                      curr_entry_in_block);

            //The ibdiagpath sets the pointer in order to take all LIDs.
            //It does not have fully discovered fabric yet and it cannot check
            //if a lid exists in the fabric.
            if (!clbck_data.m_data4) {
                // check that this lid is exist.
                IBPort *p_dstPort = p_node->p_fabric->getPortByLid(lid);
                IBVPort *p_dstVPort = p_node->p_fabric->getVPortByLid(lid);
                if ((!p_dstPort || !p_dstPort->getInSubFabric()) &&
                        !p_dstVPort                              &&
                        !p_node->p_fabric->isFLID(lid))
                continue;
            }

            pLidEntry =
                &p_ar_linear_forwarding_table->LidEntry[curr_entry_in_block];

            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                       "setLFTPortForLid will be called with "
                       "node=%s adding lid=%d, port=%d, plft=%d \n",
                        p_node->getName().c_str(), lid, pLidEntry->DefaultPort, pLFT);

            p_node->setLFTPortForLid(lid,
                                     pLidEntry->DefaultPort,
                                     pLFT);

            if (pLidEntry->LidState == AR_IB_LID_STATE_BOUNDED || //AR
                pLidEntry->LidState == AR_IB_LID_STATE_FREE ||           //AR
                p_node->isFREnabled()) {// Fault Routing (FLFR)

                if (pLidEntry->GroupNumber > p_node->getMaxARGroupNumber()) {
                    char buffer[512];
                    snprintf(buffer, sizeof(buffer),
                            "SMPARLFTGet undefined group number:%u for lid:%u pLFT:%u",
                            pLidEntry->GroupNumber, lid, pLFT);

                    m_pErrors->push_back(new FabricErrNodeWrongConfig(p_node, buffer));
                } else {
                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                               "setARLFTPortGroupForLid will be called with "
                               "node=%s adding lid=%d, group number=%d, plft=%d \n",
                               p_node->getName().c_str(), lid, pLidEntry->GroupNumber, pLFT);

                    p_node->setARLFTPortGroupForLid(lid,
                                             pLidEntry->GroupNumber,
                                             pLFT);
                }
            }
        }

        AdditionalRoutingData *p_routing_data =
                (AdditionalRoutingData *)p_node->appData3.ptr;

        if(p_routing_data)
            p_routing_data->AddARLFT(pLFT, block, p_ar_linear_forwarding_table);
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPRNSubGroupDirectionTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (!VerifyObject(p_routing_data, __LINE__))
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_routing_data->p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPRNSubGroupDirectionTableGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_routing_data->p_node, sstr.str()));

    } else {
        struct rn_sub_group_direction_tbl *p_rn_sub_group_direction_table =
            (struct rn_sub_group_direction_tbl *)p_attribute_data;

        u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;

        memcpy(&p_routing_data->sub_group_direction_table_vec[block],
               p_rn_sub_group_direction_table,
               sizeof(rn_sub_group_direction_tbl));
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPRNGenStringTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (!VerifyObject(p_routing_data, __LINE__))
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_routing_data->p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPRNGenStringTableGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_routing_data->p_node, sstr.str()));

    } else {
        struct rn_gen_string_tbl *p_rn_gen_string_table =
            (struct rn_gen_string_tbl *)p_attribute_data;

        u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;
        u_int8_t pLFT = (u_int8_t)(intptr_t)clbck_data.m_data3;

        memcpy(&p_routing_data->gen_string_table_vec[pLFT][block],
               p_rn_gen_string_table,
               sizeof(rn_gen_string_tbl));
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPRNGenBySubGroupPriorityGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (!VerifyObject(p_routing_data, __LINE__))
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_routing_data->p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPRNGenBySubGroupPriorityGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_routing_data->p_node, sstr.str()));

    } else {
        struct rn_gen_by_sub_group_prio *p_rn_gen_by_sg_priority =
            (struct rn_gen_by_sub_group_prio *)p_attribute_data;

        memcpy(&p_routing_data->gen_by_sub_group_priority,
               p_rn_gen_by_sg_priority,
               sizeof(rn_gen_by_sub_group_prio));
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPRNRcvStringGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (!VerifyObject(p_routing_data, __LINE__))
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_routing_data->p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPRNRcvStringGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_routing_data->p_node, sstr.str()));

    } else {
        struct rn_rcv_string *p_rn_rcv_string =
            (struct rn_rcv_string *)p_attribute_data;

        u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;

        memcpy(&p_routing_data->rcv_string_vec[block],
               p_rn_rcv_string,
               sizeof(rn_rcv_string));
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPRNXmitPortMaskGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (!VerifyObject(p_routing_data, __LINE__))
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_routing_data->p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPRNXmitPortMaskGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_routing_data->p_node, sstr.str()));

    } else {
        struct rn_xmit_port_mask *p_rn_xmit_port_mask =
            (struct rn_xmit_port_mask *)p_attribute_data;

        u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;

        memcpy(&p_routing_data->xmit_port_mask_vec[block],
               p_rn_xmit_port_mask,
               sizeof(rn_xmit_port_mask));
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VSPortRNCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSPortRNCountersClear." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VSPortRNCountersGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort* p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSPortRNCountersGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct port_rn_counters *p_rn_counters =
                (struct port_rn_counters *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addRNCounters(p_port, *p_rn_counters);
        if (rc) {
            SetLastError("Failed to add VS_RNCounters for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VSPortRoutingDecisionCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSPortRoutingDecisionCountersClear." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VSPortRoutingDecisionCountersGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort* p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSPortRoutingDecisionCountersGetClbck." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {
        port_routing_decision_counters *p_counters =
                (port_routing_decision_counters*)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addRoutingDecisionCounters(p_port, *p_counters);
        if (rc) {
            SetLastError("Failed to add VSPortRoutingDecisionCounters for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VSDiagnosticCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSDiagnosticCountersClear." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }
}

void IBDiagClbck::VSDiagnosticCountersPage0GetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    unsigned int latest_version = 0;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSDiagnosticDataPage0Get." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else if (m_pIBDiag->getLatestSupportedVersion(VS_MLNX_CNTRS_PAGE0,
                                                    latest_version)) {
        this->SetLastError("Failed to get latest version for "  \
                           "TransportErrorsAndFlows counters");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
    } else {
        struct VS_DiagnosticData *p_mlnx_cntrs =
            (struct VS_DiagnosticData *)p_attribute_data;
         // Between Backward and Current revisions only new fields can be added and
         // no old fields can be changed (by name,offset,print location and etc.),
         // so the provided version is compatible with ibdiagnet latest
         // supported version.
        if (p_mlnx_cntrs->BackwardRevision <= latest_version &&
                latest_version <= p_mlnx_cntrs->CurrentRevision) {
            struct VS_DC_TransportErrorsAndFlowsV2 p_page0_rev1;
            VS_DC_TransportErrorsAndFlowsV2_unpack(
                    &p_page0_rev1,
                    (uint8_t *)&(p_mlnx_cntrs->data_set));
            p_mlnx_cntrs->data_set.TransportErrorsAndFlowsV2  =
                p_page0_rev1;
            rc = m_pFabricExtendedInfo->
                addVSDiagnosticCountersPage0(p_port, *p_mlnx_cntrs);
            if (rc) {
                SetLastError("Failed to add DiagnosticCounters"  \
                             " TransportErrorsAndFlowsV2 for port=%s, err=%s",
                             p_port->getName().c_str(),
                             m_pFabricExtendedInfo->GetLastError());
                m_ErrorState = rc;
            }
        } else {
            FabricErrNodeMlnxCountersPageVer *p_curr_fabric_node_err =
                new FabricErrNodeMlnxCountersPageVer(
                        p_port->p_node,
                        VS_MLNX_CNTRS_PAGE0,
                        p_mlnx_cntrs->CurrentRevision,
                        latest_version);

            p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
            m_pErrors->push_back(p_curr_fabric_node_err);
        }
    }
}

void IBDiagClbck::VSDiagnosticCountersPage1GetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    unsigned int latest_version;
    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSDiagnosticDataPage1Get." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else if (m_pIBDiag->getLatestSupportedVersion(VS_MLNX_CNTRS_PAGE1, latest_version)) {

        this->SetLastError("Failed to get latest version for HCAExtendedFlows counters");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;

    } else {
        struct VS_DiagnosticData *p_mlnx_cntrs =
            (struct VS_DiagnosticData *)p_attribute_data;

         if (p_mlnx_cntrs->CurrentRevision == 0) {
             // Page not supported on device
             FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                 new FabricErrNodeNotSupportCap(p_port->p_node,
                     "This device does not support Diagnostic Counters Page 1");

             p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
             m_pErrors->push_back(p_curr_fabric_node_err);

         } else {

            struct VS_DC_Page1LatestVersion p_page1_last_ver;
            VS_DC_Page1LatestVersion_unpack(
                    &p_page1_last_ver, (uint8_t *)&(p_mlnx_cntrs->data_set));
            p_mlnx_cntrs->data_set.Page1LatestVersion  = p_page1_last_ver;
            rc = m_pFabricExtendedInfo->addVSDiagnosticCountersPage1(
                    p_port, *p_mlnx_cntrs);
            if (rc) {
                SetLastError("Failed to add DiagnosticCounters Page1" \
                             " for port=%s, err=%s",
                             p_port->getName().c_str(),
                             m_pFabricExtendedInfo->GetLastError());
                m_ErrorState = rc;
            }

            if (p_mlnx_cntrs->BackwardRevision > latest_version ||
                    latest_version > p_mlnx_cntrs->CurrentRevision) {
                // Different versions of the page between node FW and ibdiagnet
                FabricErrNodeMlnxCountersPageVer *p_curr_fabric_node_err =
                    new FabricErrNodeMlnxCountersPageVer(
                            p_port->p_node,
                            VS_MLNX_CNTRS_PAGE1,
                            p_mlnx_cntrs->CurrentRevision,
                            latest_version);

                p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                m_pErrors->push_back(p_curr_fabric_node_err);
            }
        }
    }
}

void IBDiagClbck::VSDiagnosticCountersPage255GetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    unsigned int latest_version;
    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSDiagnosticDataPage255Get." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else if (m_pIBDiag->getLatestSupportedVersion(VS_MLNX_CNTRS_PAGE255, latest_version)) {

        this->SetLastError("Failed to get latest version for DiagnosticCounters Page255");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;

    } else {
        struct VS_DiagnosticData *p_mlnx_cntrs = (struct VS_DiagnosticData *)p_attribute_data;

        if (p_mlnx_cntrs->CurrentRevision == 0) {
            // Page not supported on device
            FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                new FabricErrNodeNotSupportCap(p_port->p_node,
                    "This device does not support Diagnostic Counters Page 255");

            p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
            m_pErrors->push_back(p_curr_fabric_node_err);

        } else {

            struct VS_DC_Page255LatestVersion p_page255;
            VS_DC_Page255LatestVersion_unpack(&p_page255, (uint8_t *)&(p_mlnx_cntrs->data_set));
            p_mlnx_cntrs->data_set.Page255LatestVersion = p_page255;

            rc = m_pFabricExtendedInfo->addVSDiagnosticCountersPage255(p_port, *p_mlnx_cntrs);
            if (rc) {
                SetLastError("Failed to add DiagnosticCounters Page255 for port=%s, err=%s",
                        p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
                m_ErrorState = rc;
            }

            if (p_mlnx_cntrs->BackwardRevision > latest_version ||
                    latest_version > p_mlnx_cntrs->CurrentRevision) {
                // Different versions of the page between node FW and ibdiagnet
                FabricErrNodeMlnxCountersPageVer *p_curr_fabric_node_err =
                    new FabricErrNodeMlnxCountersPageVer(p_port->p_node,
                                                         VS_MLNX_CNTRS_PAGE255,
                                                         p_mlnx_cntrs->CurrentRevision,
                                                         latest_version);

                p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                m_pErrors->push_back(p_curr_fabric_node_err);
            }
        }
    }
}


void IBDiagClbck::SMPSwitchInfoGetClbck(const clbck_data_t &clbck_data,
                                        int rec_status,
                                        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!p_node) {
        this->SetLastError("Failed to get Node from the data provided to the callback");
        m_ErrorState = IBDIAG_ERR_CODE_CHECK_FAILED;
        IBDIAG_RETURN_VOID;
    }

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPSwitchInfoMad." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

        IBDIAG_RETURN_VOID;
    }

    SMP_SwitchInfo *p_switch_info = (SMP_SwitchInfo*)p_attribute_data;
    m_pFabricExtendedInfo->addSMPSwitchInfo(p_node, *p_switch_info);

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPPortInfoGetClbck(const clbck_data_t &clbck_data,
                                      int rec_status,
                                      void *p_attribute_data)
{
    IBDIAG_ENTER;
    //this value has been used extensively hardcoded
    //so, we define it locally
    static const int Max_Lid_Value = 0xc000;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!p_port) {
        this->SetLastError("Failed to get port the data provided to the callback");
        m_ErrorState = IBDIAG_ERR_CODE_CHECK_FAILED;
        IBDIAG_RETURN_VOID;
    }

    IBNode *p_node = p_port->p_node;
    if (!p_node) {
        //port's name depends on its node; but node object is NULL
        this->SetLastError("Failed to get node from the port provided to the callback;"
                           "port GUID: " U64H_FMT, p_port->guid_get());
        m_ErrorState = IBDIAG_ERR_CODE_CHECK_FAILED;
        IBDIAG_RETURN_VOID;
    }

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPPortInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

        IBDIAG_RETURN_VOID;
    }

    SMP_PortInfo *p_port_info = (SMP_PortInfo*)p_attribute_data;

    SMP_PortInfo *p_stored_port_info = m_pFabricExtendedInfo->getSMPPortInfo(p_port->createIndex);
    if (p_stored_port_info) {
        //replace with valid data
        *p_stored_port_info = *p_port_info;
    } else {
        //port info was not received during initial discovery, fill it now
        m_pFabricExtendedInfo->addSMPPortInfo(p_port, *p_port_info);

        //if we received port info for port 0 of the switch only now, we need to set base_lid and
        //lmc for all switch ports
        if (p_node->type == IB_SW_NODE && p_port->num == 0) {
            p_port->setCapMask(p_port_info->CapMsk);
            p_port->setCapMask2(p_port_info->CapMsk2);

            for (phys_port_t i = 0; i < p_node->numPorts; ++i) {
                IBPort *p_sw_port = p_node->getPort(i);
                if (!p_sw_port)
                    continue;

                p_sw_port->base_lid = p_port_info->LID;
                p_sw_port->lmc = p_port_info->LMC;
            }
        }
    }
    p_port->setPortInfoMadWasSent(true);

    u_int32_t cap_mask = p_port_info->CapMsk;
    u_int16_t cap_mask2 = p_port_info->CapMsk2;

    if (p_node->type == IB_SW_NODE) {
       IBPort *p_zeroPort = p_node->getPort(0);
       if (!p_zeroPort) {
           this->SetLastError("Failed to get port=0 from the node: %s",
                              p_node->name.c_str());
           m_ErrorState = IBDIAG_ERR_CODE_CHECK_FAILED;
           IBDIAG_RETURN_VOID;
       }

       cap_mask = p_zeroPort->getCapMask();
       cap_mask2 = p_zeroPort->getCapMask2();
    }

    IBLinkSpeed correct_speed = m_pFabricExtendedInfo->getCorrectSpeed(*p_port_info,
                                                                       cap_mask,
                                                                       cap_mask2);

    //direct_route_t* p_direct_route = (direct_route_t *)clbck_data.m_data2;

    p_port->set_internal_speed(correct_speed);
    p_port->set_internal_width((IBLinkWidth)p_port_info->LinkWidthActv);

    if (p_node->type != IB_SW_NODE){
        p_port->set_internal_state((IBPortState)p_port_info->PortState);

        lid_t numLidsPerPort = (lid_t)((lid_t)1 << p_port_info->LMC);
        if (p_port_info->LID >= Max_Lid_Value                        ||
            ((lid_t)p_port_info->LID + numLidsPerPort) >= Max_Lid_Value) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s invalid lid:%u lmc:%u\n",
                    p_node->name.c_str(), p_port_info->LID, p_port_info->LMC);

            m_pErrors->push_back(
                    new FabricErrNodeInvalidLid(p_node, p_port->num,
                                                p_port_info->LID, p_port_info->LMC));

            IBDIAG_RETURN_VOID;
        }

       //direct_route_t* p_direct_route = (direct_route_t *)clbck_data.m_data2;

       p_port->lmc = p_port_info->LMC;
       p_port->base_lid = p_port_info->LID;

       IBFabric *p_fabric = m_pIBDiag->GetDiscoverFabricPtr();
       //todo consider place in function
       for (lid_t l = p_port->base_lid; l < p_port->base_lid + (lid_t)numLidsPerPort; l++) {
            if (l > IB_MAX_UCAST_LID)
                 break;

            p_fabric->setLidPort(l, p_port);
       }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPNodeDescGetClbck(const clbck_data_t &clbck_data,
                                      int rec_status,
                                      void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!p_node) {
        this->SetLastError("Failed to get node info from data provided to callback");
        m_ErrorState = IBDIAG_ERR_CODE_CHECK_FAILED;

        IBDIAG_RETURN_VOID;
    }

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPNodeDescGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

        IBDIAG_RETURN_VOID;
    }

    SMP_NodeDesc *p_node_desc = (SMP_NodeDesc*)p_attribute_data;
    string error;
    int rc = m_pIBDiag->GetDiscoverFabricPtr()->renameNode(p_node,
                                                  (char*)p_node_desc->Byte, error);
    if (rc) {
        SetLastError(error.c_str());
        m_ErrorState = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN_VOID;
}


void IBDiagClbck::SMPNodeInfoGetClbck(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data){

    IBDIAG_ENTER;

    if (m_ErrorState || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) //bits [8-15] are class specific - don't need them here
        IBDIAG_RETURN_VOID;

    IBDiag::NodeInfoSendData* p_send_data = (IBDiag::NodeInfoSendData*)clbck_data.m_data1;
    if (!p_send_data)
       IBDIAG_RETURN_VOID;

    IBDiag::DirectRouteAndNodeInfo* p_route_node_info =
                    (IBDiag::DirectRouteAndNodeInfo*)clbck_data.m_data2;
    if (!p_route_node_info)
        IBDIAG_RETURN_VOID;

    SMP_NodeInfo *p_node_info = (SMP_NodeInfo*)p_attribute_data;
    p_route_node_info->node_info =  *p_node_info;
    p_route_node_info->is_filled = true;

    m_pIBDiag->SendNodeInfoMad(*p_send_data);

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::ParsePhysicalHierarchyInfo(SMP_HierarchyInfo *p_hierarchy_info, IBNode *p_node)
{
    IBDIAG_ENTER;

    if (p_hierarchy_info->ActiveLevels == 0)
        IBDIAG_RETURN_VOID;

    // 0 - Device; 1 - Board; 2 - System; 3 - Rack; 4 - Room; 5 - Campus
    vector<int> hierarchy_data;
    hierarchy_data.resize(6, -1); // 6 - number of possible records
    int board_type = -1;
    int system_type = -1;

    #define PARSE_PHYSICAL_HIERARCHY_DATA(value)                    \
    do {                                                            \
        u_int8_t i = (u_int8_t)((value) >> 28);                     \
                                                                    \
        if (i < hierarchy_data.size()) {                            \
            if (i == 1) /* 1 - Board - BoardType 0:1 */             \
                board_type = (value) & 0x3;                         \
            if (i == 2) /* 2 - System - SysType 0:1 */              \
                system_type = (value) & 0x3;                        \
                                                                    \
            hierarchy_data[i] = ((value) >> 8) & 0xFF;              \
        } else {                                                    \
            /* TODO::ADD ERROR */                                   \
        }                                                           \
    } while(0)

    switch (p_hierarchy_info->ActiveLevels) {
        case 13: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level12Data);
        case 12: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level11Data);
        case 11: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level10Data);
        case 10: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level9Data);
        case 9: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level8Data);
        case 8: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level7Data);
        case 7: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level6Data);
        case 6: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level5Data);
        case 5: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level4Data);
        case 4: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level3Data);
        case 3: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level2Data);
        case 2: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level1Data);
        case 1: PARSE_PHYSICAL_HIERARCHY_DATA(p_hierarchy_info->Level0Data);
            p_node->p_physical_hierarchy_info =
                new PhysicalHierarchyInfo(
                    hierarchy_data[5],
                    hierarchy_data[4],
                    hierarchy_data[3],
                    system_type,
                    hierarchy_data[2],
                    board_type,
                    hierarchy_data[1],
                    hierarchy_data[0]);
            break;
        default:
            // TODO::ADD ERROR
            break;
    }
    #undef PARSE_PHYSICAL_HIERARCHY_DATA

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::ParsePortHierarchyInfo(SMP_HierarchyInfo *p_hierarchy_info, IBPort *p_port)
{
    IBDIAG_ENTER;

    if (p_hierarchy_info->ActiveLevels == 0)
        IBDIAG_RETURN_VOID;

    // 0 - Split; 1 - Port; 2 - Cage; 3 - ASIC; 4 - Slot; 5 - Type; 6 - BDF;
    vector<int> hierarchy_data;
    hierarchy_data.resize(7, -1);

    #define PARSE_PORT_HIERARCHY_DATA(value)                        \
    do {                                                            \
        u_int8_t i = (u_int8_t)((value) >> 24);                     \
                                                                    \
        if (i < hierarchy_data.size()) {                            \
            switch (i) {                                            \
            case 5: /* 5 - Type - different layout */               \
                hierarchy_data[i] = (value) & 0x01;                 \
                break;                                              \
            case 4: /* 4 - Slot - save for later parsing */         \
                hierarchy_data[i] = (value) & 0xFFFFFF;             \
                break;                                              \
            default:                                                \
                hierarchy_data[i] = ((value) >> 8) & 0xFFFF;        \
                break;                                              \
            }                                                       \
        } else {                                                    \
            /* TODO::ADD ERROR */                                   \
        }                                                           \
    } while(0)

    switch (p_hierarchy_info->ActiveLevels) {
        case 13: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level12Data);
        case 12: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level11Data);
        case 11: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level10Data);
        case 10: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level9Data);
        case 9: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level8Data);
        case 8: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level7Data);
        case 7: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level6Data);
        case 6: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level5Data);
        case 5: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level4Data);
        case 4: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level3Data);
        case 3: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level2Data);
        case 2: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level1Data);
        case 1: PARSE_PORT_HIERARCHY_DATA(p_hierarchy_info->Level0Data);
            p_port->p_port_hierarchy_info =
                new PortHierarchyInfo(p_port->p_node->type, hierarchy_data,
                        HIERARCHY_TEMPLATE_GUID_PORT_0x03);
            break;
        default:
            // TODO::ADD ERROR
            break;
    }
    #undef PARSE_PORT_HIERARCHY_DATA

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::ParseXDRPortHierarchyInfo(SMP_HierarchyInfo *p_hierarchy_info, IBPort *p_port,
                          uint64_t template_guid){
    IBDIAG_ENTER;

    if (p_hierarchy_info->ActiveLevels == 0)
        IBDIAG_RETURN_VOID;

    // 0 - PortType; 1 - AsicName; 2 - IBport; 3 - SwitchCage; 4 - Port (IPIL); 5 - Split; 6 - Asic;
    // 7 - Slot; 8 - Type; 9 - BDF; 10 - Plane; 11 - NumOfPlanes; 12 - APort
    vector<int> hierarchy_data;
    hierarchy_data.resize(13, -1);

    #define PARSE_XDR_PORT_HIERARCHY_DATA(value)                    \
    do {                                                            \
        u_int8_t i = (u_int8_t)((value) >> 24);                     \
                                                                    \
        if (i < hierarchy_data.size()) {                            \
            switch (i) {                                            \
            /* 16 Bit layout */                                     \
            case 1:  /* 1 - AsicName */                             \
            case 2:  /* 2 - IBport */                               \
            case 3:  /* 3 - Switch Cage */                          \
            case 9:  /* 12 - BDF */                                 \
            case 10: /* 10 - Plane */                               \
            case 11: /* 11 - NumOfPlanes */                         \
            case 12: /* 12 - APort */                               \
                hierarchy_data[i] = ((value) >> 8) & 0xFFFF;        \
                break;                                              \
            case 8: /* 8 - Type - different layout */               \
                hierarchy_data[i] = (value) & 0xFF;                 \
                break;                                              \
            case 7: /* 7 - Slot - save the int as is */             \
                hierarchy_data[i] = static_cast<int>(value);        \
                break;                                              \
            default:                                                \
                /* 8 Bit layout */                                  \
                hierarchy_data[i] = ((value) >> 8) & 0xFF;          \
                break;                                              \
            }                                                       \
        } else {                                                    \
            /* TODO::ADD ERROR */                                   \
        }                                                           \
    } while(0)

    switch (p_hierarchy_info->ActiveLevels) {
        case 13: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level12Data);
        case 12: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level11Data);
        case 11: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level10Data);
        case 10: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level9Data);
        case 9: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level8Data);
        case 8: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level7Data);
        case 7: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level6Data);
        case 6: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level5Data);
        case 5: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level4Data);
        case 4: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level3Data);
        case 3: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level2Data);
        case 2: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level1Data);
        case 1: PARSE_XDR_PORT_HIERARCHY_DATA(p_hierarchy_info->Level0Data);
            p_port->p_port_hierarchy_info =
                new PortHierarchyInfo(p_port->p_node->type, hierarchy_data,
                        template_guid);
            break;
        default:
            // TODO::ADD ERROR
            break;
    }
    #undef PARSE_PORT_HIERARCHY_DATA

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPHierarchyInfoGetClbck(const clbck_data_t &clbck_data,
                                           int rec_status,
                                           void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_HIERARCHY_INFO_FAILED(p_port->p_node->appData1.val))
            //already reported error
            IBDIAG_RETURN_VOID;

        //mark we already checked for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_HIERARCHY_INFO;

        stringstream sstr;
        sstr << "SMPHierarchyInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {
        SMP_HierarchyInfo *p_hierarchy_info = (SMP_HierarchyInfo *)p_attribute_data;

        u_int8_t index = (u_int8_t)((uintptr_t)clbck_data.m_data3 & 0xFF);

        if (p_hierarchy_info->ActiveLevels &&
            (p_hierarchy_info->TemplateGUID == HIERARCHY_TEMPLATE_GUID_PHYSICAL  ||
             p_hierarchy_info->TemplateGUID == HIERARCHY_TEMPLATE_GUID_PORT_0x03 ||
             p_hierarchy_info->TemplateGUID == HIERARCHY_TEMPLATE_GUID_PORT_0x04 ||
             p_hierarchy_info->TemplateGUID == HIERARCHY_TEMPLATE_GUID_PORT_0x05)) {
            if (p_hierarchy_info->TemplateGUID == HIERARCHY_TEMPLATE_GUID_PHYSICAL &&
                    p_port->num == 0) {
                ParsePhysicalHierarchyInfo(p_hierarchy_info, p_port->p_node);
                // Ignore MaxActiveIndex, we are expecting only 1 physical hierarchy per node
                IBDIAG_RETURN_VOID;
            }

            if (p_hierarchy_info->TemplateGUID == HIERARCHY_TEMPLATE_GUID_PORT_0x03 &&
                    p_port->num != 0) {
                ParsePortHierarchyInfo(p_hierarchy_info, p_port);
                // Ignore MaxActiveIndex, we are expecting only 1 port hierarchy per port
                IBDIAG_RETURN_VOID;
            }

            if ((p_hierarchy_info->TemplateGUID == HIERARCHY_TEMPLATE_GUID_PORT_0x04 ||
                p_hierarchy_info->TemplateGUID == HIERARCHY_TEMPLATE_GUID_PORT_0x05) &&
                p_port->num != 0) {
                ParseXDRPortHierarchyInfo(p_hierarchy_info, p_port, p_hierarchy_info->TemplateGUID);
                // Ignore MaxActiveIndex, we are expecting only 1 port hierarchy per port
                IBDIAG_RETURN_VOID;
            }

            // If we got here, that means we got Physical or Port Hierarchy Info for wrong port
            m_pErrors->push_back(
                    new FabricErrHierarchyTemplateMismatch(
                            p_port, p_hierarchy_info->TemplateGUID, index));
        }

        if (index >= p_hierarchy_info->MaxActiveIndex)
            IBDIAG_RETURN_VOID;

        Ibis *        ibis_obj               = (Ibis *)clbck_data.m_data4;
        clbck_data_t  new_clbck_data         = clbck_data;
                      new_clbck_data.m_data3 = (void *)((uintptr_t)index + 1);

        ((ProgressBar *)clbck_data.m_p_progress_bar)->push(p_port);

        ibis_obj->SMPHierarchyInfoMadGetByDirect(
            (direct_route_t *)new_clbck_data.m_data2,
            p_port->num,
            (u_int8_t)((u_int64_t)(new_clbck_data.m_data3) & 0xFF),
            p_hierarchy_info,
            &new_clbck_data);
    }

    IBDIAG_RETURN_VOID;
}


void IBDiagClbck::SMPPortInfoExtendedGetClbck(const clbck_data_t &clbck_data,
                                              int rec_status,
                                              void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_PORT_INFO_EXTENDED_FAILED(p_port->p_node->appData1.val))
            //already reported error
            IBDIAG_RETURN_VOID;

        //mark we already checked for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_PORT_INFO_EXTENDED;

        stringstream sstr;
        sstr << "SMPPortInfoExtendedGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {
        SMP_PortInfoExtended *p_port_info_ext =
            (SMP_PortInfoExtended *)p_attribute_data;

        //update the fec mode of the port
        if (PORT_INFO_EXT_FEC_MODE_SUPPORT(p_port_info_ext->CapMsk)) {
            // convert from bit mask to real value
            IBFECMode fec_value = fec_mask2value(p_port_info_ext->FECModeActive);

            if (fec_value == IB_FEC_NA)
                m_pErrors->push_back(new FabricErrPortInvalidValue(p_port, "Got wrong fec_mode_act from FW"));

            p_port->set_fec_mode(fec_value);
        }

        //Save smp port info extended attribute in the fabric
        m_ErrorState = m_pFabricExtendedInfo->addSMPPortInfoExtended(
                                                p_port, *p_port_info_ext);
        if (m_ErrorState)
            this->SetLastError("Failed to store port info extended"
                               " for port %s, err=%s",
                    p_port->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::IBDiagSMPVirtualizationInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on "
                   "Virtualization Info on node %s\n",
                   p_port->p_node->name.c_str());

        stringstream sstr;
        sstr << "SMPVirtualizationInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {
        struct SMP_VirtualizationInfo *p_virtual_info =
            (struct SMP_VirtualizationInfo *)p_attribute_data;

        if (p_virtual_info->vport_index_top > p_virtual_info->vport_cap)
            m_pErrors->push_back(
                new FabricErrVPortIvalidTopIndex(p_port,
                                                 p_virtual_info->vport_cap,
                                                 p_virtual_info->vport_index_top));
        else {
            rc = m_pFabricExtendedInfo->addSMPVirtualizationInfo(
                    p_port, *p_virtual_info);
            if (rc) {
                SetLastError("Failed to add Virtualization Info" \
                        " for port=%s, err=%s",
                        p_port->getName().c_str(),
                        m_pFabricExtendedInfo->GetLastError());
                m_ErrorState = rc;
            }
        }
    }
}

void IBDiagClbck::IBDiagSMPVPortStateGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    u_int8_t block = (u_int8_t)(intptr_t)clbck_data.m_data2;
    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VPort State on node %s\n",
                   p_port->p_node->name.c_str());

        stringstream sstr;
        sstr << "SMPVPortStateGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {
        struct SMP_VPortState *p_vport_state_block =
            (struct SMP_VPortState *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPVPortState(
                p_port, *p_vport_state_block, block);
        if (rc) {
            SetLastError("Failed to add VPort State" \
                    " for port=%s, err=%s",
                    p_port->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::IBDiagSMPVPortInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VPort Info on node %s\n",
                   p_port->p_node->name.c_str());

        stringstream sstr;
        sstr << "SMPVPortInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {
        struct SMP_VPortInfo *p_vport_info =
            (struct SMP_VPortInfo *)p_attribute_data;
        //need the vport num to insert the IBVPort into VPorts map.
        virtual_port_t vportnum = (virtual_port_t)(intptr_t)clbck_data.m_data2;

        //insert pair port-number and  vport.
        IBVPort *p_vport = m_pIBDiag->GetDiscoverFabricPtr()->makeVPort(
                p_port,
                vportnum,
                p_vport_info->vport_guid,
                (IBPortState)p_vport_info->vport_state);

        if (p_vport_info->lid_required) {
            lid_t vlid = p_vport_info->vport_lid;

            if (vlid > IB_MAX_UCAST_LID) {
                m_pErrors->push_back(new FabricErrVPortInvalidLid(p_port, p_vport, vlid));
                delete p_vport;
                return;
            }

            p_vport->set_vlid(p_vport_info->vport_lid);
            p_vport->getIBFabricPtr()->setLidVPort(p_vport_info->vport_lid, p_vport);
        }
        else
          p_vport->set_lid_by_vport_index(p_vport_info->lid_by_vport_idx);

        p_port->VPorts.insert(pair_vportnum_vport(vportnum, p_vport));
        rc = m_pFabricExtendedInfo->addSMPVPortInfo(p_vport, *p_vport_info);
        if (rc) {
            SetLastError("Failed to add VPort Info for port=%s, vport"  \
                         " number=%d, err=%s",
                         p_port->getName().c_str(),
                         vportnum,
                         m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::IBDiagSMPVPortGUIDInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VPort GUID Info on node %s\n",
                   p_port->p_node->name.c_str());

        stringstream sstr;
        sstr << "SMPVPortGUIDInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

        p_port->p_node->appData1.val = 1;

    } else {
        IBVPort *p_vport = (IBVPort *)clbck_data.m_data2;
        struct SMP_VPortGUIDInfo *p_vport_guid_info =
            (struct SMP_VPortGUIDInfo *)p_attribute_data;
        u_int16_t guid_info_block = (u_int16_t)(intptr_t)clbck_data.m_data3;
        rc = m_pFabricExtendedInfo->addSMPVPortGUIDInfo(p_vport,
                                                        *p_vport_guid_info,
                                                        guid_info_block);
        if (rc) {
            SetLastError("Failed to add VPort GUID Info" \
                    " for vport=%s, err=%s",
                    p_vport->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::IBDiagSMPVNodeInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VNode Info on node %s\n",
                   p_port->p_node->name.c_str());

        stringstream sstr;
        sstr << "SMPVNodeInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        IBVPort *p_vport = (IBVPort *)clbck_data.m_data2;
        if (!VerifyObject(p_vport, __LINE__))
            return;

        struct SMP_VNodeInfo *p_vnode_info =
            (struct SMP_VNodeInfo *)p_attribute_data;

        IBVNode *p_vnode = m_pIBDiag->GetDiscoverFabricPtr()->makeVNode(
                                                p_vnode_info->vnode_guid,
                                                p_vnode_info->vnum_ports,
                                                p_vport,
                                                p_vnode_info->vlocal_port_num);
        if (!p_vnode) {
            SetLastError("Failed to create new vnode for port=%s vport=%s",
                         p_port->getName().c_str(),
                         p_vport->getName().c_str());
            IBDIAG_RETURN_VOID;
        }
        p_vport->setVNodePtr(p_vnode);
        p_vport->set_vlocal_port_num(p_vnode_info->vlocal_port_num);
        rc = m_pFabricExtendedInfo->addSMPVNodeInfo(p_vnode, *p_vnode_info);
        if (rc) {
            SetLastError("Failed to add VNode Info" \
                         " for port=%s, vport=%s, err=%s",
                        p_port->getName().c_str(),
                        p_vport->getName().c_str(),
                        m_pFabricExtendedInfo->GetLastError());
        }
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::IBDiagSMPVNodeDescriptionGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPVNodeDescriptionGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        IBVNode *p_vnode = (IBVNode *)clbck_data.m_data2;
        if (!VerifyObject(p_vnode, __LINE__))
            return;

        struct SMP_NodeDesc *p_vnode_description =
            (struct SMP_NodeDesc *)p_attribute_data;

        p_vnode->setDescription((char *)p_vnode_description->Byte);
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::IBDiagSMPVPortPKeyGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VPort PKey Table on node %s\n",
                   p_port->p_node->name.c_str());

        stringstream sstr;
        sstr << "SMPVPortPKeyTableGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {
        IBVPort *p_vport = (IBVPort *)clbck_data.m_data2;
        if (!VerifyObject(p_vport, __LINE__))
            return;

        struct SMP_PKeyTable *p_vport_pkey_tbl =
            (struct SMP_PKeyTable *)p_attribute_data;
        u_int16_t pkey_block = (u_int16_t)(intptr_t)clbck_data.m_data3;
        rc = m_pFabricExtendedInfo->addSMPVPortPKeyTable(p_vport,
                                                         *p_vport_pkey_tbl,
                                                         pkey_block);
        if (rc) {
            SetLastError("Failed to add VPort PKey Table" \
                    " for port=%s, err=%s",
                    p_vport->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::IBDiagSMPQoSConfigVLGetClbck(const clbck_data_t &clbck_data,
                                               int rec_status,
                                               void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPQosConfigVLGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct SMP_QosConfigVL *p_qos_config_vl = (struct SMP_QosConfigVL *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPQosConfigVL(p_port, *p_qos_config_vl);

        if (rc) {
            SetLastError("Failed to add SMP_QosConfigVL for node=%s, port=%u, err=%s",
                         p_port->p_node->getName().c_str(),
                         (u_int8_t)p_port->num,
                         m_pFabricExtendedInfo->GetLastError());

            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::IBDiagSMPQoSConfigSLGetClbck(const clbck_data_t &clbck_data,
                                               int rec_status,
                                               void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPQosConfigSLGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct SMP_QosConfigSL *p_qos_config_sl = (struct SMP_QosConfigSL *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPQosConfigSL(p_port, *p_qos_config_sl);

        if (rc) {
            SetLastError("Failed to add SMP_QosConfigSL for node=%s, port=%u, err=%s",
                          p_port->p_node->getName().c_str(),
                          (u_int8_t)p_port->num,
                          m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::IBDiagSMPVPortQoSConfigSLGetClbck(const clbck_data_t &clbck_data,
                                                    int rec_status,
                                                    void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    IBVPort *p_vport = (IBVPort *)clbck_data.m_data2;
    if (!VerifyObject(p_vport, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPQosConfigSLGet for vport: "
             << (u_int32_t)p_vport->getVPortNum()
             << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct SMP_QosConfigSL *p_qos_config_sl = (struct SMP_QosConfigSL *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPVPortQosConfigSL(p_vport, *p_qos_config_sl);

        if (rc) {
            SetLastError("Failed to add SMP_QosConfigSL for vport=%hu node=%s, port=%u, err=%s",
                          p_vport->getVPortNum(),
                          p_port->p_node->getName().c_str(),
                          (u_int8_t)p_port->num,
                          m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::IBDiagSMPTempSensingGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_node, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPTempSensingGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct SMP_TempSensing *p_temp_sense =
            (struct SMP_TempSensing *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPTempSensing(
                p_node, *p_temp_sense);
        if (rc) {
            SetLastError("Failed to add Virtualization Info" \
                    " for node=%s, err=%s",
                    p_node->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::SMPRouterInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_node, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPRouterInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct SMP_RouterInfo *p_router_info =
            (struct SMP_RouterInfo *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPRouterInfo(
                p_node, *p_router_info);
        if (rc) {
            SetLastError("Failed to add Router Info" \
                    " for node=%s, err=%s",
                    p_node->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::SMPAdjRouterTableGetClbck(const clbck_data_t &clbck_data,
                                            int rec_status,
                                            void *p_attribute_data)
{
    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_node, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPAdjSiteLocalSubnTblGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        struct SMP_AdjSiteLocalSubnTbl *p_router_tbl =
            (struct SMP_AdjSiteLocalSubnTbl *)p_attribute_data;
        u_int8_t rtr_block = (u_int8_t)(intptr_t)clbck_data.m_data2;

        rc = m_pFabricExtendedInfo->addSMPAdjSiteLocalSubnTbl(p_node,
                                                              *p_router_tbl,
                                                              rtr_block);
        if (rc) {
            SetLastError("Failed to add SMPAdjSiteLocalSubnTbl router table"
                         " for node=%s, err=%s",
                         p_node->getName().c_str(),
                         m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::SMPNextHopRouterTableGetClbck(const clbck_data_t &clbck_data,
                                                int rec_status,
                                                void *p_attribute_data)
{
    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_node, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPNextHopRouterTableGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {
        struct SMP_NextHopTbl *p_router_tbl =
            (struct SMP_NextHopTbl*)p_attribute_data;
        u_int32_t rtr_block = (u_int32_t)(intptr_t)clbck_data.m_data2;

        rc = m_pFabricExtendedInfo->addSMPNextHopTbl(p_node,
                                                     *p_router_tbl,
                                                     rtr_block);
        if (rc) {
            SetLastError("Failed to add SMPNextHop router table"
                         " for node=%s, err=%s",
                         p_node->getName().c_str(),
                         m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::SMPAdjRouterLIDInfoTableGetClbck(const clbck_data_t &clbck_data,
                                                   int rec_status,
                                                   void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPAdjRouterLIDInfoTableGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

        IBDIAG_RETURN_VOID;
    }

    SMP_AdjSubnetsRouterLIDInfoTable *p_subnet_tbl =
            (SMP_AdjSubnetsRouterLIDInfoTable*)p_attribute_data;
    u_int8_t block_index = (u_int8_t)(intptr_t)clbck_data.m_data2;

    int rc = m_pFabricExtendedInfo->addSMPAdjSubnetRouterLIDInfoTbl(p_node,
                                                                *p_subnet_tbl,
                                                                block_index);
    if (rc) {
        SetLastError("Failed to add SMP_AdjSubnetsRouterLIDInfoTable for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        m_ErrorState = rc;
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPARGroupToRouterLidTableGetClbck(const clbck_data_t &clbck_data,
                                                     int rec_status,
                                                     void *p_attribute_data)
{
    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPARGroupToRouterLidTableGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

        IBDIAG_RETURN_VOID;
    }

    SMP_ARGroupToRouterLIDTable *p_table = (SMP_ARGroupToRouterLIDTable*)p_attribute_data;
    u_int8_t block_index = (u_int8_t)(intptr_t)clbck_data.m_data2;
    int rc = m_pFabricExtendedInfo->addARGroupToRouterLIDTbl(p_node, *p_table, block_index);
    if (rc) {
        SetLastError("Failed to add SMP_ARGroupToRouterLIDTable for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        m_ErrorState = rc;
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPRouterLIDTableGetClbck(const clbck_data_t &clbck_data,
                                            int rec_status,
                                            void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "SMPRouterLIDTablGetClbck." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

        IBDIAG_RETURN_VOID;
    }

    SMP_RouterLIDTable *p_router_lid_table = (SMP_RouterLIDTable*)p_attribute_data;
    u_int8_t block_index = (u_int8_t)(intptr_t)clbck_data.m_data2;

    int rc = m_pFabricExtendedInfo->addSMPRouterLIDITbl(p_node,
                                                        *p_router_lid_table,
                                                        block_index);
    if (rc) {
        SetLastError("Failed to add SMP_RouterLIDTable for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        m_ErrorState = rc;
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SharpMngrClassPortInfoClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data2);
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    if (!p_node) {
        SetLastError("Failed to get IBNode ptr");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
        return;
    }

    if (!(rec_status & 0x00ff)) { //bits [8-15] are class specific - don't need them here
        // Not checking the return status as it cam
        if (!m_p_sharp_mngr) {
            SetLastError("Failed to get SharpMngr ptr");
            m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
        } else {
            m_p_sharp_mngr->AddSharpSupportedNodes(p_node);
            m_p_sharp_mngr->AddClassPortInfo(p_port->base_lid,
                                             (IB_ClassPortInfo *)p_attribute_data);
        }

    } else {

        m_num_errors++;

        stringstream sstr;
        sstr << "AMClassPortInfoGetClbck." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));
    }
}


void IBDiagClbck::SharpMngrTreeConfigClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    SharpAggNode *p_sharp_agg_node = (SharpAggNode *)clbck_data.m_data1;
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar,
                                               p_sharp_agg_node->GetIBPort());
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!p_port) {
        SetLastError("Failed to get IBPort for Aggregation Node");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
        IBDIAG_RETURN_VOID;
    }

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "AMTreeConfigGetClbck." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_port->p_node, sstr.str()));

        m_num_errors++;

    } else {

        if (!m_p_sharp_mngr) {
            SetLastError("Failed to get SharpMngr ptr");
            m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
            IBDIAG_RETURN_VOID;
        }

        struct AM_TreeConfig *p_tree_config =
            (struct AM_TreeConfig *)p_attribute_data;

        u_int16_t tree_id = (u_int16_t)(uintptr_t)clbck_data.m_data2;
        //if tree_state is DISEBLED we don't need to create new TreeNode for this TreeID
        if (!p_tree_config->tree_state) {
            IBDIAG_RETURN_VOID;
        }

        if (tree_id != p_tree_config->tree_id) {
            m_num_errors++;
            m_pErrors->push_back(new SharpErrNodeTreeIDNotMatchGetRespondTreeID(p_port->p_node, tree_id));
        }
        // The number of children that was read in the previous calls
        u_int8_t total_children_query_num = (u_int8_t)(uintptr_t)clbck_data.m_data3;

        SharpTreeNode *p_sharp_tree_node =
            new SharpTreeNode(p_sharp_agg_node, tree_id, *p_tree_config);

        int rc  = p_sharp_agg_node->AddSharpTreeNode(p_sharp_tree_node, tree_id);
        if (rc) {
            SetLastError("Failed to add SharpTreeNode for node=%s",
                    p_port->p_node->getName().c_str());
            m_ErrorState = rc;
        }

        m_p_sharp_mngr->UpdateFabricMaxTreeIdx(tree_id);

        if (p_tree_config->parent_qpn)
            p_sharp_tree_node->SetSharpParentTreeEdge(new SharpTreeEdge(p_tree_config->parent_qpn, 0));

        else {
            int rc = p_sharp_agg_node->AddTreeRoot(tree_id, p_sharp_tree_node);

            // for old sharp version (no ad-hoc) we set a warning if the root tree_id
            // already exist in the fabric and not just on the AN.
            bool is_ad_hoc = (bool)(uintptr_t)clbck_data.m_data4;
            if (!is_ad_hoc)
                rc = m_p_sharp_mngr->AddRootID(tree_id);

            if (rc) {
                m_num_warnings++;
                m_pErrors->push_back(new SharpErrRootTreeNodeAlreadyExistsForTreeID(p_port->p_node, tree_id));
            }
        }

        // total_children_query_num - The total number of children for all the call
        // for this node and tree_id.
        // Total_children_query_num value need to go to the next call of this
        // Clbck function.
        u_int8_t i = 0;
        for (i = 0;
             (i < p_tree_config->num_of_children) && (i < MAX_CHILD_IDX_IN_TREE_CONFIG_MAD);
             i++) {
            SharpTreeEdge *p_sharp_tree_eage =
                new SharpTreeEdge(p_tree_config->child_qp[i].child_qpn,
                                  p_tree_config->child_qp[i].child_index);
            rc = p_sharp_tree_node->AddSharpTreeEdge(p_sharp_tree_eage,
                                                     (u_int8_t)(i + total_children_query_num));
        }

        total_children_query_num = (u_int8_t)(total_children_query_num + i);
        // need to check against the num_of_children
        if (p_tree_config->record_locator != p_sharp_agg_node->GetMaxNumQps()) {
            struct AM_TreeConfig tree_config;
            CLEAR_STRUCT(tree_config);
            clbck_data_t tree_config_clbck_data;
            CLEAR_STRUCT(tree_config_clbck_data);
            tree_config_clbck_data.m_p_obj = &ibDiagClbck;
            tree_config_clbck_data.m_handle_data_func =
                &forwardClbck<IBDiagClbck, &IBDiagClbck::SharpMngrTreeConfigClbck>;
            tree_config_clbck_data.m_data1 = p_sharp_agg_node;
            tree_config_clbck_data.m_data2 = (void*)(uintptr_t)tree_id;
            tree_config_clbck_data.m_data3 = (void*)(uintptr_t)total_children_query_num;
            tree_config_clbck_data.m_p_progress_bar = clbck_data.m_p_progress_bar;

            tree_config.record_locator = p_tree_config->record_locator;
            tree_config.tree_id = tree_id;
            tree_config.num_of_children = MAX_CHILD_IDX_IN_TREE_CONFIG_MAD;

            ((ProgressBar *)clbck_data.m_p_progress_bar)->push(p_port);
            m_pIBDiag->GetIbisPtr()->AMTreeConfigGet(p_port->base_lid,
                                                     DEFAULT_SL,
                                                     p_port->am_key.GetKey(),
                                                     p_sharp_agg_node->getClassVersion(),
                                                     &tree_config,
                                                     &tree_config_clbck_data);
        }

        if (rc) {
          SetLastError("Failed to add SharpTreeEdge for node=%s, err=%s",
                  p_port->p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::SharpMngrQPCConfigClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    SharpAggNode *p_sharp_agg_node = (SharpAggNode *)clbck_data.m_data1;
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar,
                                               p_sharp_agg_node->GetIBPort());
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!p_port) {
        SetLastError("Failed to get IBPort for Aggregation Node");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
        IBDIAG_RETURN_VOID;
    }

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "AMQPCConfigGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_port->p_node, sstr.str()));

        m_num_errors++;

    } else {
        struct AM_QPCConfig *p_qpc_config =
            (struct AM_QPCConfig *)p_attribute_data;

         SharpTreeEdge *pSharpTreeEdge = (SharpTreeEdge *)clbck_data.m_data2;
         pSharpTreeEdge->SetQPCConfig(*p_qpc_config);
    }
}

void IBDiagClbck::SharpMngrResetPerfCountersClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    SharpAggNode *p_sharp_agg_node = (SharpAggNode *)clbck_data.m_data1;
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar,
                                               p_sharp_agg_node->GetIBPort());
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (!p_port) {
             SetLastError("Failed to get IBPort for Aggregation Node");
             m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
             IBDIAG_RETURN_VOID;
        }

        m_num_errors++;

        stringstream sstr;
        sstr << "AMPerfCountersSet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_port->p_node, sstr.str()));
    }
}

void IBDiagClbck::SharpMngrPerfCountersClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    SharpAggNode *p_sharp_agg_node = (SharpAggNode *)clbck_data.m_data1;
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar,
                                               p_sharp_agg_node->GetIBPort());
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!p_port) {
        SetLastError("Failed to get IBPort for Aggregation Node");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
        IBDIAG_RETURN_VOID;
    }

    counters_mode mode = (counters_mode)(uintptr_t)clbck_data.m_data2;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        m_num_errors++;

        stringstream sstr;
        sstr << "AMPerfCountersGet - Mode: "
             << sharp_counters_mode2char(mode)
             << IBIS_STATUS(rec_status);

        m_pErrors->push_back(new FabricErrNodeNotRespond(p_port->p_node, sstr.str()));

    } else {
        struct AM_PerformanceCounters *p_perf_counters =
            (struct AM_PerformanceCounters *)p_attribute_data;

        p_sharp_agg_node->SetPerfCounters(p_perf_counters, mode);
    }
}

void IBDiagClbck::SharpMngrHBAPerfCountersClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    SharpAggNode *p_sharp_agg_node = (SharpAggNode *)clbck_data.m_data1;
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar,
                                               p_sharp_agg_node->GetIBPort());
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!p_port) {
        SetLastError("Failed to get IBPort for Aggregation Node");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
        IBDIAG_RETURN_VOID;
    }

    IBPort *p_sw_port = (IBPort *)clbck_data.m_data3;
    if (!p_sw_port) {
            SetLastError("Failed to get IBPort for Switch connected to Aggregation Node: "
                         U64H_FMT, p_port->p_node->guid_get());
            m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
            IBDIAG_RETURN_VOID;
    }

    counters_mode mode = (counters_mode)(uintptr_t)clbck_data.m_data2;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "AMHBAPerfCountersGet - Mode: "
             << sharp_counters_mode2char(mode)
             << IBIS_STATUS(rec_status);

        m_pErrors->push_back(new FabricErrPortNotRespond(p_sw_port, sstr.str()));

        m_num_errors++;

    } else {
        struct AM_PerformanceCounters *p_hba_perf_counters =
            (struct AM_PerformanceCounters *)p_attribute_data;
        p_sharp_agg_node->SetHBAPerfCounters(p_hba_perf_counters, p_sw_port);
    }
}

void IBDiagClbck::SharpMngrANInfoClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    SharpAggNode *p_sharp_agg_node = (SharpAggNode *)clbck_data.m_data1;
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar,
                                               p_sharp_agg_node->GetIBPort());
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!p_port) {
        SetLastError("Failed to get IBPort for Aggregation Node");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
        IBDIAG_RETURN_VOID;
    }

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "AMANInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_port->p_node, sstr.str()));

        m_num_errors++;

    } else {
        struct AM_ANInfo *p_an_info =
            (struct AM_ANInfo *)p_attribute_data;

        p_sharp_agg_node->SetANInfo(p_an_info);
    }
}

void IBDiagClbck::SharpMngrANActiveJobsClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    SharpAggNode *p_sharp_agg_node = (SharpAggNode *)clbck_data.m_data1;
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar,
                                               p_sharp_agg_node->GetIBPort());
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!p_port) {
        SetLastError("Failed to get IBPort for Aggregation Node");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
        IBDIAG_RETURN_VOID;
    }

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "ANActiveJobsGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_port->p_node, sstr.str()));

        m_num_errors++;

    } else {
        AM_ANActiveJobs *p_an_jobs = (AM_ANActiveJobs*)p_attribute_data;
        p_sharp_agg_node->SetANActiveJobs(p_an_jobs);
    }
}


/*** CC ***/
void IBDiagClbck::CCEnhancedCongestionInfoGetClbck(const clbck_data_t &clbck_data,
                                                   int rec_status,
                                                   void *p_attribute_data)
{
    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_node, __LINE__))
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    int rc = 0;

    if (rec_status)
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, "CCEnhancedCongestionInfoGet"));

    else {

        CC_EnhancedCongestionInfo *p_cc_info = (CC_EnhancedCongestionInfo *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addCCEnhancedCongestionInfo(p_node, *p_cc_info);

        if (rc) {
            SetLastError("Failed to add CC_EnhancedCongestionInfo for node=%s, err=%s",
                          p_node->getName().c_str(),
                          m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

// CC switch
void IBDiagClbck::CCSwitchGeneralSettingsGetClbck(const clbck_data_t &clbck_data,
                                                  int rec_status,
                                                  void *p_attribute_data)
{
    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_node, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CCSwitchGeneralSettingsGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        CC_CongestionSwitchGeneralSettings *p_cc_switch_general_settings =
                (CC_CongestionSwitchGeneralSettings *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addCCSwitchGeneralSettings(p_node, *p_cc_switch_general_settings);

        if (rc) {
            SetLastError("Failed to add CC_CongestionSwitchGeneralSettings for node=%s, err=%s",
                          p_node->getName().c_str(),
                          m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::CCPortProfileSettingsGetClbck(const clbck_data_t &clbck_data,
                                                int rec_status,
                                                void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    u_int8_t vl = (u_int8_t)(u_int64_t)clbck_data.m_data2;
    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CCPortProfileSettingsGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        CC_CongestionPortProfileSettings *p_cc_port_profile_settings =
                (CC_CongestionPortProfileSettings *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addCCPortProfileSettings(p_port, vl, *p_cc_port_profile_settings);

        if (rc) {
            SetLastError("Failed to add CC_CongestionPortProfileSettings for port=%s, VL=%u, err=%s",
                          p_port->getName().c_str(), vl,
                          m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::CCSLMappingSettingsGetClbck(const clbck_data_t &clbck_data,
                                              int rec_status,
                                              void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CCSLMappingSettingsGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        CC_CongestionSLMappingSettings *p_cc_sl_mapping_settings =
                (CC_CongestionSLMappingSettings *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addCCSLMappingSettings(p_port, *p_cc_sl_mapping_settings);

        if (rc) {
            SetLastError("Failed to add CC_CongestionSLMappingSettings for port=%s, err=%s",
                          p_port->getName().c_str(),
                          m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

// CC HCA
void IBDiagClbck::CCHCAGeneralSettingsGetClbck(const clbck_data_t &clbck_data,
                                               int rec_status,
                                               void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CCHCAGeneralSettingsGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        CC_CongestionHCAGeneralSettings *p_cc_hca_general_settings =
                (CC_CongestionHCAGeneralSettings *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addCCHCAGeneralSettings(p_port, *p_cc_hca_general_settings);

        if (rc) {
            SetLastError("Failed to add CC_CongestionHCAGeneralSettings for port=%s, err=%s",
                          p_port->getName().c_str(),
                          m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::CCHCARPParametersGetClbck(const clbck_data_t &clbck_data,
                                            int rec_status,
                                            void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CCHCARPParametersGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        CC_CongestionHCARPParameters *p_cc_hca_rp_parameters =
                (CC_CongestionHCARPParameters *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addCCHCARPParameters(p_port, *p_cc_hca_rp_parameters);

        if (rc) {
            SetLastError("Failed to add CC_CongestionHCARPParameters for port=%s, err=%s",
                          p_port->getName().c_str(),
                          m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::CCHCANPParametersGetClbck(const clbck_data_t &clbck_data,
                                            int rec_status,
                                            void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CCHCANPParametersGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        CC_CongestionHCANPParameters *p_cc_hca_np_parameters =
                (CC_CongestionHCANPParameters *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addCCHCANPParameters(p_port, *p_cc_hca_np_parameters);

        if (rc) {
            SetLastError("Failed to add CC_CongestionHCANPParameters for port=%s, err=%s",
                          p_port->getName().c_str(),
                          m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::CCHCAStatisticsQueryGetClbck(const clbck_data_t &clbck_data,
                                               int rec_status,
                                               void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CCHCAStatisticsQueryGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        CC_CongestionHCAStatisticsQuery *p_cc_hca_statistics_query =
                (CC_CongestionHCAStatisticsQuery *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addCCHCAStatisticsQuery(p_port, *p_cc_hca_statistics_query);

        if (rc) {
            SetLastError("Failed to add CC_CongestionHCAStatisticsQuery for port=%s, err=%s",
                          p_port->getName().c_str(),
                          m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::CCPerSLVLGetClbck(const clbck_data_t &clbck_data,
                                    int rec_status,
                                    void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_pFabricExtendedInfo ||
        !m_p_capability_module)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    class CountersPerSLVL * p_cntrs_per_slvl = (CountersPerSLVL*)clbck_data.m_data2;
    IBNode* p_node = p_port->p_node;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_CC_SLVL_CNTR_FAILED(p_node, p_cntrs_per_slvl->GetAttrCapBit()))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_node->appData1.val |= p_cntrs_per_slvl->GetAttrCapBit();

        stringstream sstr;
        sstr << "CCPerSLVLCountersGet "
             << p_cntrs_per_slvl->GetCntrHeader()
             << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct PM_PortRcvXmitCntrsSlVl *p_per_slvl_ctrs =
            (PM_PortRcvXmitCntrsSlVl *)p_attribute_data;

        pair_ibport_slvl_cntr_data_t sl_vl_data(p_port, *p_per_slvl_ctrs);

        p_cntrs_per_slvl->m_set_port_data_update.insert(sl_vl_data);

        if (m_ErrorState)
            this->SetLastError("Failed to store CC per slvl counter for port %s, err=%s",
                               p_port->getName().c_str(),
                               m_pFabricExtendedInfo->GetLastError());
    }

    IBDIAG_RETURN_VOID;
}

// CC HCA Algo
void IBDiagClbck::CC_HCA_AlgoConfigSupGetClbck(const clbck_data_t &clbck_data,
                                               int rec_status,
                                               void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (!p_port)
        return;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CC_HCA_AlgoConfigSupGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        CC_CongestionHCAAlgoConfig *p_cc_hca_algo_config =
                (CC_CongestionHCAAlgoConfig *)p_attribute_data;

        rc = m_pFabricExtendedInfo->addCC_HCA_AlgoConfigSup(p_port,
                                                            *p_cc_hca_algo_config);

        if (rc) {
            SetLastError("Failed to add CC_CongestionHCAAlgoConfigSup for port=%s, err=%s",
                         p_port->getName().c_str(),
                         m_pFabricExtendedInfo->GetLastError());

            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::CC_HCA_AlgoConfigGetClbck(const clbck_data_t &clbck_data,
                                            int rec_status,
                                            void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (!p_port)
        return;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CC_HCA_AlgoConfigGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        CC_CongestionHCAAlgoConfig *p_cc_hca_algo_config =
                (CC_CongestionHCAAlgoConfig *)p_attribute_data;
        u_int8_t algo_slot = (u_int8_t)(intptr_t)clbck_data.m_data2;

        rc = m_pFabricExtendedInfo->addCC_HCA_AlgoConfig(p_port,
                                                         *p_cc_hca_algo_config,
                                                         algo_slot);

        if (rc) {
            SetLastError("Failed to add CC_CongestionHCAAlgoConfig for port=%s, err=%s",
                         p_port->getName().c_str(),
                         m_pFabricExtendedInfo->GetLastError());

            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::CC_HCA_AlgoConfigParamsGetClbck(const clbck_data_t &clbck_data,
                                                  int rec_status,
                                                  void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (!p_port)
        return;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CC_HCA_AlgoConfigParamsGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        CC_CongestionHCAAlgoConfigParams *p_cc_hca_algo_config_params =
                (CC_CongestionHCAAlgoConfigParams *)p_attribute_data;
        u_int8_t algo_slot = (u_int8_t)(intptr_t)clbck_data.m_data2;

        rc = m_pFabricExtendedInfo->addCC_HCA_AlgoConfigParams(p_port,
                                                               *p_cc_hca_algo_config_params,
                                                               algo_slot);

        if (rc) {
            SetLastError("Failed to add CC_CongestionHCAAlgoConfigParams for port=%s, err=%s",
                         p_port->getName().c_str(),
                         m_pFabricExtendedInfo->GetLastError());

            m_ErrorState = rc;
        }
    }
}

void IBDiagClbck::CC_HCA_AlgoCountersGetClbck(const clbck_data_t &clbck_data,
                                              int rec_status,
                                              void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (!p_port)
        return;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    int rc = 0;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "CC_HCA_AlgoCountersGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        CC_CongestionHCAAlgoCounters *p_cc_hca_algo_counters =
                (CC_CongestionHCAAlgoCounters *)p_attribute_data;
        u_int8_t algo_slot = (u_int8_t)(intptr_t)clbck_data.m_data2;

        rc = m_pFabricExtendedInfo->addCC_HCA_AlgoCounters(p_port,
                                                           *p_cc_hca_algo_counters,
                                                           algo_slot);

        if (rc) {
            SetLastError("Failed to add CC_CongestionHCAAlgoCounters for port=%s, err=%s",
                         p_port->getName().c_str(),
                         m_pFabricExtendedInfo->GetLastError());

            m_ErrorState = rc;
        }
    }
}

// Fast Recovery
void IBDiagClbck::VSCreditWatchdogTimeoutCountersGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort* p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSCreditWatchdogTimeoutCountersGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct VS_CreditWatchdogTimeoutCounters *p_credit_wd_counters =
                (struct VS_CreditWatchdogTimeoutCounters *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addCreditWatchdogTimeoutCounters(p_port, *p_credit_wd_counters);
        if (rc) {
            SetLastError("Failed to add VS_CreditWatchdogTimeoutCounters for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VSCreditWatchdogTimeoutCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        stringstream sstr;
        sstr << "VSCreditWatchdogTimeoutCountersClearClbck." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }
}

void IBDiagClbck::VSFastRecoveryCountersGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort* p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_FR_COUNTERS_FAILED(p_port->p_node->appData1.val))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_FR_COUNTERS;

        stringstream sstr;
        sstr << "VSFastRecoveryCountersGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct VS_FastRecoveryCounters *p_fast_recovery_counters =
                (struct VS_FastRecoveryCounters *)p_attribute_data;

        u_int8_t trigger = (u_int8_t)(intptr_t)clbck_data.m_data2;

        int rc = m_pFabricExtendedInfo->addFastRecoveryCounters(p_port,
                                                                *p_fast_recovery_counters,
                                                                trigger);
        if (rc) {
            SetLastError("Failed to add VS_FastRecoveryCounters for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VSFastRecoveryCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    if (!VerifyObject(p_port, __LINE__))
        return;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_FR_COUNTERS_FAILED(p_port->p_node->appData1.val))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_FR_COUNTERS;

        stringstream sstr;
        sstr << "VSFastRecoveryCountersClearClbck." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));
    }
}

void IBDiagClbck::VSPortGeneralCountersClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_PORT_GENERAL_COUNTERS_FAILED(p_port->p_node->appData1.val))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_PORT_GENERAL_COUNTERS;
        stringstream sstr;
        sstr << "VSPortGeneralCountersGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct VS_PortGeneralCounters *p_vs_port_general_counters =
            (struct VS_PortGeneralCounters *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addVSPortGeneralCounters(p_port, *p_vs_port_general_counters);
        if (rc) {
            SetLastError("Failed to add VS_PortGeneralCounters for port=%s, err=%s",
                         p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VSPerformanceHistogramInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode* p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        stringstream sstr;
        sstr << "VSPerformanceHistogramInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct VS_PerformanceHistogramInfo *p_perf_hist =
            (struct VS_PerformanceHistogramInfo *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addPerformanceHistogramInfo(p_node, *p_perf_hist);
        if (rc) {
            SetLastError("Failed to add VS_PerformanceHistogramInfo for node=%s, err=%s",
                         p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VSPerformanceHistogramBufferControlClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort* p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    
    u_int8_t vl = (u_int8_t)(intptr_t)clbck_data.m_data2;
    u_int8_t dir = (u_int8_t)(intptr_t)clbck_data.m_data3;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        stringstream sstr;
        sstr << "VSPerformanceHistogramBufferControlGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortVLNotRespond(p_port, vl, sstr.str()));

    } else {

        struct VS_PerformanceHistogramBufferControl *p_perf_hist_buff_info =
            (struct VS_PerformanceHistogramBufferControl *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addPerformanceHistogramBufferControl(p_port, *p_perf_hist_buff_info, vl, dir);
        if (rc) {
            SetLastError("Failed to add VS_VSPerformanceHistogramBufferControl for port=%s, err=%s",
                         p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::VS_PerformanceHistogramBufferDataClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
        IBDIAG_ENTER;

    IBPort* p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    u_int8_t vl = (u_int8_t)(intptr_t)clbck_data.m_data2;
    u_int8_t dir = (u_int8_t)(intptr_t)clbck_data.m_data3;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        stringstream sstr;
        sstr << "VSPerformanceHistogramBufferDataGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortVLNotRespond(p_port,vl, sstr.str()));

    } else {

        struct VS_PerformanceHistogramBufferData *p_perf_hist_buff_info_data =
            (struct VS_PerformanceHistogramBufferData *)p_attribute_data;


        int rc = m_pFabricExtendedInfo->addPerformanceHistogramBufferData(p_port, *p_perf_hist_buff_info_data, vl, dir);
        if (rc) {
            SetLastError("Failed to add VS_VSPerformanceHistogramBufferData for port=%s, err=%s",
                         p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;

}

void IBDiagClbck::SMP_ProfilesConfigGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode* p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_FR_PROFILES_CONFIG_FAILED(p_node->appData1.val))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_node->appData1.val |= NOT_SUPPORT_FR_PROFILES_CONFIG;

        stringstream sstr;
        sstr << "SMP_ProfilesConfigGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct SMP_ProfilesConfig *p_profiles_config =
            (struct SMP_ProfilesConfig *)p_attribute_data;

        u_int32_t block = (u_int32_t)(intptr_t)clbck_data.m_data2;

        for (int i = 0;
                i < PORT_PROFILES_IN_BLOCK &&
                    (block * PORT_PROFILES_IN_BLOCK + i) <= p_node->numPorts;
                ++i) {

            IBPort* p_port = p_node->getPort((phys_port_t)(block * PORT_PROFILES_IN_BLOCK + i));
            if (p_port && p_port->isSpecialPort())
                continue;

            p_node->fast_recovery_profiles.insert(p_profiles_config->port_profile[i]);
        }

        int rc = m_pFabricExtendedInfo->addProfilesConfig(p_node, *p_profiles_config, block);
        if (rc) {
            SetLastError("Failed to add SMP_ProfilesConfig for node=%s, err=%s",
                         p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMP_CreditWatchdogConfigGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode* p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_FR_CREDIT_WATCHDOG_CONFIG_FAILED(p_node->appData1.val))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_node->appData1.val |= NOT_SUPPORT_FR_CREDIT_WATCHDOG_CONFIG;

        stringstream sstr;
        sstr << "SMP_CreditWatchdogConfigGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct SMP_CreditWatchdogConfig *p_credit_wd_config =
            (struct SMP_CreditWatchdogConfig *)p_attribute_data;

        u_int32_t profile = (u_int32_t)(intptr_t)clbck_data.m_data2;

        int rc = m_pFabricExtendedInfo->addCreditWatchdogConfig(p_node, *p_credit_wd_config, profile);
        if (rc) {
            SetLastError("Failed to add SMP_CreditWatchdogConfig for node=%s, err=%s",
                         p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMP_BERConfigGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode* p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_FR_BER_CONFIG_FAILED(p_node->appData1.val))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_node->appData1.val |= NOT_SUPPORT_FR_BER_CONFIG;

        stringstream sstr;
        sstr << "SMP_BERConfigGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct SMP_BERConfig *p_ber_config = (struct SMP_BERConfig *)p_attribute_data;
        u_int32_t profile = (u_int32_t)(intptr_t)clbck_data.m_data2;
        u_int32_t ber = (u_int32_t)(intptr_t)clbck_data.m_data3;

        int rc = m_pFabricExtendedInfo->addBERConfig(p_node, *p_ber_config, profile, ber);
        if (rc) {
            SetLastError("Failed to add SMP_BERConfig for node=%s, err=%s",
                         p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

// Entry Plane Filter
void IBDiagClbck::SMPEntryPlaneFilterGetClbck(const clbck_data_t &clbck_data,
                                              int rec_status,
                                              void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort* p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_ENTRY_PLANE_FILTER_FAILED(p_port->p_node->appData1.val))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_ENTRY_PLANE_FILTER;

        stringstream sstr;
        sstr << "SMPEntryPlaneFilterGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct SMP_EntryPlaneFilterConfig *p_entry_plane_filter =
                (struct SMP_EntryPlaneFilterConfig *)p_attribute_data;

        IBNode* p_node = p_port->p_node;
        u_int8_t plane = (u_int8_t)(intptr_t)clbck_data.m_data2;
        u_int8_t block = (u_int8_t)(intptr_t)clbck_data.m_data3;
        u_int32_t offset = block * ENTRY_PLANE_FILTER_EGRESS_BLOCK_SIZE;

        // save the filter on the switch
        for (u_int32_t i = 0; i < 32; ++i) {
            p_node->EPF[p_port->num][plane][i + offset] =
                    (p_entry_plane_filter->egress_port_bitmask.Mask_31_0 >> i) & 1;

            p_node->EPF[p_port->num][plane][i + 32 + offset] =
                    (p_entry_plane_filter->egress_port_bitmask.Mask_63_32 >> i) & 1;

            p_node->EPF[p_port->num][plane][i + 64 + offset] =
                    (p_entry_plane_filter->egress_port_bitmask.Mask_95_64 >> i) & 1;

            p_node->EPF[p_port->num][plane][i + 96 + offset] =
                    (p_entry_plane_filter->egress_port_bitmask.Mask_127_96 >> i) & 1;

            p_node->EPF[p_port->num][plane][i + 128 + offset] =
                    (p_entry_plane_filter->egress_port_bitmask.Mask_159_128 >> i) & 1;

            p_node->EPF[p_port->num][plane][i + 160 + offset] =
                    (p_entry_plane_filter->egress_port_bitmask.Mask_191_160 >> i) & 1;

            p_node->EPF[p_port->num][plane][i + 192 + offset] =
                    (p_entry_plane_filter->egress_port_bitmask.Mask_223_192 >> i) & 1;

            p_node->EPF[p_port->num][plane][i + 224 + offset] =
                    (p_entry_plane_filter->egress_port_bitmask.Mask_255_224 >> i) & 1;
        }
    }

    IBDIAG_RETURN_VOID;
}

// Rail Filter
void IBDiagClbck::SMPRailFilterGetClbck(const clbck_data_t &clbck_data,
                                        int rec_status,
                                        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort* p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here

        if (IS_SUPPORT_RAIL_FILTER_FAILED(p_port->p_node->appData1.val))
            IBDIAG_RETURN_VOID; // already reported error

        // mark that we already check for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_RAIL_FILTER;

        stringstream sstr;
        sstr << "SMPRailFilterGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct SMP_RailFilterConfig *p_rail_filter =
                (struct SMP_RailFilterConfig *)p_attribute_data;

        u_int8_t block = (u_int8_t)(intptr_t)clbck_data.m_data2;
        u_int32_t offset = block * RAIL_FILTER_EGRESS_BLOCK_SIZE;

        p_port->railFilterData.VLMask = p_rail_filter->VLMask;
        p_port->railFilterData.MCEnable = p_rail_filter->MCEnable;
        p_port->railFilterData.UCEnable = p_rail_filter->UCEnable;

        // save the filter on the switch port
        for (u_int32_t i = 0; i < 32; ++i) {
            p_port->railFilterData.outPorts[i + offset] =
                    (p_rail_filter->EgressPortMask.Mask_31_0 >> i) & 1;

            p_port->railFilterData.outPorts[i + 32 + offset] =
                    (p_rail_filter->EgressPortMask.Mask_63_32 >> i) & 1;

            p_port->railFilterData.outPorts[i + 64 + offset] =
                    (p_rail_filter->EgressPortMask.Mask_95_64 >> i) & 1;

            p_port->railFilterData.outPorts[i + 96 + offset] =
                    (p_rail_filter->EgressPortMask.Mask_127_96 >> i) & 1;

            p_port->railFilterData.outPorts[i + 128 + offset] =
                    (p_rail_filter->EgressPortMask.Mask_159_128 >> i) & 1;

            p_port->railFilterData.outPorts[i + 160 + offset] =
                    (p_rail_filter->EgressPortMask.Mask_191_160 >> i) & 1;

            p_port->railFilterData.outPorts[i + 192 + offset] =
                    (p_rail_filter->EgressPortMask.Mask_223_192 >> i) & 1;

            p_port->railFilterData.outPorts[i + 224 + offset] =
                    (p_rail_filter->EgressPortMask.Mask_255_224 >> i) & 1;
        }

        p_port->railFilterData.is_initialized = true;
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPEndPortPlaneFilterGetClbck(const clbck_data_t &clbck_data,
                                                int rec_status,
                                                void *p_attribute_data)
{
    IBDIAG_ENTER;
    IBNode* p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        stringstream sstr;
        sstr << "SMPEndPortPlaneFilterGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct SMP_EndPortPlaneFilterConfig *p_e_port_plane_filter =
                (struct SMP_EndPortPlaneFilterConfig *)p_attribute_data;

        p_node->EndPortPlaneFilter[1] =
            p_e_port_plane_filter->end_port_plane_filter_entry_0.lid & 0xffff;
        p_node->EndPortPlaneFilter[2] =
            p_e_port_plane_filter->end_port_plane_filter_entry_1.lid & 0xffff;
        p_node->EndPortPlaneFilter[3] =
            p_e_port_plane_filter->end_port_plane_filter_entry_2.lid & 0xffff;
        p_node->EndPortPlaneFilter[4] =
            p_e_port_plane_filter->end_port_plane_filter_entry_3.lid & 0xffff;
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMP_ExtendedSwitchInfoClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        stringstream sstr;
        sstr << "SMPExtendedSwitchInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct SMP_ExtendedSwitchInfo *p_smp_extended_switch_info =
            (struct SMP_ExtendedSwitchInfo *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addSMP_ExtendedSwitchInfo(p_node, *p_smp_extended_switch_info);
        if (rc) {
            SetLastError("Failed to add SMP_ExtendedSwitchInfo for node=%s, err=%s",
                         p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::SMPChassisInfoClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_node, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        stringstream sstr;
        sstr << "SMPChassisInfoGet." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrNodeNotRespond(p_node, sstr.str()));

    } else {

        struct SMP_ChassisInfo *p_smp_chassis_info = (struct SMP_ChassisInfo*)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addSMPChassisInfo(p_node, *p_smp_chassis_info);
        if (rc) {
            SetLastError("Failed to add SMP_ChassisInfo for node=%s, err=%s",
                         p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiagClbck::PMPortSamplesResultClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data1);

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        IBDIAG_RETURN_VOID;

    if (!VerifyObject(p_port, __LINE__))
        IBDIAG_RETURN_VOID;

    if (rec_status & 0x00ff) { //bits [8-15] are class specific - don't need them here
        stringstream sstr;
        sstr << "PM_PortSamplesResult." << IBIS_STATUS(rec_status);
        m_pErrors->push_back(new FabricErrPortNotRespond(p_port, sstr.str()));

    } else {

        struct PM_PortSamplesResult *p_pm_port_samples_result =
            (struct PM_PortSamplesResult *)p_attribute_data;

        int rc = m_pFabricExtendedInfo->addPMPortSamplesResult(p_port, *p_pm_port_samples_result);
        if (rc) {
            SetLastError("Failed to add PM_PortSamplesResult for port=%s, err=%s",
                         p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }

    IBDIAG_RETURN_VOID;
}

