/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <infiniband/ibutils/release_containers.h>
#include "infiniband/ibutils/output_stream_primitives.h"
#include "ibdiagnet/ibdiagapp_types.h"
#include <math.h>

extern IBDiagClbck ibDiagClbck;

template <class T> T PM_Delta(T val1, T val2) {
    return static_cast<T>(val1 - val2);
}

#define DELTA(X) { \
        delta.X = PM_Delta(curr.X, prev.X); \
        if (prev.X > curr.X) \
            negative_ss << ' ' << #X; \
}

#define IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset,            \
                                        node_type)                        \
((((check_counters_bitset & CHECK_EXT_SPEEDS_COUNTERS_ON_SW) && node_type == IB_SW_NODE) ||                  \
  ((check_counters_bitset & CHECK_EXT_SPEEDS_COUNTERS_ON_ALL))) ? true : false)

#define IS_TO_PRINT_LLR_COUNTERS(check_counters_bitset) \
((check_counters_bitset & PRINT_LLR_COUNTERS) ? true : false)


#define IS_SUPPORT_XMIT_WAIT(cap_mask) ((cap_mask) & 0x1000)

#define IS_SUPPORT_EXTENDED_COUNTERS(cap_mask) ((cap_mask) & (1 << IsExtendedWidthSupported))
#define IS_SUPPORT_EXTENDED_COUNTERS_NO_IETF(cap_mask) ((cap_mask) & (1 << IsExtendedWidthSupportedNoIETF))
#define IS_SUPPORT_ADDITIONAL_EXTENDED_COUNTERS(cap_mask) ((cap_mask) & (1 << IsAdditionalPortCountersExtendedSupported))

#define IS_SUPPORT_EXTENDED_SPEEDS_COUNTERS(cap_mask) ((cap_mask) & IBIS_IB_PORT_CAP_HAS_EXT_SPEEDS)
//check that Performance Management ClassPortInfo:CapabilityMask has IsRSFECCountersSupported
#define IS_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS(cap_mask) ((cap_mask) & 0x4000)

#define MAX_LANE_NUMBER 12

/* Defines the source of counter in the Counter DB (below) */
typedef enum counter_mad_src
{
    PM_PORT_CNT_MAD = 0,
    PM_PORT_CNT_EXT_MAD,
    PM_PORT_EXT_SPEEDS_CNTS_MAD,
    PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD,
    //most of the llr shouldn't be checked
    VS_PORT_LLR_CNTS_MAD_HIDDEN,
    PM_PORT_CALCULATED_COUNTER,
    //for llr counters that are checked (max_retransmission_rate)
    VS_PORT_LLR_CNTS_MAD
} counter_mad_src_t;

/****************************************************/
/* global counters array */
typedef struct pm_counter {
    string name;
    u_int64_t overflow_value;
    u_int8_t real_size;             /* in bytes */
    u_int8_t struct_offset;         /* in bytes */
    u_int16_t cap_mask_bits;        /* which cap mask supports this counter */
    u_int8_t diff_check_threshold;  /* threshold between 2 samples */
    int16_t threshold;            /* counter threshold (currently only for retransmission rate) */
    counter_mad_src_t counter_src;  /* source of counter (in which struct to find it) */
    u_int32_t attribute;
} pm_counter_t;

enum PMCounterAttribute {
    PM_COUNTER_DEFAULT_ATTRIBUTE = 0x0,
    PM_COUNTER_IS_ERROR = 0x1,
    PM_COUNTER_IS_ERROR_WHEN_NO_PLR = 0x2,
    PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL = 0x4, /* is the counter subject to keyword 'all' in -PC */
    PM_COUNTER_IS_SUBJECT_TO_CAP_MASK_2 = 0x8   // counter subject to ClassPortInfo:CapMask2
};

#define PM_COUNTERS_ARR_SIZE 92
static pm_counter_t pm_counters_arr[PM_COUNTERS_ARR_SIZE] = {
        /* 16 bits */
        {"symbol_error_counter",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, SymbolErrorCounter),
                0x0, 1, -1, PM_PORT_CNT_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_rcv_remote_physical_errors",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvRemotePhysicalErrors),
                0x0, 1, -1, PM_PORT_CNT_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_rcv_errors",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvErrors),
                0x0, 1, -1, PM_PORT_CNT_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_xmit_discard",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortXmitDiscards),
                0x0, 10, -1, PM_PORT_CNT_MAD , PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_rcv_switch_relay_errors",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvSwitchRelayErrors),
                0x0, 1, -1, PM_PORT_CNT_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"vl15_dropped",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, VL15Dropped),
                0x0, 10, -1, PM_PORT_CNT_MAD , PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},

        /* 8 bits */
        {"link_error_recovery_counter",
                OVERFLOW_VAL_8_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, LinkErrorRecoveryCounter),
                0x0, 1, -1, PM_PORT_CNT_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"link_down_counter",
                OVERFLOW_VAL_8_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, LinkDownedCounter),
                0x0, 1, -1, PM_PORT_CNT_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_xmit_constraint_errors",
                OVERFLOW_VAL_8_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortXmitConstraintErrors),
                0x0, 10, -1, PM_PORT_CNT_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_rcv_constraint_errors",
                OVERFLOW_VAL_8_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvConstraintErrors),
                0x0, 10, -1, PM_PORT_CNT_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},

        /* 4 bits - in struct the size is 1 byte */
        {"local_link_integrity_errors",
                OVERFLOW_VAL_4_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, LocalLinkIntegrityErrors),
                0x0, 10, -1, PM_PORT_CNT_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"excessive_buffer_errors",
                OVERFLOW_VAL_4_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, ExcessiveBufferOverrunErrors),
                0x0, 10, -1, PM_PORT_CNT_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},

        /* 32 bits */
        {"port_xmit_data",
                OVERFLOW_VAL_NONE, 4, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortXmitData),
                0x0, 0, -1, PM_PORT_CNT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_rcv_data",
                OVERFLOW_VAL_NONE, 4, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvData),
                0x0, 0, -1, PM_PORT_CNT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_xmit_pkts",
                OVERFLOW_VAL_NONE, 4, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortXmitPkts),
                0x0, 0, -1, PM_PORT_CNT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_rcv_pkts",
                OVERFLOW_VAL_NONE, 4, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvPkts),
                0x0, 0, -1, PM_PORT_CNT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_xmit_wait",
                OVERFLOW_VAL_NONE, 4, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortXmitWait),
                0x1000, 0, -1, PM_PORT_CNT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},

        /* 64 bits */
        {"port_xmit_data_extended",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortXmitData),
                0x0400 | 0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_rcv_data_extended",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortRcvData),
                0x0400 | 0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_xmit_pkts_extended",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortXmitPkts),
                0x0400 | 0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_rcv_pkts_extended",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortRcvPkts),
                0x0400 | 0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_unicast_xmit_pkts",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortUniCastXmitPkts),
                0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_unicast_rcv_pkts",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortUniCastRcvPkts),
                0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_multicast_xmit_pkts",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortMultiCastXmitPkts),
                0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_multicast_rcv_pkts",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortMultiCastRcvPkts),
                0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"symbol_error_counter_extended",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, SymbolErrorCounter),
                0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, PM_COUNTER_IS_ERROR |
                PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL | PM_COUNTER_IS_SUBJECT_TO_CAP_MASK_2},

        /* Extended speeds counters */
        {"sync_header_err_cnt",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, SyncHeaderErrorCounter),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_ERROR |
                PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"unknown_block_cnt",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, UnknownBlockCounter),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_ERROR |
                PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane0",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[0]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane1",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[1]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane2",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[2]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane3",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[3]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane4",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[4]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane5",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[5]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane6",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[6]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane7",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[7]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane8",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[8]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane9",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[9]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane10",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[10]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"error_detection_counter_lane11",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[11]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane0",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[0]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane1",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[1]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane2",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[2]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane3",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[3]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane4",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[4]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane5",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[5]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane6",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[6]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane7",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[7]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane8",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[8]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane9",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[9]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane10",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[10]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_correctable_block_counter_lane11",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[11]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane0",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[0]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane1",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[1]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane2",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[2]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane3",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[3]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane4",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[4]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane5",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[5]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane6",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[6]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane7",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[7]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane8",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[8]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane9",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[9]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane10",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[10]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_uncorrectable_block_counter_lane11",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[11]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},

        /* LLR counters */
        /* For VS_PORT_LLR_CNTS_MAD_HIDDEN counters we'll not report overflow, threshold and threshold-diff */
        {"port_rcv_cells",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortRcvCells),
                0, 0, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_rcv_cell_for_retry",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortRcvCellForRetry),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_rcv_retry",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortRcvRetry),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_xmit_cells",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortXmitCells),
                0, 0, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_DEFAULT_ATTRIBUTE},
        {"port_xmit_retry_cells",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortXmitRetryCells),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_xmit_retry",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortXmitRetry),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_symbol_error",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortSymbolError),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_error_detection_counter_lane0",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortErrorDetectionCounterLane0),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_error_detection_counter_lane1",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortErrorDetectionCounterLane1),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_error_detection_counter_lane2",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortErrorDetectionCounterLane2),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_error_detection_counter_lane3",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortErrorDetectionCounterLane3),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"max_retransmission_rate",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, MaxRetransmissionRate),
                0, 1, 500, VS_PORT_LLR_CNTS_MAD, PM_COUNTER_DEFAULT_ATTRIBUTE},
        /* Calculated Counters */
        {"retransmission_per_sec",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct PM_PortCalcCounters, RetransmissionPerSec),
                0, 1, 500, PM_PORT_CALCULATED_COUNTER, PM_COUNTER_DEFAULT_ATTRIBUTE},

        /* Extended Speeds counters when RS-FEC is active */
        {"fec_corrected_symbol_counter_lane0",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[0]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane1",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[1]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane2",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[2]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane3",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[3]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane4",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[4]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane5",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[5]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane6",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[6]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane7",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[7]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane8",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[8]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane9",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[9]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane10",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[10]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"fec_corrected_symbol_counter_lane11",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[11]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_fec_correctable_block_counter",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, PortFECCorrectableBlockCounter),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_fec_uncorrectable_block_counter",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, PortFECUncorrectableBlockCounter),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_ERROR_WHEN_NO_PLR |
                PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
        {"port_fec_corrected_symbol_counter",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, PortFECCorrectedSymbolCounter),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL},
};

/****************************************************/
static long double CalcBERFireCodeFEC(struct PM_PortExtendedSpeedsCounters *p_counters1,
                                      struct PM_PortExtendedSpeedsCounters *p_counters2)
{
    IBDIAG_ENTER;
    long double ber = 0.0;
    long double sum_crr_cntr1= 0.0, sum_uncrr_cntr1 = 0.0;
    long double sum_crr_cntr2= 0.0, sum_uncrr_cntr2 = 0.0;

    for (uint8_t  i = 0; i < MAX_LANE_NUMBER; i++) {
        sum_uncrr_cntr1 += p_counters1->FECUncorrectableBlockCounterLane[i];
        sum_crr_cntr1 += p_counters1->FECCorrectableBlockCountrLane[i];
    }

    for (uint8_t  i = 0; i < MAX_LANE_NUMBER; i++) {
        sum_uncrr_cntr2 += p_counters2->FECUncorrectableBlockCounterLane[i];
        sum_crr_cntr2 += p_counters2->FECCorrectableBlockCountrLane[i];
    }

    ber = (sum_crr_cntr2 - sum_crr_cntr1) +
          K_FCODE *
          (sum_uncrr_cntr2 - sum_uncrr_cntr1);

    IBDIAG_RETURN(ber);
}

static long double EDPLSum(struct PM_PortExtendedSpeedsCounters *p_counters1,
                           struct PM_PortExtendedSpeedsCounters *p_counters2)
{
    IBDIAG_ENTER;
    //sum of edpls over lanes/time
    long double ber = 0.0;
    for (int i = 0; i < MAX_LANE_NUMBER; i++)
        ber += p_counters2->ErrorDetectionCounterLane[i];

    for (int i = 0; i < MAX_LANE_NUMBER; i++)
        ber -= p_counters1->ErrorDetectionCounterLane[i];

    IBDIAG_RETURN(ber);
}

static bool is_counter_supported(u_int16_t node_cap_mask, u_int16_t counter_cap_mask)
{
    IBDIAG_ENTER;
    if (counter_cap_mask == 0)              //supported
        IBDIAG_RETURN(true);
    if (node_cap_mask & counter_cap_mask)
        IBDIAG_RETURN(true);
    IBDIAG_RETURN(false);
}

//Returns: SUCCESS_CODE / IBDIAG_ERR_CODE_CHECK_FAILED
int counter_name_2_index(string name, int &index) {
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    static map < string, int > name2index;
    map < string, int >::iterator it;

    // first time the function is called
    if ( !name2index.size() ) {
        for ( int i = 0; i < PM_COUNTERS_ARR_SIZE; i++ ) {
            name2index[pm_counters_arr[i].name] = i;
        }
    }

    // check if the given name is a valid counter name
    it = name2index.find(name);

    if ( it == name2index.end() ) {
        rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        index = -1;
    } else {
        index = it->second;
    }
    IBDIAG_RETURN(rc);
}

int get_value(void* buff, u_int8_t size, u_int64_t &value)
{
    IBDIAG_ENTER;
    int rc = 0;
    switch (size) {
        case 1:
            value = *(u_int8_t *)(buff);
            break;
        case 2:
            value = *(u_int16_t *)(buff);
            break;
        case 4:
            value = *(u_int32_t *)(buff);
            break;
        case 8:
            value = *(u_int64_t *)(buff);
            break;
        default:
            rc = 1;
            value = 0;
    }

    IBDIAG_RETURN(rc);
}

static bool is_report_as_error(pm_counter_t &pm_counter, IBPort *p_port)
{
    IBDIAG_ENTER;

    if (p_port->get_fec_mode() == IB_FEC_NO_FEC || p_port->get_fec_mode() == IB_FEC_NA)
        IBDIAG_RETURN(true);

    if (pm_counter.attribute & PM_COUNTER_IS_ERROR)
        IBDIAG_RETURN(true);

    if ((pm_counter.attribute & PM_COUNTER_IS_ERROR_WHEN_NO_PLR) &&
        !(p_port->IsPLR()))
        IBDIAG_RETURN(true);

    IBDIAG_RETURN(false);
}

/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::DumpPortExtendedSpeedsCounters(ostream &sout,
                                            bool en_per_lane_cnts,
                                            IBPort *p_curr_port,
                                            u_int32_t port_idx)
{
    IBDIAG_ENTER;
    struct PM_PortExtendedSpeedsCounters *p_ext_speeds_cnts = NULL;
    struct PM_PortExtendedSpeedsRSFECCounters *p_ext_speeds_rsfec_cnts = NULL;
    char buffer[2096] = {'\0'};
    u_int32_t error_detection_cnt_total = 0;
    u_int16_t sync_header_error_counter = 0;
    u_int16_t unknown_block_counter = 0;
    u_int64_t correctable_block_cnt_total = 0;
    u_int64_t uncorrectable_block_cnt_total = 0;
    u_int64_t corrected_symbol_cnt_total = 0;

    IBLinkWidth p_curr_port_width = p_curr_port->get_common_width();
    int lane_number = LinkWidthToLane(p_curr_port_width);

    p_ext_speeds_cnts = this->fabric_extended_info.
        getPMPortExtSpeedsCounters(port_idx);
    p_ext_speeds_rsfec_cnts =
        this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(port_idx);

    // Counters not available
    if (p_ext_speeds_cnts) {
        for (int i = 0; i < lane_number; i++) {
            error_detection_cnt_total +=
                p_ext_speeds_cnts->ErrorDetectionCounterLane[i];
            correctable_block_cnt_total +=
                p_ext_speeds_cnts->FECCorrectableBlockCountrLane[i];
            uncorrectable_block_cnt_total +=
                p_ext_speeds_cnts->FECUncorrectableBlockCounterLane[i];
        }
        sync_header_error_counter =
            p_ext_speeds_cnts->SyncHeaderErrorCounter;
        unknown_block_counter = p_ext_speeds_cnts->UnknownBlockCounter;
    } else if (p_ext_speeds_rsfec_cnts) {
        for (int i = 0; i < lane_number; i++)
            corrected_symbol_cnt_total +=
                p_ext_speeds_rsfec_cnts->FECCorrectedSymbolCounterLane[i];
        sync_header_error_counter =
            p_ext_speeds_rsfec_cnts->SyncHeaderErrorCounter;
        unknown_block_counter =
            p_ext_speeds_rsfec_cnts->UnknownBlockCounter;
    }

    //common counters
    if (p_ext_speeds_cnts || p_ext_speeds_rsfec_cnts) {
        snprintf(buffer, sizeof(buffer),
                 "sync_header_error_counter=" U32H_FMT "\n" \
                 "unknown_block_counter=" U32H_FMT "\n",
                 sync_header_error_counter,
                 unknown_block_counter);
        sout << buffer;
    } else {
        snprintf(buffer, sizeof(buffer),
                 "sync_header_error_counter=NA\n" \
                 "unknown_block_counter=NA\n");
        sout << buffer;
    }


    //RS-FEC mode
    if (isRSFEC(p_curr_port->get_fec_mode())) {
        if (p_ext_speeds_rsfec_cnts) {
            snprintf(buffer, sizeof(buffer),
                     "fec_corrected_symbol_counter_total=" U64H_FMT "\n",
                     corrected_symbol_cnt_total);
            sout << buffer;
            if (en_per_lane_cnts) {
                for (int i = 0; i < lane_number; i++) {
                    snprintf(buffer, sizeof(buffer), "%s[%d]=" U32H_FMT "\n",
                             "fec_corrected_symbol_counter_lane", i,
                             p_ext_speeds_rsfec_cnts->
                             FECCorrectedSymbolCounterLane[i]);
                    sout << buffer;
                }
            }
            snprintf(buffer, sizeof(buffer),
                     "port_fec_correctable_block_counter=" U32H_FMT "\n"\
                     "port_fec_uncorrectable_block_counter=" U32H_FMT "\n"\
                     "port_fec_corrected_symbol_counter=" U32H_FMT "\n",
                     p_ext_speeds_rsfec_cnts->PortFECCorrectableBlockCounter,
                     p_ext_speeds_rsfec_cnts->PortFECUncorrectableBlockCounter,
                     p_ext_speeds_rsfec_cnts->PortFECCorrectedSymbolCounter);
            sout << buffer;
        } else {
            snprintf(buffer, sizeof(buffer),
                     "fec_corrected_symbol_counter_total=NA\n");
            sout << buffer;
            if (en_per_lane_cnts) {
                for (int i = 0; i < lane_number; i++) {
                    snprintf(buffer, sizeof(buffer), "%s[%d]=NA\n",
                             "fec_corrected_symbol_counter_lane", i);
                    sout << buffer;
                }
            }
            snprintf(buffer, sizeof(buffer),
                     "port_fec_correctable_block_counter=NA\n"\
                     "port_fec_uncorrectable_block_counter=NA\n"\
                     "port_fec_corrected_symbol_counter=NA\n");
            sout << buffer;
        }
    } else { //not RS-FEC
        if (p_ext_speeds_cnts) {
            snprintf(buffer, sizeof(buffer),
                     "error_detection_counter_total=" U32H_FMT "\n",
                     error_detection_cnt_total);
            sout << buffer;
            //when fec is "No FEC", these counters are irrelevant
            if (p_curr_port->get_fec_mode() != IB_FEC_NO_FEC) {
                snprintf(buffer, sizeof(buffer),
                         "fec_correctable_block_counter_total=" U64H_FMT "\n" \
                         "fec_uncorrectable_block_counter_total=" U64H_FMT "\n",
                         correctable_block_cnt_total,
                         uncorrectable_block_cnt_total);
                sout << buffer;
            }
            if (en_per_lane_cnts) {
                for (int i = 0; i < lane_number; i++) {
                    snprintf(buffer, sizeof(buffer),
                        "%s[%d]=" U32H_FMT "\n",
                        "error_detection_counter_lane", i,
                        p_ext_speeds_cnts->ErrorDetectionCounterLane[i]);
                    sout << buffer;
                }
                //when fec is "No FEC", these counters are irrelevant
                if (p_curr_port->get_fec_mode() != IB_FEC_NO_FEC) {
                    for (int i = 0; i < lane_number; i++) {
                        snprintf(buffer, sizeof(buffer),
                            "%s[%d]=" U32H_FMT "\n",
                            "fec_correctable_block_counter_lane", i,
                            p_ext_speeds_cnts->FECCorrectableBlockCountrLane[i]);
                        sout << buffer;
                    }
                    for (int i = 0; i < lane_number; i++) {
                        snprintf(buffer, sizeof(buffer),
                            "%s[%d]=" U32H_FMT "\n",
                            "fec_uncorrectable_block_counter_lane", i,
                            p_ext_speeds_cnts->
                            FECUncorrectableBlockCounterLane[i]);
                        sout << buffer;
                    }
                }
            }
        } else {
            sout << "error_detection_counter_total=NA\n";

            //when fec is "No FEC", these counters are irrelevant
            if (p_curr_port->get_fec_mode() != IB_FEC_NO_FEC)
                sout << "fec_correctable_block_counter_total=NA\n" \
                        "fec_uncorrectable_block_counter_total=NA\n";
            if (en_per_lane_cnts) {
                for (int i = 0; i < lane_number; i++) {
                    snprintf(buffer, sizeof(buffer), "%s[%d]=NA\n",
                             "error_detection_counter_lane", i);
                    sout << buffer;
                }
                //when fec is "No FEC", these counters are irrelevant
                if (p_curr_port->get_fec_mode() != IB_FEC_NO_FEC) {
                    for (int i = 0; i < lane_number; i++) {
                        snprintf(buffer, sizeof(buffer), "%s[%d]=NA\n",
                                 "fec_correctable_block_counter_lane", i);
                        sout << buffer;
                    }
                    for (int i = 0; i < lane_number; i++) {
                        snprintf(buffer, sizeof(buffer), "%s[%d]=NA\n",
                                 "fec_uncorrectable_block_counter_lane", i);
                        sout << buffer;
                    }
                }
            }
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpAPortCounters(APort* p_aport,
                               ofstream &sout,
                               u_int32_t check_counters_bitset,
                               bool en_per_lane_cnts)
{
    if (!p_aport)
        return;

    for (size_t i = 1; i < p_aport->ports.size(); ++i) {
        if (!p_aport->ports[i])
            return;
    }
    // stringstream for every plane, plane 1 is at index 0.
    vector<std::stringstream> plane_counters_streams(p_aport->ports.size() - 1);
    bool keep_merging = true;

    for (size_t i = 1; i < p_aport->ports.size(); ++i) {
        // get the port counters to this plane,
        DumpPortCounters(p_aport->ports[i],
                         plane_counters_streams[i-1],
                         check_counters_bitset,
                         en_per_lane_cnts,
                         false);
        // if a plane didn't write anything, exit
        if (plane_counters_streams[i-1].tellp() == std::streampos(0))
            return;
    }

    sout << "-------------------------------------------------------" << endl;
    sout << "APort Name=" << p_aport->getAggregatedLabel() << " SystemGUID="\
         << p_aport->ports[1]->p_node->guid_get() << endl;
    sout << "-------------------------------------------------------" << endl;
        
    // merge all port counters from all planes, add one line from each stream
    // until we hit the end of the stream
    while (keep_merging) {
        for (size_t i = 0; i < plane_counters_streams.size(); ++i) {
            string line;
            if (std::getline(plane_counters_streams[i], line)) {
                size_t eq_sign = line.find_first_of('=');
                if (eq_sign == string::npos)
                    continue;
                // insert plane number before '='
                sout << line.substr(0, eq_sign) << DEC(i + 1)
                     << line.substr(eq_sign) << endl;
            } else
                // TODO: check if we have other lines in other planes
                keep_merging = false;
        }
    }
}

void IBDiag::DumpPortCounters(IBPort* p_port,
                              ostream &sout,
                              u_int32_t check_counters_bitset,
                              bool en_per_lane_cnts,
                              bool print_header)
{
    IBDIAG_ENTER;
    struct PM_PortCounters *p_curr_port_counters = NULL;
    struct PM_PortCountersExtended *p_curr_port_counters_extended = NULL;
    struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics = NULL;
    struct PM_PortCalcCounters *p_port_calc_counters = NULL;
    struct PM_PortRcvErrorDetails *p_port_rcv_error_details = NULL;
    struct PM_PortXmitDiscardDetails *p_port_xmit_discard_details = NULL;

    char buffer[2096];

    if (!p_port || !p_port->getInSubFabric())
        return;

    u_int32_t port_index = p_port->createIndex;

    p_curr_port_counters = this->fabric_extended_info.getPMPortCounters(port_index);
    if (!p_curr_port_counters)      //info not exist
        return;

    if (print_header) {
        snprintf(buffer, sizeof(buffer),
            "Port=" U32D_FMT    \
            " Lid=" U16H_FMT    \
            " GUID=" U64H_FMT   \
            " Device=" U32D_FMT \
            " Port Name=" STR_FMT,
            p_port->num,
            p_port->base_lid,
            p_port->guid_get(),
            p_port->p_node->devId,
            p_port->getName().c_str());
        sout << "-------------------------------------------------------" << endl;
        sout << buffer << endl;
        sout << "-------------------------------------------------------" << endl;
    }

    snprintf(buffer, sizeof(buffer),
                "link_down_counter=" U32H_FMT "\n"  \
                "link_error_recovery_counter=" U32H_FMT "\n"  \
                "symbol_error_counter=" U32H_FMT "\n"  \
                "port_rcv_remote_physical_errors=" U32H_FMT "\n"  \
                "port_rcv_errors=" U32H_FMT "\n"  \
                "port_xmit_discard=" U32H_FMT "\n"  \
                "port_rcv_switch_relay_errors=" U32H_FMT "\n"  \
                "excessive_buffer_errors=" U32H_FMT "\n"  \
                "local_link_integrity_errors=" U32H_FMT "\n"  \
                "port_rcv_constraint_errors=" U32H_FMT "\n"  \
                "port_xmit_constraint_errors=" U32H_FMT "\n"  \
                "vl15_dropped=" U32H_FMT "\n"  \
                "port_xmit_data=" U32H_FMT "\n"  \
                "port_rcv_data=" U32H_FMT "\n"  \
                "port_xmit_pkts=" U32H_FMT "\n"  \
                "port_rcv_pkts=" U32H_FMT "\n"  \
                "port_xmit_wait=" U32H_FMT "\n",
                p_curr_port_counters->LinkDownedCounter,
                p_curr_port_counters->LinkErrorRecoveryCounter,
                p_curr_port_counters->SymbolErrorCounter,
                p_curr_port_counters->PortRcvRemotePhysicalErrors,
                p_curr_port_counters->PortRcvErrors,
                p_curr_port_counters->PortXmitDiscards,
                p_curr_port_counters->PortRcvSwitchRelayErrors,
                p_curr_port_counters->ExcessiveBufferOverrunErrors,
                p_curr_port_counters->LocalLinkIntegrityErrors,
                p_curr_port_counters->PortRcvConstraintErrors,
                p_curr_port_counters->PortXmitConstraintErrors,
                p_curr_port_counters->VL15Dropped,
                p_curr_port_counters->PortXmitData,
                p_curr_port_counters->PortRcvData,
                p_curr_port_counters->PortXmitPkts,
                p_curr_port_counters->PortRcvPkts,
                p_curr_port_counters->PortXmitWait);
    sout << buffer;

    p_curr_port_counters_extended = this->fabric_extended_info.getPMPortCountersExtended(port_index);
    if (!p_curr_port_counters_extended) {       //info not exist
        snprintf(buffer, sizeof(buffer),
                    "port_xmit_data_extended=NA\n" \
                    "port_rcv_data_extended=NA\n"  \
                    "port_xmit_pkts_extended=NA\n" \
                    "port_rcv_pkts_extended=NA\n"  \
                    "port_unicast_xmit_pkts=NA\n"  \
                    "port_unicast_rcv_pkts=NA\n"   \
                    "port_multicast_xmit_pkts=NA\n"    \
                    "port_multicast_rcv_pkts=NA\n"
                    "symbol_error_counter_extended=NA\n"
                    "link_error_recovery_counter_extended=NA\n"
                    "link_downed_counter_extended=NA\n"
                    "port_rcv_errors_extended=NA\n"
                    "port_rcv_remote_physical_errors_extended=NA\n"
                    "port_rcv_switch_relay_errors_extended=NA\n"
                    "port_xmit_discards_extended=NA\n"
                    "port_xmit_constraint_errors_extended=NA\n"
                    "port_rcv_constraint_errors_extended=NA\n"
                    "local_link_integrity_errors_extended=NA\n"
                    "excessive_buffer_overrun_errors_extended=NA\n"
                    "vl15_dropped_extended=NA\n"
                    "port_xmit_wait_extended=NA\n"
                    "qp1_dropped_extended=NA\n");

    } else {    //info  exist
        snprintf(buffer, sizeof(buffer),
                    "port_xmit_data_extended=" U64H_FMT "\n"  \
                    "port_rcv_data_extended=" U64H_FMT "\n"  \
                    "port_xmit_pkts_extended=" U64H_FMT "\n"  \
                    "port_rcv_pkts_extended=" U64H_FMT "\n"  \
                    "port_unicast_xmit_pkts=" U64H_FMT "\n"  \
                    "port_unicast_rcv_pkts=" U64H_FMT "\n"  \
                    "port_multicast_xmit_pkts=" U64H_FMT "\n"  \
                    "port_multicast_rcv_pkts=" U64H_FMT "\n",
                    p_curr_port_counters_extended->PortXmitData,
                    p_curr_port_counters_extended->PortRcvData,
                    p_curr_port_counters_extended->PortXmitPkts,
                    p_curr_port_counters_extended->PortRcvPkts,
                    p_curr_port_counters_extended->PortUniCastXmitPkts,
                    p_curr_port_counters_extended->PortUniCastRcvPkts,
                    p_curr_port_counters_extended->PortMultiCastXmitPkts,
                    p_curr_port_counters_extended->PortMultiCastRcvPkts);

        sout << buffer;

        IB_ClassPortInfo *pm_class_port_info = this->fabric_extended_info.
                getPMClassPortInfo(p_port->p_node->createIndex);

        if (IS_SUPPORT_ADDITIONAL_EXTENDED_COUNTERS(pm_class_port_info->CapMsk2)) {
            snprintf(buffer, sizeof(buffer),
                        "symbol_error_counter_extended=" U64H_FMT "\n"
                        "link_error_recovery_counter_extended=" U64H_FMT "\n"
                        "link_downed_counter_extended=" U64H_FMT "\n"
                        "port_rcv_errors_extended=" U64H_FMT "\n"
                        "port_rcv_remote_physical_errors_extended=" U64H_FMT "\n"
                        "port_rcv_switch_relay_errors_extended=" U64H_FMT "\n"
                        "port_xmit_discards_extended=" U64H_FMT "\n"
                        "port_xmit_constraint_errors_extended=" U64H_FMT "\n"
                        "port_rcv_constraint_errors_extended=" U64H_FMT "\n"
                        "local_link_integrity_errors_extended=" U64H_FMT "\n"
                        "excessive_buffer_overrun_errors_extended=" U64H_FMT "\n"
                        "vl15_dropped_extended=" U64H_FMT "\n"
                        "port_xmit_wait_extended=" U64H_FMT "\n"
                        "qp1_dropped_extended=" U64H_FMT "\n",
                        p_curr_port_counters_extended->SymbolErrorCounter,
                        p_curr_port_counters_extended->LinkErrorRecoveryCounter,
                        p_curr_port_counters_extended->LinkDownedCounter,
                        p_curr_port_counters_extended->PortRcvErrors,
                        p_curr_port_counters_extended->PortRcvRemotePhysicalErrors,
                        p_curr_port_counters_extended->PortRcvSwitchRelayErrors,
                        p_curr_port_counters_extended->PortXmitDiscards,
                        p_curr_port_counters_extended->PortXmitConstraintErrors,
                        p_curr_port_counters_extended->PortRcvConstraintErrors,
                        p_curr_port_counters_extended->LocalLinkIntegrityErrors,
                        p_curr_port_counters_extended->ExcessiveBufferOverrunErrors,
                        p_curr_port_counters_extended->VL15Dropped,
                        p_curr_port_counters_extended->PortXmitWait,
                        p_curr_port_counters_extended->QP1Dropped);

        } else {
            snprintf(buffer, sizeof(buffer),
                        "symbol_error_counter_extended=NA\n"
                        "link_error_recovery_counter_extended=NA\n"
                        "link_downed_counter_extended=NA\n"
                        "port_rcv_errors_extended=NA\n"
                        "port_rcv_remote_physical_errors_extended=NA\n"
                        "port_rcv_switch_relay_errors_extended=NA\n"
                        "port_xmit_discards_extended=NA\n"
                        "port_xmit_constraint_errors_extended=NA\n"
                        "port_rcv_constraint_errors_extended=NA\n"
                        "local_link_integrity_errors_extended=NA\n"
                        "excessive_buffer_overrun_errors_extended=NA\n"
                        "vl15_dropped_extended=NA\n"
                        "port_xmit_wait_extended=NA\n"
                        "qp1_dropped_extended=NA\n");
        }
    }
    sout << buffer;

    // extended speeds counter
    if (IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset,
                                        p_port->p_node->type)) {
        this->DumpPortExtendedSpeedsCounters(sout, en_per_lane_cnts,
                                                p_port, port_index);
    }


    p_port_calc_counters = this->fabric_extended_info.getPMPortCalcCounters(port_index);
    if ( !p_port_calc_counters ) {
        snprintf(buffer, sizeof(buffer), "retransmission_per_sec=NA\n");
        sout << buffer;
    } else {
        snprintf(buffer, sizeof(buffer), "retransmission_per_sec=" U64H_FMT "\n",
                    p_port_calc_counters->RetransmissionPerSec);
        sout << buffer;
    }

    p_curr_port_llr_statistics = this->fabric_extended_info.getVSPortLLRStatistics(port_index);

    if (!this->capability_module.IsSupportedGMPCapability(
        p_port->p_node, EnGMPCapIsMaxRetransmissionRateSupported) ||
        !p_curr_port_llr_statistics)
        sout << "max_retransmission_rate=NA" << endl;
    else {
        snprintf(buffer, sizeof(buffer), "max_retransmission_rate=" U16H_FMT "\n",
                    p_curr_port_llr_statistics->MaxRetransmissionRate);
        sout << buffer;
    }

    // LLR counter
    if (IS_TO_PRINT_LLR_COUNTERS(check_counters_bitset)) {
        // Counters not available
        if (!p_curr_port_llr_statistics) {
            snprintf(buffer, sizeof(buffer),
                        "port_rcv_cells=NA\n" \
                        "port_rcv_cell_for_retry=NA\n" \
                        "port_rcv_retry=NA\n" \
                        "port_xmit_cells=NA\n" \
                        "port_xmit_retry_cells=NA\n" \
                        "port_xmit_retry=NA\n"
                        "port_symbol_error=NA\n"
                        "port_error_detection_counter_lane0=NA\n"
                        "port_error_detection_counter_lane1=NA\n"
                        "port_error_detection_counter_lane2=NA\n"
                        "port_error_detection_counter_lane3=NA\n");
            sout << buffer;
        }

        // Counters available
        else {

            snprintf(buffer, sizeof(buffer),
                        "port_rcv_cells=" U64H_FMT "\n" \
                        "port_rcv_cell_for_retry=" U64H_FMT "\n" \
                        "port_rcv_retry=" U64H_FMT "\n" \
                        "port_xmit_cells=" U64H_FMT "\n" \
                        "port_xmit_retry_cells=" U64H_FMT "\n" \
                        "port_xmit_retry=" U64H_FMT "\n" \
                        "port_symbol_error=" U64H_FMT "\n" \
                        "port_error_detection_counter_lane0="  U64H_FMT "\n" \
                        "port_error_detection_counter_lane1="  U64H_FMT "\n" \
                        "port_error_detection_counter_lane2="  U64H_FMT "\n" \
                        "port_error_detection_counter_lane3="  U64H_FMT "\n",
                        p_curr_port_llr_statistics->PortRcvCells,
                        p_curr_port_llr_statistics->PortRcvCellForRetry,
                        p_curr_port_llr_statistics->PortRcvRetry,
                        p_curr_port_llr_statistics->PortXmitCells,
                        p_curr_port_llr_statistics->PortXmitRetryCells,
                        p_curr_port_llr_statistics->PortXmitRetry,
                        p_curr_port_llr_statistics->PortSymbolError,
                        p_curr_port_llr_statistics->PortErrorDetectionCounterLane0,
                        p_curr_port_llr_statistics->PortErrorDetectionCounterLane1,
                        p_curr_port_llr_statistics->PortErrorDetectionCounterLane2,
                        p_curr_port_llr_statistics->PortErrorDetectionCounterLane3);
            sout << buffer;
        }
    }

    // PortRcvErrorDetails and PortXmitDiscardDetails
    p_port_rcv_error_details = this->fabric_extended_info.getPMPortRcvErrorDetails(port_index);
    p_port_xmit_discard_details = this->fabric_extended_info.getPMPortXmitDiscardDetails(port_index);

    struct PM_PortSamplesControl* p_samples_control =
            this->fabric_extended_info.getPMPortSamplesControl(p_port->createIndex);

    struct PortSampleControlOptionMask* p_option_mask = NULL;

    if (p_samples_control)
        p_option_mask = &p_samples_control->PortSampleControlOptionMask;

    if (!p_port_rcv_error_details || !p_option_mask) {
        snprintf(buffer, sizeof(buffer),
                    "port_local_physical_errors=NA\n" \
                    "port_malformed_packet_errors=NA\n" \
                    "port_buffer_overrun_errors=NA\n" \
                    "port_dlid_mapping_errors=NA\n" \
                    "port_vl_mapping_errors=NA\n" \
                    "port_looping_errors=NA\n" );
        sout << buffer;
    } else {
        (p_option_mask->PortRcvErrorDetails_PortLocalPhysicalErrors)?
            snprintf(buffer, sizeof(buffer), "port_local_physical_errors=" U16H_FMT "\n",
                    p_port_rcv_error_details->PortLocalPhysicalErrors):
            snprintf(buffer, sizeof(buffer), "port_local_physical_errors=NA\n");
        sout << buffer;

        (p_option_mask->PortRcvErrorDetails_PortMalformedPacketErrors)?
            snprintf(buffer, sizeof(buffer), "port_malformed_packet_errors=" U16H_FMT "\n",
                    p_port_rcv_error_details->PortMalformedPacketErrors):
            snprintf(buffer, sizeof(buffer), "port_malformed_packet_errors=NA\n");
        sout << buffer;

        (p_option_mask->PortRcvErrorDetails_PortBufferOverrunErrors)?
            snprintf(buffer, sizeof(buffer), "port_buffer_overrun_errors=" U16H_FMT "\n",
                    p_port_rcv_error_details->PortBufferOverrunErrors):
            snprintf(buffer, sizeof(buffer), "port_buffer_overrun_errors=NA\n");
        sout << buffer;

        (p_option_mask->PortRcvErrorDetails_PortDLIDMappingErrors)?
            snprintf(buffer, sizeof(buffer), "port_dlid_mapping_errors=" U16H_FMT "\n",
                    p_port_rcv_error_details->PortDLIDMappingErrors):
            snprintf(buffer, sizeof(buffer), "port_dlid_mapping_errors=NA\n");
        sout << buffer;

        (p_option_mask->PortRcvErrorDetails_PortVlMappingErrors)?
            snprintf(buffer, sizeof(buffer), "port_vl_mapping_errors=" U16H_FMT "\n",
                    p_port_rcv_error_details->PortVLMappingErrors):
            snprintf(buffer, sizeof(buffer), "port_vl_mapping_errors=NA\n");
        sout << buffer;

        (p_option_mask->PortRcvErrorDetails_PortLoopingErrors)?
            snprintf(buffer, sizeof(buffer), "port_looping_errors=" U16H_FMT "\n",
                    p_port_rcv_error_details->PortLoopingErrors):
            snprintf(buffer, sizeof(buffer), "port_looping_errors=NA\n");
        sout << buffer;
    }

    if (!p_port_xmit_discard_details || !p_option_mask) {
        snprintf(buffer, sizeof(buffer),
                    "port_inactive_discards=NA\n" \
                    "port_neighbor_mtu_discards=NA\n" \
                    "port_sw_lifetime_limit_discards=NA\n" \
                    "port_sw_hoq_lifetime_limit_discards=NA\n");
        sout << buffer;
    } else {
        (p_option_mask->PortXmitDiscardDetails_PortInactiveDiscards)?
            snprintf(buffer, sizeof(buffer), "port_inactive_discards=" U16H_FMT "\n",
                    p_port_xmit_discard_details->PortInactiveDiscards):
            snprintf(buffer, sizeof(buffer), "port_inactive_discards=NA\n");
        sout << buffer;

        (p_option_mask->PortXmitDiscardDetails_PortNeighborMTUDiscards)?
            snprintf(buffer, sizeof(buffer), "port_neighbor_mtu_discards=" U16H_FMT "\n",
                    p_port_xmit_discard_details->PortNeighborMTUDiscards):
            snprintf(buffer, sizeof(buffer), "port_neighbor_mtu_discards=NA\n");
        sout << buffer;

        (p_option_mask->PortXmitDiscardDetails_PortSwLifetimeLimitDiscards)?
            snprintf(buffer, sizeof(buffer), "port_sw_lifetime_limit_discards=" U16H_FMT "\n",
                    p_port_xmit_discard_details->PortSwLifetimeLimitDiscards):
            snprintf(buffer, sizeof(buffer), "port_sw_lifetime_limit_discards=NA\n");
        sout << buffer;

        (p_option_mask->PortXmitDiscardDetails_PortSwHOQLifeLimitDiscards)?
            snprintf(buffer, sizeof(buffer), "port_sw_hoq_lifetime_limit_discards=" U16H_FMT "\n",
                    p_port_xmit_discard_details->PortSwHOQLifetimeLimitDiscards):
            snprintf(buffer, sizeof(buffer), "port_sw_hoq_lifetime_limit_discards=NA\n");
        sout << buffer;
    }

    sout << endl;
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpAllPortsCounters(ofstream &sout,
                                  u_int32_t check_counters_bitset,
                                  bool en_per_lane_cnts)
{
    IBDIAG_ENTER;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort* p_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_port)
            continue;

        DumpPortCounters(p_port, sout, check_counters_bitset, en_per_lane_cnts, true);
    }

    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpAllAPortsCounters(ofstream &sout,
                                   u_int32_t check_counters_bitset,
                                   bool en_per_lane_cnts)
{
    IBDIAG_ENTER;

    for (const auto& entry : this->discovered_fabric.APortsBySysGuid) {
        if (entry.second.empty())
            continue;

        for (size_t i = 0; i < entry.second.size(); i++) {
            APort *p_aport = entry.second[i];
            if (!p_aport)
                continue;
            
            DumpAPortCounters(p_aport, sout, check_counters_bitset, en_per_lane_cnts);
        }
    }

    IBDIAG_RETURN_VOID;
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
void IBDiag::CopyPMInfoObjVector(OUT vector_p_pm_info_obj& new_pm_obj_info_vector)
{
    IBDIAG_ENTER;
    vector_p_pm_info_obj& pm_info_obj_vector =
            this->fabric_extended_info.getPMInfoObjVectorRef();

    //copy vector
    for (vector_p_pm_info_obj::iterator it = pm_info_obj_vector.begin();
            it != pm_info_obj_vector.end(); ++it)
        new_pm_obj_info_vector.push_back(*it);

    //clean vector - to need to delete because pointers are valid in new vector
    pm_info_obj_vector.clear();
    IBDIAG_RETURN_VOID;
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / IBDIAG_ERR_CODE_FABRIC_ERROR /
int IBDiag::BuildClassPortInfoDB(list_p_fabric_general_err& pm_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    static bool buildDB = true;
    if(!buildDB) //run only once
        IBDIAG_RETURN(rc);
    buildDB = false;

    INFO_PRINT("Build PMClassPortInfo\n");
    struct IB_ClassPortInfo class_port_info;
    CLEAR_STRUCT(class_port_info);
    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_node->getInSubFabric())
            continue;

        u_int32_t startPort = 0;
        u_int32_t numPorts = 0;
        if (p_node->type == IB_SW_NODE) {
            startPort = 0;
            numPorts = 1;
        } else {
            startPort = 1;
            numPorts = p_node->numPorts + 1;
        }

        // go over all base lids of this node:
        for (u_int32_t i = startPort; i < numPorts; ++i) {
            IBPort *p_port = p_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_port || (i != 0 && !p_port->is_data_worthy()))
                continue;

            struct IB_ClassPortInfo *p_class_port_info =
                    this->fabric_extended_info.getPMClassPortInfo(p_node->createIndex);

            // send class port info if not exist
            if (!p_class_port_info) {
                clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::PMClassPortInfoClbck>;
                clbck_data.m_data1 = p_node;
                progress_bar.push(p_node);
                this->ibis_obj.PMClassPortInfoGet(
                    p_port->base_lid, &class_port_info, &clbck_data);
            }
            break;
        }
        if(ibDiagClbck.GetState())
            goto exit;
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("BuildClassPortInfoDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildClassPortInfoDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!pm_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
//This may send class port info MAD if cap mask not exist.
int IBDiag::ReadCapMask(IBNode *p_node,
                        IBPort *p_port,
                        u_int16_t &cap_mask,
                        u_int32_t &port_info_cap_mask)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    struct IB_ClassPortInfo *p_class_port_info =
            this->fabric_extended_info.getPMClassPortInfo(p_node->createIndex);

    if (!p_class_port_info) {
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    cap_mask = p_class_port_info->CapMsk;
    rc = ReadPortInfoCapMask(p_node, p_port, port_info_cap_mask);
    IBDIAG_RETURN(rc);
}

int IBDiag::ReadPortInfoCapMask(IBNode *p_node,
                                IBPort *p_port,
                                u_int32_t &port_info_cap_mask,
                                u_int16_t *p_port_info_cap_mask2)
{
    IBDIAG_ENTER;

    IBPort *p_tmp_port = p_port;

    if (p_port->p_node->type == IB_SW_NODE) {
        p_tmp_port = p_port->p_node->getPort(0);
        if (!p_tmp_port) {
            this->SetLastError("DB error - can not found manage port for switch=%s\n",
                    p_port->p_node->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
    }

    struct SMP_PortInfo *p_port_info =
        this->fabric_extended_info.getSMPPortInfo(p_tmp_port->createIndex);

    if (!p_port_info) {
        this->SetLastError("DB error - can not found port info for port=%s\n",
                        p_tmp_port->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    port_info_cap_mask = p_port_info->CapMsk;

    if (p_port_info_cap_mask2)
        *p_port_info_cap_mask2 = p_port_info->CapMsk2;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

u_int8_t IBDiag::PMIsOptionalAttrSupported(IBNode* p_node, int attr_id)
{
    IBDIAG_ENTER;
    u_int8_t rc = 0;

    IBPort* p_port = p_node->getFirstPort();

    if (!p_port) // no port found
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    struct PM_PortSamplesControl* p_samples_control =
            this->fabric_extended_info.getPMPortSamplesControl(p_port->createIndex);

    if (!p_samples_control) {
        IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                   "Error DB - Can not found PortSampleControl for " \
                   "node=%s\n",
                   p_node->name.c_str());
        IBDIAG_RETURN(rc);
    }

    struct PortSampleControlOptionMask* p_option_mask = &p_samples_control->PortSampleControlOptionMask;

    switch (attr_id) {
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_DATA_VL:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_DATA_VL:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_PKT_VL:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_PKT_VL:
            rc = this->capability_module.IsSupportedGMPCapability(p_node,
                    EnGMPCapIsPerVLCountersSupported);
            break;
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_DATA_VL_EXT:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_DATA_VL_EXT:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_PKT_VL_EXT:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_PKT_VL_EXT:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_WAIT_VL_EXT:
            rc = this->capability_module.IsSupportedGMPCapability(p_node,
                    EnGMPCapIsPerVLExtendedSupported);
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_XMIT_DATA_SL:
            rc = p_option_mask->PortXmitDataSL_n;
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_RCV_DATA_SL:
            rc = p_option_mask->PortRcvDataSL_n;
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_XMIT_DATA_SL_EXT:
            rc= p_option_mask->PortXmitDataSLExt_n;
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_RCV_DATA_SL_EXT:
            rc = p_option_mask->PortRcvDataSLExt_n;
            break;

        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_VL_XMIT_FLOW_CTL_UPDATE_ERRORS:
            rc = p_option_mask->PortVlXmitFlowCtlUpdateErrors_PortVLXmitCtlUpdateErrors_n;
            break;

        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_VL_XMIT_WAIT:
            rc = p_option_mask->PortVLXmitWaitCounters_PortVLXmitWait_n;
            break;

        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_RCV_ERROR_DETAILS:
            rc = p_option_mask->PortRcvErrorDetails_PortLocalPhysicalErrors ||
                 p_option_mask->PortRcvErrorDetails_PortMalformedPacketErrors ||
                 p_option_mask->PortRcvErrorDetails_PortBufferOverrunErrors ||
                 p_option_mask->PortRcvErrorDetails_PortDLIDMappingErrors ||
                 p_option_mask->PortRcvErrorDetails_PortVlMappingErrors ||
                 p_option_mask->PortRcvErrorDetails_PortLoopingErrors;
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_XMIT_DISCARD_DETAILS:
            rc = p_option_mask->PortXmitDiscardDetails_PortInactiveDiscards ||
                 p_option_mask->PortXmitDiscardDetails_PortNeighborMTUDiscards ||
                 p_option_mask->PortXmitDiscardDetails_PortSwLifetimeLimitDiscards ||
                 p_option_mask->PortXmitDiscardDetails_PortSwHOQLifeLimitDiscards;
            break;
        default:
            break;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildPMPortSamplesControlDB(list_p_fabric_general_err& pm_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    static bool buildDB = true;
    if(!buildDB) //run only once
        IBDIAG_RETURN(rc);
    buildDB = false;

    rc = BuildClassPortInfoDB(pm_errors);
    SCREEN_PRINT("\n");
    if (rc)
        IBDIAG_RETURN(rc);

    PRINT("\n");
    INFO_PRINT("Build PMPortSampleControl\n");
    ProgressBarPorts progress_bar;
    struct PM_PortSamplesControl sample_control;
    CLEAR_STRUCT(sample_control);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortSamplesControlGetClbck>;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_node = (*nI).second;
        if (!p_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s",
                    (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_node->getInSubFabric())
            continue;

        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {

            IBPort *p_port = p_node->getPort(i);

            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_port || !p_port->is_data_worthy())
                continue;

            clbck_data.m_data1 = p_port;
            progress_bar.push(p_port);

            this->ibis_obj.PMPortSampleControlGet(p_port->base_lid,
                                                  i, // port
                                                  &sample_control,
                                                  &clbck_data);
        }

        if (ibDiagClbck.GetState())
            goto exit;
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("BuildPMPortSamplesControlDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildPMPortSamplesControlDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!pm_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::BuildPMPortSamplesResult(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortSamplesResultClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct PM_PortSamplesResult pm_port_samples_result;
    CLEAR_STRUCT(pm_port_samples_result);

   for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s",
                (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

         for (phys_port_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort(i);
            if (!p_curr_port ||!p_curr_port->is_data_worthy())
                continue;


            clbck_data.m_data1 = p_curr_port;
            progress_bar.push(p_curr_port);

            this->ibis_obj.PMPortSamplesResultGet(p_curr_port->base_lid, 
                                                  i, 
                                                  &pm_port_samples_result, 
                                                  &clbck_data);

        }
    }
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}



//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildPortCountersDB(list_p_fabric_general_err& pm_errors,
                                u_int32_t check_counters_bitset, bool send_port_samples_result)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ResetAppData();
    ibDiagClbck.Set(this, &fabric_extended_info, &pm_errors);

    int rc = BuildPMPortSamplesControlDB(pm_errors);  //include 'BuildClassPortInfoDB'
    //don't exit, if only some of the nodes failed to retrieve ClassPortInfo
    if (rc && rc != IBDIAG_ERR_CODE_FABRIC_ERROR) {
        IBDIAG_RETURN(rc);
    }

    if (send_port_samples_result) {
        rc = BuildPMPortSamplesResult(pm_errors);
        if (rc && rc != IBDIAG_ERR_CODE_FABRIC_ERROR) {
            IBDIAG_RETURN(rc);
        }
    }

    int rc2;

    SCREEN_PRINT("\n");
    PRINT("\n");
    INFO_PRINT("Build Port Counters\n");
    ProgressBarPorts progress_bar;

    control_pm_stage_flags_t control_flags;
    this->ControlGetStageFlags(this->curr_control_session, STAGE_NAME_PM, &control_flags.flags);

    // go over all nodes:
    u_int32_t port_info_cap_mask = 0;
    u_int16_t curr_pm_cap_mask = 0;
    bool to_read_cap = true;
    struct PM_PortCounters curr_port_counters;
    struct PM_PortCountersExtended curr_port_counters_extended;
    struct PM_PortExtendedSpeedsCounters curr_port_ext_speeds_counters;
    struct PM_PortExtendedSpeedsRSFECCounters    curr_port_ext_speeds_RSFEC_counters;
    struct VendorSpec_PortLLRStatistics curr_port_llr_statistics;
    struct PM_PortRcvErrorDetails curr_port_rcv_error_details;
    struct PM_PortXmitDiscardDetails curr_port_xmit_discard_details;

    CLEAR_STRUCT(curr_port_counters);
    CLEAR_STRUCT(curr_port_counters_extended);
    CLEAR_STRUCT(curr_port_ext_speeds_counters);
    CLEAR_STRUCT(curr_port_ext_speeds_RSFEC_counters);
    CLEAR_STRUCT(curr_port_llr_statistics);
    CLEAR_STRUCT(curr_port_rcv_error_details);
    CLEAR_STRUCT(curr_port_xmit_discard_details);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        //get cap mask for this node
        to_read_cap = true;

        //check LLR support by DeviceID
        struct SMP_NodeInfo * p_curr_node_info =
            this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_curr_node_info) {
            this->SetLastError("DB error - failed to get SMPNodeInfo for node: %s", p_curr_node->name.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }
        if (!this->capability_module.IsSupportedGMPCapability(
            p_curr_node, EnGMPCapIsPortLLRStatisticsSupported)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s: DeviceID %u (0x%x) Does not support LLR, skipping \n",
                       p_curr_node->name.c_str(),
                       p_curr_node_info->DeviceID,
                       p_curr_node_info->DeviceID);

            p_curr_node->appData1.val |= NOT_SUPPORT_LLR_COUNTERS;
        }

        //check PM_PortRcvErrorDetails and PM_PortXmitDiscardDetails support by DeviceID
        if (!PMIsOptionalAttrSupported(p_curr_node,
                                       IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_RCV_ERROR_DETAILS)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                       "node %s: DeviceID %u (0x%x) Does not support PortRcvErrorDetails,"
                       " skipping \n",
                       p_curr_node->name.c_str(),
                       p_curr_node_info->DeviceID,
                       p_curr_node_info->DeviceID);
            p_curr_node->appData1.val |= NOT_SUPPORT_PORT_RCV_ERROR_DETAILS;
        }

        if (!PMIsOptionalAttrSupported(p_curr_node,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_XMIT_DISCARD_DETAILS)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                       "node %s: DeviceID %u (0x%x) Does not support PortXmitDiscardDetails,"
                       " skipping \n",
                       p_curr_node->name.c_str(),
                       p_curr_node_info->DeviceID,
                       p_curr_node_info->DeviceID);
            p_curr_node->appData1.val |= NOT_SUPPORT_PORT_XMIT_DISCARD_DETAILS;
        }


        // go over all ports of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            //for switch read cap only on port 0, for others read on every port
            if (to_read_cap) {
                rc2 = ReadCapMask(p_curr_node, p_curr_port,
                                  curr_pm_cap_mask, port_info_cap_mask);
                if (rc2) {
                    rc = rc2;
                    if (rc == IBDIAG_ERR_CODE_FABRIC_ERROR) {
                        IBDIAG_LOG(TT_LOG_LEVEL_MAD, "node:%s Failed to read cap mask - skip to next node.\n",
                                   (*nI).first.c_str());
                        goto next_node;     //not reasonable to continue
                    }
                    goto exit;
                }
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node:%s pm_cap_mask:0x%x port_info_cap_mask:0x%x\n",
                           (*nI).first.c_str(), curr_pm_cap_mask, port_info_cap_mask);
                if (p_curr_node->type == IB_SW_NODE)
                    to_read_cap = false;
            }

            if (control_flags.pm.port_counters) {
                //send port counters mad get for this port
                clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortCountersGetClbck>;
                clbck_data.m_data1 = p_curr_port;
                progress_bar.push(p_curr_port);
                rc2 = this->ibis_obj.PMPortCountersGet(p_curr_port->base_lid,
                    p_curr_port->num, &curr_port_counters, &clbck_data);
                if (rc2)
                    continue;  //we will try to proceed without query for extended counters
            }

            //check if xmit wait is supported
            if (GetSpecialCAPortType(p_curr_node) == IB_NOT_SPECIAL_PORT && // skip special nodes
                !(p_curr_node->appData1.val & NOT_SUPPORT_XMIT_WAIT)) {
                if (IS_SUPPORT_XMIT_WAIT(curr_pm_cap_mask) == 0) {

                    pm_errors.push_back(
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support xmit wait counter capability"));

                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                    //mark we already checked for support
                    p_curr_node->appData1.val |= NOT_SUPPORT_XMIT_WAIT;
                }
            }

            if (control_flags.pm.port_counters_extended) {
                //check if counters extended are supported
                if (!IS_SUPPORT_EXT_PORT_FAILED(p_curr_node->appData1.val)) {
                    if ( IS_SUPPORT_EXTENDED_COUNTERS(curr_pm_cap_mask) ||
                         IS_SUPPORT_EXTENDED_COUNTERS_NO_IETF(curr_pm_cap_mask)) {

                        clbck_data.m_handle_data_func =
                            &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortCountersExtendedGetClbck>;

                        progress_bar.push(p_curr_port);
                        this->ibis_obj.PMPortCountersExtendedGet(p_curr_port->base_lid,
                            p_curr_port->num, &curr_port_counters_extended, &clbck_data);
                    } else {

                        pm_errors.push_back(
                                new FabricErrNodeNotSupportCap(p_curr_node,
                                        "This device does not support extended port counters capability"));

                        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                        //mark we already checked for support
                        p_curr_node->appData1.val |= NOT_SUPPORT_EXT_PORT_COUNTERS;
                    }
                }
            }

            if (control_flags.pm.extended_speeds) {
                //check if extended speed counters supported
                if(IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, p_curr_node->type) &&
                   !IS_SUPPORT_EXT_SPEEDS_FAILED(p_curr_node->appData1.val)) {

                    if(IS_SUPPORT_EXTENDED_SPEEDS_COUNTERS(port_info_cap_mask)) {
                        //if RS_FEC is active
                        if (isRSFEC(p_curr_port->get_fec_mode())) {
                            if (!IS_SUPPORT_EXT_SPEEDS_RSFEC_FAILED(
                                    p_curr_node->appData1.val)) {
                                if (IS_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS(
                                    curr_pm_cap_mask)) {
                                    clbck_data.m_handle_data_func =
                                        &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortExtendedSpeedsRSFECGetClbck>;
                                    progress_bar.push(p_curr_port);
                                    this->ibis_obj.
                                        PMPortExtendedSpeedsRSFECCountersGet(
                                        p_curr_port->base_lid,
                                        p_curr_port->num,
                                        &curr_port_ext_speeds_RSFEC_counters,
                                        &clbck_data);
                                } else {
                                    pm_errors.push_back(
                                            new FabricErrNodeNotSupportCap(
                                                    p_curr_node, "This device does not "
                                                    "support port extended speeds RSFEC "
                                                    "counters capability"));
                                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                                    //mark we already checked for support
                                    p_curr_node->appData1.val |=
                                        NOT_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS;
                                }
                            }
                        } else {//not RS_FEC
                            clbck_data.m_handle_data_func =
                                &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortExtendedSpeedsGetClbck>;

                            progress_bar.push(p_curr_port);
                            this->ibis_obj.PMPortExtendedSpeedsCountersGet(p_curr_port->base_lid,
                                    p_curr_port->num, &curr_port_ext_speeds_counters, &clbck_data);
                        }

                    } else {

                        pm_errors.push_back(
                                new FabricErrNodeNotSupportCap(p_curr_node,
                                        "This device does not support port extended speeds counters capability"));

                        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                        //mark we already checked for support
                        p_curr_node->appData1.val |= NOT_SUPPORT_EXT_SPEEDS_COUNTERS;
                    }
                } //ext_speeds_counters
            }

            if (control_flags.pm.llr_statistics) {
                if( !IS_SUPPORT_LLR_FAILED(p_curr_node->appData1.val)) {

                    clbck_data.m_handle_data_func =
                        &forwardClbck<IBDiagClbck, &IBDiagClbck::VSPortLLRStatisticsGetClbck>;

                    progress_bar.push(p_curr_port);
                    this->ibis_obj.VSPortLLRStatisticsGet(p_curr_port->base_lid,
                            p_curr_port->num, &curr_port_llr_statistics,
                            IS_TO_PRINT_LLR_COUNTERS(check_counters_bitset), &clbck_data);

                } //LLR
            }

            if (control_flags.pm.port_rcv_error_details) {
                // PortRcvErrorDetails
                if (!IS_SUPPORT_PORT_RCV_ERROR_DETAILS(p_curr_node->appData1.val)) {

                    clbck_data.m_handle_data_func =
                        &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortRcvErrorDetailsGetClbck>;

                    progress_bar.push(p_curr_port);
                    this->ibis_obj.PMPortRcvErrorDetailsGet(p_curr_port->base_lid,
                                                            p_curr_port->num,
                                                            &curr_port_rcv_error_details,
                                                            &clbck_data);
                }
            }

            if (control_flags.pm.port_xmit_discard_details) {
                // PortXmitDiscardDetails
                if (!IS_SUPPORT_PORT_XMIT_DISCARD_DETAILS(p_curr_node->appData1.val)) {

                    clbck_data.m_handle_data_func =
                        &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortXmitDiscardDetailsGetClbck>;

                    progress_bar.push(p_curr_port);
                    this->ibis_obj.PMPortXmitDiscardDetailsGet(p_curr_port->base_lid,
                                                               p_curr_port->num,
                                                               &curr_port_xmit_discard_details,
                                                               &clbck_data);
                }
            }

        }   //current node ports loop

        if (ibDiagClbck.GetState())
            goto exit;

next_node:
        continue;
    }   //node loop
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("BuildPortCountersDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildPortCountersDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!pm_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildHBFCounters(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::VSPortRoutingDecisionCountersGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->isHBFSupported())
            continue;

        const IBPort* p_port_zero = p_curr_node->getPort(0);
        if (!p_port_zero)
            continue;

        for (u_int8_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {

             // PortRoutingDecisionCounters counters shouldn't be sent to special ports,
             // unconnected SW ports
             IBPort* p_port = p_curr_node->getPort(port_num);
             if (!p_port                       ||
                 !p_port->is_data_worthy()     ||
                  p_port->isSpecialPort()      ||
                 !p_port->p_remotePort         ||
                 !p_port->p_remotePort->p_node)
                  continue;

             clbck_data.m_data1 = p_port;
             progress_bar.push(p_port);

             port_routing_decision_counters counters = {0};
             this->ibis_obj.VSPortRoutingDecisionCountersGet(p_port_zero->base_lid,
                                                             port_num,
                                                             &counters,
                                                             &clbck_data);

             if (ibDiagClbck.GetState())
                 break;
        } // ports
    } // switches

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::ClearHBFCounters(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
                &forwardClbck<IBDiagClbck, &IBDiagClbck::VSPortRoutingDecisionCountersClearClbck>;

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->isHBFSupported())
            continue;

        IBPort* p_port_zero = p_curr_node->getPort(0);
        if (!p_port_zero)
            continue;

        for (u_int8_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {

             // PortRoutingDecisionCounters counters shouldn't be sent to special ports,
             IBPort* p_port = p_curr_node->getPort(port_num);
             if (!p_port                       ||
                 !p_port->is_data_worthy()     ||
                  p_port->isSpecialPort()      ||
                 !p_port->p_remotePort         ||
                 !p_port->p_remotePort->p_node)
                  continue;

             clbck_data.m_data1 = p_port;

             this->ibis_obj.VSPortRoutingDecisionCountersClear(p_port_zero->base_lid,
                                                               port_num,
                                                               &clbck_data);

             if (ibDiagClbck.GetState())
                 break;
        } // ports
    }//swicthes

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

// Build RN Counters
int IBDiag::BuildRNCounters(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::VSPortRNCountersGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct port_rn_counters rn_counters;
    CLEAR_STRUCT(rn_counters);

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBPort* p_port_zero = p_curr_node->getPort(0);
        if (!p_port_zero)
            continue;

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_curr_node->getInSubFabric() || !p_curr_node->isRNSupported())
            continue;

        for (u_int8_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {

            // RN counters shouldn't be sent to special ports, unconnected SW ports
            IBPort* p_port = p_curr_node->getPort(port_num);
            if (!p_port ||
                    !p_port->is_data_worthy() ||
                    p_port->isSpecialPort() ||
                    !p_port->p_remotePort ||
                    !p_port->p_remotePort->p_node)
                continue;

            clbck_data.m_data1 = p_port;
            progress_bar.push(p_port);

            this->ibis_obj.VSPortRNCountersGet(p_port_zero->base_lid,
                                               port_num,
                                               &rn_counters,
                                               &clbck_data);

            if (ibDiagClbck.GetState())
                break;
        } // ports
    } // switches

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


// Clear RN Counters
int IBDiag::ClearRNCounters(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
                &forwardClbck<IBDiagClbck, &IBDiagClbck::VSPortRNCountersClearClbck>;

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBPort* p_port_zero = p_curr_node->getPort(0);
        if (!p_port_zero)
            continue;

        if (!p_curr_node->getInSubFabric() || !p_curr_node->isRNSupported())
            continue;

        for (u_int8_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {

            // RN counters shouldn't be sent to special ports, unconnected SW ports
            IBPort* p_port = p_curr_node->getPort(port_num);
            if (!p_port ||
                    !p_port->is_data_worthy() ||
                    p_port->isSpecialPort() ||
                    !p_port->p_remotePort ||
                    !p_port->p_remotePort->p_node)
                continue;

            clbck_data.m_data1 = p_port;

            this->ibis_obj.VSPortRNCountersClear(p_port_zero->base_lid,
                                                 port_num,
                                                 &clbck_data);

            if (ibDiagClbck.GetState())
                break;
        } // ports
    } // switches

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpRNCountersToCSV(CSVOut &csv_out, list_p_fabric_general_err& pfrn_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_RN_COUNTERS))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNumber,"
            << "port_rcv_rn_pkt,port_xmit_rn_pkt,port_rcv_rn_error,"
            << "port_rcv_switch_relay_rn_error,port_ar_trials,"
            << "pfrn_received_packet,pfrn_received_error,pfrn_xmit_packet,pfrn_start_packet"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (set_pnode::const_iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        const IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->isRNSupported())
            continue;

        const adaptive_routing_info* p_ar_info =
                this->fabric_extended_info.getARInfo(p_curr_node->createIndex);

        if (!p_ar_info)
            continue;

        for (u_int8_t port = 1;  port <= p_curr_node->numPorts ; ++port) {
             IBPort *p_port = p_curr_node->getPort(port);

             // the following counters shouldn't be printed for special ports, unconnected SW ports
             if (!p_port ||
                     !p_port->is_data_worthy() ||
                     p_port->isSpecialPort() ||
                     !p_port->p_remotePort ||
                     !p_port->p_remotePort->p_node)
                 continue;

             const port_rn_counters* p_rn_counters =
                    this->fabric_extended_info.getRNCounters(p_port->createIndex);

             if (!p_rn_counters)
                 continue;

             sstream.str("");

             sstream << PTR(p_curr_node->guid_get())                     << ","
                     << PTR(p_port->guid_get())                          << ","
                     << +p_port->num                                     << ","
                     << p_rn_counters->port_rcv_rn_pkt                   << ","
                     << p_rn_counters->port_xmit_rn_pkt                  << ","
                     << p_rn_counters->port_rcv_rn_error                 << ","
                     << p_rn_counters->port_rcv_switch_relay_rn_error    << ",";

             if (p_ar_info->is_ar_trials_supported)
                 sstream << p_rn_counters->port_ar_trials    << ",";
             else
                 sstream << "N/A,";

             if (p_ar_info->is_pfrn_supported)
                 sstream << p_rn_counters->pfrn_received_packet    << ","
                         << p_rn_counters->pfrn_received_error     << ","
                         << p_rn_counters->pfrn_xmit_packet        << ","
                         << p_rn_counters->pfrn_start_packet       ;
             else
                 sstream << "N/A,N/A,N/A,N/A";

             sstream << endl;

             csv_out.WriteBuf(sstream.str());

             // add warnings for pfrn_received_error
             if (p_ar_info->is_pfrn_supported && p_rn_counters->pfrn_received_error) {

                 pFRNReceivedErrorNotZeroErr *p_curr_err =
                         new pFRNReceivedErrorNotZeroErr(p_port, p_rn_counters->pfrn_received_error);

                 p_curr_err->SetLevel(EN_FABRIC_ERR_WARNING);
                 pfrn_errors.push_back(p_curr_err);
             }
        } //ports
    } //nodes

    csv_out.DumpEnd(SECTION_RN_COUNTERS);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteRNCounters_2_File(const string& file_name)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("RN counters 2",
                            file_name,
                            sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    IBFabric::GetSwitchLabelPortNumExplanation(sout);
    rc = this->DumpRNCounters_2_Info(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpRNCounters_2_Info(ostream& sout)
{
    IBDIAG_ENTER;

    RNMaxData rn_max_data;

    sout << "File version: 1" << endl << endl;

    for (set_pnode::const_iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        const IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_curr_node->isRNSupported() && !p_curr_node->isHBFSupported())
            continue;

        const adaptive_routing_info* p_ar_info =
                this->fabric_extended_info.getARInfo(p_curr_node->createIndex);

        for (u_int8_t port = 1;  port <= p_curr_node->numPorts ; ++port) {
             const IBPort *p_port = p_curr_node->getPort(port);

             // the following counters shouldn't be printed for special ports, unconnected SW ports
             if (!p_port ||
                 !p_port->is_data_worthy() ||
                 p_port->isSpecialPort() ||
                 !p_port->p_remotePort ||
                 !p_port->p_remotePort->p_node)
                continue;

             const port_rn_counters* p_rn_counters =
                    this->fabric_extended_info.getRNCounters(p_port->createIndex);
             const port_routing_decision_counters* p_routing_decision_counters =
                    this->fabric_extended_info.getRoutingDecisionCounters(p_port->createIndex);
             if ((!p_rn_counters || !p_ar_info) &&
                 (!p_curr_node->isHBFSupported() || !p_routing_decision_counters))
                 continue;

             sout << "---------------------------------------------------------------" << endl;
             sout << "Port=" << +p_port->num
                  << " Lid=" << p_port->base_lid
                  << " GUID=" << PTR(p_port->guid_get())
                  << " Device=" <<  p_port->p_node->devId
                  << " Port Name=" << p_port->getName() << endl;
             sout << "---------------------------------------------------------------" << endl;

             if (p_rn_counters && p_ar_info)
                 DumpRNCounters_2_Info(sout, *p_rn_counters, *p_ar_info, rn_max_data);

             if (p_curr_node->isHBFSupported() && p_routing_decision_counters)
                 DumpHBFCounters_2_Info(sout, *p_routing_decision_counters);
        }//ports
    }//nodes

    sout << "*******************************************************************************"
         << "************" << endl << endl;

    sout << "Max Values:" << endl;
    sout << "===========" << endl;

    sout << "Max Rcv RN Pkt: "    << rn_max_data.port_rcv_rn_pkt << endl
         << "Max Xmit RN Pkt: "   << rn_max_data.port_xmit_rn_pkt << endl
         << "Max Rcv RN Error: "  << rn_max_data.port_rcv_rn_error << endl
         << "Max Rcv SW Relay RN Error: " << rn_max_data.sw_relay_rn_error << endl;

    if (rn_max_data.is_port_ar_trials_supported_in_fabric)
        sout << "Max Port AR Trials: " << rn_max_data.port_ar_trials << endl;
    else
        sout << "Max Port AR Trials: N/A" << endl;

    if (rn_max_data.is_pfrn_supported_in_fabric) {
        sout << "Max Rcv pFRN Pkt: "    << rn_max_data.pfrn_received_packet  << endl
             << "Max Rcv pFRN Error: "  << rn_max_data.pfrn_received_error   << endl
             << "Max Xmit pFRN Pkt: "   << rn_max_data.pfrn_xmit_packet      << endl
             << "Max Start pFRN Pkt: "  << rn_max_data.pfrn_start_packet     << endl;
    } else {
        sout << "Max Rcv pFRN Pkt: N/A"    << endl
             << "Max Rcv pFRN Error: N/A"  << endl
             << "Max Xmit pFRN Pkt: N/A"   << endl
             << "Max Start pFRN Pkt: N/A"  << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::DumpHBFCounters_2_Info(ostream &sout,
            const port_routing_decision_counters &routing_decision_counters) const
{
    IBDIAG_ENTER;

    sout << "rx_pkt_forwarding_static="  <<
                routing_decision_counters.rx_pkt_forwarding_static    << endl
        << "rx_pkt_forwarding_hbf="      <<
                routing_decision_counters.rx_pkt_forwarding_hbf       << endl
        << "rx_pkt_forwarding_ar="       <<
                routing_decision_counters.rx_pkt_forwarding_ar        << endl
        << "rx_pkt_hbf_fallback_local="  <<
                routing_decision_counters.rx_pkt_hbf_fallback_local   << endl
        << "rx_pkt_hbf_fallback_remote=" <<
                routing_decision_counters.rx_pkt_hbf_fallback_remote  << endl
        << "rx_pkt_forwarding_hbf_sg0="  <<
                routing_decision_counters.rx_pkt_forwarding_hbf_sg0   << endl
        << "rx_pkt_forwarding_hbf_sg1="  <<
                routing_decision_counters.rx_pkt_forwarding_hbf_sg1   << endl
        << "rx_pkt_forwarding_hbf_sg2="  <<
                routing_decision_counters.rx_pkt_forwarding_hbf_sg2   << endl
        << "rx_pkt_forwarding_ar_sg0="   <<
                routing_decision_counters.rx_pkt_forwarding_ar_sg0    << endl
        << "rx_pkt_forwarding_ar_sg1="   <<
                routing_decision_counters.rx_pkt_forwarding_ar_sg1    << endl
        << "rx_pkt_forwarding_ar_sg2="   <<
                routing_decision_counters.rx_pkt_forwarding_ar_sg2    << endl;

    sout << endl;

    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpRNCounters_2_Info(ostream &sout,
            const port_rn_counters &rn_counters, const adaptive_routing_info &ar_info,
            RNMaxData &rn_max_data) const
{
    IBDIAG_ENTER;

    sout << "port_rcv_rn_pkt="  << rn_counters.port_rcv_rn_pkt   << endl
        << "port_xmit_rn_pkt="  << rn_counters.port_xmit_rn_pkt  << endl
        << "port_rcv_rn_error=" << rn_counters.port_rcv_rn_error << endl
        << "port_rcv_switch_relay_rn_error=" <<
                    rn_counters.port_rcv_switch_relay_rn_error   << endl;

        if (ar_info.is_ar_trials_supported) {
            sout << "port_ar_trials=" << rn_counters.port_ar_trials << endl;
            rn_max_data.is_port_ar_trials_supported_in_fabric = true;
        } else {
            sout << "port_ar_trials=N/A"                            << endl;
        }

        if (ar_info.is_pfrn_supported) {
            sout << "pfrn_received_packet=" << rn_counters.pfrn_received_packet << endl
                << "pfrn_received_error="   << rn_counters.pfrn_received_error  << endl
                << "pfrn_xmit_packet="      << rn_counters.pfrn_xmit_packet     << endl
                << "pfrn_start_packet="     << rn_counters.pfrn_start_packet    << endl;
            rn_max_data.is_pfrn_supported_in_fabric = true;
        } else {
            sout << "pfrn_received_packet=N/A" << endl
                << "pfrn_received_error=N/A"   << endl
                << "pfrn_xmit_packet=N/A"      << endl
                << "pfrn_start_packet=N/A"     << endl;
        }

        sout << endl;

        rn_max_data.port_rcv_rn_pkt =
                max(rn_max_data.port_rcv_rn_pkt, rn_counters.port_rcv_rn_pkt);

        rn_max_data.port_xmit_rn_pkt =
                max(rn_max_data.port_xmit_rn_pkt, rn_counters.port_xmit_rn_pkt);

        rn_max_data.port_rcv_rn_error =
                max(rn_max_data.port_rcv_rn_error, rn_counters.port_rcv_rn_error);

        rn_max_data.sw_relay_rn_error =
                max(rn_max_data.sw_relay_rn_error, rn_counters.port_rcv_switch_relay_rn_error);

        if (rn_max_data.is_port_ar_trials_supported_in_fabric)
            rn_max_data.port_ar_trials =
                    max(rn_max_data.port_ar_trials, rn_counters.port_ar_trials);

        if (ar_info.is_pfrn_supported) {
            rn_max_data.pfrn_received_packet =
                        max(rn_max_data.pfrn_received_packet, rn_counters.pfrn_received_packet);
            rn_max_data.pfrn_received_error =
                        max(rn_max_data.pfrn_received_error, rn_counters.pfrn_received_error);
            rn_max_data.pfrn_xmit_packet =
                        max(rn_max_data.pfrn_xmit_packet, rn_counters.pfrn_xmit_packet);
            rn_max_data.pfrn_start_packet =
                        max(rn_max_data.pfrn_start_packet, rn_counters.pfrn_start_packet);
        }

    IBDIAG_RETURN_VOID;
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteRNCountersFile(const string& file_name)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("RN counters",
                            file_name,
                            sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->DumpRNCountersInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(rc);
}


int IBDiag::DumpRNCountersInfo(ofstream& sout)
{
    IBDIAG_ENTER;
    char buffer[2096];

    bool is_port_ar_trials_supported_in_fabric = false;

    u_int64_t max_port_rcv_rn_pkt = 0;
    u_int64_t max_port_xmit_rn_pkt = 0;
    u_int64_t max_port_rcv_rn_error = 0;
    u_int64_t max_sw_relay_rn_error = 0;
    u_int64_t max_port_ar_trials = 0;

    sout << "File version: 2" << endl;


    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric() || !p_curr_node->isRNSupported())
            continue; //continue to next switch

        adaptive_routing_info* ar_info =
                this->fabric_extended_info.getARInfo(p_curr_node->createIndex);
        if (!ar_info) // AR info not exist
            continue;

        snprintf(buffer, sizeof(buffer), "\n\ndump_rnc: Switch " U64H_FMT, p_curr_node->guid_get());
        sout << buffer << endl << endl;

        sout << std::setw(30) << std::left << "Port" <<
                std::setw(30) << std::left << "Rcv RN Pkt" <<
                std::setw(30) << std::left << "Xmit RN Pkt" <<
                std::setw(30) << std::left << "Rcv RN Error" <<
                std::setw(30) << std::left << "Rcv SW Relay RN Error" <<
                "Ingress Port AR Trials" << endl;

        sout << "#--------------------------------------------------------------------------"
             << "---------------------------------------------------------------------------"
             << "----------------------" << endl;

        for (u_int8_t port = 1;  port <= p_curr_node->numPorts ; ++port) {

            IBPort *p_port = p_curr_node->getPort(port);

            // RN counters shouldn't be printed for special ports, unconnected SW ports
            if (!p_port ||
                    !p_port->is_data_worthy() ||
                    p_port->isSpecialPort() ||
                    !p_port->p_remotePort ||
                    !p_port->p_remotePort->p_node)
                continue; // continue to next SW port

            port_rn_counters* p_rn_counters =
                    this->fabric_extended_info.getRNCounters(p_port->createIndex);

            if (!p_rn_counters)
                continue; // port_rn_counters not exist - continue to next SW port

            sout << std::setw(30) << std::left << (int)port <<
                    std::setw(30) << std::left <<
                    p_rn_counters->port_rcv_rn_pkt <<
                    std::setw(30) << std::left <<
                    p_rn_counters->port_xmit_rn_pkt <<
                    std::setw(30) << std::left <<
                    p_rn_counters->port_rcv_rn_error <<
                    std::setw(30) << std::left <<
                    p_rn_counters->port_rcv_switch_relay_rn_error;

                if (ar_info->is_ar_trials_supported) {
                    is_port_ar_trials_supported_in_fabric = true;
                    sout << p_rn_counters->port_ar_trials << endl;
                } else {
                    sout << "N/A" << endl;
                }

            max_port_rcv_rn_pkt =
                max(max_port_rcv_rn_pkt, p_rn_counters->port_rcv_rn_pkt);

            max_port_xmit_rn_pkt =
                max(max_port_xmit_rn_pkt, p_rn_counters->port_xmit_rn_pkt);

            max_port_rcv_rn_error =
                max(max_port_rcv_rn_error, p_rn_counters->port_rcv_rn_error);

            max_sw_relay_rn_error =
                max(max_sw_relay_rn_error, p_rn_counters->port_rcv_switch_relay_rn_error);

            if (ar_info->is_ar_trials_supported)
                max_port_ar_trials =
                    max(max_port_ar_trials, p_rn_counters->port_ar_trials);
        }

        sout << "\n#*****************************************************************************"
             << "********************************************************************************"
             << "**************" << endl;

    }

    sout << "#*******************************************************************************"
         << "********************************************************************************"
         << "************" << endl;

    sout << "\nMax Values:" << endl;
    sout <<   "#==========\n\n" << endl;

    sout << "Max Rcv RN Pkt: " <<
        max_port_rcv_rn_pkt <<
        "   Max Xmit RN Pkt: " <<
        max_port_xmit_rn_pkt <<
        "   Max Rcv RN Error: " <<
        max_port_rcv_rn_error <<
        "   Max Rcv SW Relay RN Error: " <<
        max_sw_relay_rn_error <<
        "   Max Port AR Trials: ";
        if (is_port_ar_trials_supported_in_fabric)
            sout << max_port_ar_trials;
        else
           sout << "N/A";

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


bool IBDiag::HandleSpecialPorts(CountersPerSLVL * cntrs_per_slvl,
                                SMP_MlnxExtPortInfo *p_curr_mepi,
                                IBPort *p_curr_port,
                                int &rc,
                                list_p_fabric_general_err &cntrs_per_slvl_errors)
{
    stringstream ss;
    ss << "This special port does not support PM "
       << cntrs_per_slvl->GetCntrHeader()
       << " MAD. type = "
       << +p_curr_mepi->SpecialPortType;

    FabricErrPortNotSupportCap *p_curr_fabric_port_err =
        new FabricErrPortNotSupportCap(p_curr_port, ss.str());

    p_curr_fabric_port_err->SetLevel(EN_FABRIC_ERR_WARNING);
    cntrs_per_slvl_errors.push_back(p_curr_fabric_port_err);

    IBDIAG_RETURN(true);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildDBOrResetSLVLCntrs(
        list_p_fabric_general_err &cntrs_per_slvl_errors,
        bool is_reset_cntr,
        bool to_dump_sup_warn,
        CountersPerSLVL * cntrs_per_slvl)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &cntrs_per_slvl_errors,
                    NULL, &capability_module);

    int rc_psc = BuildPMPortSamplesControlDB(cntrs_per_slvl_errors);
    SCREEN_PRINT("\n");
    //don't exit, if only some of the nodes failed to retrieve PortSampleControl
    if (rc_psc && rc_psc != IBDIAG_ERR_CODE_FABRIC_ERROR) {
        IBDIAG_RETURN(rc_psc);
    }

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::GSIPerSLVLGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct PM_PortRcvXmitCntrsSlVl pm_port_rcvxmit_data_slvl;
    CLEAR_STRUCT(pm_port_rcvxmit_data_slvl);

    for (phys_port_t curr_port_idx = 1;
         curr_port_idx < this->discovered_fabric.GetMaxNodePorts() + 1;
         ++curr_port_idx) {

        for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
                nI != this->discovered_fabric.NodeByName.end(); ++nI) {

            IBNode *p_curr_node = (*nI).second;
            if (!p_curr_node) {
                this->SetLastError(
                        "DB error - found null node in NodeByName map for key = %s",
                        (*nI).first.c_str());
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            if (p_curr_node->isSpecialNode())
                continue;

            if (cntrs_per_slvl->GetAttrId() == IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_VL_XMIT_TIME_CONG_EXT &&
                    p_curr_node->type != IB_SW_NODE)
                continue;

            if (!p_curr_node->getInSubFabric())
                continue;

            //skip if this node has less ports
            if (p_curr_node->numPorts < curr_port_idx)
                continue;

            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)curr_port_idx);
            if (!p_curr_port)                       //port not exist
                continue;

            if (p_curr_port->isSpecialPort())
                continue;

            //Skip ports that are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port->is_data_worthy())
                continue;

            // don't check cap bit for PORT_VL_XMIT_FLOW_CTL_UPDATE_ERRORS.
            // sampling is currently not supported for this counter
            // so bit in PortSamplesControl.OptionMask is off.
            // todo - remove this if when cap bit is on from FW side.
            if (cntrs_per_slvl->GetAttrId() !=
                    IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_VL_XMIT_FLOW_CTL_UPDATE_ERRORS) {

                //appData1 is for marking "supported"/"not supported" attributes
                if (PM_PER_SLVL_ATTR_IS_NOT_SUPPORT_CAPABILTY(p_curr_node,
                        cntrs_per_slvl->GetAttrCapBit()))
                    continue;

                //appData2 is for marking "support checked"/"supported not checked" attributes
                if (PM_PER_SLVL_ATTR_IS_SUPPORT_NOT_CHECKED_CAPABILTY(p_curr_node,
                        cntrs_per_slvl->GetAttrCapBit())) {
                    //mark "supported checked"/"supported not checked" appdata (appData2)
                    p_curr_node->appData2.val |= cntrs_per_slvl->GetAttrCapBit();
                    if (!PMIsOptionalAttrSupported(p_curr_node, cntrs_per_slvl->GetAttrId())) {
                        //mark "supported"/"not supported" appdata (appData1)
                        p_curr_node->appData1.val |= cntrs_per_slvl->GetAttrCapBit();

                        if (to_dump_sup_warn) {
                            FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                                new FabricErrNodeNotSupportCap(p_curr_node,
                                                               "This device does not support PM " +
                                                               cntrs_per_slvl->GetCntrHeader() +
                                                               " capability");

                            p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                            cntrs_per_slvl_errors.push_back(p_curr_fabric_node_err);
                        }

                        continue;
                    }
                }
            }

            // special ports
            if (p_curr_node->type == IB_SW_NODE) {
                if (!this->capability_module.IsSupportedSMPCapability(
                            p_curr_node, EnSMPCapIsSpecialPortsMarkingSupported)) {

                    // if not checked yet for this node
                    if (!(p_curr_node->appData2.val & NOT_SUPPORT_SPECIAL_PORTS_MARKING_CHECKED)) {

                        p_curr_node->appData2.val |= NOT_SUPPORT_SPECIAL_PORTS_MARKING_CHECKED;

                        if (to_dump_sup_warn) {
                            FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                                new FabricErrNodeNotSupportCap(p_curr_node,
                                        "This device does not support Special Ports Marking capability");

                            p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                            cntrs_per_slvl_errors.push_back(p_curr_fabric_node_err);
                        }
                    }

                } else { // Special Ports Marking Supported

                    struct SMP_MlnxExtPortInfo *p_curr_mepi =
                            fabric_extended_info.getSMPMlnxExtPortInfo(p_curr_port->createIndex);
                    if (!p_curr_mepi) {
                        IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                                   "Error DB - Can not found MlnxExtPortInfo for " \
                                   "node=%s, port=%u\n",
                                   p_curr_node->name.c_str(), curr_port_idx);

                        rc = IBDIAG_ERR_CODE_DB_ERR;
                        IBDIAG_RETURN(rc);
                    }

                    if (p_curr_mepi->IsSpecialPort) {

                        switch (cntrs_per_slvl->GetAttrId()) {
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_DATA_VL:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_DATA_VL:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_PKT_VL:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_PKT_VL:

                                if (!(p_curr_mepi->SpecialPortCapabilityMask &
                                        (1 << EnSPCapIsPerVLCountersSupported))) {
                                    if (to_dump_sup_warn)
                                        if (!HandleSpecialPorts(cntrs_per_slvl, p_curr_mepi,
                                                                p_curr_port, rc, cntrs_per_slvl_errors))
                                            goto exit;

                                    continue;
                                }
                                break;

                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_DATA_VL_EXT:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_DATA_VL_EXT:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_PKT_VL_EXT:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_PKT_VL_EXT:

                                if (!(p_curr_mepi->SpecialPortCapabilityMask &
                                        (1 << EnSPCapIsPerVLExtendedSupported))) {
                                    if (to_dump_sup_warn)
                                        if (!HandleSpecialPorts(cntrs_per_slvl, p_curr_mepi,
                                                                p_curr_port, rc, cntrs_per_slvl_errors))
                                            goto exit;
                                   continue;
                                }
                                break;

                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_WAIT_VL_EXT:

                                if (!(p_curr_mepi->SpecialPortCapabilityMask &
                                        (1 << EnSPCapIsPortXmitWaitVLExtendedSupported))) {
                                    if (to_dump_sup_warn)
                                        if (!HandleSpecialPorts(cntrs_per_slvl, p_curr_mepi,
                                                                p_curr_port, rc, cntrs_per_slvl_errors))
                                            goto exit;
                                    continue;
                                }
                        }
                    }
                }
            } // special ports

            /*
             * Send the MAD
             */
            clbck_data.m_data1 = p_curr_port;
            clbck_data.m_data2 = cntrs_per_slvl;

            progress_bar.push(p_curr_port);
            if (!cntrs_per_slvl->IsVSClass()) {
                this->ibis_obj.PMPerSLVLCounters(
                        is_reset_cntr,
                        p_curr_port->base_lid,
                        curr_port_idx,
                        cntrs_per_slvl->GetAttrId(),
                        &pm_port_rcvxmit_data_slvl,
                        &clbck_data);
            } else {
                this->ibis_obj.VSPerVLCounters(
                        is_reset_cntr,
                        p_curr_port->base_lid,
                        curr_port_idx,
                        cntrs_per_slvl->GetAttrId(),
                        &pm_port_rcvxmit_data_slvl,
                        &clbck_data);
            }

            if(ibDiagClbck.GetState())
                goto exit;
        }
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of VS PerSLVLCounters Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "Retrieve of VS PerSLVLCounters Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!cntrs_per_slvl_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


static void WritePortCountersHeadersToCsv(CSVOut &csv_out, u_int32_t check_counters_bitset)
{
    IBDIAG_ENTER;

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," "PortNumber,"
            << "LinkDownedCounter," << "LinkErrorRecoveryCounter," << "SymbolErrorCounter,"
            << "PortRcvRemotePhysicalErrors," << "PortRcvErrors," << "PortXmitDiscards,"
            << "PortRcvSwitchRelayErrors," << "ExcessiveBufferOverrunErrors," << "LocalLinkIntegrityErrors,"
            << "PortRcvConstraintErrors," << "PortXmitConstraintErrors," << "VL15Dropped,"
            << "PortXmitData," << "PortRcvData,"
            << "PortXmitPkts," << "PortRcvPkts,"
            << "PortXmitWait,"
            // Port Counters Extended
            << "PortXmitDataExtended," << "PortRcvDataExtended,"
            << "PortXmitPktsExtended," << "PortRcvPktsExtended,"
            << "PortUniCastXmitPkts," << "PortUniCastRcvPkts,"
            << "PortMultiCastXmitPkts," << "PortMultiCastRcvPkts,"
            << "SymbolErrorCounterExt," << "LinkErrorRecoveryCounterExt,"
            << "LinkDownedCounterExt," << "PortRcvErrorsExt,"
            << "PortRcvRemotePhysicalErrorsExt," << "PortRcvSwitchRelayErrorsExt,"
            << "PortXmitDiscardsExt," << "PortXmitConstraintErrorsExt,"
            << "PortRcvConstraintErrorsExt," << "LocalLinkIntegrityErrorsExt,"
            << "ExcessiveBufferOverrunErrorsExt," << "VL15DroppedExt,"
            << "PortXmitWaitExt," << "QP1DroppedExt";

    if (IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, IB_SW_NODE)) {
        const char * port_extended_speeds_counters_names[] = {"ErrorDetectionCounterLane",
                                                              "FECCorrectableBlockCounterLane",
                                                              "FECUncorrectableBlockCounterLane",
                                                              "FECCorrectedSymbolCounterLane",
                                                             };

        sstream << ",SyncHeaderErrorCounter,UnknownBlockCounter";

        size_t max_ext_speeds_counters = sizeof(port_extended_speeds_counters_names) / sizeof (char *);
        for (size_t i=0 ; i < max_ext_speeds_counters ; i++) {
            for (u_int16_t j=0 ; j < MAX_LANE_NUMBER ; j++) {
                sstream << "," << port_extended_speeds_counters_names[i] << "[" << j << "]";
            }
        }

        sstream << ",PortFECCorrectableBlockCounter,"
                << "PortFECUncorrectableBlockCounter,PortFECCorrectedSymbolCounter";

    }

    // calculated counter
    sstream << ",retransmission_per_sec, max_retransmission_rate";

    // PortRcvErrorDetails counters
    sstream << ",PortLocalPhysicalErrors,PortMalformedPacketErrors"
         << ",PortBufferOverrunErrors,PortDLIDMappingErrors"
         << ",PortVLMappingErrors,PortLoopingErrors";

    // PortXmitDiscardDetails counters
    sstream << ",PortInactiveDiscards,PortNeighborMTUDiscards"
         << ",PortSwLifetimeLimitDiscards,PortSwHOQLifetimeLimitDiscards" << endl;

    csv_out.WriteBuf(sstream.str());

    IBDIAG_RETURN_VOID;
}


#define DUMP(FMT, condition, X, na_val)              \
  __DUMP_IMPL(FMT, condition, X, #X, na_val)

#define DUMP_NAMED(FMT, condition, X, NAME, na_val)             \
  __DUMP_IMPL(FMT, condition, X, NAME, na_val)

#define DUMP_EXT(FMT, condition, X, EXT, na_val)                \
  __DUMP_IMPL(FMT, condition, X, #X EXT, na_val)

#define __DUMP_IMPL(FMT, condition, X, NAME, na_val)             \
    if (!(condition) || !p_curr)                                 \
        stream << "," << na_val;                                 \
    else {                                                       \
        stream << ',' ;                                          \
        if (!p_prev)                                             \
            stream << FMT(p_curr->X);                            \
        else  {                                                  \
            if (p_prev->X > p_curr->X)  {                        \
                (*p_negative_names) << ' ' << (NAME);            \
                stream << "ERR";                                 \
            } else                                               \
                stream << FMT(PM_Delta(p_curr->X, p_prev->X));   \
        }                                                        \
    }

static void PM_PortCounters_ToCSV(ostream &stream,
                                  const PM_PortCounters *p_curr,
                                  const PM_PortCounters *p_prev,
                                  ostream *p_negative_names)
{
    IBDIAG_ENTER;

    DUMP(DEC, true, LinkDownedCounter, "-1");
    DUMP(DEC, true, LinkErrorRecoveryCounter, "-1");
    DUMP(DEC, true, SymbolErrorCounter, "-1");
    DUMP(DEC, true, PortRcvRemotePhysicalErrors, "-1");
    DUMP(DEC, true, PortRcvErrors, "-1");
    DUMP(DEC, true, PortXmitDiscards, "-1");
    DUMP(DEC, true, PortRcvSwitchRelayErrors, "-1");
    DUMP(DEC, true, ExcessiveBufferOverrunErrors, "-1");
    DUMP(DEC, true, LocalLinkIntegrityErrors, "-1");
    DUMP(DEC, true, PortRcvConstraintErrors, "-1");
    DUMP(DEC, true, PortXmitConstraintErrors, "-1");
    DUMP(DEC, true, VL15Dropped, "-1");
    DUMP(DEC, true, PortXmitData, "-1");
    DUMP(DEC, true, PortRcvData, "-1");
    DUMP(DEC, true, PortXmitPkts, "-1");
    DUMP(DEC, true, PortRcvPkts, "-1");
    DUMP(DEC, true, PortXmitWait, "-1");

    IBDIAG_RETURN_VOID;
}

static void PM_PortExtendedCounters_ToCSV(ostream &stream,
                                          IB_ClassPortInfo *pm_class_port_info,
                                          const PM_PortCountersExtended *p_curr,
                                          const PM_PortCountersExtended *p_prev,
                                          ostream *p_negative_names)
{
    IBDIAG_ENTER;

    DUMP_EXT(DEC, true, PortXmitData, "Extended", "-1");
    DUMP_EXT(DEC, true, PortRcvData, "Extended", "-1");
    DUMP_EXT(DEC, true, PortXmitPkts, "Extended", "-1");
    DUMP_EXT(DEC, true, PortRcvPkts, "Extended", "-1");

    DUMP(DEC, true, PortUniCastXmitPkts, "-1");
    DUMP(DEC, true, PortUniCastRcvPkts, "-1");
    DUMP(DEC, true, PortMultiCastXmitPkts, "-1");
    DUMP(DEC, true, PortMultiCastRcvPkts, "-1");

    bool isSupported = pm_class_port_info && IS_SUPPORT_ADDITIONAL_EXTENDED_COUNTERS(pm_class_port_info->CapMsk2);
    DUMP_EXT(DEC, isSupported, SymbolErrorCounter, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, LinkErrorRecoveryCounter, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, LinkDownedCounter, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, PortRcvErrors, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, PortRcvRemotePhysicalErrors, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, PortRcvSwitchRelayErrors, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, PortXmitDiscards, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, PortXmitConstraintErrors, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, PortRcvConstraintErrors, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, LocalLinkIntegrityErrors, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, ExcessiveBufferOverrunErrors, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, VL15Dropped, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, PortXmitWait, "Ext", "-1");
    DUMP_EXT(DEC, isSupported, QP1Dropped, "Ext", "-1");

    IBDIAG_RETURN_VOID;
}

static void Common_ToCSV(ostream &stream,
                    const PM_PortExtendedSpeedsCounters *p_curr_reg,
                    const PM_PortExtendedSpeedsCounters *p_prev_reg,
                    const PM_PortExtendedSpeedsRSFECCounters *p_curr_rsfec,
                    const PM_PortExtendedSpeedsRSFECCounters *p_prev_rsfec,
                    ostream *p_negative_names)
{
    IBDIAG_ENTER;

    if (!p_curr_reg && !p_curr_rsfec)
       stream << ",-1,-1";
    else if (p_curr_reg) {
        const PM_PortExtendedSpeedsCounters *p_curr = p_curr_reg;
        const PM_PortExtendedSpeedsCounters *p_prev = p_prev_reg;

        DUMP(DEC, true, SyncHeaderErrorCounter,  "-1");
        DUMP(DEC, true, UnknownBlockCounter, "-1");
    }
    else {
        const PM_PortExtendedSpeedsRSFECCounters *p_curr = p_curr_rsfec;
        const PM_PortExtendedSpeedsRSFECCounters *p_prev = p_prev_rsfec;

        DUMP(DEC, true, SyncHeaderErrorCounter, "-1");
        DUMP(DEC, true, UnknownBlockCounter, "-1");
    }

    IBDIAG_RETURN_VOID;
}

static string IndexedName(const char* p_name, size_t index) {
    IBDIAG_ENTER;

    stringstream  ss;
    ss << p_name << '[' << index << ']';

    IBDIAG_RETURN(ss.str());
}

static void ErrorDetectionCounterLane_ToCSV(ostream &stream,
                                        const PM_PortExtendedSpeedsCounters *p_curr,
                                        const PM_PortExtendedSpeedsCounters *p_prev,
                                        ostream *p_negative_names)
{
    IBDIAG_ENTER;

    size_t size = ARRAY_SIZE(p_curr->ErrorDetectionCounterLane);
    for (size_t i = 0; i < size; i++)
        DUMP_NAMED(DEC, true, ErrorDetectionCounterLane[i],
                IndexedName("ErrorDetectionCounterLane", i), "-1");

    IBDIAG_RETURN_VOID;
}

static void BlockCounterLane_ToCSV(ostream &stream, IBFECMode fec_mode,
                                const PM_PortExtendedSpeedsCounters *p_curr,
                                const PM_PortExtendedSpeedsCounters *p_prev,
                                ostream *p_negative_names)
{
    IBDIAG_ENTER;

    size_t size = ARRAY_SIZE(p_curr->FECCorrectableBlockCountrLane);
    for (size_t i = 0; i < size; i++)
        DUMP_NAMED(DEC, fec_mode != IB_FEC_NO_FEC, FECCorrectableBlockCountrLane[i],
                IndexedName("FECCorrectableBlockCounterLane", i), "-1");

    size = ARRAY_SIZE(p_curr->FECUncorrectableBlockCounterLane);
    for (size_t i = 0; i < size; i++)
        DUMP_NAMED(DEC, fec_mode != IB_FEC_NO_FEC, FECUncorrectableBlockCounterLane[i],
                IndexedName("FECUncorrectableBlockCounterLane", i), "-1");

    IBDIAG_RETURN_VOID;
}

static void RSFEC_ToCSV(ostream &stream,
                        const PM_PortExtendedSpeedsRSFECCounters *p_curr,
                        const PM_PortExtendedSpeedsRSFECCounters *p_prev,
                        ostream *p_negative_names)
{
    IBDIAG_ENTER;

    const size_t size = ARRAY_SIZE(p_curr->FECCorrectedSymbolCounterLane);
    for (size_t i = 0; i < size; i++)
        DUMP_NAMED(DEC, true, FECCorrectedSymbolCounterLane[i],
                IndexedName("FECCorrectedSymbolCounterLane", i), "-1");

    DUMP(DEC, true, PortFECCorrectableBlockCounter, "-1");
    DUMP(DEC, true, PortFECUncorrectableBlockCounter, "-1");
    DUMP(DEC, true, PortFECCorrectedSymbolCounter, "-1");

    IBDIAG_RETURN_VOID;
}

static void PM_PortExtendedSpeedCounter_ToCSV(ostream &stream, IBFECMode fec_mode,
                                            const PM_PortExtendedSpeedsCounters *p_curr_reg,
                                            const PM_PortExtendedSpeedsCounters *p_prev_reg,
                                            const PM_PortExtendedSpeedsRSFECCounters *p_curr_rsfec,
                                            const PM_PortExtendedSpeedsRSFECCounters *p_prev_rsfec,
                                            ostream *p_negative_names)
{
    IBDIAG_ENTER;

    Common_ToCSV(stream, p_curr_reg, p_prev_reg,
                        p_curr_rsfec, p_prev_rsfec,
                        p_negative_names);

    ErrorDetectionCounterLane_ToCSV(stream, p_curr_reg, p_prev_reg, p_negative_names);

    BlockCounterLane_ToCSV(stream, fec_mode, p_curr_reg, p_prev_reg, p_negative_names);

    RSFEC_ToCSV(stream, p_curr_rsfec, p_prev_rsfec, p_negative_names);

    IBDIAG_RETURN_VOID;
}

static void PM_PortCalcCounter_ToCSV(ostream &stream,
                                    const PM_PortCalcCounters *p_curr,
                                    const PM_PortCalcCounters *p_prev,
                                    ostream *p_negative_names)
{
    IBDIAG_ENTER;

    DUMP_NAMED(PTR, true, RetransmissionPerSec, "retransmission_per_sec", "0xfffffffffffffffe");

    IBDIAG_RETURN_VOID;
}

static void PM_PortLLRStatisticsCounter_ToCSV(ostream &stream, bool isSupported,
                                            const VendorSpec_PortLLRStatistics *p_curr,
                                            const VendorSpec_PortLLRStatistics *p_prev,
                                            ostream *p_negative_names)
{
    IBDIAG_ENTER;

    DUMP_NAMED(PTR, isSupported, MaxRetransmissionRate, "max_retransmission_rate", "-1");

    IBDIAG_RETURN_VOID;
}

static void PM_PortRcvErrorDetailsCounter_ToCSV(ostream &stream,
                                            const PortSampleControlOptionMask *p_option_mask,
                                            const PM_PortRcvErrorDetails *p_curr,
                                            const PM_PortRcvErrorDetails *p_prev,
                                            ostream *p_negative_names)
{
    IBDIAG_ENTER;

    DUMP(DEC, p_option_mask && p_option_mask->PortRcvErrorDetails_PortLocalPhysicalErrors,
                PortLocalPhysicalErrors, "-1");
    DUMP(DEC, p_option_mask && p_option_mask->PortRcvErrorDetails_PortMalformedPacketErrors,
                PortMalformedPacketErrors, "-1");
    DUMP(DEC, p_option_mask && p_option_mask->PortRcvErrorDetails_PortBufferOverrunErrors,
                PortBufferOverrunErrors, "-1");
    DUMP(DEC, p_option_mask && p_option_mask->PortRcvErrorDetails_PortDLIDMappingErrors,
                PortDLIDMappingErrors, "-1");
    DUMP(DEC, p_option_mask && p_option_mask->PortRcvErrorDetails_PortVlMappingErrors,
                PortVLMappingErrors, "-1");
    DUMP(DEC, p_option_mask && p_option_mask->PortRcvErrorDetails_PortLoopingErrors,
                PortLoopingErrors, "-1");

    IBDIAG_RETURN_VOID;
}

static void PM_PortXmitDiscardDetailsCounter_ToCSV(ostream &stream,
                                                const PortSampleControlOptionMask *p_option_mask,
                                                const PM_PortXmitDiscardDetails * p_curr,
                                                const PM_PortXmitDiscardDetails * p_prev,
                                                ostream *p_negative_names)
{
    IBDIAG_ENTER;

    static const string postfix = "";
    static const string name = "";
    DUMP(DEC, p_option_mask && p_option_mask->PortXmitDiscardDetails_PortInactiveDiscards,
                PortInactiveDiscards, "-1");
    DUMP(DEC, p_option_mask && p_option_mask->PortXmitDiscardDetails_PortNeighborMTUDiscards,
                PortNeighborMTUDiscards, "-1");
    DUMP(DEC, p_option_mask && p_option_mask->PortXmitDiscardDetails_PortSwLifetimeLimitDiscards,
                PortSwLifetimeLimitDiscards, "-1");
    DUMP(DEC, p_option_mask && p_option_mask->PortXmitDiscardDetails_PortSwHOQLifeLimitDiscards,
                PortSwHOQLifetimeLimitDiscards, "-1");

    IBDIAG_RETURN_VOID;
}

int IBDiag::DumpPortCountersDeltaCSVTable(CSVOut &csv_out,
                            const vector_p_pm_info_obj &prev_pm_info_obj_vec,
                            u_int32_t check_counters_bitset,
                            list_p_fabric_general_err& ports_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    // cannot be skipped if disabled because generate warnings
    (void) csv_out.DumpStart(SECTION_PM_DELTA);

    WritePortCountersHeadersToCsv(csv_out, check_counters_bitset);

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        if (prev_pm_info_obj_vec.size() < i + 1)    //info not exists
            continue;

        if (!prev_pm_info_obj_vec[i])               //info not exists
            continue;

        //treshlod is checked for error counters and they should be OK.
        //for the rest of counters use abs value.

        stringstream sstream;
        stringstream negative_ss;

        //regular counters
        {
            const PM_PortCounters *p_prev = prev_pm_info_obj_vec[i]->p_port_counters;
            const PM_PortCounters *p_curr = this->fabric_extended_info.getPMPortCounters(i);

            if (!p_curr || !p_prev)
                continue;

            sstream << PTR(p_curr_port->p_node->guid_get())   << ","
                    << PTR(p_curr_port->guid_get())           << ","
                    << +p_curr_port->num;

            PM_PortCounters_ToCSV(sstream, p_curr, p_prev, &negative_ss);
        }

        //For printing delta
        //we need both p_curr and p_prev pointers not being null.

        //extended_port_counters
        {
            const PM_PortCountersExtended *p_prev =
                            prev_pm_info_obj_vec[i]->p_extended_port_counters;
            const PM_PortCountersExtended *p_curr =
                            this->fabric_extended_info.getPMPortCountersExtended(i);

            IB_ClassPortInfo *pm_class_port_info = this->fabric_extended_info.
                    getPMClassPortInfo(p_curr_port->p_node->createIndex);

            if (p_prev && p_curr)
                PM_PortExtendedCounters_ToCSV(sstream, pm_class_port_info,
                                        p_curr, p_prev, &negative_ss);
            else
                PM_PortExtendedCounters_ToCSV(sstream, pm_class_port_info,
                                        NULL, NULL, &negative_ss);
        }

        if (IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, IB_SW_NODE)) {
            //ext_speeds_counters
            const PM_PortExtendedSpeedsCounters *p_prev =
                            prev_pm_info_obj_vec[i]->p_port_ext_speeds_counters;
            const PM_PortExtendedSpeedsCounters *p_curr =
                            this->fabric_extended_info.getPMPortExtSpeedsCounters(i);

            if (!p_prev || !p_curr)
                p_prev = p_curr = NULL;

            //ext_speeds_counters for rsfec
            const PM_PortExtendedSpeedsRSFECCounters *p_prev_rsfec =
                            prev_pm_info_obj_vec[i]->p_port_ext_speeds_rsfec_counters;
            const PM_PortExtendedSpeedsRSFECCounters *p_curr_rsfec =
                            this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(i);

            if (!p_prev_rsfec || !p_curr_rsfec)
                p_prev_rsfec = p_curr_rsfec = NULL;

            //first print counters common to both layouts:
            //regular extended speed counters and rsfec extended speed counters
            PM_PortExtendedSpeedCounter_ToCSV(sstream, p_curr_port->get_fec_mode(),
                                            p_curr, p_prev,
                                            p_curr_rsfec, p_prev_rsfec,
                                            &negative_ss);
        }

        //calculated counters
        {
            const PM_PortCalcCounters *p_prev =
                        prev_pm_info_obj_vec[i]->p_port_calc_counters;
            const PM_PortCalcCounters *p_curr =
                        this->fabric_extended_info.getPMPortCalcCounters(i);

            if (p_curr && p_prev)
                PM_PortCalcCounter_ToCSV(sstream, p_curr, p_prev, &negative_ss);
            else
                PM_PortCalcCounter_ToCSV(sstream, NULL, NULL, &negative_ss);
        }

        //LLRStatistics counters
        {
            const VendorSpec_PortLLRStatistics *p_prev =
                        prev_pm_info_obj_vec[i]->p_port_llr_statistics;
            const  VendorSpec_PortLLRStatistics *p_curr =
                        this->fabric_extended_info.getVSPortLLRStatistics(i);

            bool isSupported = this->capability_module.IsSupportedGMPCapability(
                p_curr_port->p_node,
                EnGMPCapIsMaxRetransmissionRateSupported);

            if (p_curr && p_prev)
                PM_PortLLRStatisticsCounter_ToCSV(sstream, isSupported,
                                            p_curr, p_prev, &negative_ss);
            else
                PM_PortLLRStatisticsCounter_ToCSV(sstream, isSupported,
                                            NULL, NULL, &negative_ss);
        }

        struct PM_PortSamplesControl* p_samples_control =
                this->fabric_extended_info.getPMPortSamplesControl(p_curr_port->createIndex);

        struct PortSampleControlOptionMask* p_option_mask = NULL;

        if (p_samples_control)
            p_option_mask = &p_samples_control->PortSampleControlOptionMask;

        //PortRcvErrorDetails
        {
            const PM_PortRcvErrorDetails *p_curr =
                    this->fabric_extended_info.getPMPortRcvErrorDetails(i);
            const PM_PortRcvErrorDetails *p_prev =
                    prev_pm_info_obj_vec[i]->p_port_rcv_error_details;

            if (p_curr && p_prev)
                PM_PortRcvErrorDetailsCounter_ToCSV(sstream, p_option_mask,
                                                p_curr, p_prev, &negative_ss);
            else
                PM_PortRcvErrorDetailsCounter_ToCSV(sstream, p_option_mask,
                                                NULL, NULL, &negative_ss);
        }

        //XmitDiscard
        {
            const  PM_PortXmitDiscardDetails *p_curr =
                    this->fabric_extended_info.getPMPortXmitDiscardDetails(i);
            const  PM_PortXmitDiscardDetails *p_prev =
                    prev_pm_info_obj_vec[i]->p_port_xmit_discard_details;

            if (p_curr && p_prev)
                PM_PortXmitDiscardDetailsCounter_ToCSV(sstream, p_option_mask,
                                                p_curr, p_prev, &negative_ss);
            else
                PM_PortXmitDiscardDetailsCounter_ToCSV(sstream, p_option_mask,
                                                NULL, NULL, &negative_ss);
        }

        sstream << endl;
        csv_out.WriteBuf(sstream.str());

        string str = negative_ss.str();
        if (!str.empty())
            ports_errors.push_back(new FabricErrPMInvalidDelta(p_curr_port, str));
    }

    csv_out.DumpEnd(SECTION_PM_DELTA);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpPMPortSamplesControlCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_PM_PORT_SAMPLES_CONTROL))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNumber,"
            << "Tick,CounterWidth,CounterMask0,CounterMasks1to9,"
            << "CounterMasks10to14,SampleMechanisms,SampleStatus,OptionMask,"
            << "VendorMask,SampleStart,SampleInterval,Tag,"
            << "CounterSelect0,CounterSelect1,CounterSelect2,CounterSelect3,"
            << "CounterSelect4,CounterSelect5,CounterSelect6,CounterSelect7,"
            << "CounterSelect8,CounterSelect9,CounterSelect10,CounterSelect11,"
            << "CounterSelect12,CounterSelect13,CounterSelect14"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {

        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        struct PM_PortSamplesControl* p_samples_control =
                this->fabric_extended_info.getPMPortSamplesControl(p_curr_port->createIndex);

        if (!p_samples_control)   // PM_PortSamplesControl not exist
            continue;

        sstream.str("");

        sstream << PTR(p_curr_port->p_node->guid_get())          << ","
                << PTR(p_curr_port->guid_get())                  << ","
                << +p_curr_port->num                             << ","
                << PTR(p_samples_control->Tick)                  << ","
                << PTR(p_samples_control->CounterWidth)          << ","
                << PTR(p_samples_control->CounterMask0)          << ","
                << PTR(p_samples_control->CounterMasks1to9)      << ","
                << PTR(p_samples_control->CounterMasks10to14)    << ","
                << PTR(p_samples_control->SampleMechanisms)      << ","
                << PTR(p_samples_control->SampleStatus)          << ","
                << PTR(PMOptionMaskTouint64(p_samples_control->PortSampleControlOptionMask)) << ","
                << PTR(p_samples_control->VendorMask)            << ","
                << PTR(p_samples_control->SampleStart)           << ","
                << PTR(p_samples_control->SampleInterval)        << ","
                << PTR(p_samples_control->Tag)                   ;

        for (int idx = 0; idx < 15; ++idx) {
            sstream << "," << PTR(p_samples_control->CounterSelect[idx]);
        }

        sstream << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_PM_PORT_SAMPLES_CONTROL);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpPMPortSamplesResultToCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_PM_PORT_SAMPLES_RESULT))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNumber,Tag,SampleStatus,"
            << "Counter0,Counter1,Counter2,Counter3,Counter4,Counter5,Counter6,"
            << "Counter7,Counter8,Counter9,Counter10,Counter11,Counter12,Counter13,Counter14" << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)
            continue;

        PM_PortSamplesResult* p_port_samples_result = this->fabric_extended_info.
                                    getPMPortSamplesResult(p_curr_port->createIndex);
        if (!p_port_samples_result)
            continue;

        sstream.str("");
        sstream  << PTR(p_curr_port->p_node->guid_get())  << ','
                 << PTR(p_curr_port->guid_get())          << ','
                 << +p_curr_port->num                     << ','
                 << +p_port_samples_result->Tag           << ','
                 << +p_port_samples_result->SampleStatus  << ','
                 << +p_port_samples_result->Counter0      << ','
                 << +p_port_samples_result->Counter1      << ','
                 << +p_port_samples_result->Counter2      << ','
                 << +p_port_samples_result->Counter3      << ','
                 << +p_port_samples_result->Counter4      << ','
                 << +p_port_samples_result->Counter5      << ','
                 << +p_port_samples_result->Counter6      << ','
                 << +p_port_samples_result->Counter7      << ','
                 << +p_port_samples_result->Counter8      << ','
                 << +p_port_samples_result->Counter9      << ','
                 << +p_port_samples_result->Counter10     << ','
                 << +p_port_samples_result->Counter11     << ','
                 << +p_port_samples_result->Counter12     << ','
                 << +p_port_samples_result->Counter13     << ','
                 << +p_port_samples_result->Counter14     << endl;


        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_PM_PORT_SAMPLES_RESULT);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

u_int64_t IBDiag::PMOptionMaskTouint64(struct PortSampleControlOptionMask &pm_option_mask)
{
    IBDIAG_ENTER;

    u_int64_t mask = 0;

    // Bit 0 Reserved
    if (pm_option_mask.PortXmitQueue_n) mask |= 1lu << 1;
    if (pm_option_mask.PortXmitDataVl_n) mask |= 1lu << 2;
    if (pm_option_mask.PortRcvDataVl_n) mask |= 1lu << 3;
    if (pm_option_mask.PortXmitPktVl_n) mask |= 1lu << 4;
    if (pm_option_mask.PortRcvPktVl_n) mask |= 1lu << 5;
    if (pm_option_mask.PortRcvErrorDetails_PortLocalPhysicalErrors) mask |= 1lu << 6;
    if (pm_option_mask.PortRcvErrorDetails_PortMalformedPacketErrors) mask |= 1lu << 7;
    if (pm_option_mask.PortRcvErrorDetails_PortBufferOverrunErrors) mask |= 1lu << 8;
    if (pm_option_mask.PortRcvErrorDetails_PortDLIDMappingErrors) mask |= 1lu << 9;
    if (pm_option_mask.PortRcvErrorDetails_PortVlMappingErrors) mask |= 1lu << 10;
    if (pm_option_mask.PortRcvErrorDetails_PortLoopingErrors) mask |= 1lu << 11;
    if (pm_option_mask.PortXmitDiscardDetails_PortInactiveDiscards) mask |= 1lu << 12;
    if (pm_option_mask.PortXmitDiscardDetails_PortNeighborMTUDiscards) mask |= 1lu << 13;
    if (pm_option_mask.PortXmitDiscardDetails_PortSwLifetimeLimitDiscards) mask |= 1lu << 14;
    if (pm_option_mask.PortXmitDiscardDetails_PortSwHOQLifeLimitDiscards) mask |= 1lu << 15;
    if (pm_option_mask.PortOpRcvCounters_PortOpRcvPkts) mask |= 1lu << 16;
    if (pm_option_mask.PortOpRcvCounters_PortOpRcvData) mask |= 1lu << 17;
    if (pm_option_mask.PortFlowCtlCounters_PortXmitFlowPkts) mask |= 1lu << 18;
    if (pm_option_mask.PortFlowCtlCounters_PortRcvFlowPkts) mask |= 1lu << 19;
    if (pm_option_mask.PortVLOpPackets_PortVLOpPackets_n) mask |= 1lu << 20;
    if (pm_option_mask.PortVLOpData_PortVLOpData_n) mask |= 1lu << 21;
    if (pm_option_mask.PortVlXmitFlowCtlUpdateErrors_PortVLXmitCtlUpdateErrors_n) mask |= 1lu << 22;
    if (pm_option_mask.PortVLXmitWaitCounters_PortVLXmitWait_n) mask |= 1lu << 23;
    if (pm_option_mask.PortUnicastXmitPkts) mask |= 1lu << 24;
    if (pm_option_mask.PortUnicastRcvPkts) mask |= 1lu << 25;
    if (pm_option_mask.PortMulticastXmitPkts) mask |= 1lu << 26;
    if (pm_option_mask.PortMulticastRcvPkts) mask |= 1lu << 27;
    if (pm_option_mask.PortExtendedSpeedsCounter) mask |= 1lu << 28;
    // Bits 29 - 47 Reserved
    if (pm_option_mask.SwPortVLCongestion_SWPortVLCongestion_n) mask |= 1lu << 48;
    if (pm_option_mask.PortRcvConCtrl_PortPktRcvFECN) mask |= 1lu << 49;
    if (pm_option_mask.PortRcvConCtrl_PortPktRcvBECN) mask |= 1lu << 50;
    if (pm_option_mask.PortSLRcvFECN_PortSLRcvFECN_n) mask |= 1lu << 51;
    if (pm_option_mask.PortSLRcvBECN_PortSLRcvBECN_n) mask |= 1lu << 52;
    if (pm_option_mask.PortXmitConCtrl_PortXmitTimeCong) mask |= 1lu << 53;
    if (pm_option_mask.PortVLXmitTimeCong_PortVLXmitTimeCong_n) mask |= 1lu << 54;
    if (pm_option_mask.PortXmitDataSL_n) mask |= 1lu << 55;
    if (pm_option_mask.PortRcvDataSL_n) mask |= 1lu << 56;
    if (pm_option_mask.PortXmitDataSLExt_n) mask |= 1lu << 57;
    if (pm_option_mask.PortRcvDataSLExt_n) mask |= 1lu << 58;
    // Bits 59-63 Reserved.

    IBDIAG_RETURN(mask);
}


struct PortSampleControlOptionMask & IBDiag::uint64ToPMOptionMask(
                  struct PortSampleControlOptionMask &pm_option_mask, u_int64_t value)
{
    IBDIAG_ENTER;

    memset(&pm_option_mask, 0, sizeof(pm_option_mask));

    pm_option_mask.PortXmitQueue_n                                              = (value >> 1) & 0x01;
    pm_option_mask.PortXmitDataVl_n                                             = (value >> 2) & 0x01;
    pm_option_mask.PortRcvDataVl_n                                              = (value >> 3) & 0x01;
    pm_option_mask.PortXmitPktVl_n                                              = (value >> 4) & 0x01;
    pm_option_mask.PortRcvPktVl_n                                               = (value >> 5) & 0x01;
    pm_option_mask.PortRcvErrorDetails_PortLocalPhysicalErrors                  = (value >> 6) & 0x01;
    pm_option_mask.PortRcvErrorDetails_PortMalformedPacketErrors                = (value >> 7) & 0x01;
    pm_option_mask.PortRcvErrorDetails_PortBufferOverrunErrors                  = (value >> 8) & 0x01;
    pm_option_mask.PortRcvErrorDetails_PortDLIDMappingErrors                    = (value >> 9) & 0x01;
    pm_option_mask.PortRcvErrorDetails_PortVlMappingErrors                      = (value >> 10) & 0x01;
    pm_option_mask.PortRcvErrorDetails_PortLoopingErrors                        = (value >> 11) & 0x01;
    pm_option_mask.PortXmitDiscardDetails_PortInactiveDiscards                  = (value >> 12) & 0x01;
    pm_option_mask.PortXmitDiscardDetails_PortNeighborMTUDiscards               = (value >> 13) & 0x01;
    pm_option_mask.PortXmitDiscardDetails_PortSwLifetimeLimitDiscards           = (value >> 14) & 0x01;
    pm_option_mask.PortXmitDiscardDetails_PortSwHOQLifeLimitDiscards            = (value >> 15) & 0x01;
    pm_option_mask.PortOpRcvCounters_PortOpRcvPkts                              = (value >> 16) & 0x01;
    pm_option_mask.PortOpRcvCounters_PortOpRcvData                              = (value >> 17) & 0x01;
    pm_option_mask.PortFlowCtlCounters_PortXmitFlowPkts                         = (value >> 18) & 0x01;
    pm_option_mask.PortFlowCtlCounters_PortRcvFlowPkts                          = (value >> 19) & 0x01;
    pm_option_mask.PortVLOpPackets_PortVLOpPackets_n                            = (value >> 20) & 0x01;
    pm_option_mask.PortVLOpData_PortVLOpData_n                                  = (value >> 21) & 0x01;
    pm_option_mask.PortVlXmitFlowCtlUpdateErrors_PortVLXmitCtlUpdateErrors_n    = (value >> 22) & 0x01;
    pm_option_mask.PortVLXmitWaitCounters_PortVLXmitWait_n                      = (value >> 23) & 0x01;
    pm_option_mask.PortUnicastXmitPkts                                          = (value >> 24) & 0x01;
    pm_option_mask.PortUnicastRcvPkts                                           = (value >> 25) & 0x01;
    pm_option_mask.PortMulticastXmitPkts                                        = (value >> 26) & 0x01;
    pm_option_mask.PortMulticastRcvPkts                                         = (value >> 27) & 0x01;
    pm_option_mask.PortExtendedSpeedsCounter                                    = (value >> 28) & 0x01;
    // Bits 29 - 47 Reserved
    pm_option_mask.SwPortVLCongestion_SWPortVLCongestion_n                      = (value >> 48) & 0x01;
    pm_option_mask.PortRcvConCtrl_PortPktRcvFECN                                = (value >> 49) & 0x01;
    pm_option_mask.PortRcvConCtrl_PortPktRcvBECN                                = (value >> 50) & 0x01;
    pm_option_mask.PortSLRcvFECN_PortSLRcvFECN_n                                = (value >> 51) & 0x01;
    pm_option_mask.PortSLRcvBECN_PortSLRcvBECN_n                                = (value >> 52) & 0x01;
    pm_option_mask.PortXmitConCtrl_PortXmitTimeCong                             = (value >> 53) & 0x01;
    pm_option_mask.PortVLXmitTimeCong_PortVLXmitTimeCong_n                      = (value >> 54) & 0x01;
    pm_option_mask.PortXmitDataSL_n                                             = (value >> 55) & 0x01;
    pm_option_mask.PortRcvDataSL_n                                              = (value >> 56) & 0x01;
    pm_option_mask.PortXmitDataSLExt_n                                          = (value >> 57) & 0x01;
    pm_option_mask.PortRcvDataSLExt_n                                           = (value >> 58) & 0x01;
    // Bits 59-63 Reserved.

    IBDIAG_RETURN(pm_option_mask);
}
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpPortCountersCSVTable(CSVOut &csv_out, u_int32_t check_counters_bitset)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_PM_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    WritePortCountersHeadersToCsv(csv_out, check_counters_bitset);
 
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        stringstream sstream;
        //regular counters
        {
            const struct PM_PortCounters *p_curr =
                    this->fabric_extended_info.getPMPortCounters(i);
            if (!p_curr)       //info not exist
                continue;

            sstream << PTR(p_curr_port->p_node->guid_get())  << ","
                    << PTR(p_curr_port->guid_get())          << ","
                    << +p_curr_port->num;

            PM_PortCounters_ToCSV(sstream, p_curr, NULL, NULL);
        }

        //extended_port_counters
        {
            const struct PM_PortCountersExtended *p_curr =
                    this->fabric_extended_info.getPMPortCountersExtended(i);

            IB_ClassPortInfo *pm_class_port_info = this->fabric_extended_info.
                    getPMClassPortInfo(p_curr_port->p_node->createIndex);

            PM_PortExtendedCounters_ToCSV(sstream, pm_class_port_info, p_curr, NULL, NULL);
        }


        if (IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, IB_SW_NODE)) {
            //ext_speeds_counters
            const struct PM_PortExtendedSpeedsCounters *p_curr =
                this->fabric_extended_info.getPMPortExtSpeedsCounters(i);
            const struct PM_PortExtendedSpeedsRSFECCounters *p_curr_rsfec =
                this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(i);

            //first print counters common to both layouts:
            //regular extended speed counters and rsfec extended speed counters
            PM_PortExtendedSpeedCounter_ToCSV(sstream, p_curr_port->get_fec_mode(),
                                            p_curr, NULL, p_curr_rsfec, NULL, NULL);
        }

        //calculated counters
        {
            const struct PM_PortCalcCounters *p_curr =
                        this->fabric_extended_info.getPMPortCalcCounters(i);
            PM_PortCalcCounter_ToCSV(sstream, p_curr, NULL, NULL);
        }

        //LLRStatistics counters
        {
            const struct VendorSpec_PortLLRStatistics *p_curr =
                        this->fabric_extended_info.getVSPortLLRStatistics(i);
            bool isSupported = this->capability_module.IsSupportedGMPCapability(
                p_curr_port->p_node,
                EnGMPCapIsMaxRetransmissionRateSupported);

            PM_PortLLRStatisticsCounter_ToCSV(sstream, isSupported, p_curr, NULL, NULL);
        }

        struct PM_PortSamplesControl* p_samples_control =
                this->fabric_extended_info.getPMPortSamplesControl(p_curr_port->createIndex);

        struct PortSampleControlOptionMask* p_option_mask = NULL;

        if (p_samples_control)
            p_option_mask = &p_samples_control->PortSampleControlOptionMask;

        //PortRcvErrorDetails counters
        {
            const struct PM_PortRcvErrorDetails *p_curr =
                    this->fabric_extended_info.getPMPortRcvErrorDetails(i);
            PM_PortRcvErrorDetailsCounter_ToCSV(sstream, p_option_mask, p_curr, NULL, NULL);
        }

        //PM_PortXmitDiscardDetails counters
        {
            const struct PM_PortXmitDiscardDetails *p_curr =
                    this->fabric_extended_info.getPMPortXmitDiscardDetails(i);
            PM_PortXmitDiscardDetailsCounter_ToCSV(sstream, p_option_mask, p_curr, NULL, NULL);
        }

        sstream << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_PM_INFO);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::ResetPortCounters(list_p_fabric_general_err& ports_errors,
                              u_int32_t check_counters_bitset)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ResetAppData();
    ibDiagClbck.Set(this, &fabric_extended_info, &ports_errors);

    int rc = BuildClassPortInfoDB(ports_errors);
    //don't exit, if only some of the nodes failed to retrieve ClassPortInfo
    if (rc && rc != IBDIAG_ERR_CODE_FABRIC_ERROR) {
        IBDIAG_RETURN(rc);
    }

    SCREEN_PRINT("\n");
    PRINT("\n");
    INFO_PRINT("Reset Port Counters\n");
    ProgressBarPorts progress_bar;

    u_int32_t port_info_cap_mask = 0;
    u_int16_t curr_pm_cap_mask = 0;
    bool to_read_cap = true;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes:
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        //read capability for new node
        to_read_cap = true;

        //check LLR support by DeviceID
        struct SMP_NodeInfo * p_curr_node_info =
            this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_curr_node_info) {
            this->SetLastError("DB error - failed to get SMPNodeInfo for node: %s", p_curr_node->name.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }
        if (!this->capability_module.IsSupportedGMPCapability(
            p_curr_node, EnGMPCapIsPortLLRStatisticsSupported)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s: DeviceID %u (0x%x) Does not support LLR, skipping \n",
                       p_curr_node->name.c_str(),
                       p_curr_node_info->DeviceID,
                       p_curr_node_info->DeviceID);

            p_curr_node->appData1.val |= NOT_SUPPORT_LLR_COUNTERS;
        }

        // go over all ports of this node:
        for (u_int32_t i = 1; i <= p_curr_node->numPorts; ++i) { //start from port 1, no need to reset port 0
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            //for switch read cap only on port 0, for others read on every port
            if (to_read_cap) {
                rc = ReadCapMask(p_curr_node, p_curr_port,
                                 curr_pm_cap_mask, port_info_cap_mask);
                if (rc) {
                    if (rc == IBDIAG_ERR_CODE_FABRIC_ERROR) {
                        break;     //not reasonable to continue, goto next_node
                    }
                    goto exit;
                }
                if (p_curr_node->type == IB_SW_NODE)
                    to_read_cap = false;
            }

            //send port counters mad set for this port
            clbck_data.m_data1 = p_curr_port;
            clbck_data.m_handle_data_func =
                &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortCountersClearClbck>;

            progress_bar.push(p_curr_port);
            int rc2 = this->ibis_obj.PMPortCountersClear(p_curr_port->base_lid,
                                                         p_curr_port->num,
                                                         &clbck_data);

            if (rc2)
                continue;  //we will try to proceed without query for extended counters

            //check if counters extended are supported
            if (!IS_SUPPORT_EXT_PORT_FAILED(p_curr_node->appData1.val)) {
                if ( IS_SUPPORT_EXTENDED_COUNTERS(curr_pm_cap_mask) ||
                     IS_SUPPORT_EXTENDED_COUNTERS_NO_IETF(curr_pm_cap_mask)) {

                    clbck_data.m_handle_data_func =
                        &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortCountersExtendedClearClbck>;

                    progress_bar.push(p_curr_port);
                    this->ibis_obj.PMPortCountersExtendedClear(p_curr_port->base_lid,
                        p_curr_port->num, &clbck_data);
                } else {
                    ports_errors.push_back(
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support extended port counters capability"));

                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                    //mark we already checked for support
                    p_curr_node->appData1.val |= NOT_SUPPORT_EXT_PORT_COUNTERS;
                }
            }

            //check if extended speed counters supported
            if(IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, p_curr_node->type) &&
               !IS_SUPPORT_EXT_SPEEDS_FAILED(p_curr_node->appData1.val)) {
                if(IS_SUPPORT_EXTENDED_SPEEDS_COUNTERS(port_info_cap_mask)) {
                    //if RS_FEC is active
                    if (isRSFEC(p_curr_port->get_fec_mode())) {
                        if (!IS_SUPPORT_EXT_SPEEDS_RSFEC_FAILED(
                            p_curr_node->appData1.val)) {
                            if (IS_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS(
                                curr_pm_cap_mask)) {
                                clbck_data.m_handle_data_func =
                                    &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortExtendedSpeedsRSFECClearClbck>;
                                progress_bar.push(p_curr_port);
                                this->ibis_obj.
                                    PMPortExtendedSpeedsRSFECCountersClear(
                                        p_curr_port->base_lid, p_curr_port->num,
                                        &clbck_data);
                            } else {
                                ports_errors.push_back(
                                        new FabricErrNodeNotSupportCap(
                                                p_curr_node, "This device does not "
                                                "support port extended speeds RSFEC "
                                                "counters capability"));
                                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                                //mark we already checked for support
                                p_curr_node->appData1.val |=
                                    NOT_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS;
                            }
                        }
                    } else { //no RS_FEC
                        clbck_data.m_handle_data_func =
                            &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortExtendedSpeedsClearClbck>;

                        progress_bar.push(p_curr_port);
                        this->ibis_obj.PMPortExtendedSpeedsCountersClear(
                                        p_curr_port->base_lid,
                                        p_curr_port->num, &clbck_data);
                    }
                } else {

                    ports_errors.push_back(
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support port extended speeds counters capability"));

                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                    //mark we already checked for support
                    p_curr_node->appData1.val |= NOT_SUPPORT_EXT_SPEEDS_COUNTERS;
                }

            } //ext_speeds_counters

            //always clear LLR counters and symbol error counter
            if(!IS_SUPPORT_LLR_FAILED(p_curr_node->appData1.val)) {
                clbck_data.m_data2 = (void*)IS_TO_PRINT_LLR_COUNTERS(check_counters_bitset);
                clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::VSPortLLRStatisticsClearClbck>;

                progress_bar.push(p_curr_port);
                this->ibis_obj.VSPortLLRStatisticsClear(
                        p_curr_port->base_lid,
                        p_curr_port->num,
                        true,
                        &clbck_data);
            }

            // PortRcvErrorDetails and PortXmitDiscardDetails

            if (!IS_SUPPORT_PORT_RCV_ERROR_DETAILS(p_curr_node->appData1.val)) {

                clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortRcvErrorDetailsClearClbck>;

               progress_bar.push(p_curr_port);
               this->ibis_obj.PMPortRcvErrorDetailsClear(p_curr_port->base_lid,
                                                         p_curr_port->num,
                                                         &clbck_data);
            }

            if (!IS_SUPPORT_PORT_XMIT_DISCARD_DETAILS(p_curr_node->appData1.val)) {

                clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::PMPortXmitDiscardDetailsClearClbck>;

                progress_bar.push(p_curr_port);
                this->ibis_obj.PMPortXmitDiscardDetailsClear(p_curr_port->base_lid,
                                                             p_curr_port->num,
                                                             &clbck_data);
            }

        } // port loop

        if (ibDiagClbck.GetState())
            goto exit;
    } //node loop
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("ResetPortCounters Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "ResetPortCounters Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!ports_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / IBDIAG_ERR_CODE_NO_MEM / ERR_CODE_INCORRECT_ARGS / IBDIAG_ERR_CODE_CHECK_FAILED
int IBDiag::CalcCounters(vector_p_pm_info_obj& prev_pm_info_obj_vec,
                         double diff_time_between_samples,
                         list_p_fabric_general_err& pm_errors) {
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    int rc2 = IBDIAG_SUCCESS_CODE;
    int llr_rcv_retry_index = -1;
    u_int64_t value1 = 0, value2 = 0;
    double rtr = 0.0;
    struct PM_PortCalcCounters curr_port_calc_counters;
    CLEAR_STRUCT(curr_port_calc_counters);


    rc = counter_name_2_index("port_rcv_retry", llr_rcv_retry_index);
    if ( rc )
        IBDIAG_RETURN(rc);

    for ( u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i ) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if ( !p_curr_port || !p_curr_port->getInSubFabric() )
            continue;
        if ( prev_pm_info_obj_vec.size() < i + 1 )    //info not exists
            continue;

        if ( !prev_pm_info_obj_vec[i] )               //info not exists
            continue;

        //llr_counters
        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics1 =
        prev_pm_info_obj_vec[i]->p_port_llr_statistics;
        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics2 =
        this->fabric_extended_info.getVSPortLLRStatistics(i);

        if (!p_curr_port_llr_statistics1 || !p_curr_port_llr_statistics2) {
            continue;
        }

        // check for overflow of the base counters
        value1 = p_curr_port_llr_statistics1->PortRcvRetry;
        value2 = p_curr_port_llr_statistics2->PortRcvRetry;

        // check for base counters overflow
        if ( (pm_counters_arr[llr_rcv_retry_index].overflow_value != OVERFLOW_VAL_NONE) &&
             (value1 == pm_counters_arr[llr_rcv_retry_index].overflow_value ||
              value2 == pm_counters_arr[llr_rcv_retry_index].overflow_value ||
              value1 > value2) ) {
            pm_errors.push_back(
                    new FabricErrPMBaseCalcCounterOverflow(p_curr_port, "retransmission_per_sec"));
            // don't store, i.e retransmission_per_sec is NA
            IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Calc rtr for (%s). value1=" U64H_FMT ", value2=" U64H_FMT ", "
                       "retransmission_per_sec=NA (base is overflown), diff_time=%f\n",
                       p_curr_port->getName().c_str(), value1, value2,
                       diff_time_between_samples);
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        } else {
            // calculate retransmission rate  and store it

            //sanity check
            if (diff_time_between_samples == 0) {
                IBDIAG_LOG(TT_LOG_LEVEL_INFO,
                           "Can't calculate the retransmission_per_sec counter, as the diff time is zero");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
            }
            rtr = (double)(value2 - value1)/ diff_time_between_samples;
            curr_port_calc_counters.RetransmissionPerSec = (u_int64_t)rtr;

            IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Calc rtr for (%s). value1=" U64H_FMT ", value2=" U64H_FMT ", "
                       "retransmission_per_sec=" U64H_FMT ", diff_time=%f\n",
                       p_curr_port->getName().c_str(), value1, value2,
                       curr_port_calc_counters.RetransmissionPerSec, diff_time_between_samples);

            rc2 = this->fabric_extended_info.addPMPortCalculatedCounters(p_curr_port, curr_port_calc_counters);
            if ( rc2 ) {
                this->SetLastError("Failed to add PM_PortCalcCounters for port=%s, err=%s",
                                   p_curr_port->getName().c_str(), this->fabric_extended_info.GetLastError());
                IBDIAG_RETURN(rc2);
            }
        }
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / IBDIAG_ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED / ERR_CODE_INCORRECT_ARGS
int IBDiag::CheckAllPMValues(list_p_fabric_general_err& pm_errors,
                             list_p_fabric_general_err &pm_threshold_overflow_warnings,
                             map_str_uint64& counters_to_threshold_map,
                             u_int32_t check_counters_bitset)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    u_int64_t all_errs_counters_threshold = 0;

    /*
     * Step 1 -
     *  Check if need to check all counters according to default threshold.
     *  If all is specified than all errors counters (not including extended)
     *  need to be checked.
     */
    bool need_to_check_all_errs_counters = (counters_to_threshold_map.find("all") != counters_to_threshold_map.end());
    if (need_to_check_all_errs_counters) {
        all_errs_counters_threshold = counters_to_threshold_map["all"];
    }

    for (int i = 0; i < PM_COUNTERS_ARR_SIZE; ++i) {
        //if not given specific threshold already
        if (counters_to_threshold_map.find(pm_counters_arr[i].name) == counters_to_threshold_map.end()) {
            // if not subject to keyword 'all' in -PC parameter, but has its own default threshold,
            // then use its own threshold even when 'all' is used
            if (!(pm_counters_arr[i].attribute & PM_COUNTER_IS_SUBJECT_TO_KEYWORD_ALL)) {
                if (pm_counters_arr[i].threshold != -1)
                    counters_to_threshold_map[pm_counters_arr[i].name] = pm_counters_arr[i].threshold;

            } else {
                // check if 'all' was given
                if (need_to_check_all_errs_counters)
                    counters_to_threshold_map[pm_counters_arr[i].name] = all_errs_counters_threshold;
            }
        }
    }

    /*
     * Step 2 -
     *  Building array of theshold by counters index.
     *  This will take O(n) but in the next steps we will be able to access
     *  the counters threshold by O(1).
     */
    u_int64_t *p_counters_thresh_by_idx[PM_COUNTERS_ARR_SIZE];
    for (int i = 0; i < PM_COUNTERS_ARR_SIZE; ++i)
        p_counters_thresh_by_idx[i] = NULL;
    for (map_str_uint64::iterator it = counters_to_threshold_map.begin();
            it != counters_to_threshold_map.end(); ++it) {
        if ((*it).first == "all")           //we already took care of that
            continue;
        for (int i = 0; i < PM_COUNTERS_ARR_SIZE; ++i) {
            if ((*it).first == pm_counters_arr[i].name) {
                p_counters_thresh_by_idx[i] = &((*it).second);
                break;
            }
            if (i == PM_COUNTERS_ARR_SIZE) {
                this->SetLastError("Unknown counter name %s was given as arg", it->first.c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
            }
        }
    }

    /*
     * Step 3 -
     *  Go over all ports
     */
    for (u_int32_t curr_port_idx = 0;
            curr_port_idx < this->fabric_extended_info.getPortsVectorSize();
            ++curr_port_idx) {
        /*
         * Step 3.1 -
         *  Get all needed data for checks
         */
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(curr_port_idx);
        //Skip ports that don't exist or are down or have undefined state
        //or are not in the scope of the fabric.
        if (!p_curr_port || !p_curr_port->is_data_worthy())
            continue;

        IBNode *p_curr_node = p_curr_port->p_node;
        if (!p_curr_node) {
            this->SetLastError("DB error - found port with null node = %s", p_curr_port->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        struct IB_ClassPortInfo *p_class_port_info =
                this->fabric_extended_info.getPMClassPortInfo(p_curr_node->createIndex);

        if (!p_class_port_info)           //already reported to user no reason to continue here
            continue;

        // Extended Speeds Capability
        //bool* tmp_ext_speeds_cap;
        //bool ext_speeds_cap;
        //tmp_ext_speeds_cap = this->fabric_extended_info.getPMPortExtSpeedsCap(p_curr_node->createIndex);
        //ext_speeds_cap = tmp_ext_speeds_cap && *tmp_ext_speeds_cap;

        struct PM_PortCounters *p_curr_port_counters =
                this->fabric_extended_info.getPMPortCounters(curr_port_idx);
        struct PM_PortCountersExtended *p_curr_port_counters_extended =
                this->fabric_extended_info.getPMPortCountersExtended(curr_port_idx);
        struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_cnts =
                this->fabric_extended_info.getPMPortExtSpeedsCounters(curr_port_idx);
        struct PM_PortExtendedSpeedsRSFECCounters
            *p_curr_port_ext_speeds_rsfec_cnts =
            this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(
                                        curr_port_idx);
        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics =
                this->fabric_extended_info.getVSPortLLRStatistics(curr_port_idx);
        struct PM_PortCalcCounters *p_curr_port_calc_counters =
            this->fabric_extended_info.getPMPortCalcCounters(curr_port_idx);

        /*
         * Step 3.2 -
         *  Go over all counters
         */
        u_int64_t value;
        u_int64_t thresh;
        list_p_pm_err curr_port_errors;
        list_p_pm_err curr_port_overflow_threshold_warning;
        for (int curr_counter_idx = 0; curr_counter_idx < PM_COUNTERS_ARR_SIZE; ++curr_counter_idx) {
            // Skip untested counters
            if( pm_counters_arr[curr_counter_idx].counter_src == PM_PORT_EXT_SPEEDS_CNTS_MAD &&
                (!IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, p_curr_node->type)))
                continue;

            // Skip LLR counters - we don't check them, only report the value
            if(pm_counters_arr[curr_counter_idx].counter_src ==
               VS_PORT_LLR_CNTS_MAD_HIDDEN)
                continue;

            /*
             * Step 3.2.1 -
             *  skip unsupported counters
             */
            if (((pm_counters_arr[curr_counter_idx].counter_src == PM_PORT_CNT_EXT_MAD ||
                    pm_counters_arr[curr_counter_idx].counter_src == PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD) &&
                    !is_counter_supported(p_class_port_info->CapMsk, pm_counters_arr[curr_counter_idx].cap_mask_bits)) ||
                    (pm_counters_arr[curr_counter_idx].counter_src == PM_PORT_EXT_SPEEDS_CNTS_MAD &&
                            !p_curr_port_ext_speeds_cnts)) {
                continue;           //go to next counter
            }

            if ((pm_counters_arr[curr_counter_idx].counter_src == PM_PORT_CNT_EXT_MAD) &&
                    (pm_counters_arr[curr_counter_idx].attribute & PM_COUNTER_IS_SUBJECT_TO_CAP_MASK_2) &&
                    !IS_SUPPORT_ADDITIONAL_EXTENDED_COUNTERS(p_class_port_info->CapMsk2))
                continue;

            /*
             * Step 3.2.2 -
             *  Get counter value - regular / extended
             */
            value = 0;
            switch (pm_counters_arr[curr_counter_idx].counter_src) {
                case PM_PORT_CNT_MAD:
                {
                    if (!p_curr_port_counters)          //info not exist
                        continue;
                    rc = get_value((u_int8_t *)p_curr_port_counters + pm_counters_arr[curr_counter_idx].struct_offset,
                            pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                case PM_PORT_CNT_EXT_MAD:
                {
                    if (!p_curr_port_counters_extended)
                        continue;                       //info not exist
                    rc = get_value((u_int8_t *)p_curr_port_counters_extended + pm_counters_arr[curr_counter_idx].struct_offset,
                                   pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                case PM_PORT_EXT_SPEEDS_CNTS_MAD:
                {
                    if (!p_curr_port_ext_speeds_cnts)
                        continue;
                    rc = get_value((u_int8_t*)p_curr_port_ext_speeds_cnts + pm_counters_arr[curr_counter_idx].struct_offset,
                            pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                case PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD:
                {
                    if (!p_curr_port_ext_speeds_rsfec_cnts)
                        continue;
                    rc = get_value((u_int8_t*)p_curr_port_ext_speeds_rsfec_cnts + pm_counters_arr[curr_counter_idx].struct_offset,
                            pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                case VS_PORT_LLR_CNTS_MAD:
                {
                    if (!p_curr_port_llr_statistics)
                        continue;
                    rc = get_value((u_int8_t*)p_curr_port_llr_statistics + pm_counters_arr[curr_counter_idx].struct_offset,
                            pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                case PM_PORT_CALCULATED_COUNTER:
                {
                    if (!p_curr_port_calc_counters) {
                        continue;
                    }
                    rc = get_value((u_int8_t*)p_curr_port_calc_counters + pm_counters_arr[curr_counter_idx].struct_offset,
                            pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                default:
                    continue;
            }

            if (rc)
                curr_port_errors.push_back(
                        new FabricErrPMCounterInvalidSize(p_curr_port,
                                pm_counters_arr[curr_counter_idx].name,
                                pm_counters_arr[curr_counter_idx].real_size));

            /*
             * Step 3.2.3 -
             *  Check if overflowed (except for calculated counters) or exceeds threshold value
             */
            if (pm_counters_arr[curr_counter_idx].counter_src != PM_PORT_CALCULATED_COUNTER  &&
                    (pm_counters_arr[curr_counter_idx].overflow_value != OVERFLOW_VAL_NONE)  &&
                    (value == pm_counters_arr[curr_counter_idx].overflow_value)              &&
                    is_report_as_error(pm_counters_arr[curr_counter_idx], p_curr_port)
                    ) {
                FabricErrPMCounterOverflow *p_curr_fabric_pm_err =
                        new FabricErrPMCounterOverflow(p_curr_port, pm_counters_arr[curr_counter_idx].name,
                                pm_counters_arr[curr_counter_idx].overflow_value);

                p_curr_fabric_pm_err->SetLevel(EN_FABRIC_ERR_WARNING);
                curr_port_overflow_threshold_warning.push_back(p_curr_fabric_pm_err);

            } else if ((p_counters_thresh_by_idx[curr_counter_idx]) &&
                        is_report_as_error(pm_counters_arr[curr_counter_idx], p_curr_port)) {

                thresh = *(p_counters_thresh_by_idx[curr_counter_idx]);

                if (value > thresh)
                    curr_port_errors.push_back(
                            new FabricErrPMCounterExceedThreshold(
                                    p_curr_port,
                                    pm_counters_arr[curr_counter_idx].name,
                                    thresh, value));
            }
        }   //counters loop

        /*
         * Step 3.3 -
         *  Merge all counters errors
         */
        if (!curr_port_errors.empty()) {
            pm_errors.push_back(new FabricErrPMCountersAll(p_curr_port, curr_port_errors));
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;

            release_container_data(curr_port_errors);
        }

        if (!curr_port_overflow_threshold_warning.empty()) {
            FabricErrPMCountersAll *p_curr_fabric_pm_err =
                    new FabricErrPMCountersAll(p_curr_port, curr_port_overflow_threshold_warning);

            p_curr_fabric_pm_err->SetLevel(EN_FABRIC_ERR_WARNING);
            pm_threshold_overflow_warnings.push_back(p_curr_fabric_pm_err);
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;

            release_container_data(curr_port_overflow_threshold_warning);
        }
    }   //ports loop

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckCountersDiff(vector_p_pm_info_obj& prev_pm_info_obj_vec,
        list_p_fabric_general_err& pm_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    u_int64_t value1, value2, diff_value;
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;
        if (prev_pm_info_obj_vec.size() < i + 1)    //info not exists
            continue;

        if (!prev_pm_info_obj_vec[i])               //info not exists
            continue;

        //regular counters
        struct PM_PortCounters *p_curr_port_counters1 = prev_pm_info_obj_vec[i]->p_port_counters;
        if (!p_curr_port_counters1)                 //info not exists
            continue;
        struct PM_PortCounters *p_curr_port_counters2 = this->fabric_extended_info.getPMPortCounters(i);
        if (!p_curr_port_counters2)                 //info not exists
            continue;

        //extended_port_counters
        struct PM_PortCountersExtended *p_curr_extended_port_counters1 =
            prev_pm_info_obj_vec[i]->p_extended_port_counters;
        struct PM_PortCountersExtended *p_curr_extended_port_counters2 =
            this->fabric_extended_info.getPMPortCountersExtended(i);

        //ext_speeds_counters
        struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_counters1 =
            prev_pm_info_obj_vec[i]->p_port_ext_speeds_counters;
        struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_counters2 =
            this->fabric_extended_info.getPMPortExtSpeedsCounters(i);

        //ext_speeds_counters for rsfec
        struct PM_PortExtendedSpeedsRSFECCounters
            *p_curr_port_ext_speeds_rsfec_counters1 =
            prev_pm_info_obj_vec[i]->p_port_ext_speeds_rsfec_counters;
        struct PM_PortExtendedSpeedsRSFECCounters
            *p_curr_port_ext_speeds_rsfec_counters2 =
            this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(i);

        //for max retransmission rate
        // For all counters except max retransmission rate,
        // counter_src = VS_PORT_LLR_CNTS_MAD_HIDDEN
        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics1 =
            prev_pm_info_obj_vec[i]->p_port_llr_statistics;
        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics2 =
            this->fabric_extended_info.getVSPortLLRStatistics(i);


        u_int8_t * p_curr_counters1 = NULL;
        u_int8_t * p_curr_counters2 = NULL;

        for (int j = 0; j < PM_COUNTERS_ARR_SIZE; ++j) {
            if (!pm_counters_arr[j].diff_check_threshold)   //checking only error counters
                continue;

            switch(pm_counters_arr[j].counter_src) {
                case PM_PORT_CNT_MAD:
                    p_curr_counters1 = (u_int8_t *)p_curr_port_counters1;
                    p_curr_counters2 = (u_int8_t *)p_curr_port_counters2;
                    break;
                case PM_PORT_CNT_EXT_MAD:
                    p_curr_counters1 = (u_int8_t *)p_curr_extended_port_counters1;
                    p_curr_counters2 = (u_int8_t *)p_curr_extended_port_counters2;
                    break;
                case PM_PORT_EXT_SPEEDS_CNTS_MAD:
                    p_curr_counters1 = (u_int8_t *)p_curr_port_ext_speeds_counters1;
                    p_curr_counters2 = (u_int8_t *)p_curr_port_ext_speeds_counters2;
                    break;
                case PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD:
                    p_curr_counters1 =
                        (u_int8_t *)p_curr_port_ext_speeds_rsfec_counters1;
                    p_curr_counters2 =
                        (u_int8_t *)p_curr_port_ext_speeds_rsfec_counters2;
                    break;
                case VS_PORT_LLR_CNTS_MAD:
                    p_curr_counters1 =
                        (u_int8_t *)p_curr_port_llr_statistics1;
                    p_curr_counters2 =
                        (u_int8_t *)p_curr_port_llr_statistics2;
                    break;

                default:
                    continue;
            }

            if (!p_curr_counters1 || !p_curr_counters2)                 //info not exists
                continue;

            value1 = value2 = 0;
            rc  = get_value(p_curr_counters1 + pm_counters_arr[j].struct_offset,
                    pm_counters_arr[j].real_size, value1);
            rc |= get_value(p_curr_counters2 + pm_counters_arr[j].struct_offset,
                    pm_counters_arr[j].real_size, value2);
            if (rc) {
                this->SetLastError("Invalid pm counter size: %s. Counter real_size is %d",
                                   pm_counters_arr[j].name.c_str(),
                                   pm_counters_arr[j].real_size);
                IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
            }

            diff_value = value2 - value1;

            bool report_as_warning = false;

            // Custom validation for PortRcvSwitchRelayErrors
            if (pm_counters_arr[j].name == "port_rcv_switch_relay_errors")
                if (p_curr_extended_port_counters1 && p_curr_extended_port_counters2) {
                    // calc diff of PortMultiCastRcvPkts
                    uint64_t prev = p_curr_extended_port_counters1->PortMultiCastRcvPkts;
                    uint64_t last = p_curr_extended_port_counters2->PortMultiCastRcvPkts;
                    uint64_t diff = prev <= last ?  last - prev : prev - last;

                    report_as_warning = diff_value <= diff;
                }

            if (diff_value >= pm_counters_arr[j].diff_check_threshold) {
                pm_errors.push_back(
                        new FabricErrPMErrCounterIncreased(
                                p_curr_port,
                                pm_counters_arr[j].name,
                                pm_counters_arr[j].diff_check_threshold,
                                diff_value,
                                report_as_warning));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        }
    }
    IBDIAG_RETURN(rc);
}

/*
 * time - in seconds units
 * symbol errors - delta of symbol error counter in this time slot
 * ber - to return calculated ber
 */
int IBDiag::CalcBER(IBPort *p_curr_port,
                    double time,
                    u_int64_t symbol_error,
                    long double &reciprocal_ber)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    long double ber = 0;
    IBLinkSpeed port_speed = p_curr_port->get_common_speed();

    if (!symbol_error) {
        reciprocal_ber = 0.0;
        IBDIAG_RETURN(rc);
    }

    //calculate BER
    /* BER = errors / data_bits ==>
     * diff_errors / delta(t) * link rate
     */
    long double link_rate =
        (long double)(CalcLinkRate(p_curr_port->get_common_width(),
                                   port_speed));

    reciprocal_ber =
        (time * link_rate) / symbol_error;

    SMP_MlnxExtPortInfo *p_curr_mlnx_ext_port_info =
    this->fabric_extended_info.getSMPMlnxExtPortInfo(p_curr_port->createIndex);

    if (p_curr_mlnx_ext_port_info &&
            (this->GetLLRActiveCellSize() ||
            (p_curr_mlnx_ext_port_info->RetransMode ==
             RETRANS_LLR_ACTIVE_CELL_64 ||
             p_curr_mlnx_ext_port_info->RetransMode ==
             RETRANS_LLR_ACTIVE_CELL_128))) {
        // if reciprocal_ber=0 need to avoid dividing by zero
        if (!reciprocal_ber) {
            reciprocal_ber = 0.0;
            IBDIAG_RETURN(rc);
        }

        unsigned int cell_size_bits = llr_active_cell2size(
                p_curr_mlnx_ext_port_info->RetransMode) * 8;

        //check that we have a valid active cell size
        if (!cell_size_bits)
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        // for llr, ber = symbol error/time
        ber = 1/reciprocal_ber;

        long double p2_errors = (1.0 -
                                 (powl((1.0 - ber), cell_size_bits) +
                                  cell_size_bits *
                                  ber *
                                  powl((1.0 - ber), (cell_size_bits - 1))));
        long double effective_ber = p2_errors * powl(2, -16.0);
        reciprocal_ber = 1/effective_ber;
    }

    IBDIAG_RETURN(rc);
}


//The formulas here are as in Raw BER in phy diag
//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::CalcPhyTest(vector_p_pm_info_obj& prev_pm_info_obj_vec,
                        double sec_between_samples,
                        CSVOut &csv_out)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    long double ber = 0.0;
    long double k_rsfec = 0.0;
    char buffer[256];
    IBFECMode port_fec = IB_FEC_NA;
    stringstream sstream;

    if (csv_out.DumpStart(SECTION_PHY_TEST))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    sstream << "NodeGUID,PortGUID,PortNumber,Value" << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize();
            ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;
        if (prev_pm_info_obj_vec.size() < i + 1)    //info not exists
            break;

        if (!prev_pm_info_obj_vec[i])               //info not exists
            continue;

        struct PM_PortCounters *p_curr_port_counters1 =
        prev_pm_info_obj_vec[i]->p_port_counters;
        if (!p_curr_port_counters1) {                 //info not exists
            rc = IBDIAG_ERR_CODE_DB_ERR;
            continue;
        }

        struct PM_PortCounters *p_curr_port_counters2 =
            this->fabric_extended_info.getPMPortCounters(i);
        if (!p_curr_port_counters2)        //info not exists
            continue;

        port_fec = p_curr_port->get_fec_mode();
        ber = 0.0;

        if (isRSFEC(port_fec)) {
            struct PM_PortExtendedSpeedsRSFECCounters *p_ext_speeds_rsfec_cnts1 =
                prev_pm_info_obj_vec[i]->p_port_ext_speeds_rsfec_counters;
            if (!p_ext_speeds_rsfec_cnts1)    //info not exists
                continue;

            struct PM_PortExtendedSpeedsRSFECCounters *p_ext_speeds_rsfec_cnts2 =
                this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(i);
            if (!p_ext_speeds_rsfec_cnts2)    //info not exists
                continue;
            SMP_MlnxExtPortInfo *p_curr_mlnx_ext_port_info =
                this->fabric_extended_info.getSMPMlnxExtPortInfo(p_curr_port->createIndex);
            if (!p_curr_mlnx_ext_port_info)
                continue;

            k_rsfec = p_curr_mlnx_ext_port_info->ActiveRSFECParity / 2 + 1;
            //calculate k_rsfec
            ber = (p_ext_speeds_rsfec_cnts2->PortFECCorrectedSymbolCounter -
                   p_ext_speeds_rsfec_cnts1->PortFECCorrectedSymbolCounter) +
                  k_rsfec *
                  (p_ext_speeds_rsfec_cnts2->PortFECUncorrectableBlockCounter -
                   p_ext_speeds_rsfec_cnts1->PortFECUncorrectableBlockCounter);
        } else {
            struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_counters1 =
                prev_pm_info_obj_vec[i]->p_port_ext_speeds_counters;
            if (!p_curr_port_ext_speeds_counters1)   //info not exists
                continue;

            struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_counters2 =
                this->fabric_extended_info.getPMPortExtSpeedsCounters(i);
            if (!p_curr_port_ext_speeds_counters2)   //info not exists
                continue;
            //undefined fec port
            if (port_fec == IB_FEC_NA)
                ber = max(CalcBERFireCodeFEC(p_curr_port_ext_speeds_counters1,
                                             p_curr_port_ext_speeds_counters2),
                          EDPLSum(p_curr_port_ext_speeds_counters1,
                                  p_curr_port_ext_speeds_counters2));
            else if (port_fec == IB_FEC_NO_FEC) {
                //up to QDR (8/10)
                if (p_curr_port->get_common_speed() < IB_LINK_SPEED_14)
                    //symbol error/time
                    ber = p_curr_port_counters2->SymbolErrorCounter -
                          p_curr_port_counters1->SymbolErrorCounter;
                else //64/66
                    //sum of edpls over lanes/time
                    ber = EDPLSum(p_curr_port_ext_speeds_counters1,
                                  p_curr_port_ext_speeds_counters2);
            } else if (port_fec == IB_FEC_FIRECODE_FEC)
                ber = CalcBERFireCodeFEC(p_curr_port_ext_speeds_counters1,
                                         p_curr_port_ext_speeds_counters2);
        }
        long double link_rate =
            (long double)(CalcLinkRate(p_curr_port->get_common_width(),
                        p_curr_port->get_common_speed()));
        long double data_rate =
            sec_between_samples * link_rate;

        if (data_rate)
            ber /= data_rate;
        else {
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto Exit;
        }

        sstream.str("");
        snprintf(buffer, sizeof(buffer),
                 U64H_FMT "," U64H_FMT "," U32D_FMT ","   \
                 "%Le",
                 p_curr_port->p_node->guid_get(),
                 p_curr_port->guid_get(),
                 p_curr_port->num,
                 ber);
        sstream << buffer << endl;
        csv_out.WriteBuf(sstream.str());
    }

Exit:
    csv_out.DumpEnd(SECTION_PHY_TEST);
    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CalcBERErrors(vector_p_pm_info_obj& prev_pm_info_obj_vec,
                          u_int64_t ber_threshold_reciprocal_val,
                          double sec_between_samples,
                          list_p_fabric_general_err& ber_errors,
                          CSVOut &csv_out)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    long double ber_value_reciprocal_val = 0.0;
    char buffer[256];
    stringstream sstream;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "BER threshold=" U64D_FMT "\n",
               ber_threshold_reciprocal_val);

    // cannot be skipped because generate warnings
    (void) csv_out.DumpStart(SECTION_BER_TEST);

    sstream << "NodeGUID,PortGUID,PortNumber,Value" << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize();
         ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;
        if (prev_pm_info_obj_vec.size() < i + 1)    //info not exists
            continue;

        if (!prev_pm_info_obj_vec[i])               //info not exists
            continue;

        struct PM_PortCounters *p_curr_port_counters1 =
        prev_pm_info_obj_vec[i]->p_port_counters;
        if (!p_curr_port_counters1) {                 //info not exists
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto Exit;
        }
        struct PM_PortCounters *p_curr_port_counters2 =
        this->fabric_extended_info.getPMPortCounters(i);
        if (!p_curr_port_counters2) {                 //info not exists
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto Exit;
        }

        //collect all errors values
        u_int64_t diff_errs_value = 0;
        diff_errs_value = p_curr_port_counters2->SymbolErrorCounter -
            p_curr_port_counters1->SymbolErrorCounter;

        rc = this->CalcBER(p_curr_port,
                           sec_between_samples,
                           diff_errs_value,
                           ber_value_reciprocal_val);
        sstream.str("");
        snprintf(buffer, sizeof(buffer),
                 U64H_FMT "," U64H_FMT "," U32D_FMT ","   \
                 "%Le",
                 p_curr_port->p_node->guid_get(),
                 p_curr_port->guid_get(),
                 p_curr_port->num,
                 ber_value_reciprocal_val == 0 ?
                    0.0 : 1.0/ber_value_reciprocal_val);
        sstream << buffer << endl;
        csv_out.WriteBuf(sstream.str());

        //info doesn't exist
        if (rc == IBDIAG_ERR_CODE_DB_ERR) {
            rc = IBDIAG_SUCCESS_CODE;
            continue;
        }

        if (rc == IBDIAG_SUCCESS_CODE && ber_value_reciprocal_val == 0) {
            if (ber_threshold_reciprocal_val == OVERFLOW_VAL_64_BIT) {    //need to report anyway
                ber_errors.push_back(new FabricErrBERIsZero(p_curr_port));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
            continue;
        }

        if ( (ber_value_reciprocal_val < ber_threshold_reciprocal_val) ||
                (ber_threshold_reciprocal_val == OVERFLOW_VAL_64_BIT) ) {

            ber_errors.push_back(
                    new FabricErrBERExceedThreshold(
                            p_curr_port,
                            ber_threshold_reciprocal_val,
                            ber_value_reciprocal_val));

            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }

Exit:
    csv_out.DumpEnd(SECTION_BER_TEST);
    IBDIAG_RETURN(rc);
}


list_string IBDiag::GetListOFPMNames()
{
    IBDIAG_ENTER;
    list_string names_list;
    for (int i = 0; i < PM_COUNTERS_ARR_SIZE; ++i)
        names_list.push_back(pm_counters_arr[i].name);
    names_list.push_back("all");
    IBDIAG_RETURN(names_list);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WritePMFile(const string &file_name,
                        u_int32_t check_counters_bitset,
                        bool en_per_lane_cnts)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Ports Counters",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    IBFabric::GetSwitchLabelPortNumExplanation(sout);
    this->DumpAllPortsCounters(sout, check_counters_bitset, en_per_lane_cnts);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::WritePMAggregatedFile(const string &file_name,
                                  u_int32_t check_counters_bitset,
                                  bool en_per_lane_cnts)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Aggregated Ports Counters",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->DumpAllAPortsCounters(sout, check_counters_bitset, en_per_lane_cnts);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpPerSLVLPortCntrsCSVTable(CSVOut &csv_out, vec_slvl_cntrs  &slvl_cntrs_vec)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    for (CountersPerSLVL* p_cntrs_per_slvl : slvl_cntrs_vec) {

        if (csv_out.DumpStart(p_cntrs_per_slvl->GetCSVSectionHeader().c_str()))
            continue;

        p_cntrs_per_slvl->DumpSLVLCntrsHeader(csv_out);
        p_cntrs_per_slvl->DumpSLVLCntrsData(csv_out, this->fabric_extended_info);

        csv_out.DumpEnd(p_cntrs_per_slvl->GetCSVSectionHeader().c_str());

    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


void CountersPerSLVL::DumpSLVLCntrsHeader(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    stringstream sstream;
    sstream << "PortName, LID, GUID";
    for (u_int32_t cnt = 0; cnt < this->m_num_fields; cnt++)
        sstream << "," << this->m_header << this->m_cntr_size << "[" <<  cnt << "]";

    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    IBDIAG_RETURN_VOID;
}

void CountersPerSLVL::DumpSLVLCntrsData(CSVOut &csv_out, IBDMExtendedInfo &fabric_extended_info)
{
    IBDIAG_ENTER;
    char buffer[1024];
    u_int8_t operationalVLs = 0;
    stringstream sstream;

    for (set_port_data_update_t::iterator it = m_set_port_data_update.begin();
            it != m_set_port_data_update.end(); it++) {

        IBPort * p_port = (*it).first;
        const struct PM_PortRcvXmitCntrsSlVl * p_per_slvl_ctrs = &((*it).second);
        sstream.str("");
        snprintf(buffer, sizeof(buffer), "%s,%d," U64H_FMT,
                      p_port->getName().c_str(),
                      p_port->base_lid,
                      p_port->guid_get());
        sstream << buffer;

        SMP_PortInfo *curr_port_info = fabric_extended_info.getSMPPortInfo(p_port->createIndex);
        if (!curr_port_info) {
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "DB error - found connected port=%s"
                       " without SMPPortInfo",
                       p_port->getName().c_str());
            IBDIAG_RETURN_VOID;
        }

        operationalVLs = get_operational_vl_num(curr_port_info->OpVLs);

        union IBDiagSLVLCntrsData ibdiag_slvl_cntrs_data;
        CLEAR_STRUCT(ibdiag_slvl_cntrs_data);

        this->Unpack(ibdiag_slvl_cntrs_data, p_per_slvl_ctrs->Data.RawData_8x132_array.data);

        if (!this->m_is_ext_cntrs)
            this->Dump(ibdiag_slvl_cntrs_data.data32, this->m_num_fields, operationalVLs, sstream);
        else
            this->Dump(ibdiag_slvl_cntrs_data.data64, this->m_num_fields, operationalVLs, sstream);

        csv_out.WriteBuf(sstream.str());
    }

    IBDIAG_RETURN_VOID;
}

void CountersPerSLVL::Dump(u_int32_t data[], size_t arrsize, u_int8_t operationalVLs, stringstream &sstream)
{
    IBDIAG_ENTER;
    for(u_int32_t i = 0; i < arrsize; i++)
        if (this->m_is_per_vl && i >= operationalVLs)
            sstream << ",NA";
        else
            sstream << "," << data[i];

    sstream << endl;
    IBDIAG_RETURN_VOID;
}

void CountersPerSLVL::Dump(struct uint64bit data[], size_t arrsize, u_int8_t operationalVLs, stringstream &sstream)
{
    IBDIAG_ENTER;
    for(u_int32_t i = 0; i < arrsize; i++)
        if (this->m_is_per_vl && i >= operationalVLs)
            sstream << ",NA";
        else {
            u_int64_t cntr64;
            cntr64 = data[i].High;
            cntr64 = cntr64 << 32;
            cntr64 |= 0xffffffff & data[i].Low;
            sstream << "," << cntr64;
        }

    sstream << endl;
    IBDIAG_RETURN_VOID;
}


// Fast Recovery
int IBDiag::BuildCreditWatchdogTimeoutCounters(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::VSCreditWatchdogTimeoutCountersGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct VS_CreditWatchdogTimeoutCounters wd_counters;
    CLEAR_STRUCT(wd_counters);

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBPort* p_port_zero = p_curr_node->getPort(0);
        if (!p_port_zero)
            continue;

        if (!p_curr_node->getInSubFabric())
            continue;

        if(!this->GetCapabilityModulePtr()->IsSupportedGMPCapability(
                p_curr_node, EnGMPCAPIsCreditWatchdogTimeoutCountersSupported))
            continue;

        for (u_int8_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {

            // send to all connected ports
            IBPort* p_port = p_curr_node->getPort(port_num);
            if (!p_port ||
                    !p_port->is_data_worthy() ||
                    p_port->isSpecialPort() ||
                    !p_port->p_remotePort)
                continue;

            clbck_data.m_data1 = p_port;
            progress_bar.push(p_port);

            this->ibis_obj.VSCreditWatchdogTimeoutCountersGet(p_port_zero->base_lid,
                                                              port_num,
                                                              &wd_counters,
                                                              &clbck_data);

            if (ibDiagClbck.GetState())
                break;
        } // ports
    } // switches

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::ClearCreditWatchdogTimeoutCounters(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::VSCreditWatchdogTimeoutCountersClearClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct VS_CreditWatchdogTimeoutCounters wd_counters;
    CLEAR_STRUCT(wd_counters);

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBPort* p_port_zero = p_curr_node->getPort(0);
        if (!p_port_zero)
            continue;

        if (!p_curr_node->getInSubFabric())
            continue;

        if(!this->GetCapabilityModulePtr()->IsSupportedGMPCapability(
                p_curr_node, EnGMPCAPIsCreditWatchdogTimeoutCountersSupported))
            continue;

        for (u_int8_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {

            // send to all connected ports
            IBPort* p_port = p_curr_node->getPort(port_num);
            if (!p_port ||
                    !p_port->is_data_worthy() ||
                    p_port->isSpecialPort() ||
                    !p_port->p_remotePort)
                continue;

            clbck_data.m_data1 = p_port;
            progress_bar.push(p_port);

            this->ibis_obj.VSCreditWatchdogTimeoutCountersClear(p_port_zero->base_lid,
                                                                port_num,
                                                                &wd_counters,
                                                                &clbck_data);

            if (ibDiagClbck.GetState())
                break;
        } // ports
    } // switches

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpCreditWatchdogTimeoutToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_CREDIT_WATCHDOG_TIMEOUT_COUNTERS))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    const int num_credit_wd_timeout_per_vl = 8;

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," "PortNumber";

    for (int i = 0; i < num_credit_wd_timeout_per_vl; ++i)
        sstream << ",credit_watchdog_timeout_per_vl_" << i;

    sstream << ",total_port_credit_watchdog_timeout" << endl;

    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {

        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        VS_CreditWatchdogTimeoutCounters* p_credit_wd =
            this->fabric_extended_info.getCreditWatchdogTimeoutCounters(i);
        if (!p_credit_wd)       //info not exist
            continue;

        sstream.str("");

        sstream << PTR(p_curr_port->p_node->guid_get())    << ","
                << PTR(p_curr_port->guid_get())            << ","
                << +p_curr_port->num                       << ",";

        for (int i = 0; i < num_credit_wd_timeout_per_vl; ++i)
            sstream << p_credit_wd->credit_watchdog_timeout_per_vl[i] << ",";

        sstream << p_credit_wd->total_port_credit_watchdog_timeout << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_CREDIT_WATCHDOG_TIMEOUT_COUNTERS);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::BuildFastRecoveryCounters(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::VSFastRecoveryCountersGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct VS_FastRecoveryCounters fast_recovery_counters;
    CLEAR_STRUCT(fast_recovery_counters);

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBPort* p_port_zero = p_curr_node->getPort(0);
        if (!p_port_zero)
            continue;

        if (!p_curr_node->getInSubFabric())
            continue;

        if(!this->GetCapabilityModulePtr()->IsSupportedGMPCapability(
                p_curr_node, EnGMPCAPIsFastRecoveryCountersSupported))
            continue;

        for (u_int8_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {

            // send to all connected ports
            IBPort* p_port = p_curr_node->getPort(port_num);
            if (!p_port ||
                    !p_port->is_data_worthy() ||
                    p_port->isSpecialPort() ||
                    !p_port->p_remotePort)
                continue;

            clbck_data.m_data1 = p_port;
            progress_bar.push(p_port);

            for (u_int8_t trigger = FR_TRIGGER_FIRST; trigger < FR_TRIGGER_LAST; ++trigger) {

                if (trigger == FR_TRIGGER_RESERVED)
                    continue;

                clbck_data.m_data2 = (void *)(uintptr_t)trigger;

                this->ibis_obj.VSFastRecoveryCountersGet(p_port_zero->base_lid,
                                                         port_num,
                                                         trigger,
                                                         &fast_recovery_counters,
                                                         &clbck_data);

                if (ibDiagClbck.GetState())
                    break;

            } // triggers
        } // ports
    } // switches

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::ClearFastRecoveryCounters(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::VSFastRecoveryCountersClearClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct VS_FastRecoveryCounters fast_recovery_counters;
    CLEAR_STRUCT(fast_recovery_counters);

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBPort* p_port_zero = p_curr_node->getPort(0);
        if (!p_port_zero)
            continue;

        if (!p_curr_node->getInSubFabric())
            continue;

        if(!this->GetCapabilityModulePtr()->IsSupportedGMPCapability(
                p_curr_node, EnGMPCAPIsFastRecoveryCountersSupported))
            continue;

        for (u_int8_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {

            // send to all connected ports
            IBPort* p_port = p_curr_node->getPort(port_num);
            if (!p_port ||
                    !p_port->is_data_worthy() ||
                    p_port->isSpecialPort() ||
                    !p_port->p_remotePort)
                continue;

            clbck_data.m_data1 = p_port;
            progress_bar.push(p_port);

            for (u_int8_t trigger = FR_TRIGGER_FIRST; trigger < FR_TRIGGER_LAST; ++trigger) {

                if (trigger == FR_TRIGGER_RESERVED)
                    continue;

                this->ibis_obj.VSFastRecoveryCountersClear(p_port_zero->base_lid,
                                                           port_num,
                                                           trigger,
                                                           &fast_recovery_counters,
                                                           &clbck_data);

                if (ibDiagClbck.GetState())
                    break;

            } // triggers
        } // ports
    } // switches

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpFastRecoveryCountersToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_FAST_RECOVERY_COUNTERS))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNumber,"
            << "trigger,counter_overflow,"
            << "num_errors,num_warnings,num_normals,"
            << "min_value,max_value,consecutive_normal,"
            << "last_value_0,last_value_1,last_value_2"
            << endl;

    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {

        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        for (u_int8_t trigger = FR_TRIGGER_FIRST; trigger < FR_TRIGGER_LAST; ++trigger) {

            if (trigger == FR_TRIGGER_RESERVED)
                continue;

            VS_FastRecoveryCounters* p_fast_recovery_cntrs =
                this->fabric_extended_info.getFastRecoveryCounters(i, trigger);
            if (!p_fast_recovery_cntrs)       //info not exist
                continue;

            sstream.str("");

            sstream << PTR(p_curr_port->p_node->guid_get())          << ","
                    << PTR(p_curr_port->guid_get())                  << ","
                    << +p_curr_port->num                             << ","
                    << +p_fast_recovery_cntrs->trigger               << ","
                    << PTR(p_fast_recovery_cntrs->counter_overflow)  << ","
                    << p_fast_recovery_cntrs->num_errors             << ","
                    << p_fast_recovery_cntrs->num_warnings           << ","
                    << p_fast_recovery_cntrs->num_normals            << ","
                    << p_fast_recovery_cntrs->min_value              << ","
                    << p_fast_recovery_cntrs->max_value              << ","
                    << p_fast_recovery_cntrs->consecutive_normal     << ","
                    << p_fast_recovery_cntrs->last_value_0           << ","
                    << p_fast_recovery_cntrs->last_value_1           << ","
                    << p_fast_recovery_cntrs->last_value_2           << endl;

            csv_out.WriteBuf(sstream.str());
        } // triggers
    } // ports

    csv_out.DumpEnd(SECTION_FAST_RECOVERY_COUNTERS);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


int IBDiag::SendVSPortGeneralCounters(list_p_fabric_general_err& retrieve_errors, bool clear)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::VSPortGeneralCountersClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct VS_PortGeneralCounters vs_port_general_counters;
    CLEAR_STRUCT(vs_port_general_counters);

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        IBPort* p_port_zero = p_curr_node->getPort(0);
        if (!p_port_zero)
            continue;

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!this->GetCapabilityModulePtr()->IsSupportedGMPCapability(
                p_curr_node, EnGMPCAPIsPortGeneralCountersSupported))
            continue;

         for (phys_port_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort(i);
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;


            clbck_data.m_data1 = p_curr_port;
            progress_bar.push(p_curr_port);
            if (clear) {
                this->ibis_obj.VSPortGeneralCountersClear(p_port_zero->base_lid, 
                                                          i, 
                                                          &vs_port_general_counters, 
                                                          &clbck_data);
            }
            else {
                this->ibis_obj.VSPortGeneralCountersGet(p_port_zero->base_lid, 
                                                        i, 
                                                        &vs_port_general_counters, 
                                                        &clbck_data);
            }

        }
    }
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpPortGeneralCountersToCSV(CSVOut &csv_out) {
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_PORT_GENERAL_COUNTERS))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNumber,port_select,capablity_mask,counter_select,"
            << "rx_icrc_error,tx_parity_error,contain_n_drain_xmit_discards,contain_n_drain_rcv_discards" << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)
            continue;

        VS_PortGeneralCounters* p_port_general_counters = this->fabric_extended_info.
                                    getVSPortGeneralCounters(p_curr_port->createIndex);
        if (!p_port_general_counters)
            continue;

        sstream.str("");
        sstream << PTR(p_curr_port->p_node->guid_get())                    << ','
                << PTR(p_curr_port->guid_get())                            << ','
                << +p_curr_port->num                                       << ','
                << +p_port_general_counters->port_select                   << ','
                << +p_port_general_counters->capablity_mask                << ','
                << +p_port_general_counters->counter_select                << ','
                << +p_port_general_counters->rx_icrc_error                 << ','
                << +p_port_general_counters->tx_parity_error               << ','
                << +p_port_general_counters->contain_n_drain_xmit_discards << ','
                << +p_port_general_counters->contain_n_drain_rcv_discards  << endl;


        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_PORT_GENERAL_COUNTERS);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::BuildPerformanceHistogramInfo(list_p_fabric_general_err& retrieve_errors)
{

    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarNodes progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::VSPerformanceHistogramInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct VS_PerformanceHistogramInfo perf_hist;
    CLEAR_STRUCT(perf_hist);

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!this->GetCapabilityModulePtr()->IsSupportedGMPCapability(
                p_curr_node, EnGMPCAPIsBufferHistogramSupported))
            continue;

        IBPort* p_port_zero = p_curr_node->getPort(0);
        if (!p_port_zero)
            continue;

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        this->ibis_obj.VSPerformanceHistogramInfoGet(p_port_zero->base_lid,
                                                     &perf_hist,
                                                     &clbck_data);
    }
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpPerformanceHistogramInfoToCSV(CSVOut &csv_out) {
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_PERFORMANCE_HISTOGRAM_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,cap_max_sample_time,cap_max_port_hist_id,cap_hist_bin_size,cap_cell_size" << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)
            continue;
        
        VS_PerformanceHistogramInfo* perf_info = this->fabric_extended_info.
                                    getPerformanceHistogramInfo(p_curr_node->createIndex);
        if (!perf_info)
            continue;
        
        sstream.str("");
        sstream << PTR(p_curr_node->guid_get()) << ","
                << PTR(perf_info->cap_max_sample_time) << ","
                << PTR(perf_info->cap_max_port_hist_id) << ","
                << PTR(perf_info->cap_hist_bin_size) << ","
                << PTR(perf_info->cap_cell_size) << endl;

        csv_out.WriteBuf(sstream.str());
    }
    
    csv_out.DumpEnd(SECTION_PERFORMANCE_HISTOGRAM_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
int IBDiag::BuildPerformanceHistogramBufferControl(list_p_fabric_general_err& retrieve_errors) {
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::VSPerformanceHistogramBufferControlClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct VS_PerformanceHistogramBufferControl perf_hist_buff_control;
    CLEAR_STRUCT(perf_hist_buff_control);

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!this->GetCapabilityModulePtr()->IsSupportedGMPCapability(
                p_curr_node, EnGMPCAPIsBufferHistogramSupported))
            continue;

        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port || 
                    !p_curr_port->is_data_worthy() || 
                    p_curr_port->isSpecialPort())
                continue;

            u_int8_t curr_vl = 0;
            u_int8_t curr_dir = 1;
            clbck_data.m_data1 = p_curr_port;
            clbck_data.m_data2 = (void *)(uintptr_t)curr_vl;
            clbck_data.m_data3 = (void *)(uintptr_t)curr_dir;
            progress_bar.push(p_curr_port);

            this->ibis_obj.VSPerformanceHistogramBufferControlGet(p_curr_port->base_lid, 
                                                                  p_curr_port->num, curr_vl, curr_dir,
                                                                  false, false, 
                                                                  &perf_hist_buff_control, 
                                                                  &clbck_data);
            
            curr_vl = 1;
            clbck_data.m_data2 = (void *)(uintptr_t)curr_vl;
            progress_bar.push(p_curr_port);
            this->ibis_obj.VSPerformanceHistogramBufferControlGet(p_curr_port->base_lid, 
                                                                  p_curr_port->num, curr_vl, curr_dir,
                                                                  false, false, 
                                                                  &perf_hist_buff_control, 
                                                                  &clbck_data);
            }
        }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpPerformanceHistogramBufferControlIntoCSV(CSVOut &csv_out) {
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_PERFORMANCE_HISTOGRAM_BUFFER_CONTROL))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNumber,vl,direction,histogram_type,sample_time,mode,hist_min_value,bin_size" << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)
            continue;
        
        sstream.str("");
        if (this->DumpPerformanceHistogramBufferControlByVLAndDir(p_curr_port, sstream, 0, 1))
            csv_out.WriteBuf(sstream.str());

        sstream.str("");
        if (this->DumpPerformanceHistogramBufferControlByVLAndDir(p_curr_port, sstream, 1, 1))
            csv_out.WriteBuf(sstream.str());
    }
    
    csv_out.DumpEnd(SECTION_PERFORMANCE_HISTOGRAM_BUFFER_CONTROL);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

bool IBDiag::DumpPerformanceHistogramBufferControlByVLAndDir(IBPort* p_port, ostream& sout, u_int8_t vl, u_int8_t dir) {
    VS_PerformanceHistogramBufferControl* perf_buff_control = 
            this->fabric_extended_info.getPerformanceHistogramBufferControl(p_port->createIndex, vl, dir);
    if (!perf_buff_control)
        return false;

    sout << PTR(p_port->p_node->guid_get()) << ","
         << PTR(p_port->guid_get()) << ","
         << +p_port->num << ","
         << +vl << ","
         << +dir << ","
         << +perf_buff_control->histogram_type << ","
         << +perf_buff_control->sample_time << ","
         << +perf_buff_control->mode << ","
         << perf_buff_control->hist_min_value << ","
         << PTR(perf_buff_control->bin_size) << endl;
    return true;
}

int IBDiag::BuildPerformanceHistogramBufferData(list_p_fabric_general_err& retrieve_errors) {
     IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::VS_PerformanceHistogramBufferDataClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct VS_PerformanceHistogramBufferData perf_hist_buff_data;
    CLEAR_STRUCT(perf_hist_buff_data);

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!this->GetCapabilityModulePtr()->IsSupportedGMPCapability(
            p_curr_node, EnGMPCAPIsBufferHistogramSupported))
            continue;

        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port || 
                    !p_curr_port->is_data_worthy() || 
                    p_curr_port->isSpecialPort())
                continue;

            u_int8_t curr_vl = 0;
            u_int8_t curr_dir = 1;
            clbck_data.m_data1 = p_curr_port;
            clbck_data.m_data2 = (void *)(uintptr_t)curr_vl;
            clbck_data.m_data3 = (void *)(uintptr_t)curr_dir;
            progress_bar.push(p_curr_port);


            this->ibis_obj.VSPerformanceHistogramBufferDataGet(p_curr_port->base_lid, 
                                                               p_curr_port->num, curr_vl, 
                                                               curr_dir, false, 
                                                               &perf_hist_buff_data, 
                                                               &clbck_data);

            curr_vl = 1;
            clbck_data.m_data2 = (void *)(uintptr_t)curr_vl;
            clbck_data.m_data3 = (void *)(uintptr_t)curr_dir;
            progress_bar.push(p_curr_port);

            this->ibis_obj.VSPerformanceHistogramBufferDataGet(p_curr_port->base_lid, 
                                                               p_curr_port->num, curr_vl, 
                                                               curr_dir, false, 
                                                               &perf_hist_buff_data, 
                                                               &clbck_data);
            }

        }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpPerformanceHistogramBufferDataToCSV(CSVOut &csv_out) {
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_PERFORMANCE_HISTOGRAM_BUFFER_DATA))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNumber,vl,direction,histogram_type,min_sampled,max_sampled,"
            << "bin[0],bin[1],bin[2],bin[3],bin[4],bin[5],bin[6],bin[7],bin[8],bin[9]" << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort* p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)
            continue;
        
        sstream.str("");
        if (this->DumpPerformanceHistogramBufferDataByVLAndDir(p_curr_port, sstream, 0, 1))
            csv_out.WriteBuf(sstream.str());

        sstream.str("");
        if (this->DumpPerformanceHistogramBufferDataByVLAndDir(p_curr_port, sstream, 1, 1))
            csv_out.WriteBuf(sstream.str());
    }
    
    csv_out.DumpEnd(SECTION_PERFORMANCE_HISTOGRAM_BUFFER_DATA);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

bool IBDiag::DumpPerformanceHistogramBufferDataByVLAndDir(IBPort* p_port, ostream& sout, u_int8_t vl, u_int8_t dir) {
    IBDIAG_ENTER;
    VS_PerformanceHistogramBufferData* perf_buff_data = this->fabric_extended_info.
                                    getPerformanceHistogramBufferData(p_port->createIndex, vl, dir);
    if (!perf_buff_data)
        IBDIAG_RETURN(false);
    
    sout << PTR(p_port->p_node->guid_get()) << ","
         << PTR(p_port->guid_get()) << ","
         << +p_port->num << ","
         << +vl << ","
         << +dir << ","
         << +perf_buff_data->histogram_type << ","
         << perf_buff_data->min_sampled << ","
         << perf_buff_data->max_sampled;

    for (size_t i = 0; i < ARRAY_SIZE(perf_buff_data->bin) ; ++i) {
        u_int64_t entry = perf_buff_data->bin[i].High;
        entry = entry  << 32;
        entry |= 0xffffffff & perf_buff_data->bin[i].Low;
        sout << ',' << entry;
    }
    sout << endl;
    IBDIAG_RETURN(true);
}
