/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <typeinfo>
#include <math.h>
#include <assert.h>
#include "phy_diag.h"
#include "acc_reg.h"
#include "phy_diag_errs.h"
#include "infiniband/ibdm/cable/PrtlRecord.h"
#include "phy_diag_types.h"
#include "infiniband/ibutils/output_stream_primitives.h"


template<typename _Type, int (_Type::*_Method)(const clbck_data_t &, int , void*)>
    void AccRegHandler_ForwardClbck(const clbck_data_t &clbck_data,
                      int rec_status,
                      void *p_attribute_data)
{
    //wW cast to a reference in order to avoid having unchecked NULL ptr
    //which dynamic_cast my return.
    //In case of casting to the reference an exception is thrown.
    (dynamic_cast<_Type&>
        (*reinterpret_cast<AccRegHandler*>(clbck_data.m_p_obj)->GetPReg()).*_Method)
              (clbck_data, rec_status, p_attribute_data);
}

static inline bool IsSupportAccRegFailed(IBNode* p_node,
                                         u_int8_t acc_reg_type,
                                         u_int8_t not_support_bit)
{
    return NSB_BITSET(p_node)[acc_reg_type] || NSB_BITSET(p_node)[not_support_bit];
}

bool keycomp(AccRegKey *k1, AccRegKey *k2)
{
    return ((*k1) < (*k2));
};

Register::Register(PhyDiag * phy_diag,
                   uint32_t reg_id,
                   const unpack_data_func_t p_unpack_func,
                   string sn,
                   string name,
                   uint32_t fn,
                   uint64_t nsb,
                   string header,
                   PhyPluginSupportedNodesType support_nodes,
                   bool dump_enabled,
                   bool retrive_disconnected,
                   AccRegVia_t acc_reg_type,
                   AccRegVia_t acc_type_priority)

    : m_phy_diag(phy_diag), m_register_id(reg_id), m_fields_num(fn), m_not_supported_bit(nsb),
      m_section_name(sn), m_header(header), m_name(name),
      m_support_nodes(support_nodes), m_retrive_disconnected(retrive_disconnected),
      m_dump_enabled(dump_enabled), m_acc_reg_type(acc_reg_type), unpack_data_func(p_unpack_func)

{
    IBDIAGNET_ENTER;

    m_acc_type_priority = PhyDiag::acc_reg_priority == NOT_SUP_ACC_REG ?
            acc_type_priority : PhyDiag::acc_reg_priority;

    IBDIAGNET_RETURN_VOID;
}

bool Register::IsRegSupportNodeType(const IBNode* node)
{
    IBDIAGNET_ENTER;
    bool rc = IsPhyPluginSupportNodeType(m_support_nodes, node);
    IBDIAGNET_RETURN(rc);
}


void Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    sstream << handler_header;
    if (this->GetHeader().empty())
        for (u_int32_t cnt = 0; cnt < this->GetFieldsNum(); cnt++)
            sstream << ",field" <<  cnt;
    else
        sstream << this->GetHeader();

    IBDIAGNET_RETURN_VOID;
}

void Register::GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices)
{
    IBDIAGNET_ENTER;
    assert(true); // make sure that this function called only from
                  // derived registers that overrun this function.

    IBDIAGNET_RETURN_VOID;
}

int Register::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors)
{
    IBDIAGNET_ENTER;
    assert(true); // make sure that this function called only from
                  // derived classes that overrun this function.

    IBDIAGNET_RETURN(IBDIAG_ERR_CODE_ILLEGAL_FUNCTION_CALL);
}

int Register::SensorsBitsToList(u_int64_t sensors,
                                u_int8_t max_sensors,
                                list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    u_int64_t mask = 1;

    for (u_int8_t i = 0; i < max_sensors; ++i, mask <<= 1) {
        if (mask & sensors)
            sensors_list.push_back(i);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int Register::SensorsCountToList(u_int8_t sensor_count, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;

    for (u_int8_t i = 0; i < sensor_count; ++i)
        sensors_list.push_back(i);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int Register::BuildDB(AccRegHandler* p_handler,
                      list_p_fabric_general_err &phy_errors,
                      ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    assert(false); // make sure that this function called only from
                   // derived classes that overrun this function.

    IBDIAGNET_RETURN(IBDIAG_ERR_CODE_ILLEGAL_FUNCTION_CALL);
}

void Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                  export_data_phy_node_t *p_export_data_phy_node,
                                  struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

bool Register::CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                export_data_phy_node_t *p_export_data_phy_node,
                                AccRegKey *p_key)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(false);
}

//*****************************************
//***     PRTL Register functions       ***
//*****************************************
PRTLRegister::PRTLRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
            ACCESS_REGISTER_ID_PRTL,
            (const unpack_data_func_t)prtl_reg_unpack,
            ACC_REG_PRTL_INTERNAL_SECTION_NAME,
            ACC_REG_PRTL_NAME,
            ACC_REG_PRTL_FIELDS_NUM,//how many output columns of the register are going to print
            NOT_SUPPORT_PRTL,
            ACC_REG_PRTL_FIELDS_HEADER,
            SUPPORT_SW_CA, //should support HCA and Switches
            true,
            false,
            VIA_SMP,
            VIA_GMP)
{
}

int PRTLRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    AccRegKeyPort * key = dynamic_cast<AccRegKeyPort*>(p_key);
    if (!key || !p_node)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    if (p_node->serdes_version < SerDes_7nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    //On switches CAP mask should be taken from "0 port"
    const IBPort *p_port = p_node->getPort(p_node->type == IB_SW_NODE ? 0 : key->port_num);
    if (!p_port)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    const SMP_PortInfo * portInfo =
           this->m_phy_diag->GetIBDiag()->GetIBDMExtendedInfoPtr()->getSMPPortInfo(p_port->createIndex);
    if (!portInfo || !IS_SUPPORT_LINK_ROUND_TRIP_LATENCY(portInfo->CapMsk))
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    //sent any time cause it used for cable's length calculation even
    //if a cable is fetched by the "Cable" plugin

    struct prtl_reg prtl;
    CLEAR_STRUCT(prtl);
    prtl.pnat =  ACC_REG_PNAT_IB_PORT;
    prtl.local_port = ((AccRegKeyPort *)p_key)->port_num;

    prtl_reg_pack(&prtl, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

#ifdef COPY
    #error "The PRTL COPY macros has been already defined"
#else
    #define COPY(value) p_port->p_prtl->value = p_reg->value;
#endif

int PRTLRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer)
{
    IBDIAGNET_ENTER;
    int rc = Register::UnpackData(p_key, data_to_unpack, unpacked_buffer);
    if (rc)
        IBDIAG_RETURN(rc);

    const AccRegKeyPort *p_regKey = dynamic_cast<const AccRegKeyPort*>(p_key);
    if (!p_regKey)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    IBPort *p_port = this->m_phy_diag->GetPort(p_regKey->node_guid, p_regKey->port_num);
    if (!p_port || p_port->p_prtl)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    const prtl_reg *p_reg = reinterpret_cast<prtl_reg*>(data_to_unpack);
    p_port->p_prtl = new PrtlRecord();
    COPY(lp_msb);
    COPY(local_port);
    COPY(rtt_support);
    COPY(latency_accuracy);
    COPY(latency_res);
    COPY(local_phy_latency);
    COPY(local_mod_dp_latency);
    COPY(round_trip_latency);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PRTLRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                   << '#'
            << handler_header         << ','
            << "lp_msb"               << ','
            << "pnat"                 << ','
            << "local_port"           << ','
            << "rtt_support"          << ','
            << "latency_accuracy"     << ','
            << "latency_res"          << ','
            << "local_phy_latency"    << ','
            << "local_mod_dp_latency" << ','
            << "round_trip_latency";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PRTLRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct prtl_reg &prtl = areg.regs.prtl;
    sstream << +prtl.lp_msb              << ','
            << +prtl.pnat                << ','
            << +prtl.local_port          << ','
            << +prtl.rtt_support         << ','
            << +prtl.latency_accuracy    << ','
            << +prtl.latency_res         << ','
            << prtl.local_phy_latency    << ','
            << prtl.local_mod_dp_latency << ','
            << prtl.round_trip_latency   << endl;

    IBDIAGNET_RETURN_VOID;
}

void PRTLRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_prtl_reg = &areg.regs.prtl;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***     PCAM Register functions       ***
//*****************************************
PCAMRegister::PCAMRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PCAM,
                (const unpack_data_func_t)pcam_reg_unpack,
                ACC_REG_PCAM_INTERNAL_SECTION_NAME,
                ACC_REG_PCAM_NAME,
                ACC_REG_PCAM_FIELDS_NUM,//how many output columns of the register are going to print
                NOT_SUPPORT_PCAM,
                ACC_REG_PCAM_FIELDS_HEADER,
                SUPPORT_CA,
                true,
                false,
                VIA_GMP,
                VIA_GMP),
      m_access_reg_group(0), m_feature_group(0)
{
}

AccRegVia_t PCAMRegister::Validation(IBNode *p_node, int &rc)
{
    //TODO take care of
    //HCA_CAP.PCAM == 1

    IBDIAG_RETURN(Register::Validation(p_node, rc));
}

int PCAMRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pcam_reg pcam;
    CLEAR_STRUCT(pcam);

    pcam.access_reg_group   = m_access_reg_group;
    pcam.feature_group      = m_feature_group;

    pcam_reg_pack(&pcam, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PCAMRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer)
{
    IBDIAGNET_ENTER;
    int rc = Register::UnpackData(p_key, data_to_unpack, unpacked_buffer);

    if (!rc) {
        PHYNodeData* data = this->m_phy_diag->GetPHYNodeData(p_key->node_guid);
        if (data && !data->pcam)
            data->pcam = new pcam_reg(*(pcam_reg*)data_to_unpack);
    }

    IBDIAG_RETURN(rc);
}

void PCAMRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                        << '#'
            << handler_header              << ","
            << "access_reg_group"          << ","
            << "feature_group"             << "," 
            << "port_access_reg_cap_mask"  << ","
            << "feature_cap_mask";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PCAMRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct pcam_reg &pcam = areg.regs.pcam;
    sstream << +pcam.access_reg_group      << ','
            << +pcam.feature_group         << ',';

    for (size_t i = 0; i < sizeof(pcam.port_access_reg_cap_mask); i++) {
        if (!(i % 8))
            sstream << ' ';

        sstream << +pcam.port_access_reg_cap_mask[i];
    }

    sstream << ',';

    for (size_t i = 0; i < sizeof(pcam.feature_cap_mask); i++) {
        if (!(i % 8))
            sstream << ' ';

        sstream << +pcam.feature_cap_mask[i];
    }

    sstream << endl;

    IBDIAG_RETURN_VOID;
}

void PCAMRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node)
        p_export_data_phy_node->p_pcam_reg = &areg.regs.pcam;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* PPLL Register functions ***
//*****************************************
PPLLRegister::PPLLRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PPLL,
                (const unpack_data_func_t)ppll_reg_unpack,
                ACC_REG_PPLL_INTERNAL_SECTION_NAME,
                ACC_REG_PPLL_NAME,
                ACC_REG_PPLL_MAX_FIELDS_NUM,
                NOT_SUPPORT_PPLL)
{
}

int PPLLRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct ppll_reg ppll;
    CLEAR_STRUCT(ppll);
    ppll.pll_group = ((AccRegKeyGroup *)p_key)->group_num;
    ppll_reg_pack(&ppll, data);
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPLLRegister::Header_Dump_28nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "version_28nm=1"    << ','
            << "num_pll_groups"    << ','
            << "pll_group"         << ','
            << "pci_oob_pll"       << ','
            << "num_plls"          << ','
            << "ae";

    for (size_t i = 0; i < ARRAY_SIZE(ppll_28nm::pll_status); ++i)
        sstream << ',' << "pll_28[" << i << "].lock_cal"
                << ',' << "pll_28[" << i << "].lock_status"
                << ',' << "pll_28[" << i << "].algo_f_ctrl"
                << ',' << "pll_28[" << i << "].analog_algo_num_var"
                << ',' << "pll_28[" << i << "].f_ctrl_measure"
                << ',' << "pll_28[" << i << "].analog_var"
                << ',' << "pll_28[" << i << "].high_var"
                << ',' << "pll_28[" << i << "].low_var"
                << ',' << "pll_28[" << i << "].mid_var";

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Header_Dump_16nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "version_16nm=3"    << ','
            << "num_pll_groups"    << ','
            << "pll_group"         << ','
            << "pci_oob_pll"       << ','
            << "num_plls";

    for (size_t i = 0; i < ARRAY_SIZE(ppll_reg_16nm::pll_status); ++i)
        sstream << ',' << "pll_16[" << i << "].lock_status"
                << ',' << "pll_16[" << i << "].pll_pwrup"
                << ',' << "pll_16[" << i << "].lock_pulse"
                << ',' << "pll_16[" << i << "].lock_lost_counter"
                << ',' << "pll_16[" << i << "].ae"
                << ',' << "pll_16[" << i << "].clear_cause"
                << ',' << "pll_16[" << i << "].lock_clk_val_cause"
                << ',' << "pll_16[" << i << "].plllock_clk_val"
                << ',' << "pll_16[" << i << "].clock_valid"
                << ',' << "pll_16[" << i << "].pll_speed"
                << ',' << "pll_16[" << i << "].lock"
                << ',' << "pll_16[" << i << "].analog_var"
                << ',' << "pll_16[" << i << "].high_var"
                << ',' << "pll_16[" << i << "].low_var"
                << ',' << "pll_16[" << i << "].mid_var"
                << ',' << "pll_16[" << i << "].fctrl_measure";

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "version_7nm=4"                        << ','
            << "num_pll_groups"                       << ','
            << "pll_group"                            << ','
            << "pci_oob_pll"                          << ','
            << "num_plls"                             << ','
            << "pll_ugl_state"                        << ','
            << "lokcdet_error_count"                  << ','
            << "lockdet_err_measure_done"             << ','
            << "lockdet_err_state"                    << ','
            << "lockdet_cnt_based_lock"               << ','
            << "lockdet_err_cnt_unlocked_sticky"      << ','
            << "lockdet_err_cnt_locked_sticky"        << ','
            << "lockdet_err_cnt_active_lock"          << ','
            << "cal_internal_state"                   << ','
            << "cal_main_state"                       << ','
            << "pll_lockdet_state"                    << ','
            << "pwr_up_state"                         << ','
            << "cal_abort_sticky"                     << ','
            << "cal_abort"                            << ','
            << "cal_valid_sticky"                     << ','
            << "dco_coarse"                           << ','
            << "cal_done"                             << ','
            << "cal_en"                               << ','
            << "dco_coarse_bin"                       << ','
            << "cal_error_counter"                    << ','
            << "lma_rcal_code"                        << ','
            << "lma_rcal_tx_p_code"                   << ','
            << "lma_rcal_tx_n_code"                   << ','
            << "rcal_done_valid_sticky"               << ','
            << "rcal_tx_p_valid_sticky"               << ','
            << "rcal_tx_n_valid_sticky"               << ','
            << "rcal_fsm_done"                        << ','
            << "speedo_valid"                         << ','
            << "cvb_trim_valid"                       << ','
            << "bg_trim_valid"                        << ','
            << "speedo"                               << ','
            << "cvb_trim"                             << ','
            << "bg_trim";

    IBDIAGNET_RETURN_VOID
}

void PPLLRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    IBDIAGNET_ENTER;

    sstream << "version_5nm=5"                        << ','
            << "num_pll_groups"                       << ','
            << "pll_group"                            << ','
            << "pci_oob_pll"                          << ','
            << "num_plls"                             << ','
            << "lockdet_err_measure_done"             << ','
            << "lockdet_err_state"                    << ','
            << "lockdet_cnt_based_lock"               << ','
            << "lockdet_err_cnt_unlocked_sticky"      << ','
            << "lockdet_err_cnt_locked_sticky"        << ','
            << "lockdet_err_cnt_active_lock"          << ','
            << "dlf2_phlock_det_status_sticky"        << ','
            << "lockdet_error_count"                  << ','
            << "speedo_valid"                         << ','
            << "cvb_trim_valid"                       << ','
            << "bg_trim_valid"                        << ','
            << "lockdet_mode_retimer"                 << ','
            << "speedo"                               << ','
            << "cvb_trim"                             << ','
            << "bg_trim";

    IBDIAGNET_RETURN_VOID
}

void PPLLRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_28nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct ppll_reg &ppll = areg.regs.ppll;

    sstream << hex
            << "0x" << +ppll.version           << ','
            << "0x" << +ppll.num_pll_groups    << ','
            << "0x" << +ppll.pll_group         << ','
            << "0x" << +ppll.pci_oob_pll       << ','
            << "0x" << +ppll.num_plls          << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (ppll.version) {

        case 1: // 28nm
            Dump_28nm(ppll, sstream);
            break;

        case 3: // 16nm
            Dump_16nm(ppll, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(ppll, sstream);
            break;

        case 5: // 5nm
            Dump_5nm(ppll, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +ppll.version << '"';
            for (int i = 0; i < ACC_REG_PPLL_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for PPLL: %d, on node: " U64H_FMT ".\n",
                        ppll.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_28nm(const struct ppll_reg &ppll, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    ppll_28nm ppll_28;
    ppll_28nm_unpack(&ppll_28, ppll.page_data.ppll_data_set.data);

    sstream << "0x" << +ppll_28.ae;

    int pll_status_size = ARRAY_SIZE(ppll_28.pll_status);

    for (int i = 0; i < pll_status_size; ++i) {
        sstream << ',';
        Dump_pll_28nm(ppll_28.pll_status[i], sstream);
    }

    for (int i = 0; i < ACC_REG_PPLL_MAX_FIELDS_NUM - ACC_REG_PPLL28NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_16nm(const struct ppll_reg &ppll, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    ppll_reg_16nm ppll_16;
    ppll_reg_16nm_unpack(&ppll_16, ppll.page_data.ppll_data_set.data);

    int pll_status_size = ARRAY_SIZE(ppll_16.pll_status);

    for (int i = 0; i < pll_status_size; ++i) {
        if (i)
            sstream << ',';
        Dump_pll_16nm(ppll_16.pll_status[i], sstream);
    }

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_7nm(const struct ppll_reg &ppll, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    ppll_reg_7nm ppll_7;
    ppll_reg_7nm_unpack(&ppll_7, ppll.page_data.ppll_data_set.data);

    sstream << "0x" << +ppll_7.pll_ugl_state                        << ','
            << "0x" << +ppll_7.lokcdet_error_count                  << ','
            << "0x" << +ppll_7.lockdet_err_measure_done             << ','
            << "0x" << +ppll_7.lockdet_err_state                    << ','
            << "0x" << +ppll_7.lockdet_cnt_based_lock               << ','
            << "0x" << +ppll_7.lockdet_err_cnt_unlocked_sticky      << ','
            << "0x" << +ppll_7.lockdet_err_cnt_locked_sticky        << ','
            << "0x" << +ppll_7.lockdet_err_cnt_active_lock          << ','
            << "0x" << +ppll_7.cal_internal_state                   << ','
            << "0x" << +ppll_7.cal_main_state                       << ','
            << "0x" << +ppll_7.pll_lockdet_state                    << ','
            << "0x" << +ppll_7.pwr_up_state                         << ','
            << "0x" << +ppll_7.cal_abort_sticky                     << ','
            << "0x" << +ppll_7.cal_abort                            << ','
            << "0x" << +ppll_7.cal_valid_sticky                     << ','
            << "0x" << +ppll_7.dco_coarse                           << ','
            << "0x" << +ppll_7.cal_done                             << ','
            << "0x" << +ppll_7.cal_en                               << ','
            << "0x" << +ppll_7.dco_coarse_bin                       << ','
            << "0x" << +ppll_7.cal_error_counter                    << ','
            << "0x" << +ppll_7.lma_rcal_code                        << ','
            << "0x" << +ppll_7.lma_rcal_tx_p_code                   << ','
            << "0x" << +ppll_7.lma_rcal_tx_n_code                   << ','
            << "0x" << +ppll_7.rcal_done_valid_sticky               << ','
            << "0x" << +ppll_7.rcal_tx_p_valid_sticky               << ','
            << "0x" << +ppll_7.rcal_tx_n_valid_sticky               << ','
            << "0x" << +ppll_7.rcal_fsm_done                        << ','
            << "0x" << +ppll_7.speedo_valid                         << ','
            << "0x" << +ppll_7.cvb_trim_valid                       << ','
            << "0x" << +ppll_7.bg_trim_valid                        << ','
            << "0x" << +ppll_7.speedo                               << ','
            << "0x" << +ppll_7.cvb_trim                             << ','
            << "0x" << +ppll_7.bg_trim                              ;

    for (int i = 0; i < ACC_REG_PPLL_MAX_FIELDS_NUM - ACC_REG_PPLL7NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_5nm(const struct ppll_reg &ppll, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    ppll_reg_5nm ppll_5;
    ppll_reg_5nm_unpack(&ppll_5, ppll.page_data.ppll_data_set.data);

    sstream << "0x" << +ppll_5.lockdet_err_measure_done                    << ','
            << "0x" << +ppll_5.lockdet_err_state                           << ','
            << "0x" << +ppll_5.lockdet_cnt_based_lock                      << ','
            << "0x" << +ppll_5.lockdet_err_cnt_unlocked_sticky             << ','
            << "0x" << +ppll_5.lockdet_err_cnt_locked_sticky               << ','
            << "0x" << +ppll_5.lockdet_err_cnt_active_lock                 << ','
            << "0x" << +ppll_5.dlf2_phlock_det_status_sticky               << ','
            << "0x" << +ppll_5.lockdet_error_count                         << ','
            << "0x" << +ppll_5.speedo_valid                                << ','
            << "0x" << +ppll_5.cvb_trim_valid                              << ','
            << "0x" << +ppll_5.bg_trim_valid                               << ','
            << "0x" << +ppll_5.lockdet_mode_retimer                        << ','
            << "0x" << +ppll_5.speedo                                      << ','
            << "0x" << +ppll_5.cvb_trim                                    << ','
            << "0x" << +ppll_5.bg_trim;

    for (int i = 0; i < ACC_REG_PPLL_MAX_FIELDS_NUM - ACC_REG_PPLL5NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_pll_28nm(const struct pll_28nm &pll_28, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "0x" << +pll_28.lock_cal               << ','
            << "0x" << +pll_28.lock_status            << ','
            << "0x" << +pll_28.algo_f_ctrl            << ','
            << "0x" << +pll_28.analog_algo_num_var    << ','
            << "0x" << +pll_28.f_ctrl_measure         << ','
            << "0x" << +pll_28.analog_var             << ','
            << "0x" << +pll_28.high_var               << ','
            << "0x" << +pll_28.low_var                << ','
            << "0x" << +pll_28.mid_var                ;

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_pll_16nm(const struct pll_status_data &pll_16, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "0x" << +pll_16.lock_status             << ','
            << "0x" << +pll_16.pll_pwrup               << ','
            << "0x" << +pll_16.lock_pulse              << ','
            << "0x" << +pll_16.lock_lost_counter       << ','
            << "0x" << +pll_16.ae                      << ','
            << "0x" << +pll_16.clear_cause             << ','
            << "0x" << +pll_16.lock_clk_val_cause      << ','
            << "0x" << +pll_16.plllock_clk_val         << ','
            << "0x" << +pll_16.clock_valid             << ','
            << "0x" << +pll_16.pll_speed               << ','
            << "0x" << +pll_16.lock                    << ','
            << "0x" << +pll_16.analog_var              << ','
            << "0x" << +pll_16.high_var                << ','
            << "0x" << +pll_16.low_var                 << ','
            << "0x" << +pll_16.mid_var                 << ','
            << "0x" << +pll_16.fctrl_measure           ;

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node) {
        struct ppll_reg *p_ppll_reg = new struct ppll_reg(areg.regs.ppll);

        switch (p_ppll_reg->version) {
            case 0: // 28nm
                ppll_28nm_unpack(&p_ppll_reg->page_data.ppll_28nm, areg.regs.ppll.page_data.ppll_data_set.data);
                break;
            case 1: // 16nm
                ppll_reg_16nm_unpack(&p_ppll_reg->page_data.ppll_reg_16nm, areg.regs.ppll.page_data.ppll_data_set.data);
                break;
            case 4: // 7nm
                ppll_reg_7nm_unpack(&p_ppll_reg->page_data.ppll_reg_7nm, areg.regs.ppll.page_data.ppll_data_set.data);
                break;
            case 5: // 5nm
                ppll_reg_5nm_unpack(&p_ppll_reg->page_data.ppll_reg_5nm, areg.regs.ppll.page_data.ppll_data_set.data);
                break;
            default:
                break;
        }

        p_export_data_phy_node->p_ppll_reg[((AccRegKeyGroup *)p_key)->group_num] = p_ppll_reg;
    }

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* PTAS Register functions ***
//*****************************************
PTASRegister::PTASRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PTAS,
                (const unpack_data_func_t)ptas_reg_unpack,
                ACC_REG_PTAS_INTERNAL_SECTION_NAME,
                ACC_REG_PTAS_NAME,
                ACC_REG_PTAS_FIELDS_NUM,
                NOT_SUPPORT_PTAS)
{
}

int PTASRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PTASRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                             <<  '#'
            << handler_header                   <<  ','
            << "algorithm_options"              <<  ','
            << "repetitions_mode"               <<  ','
            << "num_of_repetitions"             <<  ','
            << "grade_version"                  <<  ','
            << "height_grade_type"              <<  ','
            << "phase_grade_type"               <<  ','
            << "height_grade_weight"            <<  ','
            << "phase_grade_weight"             <<  ','
            << "gisim_measure_bits"             <<  ','
            << "adaptive_tap_measure_bits"      <<  ','
            << "ber_bath_high_error_threshold"  <<  ','
            << "ber_bath_mid_error_threshold"   <<  ','
            << "ber_bath_low_error_threshold"   <<  ','
            << "one_ratio_high_threshold"       <<  ','
            << "one_ratio_high_mid_threshold"   <<  ','
            << "one_ratio_low_mid_threshold"    <<  ','
            << "one_ratio_low_threshold"        <<  ','
            << "ndeo_error_threshold"           <<  ','
            << "mixer_offset_step_size"         <<  ','
            << "mix90_phase_for_voltage_bath"   <<  ','
            << "mixer_offset_start"             <<  ','
            << "mixer_offset_end"               <<  ','
            << "ber_test_time";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PTASRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    char buffer[1024] = {0};
    const struct ptas_reg & p_ptas = areg.regs.ptas;
    //register fields
    snprintf(buffer, sizeof(buffer),
             UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
             UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
             UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
             UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
             UH_FMT "," UH_FMT "," UH_FMT ,
             p_ptas.algorithm_options,
             p_ptas.repetitions_mode,
             p_ptas.num_of_repetitions,
             p_ptas.grade_version,
             p_ptas.height_grade_type,
             p_ptas.phase_grade_type,
             p_ptas.height_grade_weight,
             p_ptas.phase_grade_weight,
             p_ptas.gisim_measure_bits,
             p_ptas.adaptive_tap_measure_bits,
             p_ptas.ber_bath_high_error_threshold,
             p_ptas.ber_bath_mid_error_threshold,
             p_ptas.ber_bath_low_error_threshold,
             p_ptas.one_ratio_high_threshold,
             p_ptas.one_ratio_high_mid_threshold,
             p_ptas.one_ratio_low_mid_threshold,
             p_ptas.one_ratio_low_threshold,
             p_ptas.ndeo_error_threshold,
             p_ptas.mixer_offset_step_size,
             p_ptas.mix90_phase_for_voltage_bath,
             p_ptas.mixer_offset_start,
             p_ptas.mixer_offset_end,
             p_ptas.ber_test_time);
    sstream << buffer << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* PMDR Register functions ***
//*****************************************
PMDRRegister::PMDRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PMDR,
                (const unpack_data_func_t)pmdr_reg_unpack,
                ACC_REG_PMDR_INTERNAL_SECTION_NAME,
                ACC_REG_PMDR_NAME,
                ACC_REG_PMDR_FIELDS_NUM,
                NOT_SUPPORT_PMDR),
       m_pnat(ACC_REG_PNAT_IB_PORT)
{
}

int PMDRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pmdr_reg pmdr;
    CLEAR_STRUCT(pmdr);
    pmdr.local_port = ((AccRegKeyPort *)p_key)->port_num;
    pmdr.pnat = this->m_pnat;
    pmdr_reg_pack(&pmdr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PMDRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS    
    sstream << endl                    << '#'
            << handler_header          << ','
            << "status"                << ','
            << "version"               << ','
            << "gb_valid"              << ','
            << "mcm_tile_valid"        << ','
            << "pport"                 << ','
            << "module"                << ','
            << "cluster"               << ','
            << "clp_4x"                << ','
            << "clp_1x"                << ','
            << "split"                 << ','
            << "swid"                  << ','
            << "module_lane_mask"      << ','
            << "ib_port"               << ','
            << "label_port_query"      << ','
            << "local_port_query"      << ','
            << "usr_sd_4x_tx"          << ','
            << "usr_sd_4x_rx"          << ','
            << "sd_4x_tx"              << ','
            << "sd_4x_rx"              << ','
            << "sd_lane_mask_tx"       << ','
            << "sd_lane_mask_rx"       << ','
            << "slot_index"            << ','
            << "tile_cluster"          << ','
            << "mcm_tile_num"          << ','
            << "tile_pport"            << ','
            << "gearbox_die_num"       << ','
            << "vl_num"                << ','
            << "lane7_physical_rx"     << ','
            << "lane6_physical_rx"     << ','
            << "lane5_physical_rx"     << ','
            << "lane4_physical_rx"     << ','
            << "lane3_physical_rx"     << ','
            << "lane2_physical_rx"     << ','
            << "lane1_physical_rx"     << ','
            << "lane0_physical_rx"     << ','
            << "pll_index"             << ','
            << "lane7_physical_tx"     << ','
            << "lane6_physical_tx"     << ','
            << "lane5_physical_tx"     << ','
            << "lane4_physical_tx"     << ','
            << "lane3_physical_tx"     << ','
            << "lane2_physical_tx"     << ','
            << "lane1_physical_tx"     << ','
            << "lane0_physical_tx"     << ','
            << "VL7_lane_map"          << ','
            << "VL6_lane_map"          << ','
            << "VL5_lane_map"          << ','
            << "VL4_lane_map"          << ','
            << "VL3_lane_map"          << ','
            << "VL2_lane_map"          << ','
            << "VL1_lane_map"          << ','
            << "VL0_lane_map"          << ','
            << "VL15_lane_map"         << ','
            << "VL14_lane_map"         << ','
            << "VL13_lane_map"         << ','
            << "VL12_lane_map"         << ','
            << "VL11_lane_map"         << ','
            << "VL10_lane_map"         << ','
            << "VL9_lane_map"          << ','
            << "VL8_lane_map"          << ','
            << "VL23_lane_map"         << ','
            << "VL22_lane_map"         << ','
            << "VL21_lane_map"         << ','
            << "VL20_lane_map"         << ','
            << "VL19_lane_map"         << ','
            << "VL18_lane_map"         << ','
            << "VL17_lane_map"         << ','
            << "VL16_lane_map"         << ','
            << "VL31_lane_map"         << ','
            << "VL30_lane_map"         << ','
            << "VL29_lane_map"         << ','
            << "VL28_lane_map"         << ','
            << "VL27_lane_map"         << ','
            << "VL26_lane_map"         << ','
            << "VL25_lane_map"         << ','
            << "VL24_lane_map";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PMDRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct pmdr_reg &pmdr = areg.regs.pmdr;

    //register fields
    sstream << +pmdr.status                << ","
            << +pmdr.version               << ","
            << +pmdr.gb_valid              << ","
            << +pmdr.mcm_tile_valid        << ","
            << +pmdr.pport                 << ","
            << +pmdr.module                << ","
            << +pmdr.cluster               << ","
            << +pmdr.clp_4x                << ","
            << +pmdr.clp_1x                << ","
            << +pmdr.split                 << ","
            << +pmdr.swid                  << ","
            << +pmdr.module_lane_mask      << ","
            << +pmdr.ib_port               << ","
            << +pmdr.label_port_query      << ","
            << +pmdr.local_port_query      << ","
            << +pmdr.usr_sd_4x_tx          << ","
            << +pmdr.usr_sd_4x_rx          << ","
            << +pmdr.sd_4x_tx              << ","
            << +pmdr.sd_4x_rx              << ","
            << +pmdr.sd_lane_mask_tx       << ","
            << +pmdr.sd_lane_mask_rx       << ","
            << +pmdr.slot_index            << ","
            << +pmdr.tile_cluster          << ","
            << +pmdr.mcm_tile_num          << ","
            << +pmdr.tile_pport            << ","
            << +pmdr.gearbox_die_num       << ","
            << +pmdr.vl_num                << ","
            << +pmdr.lane7_physical_rx     << ","
            << +pmdr.lane6_physical_rx     << ","
            << +pmdr.lane5_physical_rx     << ","
            << +pmdr.lane4_physical_rx     << ","
            << +pmdr.lane3_physical_rx     << ","
            << +pmdr.lane2_physical_rx     << ","
            << +pmdr.lane1_physical_rx     << ","
            << +pmdr.lane0_physical_rx     << ","
            << +pmdr.pll_index             << ","
            << +pmdr.lane7_physical_tx     << ","
            << +pmdr.lane6_physical_tx     << ","
            << +pmdr.lane5_physical_tx     << ","
            << +pmdr.lane4_physical_tx     << ","
            << +pmdr.lane3_physical_tx     << ","
            << +pmdr.lane2_physical_tx     << ","
            << +pmdr.lane1_physical_tx     << ","
            << +pmdr.lane0_physical_tx     << ","
            << +pmdr.VL7_lane_map          << ","
            << +pmdr.VL6_lane_map          << ","
            << +pmdr.VL5_lane_map          << ","
            << +pmdr.VL4_lane_map          << ","
            << +pmdr.VL3_lane_map          << ","
            << +pmdr.VL2_lane_map          << ","
            << +pmdr.VL1_lane_map          << ","
            << +pmdr.VL0_lane_map          << ","
            << +pmdr.VL15_lane_map         << ","
            << +pmdr.VL14_lane_map         << ","
            << +pmdr.VL13_lane_map         << ","
            << +pmdr.VL12_lane_map         << ","
            << +pmdr.VL11_lane_map         << ","
            << +pmdr.VL10_lane_map         << ","
            << +pmdr.VL9_lane_map          << ","
            << +pmdr.VL8_lane_map          << ","
            << +pmdr.VL23_lane_map         << ","
            << +pmdr.VL22_lane_map         << ","
            << +pmdr.VL21_lane_map         << ","
            << +pmdr.VL20_lane_map         << ","
            << +pmdr.VL19_lane_map         << ","
            << +pmdr.VL18_lane_map         << ","
            << +pmdr.VL17_lane_map         << ","
            << +pmdr.VL16_lane_map         << ","
            << +pmdr.VL31_lane_map         << ","
            << +pmdr.VL30_lane_map         << ","
            << +pmdr.VL29_lane_map         << ","
            << +pmdr.VL28_lane_map         << ","
            << +pmdr.VL27_lane_map         << ","
            << +pmdr.VL26_lane_map         << ","
            << +pmdr.VL25_lane_map         << ","
            << +pmdr.VL24_lane_map
            << endl;

    IBDIAGNET_RETURN_VOID;
}

void PMDRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pmdr_reg = &areg.regs.pmdr;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* PMCR Register functions ***
//*****************************************
PMCRRegister::PMCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PMCR,
                (const unpack_data_func_t)pmcr_reg_unpack,
                ACC_REG_PMCR_INTERNAL_SECTION_NAME,
                ACC_REG_PMCR_NAME,
                ACC_REG_PMCR_FIELDS_NUM,
                NOT_SUPPORT_PMCR)
{
}

int PMCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pmcr_reg pmcr;
    CLEAR_STRUCT(pmcr);
    pmcr.local_port = ((AccRegKeyPort *)p_key)->port_num;
    pmcr_reg_pack(&pmcr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PMCRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                               << '#'
            << handler_header                     << ','
            << "cs_sel"                           << ','
            << "local_port"                       << ','
            << "lp_msb"                           << ','
            << "error_code_response"              << ','
            << "c_db"                             << ','
            << "tx_disable_override_value"        << ','
            << "tx_disable_override_cntl"         << ','
            << "cdr_override_value"               << ','
            << "cdr_override_cntl"                << ','
            << "rx_los_override_admin"            << ','
            << "rx_los_override_cntl"             << ','
            << "rx_amp_override_value"            << ','
            << "rx_amp_override_cntl"             << ','
            << "rx_post_emp_override_value"       << ','
            << "rx_post_emp_override_cntl"        << ','
            << "rx_emp_override_value"            << ','
            << "rx_emp_override_cntl"             << ','
            << "tx_adaptive_override_cntrl"       << ','
            << "tx_adaptive_override_value"       << ','
            << "tx_equ_override_value"            << ','
            << "tx_equ_override_cntl"             << ','
            << "rs_override_value"                << ','
            << "rs_override_cntl"                 << ','
            << "si_override_cntrl"                << ','
            << "ap_sel_override_value"            << ','
            << "ap_sel_override_cntrl";
#endif

    IBDIAGNET_RETURN_VOID;

}

void PMCRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct pmcr_reg &pmcr = areg.regs.pmcr;

    //register fields
    sstream << +pmcr.cs_sel                           << ","
            << +pmcr.local_port                       << ","
            << +pmcr.lp_msb                           << ","
            << +pmcr.error_code_response              << ","
            << +pmcr.c_db                             << ","
            << +pmcr.tx_disable_override_value        << ","
            << +pmcr.tx_disable_override_cntl         << ","
            << +pmcr.cdr_override_value               << ","
            << +pmcr.cdr_override_cntl                << ","
            << +pmcr.rx_los_override_admin            << ","
            << +pmcr.rx_los_override_cntl             << ","
            << +pmcr.rx_amp_override_value            << ","
            << +pmcr.rx_amp_override_cntl             << ","
            << +pmcr.rx_post_emp_override_value       << ","
            << +pmcr.rx_post_emp_override_cntl        << ","
            << +pmcr.rx_emp_override_value            << ","
            << +pmcr.rx_emp_override_cntl             << ","
            << +pmcr.tx_adaptive_override_cntrl       << ","
            << +pmcr.tx_adaptive_override_value       << ","
            << +pmcr.tx_equ_override_value            << ","
            << +pmcr.tx_equ_override_cntl             << ","
            << +pmcr.rs_override_value                << ","
            << +pmcr.rs_override_cntl                 << ","
            << +pmcr.si_override_cntrl                << ","
            << +pmcr.ap_sel_override_value            << ","
            << +pmcr.ap_sel_override_cntrl            << endl;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* PTYS Register functions ***
//*****************************************
PTYSRegister::PTYSRegister(PhyDiag * phy_diag)
    : Register(phy_diag, ACCESS_REGISTER_ID_PTYS,
                (const unpack_data_func_t)ptys_reg_unpack,
                ACC_REG_PTYS_INTERNAL_SECTION_NAME,
                ACC_REG_PTYS_NAME,
                ACC_REG_PTYS_FIELDS_NUM,
                NOT_SUPPORT_PTYS,
                "",
                SUPPORT_SW_CA,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP),
    m_pnat(ACC_REG_PNAT_IB_PORT)
{
}

int PTYSRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct ptys_reg ptys;
    CLEAR_STRUCT(ptys);
    ptys.local_port = ((AccRegKeyPort *)p_key)->port_num;
    ptys.pnat = this->m_pnat;
    ptys_reg_pack(&ptys, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PTYSRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                             << '#'
            << handler_header                   << ','
            << "reserved_high"                  << ','
            << "an_disable_admin"               << ','
            << "an_disable_cap"                 << ','
            << "force_tx_aba_param"             << ','
            << "local_port"                     << ','
            << "pnat"                           << ','
            << "lp_msb"                         << ','
            << "port_type"                      << ','
            << "proto_mask"                     << ','
            << "an_status"                      << ','
            << "max_port_rate"                  << ','
            << "data_rate_oper"                 << ','
            << "ext_eth_proto_capability"       << ','
            << "eth_proto_capability"           << ','
            << "ib_link_width_capability"       << ','
            << "ib_proto_capability"            << ','
            << "ext_eth_proto_admin"            << ','
            << "eth_proto_admin"                << ','
            << "ib_link_width_admin"            << ','
            << "ib_proto_admin"                 << ','
            << "ext_eth_proto_oper"             << ','
            << "eth_proto_oper"                 << ','
            << "ib_link_width_oper"             << ','
            << "ib_proto_oper"                  << ','
            << "force_kr_frames_cap"            << ','
            << "force_kr_frames_admin"          << ','
            << "connector_type"                 << ','
            << "eth_proto_lp_advertise";
#endif

    IBDIAGNET_RETURN_VOID;

}


void PTYSRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct ptys_reg &ptys = areg.regs.ptys;

    //register fields
    sstream << +ptys.reserved_high                  << ","
            << +ptys.an_disable_admin               << ","
            << +ptys.an_disable_cap                 << ","
            << +ptys.force_tx_aba_param             << ","
            << +ptys.local_port                     << ","
            << +ptys.pnat                           << ","
            << +ptys.lp_msb                         << ","
            << +ptys.port_type                      << ","
            << +ptys.proto_mask                     << ","
            << +ptys.an_status                      << ","
            << +ptys.max_port_rate                  << ","
            << +ptys.data_rate_oper                 << ","
            << +ptys.ext_eth_proto_capability       << ","
            << +ptys.eth_proto_capability           << ","
            << +ptys.ib_link_width_capability       << ","
            << +ptys.ib_proto_capability            << ","
            << +ptys.ext_eth_proto_admin            << ","
            << +ptys.eth_proto_admin                << ","
            << +ptys.ib_link_width_admin            << ","
            << +ptys.ib_proto_admin                 << ","
            << +ptys.ext_eth_proto_oper             << ","
            << +ptys.eth_proto_oper                 << ","
            << +ptys.ib_link_width_oper             << ","
            << +ptys.ib_proto_oper                  << ","
            << +ptys.force_kr_frames_cap            << ","
            << +ptys.force_kr_frames_admin          << ","
            << +ptys.connector_type                 << ","
            << +ptys.eth_proto_lp_advertise         << endl;

    IBDIAGNET_RETURN_VOID;
}

void PTYSRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ptys_reg = &areg.regs.ptys;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************ PPHCR Register functions ***
//*****************************************
PPHCRRegister::PPHCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PPHCR,
                (const unpack_data_func_t)pphcr_reg_unpack,
                ACC_REG_PPHCR_INTERNAL_SECTION_NAME,
                ACC_REG_PPHCR_NAME,
                ACC_REG_PPHCR_FIELDS_NUM,
                NOT_SUPPORT_PPHCR,
                "",
                SUPPORT_SW_CA,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP),
        m_pnat(ACC_REG_PNAT_IB_PORT)
{
}

int PPHCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pphcr_reg pphcr;
    CLEAR_STRUCT(pphcr);
    pphcr.local_port = ((AccRegKeyPort *)p_key)->port_num;
    pphcr.pnat = this->m_pnat;
    pphcr_reg_pack(&pphcr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPHCRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                    << '#'
            << handler_header          << ','
            << "we"                    << ','
            << "lp_msb"                << ','
            << "pnat"                  << ','
            << "local_port"            << ','
            << "active_hist_type"      << ','
            << "hist_type"             << ','
            << "num_of_bins"           << ','
            << "hist_min_measurement"  << ','
            << "hist_max_measurement"  << ','
            << "bin_range_write_mask";

    for (size_t i = 0; i < ARRAY_SIZE(pphcr_reg::bin_range); i++)
         sstream << ',' << "bin_range[" << i << "].low_val"
                 << ',' << "bin_range[" << i << "].high_val";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPHCRRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct pphcr_reg &pphcr = areg.regs.pphcr;

    //register fields
    sstream << +pphcr.we                    << ","
            << +pphcr.lp_msb                << ","
            << +pphcr.pnat                  << ","
            << +pphcr.local_port            << ","
            << +pphcr.active_hist_type      << ","
            << +pphcr.hist_type             << ","
            << +pphcr.num_of_bins           << ","
            << +pphcr.hist_min_measurement  << ","
            << +pphcr.hist_max_measurement  << ","
            << DEC(pphcr.bin_range_write_mask);

    int bin_range_size = ARRAY_SIZE(pphcr.bin_range);
    for (int i = 0; i < bin_range_size; i++) {

         sstream << "," << +pphcr.bin_range[i].low_val
                 << "," << +pphcr.bin_range[i].high_val;
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void PPHCRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node,
                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pphcr_reg = &areg.regs.pphcr;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************ PEUCG Register functions ***
//*****************************************
PEUCGRegister::PEUCGRegister(PhyDiag * phy_diag, string section_name)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_PEUCG,
               (const unpack_data_func_t)peucg_reg_unpack,
               section_name,
               ACC_REG_PEUCG_NAME,
               ACC_REG_PEUCG_FIELDS_NUM,
               NOT_SUPPORT_PEUCG,
               "",
               SUPPORT_SW_CA,
               true,  // dump enabled
               false, // disconnected ports
               VIA_GMP),
      m_pnat(ACC_REG_PNAT_IB_PORT),
      m_cur_mad_builder(NULL)
{
}

void PEUCGRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "unit"             << ","
            << "local_port"       << ","
            << "pnat"             << ","
            << "lp_msb"           << ","
            << "lane"             << ","
            << "status"           << ","
            << "payload_size"     << ","
            << "db"               << ","
            << "clr"              << ","
            << "enum_init"        << ","
            << "num_of_entries"   << ","
            << "db_index";

    for (size_t i = 0; i < ARRAY_SIZE(peucg_reg::page_data); i++)
        sstream << "," << "page_data[" << i << "].address"
                << "," << "page_data[" << i << "].rxtx"
                << "," << "page_data[" << i << "].payload_data";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PEUCGRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct peucg_reg &peucg = areg.regs.peucg;

    //register fields
    sstream << +peucg.unit               << ","
            << +peucg.local_port         << ","
            << +peucg.pnat               << ","
            << +peucg.lp_msb             << ","
            << +peucg.lane               << ","
            << +peucg.status             << ","
            << +peucg.payload_size       << ","
            << +peucg.db                 << ","
            << +peucg.clr                << ","
            << +peucg.enum_init          << ","
            << peucg.num_of_entries      << ","
            << peucg.db_index;

    u_int8_t page_data_size = ARRAY_SIZE(peucg.page_data);
    u_int8_t i = 0;

    for ( ; i < peucg.payload_size; ++i) {
        sstream << "," << peucg.page_data[i].address
                << "," << +peucg.page_data[i].rxtx
                << "," << peucg.page_data[i].payload_data;
    }

    for ( ; i < page_data_size; ++i) {
        sstream << "," << "N/A"
                << "," << "N/A"
                << "," << "N/A";
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//**** PEUCG Version Register functions ***
//*****************************************
PEUCG_Ver_Register::PEUCG_Ver_Register(PhyDiag * phy_diag)
    : PEUCGRegister(phy_diag, ACC_REG_PEUCG_VER_INTERNAL_SECTION_NAME)
{
}

int PEUCG_Ver_Register::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct peucg_reg peucg;
    CLEAR_STRUCT(peucg);

    // for version - all fields are 0, except:
    peucg.pnat = this->m_pnat;
    peucg.payload_size = 1;
    peucg.local_port = 1;

    peucg_reg_pack(&peucg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PEUCG_Ver_Register::HandleData(IBNode *p_node, AccRegKey* p_key, acc_reg_data& reg_data) const
{
    IBDIAGNET_ENTER;

    if (!p_node || !p_node->p_phy_data)
        IBDIAGNET_RETURN_VOID;

    PHYNodeData* phy_data = (PHYNodeData*)p_node->p_phy_data;

    // if we already get uphy version we'll use existing data.
    if (phy_data->uphy)
        IBDIAGNET_RETURN_VOID;

    phy_data->uphy = new PHYNodeData::UPHYData();

    phy_data->uphy->cln.version = new UPHY::Version(reg_data.regs.peucg.page_data[0].payload_data & 0x0f,
            (reg_data.regs.peucg.page_data[0].payload_data & 0xf0) >> 4);

    phy_data->uphy->dln.version = new UPHY::Version(reg_data.regs.peucg.page_data[0].payload_data & 0x0f,
            (reg_data.regs.peucg.page_data[0].payload_data & 0xf0) >> 4);

    phy_data->uphy->cln.dataset = UPHY::DB::instance().get_compatible(
            UPHY::DataSetType_CLN, *phy_data->uphy->cln.version);

    phy_data->uphy->dln.dataset = UPHY::DB::instance().get_compatible(
            UPHY::DataSetType_DLN, *phy_data->uphy->dln.version);

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//****** PEUCG DLN Register functions *****
//*****************************************
PEUCG_DLN_Register::PEUCG_DLN_Register(PhyDiag * phy_diag) :
        PEUCGRegister(phy_diag, ACC_REG_PEUCG_DLN_INTERNAL_SECTION_NAME)
{
}

int PEUCG_DLN_Register::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct peucg_reg peucg;
    CLEAR_STRUCT(peucg);

    UPHY::DataKey *p_cur_key = (UPHY::DataKey*)p_key;

    AccRegKeyPortLane* p_plkey = (AccRegKeyPortLane*)(p_cur_key->base_key());

    peucg.local_port = p_plkey->port_num;
    peucg.lane = p_plkey->lane;
    peucg.pnat = this->m_pnat;

    if (m_cur_mad_builder)
        m_cur_mad_builder->next(&peucg);

    peucg_reg_pack(&peucg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PEUCG_DLN_Register::BuildDB(AccRegHandler* p_handler,
                                list_p_fabric_general_err &phy_errors,
                                ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    for (map_str_pnode::iterator nI = m_phy_diag->GetFabric()->NodeByName.begin();
         nI != m_phy_diag->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            m_phy_diag->SetLastError("DB error - found null node in"  \
                                           " NodeByName map for key = %s",
                                           nI->first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            IBDIAGNET_RETURN(rc);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        PHYNodeData::UPHYData* cur_uphy_data = ((PHYNodeData*)(p_curr_node->p_phy_data))->uphy;
        if (!cur_uphy_data || !cur_uphy_data->dln.dataset)
            continue;

        // go over all ports of this node:
        for (phys_port_t i = 1; i < p_curr_node->numPorts + 1; ++i) {

            IBPort *p_curr_port = p_curr_node->getPort(i);
            //port doesn't exist or down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            if (p_curr_port->isSpecialPort())
                continue;

            for (u_int32_t ln = 0; ln < LANE_NUM; ln++) {

                AccessRegister acc_reg;
                CLEAR_STRUCT(acc_reg);

                uint8_t idx = 0; // for the dln key to be unique.

                this->m_cur_mad_builder = new UPHY::MadBuilder(*cur_uphy_data->dln.dataset, m_phy_diag->GetUphyDumps());

                while (!this->m_cur_mad_builder->empty()) {

                    AccRegKeyPortLane *p_plkey = new AccRegKeyPortLane(
                            p_curr_node->guid_get(),
                            p_curr_port->guid_get(),
                            i,
                            ln,
                            0); // idx in lane - not relevant

                    UPHY::DataKey* p_cur_key = new UPHY::DataKey(p_plkey, idx++);

                    rc = p_handler->SendAccReg(acc_reg_via,
                                               p_curr_node,
                                               i,
                                               p_curr_port->base_lid,
                                               acc_reg,
                                               p_cur_key,
                                               p_progress_bar);

                    if (rc == IBDIAG_ERR_CODE_DB_ERR) {
                        delete this->m_cur_mad_builder;
                        this->m_cur_mad_builder = NULL;
                        IBDIAGNET_RETURN(rc);
                    }
                } // elements blocks

                delete this->m_cur_mad_builder;
                this->m_cur_mad_builder = NULL;
            } // lane
        } // port
    } // node

    IBDIAGNET_RETURN(rc);
}


//*****************************************
//****** PEUCG CLN Register functions *****
//*****************************************
PEUCG_CLN_Register::PEUCG_CLN_Register(PhyDiag * phy_diag, map_akey_areg *pmdr_map) :
        PEUCGRegister(phy_diag, ACC_REG_PEUCG_CLN_INTERNAL_SECTION_NAME),
        p_pmdr_map(pmdr_map)
{
}

int PEUCG_CLN_Register::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct peucg_reg peucg;
    CLEAR_STRUCT(peucg);

    UPHY::DataKey *p_cur_key = (UPHY::DataKey*)p_key;
    AccRegKeyGroup* p_gkey = (AccRegKeyGroup*)(p_cur_key->base_key());

    peucg.unit = 1; // 1 - common pll (CLN)
    peucg.local_port = p_gkey->group_num; // for CLN local_port is pll_index

    if (m_cur_mad_builder)
        m_cur_mad_builder->next(&peucg);

    peucg_reg_pack(&peucg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PEUCG_CLN_Register::BuildDB(AccRegHandler* p_handler,
                                list_p_fabric_general_err &phy_errors,
                                ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    // 1. go over all PMDRs and make a map of node and set of PLLs
    map < u_int64_t , set<u_int8_t> > node2plls;

    for (map_akey_areg::iterator pmdr_iter = p_pmdr_map->begin();
            pmdr_iter != p_pmdr_map->end(); pmdr_iter++) {

        AccRegKeyPort *p_port_key = (AccRegKeyPort*)pmdr_iter->first;
        pmdr_reg pmdr = pmdr_iter->second.regs.pmdr;

        node2plls[p_port_key->node_guid].insert(pmdr.pll_index);
    }

    // 2. go over all nodes and send CLN to all PLLs
    for (map < u_int64_t , set<u_int8_t> >::iterator nodes_iter = node2plls.begin();
            nodes_iter != node2plls.end(); nodes_iter++) {

        IBNode *p_curr_node =
                m_phy_diag->GetFabric()->getNodeByGuid(nodes_iter->first);

        if (!p_curr_node) {
            m_phy_diag->SetLastError("DB error - found null node in "
                    "NodeByName map for key = " U64H_FMT, nodes_iter->first);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        PHYNodeData::UPHYData* cur_uphy_data = ((PHYNodeData*)(p_curr_node->p_phy_data))->uphy;
        if (!cur_uphy_data || !cur_uphy_data->cln.dataset)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        // go over all PLLs of this node:
        for (set<u_int8_t>::iterator plls_iter = nodes_iter->second.begin();
                plls_iter != nodes_iter->second.end(); plls_iter++) {

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            uint8_t idx = 0; // for the cln key to be unique.

            this->m_cur_mad_builder = new UPHY::MadBuilder(*cur_uphy_data->cln.dataset, m_phy_diag->GetUphyDumps());

            while (!this->m_cur_mad_builder->empty()) {

                AccRegKeyGroup *p_gkey = new AccRegKeyGroup(
                        p_curr_node->guid_get(),
                        *plls_iter);

                UPHY::DataKey* p_cur_key = new UPHY::DataKey(p_gkey, idx++);

                rc = p_handler->SendAccReg(acc_reg_via,
                        p_curr_node,
                        0, // port 0
                        lid,
                        acc_reg,
                        p_cur_key,
                        p_progress_bar);

                if (rc == IBDIAG_ERR_CODE_DB_ERR) {
                    delete this->m_cur_mad_builder;
                    this->m_cur_mad_builder = NULL;
                    IBDIAGNET_RETURN(rc);
                }
            }

            delete this->m_cur_mad_builder;
            this->m_cur_mad_builder = NULL;
        }
    }

    IBDIAGNET_RETURN(rc);
}


//*****************************************
//********   SLRegister functions    ******
//*****************************************
SLRegister::SLRegister(PhyDiag * phy_diag,
                       uint32_t reg_id,
                       const unpack_data_func_t p_unpack_func,
                       string sn,
                       string name,
                       uint32_t fn,
                       uint64_t nsb,
                       map_akey_areg *mpein_map,
                       map_akey_areg *mpir_map)
    : Register(phy_diag,
               reg_id,
               p_unpack_func,
               sn,
               name,
               fn,
               nsb),
      p_mpein_map(mpein_map),
      p_mpir_map(mpir_map){}

/*
 * this register is for slrg slrp sltp registers in pci mode.
 * it goes over all valid MPIR queries and get the above registers.
 * p_handler - contain the MPIR map.
 * */
int SLRegister::BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    p_handler->SetHeader(PCI_PORT_LANE_HEADER);

    for(map_akey_areg::iterator mpir_iter = p_mpir_map->begin();
            mpir_iter != p_mpir_map->end(); mpir_iter++) {

        AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)mpir_iter->first;
        mpir_reg mpir = mpir_iter->second.regs.mpir;
        mpein_reg mpein = p_mpein_map->find(p_dpnkey)->second.regs.mpein;

        IBNode *p_curr_node =
                m_phy_diag->GetFabric()->getNodeByGuid(p_dpnkey->node_guid);

        if (!p_curr_node) {
            m_phy_diag->SetLastError("DB error - found null node in "
                    "NodeByName map for key = " U64H_FMT, p_dpnkey->node_guid);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        AccRegVia_t acc_reg_via = this->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        IBPort *p_curr_port = NULL;
        phys_port_t port_num = 1;
        for ( ; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
            p_curr_port = p_curr_node->getPort(port_num);

            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;
            break; // use the first valid port.
        }

        if (!p_curr_port) {
            m_phy_diag->SetLastError("DB error - failed to find valid port for node %s",
                                                p_curr_node->getName().c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        for (u_int8_t lane = 0; lane < mpein.link_width_active; lane++) {

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            AccRegKeyPCIPortLane *p_pciplkey = new AccRegKeyPCIPortLane(
                    p_dpnkey->node_guid,
                    mpir.local_port,
                    lane);

            rc = p_handler->SendAccReg(acc_reg_via,
                                       p_curr_node,
                                       port_num,
                                       p_curr_port->base_lid,
                                       acc_reg,
                                       p_pciplkey,
                                       p_progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR)
                IBDIAGNET_RETURN(rc);

        } // lane loop
    } // mpir loop

    IBDIAGNET_RETURN(rc);
}

bool SLRegister::CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                  export_data_phy_node_t *p_export_data_phy_node,
                                  AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node &&
            p_key->node_guid == p_export_data_phy_node->node_guid)
        IBDIAGNET_RETURN(true);

    IBDIAGNET_RETURN(false);
}


//*****************************************
//***     SLSIR Register functions      ***
//*****************************************
SLSIRRegister::SLSIRRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(phy_diag,
               ACCESS_REGISTER_ID_SLSIR,
               (const unpack_data_func_t)slsir_reg_unpack,
               sn,
               ACC_REG_SLSIR_NAME,
               ACC_REG_SLSIR7NM_FIELDS_NUM,
               NOT_SUPPORT_SLSIR,
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLSIRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slsir_reg slsir;
    CLEAR_STRUCT(slsir);

    slsir.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slsir.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slsir.pnat = m_pnat;

    slsir_reg_pack(&slsir, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLSIRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                     << '#'
            << handler_header           << ','
            << "version"                << ','
            << "local_port"             << ','
            << "pnat"                   << ','
            << "lp_msb"                 << ','
            << "lane"                   << ','
            << "port_type"              << ','
            << "nop_rsunf_error"        << ','
            << "nop_rsovf_error"        << ','
            << "nop_dsunf_error"        << ','
            << "nop_dsovf_error"        << ','
            << "peq_adc_overload"       << ','
            << "feq_adc_overload"       << ','
            << "cdr_error"              << ','
            << "imem_chksm_error"       << ','
            << "rx_ugl_state"           << ','
            << "rx_eom_ugl_state"       << ','
            << "rx_cal_ugl_state"       << ','
            << "rx_eq_ugl_state"        << ','
            << "tx_ugl_state"           << ','
            << "sd_hits_cnt"            << ','
            << "sd_iter_cnt"            << ','
            << "rd_iter_cnt"            << ','
            << "ae_state"               << ','
            << "rx_init_abort_cnt"      << ','
            << "rx_init_done_cnt"       << ','
            << "cdr_abort_cnt"          << ','
            << "cdr_done_cnt"           << ','
            << "cal_abort_cnt"          << ','
            << "cal_done_cnt"           << ','
            << "eq_abort_cnt"           << ','
            << "eq_done_cnt"            << ','
            << "eom_abort_cnt"          << ','
            << "eom_done_cnt";
#endif

    IBDIAGNET_RETURN_VOID;

}

void SLSIRRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct slsir_reg &slsir = areg.regs.slsir;

    sstream << +slsir.version                << ','
            << +slsir.local_port             << ','
            << +slsir.pnat                   << ','
            << +slsir.lp_msb                 << ','
            << +slsir.lane                   << ','
            << +slsir.port_type              << ','
            << +slsir.nop_rsunf_error        << ','
            << +slsir.nop_rsovf_error        << ','
            << +slsir.nop_dsunf_error        << ','
            << +slsir.nop_dsovf_error        << ','
            << +slsir.peq_adc_overload       << ','
            << +slsir.feq_adc_overload       << ','
            << +slsir.cdr_error              << ','
            << +slsir.imem_chksm_error       << ','
            << +slsir.rx_ugl_state           << ','
            << +slsir.rx_eom_ugl_state       << ','
            << +slsir.rx_cal_ugl_state       << ','
            << +slsir.rx_eq_ugl_state        << ','
            << +slsir.tx_ugl_state           << ','
            << +slsir.sd_hits_cnt            << ','
            << +slsir.sd_iter_cnt            << ','
            << +slsir.rd_iter_cnt            << ','
            << +slsir.ae_state               << ','
            << +slsir.rx_init_abort_cnt      << ','
            << +slsir.rx_init_done_cnt       << ','
            << +slsir.cdr_abort_cnt          << ','
            << +slsir.cdr_done_cnt           << ','
            << +slsir.cal_abort_cnt          << ','
            << +slsir.cal_done_cnt           << ','
            << +slsir.eq_abort_cnt           << ','
            << +slsir.eq_done_cnt            << ','
            << +slsir.eom_abort_cnt          << ','
            << +slsir.eom_done_cnt           << endl;

    IBDIAGNET_RETURN_VOID;
}

AccRegVia_t SLSIRRegister::Validation(IBNode *p_node, int &rc)
{
    if (p_node->serdes_version < SerDes_7nm)
        return NOT_SUP_ACC_REG;

    return Register::Validation(p_node, rc);
}

void SLSIRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node,
                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port && this->m_pnat == ACC_REG_PNAT_IB_PORT)
        p_export_data_phy_port->p_slsir_reg[((AccRegKeyPortLane *)p_key)->lane] = &areg.regs.slsir;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* SLRP Register functions ***
//*****************************************
SLRPRegister::SLRPRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(phy_diag,
               ACCESS_REGISTER_ID_SLRP,
               (const unpack_data_func_t)slrp_reg_unpack,
               sn,
               ACC_REG_SLRP_NAME,
               ACC_REG_SLRP_MAX_FIELDS_NUM,
               NOT_SUPPORT_SLRP,
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLRPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slrp_reg slrp;
    CLEAR_STRUCT(slrp);
    slrp.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slrp.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slrp.pnat = m_pnat;
    slrp_reg_pack(&slrp, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLRPRegister::Header_Dump_40nm_28nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                    << ','
            << "version_40nm=0_28nm=1"     << ','
            << "local_port"                << ','
            << "pnat"                      << ','
            << "lp_msb"                    << ','
            << "lane"                      << ','
            << "port_type"                 << ','
            << "ib_sel"                    << ','
            << "dp_sel"                    << ','
            << "dp90sel"                   << ','
            << "mix90phase"                << ','
            << "ffe_tap0"                  << ','
            << "ffe_tap1"                  << ','
            << "ffe_tap2"                  << ','
            << "ffe_tap3"                  << ','
            << "ffe_tap4"                  << ','
            << "ffe_tap5"                  << ','
            << "ffe_tap6"                  << ','
            << "ffe_tap7"                  << ','
            << "ffe_tap8"                  << ','
            << "mixerbias_tap_amp"         << ','
            << "ffe_tap_en"                << ','
            << "ffe_tap_offset0"           << ','
            << "ffe_tap_offset1"           << ','
            << "slicer_offset0"            << ','
            << "mixer_offset0"             << ','
            << "mixer_offset1"             << ','
            << "mixerbgn_inp"              << ','
            << "mixerbgn_inn"              << ','
            << "mixerbgn_refp"             << ','
            << "mixerbgn_refn"             << ','
            << "sel_slicer_lctrl_h"        << ','
            << "sel_slicer_lctrl_l"        << ','
            << "ref_mixer_vreg"            << ','
            << "slicer_gctrl"              << ','
            << "lctrl_input"               << ','
            << "mixer_offset_cm1"          << ','
            << "common_mode"               << ','
            << "mixer_offset_cm0"          << ','
            << "slicer_offset_cm"          ;

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Header_Dump_16nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                                   << ','
            << "version_16nm=3"                           << ','
            << "local_port"                               << ','
            << "pnat"                                     << ','
            << "lp_msb"                                   << ','
            << "lane"                                     << ','
            << "port_type"                                << ','
            << "mixer_offset_up"                          << ','
            << "mixer_offset_down"                        << ','
            << "sel_enc"                                  << ','
            << "dp_sel"                                   << ','
            << "sel_ref_in_mixer_biasgen_up"              << ','
            << "sel_ref_in_mixer_biasgen_mid"             << ','
            << "sel_ref_in_mixer_biasgen_down"            << ','
            << "sel_ref_in_mixer_biasgen_eo"              << ','
            << "sel_ref_out_mixer_biasgen_up"             << ','
            << "sel_ref_out_mixer_biasgen_mid"            << ','
            << "sel_ref_out_mixer_biasgen_down"           << ','
            << "sel_ref_out_mixer_biasgen_eo"             << ','
            << "sel_amp_mixer_biasgen_up"                 << ','
            << "sel_amp_mixer_biasgen_mid"                << ','
            << "sel_amp_mixer_biasgen_down"               << ','
            << "sel_amp_mixer_biasgen_eo"                 << ','
            << "mixer_offset_eye_scan"                    << ','
            << "gctrl_bin_bgn0_n"                         << ','
            << "gctrl_bin_bgn0_p"                         << ','
            << "gctrl_bin_bgn1_n"                         << ','
            << "gctrl_bin_bgn1_p"                         << ','
            << "sel_digital_gctrl"                        << ','
            << "sel_bias2ctle"                            << ','
            << "selgc_ref_op0_alev_ctle"                  << ','
            << "selgc_ref_op1_alev_ctle"                  << ','
            << "sel_bias_en_0"                            << ','
            << "sel_bias_en_1"                            << ','
            << "sel_bias_en_2"                            << ','
            << "sel_bias_en_3"                            << ','
            << "sel_bias_en_4"                            << ','
            << "sel_bias_en_5"                            << ','
            << "sel_bias_en_6"                            << ','
            << "sel_gctrln_en_0"                          << ','
            << "sel_gctrln_en_1"                          << ','
            << "sel_gctrln_en_2"                          << ','
            << "sel_gctrln_en_3"                          << ','
            << "sel_gctrln_en_4"                          << ','
            << "sel_gctrln_en_5"                          << ','
            << "sel_gctrln_en_6"                          << ','
            << "sel_gctrlp_en_0"                          << ','
            << "sel_gctrlp_en_1"                          << ','
            << "sel_gctrlp_en_2"                          << ','
            << "sel_gctrlp_en_3"                          << ','
            << "sel_gctrlp_en_4"                          << ','
            << "sel_gctrlp_en_5"                          << ','
            << "sel_gctrlp_en_6"                          << ','
            << "crnt_bgn_offset_p1"                       << ','
            << "crnt_bgn_offset_n1"                       << ','
            << "crnt_bgn_offset_p0"                       << ','
            << "crnt_bgn_offset_n0"                       ;

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                       << ','
            << "version_7nm=4"                << ','
            << "local_port"                   << ','
            << "pnat"                         << ','
            << "lp_msb"                       << ','
            << "lane"                         << ','
            << "port_type"                    << ','
            << "adc_recording_admin"          << ','
            << "adc_recording_status"         << ','
            << "adc_rocording_lanes"          << ','
            << "edge_vos_ccal_en"             << ','
            << "adc_gain_shift_auto"          << ','
            << "ccal_mode"                    << ','
            << "ctle_override_ctrl"           << ','
            << "vga_override_ctrl"            << ','
            << "adc_vos_override_ctrl"        << ','
            << "adc_gain_override_ctrl"       << ','
            << "phos_override_ctrl"           << ','
            << "cal_error_cnt"                << ','
            << "ccal_state"                   << ','
            << "ccal_op"                      << ','
            << "phos"                         << ','
            << "vga_vos1"                     << ','
            << "vga_vos0"                     << ','
            << "ctle_vos0"                    ;
    for (size_t i = 0; i < ARRAY_SIZE(slrp_7nm::adc_vos); ++i)
        sstream << ',' << "adc_vos[" << i << "]";
    for (size_t i = 0; i < ARRAY_SIZE(slrp_7nm::adc_gos); ++i)
        sstream << ',' << "adc_gos[" << i << "]";
    
    sstream << ',' 
            << "phos3"                        << ','
            << "phos2"                        << ','
            << "phos1"                        << ','
            << "phos0"                        << ','
            << "phos7"                        << ','
            << "phos6"                        << ','
            << "phos5"                        << ','
            << "phos4"                        ;
            
    IBDIAG_RETURN_VOID;
}

void SLRPRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                       << ','
            << "version_5nm=5"                << ','
            << "local_port"                   << ','
            << "pnat"                         << ','
            << "lp_msb"                       << ','
            << "lane"                         << ','
            << "port_type"                    << ','
            << "ctle_vos_val"                 << ','
            << "ctle_vos_stg1"                << ','
            << "ctle_vos_aux"                 << ','
            << "adc_recording_admin"          << ','
            << "adc_recording_status"         << ','
            << "fixed_val"                    << ','
            << "fine_val"                     << ','
            << "coarse_val";

    for (int i = 0; i < 16; ++i)
        sstream << ',' << "tah_vos[" << i << ']';

    for (int i = 0; i < 16; ++i)
        sstream << ',' << "adc_vos[" << i << ']';

    for (int i = 0; i < 16; ++i)
        sstream << ',' << "adc_gos[" << i << ']';

    IBDIAG_RETURN_VOID;
}

void SLRPRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_40nm_28nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct slrp_reg & slrp = areg.regs.slrp;

    sstream << +slrp.status           << ','
            << +slrp.version          << ','
            << +slrp.local_port       << ','
            << +slrp.pnat             << ','
            << +slrp.lp_msb           << ','
            << +slrp.lane             << ','
            << +slrp.port_type        << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (slrp.version) {

        case 0: // 40nm
        case 1: // 28nm
            Dump_40nm_28nm(slrp, sstream);
            break;

        case 3: // 16nm
            Dump_16nm(slrp, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(slrp, sstream);
            break;

        case 5: // 5nm_scc
            Dump_5nm_scc(slrp, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +slrp.version << '"';
            for (int i = 0; i < ACC_REG_SLRP_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLRP: %d, on node: " U64H_FMT ".\n",
                        slrp.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << endl;
    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Dump_40nm_28nm(const struct slrp_reg &slrp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrp_40nm_28nm slrp_40_28;
    slrp_40nm_28nm_unpack(&slrp_40_28, slrp.page_data.slrp_data_set.data);

    sstream << +slrp_40_28.ib_sel                    << ','
            << +slrp_40_28.dp_sel                    << ','
            << +slrp_40_28.dp90sel                   << ','
            << +slrp_40_28.mix90phase                << ','
            << +slrp_40_28.ffe_tap0                  << ','
            << +slrp_40_28.ffe_tap1                  << ','
            << +slrp_40_28.ffe_tap2                  << ','
            << +slrp_40_28.ffe_tap3                  << ','
            << +slrp_40_28.ffe_tap4                  << ','
            << +slrp_40_28.ffe_tap5                  << ','
            << +slrp_40_28.ffe_tap6                  << ','
            << +slrp_40_28.ffe_tap7                  << ','
            << +slrp_40_28.ffe_tap8                  << ','
            << +slrp_40_28.mixerbias_tap_amp         << ','
            << +slrp_40_28.ffe_tap_en                << ','
            << +slrp_40_28.ffe_tap_offset0           << ','
            << +slrp_40_28.ffe_tap_offset1           << ','
            << +slrp_40_28.slicer_offset0            << ','
            << +slrp_40_28.mixer_offset0             << ','
            << +slrp_40_28.mixer_offset1             << ','
            << +slrp_40_28.mixerbgn_inp              << ','
            << +slrp_40_28.mixerbgn_inn              << ','
            << +slrp_40_28.mixerbgn_refp             << ','
            << +slrp_40_28.mixerbgn_refn             << ','
            << +slrp_40_28.sel_slicer_lctrl_h        << ','
            << +slrp_40_28.sel_slicer_lctrl_l        << ','
            << +slrp_40_28.ref_mixer_vreg            << ','
            << +slrp_40_28.slicer_gctrl              << ','
            << +slrp_40_28.lctrl_input               << ','
            << +slrp_40_28.mixer_offset_cm1          << ','
            << +slrp_40_28.common_mode               << ','
            << +slrp_40_28.mixer_offset_cm0          << ','
            << +slrp_40_28.slicer_offset_cm          ;

    for (int i = 0; i < ACC_REG_SLRP_MAX_FIELDS_NUM - ACC_REG_SLRP28NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Dump_16nm(const struct slrp_reg &slrp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;
    slrp_16nm slrp_16;
    slrp_16nm_unpack(&slrp_16, slrp.page_data.slrp_data_set.data);

    sstream << +slrp_16.mixer_offset_up                          << ','
            << +slrp_16.mixer_offset_down                        << ','
            << +slrp_16.sel_enc                                  << ','
            << +slrp_16.dp_sel                                   << ','
            << +slrp_16.sel_ref_in_mixer_biasgen_up              << ','
            << +slrp_16.sel_ref_in_mixer_biasgen_mid             << ','
            << +slrp_16.sel_ref_in_mixer_biasgen_down            << ','
            << +slrp_16.sel_ref_in_mixer_biasgen_eo              << ','
            << +slrp_16.sel_ref_out_mixer_biasgen_up             << ','
            << +slrp_16.sel_ref_out_mixer_biasgen_mid            << ','
            << +slrp_16.sel_ref_out_mixer_biasgen_down           << ','
            << +slrp_16.sel_ref_out_mixer_biasgen_eo             << ','
            << +slrp_16.sel_amp_mixer_biasgen_up                 << ','
            << +slrp_16.sel_amp_mixer_biasgen_mid                << ','
            << +slrp_16.sel_amp_mixer_biasgen_down               << ','
            << +slrp_16.sel_amp_mixer_biasgen_eo                 << ','
            << +slrp_16.mixer_offset_eye_scan                    << ','
            << +slrp_16.gctrl_bin_bgn0_n                         << ','
            << +slrp_16.gctrl_bin_bgn0_p                         << ','
            << +slrp_16.gctrl_bin_bgn1_n                         << ','
            << +slrp_16.gctrl_bin_bgn1_p                         << ','
            << +slrp_16.sel_digital_gctrl                        << ','
            << +slrp_16.sel_bias2ctle                            << ','
            << +slrp_16.selgc_ref_op0_alev_ctle                  << ','
            << +slrp_16.selgc_ref_op1_alev_ctle                  << ','
            << +slrp_16.sel_bias_en_0                            << ','
            << +slrp_16.sel_bias_en_1                            << ','
            << +slrp_16.sel_bias_en_2                            << ','
            << +slrp_16.sel_bias_en_3                            << ','
            << +slrp_16.sel_bias_en_4                            << ','
            << +slrp_16.sel_bias_en_5                            << ','
            << +slrp_16.sel_bias_en_6                            << ','
            << +slrp_16.sel_gctrln_en_0                          << ','
            << +slrp_16.sel_gctrln_en_1                          << ','
            << +slrp_16.sel_gctrln_en_2                          << ','
            << +slrp_16.sel_gctrln_en_3                          << ','
            << +slrp_16.sel_gctrln_en_4                          << ','
            << +slrp_16.sel_gctrln_en_5                          << ','
            << +slrp_16.sel_gctrln_en_6                          << ','
            << +slrp_16.sel_gctrlp_en_0                          << ','
            << +slrp_16.sel_gctrlp_en_1                          << ','
            << +slrp_16.sel_gctrlp_en_2                          << ','
            << +slrp_16.sel_gctrlp_en_3                          << ','
            << +slrp_16.sel_gctrlp_en_4                          << ','
            << +slrp_16.sel_gctrlp_en_5                          << ','
            << +slrp_16.sel_gctrlp_en_6                          << ','
            << +slrp_16.crnt_bgn_offset_p1                       << ','
            << +slrp_16.crnt_bgn_offset_n1                       << ','
            << +slrp_16.crnt_bgn_offset_p0                       << ','
            << +slrp_16.crnt_bgn_offset_n0                       ;

    for (int i = 0; i < ACC_REG_SLRP_MAX_FIELDS_NUM - ACC_REG_SLRP16NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Dump_7nm(const struct slrp_reg &slrp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;
    slrp_7nm slrp_7;
    slrp_7nm_unpack(&slrp_7, slrp.page_data.slrp_data_set.data);

    sstream << +slrp_7.adc_recording_admin          << ','
            << +slrp_7.adc_recording_status         << ','
            << +slrp_7.adc_rocording_lanes          << ','
            << +slrp_7.edge_vos_ccal_en             << ','
            << +slrp_7.adc_gain_shift_auto          << ','
            << +slrp_7.ccal_mode                    << ','
            << +slrp_7.ctle_override_ctrl           << ','
            << +slrp_7.vga_override_ctrl            << ','
            << +slrp_7.adc_vos_override_ctrl        << ','
            << +slrp_7.adc_gain_override_ctrl       << ','
            << +slrp_7.phos_override_ctrl           << ','
            << +slrp_7.cal_error_cnt                << ','
            << +slrp_7.ccal_state                   << ','
            << +slrp_7.ccal_op                      << ','
            << +slrp_7.phos                         << ','
            << +slrp_7.vga_vos1                     << ','
            << +slrp_7.vga_vos0                     << ','
            << +slrp_7.ctle_vos0                    ;
    for (size_t i = 0; i < ARRAY_SIZE(slrp_7.adc_vos); ++i) {
        sstream << ',' << +slrp_7.adc_vos[i];
    }
    for (size_t i = 0; i < ARRAY_SIZE(slrp_7.adc_gos); ++i) {
        sstream << ',' << +slrp_7.adc_gos[i];
    }
    sstream << ','
            << +slrp_7.phos3                        << ','
            << +slrp_7.phos2                        << ','
            << +slrp_7.phos1                        << ','
            << +slrp_7.phos0                        << ','
            << +slrp_7.phos7                        << ','
            << +slrp_7.phos6                        << ','
            << +slrp_7.phos5                        << ','
            << +slrp_7.phos4                        ;

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Dump_5nm_scc(const struct slrp_reg &slrp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;
    slrp_5nm slrp_5;
    slrp_5nm_unpack(&slrp_5, slrp.page_data.slrp_data_set.data);

    sstream << +slrp_5.ctle_vos_val                  << ','
            << +slrp_5.ctle_vos_stg1                 << ','
            << +slrp_5.ctle_vos_aux                  << ','
            << +slrp_5.adc_recording_admin           << ','
            << +slrp_5.adc_recording_status          << ','
            << +slrp_5.fixed_val                     << ','
            << +slrp_5.fine_val                      << ','
            << +slrp_5.coarse_val                    ;

    for (int i = 0; i < 16; ++i) {
        sstream << ',' << +slrp_5.tah_vos[i];
    }
    for (int i = 0; i < 16; ++i) {
        sstream << ',' << +slrp_5.adc_vos[i];
    }
    for (int i = 0; i < 16; ++i) {
        sstream << ',' << +slrp_5.adc_gos[i];
    }

    for (int i = 0; i < ACC_REG_SLRP_MAX_FIELDS_NUM - ACC_REG_SLRP5NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if ((p_export_data_phy_port && this->m_pnat == ACC_REG_PNAT_IB_PORT) ||
            (p_export_data_phy_node && this->m_pnat == ACC_REG_PNAT_OOB_PORT)) {
        struct slrp_reg *p_slrp_reg = new struct slrp_reg(areg.regs.slrp);

        switch (p_slrp_reg->version) {
            case 0: // 40nm
            case 1: // 28nm
                slrp_40nm_28nm_unpack(&p_slrp_reg->page_data.slrp_40nm_28nm, areg.regs.slrp.page_data.slrp_data_set.data);
                break;
            case 3: // 16nm
                slrp_16nm_unpack(&p_slrp_reg->page_data.slrp_16nm, areg.regs.slrp.page_data.slrp_data_set.data);
                break;
            case 4: // 7nm
                slrp_7nm_unpack(&p_slrp_reg->page_data.slrp_7nm, areg.regs.slrp.page_data.slrp_data_set.data);
                break;
            case 5: // 5nm
                slrp_5nm_unpack(&p_slrp_reg->page_data.slrp_5nm, areg.regs.slrp.page_data.slrp_data_set.data);
                break;
            default:
                break;
        }

        if (this->m_pnat == ACC_REG_PNAT_IB_PORT) {
            p_export_data_phy_port->p_slrp_reg[((AccRegKeyPortLane *)p_key)->lane] = p_slrp_reg;
        } else if (this->m_pnat == ACC_REG_PNAT_OOB_PORT) {
            AccRegKeyPCIPortLane *p_pcipl_key = (AccRegKeyPCIPortLane *)p_key;
            p_export_data_phy_node->p_slrp_reg[p_pcipl_key->port_num] = p_slrp_reg;
        }
    }

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* SLTP Register functions ***
//*****************************************
SLTPRegister::SLTPRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(phy_diag,
               ACCESS_REGISTER_ID_SLTP,
               (const unpack_data_func_t)sltp_reg_unpack,
               sn,
               ACC_REG_SLTP_NAME,
               ACC_REG_SLTP_MAX_FIELDS_NUM,
               NOT_SUPPORT_SLTP,
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLTPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct sltp_reg sltp;
    CLEAR_STRUCT(sltp);
    sltp.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    sltp.lane = ((AccRegKeyPortLane *)p_key)->lane;
    sltp.pnat = m_pnat;
    sltp_reg_pack(&sltp, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLTPRegister::Header_Dump_40nm_28nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                  << ','
            << "version_40nm=0_28nm=1"   << ','
            << "local_port"              << ','
            << "pnat"                    << ','
            << "tx_policy"               << ','
            << "conf_mod"                << ','
            << "lane"                    << ','
            << "lane_speed"              << ','
            << "port_type"               << ','
            << "c_db"                    << ','            
            << "polarity"                << ','
            << "ob_tap0"                 << ','
            << "ob_tap1"                 << ','
            << "ob_tap2"                 << ','
            << "ob_leva"                 << ','
            << "ob_preemp_mode"          << ','
            << "ob_reg"                  << ','
            << "ob_bias"                 << ','
            << "ob_norm"                 << ','
            << "ob_bad_stat";

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Header_Dump_16nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"             << ','
            << "version_16nm=3"     << ','
            << "local_port"         << ','
            << "pnat"               << ','
            << "tx_policy"          << ','
            << "conf_mod"           << ','
            << "lane"               << ','
            << "lane_speed"         << ','
            << "port_type"          << ','
            << "c_db"               << ','
            << "pre_2_tap"          << ','
            << "pre_tap"            << ','
            << "main_tap"           << ','
            << "post_tap"           << ','
            << "ob_m2lp"            << ','
            << "ob_amp"             << ','
            << "ob_alev_out"        << ','
            << "ob_bad_stat"        << ','
            << "obplev"             << ','
            << "obnlev"             << ','
            << "regn_bfm1p"         << ','
            << "regp_bfm1n"         << ','
            << "blev"               << ','
            << "tx_alev"            << ','
            << "alev_plus_bfm2"     << ','
            << "alev_minus_bfm2";

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"             << ','
            << "version_7nm=4"      << ','
            << "local_port"         << ','
            << "pnat"               << ','
            << "tx_policy"          << ','
            << "conf_mod"           << ','
            << "lane"               << ','
            << "lane_speed"         << ','
            << "port_type"          << ','
            << "c_db"               << ','
            << "vs_peer_db"         << ','
            << "drv_amp"            << ','
            << "ob_bad_stat"        << ','
            << "fir_pre3"           << ','
            << "fir_pre2"           << ','
            << "fir_pre1"           << ','
            << "fir_main"           << ','
            << "fir_post1";

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"             << ','
            << "version_5nm=5"      << ','
            << "local_port"         << ','
            << "pnat"               << ','
            << "tx_policy"          << ','
            << "conf_mod"           << ','
            << "lane"               << ','
            << "lane_speed"         << ','
            << "port_type"          << ','
            << "c_db"               << ','
            << "ob_bad_stat"        << ','
            << "tap3"               << ','
            << "tap2"               << ','
            << "tap1"               << ','
            << "tap0"               << ','
            << "tap7"               << ','
            << "tap6"               << ','
            << "tap5"               << ','
            << "tap4"               << ','
            << "tap11"              << ','
            << "tap10"              << ','
            << "tap9"               << ','
            << "tap8"               << ','
            << "upper_eye_amp"      << ','
            << "middle_eye_amp"     << ','
            << "lower_eye_amp";

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_40nm_28nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct sltp_reg &sltp = areg.regs.sltp;

    sstream << hex
            << "0x" << +sltp.status        << ','
            << "0x" << +sltp.version       << ','
            << "0x" << +sltp.local_port    << ','
            << "0x" << +sltp.pnat          << ','
            << "0x" << +sltp.tx_policy     << ','
            << "0x" << +sltp.conf_mod      << ','
            << "0x" << +sltp.lane          << ','
            << "0x" << +sltp.lane_speed    << ','
            << "0x" << +sltp.port_type     << ','
            << "0x" << +sltp.c_db          << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (sltp.version) {

        case 0: // 40nm
        case 1: // 28nm
            Dump_40nm_28nm(sltp, sstream);
            break;

        case 3: // 16nm
            Dump_16nm(sltp, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(sltp, sstream);
            break;
        case 5: // 5nm
            Dump_5nm(sltp, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +sltp.version << '"';
            for (int i = 0; i < ACC_REG_SLTP_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLTP: %d, on node: " U64H_FMT ".\n",
                        sltp.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Dump_40nm_28nm(const struct sltp_reg &sltp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sltp_28nm_40nm sltp_40_28;
    sltp_28nm_40nm_unpack(&sltp_40_28, sltp.page_data.sltp_data_set.data);

    sstream << "0x" << +sltp_40_28.polarity           << ','
            << "0x" << +sltp_40_28.ob_tap0            << ','
            << "0x" << +sltp_40_28.ob_tap1            << ','
            << "0x" << +sltp_40_28.ob_tap2            << ','
            << "0x" << +sltp_40_28.ob_leva            << ','
            << "0x" << +sltp_40_28.ob_preemp_mode     << ','
            << "0x" << +sltp_40_28.ob_reg             << ','
            << "0x" << +sltp_40_28.ob_bias            << ','
            << "0x" << +sltp_40_28.ob_norm            << ','
            << "0x" << +sltp_40_28.ob_bad_stat        ;

    for (int i = 0; i < ACC_REG_SLTP_MAX_FIELDS_NUM - ACC_REG_SLTP28NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Dump_16nm(const struct sltp_reg &sltp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sltp_16nm sltp_16;
    sltp_16nm_unpack(&sltp_16, sltp.page_data.sltp_data_set.data);

    sstream << "0x" << +sltp_16.pre_2_tap          << ','
            << "0x" << +sltp_16.pre_tap            << ','
            << "0x" << +sltp_16.main_tap           << ','
            << "0x" << +sltp_16.post_tap           << ','
            << "0x" << +sltp_16.ob_m2lp            << ','
            << "0x" << +sltp_16.ob_amp             << ','
            << "0x" << +sltp_16.ob_alev_out        << ','
            << "0x" << +sltp_16.ob_bad_stat        << ','
            << "0x" << +sltp_16.obplev             << ','
            << "0x" << +sltp_16.obnlev             << ','
            << "0x" << +sltp_16.regn_bfm1p         << ','
            << "0x" << +sltp_16.regp_bfm1n         << ','
            << "0x" << +sltp_16.blev               << ','
            << "0x" << +sltp_16.tx_alev            << ','
            << "0x" << +sltp_16.alev_plus_bfm2     << ','
            << "0x" << +sltp_16.alev_minus_bfm2    ;

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Dump_7nm(const struct sltp_reg &sltp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sltp_7nm sltp_7;
    sltp_7nm_unpack(&sltp_7, sltp.page_data.sltp_data_set.data);

    sstream << "0x" << +sltp_7.vs_peer_db  << ','
            << "0x" << +sltp_7.drv_amp     << ','
            << "0x" << +sltp_7.ob_bad_stat << ','
            << "0x" << +sltp_7.fir_pre3    << ','
            << "0x" << +sltp_7.fir_pre2    << ','
            << "0x" << +sltp_7.fir_pre1    << ','
            << "0x" << +sltp_7.fir_main    << ','
            << "0x" << +sltp_7.fir_post1   ;

    for (int i = 0; i < ACC_REG_SLTP_MAX_FIELDS_NUM - ACC_REG_SLTP7NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Dump_5nm(const struct sltp_reg &sltp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sltp_5nm sltp_5;
    sltp_5nm_unpack(&sltp_5, sltp.page_data.sltp_data_set.data);

    sstream << "0x" << +sltp_5.ob_bad_stat         << ','
            << "0x" << +sltp_5.tap3                << ','
            << "0x" << +sltp_5.tap2                << ','
            << "0x" << +sltp_5.tap1                << ','
            << "0x" << +sltp_5.tap0                << ','
            << "0x" << +sltp_5.tap7                << ','
            << "0x" << +sltp_5.tap6                << ','
            << "0x" << +sltp_5.tap5                << ','
            << "0x" << +sltp_5.tap4                << ','
            << "0x" << +sltp_5.tap11               << ','
            << "0x" << +sltp_5.tap10               << ','
            << "0x" << +sltp_5.tap9                << ','
            << "0x" << +sltp_5.tap8                << ','
            << "0x" << +sltp_5.upper_eye_amp       << ','
            << "0x" << +sltp_5.middle_eye_amp      << ','
            << "0x" << +sltp_5.lower_eye_amp;

#if ACC_REG_SLTP_MAX_FIELDS_NUM < ACC_REG_SLTP5NM_FIELDS_NUM
#error "Number of field of SLTP 5nm more of max fields of SLTP access register"
#endif

#if ACC_REG_SLTP_MAX_FIELDS_NUM - ACC_REG_SLTP5NM_FIELDS_NUM
    for (int i = 0; i < ACC_REG_SLTP_MAX_FIELDS_NUM - ACC_REG_SLTP5NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if ((p_export_data_phy_port && this->m_pnat == ACC_REG_PNAT_IB_PORT) ||
            (p_export_data_phy_node && this->m_pnat == ACC_REG_PNAT_OOB_PORT)) {
        struct sltp_reg *p_sltp_reg = new struct sltp_reg(areg.regs.sltp);

        switch (p_sltp_reg->version) {
            case 0: // 40nm
            case 1: // 28nm
                sltp_28nm_40nm_unpack(&p_sltp_reg->page_data.sltp_28nm_40nm, areg.regs.sltp.page_data.sltp_data_set.data);
                break;
            case 3: // 16nm
                sltp_16nm_unpack(&p_sltp_reg->page_data.sltp_16nm, areg.regs.sltp.page_data.sltp_data_set.data);
                break;
            case 4: // 7nm
                sltp_7nm_unpack(&p_sltp_reg->page_data.sltp_7nm, areg.regs.sltp.page_data.sltp_data_set.data);
                break;
            case 5: // 5nm
                sltp_5nm_unpack(&p_sltp_reg->page_data.sltp_5nm, areg.regs.sltp.page_data.sltp_data_set.data);
                break;
            default:
                break;
        }

        if (this->m_pnat == ACC_REG_PNAT_IB_PORT) {
            p_export_data_phy_port->p_sltp_reg[((AccRegKeyPortLane *)p_key)->lane] = p_sltp_reg;
        } else if (this->m_pnat == ACC_REG_PNAT_OOB_PORT) {
            AccRegKeyPCIPortLane *p_pcipl_key = (AccRegKeyPCIPortLane *)p_key;
            p_export_data_phy_node->p_sltp_reg[p_pcipl_key->port_num] = p_sltp_reg;
        }
    }

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* SLRG Register functions ***
//*****************************************
SLRGRegister::SLRGRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(phy_diag,
               ACCESS_REGISTER_ID_SLRG,
               (const unpack_data_func_t)slrg_reg_unpack,
               sn,
               ACC_REG_SLRG_NAME,
               ACC_REG_SLRG_MAX_FIELDS_NUM,
               NOT_SUPPORT_SLRG,
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLRGRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slrg_reg slrg;
    CLEAR_STRUCT(slrg);

    slrg.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slrg.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slrg.pnat = m_pnat;
    slrg_reg_pack(&slrg, data);

    if (p_node &&
            (p_node->serdes_version == SerDes_7nm || p_node->serdes_version == SerDes_5nm)) {
        slrg.page_data.slrg_7nm.fom_measurment = 0xF;
        slrg_7nm_pack(&slrg.page_data.slrg_7nm, data + 4); // +4 - packed SLRG header size
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLRGRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
     IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_40nm_28nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Header_Dump_16nm(stringstream &sstream) const
{
    IBDIAG_ENTER;

    sstream << "status"                << ','
            << "version_16nm=3"        << ','
            << "local_port"            << ','
            << "pnat"                  << ','
            << "lp_msb"                << ','
            << "lane"                  << ','
            << "port_type"             << ','
            << "test_mode"             << ','
            << "grade_lane_speed"      << ','
            << "grade_version"         << ','
            << "grade"                 << ','
            << "height_eo_pos_up"      << ','
            << "height_eo_neg_up"      << ','
            << "phase_eo_pos_up"       << ','
            << "phase_eo_neg_up"       << ','
            << "height_eo_pos_mid"     << ','
            << "height_eo_neg_mid"     << ','
            << "phase_eo_pos_mid"      << ','
            << "phase_eo_neg_mid"      << ','
            << "height_eo_pos_low"     << ','
            << "height_eo_neg_low"     << ','
            << "phase_eo_pos_low"      << ','
            << "phase_eo_neg_low"      << ','
            << "offset_units"          << ','
            << "phase_units"           << ','
            << "height_grade_type"     << ','
            << "phase_grade_type"      << ','
            << "mid_eye_grade"         << ','
            << "up_eye_grade"          << ','
            << "dn_eye_grade";

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Header_Dump_40nm_28nm(stringstream &sstream) const
{
    IBDIAG_ENTER;

    sstream << "status"                   << ','
            << "version_40nm=0_28nm=1"    << ','
            << "local_port"               << ','
            << "pnat"                     << ','
            << "lp_msb"                   << ','
            << "lane"                     << ','
            << "port_type"                << ','
            << "test_mode"                << ','
            << "grade_lane_speed"         << ','
            << "grade_version"            << ','
            << "grade"                    << ','
            << "offset_units"             << ','
            << "phase_units"              << ','
            << "height_grade_type"        << ','
            << "height_grade"             << ','
            << "height_dz"                << ','
            << "height_dv"                << ','
            << "height_sigma"             << ','
            << "height_eo_pos"            << ','
            << "height_eo_neg"            << ','
            << "phase_grade_type"         << ','
            << "phase_grade"              << ','
            << "phase_eo_pos"             << ','
            << "phase_eo_neg"             << ','
            << "ffe_set_tested"           << ','
            << "test_errors_per_lane";

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                   << ','
            << "version_7nm=4"            << ','
            << "local_port"               << ','
            << "pnat"                     << ','
            << "lp_msb"                   << ','
            << "lane"                     << ','
            << "port_type"                << ','
            << "test_mode"                << ','
            << "fom_measurment"           << ','
            << "fom_mode"                 << ','
            << "initial_fom"              << ','
            << "last_fom"                 << ','
            << "upper_eye"                << ','
            << "mid_eye"                  << ','
            << "lower_eye"                << ','
            << "fom_ext_conf"             << ','
            << "fom_ext_conf_cap"         << ','
            << "nblks_max"                << ','
            << "nerrs_min";

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                   << ','
            << "version_5nm=5"            << ','
            << "local_port"               << ','
            << "pnat"                     << ','
            << "lp_msb"                   << ','
            << "lane"                     << ','
            << "port_type"                << ','
            << "test_mode"                << ','
            << "fom_measurment"           << ','
            << "fom_mode"                 << ','
            << "initial_fom"              << ','
            << "last_fom"                 << ','
            << "upper_eye"                << ','
            << "mid_eye"                  << ','
            << "lower_eye"                << ','
            << "fom_ext_conf"             << ','
            << "fom_ext_conf_cap"         << ','
            << "nblks_max"                << ','
            << "nerrs_min";

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct slrg_reg & slrg = areg.regs.slrg;

    sstream << +slrg.status         << ','
            << +slrg.version        << ','
            << +slrg.local_port     << ','
            << +slrg.pnat           << ','
            << +slrg.lp_msb         << ','
            << +slrg.lane           << ','
            << +slrg.port_type      << ','
            << +slrg.test_mode      << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (slrg.version) {

    case 0: // 40nm
    case 1: // 28nm
        Dump_40nm_28nm(slrg, sstream);
        break;

    case 3: // 16nm
        Dump_16nm(slrg, sstream);
        break;

    case 4: // 7nm
        Dump_7nm(slrg, sstream);
        break;

    case 5: // 5nm
        Dump_5nm(slrg, sstream);
        break;

    default:
        static bool should_print = true;

        sstream << '"' << "Unknown version " << +slrg.version << '"';
        for (int i = 0; i < ACC_REG_SLRG_MAX_FIELDS_NUM - 1; ++i)
            sstream << ",NA"; // add NAs to fill the empty fields

        if (should_print) {
            WARN_PRINT("Unknown version for SLRG: %d, on node: " U64H_FMT ".\n",
                    slrg.version, key.node_guid);
            should_print = false;
        }
    }

    sstream << endl;
    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Dump_40nm_28nm(const struct slrg_reg &slrg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrg_40nm_28nm slrg_40_28;
    slrg_40nm_28nm_unpack(&slrg_40_28, slrg.page_data.slrg_data_set.data);

    sstream << +slrg_40_28.grade_lane_speed         << ','
            << +slrg_40_28.grade_version            << ','
            << +slrg_40_28.grade                    << ','
            << +slrg_40_28.offset_units             << ','
            << +slrg_40_28.phase_units              << ','
            << +slrg_40_28.height_grade_type        << ','
            << +slrg_40_28.height_grade             << ','
            << +slrg_40_28.height_dz                << ','
            << +slrg_40_28.height_dv                << ','
            << +slrg_40_28.height_sigma             << ','
            << +slrg_40_28.height_eo_pos            << ','
            << +slrg_40_28.height_eo_neg            << ','
            << +slrg_40_28.phase_grade_type         << ','
            << +slrg_40_28.phase_grade              << ','
            << +slrg_40_28.phase_eo_pos             << ','
            << +slrg_40_28.phase_eo_neg             << ','
            << +slrg_40_28.ffe_set_tested           << ','
            << +slrg_40_28.test_errors_per_lane     ;

    for (int i = 0; i < ACC_REG_SLRG_MAX_FIELDS_NUM - ACC_REG_SLRG28NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Dump_16nm(const struct slrg_reg &slrg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrg_16nm slrg_16;
    slrg_16nm_unpack(&slrg_16, slrg.page_data.slrg_data_set.data);

    sstream << +slrg_16.grade_lane_speed      << ','
            << +slrg_16.grade_version         << ','
            << +slrg_16.grade                 << ','
            << +slrg_16.height_eo_pos_up      << ','
            << +slrg_16.height_eo_neg_up      << ','
            << +slrg_16.phase_eo_pos_up       << ','
            << +slrg_16.phase_eo_neg_up       << ','
            << +slrg_16.height_eo_pos_mid     << ','
            << +slrg_16.height_eo_neg_mid     << ','
            << +slrg_16.phase_eo_pos_mid      << ','
            << +slrg_16.phase_eo_neg_mid      << ','
            << +slrg_16.height_eo_pos_low     << ','
            << +slrg_16.height_eo_neg_low     << ','
            << +slrg_16.phase_eo_pos_low      << ','
            << +slrg_16.phase_eo_neg_low      << ','
            << +slrg_16.offset_units          << ','
            << +slrg_16.phase_units           << ','
            << +slrg_16.height_grade_type     << ','
            << +slrg_16.phase_grade_type      << ','
            << +slrg_16.mid_eye_grade         << ','
            << +slrg_16.up_eye_grade          << ','
            << +slrg_16.dn_eye_grade          ;

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Dump_7nm(const struct slrg_reg &slrg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrg_7nm slrg_7;
    slrg_7nm_unpack(&slrg_7, slrg.page_data.slrg_data_set.data);

    sstream << +slrg_7.fom_measurment    << ','
            << +slrg_7.fom_mode          << ','
            << +slrg_7.initial_fom       << ','
            << +slrg_7.last_fom          << ','
            << +slrg_7.upper_eye         << ','
            << +slrg_7.mid_eye           << ','
            << +slrg_7.lower_eye         << ','
            << +slrg_7.fom_ext_conf      << ','
            << +slrg_7.fom_ext_conf_cap  << ','
            << +slrg_7.nblks_max         << ','
            << +slrg_7.nerrs_min;

    for (int i = 0; i < ACC_REG_SLRG_MAX_FIELDS_NUM - ACC_REG_SLRG7NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Dump_5nm(const struct slrg_reg &slrg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrg_5nm slrg_5;
    slrg_5nm_unpack(&slrg_5, slrg.page_data.slrg_data_set.data);

    sstream << +slrg_5.fom_measurment       << ','
            << +slrg_5.fom_mode             << ','
            << +slrg_5.initial_fom          << ','
            << +slrg_5.last_fom             << ','
            << +slrg_5.upper_eye            << ','
            << +slrg_5.mid_eye              << ','
            << +slrg_5.lower_eye            << ','
            << +slrg_5.fom_ext_conf         << ','
            << +slrg_5.fom_ext_conf_cap     << ','
            << +slrg_5.nblks_max            << ','
            << +slrg_5.nerrs_min;

    for (int i = 0; i < ACC_REG_SLRG_MAX_FIELDS_NUM - ACC_REG_SLRG5NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if ((p_export_data_phy_port && this->m_pnat == ACC_REG_PNAT_IB_PORT) ||
            (p_export_data_phy_node && this->m_pnat == ACC_REG_PNAT_OOB_PORT)) {
        struct slrg_reg *p_slrg_reg = new struct slrg_reg(areg.regs.slrg);

        switch (p_slrg_reg->version) {
            case 0: // 40nm
            case 1: // 28nm
                slrg_40nm_28nm_unpack(&p_slrg_reg->page_data.slrg_40nm_28nm, areg.regs.slrg.page_data.slrg_data_set.data);
                break;
            case 3: // 16nm
                slrg_16nm_unpack(&p_slrg_reg->page_data.slrg_16nm, areg.regs.slrg.page_data.slrg_data_set.data);
                break;
            case 4: // 7nm
            case 5: // 5nm
                slrg_7nm_unpack(&p_slrg_reg->page_data.slrg_7nm, areg.regs.slrg.page_data.slrg_data_set.data);
                break;
            default:
                break;
        }

        if (this->m_pnat == ACC_REG_PNAT_IB_PORT) {
            p_export_data_phy_port->p_slrg_reg[((AccRegKeyPortLane *)p_key)->lane] = p_slrg_reg;
        } else if (this->m_pnat == ACC_REG_PNAT_OOB_PORT) {
            AccRegKeyPCIPortLane *p_pcipl_key = (AccRegKeyPCIPortLane *)p_key;
            p_export_data_phy_node->p_slrg_reg[p_pcipl_key->port_num] = p_slrg_reg;
        }
    }

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLLM Register functions      ***
//*****************************************
SLLMRegister::SLLMRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map,
                           map_akey_areg *mpir_map):
    SLRegister(phy_diag,
               ACCESS_REGISTER_ID_SLLM,
               (const unpack_data_func_t)sllm_reg_unpack,
               sn,
               ACC_REG_SLLM_NAME,
               ACC_REG_SLLM7NM_FIELDS_NUM,
               NOT_SUPPORT_SLLM,
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLLMRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct sllm_reg sllm;
    CLEAR_STRUCT(sllm);
    sllm.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    sllm.lane = ((AccRegKeyPortLane *)p_key)->lane;
    sllm.pnat = this->m_pnat;
    sllm.port_type = 0;
    sllm_reg_pack(&sllm, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLLMRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLLMRegister::Header_Dump_7nm(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "status"                 << ','
            << "version_7nm=4"          << ','
            << "local_port"             << ','
            << "pnat"                   << ','
            << "lp_msb"                 << ','
            << "lane"                   << ','
            << "port_type"              << ','
            << "c_db"                   << ','
            << "ctle_peq_en"            << ','
            << "peq_tsense_en"          << ','
            << "peq_f1_adapt_skip"      << ','
            << "vref_peq_en"            << ','
            << "dffe_peq_scout_skip"    << ','
            << "peq_train_mode"         << ','
            << "peq_vref_iters"         << ','
            << "peq_adc_vref_step"      << ','
            << "dffe_peq_en"            << ','
            << "peq_dffe_iters"         << ','
            << "peq_dffe_delay"         << ','
            << "ctle_peq_cnt"           << ','
            << "ber_mon_mantissa"       << ','
            << "ber_mon_exp";

    IBDIAGNET_RETURN_VOID;
}

void SLLMRegister::Header_Dump_16nm(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "status"               << ','
            << "version_16nm=3"       << ','
            << "local_port"           << ','
            << "pnat"                 << ','
            << "lp_msb"               << ','
            << "lane"                 << ','
            << "port_type"            << ','
            << "c_db"                 << ','
            << "lm_active"            << ','
            << "lm_was_active"        << ','
            << "pib_gw_lock"          << ','
            << "lm_en"                << ','
            << "lm_clk90_fl_err_max"  << ','
            << "lm_clk90_fl_err_acc"  << ','
            << "ib3_max_lm_90_tries"  << ','
            << "lm_counter_up"        << ','
            << "ib1_max_lm_90_tries"  << ','
            << "lm_counter_mid"       << ','
            << "lm_counter_dn";

    IBDIAG_RETURN_VOID;
}

void SLLMRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct sllm_reg &sllm = areg.regs.sllm;
    sstream << +sllm.status         << ','
            << +sllm.version        << ','
            << +sllm.local_port     << ','
            << +sllm.pnat           << ','
            << +sllm.lp_msb         << ','
            << +sllm.lane           << ','
            << +sllm.port_type      << ','
            << +sllm.c_db           << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (sllm.version) {

        case 3: // 16nm
            Dump_16nm(sllm, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(sllm, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +sllm.version << '"';
            for (int i = 0; i < ACC_REG_SLLM_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLLM: %d, on node: " U64H_FMT ".\n",
                        sllm.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void SLLMRegister::Dump_16nm(const struct sllm_reg &sllm, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sllm_16nm sllm_16;
    sllm_16nm_unpack(&sllm_16, sllm.page_data.sllm_data_set.data);

    sstream << +sllm_16.lm_active                   << ','
            << +sllm_16.lm_was_active               << ','
            << +sllm_16.pib_gw_lock                 << ','
            << +sllm_16.lm_en                       << ','
            << +sllm_16.lm_clk90_fl_err_max         << ','
            << +sllm_16.lm_clk90_fl_err_acc         << ','
            << +sllm_16.ib3_max_lm_90_tries         << ','
            << +sllm_16.lm_counter_up               << ','
            << +sllm_16.ib1_max_lm_90_tries         << ','
            << +sllm_16.lm_counter_mid              << ','
            << +sllm_16.lm_counter_dn               << ','
            << "NA,NA,NA";  // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLLMRegister::Dump_7nm(const struct sllm_reg &sllm, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sllm_7nm sllm_7;
    sllm_7nm_unpack(&sllm_7, sllm.page_data.sllm_data_set.data);

    sstream << +sllm_7.ctle_peq_en                 << ','
            << +sllm_7.peq_tsense_en               << ','
            << +sllm_7.peq_f1_adapt_skip           << ','
            << +sllm_7.vref_peq_en                 << ','
            << +sllm_7.dffe_peq_scout_skip         << ','
            << +sllm_7.peq_train_mode              << ','
            << +sllm_7.peq_vref_iters              << ','
            << +sllm_7.peq_adc_vref_step           << ','
            << +sllm_7.dffe_peq_en                 << ','
            << +sllm_7.peq_dffe_iters              << ','
            << +sllm_7.peq_dffe_delay              << ','
            << +sllm_7.ctle_peq_cnt                << ','
            << +sllm_7.ber_mon_mantissa            << ','
            << +sllm_7.ber_mon_exp                 ;

    IBDIAGNET_RETURN_VOID;
}

void SLLMRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port) {
        struct sllm_reg *p_sllm_reg = new struct sllm_reg(areg.regs.sllm);

        switch (p_sllm_reg->version) {
            case 3: // 16nm
                sllm_16nm_unpack(&p_sllm_reg->page_data.sllm_16nm, areg.regs.sllm.page_data.sllm_data_set.data);
                break;
            case 4: // 7nm
                sllm_7nm_unpack(&p_sllm_reg->page_data.sllm_7nm, areg.regs.sllm.page_data.sllm_data_set.data);
                break;
            default:
                break;
        }

        p_export_data_phy_port->p_sllm_reg[((AccRegKeyPortLane *)p_key)->lane] = p_sllm_reg;
    }

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* MPEIN Register functions ***
//*****************************************
MPEINRegister::MPEINRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MPEIN,
                (const unpack_data_func_t)mpein_reg_unpack,
                ACC_REG_MPEIN_INTERNAL_SECTION_NAME,
                ACC_REG_MPEIN_NAME,
                ACC_REG_MPEIN_FIELDS_NUM,
                NOT_SUPPORT_MPEIN)
{
}


int MPEINRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)p_key;
    struct mpein_reg mpein;
    CLEAR_STRUCT(mpein);

    mpein.depth = p_dpnkey->depth;
    mpein.pcie_index = p_dpnkey->pci_idx;
    mpein.node = p_dpnkey->pci_node;
    mpein_reg_pack(&mpein, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MPEINRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                           << '#'
            << handler_header                 << ','
            << "capability_mask"              << ','
            << "link_width_enabled"           << ','
            << "link_speed_enabled"           << ','
            << "lane0_physical_position"      << ','
            << "link_width_active"            << ','
            << "link_speed_active"            << ','
            << "num_of_pfs"                   << ','
            << "num_of_vfs"                   << ','
            << "bdf0"                         << ','
            << "max_read_request_size"        << ','
            << "max_payload_size"             << ','
            << "pwr_status"                   << ','
            << "port_type"                    << ','
            << "lane_reversal"                << ','
            << "link_peer_max_speed"          << ','
            << "pci_power"                    << ','
            << "device_status"                << ','
            << "port_state"                   << ','
            << "receiver_detect_result";
#endif
    IBDIAGNET_RETURN_VOID;
}

void MPEINRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct mpein_reg &p_mpein = areg.regs.mpein;

    // HEADER
    // NodeGuid, PCIIndex, Depth, PCINode,
    // capability_mask, link_width_enabled, link_speed_enabled, lane0_physical_position,
    // link_width_active, link_speed_active, num_of_pfs, num_of_vfs, bdf0, max_read_request_size,
    // max_payload_size, pwr_status, port_type, lane_reversal, link_peer_max_speed,
    // pci_power, device_status, port_state, receiver_detect_result

    sstream << hex
            << "0x" << +p_mpein.capability_mask              << ','
            << "0x" << +p_mpein.link_width_enabled           << ','
            << "0x" << +p_mpein.link_speed_enabled           << ','
            << "0x" << +p_mpein.lane0_physical_position      << ','
            << "0x" << +p_mpein.link_width_active            << ','
            << "0x" << +p_mpein.link_speed_active            << ','
            << "0x" << +p_mpein.num_of_pfs                   << ','
            << "0x" << +p_mpein.num_of_vfs                   << ','
            << "0x" << +p_mpein.bdf0                         << ','
            << "0x" << +p_mpein.max_read_request_size        << ','
            << "0x" << +p_mpein.max_payload_size             << ','
            << "0x" << +p_mpein.pwr_status                   << ','
            << "0x" << +p_mpein.port_type                    << ','
            << "0x" << +p_mpein.lane_reversal                << ','
            << "0x" << +p_mpein.link_peer_max_speed          << ','
            << "0x" << +p_mpein.pci_power                    << ','
            << "0x" << +p_mpein.device_status                << ','
            << "0x" << +p_mpein.port_state                   << ','
            << "0x" << +p_mpein.receiver_detect_result
            << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

/*
 * Build DB in the following order:
 * For each CA
 *   For each PCI index
 *     For each PCI depth
 *       For each PCI node
*/
int MPEINRegister::BuildDB(AccRegHandler* p_handler,
                           list_p_fabric_general_err &phy_errors,
                           ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    // clbck_data.m_p_obj            - handler
    // clbck_data.m_data1            - node
    // clbck_data.m_data2            - key
    // clbck_data.m_data3            - port
    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = &AccRegHandler_ForwardClbck<MPEINRegister,
                &MPEINRegister::LoopPCIDepth>;

    clbck_data.m_p_obj = p_handler;

    p_handler->SetHeader(PCI_DPN_HEADER);

    for (map_str_psys::iterator sys_iter = m_phy_diag->GetFabric()->SystemByName.begin();
            sys_iter != m_phy_diag->GetFabric()->SystemByName.end(); sys_iter++) {

        IBSystem *p_curr_sys = sys_iter->second;
        if (!p_curr_sys) {
            m_phy_diag->SetLastError("DB error - found null node in"  \
                    " SystemByName map for key = %s",
                    sys_iter->first.c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        set<uint64_t> sys_guids;

        for (map_str_pnode::iterator it = p_curr_sys->NodeByName.begin();
             it != p_curr_sys->NodeByName.end(); it++) {

            IBNode *p_curr_node = it->second;
            if (!p_curr_node) {
                m_phy_diag->SetLastError("DB error - found null node in"  \
                        " NodeByName map in IBSystem for key = %s",
                        p_curr_sys->NodeByName.begin()->first.c_str());
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            if (p_curr_node->type == IB_SW_NODE)
                continue;

            // skip ANs
            if (p_curr_node->isSpecialNode())
                continue;

            // todo - filter node not in scope already in the SystemByName map
            //        as in NodeByName map.
            // skip not in scope
            if (!p_curr_node->getInSubFabric())
                continue;

            AccRegVia_t acc_reg_via = Validation(p_curr_node, rc);

            if (acc_reg_via == NOT_SUP_ACC_REG)
                continue;

            pair<set<uint64_t>::iterator,bool> ret = sys_guids.insert(p_curr_node->system_guid_get());
            if (!ret.second)
                continue;

            clbck_data.m_data1 = p_curr_node;
            clbck_data.m_data4 = (void*)(u_int64_t)acc_reg_via;

            for (u_int32_t port_num = 1; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)port_num);

                if (!p_curr_port || !p_curr_port->is_data_worthy())
                   continue;

                clbck_data.m_data3 = p_curr_port;

                for (u_int8_t idx = 0; idx < MAX_PCI_IDX; ++idx) {

                    AccRegKeyDPN *p_dpnkey = new AccRegKeyDPN(
                            p_curr_port, // port
                            0,        // depth
                            idx,      // pci index
                            0);       // pci node

                    clbck_data.m_data2 = p_dpnkey;

                    AccessRegister acc_reg;
                    CLEAR_STRUCT(acc_reg);

                    rc = p_handler->SendAccReg(acc_reg_via,
                                          p_curr_node,
                                          p_curr_port->num,
                                          p_curr_port->base_lid,
                                          acc_reg,
                                          p_dpnkey,
                                          p_progress_bar,
                                          &clbck_data);

                    if (rc == IBDIAG_ERR_CODE_DB_ERR)
                        IBDIAGNET_RETURN(rc);

                } // pci index loop

                break; // send only for one port per node
            } // port loop
        } // nodes loop
    } // system loop

    IBDIAGNET_RETURN(rc);
}

int MPEINRegister::LoopPCIDepth(const clbck_data_t &clbck_data,
                                int rec_status,
                                void *p_attribute_data)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    AccRegHandler *p_handler = (AccRegHandler*)clbck_data.m_p_obj;
    IBNode *p_curr_node = (IBNode*)clbck_data.m_data1;
    AccRegKeyDPN *p_dpnkey_old = (AccRegKeyDPN *)clbck_data.m_data2;
    IBPort *p_curr_port = (IBPort*)clbck_data.m_data3;
    AccRegVia_t acc_reg_via = (AccRegVia_t)(u_int64_t)clbck_data.m_data4;

    direct_route_t *p_curr_direct_route = m_phy_diag->GetIBDiag()->GetDR(p_curr_node);
    if (!p_curr_direct_route) {
        m_phy_diag->SetLastError("DB error - current direct route is NULL. Node GUID="
                U64H_FMT ", node name %s.", p_curr_node->guid_get(), p_curr_node->getName().c_str());

        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    mpein_reg mpein;
    CLEAR_STRUCT(mpein);

    // invoke handler callback in order to handle return status and insert reply to the map
    if (acc_reg_via == VIA_SMP) {
        rc = p_handler->SMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);
        if (rc)
            IBDIAGNET_RETURN(rc);

        SMP_AccessRegister *p_acc_reg_smp = (SMP_AccessRegister*)p_attribute_data;
        mpein_reg_unpack(&mpein, p_acc_reg_smp->reg.data);
    }

    if (acc_reg_via == VIA_GMP){
        rc = p_handler->GMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);
        if (rc)
            IBDIAGNET_RETURN(rc);

        struct GMP_AccessRegister *p_acc_reg_gmp = (struct GMP_AccessRegister *)p_attribute_data;
        mpein_reg_unpack(&mpein, p_acc_reg_gmp->reg.data);
    }

    clbck_data_t clbck_data_new = clbck_data;
    clbck_data_new.m_handle_data_func = &AccRegHandler_ForwardClbck<MPEINRegister,
                   &MPEINRegister::LoopPCINode>;

    if (mpein.port_type == PCI_PORT_TYPE_US) {

        for (u_int8_t depth = 0; depth < MAX_PCI_DEPTH; ++depth) {

            AccRegKeyDPN *p_dpnkey = new AccRegKeyDPN(
                    p_curr_port,             // port
                    depth,                   // depth
                    p_dpnkey_old->pci_idx,   // pci index
                    0);                      // pci node

            clbck_data_new.m_data2 = p_dpnkey;

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            rc = p_handler->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  p_curr_port->num,
                                  p_curr_port->base_lid,
                                  acc_reg,
                                  p_dpnkey,
                                  (ProgressBar *)clbck_data.m_p_progress_bar,
                                  &clbck_data_new);

            if (rc == IBDIAG_ERR_CODE_DB_ERR)
                IBDIAGNET_RETURN(rc);
        }
    }

    IBDIAGNET_RETURN(rc);
}

int MPEINRegister::LoopPCINode(const clbck_data_t &clbck_data,
                                int rec_status,
                                void *p_attribute_data)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    AccRegHandler *p_handler = (AccRegHandler*)clbck_data.m_p_obj;
    IBNode *p_curr_node = (IBNode*)clbck_data.m_data1;
    AccRegKeyDPN *p_dpnkey_old = (AccRegKeyDPN *)clbck_data.m_data2;
    IBPort *p_curr_port = (IBPort*)clbck_data.m_data3;
    AccRegVia_t acc_reg_via = (AccRegVia_t)(u_int64_t)clbck_data.m_data4;

    direct_route_t *p_curr_direct_route = m_phy_diag->GetIBDiag()->GetDR(p_curr_node);
    if (!p_curr_direct_route) {
        m_phy_diag->SetLastError("DB error - current direct route is NULL. Node GUID="
                U64H_FMT ", node name %s.", p_curr_node->guid_get(), p_curr_node->getName().c_str());

        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    mpein_reg mpein;
    CLEAR_STRUCT(mpein);

    // invoke handler callback in order to handle return status and insert reply to the map
    if (acc_reg_via == VIA_SMP) {
        rc = p_handler->SMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);
        if (rc)
            IBDIAGNET_RETURN(rc);

        SMP_AccessRegister *p_acc_reg_smp = (SMP_AccessRegister*)p_attribute_data;
        mpein_reg_unpack(&mpein, p_acc_reg_smp->reg.data);
    }

    if (acc_reg_via == VIA_GMP){
        rc = p_handler->GMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);
        if (rc)
            IBDIAGNET_RETURN(rc);

        struct GMP_AccessRegister *p_acc_reg_gmp = (struct GMP_AccessRegister *)p_attribute_data;
        mpein_reg_unpack(&mpein, p_acc_reg_gmp->reg.data);
    }

    if (mpein.port_type == PCI_PORT_TYPE_DS) {

        for (u_int8_t node = 0; node < MAX_PCI_NODES; ++node) {

            AccRegKeyDPN *p_dpnkey = new AccRegKeyDPN(
                    p_curr_port,                // port
                    p_dpnkey_old->depth,        // depth
                    p_dpnkey_old->pci_idx,      // pci index
                    node);                      // pci node

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            rc = p_handler->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  p_curr_port->num,
                                  p_curr_port->base_lid,
                                  acc_reg,
                                  p_dpnkey,
                                  (ProgressBar *)clbck_data.m_p_progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR)
                IBDIAGNET_RETURN(rc);
        }
    }

    IBDIAGNET_RETURN(rc);
}

bool MPEINRegister::IsValidPakcet(const struct acc_reg_data & reg_data) const
{
    u_int8_t port_type = reg_data.regs.mpein.port_type;

    if (port_type == PCI_PORT_TYPE_EP ||
        port_type == PCI_PORT_TYPE_US ||
        port_type == PCI_PORT_TYPE_DS)
        return true;

    return false;
}

void MPEINRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node,
                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node) {
        AccRegKeyDPN *p_dpn_key = (AccRegKeyDPN *)p_key;
        p_export_data_phy_node->p_mpein_reg[p_dpn_key->pci_idx][p_dpn_key->depth][p_dpn_key->pci_node] = &areg.regs.mpein;
    }

    IBDIAGNET_RETURN_VOID;
}

bool MPEINRegister::CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                     export_data_phy_node_t *p_export_data_phy_node,
                                     AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node &&
            p_key->node_guid == p_export_data_phy_node->node_guid)
        IBDIAGNET_RETURN(true);

    IBDIAGNET_RETURN(false);
}

//*****************************************
//***      MPIR Register functions      ***
//*****************************************
MPIRRegister::MPIRRegister(PhyDiag * phy_diag, map_akey_areg &mpein_map) :
    Register(phy_diag,
             ACCESS_REGISTER_ID_MPIR,
             (const unpack_data_func_t)mpir_reg_unpack,
             ACC_REG_MPIR_INTERNAL_SECTION_NAME,
             ACC_REG_MPIR_NAME,
             ACC_REG_MPIR_FIELDS_NUM,
             NOT_SUPPORT_MPIR),
    mpein_map(mpein_map) // this register should be sent to
                         // all valid mpein registers.
{
}

int MPIRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    AccRegKeyDPN* p_dpnkey = (AccRegKeyDPN*)p_key;
    mpir_reg mpir;
    CLEAR_STRUCT(mpir);

    mpir.depth = p_dpnkey->depth;
    mpir.pcie_index = p_dpnkey->pci_idx;
    mpir.node = p_dpnkey->pci_node;
    mpir_reg_pack(&mpir, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MPIRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl              << '#'
            << handler_header    << ','
            << "host_buses"      << ','
            << "sdm"             << ','
            << "subordinate_bus" << ','
            << "secondary_bus"   << ','
            << "device"          << ','
            << "bus"             << ','
            << "local_port"      << ','
            << "slot_number"     << ','
            << "num_con_devices" << ','
            << "slot_cap";
#endif
    IBDIAGNET_RETURN_VOID
}

void MPIRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct mpir_reg & mpir = areg.regs.mpir;

    sstream << +mpir.host_buses << ','
            << +mpir.sdm << ','
            << +mpir.subordinate_bus << ','
            << +mpir.secondary_bus << ','
            << +mpir.device << ','
            << +mpir.bus << ','
            << +mpir.local_port << ','
            << +mpir.slot_number << ','
            << +mpir.num_con_devices << ','
            << +mpir.slot_cap << endl;

    IBDIAGNET_RETURN_VOID;
}

int MPIRRegister::BuildDB(AccRegHandler* p_handler,
                          list_p_fabric_general_err &phy_errors,
                          ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func =  &forwardClbck<AccRegHandler,
                    &AccRegHandler::PCIAddressCollectGetClbck>;
    clbck_data.m_p_obj = p_handler;

    p_handler->SetHeader(PCI_DPN_HEADER);

    // go over all the mpein MADs and send mpir accordingly
    for (map_akey_areg::iterator mpein_iter = mpein_map.begin();
         mpein_iter != mpein_map.end(); mpein_iter++) {

        AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)mpein_iter->first;

        // get node
        IBNode *p_curr_node =
                m_phy_diag->GetFabric()->getNodeByGuid(p_dpnkey->node_guid);
        if (!p_curr_node) {
            m_phy_diag->SetLastError("DB error - found null node in "
                                     "NodeByName map for key = " U64H_FMT,
                                                p_dpnkey->node_guid);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        AccRegVia_t acc_reg_via = Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        // get port
        IBPort *p_curr_port = NULL;
        for (u_int32_t port_num = 1; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
            p_curr_port = p_curr_node->getPort((phys_port_t)port_num);

            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;
            break; // use the first valid port.
        }

        if (!p_curr_port) {
            m_phy_diag->SetLastError("DB error - failed to find valid port for node %s",
                                     p_curr_node->getName().c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        AccessRegister acc_reg;
        CLEAR_STRUCT(acc_reg);

        AccRegKeyDPN *p_dpnkey_new = new AccRegKeyDPN(p_dpnkey->p_port,
                                                      p_dpnkey->depth,
                                                      p_dpnkey->pci_idx,
                                                      p_dpnkey->pci_node);

        clbck_data.m_data1 = p_curr_node;
        clbck_data.m_data2 = p_dpnkey_new;
        clbck_data.m_data4 = (void*)(uint64_t)acc_reg_via;

        rc = p_handler->SendAccReg(acc_reg_via,
                                   p_curr_node,
                                   0, // port 0
                                   p_curr_port->base_lid,
                                   acc_reg,
                                   p_dpnkey_new,
                                   p_progress_bar,
                                   &clbck_data);

        if (rc == IBDIAG_ERR_CODE_DB_ERR)
            IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(rc);
}


//*****************************************
//***      MFCR Register functions      ***
//*****************************************
MFCRRegister::MFCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MFCR,
                (const unpack_data_func_t)mfcr_reg_unpack,
                ACC_REG_MFCR_INTERNAL_SECTION_NAME,
                ACC_REG_MFCR_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MFCR,
                AVAILABLE_FANS,
                SUPPORT_SW,
                false)
{
}

int MFCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MFCRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct mfcr_reg & p_mfcr = areg.regs.mfcr;
    sstream << hex << p_mfcr.tacho_active << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

int MFCRRegister::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    struct mfcr_reg *p_mfcr = &areg.regs.mfcr;
    int rc  = SensorsBitsToList(p_mfcr->tacho_active, MAX_FANS, sensors_list);
    IBDIAGNET_RETURN(rc);
}

//*****************************************
//***      MFSM Register functions      ***
//*****************************************
MFSMRegister::MFSMRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MFSM,
                (const unpack_data_func_t)mfsm_reg_unpack,
                ACC_REG_MFSM_INTERNAL_SECTION_NAME,
                ACC_REG_MFSM_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MFSM,
                FANS_SPEED_HEADER,
                SUPPORT_SW)
{
}

int MFSMRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct mfsm_reg mfsm;
    CLEAR_STRUCT(mfsm);
    mfsm.tacho = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mfsm_reg_pack(&mfsm, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MFSMRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    char buffer[1024] = {0};
    const struct mfsm_reg & p_mfsm = areg.regs.mfsm;

    snprintf(buffer, sizeof(buffer), U32D_FMT, p_mfsm.rpm);

    sstream << buffer << endl;
    IBDIAGNET_RETURN_VOID;
}

void MFSMRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node)
        p_export_data_phy_node->p_mfsm_reg[((AccRegKeyNodeSensor *)p_key)->sensor_id] = &areg.regs.mfsm;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MFSL Register functions      ***
//*****************************************
MFSLRegister::MFSLRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MFSL,
                (const unpack_data_func_t)mfsl_reg_unpack,
                ACC_REG_MFSL_INTERNAL_SECTION_NAME,
                ACC_REG_MFSL_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MFSL,
                FANS_THRESHOLDS_HEADER,
                SUPPORT_SW)
{
}

int MFSLRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct mfsl_reg mfsl;
    CLEAR_STRUCT(mfsl);
    mfsl.tacho = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mfsl_reg_pack(&mfsl, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MFSLRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    char buffer[1024] = {0};
    const struct mfsl_reg & p_mfsl = areg.regs.mfsl;

    snprintf(buffer, sizeof(buffer), U32D_FMT "," U32D_FMT, p_mfsl.tach_min, p_mfsl.tach_max);

    sstream << buffer << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      FORE Register functions      ***
//*****************************************
FORERegister::FORERegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_FORE,
                (const unpack_data_func_t)fore_reg_unpack,
                ACC_REG_FORE_INTERNAL_SECTION_NAME,
                ACC_REG_FORE_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_FORE,
                FANS_ALERT_HEADER,
                SUPPORT_SW)
{
}

int FORERegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void FORERegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct fore_reg & p_fore = areg.regs.fore;
    stringstream under_limit;
    stringstream over_limit;
    u_int16_t mask = 1;

    if (!p_fore.fan_under_limit && !p_fore.fan_over_limit) {
        sstream << "-1,-1" << endl;
        IBDIAGNET_RETURN_VOID;
    }

    for (int i = 0; i < MAX_FANS; ++i, mask <<= 1) {
        if (mask & p_fore.fan_under_limit)
            under_limit << i << "|";
        if (mask & p_fore.fan_over_limit)
            over_limit << i << "|";
    }

    string under;
    string over;

    if (!p_fore.fan_under_limit)
        under = "-1";
    else {
        under = under_limit.str();
        under = under.substr(0, under.size() - 1);

    }

    if (!p_fore.fan_over_limit)
        over = "-1";
    else {
        over = over_limit.str();
        over = over.substr(0, over.size() - 1);

    }

    sstream << under << "," << over << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***     MFNR Register functions       ***
//*****************************************
MFNRRegister::MFNRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MFNR,
                (const unpack_data_func_t)mfnr_reg_unpack,
                ACC_REG_MFNR_INTERNAL_SECTION_NAME,
                ACC_REG_MFNR_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MFNR,
                FANS_NUMBERS_HEADER,
                SUPPORT_SW)
{
}

int MFNRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct mfnr_reg mfnr;
    CLEAR_STRUCT(mfnr);
    mfnr.fan_index = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mfnr_reg_pack(&mfnr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MFNRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct mfnr_reg & p_mfnr = areg.regs.mfnr;
    sstream << p_mfnr.serial_number << endl;
    IBDIAGNET_RETURN_VOID;
}

//******************************************
//***      MTCAP Register functions      ***
//******************************************
MTCAPRegister::MTCAPRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MTCAP,
                (const unpack_data_func_t)mtcap_reg_unpack,
                ACC_REG_MTCAP_INTERNAL_SECTION_NAME,
                ACC_REG_MTCAP_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MTCAP,
                TEMPERATURE_SENSORS_COUNT,
                SUPPORT_SW,
                false) // dump enabled
{
}

int MTCAPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MTCAPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    char buffer[32] = {0};
    const struct mtcap_reg & p_mtcap = areg.regs.mtcap;
    snprintf(buffer, sizeof(buffer), U32D_FMT, p_mtcap.sensor_count);
    sstream << buffer << endl;

    IBDIAGNET_RETURN_VOID;
}

int MTCAPRegister::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    struct mtcap_reg *p_mtcap = &areg.regs.mtcap;
    IBDIAGNET_RETURN(SensorsCountToList(p_mtcap->sensor_count, sensors_list));
}

//*****************************************
//***      MTMP Register functions      ***
//*****************************************
MTMPRegister::MTMPRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MTMP,
                (const unpack_data_func_t)mtmp_reg_unpack,
                ACC_REG_MTMP_INTERNAL_SECTION_NAME,
                ACC_REG_MTMP_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MTMP,
                TEMPERATURE_SENSORS_HEADER,
                SUPPORT_SW)
{
}

int MTMPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct mtmp_reg mtmp;
    CLEAR_STRUCT(mtmp);
    mtmp.sensor_index = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mtmp_reg_pack(&mtmp, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MTMPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    char buffer[1024] = {0};
    const struct mtmp_reg & p_mtmp = areg.regs.mtmp;

    int16_t temperature = p_mtmp.temperature;
    int16_t max_temperature = p_mtmp.max_temperature;
    int16_t threshold_lo = p_mtmp.temperature_threshold_lo;
    int16_t threshold_hi = p_mtmp.temperature_threshold_hi;

    snprintf(buffer, sizeof(buffer), FLT_FMT "," FLT_FMT "," FLT_FMT "," FLT_FMT,
            (float)temperature / 8, (float)max_temperature / 8,
            (float)threshold_lo / 8, (float)threshold_hi / 8);

    sstream << p_mtmp.sensor_name_hi << p_mtmp.sensor_name_lo << "," << buffer << endl;

    IBDIAGNET_RETURN_VOID;
}

void MTMPRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node)
        p_export_data_phy_node->p_mtmp_reg[((AccRegKeyNodeSensor *)p_key)->sensor_id] = &areg.regs.mtmp;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MTWE Register functions      ***
//*****************************************
MTWERegister::MTWERegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MTWE,
                (const unpack_data_func_t)mtwe_reg_unpack,
                ACC_REG_MTWE_INTERNAL_SECTION_NAME,
                ACC_REG_MTWE_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MTWE,
                TEMPERATURE_SENSORS_ALERT_HEADER,
                SUPPORT_SW)
{
}

int MTWERegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MTWERegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct mtwe_reg & p_mtwe = areg.regs.mtwe;
    stringstream sensor_warning;
    const int num_of_warnings = 4;
    const int bits_per_warning = 32;
    u_int32_t mask = 1;
    u_int32_t sensor_warnings[num_of_warnings];
    string warning;

    if (!p_mtwe.sensor_warning0 && !p_mtwe.sensor_warning1 &&
        !p_mtwe.sensor_warning2 && !p_mtwe.sensor_warning3 ) {
        sstream << "-1" << endl;
        IBDIAGNET_RETURN_VOID;
    }

    sensor_warnings[0] = p_mtwe.sensor_warning0;
    sensor_warnings[1] = p_mtwe.sensor_warning1;
    sensor_warnings[2] = p_mtwe.sensor_warning2;
    sensor_warnings[3] = p_mtwe.sensor_warning3;

    for (int i = 0; i < num_of_warnings; ++i) {
        for (int j = 0; j < bits_per_warning; ++j, mask <<= 1) {
            if (mask & sensor_warnings[i])
                sensor_warning << (i * bits_per_warning + j) << "|";
        }
        mask = 1;
    }

    warning = sensor_warning.str();
    warning = warning.substr(0, warning.size() - 1);
    sstream << warning << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***     MVCAP Register functions     ***
//*****************************************
MVCAPRegister::MVCAPRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MVCAP,
                (const unpack_data_func_t)mvcap_reg_unpack,
                ACC_REG_MVCAP_INTERNAL_SECTION_NAME,
                ACC_REG_MVCAP_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MVCAP,
                AVAILABLE_POWER_SENSORS,
                SUPPORT_SW,
                false)
{
}

int MVCAPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MVCAPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct mvcap_reg & p_mvcap = areg.regs.mvcap;
    sstream << hex << p_mvcap.sensor_map << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

int MVCAPRegister::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    struct mvcap_reg *p_mvcap = &areg.regs.mvcap;
    int rc = SensorsBitsToList(p_mvcap->sensor_map, MAX_VOLTAGE_CURRENT_SENSORS, sensors_list);
    IBDIAGNET_RETURN(rc);
}

//*****************************************
//***      MVCR Register functions      ***
//*****************************************
MVCRRegister::MVCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MVCR,
                (const unpack_data_func_t)mvcr_reg_unpack,
                ACC_REG_MVCR_INTERNAL_SECTION_NAME,
                ACC_REG_MVCR_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MVCR,
                POWER_SENSORS_HEADER,
                SUPPORT_SW)
{
}

int MVCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct mvcr_reg mvcr;
    CLEAR_STRUCT(mvcr);
    mvcr.sensor_index = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mvcr_reg_pack(&mvcr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MVCRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    char buffer[128] = {0};
    char name[128] = {0};
    const struct mvcr_reg & p_mvcr = areg.regs.mvcr;

    memcpy(name, &p_mvcr.sensor_name, sizeof(u_int64_t));

    // todo change current from 'N/A' to real value when supported.
    snprintf(buffer, sizeof(buffer), FLT_FMT "," "N/A",
             p_mvcr.voltage_sensor_value / 100.0);

    sstream << name << "," << buffer << endl;
    IBDIAGNET_RETURN_VOID;
}

void MVCRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node)
        p_export_data_phy_node->p_mvcr_reg[((AccRegKeyNodeSensor *)p_key)->sensor_id] = &areg.regs.mvcr;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MSPS Register functions      ***
//*****************************************
MSPSRegister::MSPSRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MSPS,
                (const unpack_data_func_t)msps_reg_unpack,
                ACC_REG_MSPS_INTERNAL_SECTION_NAME,
                ACC_REG_MSPS_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MSPS,
                POWER_SUPPLIES_HEADER,
                SUPPORT_SW,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP)
{
}

int MSPSRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

static string DumpPSU(const struct psu_status & psu)
{
    string psu_status;

    switch(psu.present) {
        case 0 :
            psu_status += ",N/A,N/A";
            break;
        case 1 :
            psu_status += ",Yes,No";
            break;
        case 2 :
            psu_status += ",No,No";
            break;
        case 5 :
            psu_status += ",Yes,Yes";
            break;
        case 6 :
            psu_status += ",No,Yes";
            break;
        default:
            psu_status += ",,";
    }

    switch(psu.ac_status) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",Yes";
            break;
        case 2 :
            psu_status += ",No";
            break;
        default:
            psu_status += ",";
    }

    switch(psu.dc_status) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",OK";
            break;
        case 2 :
            psu_status += ",Error";
            break;
        default:
            psu_status += ",";
    }

    switch(psu.alert) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",No";
            break;
        case 2 :
            psu_status += ",Yes";
            break;
        default:
            psu_status += ",";
    }

    switch(psu.fan) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",No Fan";
            break;
        case 2 :
            psu_status += ",OK";
            break;
        case 3 :
            psu_status += ",Error";
            break;
        default:
            psu_status += ",";
    }

    switch(psu.temp) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",OK";
            break;
        case 2 :
            psu_status += ",Overtemp";
            break;
        default:
            psu_status += ",";
    }

    psu_status += ',';
    psu_status += to_string(psu.power_cap);

    if (psu.power_cap == 2) {
        psu_status += ',';
        psu_status += to_string(psu.power_consumption);
    } else
        psu_status += ",N/A";

    if (psu.snp) {
        psu_status += ',';
        psu_status += psu.serial_number;
    } else
        psu_status += ",N/A";

    return psu_status;
}

void MSPSRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct msps_reg & p_msps = areg.regs.msps;
    string node_guid = sstream.str(); // get the node guid from the key

    sstream << "1" << DumpPSU(p_msps.psu0) << endl
            << node_guid << "2" << DumpPSU(p_msps.psu1) << endl;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MSGI Register functions      ***
//*****************************************
MSGIRegister::MSGIRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MSGI,
                (const unpack_data_func_t)msgi_reg_unpack,
                ACC_REG_MSGI_INTERNAL_SECTION_NAME,
                ACC_REG_MSGI_NAME,
                UNUSED_FIELDS_NUM,
                NOT_SUPPORT_MSGI,
                SYSTEM_GENERAL_INFORMATION_HEADER,
                SUPPORT_SW,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP)
{
}

int MSGIRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MSGIRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct msgi_reg & p_msgi = areg.regs.msgi;

    sstream << p_msgi.serial_number << ','
            << p_msgi.part_number << ','
            << p_msgi.revision << ','
            << '\"' << p_msgi.product_name << '\"' << endl;

    IBDIAGNET_RETURN_VOID;
}

void MSGIRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node)
        p_export_data_phy_node->p_msgi_reg = &areg.regs.msgi;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLRIP Register functions     ***
//*****************************************
SLRIPRegister::SLRIPRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map)
    : SLRegister(phy_diag,
                ACCESS_REGISTER_ID_SLRIP,
                (const unpack_data_func_t)slrip_reg_unpack,
                sn,
                ACC_REG_SLRIP_NAME,
                ACC_REG_SLRIP_MAX_FIELDS_NUM,
                NOT_SUPPORT_SLRIP,
                mpein_map,
                mpir_map),
     m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLRIPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slrip_reg slrip;
    CLEAR_STRUCT(slrip);

    slrip.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slrip.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slrip.pnat = m_pnat;

    // ib_sel is just for 16nm version
    struct slrip_16nm slrip_16;
    CLEAR_STRUCT(slrip_16);
    slrip_16.ib_sel = ((AccRegKeyPortLane *)p_key)->idx_in_lane;
    slrip_16nm_pack(&slrip_16, slrip.page_data.slrip_data_set.data);

    slrip_reg_pack(&slrip, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLRIPRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);

#endif

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Header_Dump_16nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "stauts"               << ','
            << "version_16nm=0"       << ','
            << "local_port"           << ','
            << "p_nat"                << ','
            << "lp_msb"               << ','
            << "lane"                 << ','
            << "port_type"            << ','
            << "ib_sel"               << ','
            << "ffe_tap0"             << ','
            << "ffe_tap1"             << ','
            << "ffe_tap2"             << ','
            << "ffe_tap3"             << ','
            << "ffe_tap4"             << ','
            << "ffe_tap5"             << ','
            << "ffe_tap6"             << ','
            << "ffe_tap7"             << ','
            << "ffe_tap8"             << ','
            << "sel_enc0"             << ','
            << "ffe_tap_en"           << ','
            << "sel_enc1"             << ','
            << "sel_enc2"             << ','
            << "mixer_offset0"        << ','
            << "mixer_offset1"        << ','
            << "saved0_sel_enc0"      << ','
            << "saved0_mixer_offset0" << ','
            << "saved1_sel_enc0"      << ','
            << "saved1_mixer_offset0";

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "stauts"                      << ','
            << "version_7nm=4"               << ','
            << "local_port"                  << ','
            << "p_nat"                       << ','
            << "lp_msb"                      << ','
            << "lane"                        << ','
            << "port_type"                   << ','
            << "feq_train_mode"              << ','
            << "term_lpf_override_ctrl"      << ','
            << "vref_val_mode_ctrl"          << ','
            << "dffe_override_ctrl"          << ','
            << "vga_gain_override_ctrl"      << ','
            << "hf_gain_override_ctrl"       << ','
            << "mf_pole_override_ctrl"       << ','
            << "mf_gain_override_ctrl"       << ','
            << "lf_pole_override_ctrl"       << ','
            << "lf_gain_override_ctrl"       << ','
            << "dc_gain_override_ctrl"       << ','
            << "term_attn_override_ctrl"     << ','
            << "f1_override_ctrl"            << ','
            << "f2_override_ctrl"            << ','
            << "f3_override_ctrl"            << ','
            << "f4_override_ctrl"            << ','
            << "f5_override_ctrl"            << ','
            << "f6_override_ctrl"            << ','
            << "f7_override_ctrl"            << ','
            << "f8_override_ctrl"            << ','
            << "fm1_override_ctrl"           << ','
            << "fm2_override_ctrl"           << ','
            << "fm3_override_ctrl"           << ','
            << "eth_override_ctrl"           << ','
            << "mf_pole"                     << ','
            << "mf_gain"                     << ','
            << "lf_pole"                     << ','
            << "lf_gain"                     << ','
            << "term_attn_ctrl"              << ','
            << "hf_gain"                     << ','
            << "dc_gain"                     << ','
            << "adc_vref_val"                << ','
            << "vga_gain"                    << ','
            << "ffe_fm1"                     << ','
            << "ffe_fm2"                     << ','
            << "ffe_fm3"                     << ','
            << "ffe_f4 "                     << ','
            << "ffe_f3"                      << ','
            << "ffe_f2"                      << ','
            << "ffe_f1"                      << ','
            << "ffe_f8"                      << ','
            << "ffe_f7"                      << ','
            << "ffe_f6"                      << ','
            << "ffe_f5"                      << ','
            << "dffe_dsel3"                  << ','
            << "dffe_coef3"                  << ','
            << "dffe_dsel2"                  << ','
            << "dffe_coef2"                  << ','
            << "dffe_dsel1"                  << ','
            << "dffe_coef1"                  << ','
            << "dffe_dsel0"                  << ','
            << "dffe_coef0"                  << ','
            << "dffe_dsel7"                  << ','
            << "dffe_coef7"                  << ','
            << "dffe_dsel6"                  << ','
            << "dffe_coef6"                  << ','
            << "dffe_dsel5"                  << ','
            << "dffe_coef5"                  << ','
            << "dffe_dsel4"                  << ','
            << "dffe_coef4"                  << ','
            << "dffe_dsel8"                  << ','
            << "dffe_coef8"                  << ','
            << "th1"                         << ','
            << "th0"                         << ','
            << "th3"                         << ','
            << "th2"                         << ','
            << "th5"                         << ','
            << "th4"                         << ','
            << "th7"                         << ','
            << "th6"                         << ','
            << "th9"                         << ','
            << "th8"                         << ','
            << "th11"                        << ','
            << "th10"                        << ','
            << "th13"                        << ','
            << "th12"                        << ','
            << "th15"                        << ','
            << "th14"                        << ','
            << "th17"                        << ','
            << "th16"                        << ','
            << "th19"                        << ','
            << "th18"                        << ','
            << "th21"                        << ','
            << "th20";

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "stauts"                             << ','
            << "version_5nm=5"                      << ','
            << "local_port"                         << ','
            << "p_nat"                              << ','
            << "lp_msb"                             << ','
            << "lane"                               << ','
            << "port_type"                          << ','
            << "tah_amp_gain"                       << ','
            << "lf_gain"                            << ','
            << "lf_pol"                             << ','
            << "mf_gain"                            << ','
            << "mf_pol"                             << ','
            << "dc_gain"                            << ','
            << "hf_gain"                            << ','
            << "ffe_fm3"                            << ','
            << "ffe_fm4"                            << ','
            << "ffe_fm5"                            << ','
            << "ffe_f1"                             << ','
            << "ffe_fm1"                            << ','
            << "ffe_fm2"                            << ','
            << "ffe_f4"                             << ','
            << "ffe_f3"                             << ','
            << "ffe_f2"                             << ','
            << "ffe_f8"                             << ','
            << "ffe_f7"                             << ','
            << "ffe_f6"                             << ','
            << "ffe_f5"                             << ','
            << "ffe_f12"                            << ','
            << "ffe_f11"                            << ','
            << "ffe_f10"                            << ','
            << "ffe_f9"                             << ','
            << "dffe_coef2"                         << ','
            << "dffe_dsel2"                         << ','
            << "dffe_coef1"                         << ','
            << "dffe_dsel1"                         << ','
            << "dffe_coef0"                         << ','
            << "dffe_dsel0"                         << ','
            << "dffe_coef5"                         << ','
            << "dffe_dsel5"                         << ','
            << "dffe_coef4"                         << ','
            << "dffe_dsel4"                         << ','
            << "dffe_coef3"                         << ','
            << "dffe_dsel3"                         << ','
            << "dffe_coef8"                         << ','
            << "dffe_dsel8"                         << ','
            << "dffe_coef7"                         << ','
            << "dffe_dsel7"                         << ','
            << "dffe_coef6"                         << ','
            << "dffe_dsel6"                         << ','
            << "dffe_coef11"                        << ','
            << "dffe_dsel11"                        << ','
            << "dffe_coef10"                        << ','
            << "dffe_dsel10"                        << ','
            << "dffe_coef9"                         << ','
            << "dffe_dsel9"                         << ','
            << "dffe_coef14"                        << ','
            << "dffe_dsel14"                        << ','
            << "dffe_coef13"                        << ','
            << "dffe_dsel13"                        << ','
            << "dffe_coef12"                        << ','
            << "dffe_dsel12"                        << ','
            << "fb_slicer_th1"                      << ','
            << "fb_slicer_th0"                      << ','
            << "dffe_coef15"                        << ','
            << "dffe_dsel15"                        << ','
            << "fb_slicer_th4"                      << ','
            << "fb_slicer_th3"                      << ','
            << "fb_slicer_th2"                      << ','
            << "fb_slicer_th7"                      << ','
            << "fb_slicer_th6"                      << ','
            << "fb_slicer_th5"                      << ','
            << "fb_slicer_th10"                     << ','
            << "fb_slicer_th9"                      << ','
            << "fb_slicer_th8"                      << ','
            << "ff_slicer_th0"                      << ','
            << "fb_slicer_th12"                     << ','
            << "fb_slicer_th11"                     << ','
            << "ff_slicer_th3"                      << ','
            << "ff_slicer_th2"                      << ','
            << "ff_slicer_th1"                      << ','
            << "ff_slicer_th6"                      << ','
            << "ff_slicer_th5"                      << ','
            << "ff_slicer_th4"                      << ','
            << "ff_slicer_th9"                      << ','
            << "ff_slicer_th8"                      << ','
            << "ff_slicer_th7"                      << ','
            << "ff_slicer_th12"                     << ','
            << "ff_slicer_th11"                     << ','
            << "ff_slicer_th10";

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct slrip_reg & slrip = areg.regs.slrip;

    sstream << +slrip.status          << ','
            << +slrip.version         << ','
            << +slrip.local_port      << ','
            << +slrip.pnat            << ','
            << +slrip.lp_msb          << ','
            << +slrip.lane            << ','
            << +slrip.port_type       << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (slrip.version) {

        case 0: // 16nm
            Dump_16nm(slrip, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(slrip, sstream);
            break;

        case 5: // 5nm
            Dump_5nm(slrip, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +slrip.version << '"';
            for (int i = 0; i < ACC_REG_SLRIP_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLRIP: %d, on node: " U64H_FMT ".\n",
                        slrip.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Dump_16nm(const struct slrip_reg &slrip, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrip_16nm slrip_16;
    slrip_16nm_unpack(&slrip_16, slrip.page_data.slrip_data_set.data);

    sstream << +slrip_16.ib_sel                         << ','
            << +slrip_16.ffe_tap0                       << ','
            << +slrip_16.ffe_tap1                       << ','
            << +slrip_16.ffe_tap2                       << ','
            << +slrip_16.ffe_tap3                       << ','
            << +slrip_16.ffe_tap4                       << ','
            << +slrip_16.ffe_tap5                       << ','
            << +slrip_16.ffe_tap6                       << ','
            << +slrip_16.ffe_tap7                       << ','
            << +slrip_16.ffe_tap8                       << ','
            << +slrip_16.sel_enc0                       << ','
            << +slrip_16.ffe_tap_en                     << ','
            << +slrip_16.sel_enc1                       << ','
            << +slrip_16.sel_enc2                       << ','
            << +slrip_16.mixer_offset0                  << ','
            << +slrip_16.mixer_offset1                  << ','
            << +slrip_16.saved0_sel_enc0                << ','
            << +slrip_16.saved0_mixer_offset0           << ','
            << +slrip_16.saved1_sel_enc0                << ','
            << +slrip_16.saved1_mixer_offset0           ;

    for (int i = 0; i < ACC_REG_SLRIP_MAX_FIELDS_NUM - ACC_REG_SLRIP16NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Dump_7nm(const struct slrip_reg &slrip, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrip_7nm slrip_7;
    slrip_7nm_unpack(&slrip_7, slrip.page_data.slrip_data_set.data);

    sstream << +slrip_7.feq_train_mode              << ','
            << +slrip_7.term_lpf_override_ctrl      << ','
            << +slrip_7.vref_val_mode_ctrl          << ','
            << +slrip_7.dffe_override_ctrl          << ','
            << +slrip_7.vga_gain_override_ctrl      << ','
            << +slrip_7.hf_gain_override_ctrl       << ','
            << +slrip_7.mf_pole_override_ctrl       << ','
            << +slrip_7.mf_gain_override_ctrl       << ','
            << +slrip_7.lf_pole_override_ctrl       << ','
            << +slrip_7.lf_gain_override_ctrl       << ','
            << +slrip_7.dc_gain_override_ctrl       << ','
            << +slrip_7.term_attn_override_ctrl     << ','
            << +slrip_7.f1_override_ctrl            << ','
            << +slrip_7.f2_override_ctrl            << ','
            << +slrip_7.f3_override_ctrl            << ','
            << +slrip_7.f4_override_ctrl            << ','
            << +slrip_7.f5_override_ctrl            << ','
            << +slrip_7.f6_override_ctrl            << ','
            << +slrip_7.f7_override_ctrl            << ','
            << +slrip_7.f8_override_ctrl            << ','
            << +slrip_7.fm1_override_ctrl           << ','
            << +slrip_7.fm2_override_ctrl           << ','
            << +slrip_7.fm3_override_ctrl           << ','
            << +slrip_7.eth_override_ctrl           << ','
            << +slrip_7.mf_pole                     << ','
            << +slrip_7.mf_gain                     << ','
            << +slrip_7.lf_pole                     << ','
            << +slrip_7.lf_gain                     << ','
            << +slrip_7.term_attn_ctrl              << ','
            << +slrip_7.hf_gain                     << ','
            << +slrip_7.dc_gain                     << ','
            << +slrip_7.adc_vref_val                << ','
            << +slrip_7.vga_gain                    << ','
            << +slrip_7.ffe_fm1                     << ','
            << +slrip_7.ffe_fm2                     << ','
            << +slrip_7.ffe_fm3                     << ','
            << +slrip_7.ffe_f4                      << ','
            << +slrip_7.ffe_f3                      << ','
            << +slrip_7.ffe_f2                      << ','
            << +slrip_7.ffe_f1                      << ','
            << +slrip_7.ffe_f8                      << ','
            << +slrip_7.ffe_f7                      << ','
            << +slrip_7.ffe_f6                      << ','
            << +slrip_7.ffe_f5                      << ','
            << +slrip_7.dffe_dsel3                  << ','
            << +slrip_7.dffe_coef3                  << ','
            << +slrip_7.dffe_dsel2                  << ','
            << +slrip_7.dffe_coef2                  << ','
            << +slrip_7.dffe_dsel1                  << ','
            << +slrip_7.dffe_coef1                  << ','
            << +slrip_7.dffe_dsel0                  << ','
            << +slrip_7.dffe_coef0                  << ','
            << +slrip_7.dffe_dsel7                  << ','
            << +slrip_7.dffe_coef7                  << ','
            << +slrip_7.dffe_dsel6                  << ','
            << +slrip_7.dffe_coef6                  << ','
            << +slrip_7.dffe_dsel5                  << ','
            << +slrip_7.dffe_coef5                  << ','
            << +slrip_7.dffe_dsel4                  << ','
            << +slrip_7.dffe_coef4                  << ','
            << +slrip_7.dffe_dsel8                  << ','
            << +slrip_7.dffe_coef8                  << ','
            << +slrip_7.th1                         << ','
            << +slrip_7.th0                         << ','
            << +slrip_7.th3                         << ','
            << +slrip_7.th2                         << ','
            << +slrip_7.th5                         << ','
            << +slrip_7.th4                         << ','
            << +slrip_7.th7                         << ','
            << +slrip_7.th6                         << ','
            << +slrip_7.th9                         << ','
            << +slrip_7.th8                         << ','
            << +slrip_7.th11                        << ','
            << +slrip_7.th10                        << ','
            << +slrip_7.th13                        << ','
            << +slrip_7.th12                        << ','
            << +slrip_7.th15                        << ','
            << +slrip_7.th14                        << ','
            << +slrip_7.th17                        << ','
            << +slrip_7.th16                        << ','
            << +slrip_7.th19                        << ','
            << +slrip_7.th18                        << ','
            << +slrip_7.th21                        << ','
            << +slrip_7.th20                        ;

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Dump_5nm(const struct slrip_reg &slrip, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrip_5nm slrip_5;
    slrip_5nm_unpack(&slrip_5, slrip.page_data.slrip_data_set.data);
    sstream << +slrip_5.tah_amp_gain                       << ','
            << +slrip_5.lf_gain                            << ','
            << +slrip_5.lf_pol                             << ','
            << +slrip_5.mf_gain                            << ','
            << +slrip_5.mf_pol                             << ','
            << +slrip_5.dc_gain                            << ','
            << +slrip_5.hf_gain                            << ','
            << +slrip_5.ffe_fm3                            << ','
            << +slrip_5.ffe_fm4                            << ','
            << +slrip_5.ffe_fm5                            << ','
            << +slrip_5.ffe_f1                             << ','
            << +slrip_5.ffe_fm1                            << ','
            << +slrip_5.ffe_fm2                            << ','
            << +slrip_5.ffe_f4                             << ','
            << +slrip_5.ffe_f3                             << ','
            << +slrip_5.ffe_f2                             << ','
            << +slrip_5.ffe_f8                             << ','
            << +slrip_5.ffe_f7                             << ','
            << +slrip_5.ffe_f6                             << ','
            << +slrip_5.ffe_f5                             << ','
            << +slrip_5.ffe_f12                            << ','
            << +slrip_5.ffe_f11                            << ','
            << +slrip_5.ffe_f10                            << ','
            << +slrip_5.ffe_f9                             << ','
            << +slrip_5.dffe_coef2                         << ','
            << +slrip_5.dffe_dsel2                         << ','
            << +slrip_5.dffe_coef1                         << ','
            << +slrip_5.dffe_dsel1                         << ','
            << +slrip_5.dffe_coef0                         << ','
            << +slrip_5.dffe_dsel0                         << ','
            << +slrip_5.dffe_coef5                         << ','
            << +slrip_5.dffe_dsel5                         << ','
            << +slrip_5.dffe_coef4                         << ','
            << +slrip_5.dffe_dsel4                         << ','
            << +slrip_5.dffe_coef3                         << ','
            << +slrip_5.dffe_dsel3                         << ','
            << +slrip_5.dffe_coef8                         << ','
            << +slrip_5.dffe_dsel8                         << ','
            << +slrip_5.dffe_coef7                         << ','
            << +slrip_5.dffe_dsel7                         << ','
            << +slrip_5.dffe_coef6                         << ','
            << +slrip_5.dffe_dsel6                         << ','
            << +slrip_5.dffe_coef11                        << ','
            << +slrip_5.dffe_dsel11                        << ','
            << +slrip_5.dffe_coef10                        << ','
            << +slrip_5.dffe_dsel10                        << ','
            << +slrip_5.dffe_coef9                         << ','
            << +slrip_5.dffe_dsel9                         << ','
            << +slrip_5.dffe_coef14                        << ','
            << +slrip_5.dffe_dsel14                        << ','
            << +slrip_5.dffe_coef13                        << ','
            << +slrip_5.dffe_dsel13                        << ','
            << +slrip_5.dffe_coef12                        << ','
            << +slrip_5.dffe_dsel12                        << ','
            << +slrip_5.fb_slicer_th1                      << ','
            << +slrip_5.fb_slicer_th0                      << ','
            << +slrip_5.dffe_coef15                        << ','
            << +slrip_5.dffe_dsel15                        << ','
            << +slrip_5.fb_slicer_th4                      << ','
            << +slrip_5.fb_slicer_th3                      << ','
            << +slrip_5.fb_slicer_th2                      << ','
            << +slrip_5.fb_slicer_th7                      << ','
            << +slrip_5.fb_slicer_th6                      << ','
            << +slrip_5.fb_slicer_th5                      << ','
            << +slrip_5.fb_slicer_th10                     << ','
            << +slrip_5.fb_slicer_th9                      << ','
            << +slrip_5.fb_slicer_th8                      << ','
            << +slrip_5.ff_slicer_th0                      << ','
            << +slrip_5.fb_slicer_th12                     << ','
            << +slrip_5.fb_slicer_th11                     << ','
            << +slrip_5.ff_slicer_th3                      << ','
            << +slrip_5.ff_slicer_th2                      << ','
            << +slrip_5.ff_slicer_th1                      << ','
            << +slrip_5.ff_slicer_th6                      << ','
            << +slrip_5.ff_slicer_th5                      << ','
            << +slrip_5.ff_slicer_th4                      << ','
            << +slrip_5.ff_slicer_th9                      << ','
            << +slrip_5.ff_slicer_th8                      << ','
            << +slrip_5.ff_slicer_th7                      << ','
            << +slrip_5.ff_slicer_th12                     << ','
            << +slrip_5.ff_slicer_th11                     << ','
            << +slrip_5.ff_slicer_th10                     ;

    for (int i = 0; i < ACC_REG_SLRIP_MAX_FIELDS_NUM - ACC_REG_SLRIP5NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}
void SLRIPRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node,
                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port) {
        struct slrip_reg *p_slrip_reg = new struct slrip_reg(areg.regs.slrip);

        switch (p_slrip_reg->version) {
            case 0: // 16nm
                slrip_16nm_unpack(&p_slrip_reg->page_data.slrip_16nm, areg.regs.slrip.page_data.slrip_data_set.data);
                break;
            case 4: // 7nm
                slrip_7nm_unpack(&p_slrip_reg->page_data.slrip_7nm, areg.regs.slrip.page_data.slrip_data_set.data);
                break;
            case 5: // 5nm
                slrip_5nm_unpack(&p_slrip_reg->page_data.slrip_5nm, areg.regs.slrip.page_data.slrip_data_set.data);
                break;
            default:
                break;
        }

        AccRegKeyPortLane *p_pl_key = (AccRegKeyPortLane *)p_key;
        p_export_data_phy_port->p_slrip_reg[p_pl_key->lane][p_pl_key->idx_in_lane] = p_slrip_reg;
    }

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLREG Register functions     ***
//*****************************************
SLREGRegister::SLREGRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_SLREG,
                (const unpack_data_func_t)slreg_reg_unpack,
                ACC_REG_SLREG_INTERNAL_SECTION_NAME,
                ACC_REG_SLREG_NAME,
                ACC_REG_SLREG7NM_FIELDS_NUM,
                NOT_SUPPORT_SLREG)
{
}

int SLREGRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slreg_reg slreg;
    CLEAR_STRUCT(slreg);
    slreg.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slreg.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slreg.pnat = ACC_REG_PNAT_IB_PORT;
    slreg.eidx = ((AccRegKeyPortLane *)p_key)->idx_in_lane;
    slreg_reg_pack(&slreg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLREGRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) 
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

#endif

    IBDIAGNET_RETURN_VOID;
}

void SLREGRegister::Header_Dump_16nm(stringstream& sstream) const
{
    IBDIAGNET_ENTER;
    sstream << "eidx"                                   << ','
            << "status"                                 << ','
            << "version_16nm=0"                         << ','
            << "local_port"                             << ','
            << "pnat"                                   << ','
            << "lp_msb"                                 << ','
            << "lane"                                   << ','
            << "port_type"                              << ','
            << "initial_error"                          << ','
            << "eye_grade"                              << ','
            << "post_error"                             << ','
            << "initial_area"                           << ','
            << "post_area"                              << ','
            << "post_area_after_centering"              << ','
            << "gd_iter_num_errors"                     << ','
            << "gd_iter_num_area"                       << ','
            << "post_sel_enc_slave_centering"           << ','
            << "post_sel_enc_slave"                     << ','
            << "initial_sel_enc_slave"                  << ','
            << "ffe_gain"                               << ','
            << "cur_phase_ib"                           << ','
            << "rect_size"                              << ','
            << "sel_enc_tune_area_failure"              << ','
            << "area_zero_event_counter"                << ','
            << "monotonic_reached_low_limit"            << ','
            << "monotonic_reached_high_limit"           << ','
            << "monotonic_flow_height_inconsistent"     << ','
            << "update_single_gain_bad_value"           << ','
            << "init_all_gains_bad_value";

    IBDIAGNET_RETURN_VOID;
}

void SLREGRegister::Header_Dump_7nm(stringstream& sstream) const 
{
    IBDIAGNET_ENTER;
    sstream << "eidx"                           << ','
            << "status"                         << ','
            << "version_7nm=4"                  << ','
            << "local_port"                     << ','
            << "pnat"                           << ','
            << "lp_msb"                         << ','
            << "lane"                           << ','
            << "port_type"                      << ','
            << "eom_en"                         << ','
            << "eom_status"                     << ','
            << "sth_start_mode"                 << ','
            << "eom_vsign"                      << ','
            << "fom_timeout"                    << ','
            << "ber_grad_inv"                   << ','
            << "eom_mode"                       << ','
            << "vos_adc_test_en"                << ','
            << "gos_adc_test_en"                << ','
            << "vos_vga_test_en"                << ','
            << "vos_ctle_test_en"               << ','
            << "latch_phos_test_en"             << ','
            << "latch_adc_vos_test_en"          << ','
            << "latch_adc_gos_test_en"          << ','
            << "fom_retest_mode"                << ','
            << "phos_test_en"                   << ','
            << "latch_test_mode"                << ','
            << "vos_test_mode"                  << ','
            << "fom_test_mode"                  << ','
            << "phos_threshold"                 << ','
            << "fom_threshold"                  << ','
            << "adc_vos_threshold"              << ','
            << "adc_gos_threshold"              << ','
            << "vga_vos_threshold"              << ','
            << "ctle_vos_threshold"             << ','
            << "fom_mode"                       << ','
            << "eom_fmr_sel"                    << ','
            << "eom_fmr_ovrd"                   << ','
            << "fom_norm_max_en"                << ','
            << "eom_grad"                       << ','
            << "ber_mode"                       << ','
            << "nblks_max"                      << ','
            << "nerrs_min"                      << ','
            << "eyeo_range"                     << ','
            << "ber_eye_sel"                    << ','
            << "ber_t"                          << ','
            << "ber_v"                          << ','
            << "margin_en"                      << ','
            << "margin_v"                       << ','
            << "margin_t";

    IBDIAGNET_RETURN_VOID;
}


void SLREGRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct slreg_reg & slreg = areg.regs.slreg;

    sstream << "eidx_" << +slreg.eidx << ','
            << +slreg.status          << ','
            << +slreg.version         << ','
            << +slreg.local_port      << ','
            << +slreg.pnat            << ','
            << +slreg.lp_msb          << ','
            << +slreg.lane            << ','
            << +slreg.port_type       << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (slreg.version) {

        case 0: // 16nm
            Dump_16nm(slreg, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(slreg, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +slreg.version << '"';
            for (int i = 0; i < ACC_REG_SLREG_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLREG: %d, on node: " U64H_FMT ".\n",
                        slreg.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void SLREGRegister::Dump_16nm(const struct slreg_reg &slreg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slreg_16nm slreg_16;
    slreg_16nm_unpack(&slreg_16, slreg.page_data.slreg_data_set.data);

    sstream << +slreg_16.initial_error                           << ','
            << +slreg_16.eye_grade                               << ','
            << +slreg_16.post_error                              << ','
            << +slreg_16.initial_area                            << ','
            << +slreg_16.post_area                               << ','
            << +slreg_16.post_area_after_centering               << ','
            << +slreg_16.gd_iter_num_errors                      << ','
            << +slreg_16.gd_iter_num_area                        << ','
            << +slreg_16.post_sel_enc_slave_centering            << ','
            << +slreg_16.post_sel_enc_slave                      << ','
            << +slreg_16.initial_sel_enc_slave                   << ','
            << +slreg_16.ffe_gain                                << ','
            << +slreg_16.cur_phase_ib                            << ','
            << +slreg_16.rect_size                               << ','
            << +slreg_16.sel_enc_tune_area_failure               << ','
            << +slreg_16.area_zero_event_counter                 << ','
            << +slreg_16.monotonic_reached_low_limit             << ','
            << +slreg_16.monotonic_reached_high_limit            << ','
            << +slreg_16.monotonic_flow_height_inconsistent      << ','
            << +slreg_16.update_single_gain_bad_value            << ','
            << +slreg_16.init_all_gains_bad_value                << ','
            << "NA,NA,NA,NA,NA,NA,NA,NA,NA,NA"                   << ','
            << "NA,NA,NA,NA,NA,NA,NA,NA,NA"                      ;  // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLREGRegister::Dump_7nm(const struct slreg_reg &slreg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slreg_7nm slreg_7;
    slreg_7nm_unpack(&slreg_7, slreg.page_data.slreg_data_set.data);

    sstream << +slreg_7.eom_en                          << ','
            << +slreg_7.eom_status                      << ','
            << +slreg_7.sth_start_mode                  << ','
            << +slreg_7.eom_vsign                       << ','
            << +slreg_7.fom_timeout                     << ','
            << +slreg_7.ber_grad_inv                    << ','
            << +slreg_7.eom_mode                        << ','
            << +slreg_7.vos_adc_test_en                 << ','
            << +slreg_7.gos_adc_test_en                 << ','
            << +slreg_7.vos_vga_test_en                 << ','
            << +slreg_7.vos_ctle_test_en                << ','
            << +slreg_7.latch_phos_test_en              << ','
            << +slreg_7.latch_adc_vos_test_en           << ','
            << +slreg_7.latch_adc_gos_test_en           << ','
            << +slreg_7.fom_retest_mode                 << ','
            << +slreg_7.phos_test_en                    << ','
            << +slreg_7.latch_test_mode                 << ','
            << +slreg_7.vos_test_mode                   << ','
            << +slreg_7.fom_test_mode                   << ','
            << +slreg_7.phos_threshold                  << ','
            << +slreg_7.fom_threshold                   << ','
            << +slreg_7.adc_vos_threshold               << ','
            << +slreg_7.adc_gos_threshold               << ','
            << +slreg_7.vga_vos_threshold               << ','
            << +slreg_7.ctle_vos_threshold              << ','
            << +slreg_7.fom_mode                        << ','
            << +slreg_7.eom_fmr_sel                     << ','
            << +slreg_7.eom_fmr_ovrd                    << ','
            << +slreg_7.fom_norm_max_en                 << ','
            << +slreg_7.eom_grad                        << ','
            << +slreg_7.ber_mode                        << ','
            << +slreg_7.nblks_max                       << ','
            << +slreg_7.nerrs_min                       << ','
            << +slreg_7.eyeo_range                      << ','
            << +slreg_7.ber_eye_sel                     << ','
            << +slreg_7.ber_t                           << ','
            << +slreg_7.ber_v                           << ','
            << +slreg_7.margin_en                       << ','
            << +slreg_7.margin_v                        << ','
            << +slreg_7.margin_t                        ;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLCCT Register functions     ***
//*****************************************
SLCCTRegister::SLCCTRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_SLCCT,
                (const unpack_data_func_t)slcct_reg_unpack,
                ACC_REG_SLCCT_INTERNAL_SECTION_NAME,
                ACC_REG_SLCCT_NAME,
                ACC_REG_SLCCT_FIELDS_NUM,
                NOT_SUPPORT_SLCCT)
{
}

int SLCCTRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slcct_reg slcct;
    CLEAR_STRUCT(slcct);
    slcct.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slcct.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slcct.pnat = ACC_REG_PNAT_IB_PORT;
    slcct.conf_index = ((AccRegKeyPortLane *)p_key)->idx_in_lane? SLCCT_CTLE_ARR_SIZE : 0;
    slcct.num_of_idx = SLCCT_CTLE_ARR_SIZE;

    slcct_reg_pack(&slcct, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLCCTRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "pnat"             << ','
            << "version"          << ','
            << "status"           << ','
            << "conf_index"       << ','
            << "num_of_idx";

    for (size_t i = 0; i < ARRAY_SIZE(slcct_reg::ctle_calib_set); ++i)
        sstream << ',' << "ctle_calib_set[" << i << "].bgn_offset_n0"
                << ',' << "ctle_calib_set[" << i << "].bgn_offset_p0"
                << ',' << "ctle_calib_set[" << i << "].bgn_offset_n1"
                << ',' << "ctle_calib_set[" << i << "].bgn_offset_p1";
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLCCTRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct slcct_reg & p_slcct = areg.regs.slcct;

    sstream << +p_slcct.pnat << ','
            << +p_slcct.version << ','
            << +p_slcct.status << ','
            << +p_slcct.conf_index << ','
            << +p_slcct.num_of_idx;

    for (int i = 0; i < p_slcct.num_of_idx; ++i) {
        sstream << ',' << +p_slcct.ctle_calib_set[i].bgn_offset_n0
                << ',' << +p_slcct.ctle_calib_set[i].bgn_offset_p0
                << ',' << +p_slcct.ctle_calib_set[i].bgn_offset_n1
                << ',' << +p_slcct.ctle_calib_set[i].bgn_offset_p1;
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      PPAMP Register functions     ***
//*****************************************
PPAMPRegister::PPAMPRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PPAMP,
                (const unpack_data_func_t)ppamp_reg_unpack,
                ACC_REG_PPAMP_INTERNAL_SECTION_NAME,
                ACC_REG_PPAMP_NAME,
                ACC_REG_PPAMP_FIELDS_NUM,
                NOT_SUPPORT_PPAMP)
{
}

int PPAMPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct ppamp_reg ppamp;
    CLEAR_STRUCT(ppamp);
    ppamp.opamp_group_type = ((AccRegKeyTypeGroup *)p_key)->group_type;
    ppamp.opamp_group = ((AccRegKeyTypeGroup *)p_key)->group_num;
    ppamp.start_index = ((AccRegKeyTypeGroup *)p_key)->start_index;
    ppamp.num_of_indices = ((AccRegKeyTypeGroup *)p_key)->num_of_indices;
    ppamp_reg_pack(&ppamp, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPAMPRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "max_opamp_group"    << ','
            << "max_num_of_indices";
    
    for (int i = 0; i < PPAMP_IDX_DATA_SIZE; ++i)
        sstream << ',' << "index_data[" << i << ']';

#endif

    IBDIAGNET_RETURN_VOID;
}

void PPAMPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct ppamp_reg & p_ppamp = areg.regs.ppamp;

    sstream << +p_ppamp.max_opamp_group << ','
            << p_ppamp.max_num_of_indices;

    for (int i = 0; i < PPAMP_IDX_DATA_SIZE; ++i) {
        sstream << ',' << p_ppamp.index_data[i];
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void PPAMPRegister::GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices)
{
    IBDIAGNET_ENTER;

    group = areg.regs.ppamp.max_opamp_group;
    indices = areg.regs.ppamp.max_num_of_indices;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* PDDR Register functions ***
//*****************************************
int PDDRRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer)
{
    IBDIAGNET_ENTER;

    pddr_reg tmp;
    pddr_reg_unpack(&tmp, unpacked_buffer);

    unpack_data_func(data_to_unpack, tmp.page_data.raw.data);

    reg_raw_data *p_raw_data = reinterpret_cast<reg_raw_data *>(data_to_unpack);
    p_raw_data->module_info_ext = tmp.module_info_ext;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PDDRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pddr_reg pddr;
    CLEAR_STRUCT(pddr);

    pddr.port_type          = this->m_port_type;
    pddr.lp_msb             = this->m_lp_msb;
    pddr.pnat               = this->m_pnat;
    pddr.local_port         = ((AccRegKeyPort *)p_key)->port_num;
    pddr.page_select        = this->m_page_select;
    pddr.module_info_ext    = this->m_module_info_ext;

    pddr_reg_pack(&pddr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


PDDRModuleInfoRegister::PDDRModuleInfoRegister(PhyDiag * phy_diag)
    : PDDRRegister(phy_diag,
                   0x03,
                   (const unpack_data_func_t)DDModuleInfo_unpack,
                   ACC_REG_PDDR_MODULE_INTERNAL_SECTION_NAME,
                   ACC_REG_PDDR_MODULE_NAME,
                   ACC_REG_PDDR_MODULE_FIELDS_NUM,
                   ACC_REG_PDDR_MODULE_HEADER,
                   SUPPORT_SW_CA,
                   false,
                   phy_diag->GetToGetCableDisconnectedPorts())
{
    this->m_module_info_ext = 1;
}

int PDDRModuleInfoRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    AccRegKeyPort * key = dynamic_cast<AccRegKeyPort*>(p_key);

    if (!key || !p_node)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    IBPort * p_port = p_node->getPort(key->port_num);

    if (!p_port)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    if (p_port->p_combined_cable && !m_phy_diag->GetPhyCableFullData())
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    IBDIAGNET_RETURN(PDDRRegister::PackData(p_key, data, p_node));
}

void PDDRModuleInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ','
            << "cable_technology"                  << ','
            << "cable_breakout"                    << ','
            << "ext_ethernet_compliance_code"      << ','
            << "ethernet_compliance_code"          << ','
            << "cable_type"                        << ','
            << "cable_vendor"                      << ','
            << "cable_length"                      << ','
            << "cable_identifier"                  << ','
            << "cable_power_class"                 << ','
            << "max_power"                         << ','
            << "cable_rx_amp"                      << ','
            << "cable_rx_emphasis"                 << ','
            << "cable_tx_equalization"             << ','
            << "cable_attenuation_25g"             << ','
            << "cable_attenuation_12g"             << ','
            << "cable_attenuation_7g"              << ','
            << "cable_attenuation_5g"              << ','
            << "cable_rx_post_emphasis"            << ','
            << "rx_cdr_cap"                        << ','
            << "tx_cdr_cap"                        << ','
            << "rx_cdr_state"                      << ','
            << "tx_cdr_state"                      << ','
            << "vendor_name"                       << ','
            << "vendor_pn"                         << ','
            << "module_info"                       << ','
            << "+fw_hi.fw_mid.fw_lo"               << ','
            << "vendor_sn"                         << ','
            << "temperature"                       << ','
            << "voltage"                           << ','
            << "rx_power_lane0"                    << ','
            << "rx_power_lane1"                    << ','
            << "rx_power_lane2"                    << ','
            << "rx_power_lane3"                    << ','
            << "rx_power_lane4"                    << ','
            << "rx_power_lane5"                    << ','
            << "rx_power_lane6"                    << ','
            << "rx_power_lane7"                    << ','
            << "tx_power_lane0"                    << ','
            << "tx_power_lane1"                    << ','
            << "tx_power_lane2"                    << ','
            << "tx_power_lane3"                    << ','
            << "tx_power_lane4"                    << ','
            << "tx_power_lane5"                    << ','
            << "tx_power_lane6"                    << ','
            << "tx_power_lane7"                    << ','
            << "tx_bias_lane0"                     << ','
            << "tx_bias_lane1"                     << ','
            << "tx_bias_lane2"                     << ','
            << "tx_bias_lane3"                     << ','
            << "tx_bias_lane4"                     << ','
            << "tx_bias_lane5"                     << ','
            << "tx_bias_lane6"                     << ','
            << "tx_bias_lane7"                     << ','
            << "temperature_high_th"               << ','
            << "temperature_low_th"                << ','
            << "voltage_high_th"                   << ','
            << "voltage_low_th"                    << ','
            << "rx_power_high_th"                  << ','
            << "rx_power_low_th"                   << ','
            << "tx_power_high_th"                  << ','
            << "tx_power_low_th"                   << ','
            << "tx_bias_high_th"                   << ','
            << "tx_bias_low_th"                    << ','
            << "module_st"                         << ','
            << "rx_power_type"                     << ','
            << "did_cap"                           << ','
            << "smf_length"                        << ','
            << "wavelength"                        << ','
            << "active_set_host_compliance_code"   << ','
            << "active_set_media_compliance_code"  << ','
            << "ib_compliance_code"                << ','
            << "nbr_250"                           << ','
            << "nbr_100"                           << ','
            << "monitor_cap_mask"                  << ','
            << "ib_width"                          << ','
            << "dp_st_lane0"                       << ','
            << "dp_st_lane1"                       << ','
            << "dp_st_lane2"                       << ','
            << "dp_st_lane3"                       << ','
            << "dp_st_lane4"                       << ','
            << "dp_st_lane5"                       << ','
            << "dp_st_lane6"                       << ','
            << "dp_st_lane7"                       << ','
            << "length_om2"                        << ','
            << "length_om3"                        << ','
            << "length_om4"                        << ','
            << "length_om5"                        << ','
            << "length_om1"                        << ','
            << "wavelength_tolerance"              << ','
            << "memory_map_rev"                    << ','
            << "memory_map_compliance"             << ','
            << "date_code"                         << ','
            << "connector_type"                    << ','
            << "vendor_oui"                        << ','
            << "rx_input_valid"                    << ','
            << "rx_output_valid"                   << ','
            << "tx_input_freq_sync"                << ','
            << "error_code"                        << ','
            << "cdr_vendor"                        << ','
            << "max_fiber_length";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PDDRModuleInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    DiagnosticDataModuleInfo::DumpModuleInfoRawData(sstream, areg.regs.pddr_module);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

PDDRLinkUpInfoRegister::PDDRLinkUpInfoRegister(PhyDiag * phy_diag)
    : PDDRRegister(phy_diag, 
                   0x08, 
                   (const unpack_data_func_t)pddr_link_up_info_page_unpack,
                   ACC_REG_PDDR_LINKUP_INTERNAL_SECTION_NAME,
                   ACC_REG_PDDR_LINKUP_NAME,
                   ACC_REG_PDDR_LINKUP_FIELDS_NUM,
                   ACC_REG_PDDR_LINKUP_HEADER,
                   SUPPORT_SW_CA)
{
}

void PDDRLinkUpInfoRegister::DumpLayout(stringstream &sstream, const struct pddr_link_up_info_page &linkup_info)
{
    IBDIAGNET_ENTER;
    sstream << +linkup_info.up_reason_mng << ','
            << +linkup_info.up_reason_drv << ','
            << +linkup_info.up_reason_pwr << ','
            << +linkup_info.time_to_link_up << ','
            << +linkup_info.lt_cnt << ','
            << +linkup_info.fast_link_up_status << ','
            << +linkup_info.time_to_link_up_phy_up_to_active << ','
            << +linkup_info.time_to_link_up_sd_to_phy_up << ','
            << +linkup_info.time_to_link_up_disable_to_sd << ','
            << +linkup_info.time_to_link_up_disable_to_pd;
    IBDIAGNET_RETURN_VOID;
}

void PDDRLinkUpInfoRegister::DumpLinkUpInfoRealHeader(stringstream &sstream, const string& header)
{
    IBDIAGNET_ENTER;
#ifdef SHOW_PHY_HEADERS
    sstream << '#' << header                         << ','
            << "up_reason_mng"                       << ','
            << "up_reason_drv"                       << ','
            << "up_reason_pwr"                       << ','
            << "time_to_link_up"                     << ','
            << "lt_cnt"                              << ','
            << "fast_link_up_status"                 << ','
            << "time_to_link_up_phy_up_to_active"    << ','
            << "time_to_link_up_sd_to_phy_up"        << ','
            << "time_to_link_up_disable_to_sd"       << ','
            << "time_to_link_up_disable_to_pd";

#endif
    IBDIAGNET_RETURN_VOID;
}

void PDDRLinkUpInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
    sstream << std::endl;
    DumpLinkUpInfoRealHeader(sstream, handler_header);
    IBDIAGNET_RETURN_VOID;
}

void PDDRLinkUpInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    DumpLayout(sstream, areg.regs.pddr_link_up);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

PDDRLatchedFlagInfoRegister::PDDRLatchedFlagInfoRegister(PhyDiag * phy_diag)
    : PDDRRegister(phy_diag,
                   0x09,
                   (const unpack_data_func_t)DDLatchedFlagInfo_unpack,
                   ACC_REG_PDDR_LATCHED_INTERNAL_SECTION_NAME,
                   ACC_REG_PDDR_LATCHED_NAME,
                   ACC_REG_PDDR_LATCHED_FIELDS_NUM,
                   ACC_REG_PDDR_LATCHED_HEADER,
                   SUPPORT_SW_CA,
                   false,
                   phy_diag->GetToGetCableDisconnectedPorts())
{
}

int PDDRLatchedFlagInfoRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    AccRegKeyPort * key = dynamic_cast<AccRegKeyPort*>(p_key);

    if (!key || !p_node)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    IBPort * p_port = p_node->getPort(key->port_num);

    if (!p_port)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    if (p_port->p_combined_cable && !m_phy_diag->GetPhyCableFullData())
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    IBDIAGNET_RETURN(PDDRRegister::PackData(p_key, data, p_node));
}

void PDDRLatchedFlagInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                       << '#'
            << handler_header             << ','
            << "dp_fw_fault"              << ','
            << "mod_fw_fault"             << ','
            << "vcc_flags"                << ','
            << "temp_flags"               << ','
            << "tx_ad_eq_fault"           << ','
            << "tx_cdr_lol"               << ','
            << "tx_los"                   << ','
            << "tx_fault"                 << ','
            << "tx_power_lo_war"          << ','
            << "tx_power_hi_war"          << ','
            << "tx_power_lo_al"           << ','
            << "tx_power_hi_al"           << ','
            << "tx_bias_lo_war"           << ','
            << "tx_bias_hi_war"           << ','
            << "tx_bias_lo_al"            << ','
            << "tx_bias_hi_al"            << ','
            << "rx_cdr_lol"               << ','
            << "rx_los"                   << ','
            << "rx_power_lo_war"          << ','
            << "rx_power_hi_war"          << ','
            << "rx_power_lo_al"           << ','
            << "rx_power_hi_al"           << ','
            << "rx_output_valid_change"   << ','
            << "rx_input_valid_change";
#endif
    IBDIAGNET_RETURN_VOID;
}


void PDDRLatchedFlagInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoRawData(sstream, areg.regs.pddr_latched);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

PDDRPhyInfoRegister::PDDRPhyInfoRegister(PhyDiag * phy_diag)
    :PDDRRegister(phy_diag, 
                  0x02, 
                  (const unpack_data_func_t)pddr_phy_info_page_unpack, 
                  ACC_REG_PDDR_PHY_INFO_INTERNAL_SECTION_NAME, 
                  ACC_REG_PDDR_PHY_INFO_NAME, 
                  ACC_REG_PDDR_PHY_INFO_FIELDS_NUM,
                  ACC_REG_PDDR_PHY_INFO_HEADER, 
                  SUPPORT_SW_CA)
{

}
void PDDRPhyInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
    sstream << std::endl;
    DumpPhyInfoRealHeader(sstream, handler_header);
    IBDIAGNET_RETURN_VOID;
}

void PDDRPhyInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    DumpLayout(sstream, areg.regs.pddr_phyinfo);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PDDRPhyInfoRegister::DumpPhyInfoRealHeader(stringstream &sstream, const string& header)
{
#ifdef SHOW_PHY_HEADERS
    sstream << '#' << header                               << ','
            << "remote_device_type"                        << ','
            << "negotiation_mask"                          << ','
            << "port_notifications"                        << ','
            << "num_of_negotiation_attempts"               << ','
            << "ib_revision"                               << ','
            << "lp_ib_revision"                            << ','
            << "hw_link_phy_state"                         << ','
            << "phy_manger_disable_mask"                   << ','
            << "pcs_phy_state"                             << ','
            << "lp_proto_enabled"                          << ','
            << "lp_fec_mode_support"                       << ','
            << "lp_fec_mode_request"                       << ','
            << "ib_last_link_down_reason"                  << ','
            << "eth_last_link_down_lane"                   << ','
            << "speed_deg_db"                              << ','
            << "degrade_grade_lane_0"                      << ','
            << "degrade_grade_lane_1"                      << ','
            << "degrade_grade_lane_2"                      << ','
            << "degrade_grade_lane_3"                      << ','
            << "kr_startup_debug_indications_4"            << ','
            << "kr_startup_debug_indications_5"            << ','
            << "kr_startup_debug_indications_6"            << ','
            << "kr_startup_debug_indications_7"            << ','
            << "num_of_presets_tested_lane0"               << ','
            << "num_of_presets_tested_lane1"               << ','
            << "num_of_presets_tested_lane2"               << ','
            << "num_of_presets_tested_lane3"               << ','
            << "num_of_presets_tested_lane_4"              << ','
            << "num_of_presets_tested_lane_5"              << ','
            << "num_of_presets_tested_lane_6"              << ','
            << "num_of_presets_tested_lane_7"              << ','
            << "kr_startup_fsm_lane"                       << ','
            << "eth_an_debug_indication"                   << ','
            << "cdr_not_locked_cnt"                        << ','
            << "fall_from_cfg_idle_cnt"                    << ','
            << "fall_from_cfg_idle_cdr_cnt"                << ','
            << "eth_an_watchdog_cnt"                       << ','
            << "rounds_waited_for_peer_to_end_test"        << ','
            << "ib_phy_fsm_state_trace"                    << ','
            << "kr_startup_debug_indications_0"            << ','
            << "kr_startup_debug_indications_1"            << ','
            << "kr_startup_debug_indications_2"            << ','
            << "kr_startup_debug_indications_3"            << ','
            << "tx_tuning_stages_lane_0"                   << ','
            << "tx_tuning_stages_lane_1"                   << ','
            << "tx_tuning_stages_lane_2"                   << ','
            << "tx_tuning_stages_lane_3"                   << ','
            << "tx_tuning_stages_lane_4"                   << ','
            << "tx_tuning_stages_lane_5"                   << ','
            << "tx_tuning_stages_lane_6"                   << ','
            << "tx_tuning_stages_lane_7"                   << ','
            << "plu_rx_polarity"                           << ','
            << "plu_tx_polarity"                           << ','
            << "plu_rx_pwrup"                              << ','
            << "plu_tx_pwrup"                              << ','
            << "signal_detected"                           << ','
            << "plr_outstanding_cells"                     << ','
            << "plu_rx_speed"                              << ','
            << "plu_tx_speed"                              << ','
            << "sd_valid"                                  << ','
            << "ib_cfg_delay_timeout"                      << ','
            << "irisc_status"                              << ','
            << "stamping_reason"                           << ','
            << "kr_frame_lock_tuning_failure_events_count" << ','
            << "kr_full_tuning_failure_count"              << ','
            << "ib_debug_indication"                       << ','
            << "pm_debug_indication"                       << ','
            << "pd_cat_val"                                << ','
            << "pd_catastrophic_enum"                      << ','
            << "hst_cat_val"                               << ','
            << "hst_catastrophic_enum"                     << ','
            << "an_cat_val"                                << ','
            << "an_catastrophic_enum"                      << ','
            << "pm_cat_val"                                << ','
            << "pm_catastrophic_enum"                      << ','
            << "pd_debug_indication"                       << ','
            << "psi_collision1"                            << ','
            << "psi_collision2"                            << ','
            << "hst_mismatch_reason"                       << ','
            << "fp_signal_detect_count"                    << ','
            << "pd_count"                                  << ','
            << "nlpn_debug_ind_mask"                       << ','
            << "phy2mod_speed_req"                         << ','
            << "eeprom_prsnt"                              << ','
            << "no_dme_mod"                                << ','
            << "one_pll_mod"                               << ','
            << "phy2mod_ack_lanes"                         << ','
            << "phy2mod_deactivate_lanes"                  << ','
            << "port_xmit_pkts_inc_s1"                     << ','
            << "port_xmit_pkts_inc_s0"                     << ','
            << "cnt_rx_frame_received_ok_s1"               << ','
            << "cnt_rx_frame_received_ok_s0"               << ','
            << "macsec_en_plt1_s1"                         << ','
            << "macsec_en_plt1_s0"                         << ','
            << "macsec_en_plt0_s1"                         << ','
            << "macsec_en_plt0_s0"                         << ','
            << "reconciliation_mux_plt1"                   << ','
            << "reconciliation_mux_plt0"                   << ','
            << "tx_bypass_mux_plt1"                        << ','
            << "tx_bypass_mux_plt0"                        << ','
            << "rx_bypass_mux_plt1"                        << ','
            << "rx_bypass_mux_plt0"                        << ','
            << "plr_rtt_hdr_threshold"                     << ','
            << "plr_rtt_ndr_threshold"                     << ','
            << "plr_rtt_xdr_threshold";
#endif
}

void PDDRPhyInfoRegister::DumpLayout(stringstream &sstream, const struct pddr_phy_info_page &phy_info)
{
    IBDIAGNET_ENTER;
    #define ARRAY_TO_32_BIT(arr) (u_int32_t)(arr[0] << 24 | (arr[1] << 16) | (arr[2] << 8) | (arr[3]))
    #define ARRAY_TO_64_BIT(arr) (((uint64_t)(((uint8_t *)(arr))[7]) <<  0)+\
                    ((uint64_t)(((uint8_t *)(arr))[6]) <<  8)+\
                    ((uint64_t)(((uint8_t *)(arr))[5]) << 16)+\
                    ((uint64_t)(((uint8_t *)(arr))[4]) << 24)+\
                    ((uint64_t)(((uint8_t *)(arr))[3]) << 32)+\
                    ((uint64_t)(((uint8_t *)(arr))[2]) << 40)+\
                    ((uint64_t)(((uint8_t *)(arr))[1]) << 48)+\
                    ((uint64_t)(((uint8_t *)(arr))[0]) << 56))
    sstream << +phy_info.remote_device_type << ','
            << +phy_info.negotiation_mask << ','
            << +phy_info.port_notifications << ','
            << phy_info.num_of_negotiation_attempts << ','
            << +phy_info.ib_revision << ','
            << +phy_info.lp_ib_revision << ','
            << +phy_info.hw_link_phy_state << ','
            << phy_info.phy_manager_disable_mask << ','
            << phy_info.pcs_phy_state << ','
            << phy_info.lp_proto_enabled << ','
            << phy_info.lp_fec_mode_support << ','
            << phy_info.lp_fec_mode_request << ','
            << phy_info.ib_last_link_down_reason << ','
            << PTR(ARRAY_TO_32_BIT(phy_info.eth_last_link_down_lane)) << ','
            << phy_info.speed_deg_db << ','
            << phy_info.degrade_grade_lane_0 << ','
            << phy_info.degrade_grade_lane_1 << ','
            << phy_info.degrade_grade_lane_2 << ','
            << phy_info.degrade_grade_lane_3 << ','
            << +phy_info.num_of_presets_tested_lane_4 << ','
            << +phy_info.num_of_presets_tested_lane_5 << ','
            << +phy_info.num_of_presets_tested_lane_6 << ','
            << +phy_info.num_of_presets_tested_lane_7 << ','
            << +phy_info.kr_startup_debug_indications_4 << ','
            << +phy_info.kr_startup_debug_indications_5 << ','
            << +phy_info.kr_startup_debug_indications_6 << ','
            << +phy_info.kr_startup_debug_indications_7 << ','
            << +phy_info.num_of_presets_tested_lane0 << ','
            << +phy_info.num_of_presets_tested_lane1 << ','
            << +phy_info.num_of_presets_tested_lane2 << ','
            << +phy_info.num_of_presets_tested_lane3 << ','
            << PTR(ARRAY_TO_64_BIT(phy_info.kr_startup_fsm_lane)) << ','
            << +phy_info.eth_an_debug_indication << ','
            << +phy_info.cdr_not_locked_cnt << ','
            << +phy_info.fall_from_cfg_idle_cnt << ','
            << +phy_info.fall_from_cfg_idle_cdr_cnt << ','
            << +phy_info.eth_an_watchdog_cnt << ','
            << +phy_info.rounds_waited_for_peer_to_end_test << ','
            << phy_info.ib_phy_fsm_state_trace << ','
            << +phy_info.kr_startup_debug_indications_0 << ','
            << +phy_info.kr_startup_debug_indications_1 << ','
            << +phy_info.kr_startup_debug_indications_2 << ','
            << +phy_info.kr_startup_debug_indications_3 << ','
            << +phy_info.tx_tuning_stages_lane[0] << ','
            << +phy_info.tx_tuning_stages_lane[1] << ','
            << +phy_info.tx_tuning_stages_lane[2] << ','
            << +phy_info.tx_tuning_stages_lane[3] << ','
            << +phy_info.tx_tuning_stages_lane[4] << ','
            << +phy_info.tx_tuning_stages_lane[5] << ','
            << +phy_info.tx_tuning_stages_lane[6] << ','
            << +phy_info.tx_tuning_stages_lane[7] << ','
            << +phy_info.plu_rx_polarity << ','
            << +phy_info.plu_tx_polarity << ','
            << +phy_info.plu_rx_pwrup << ','
            << +phy_info.plu_tx_pwrup << ','
            << +phy_info.signal_detected << ','
            << +phy_info.plr_outstanding_cells << ','
            << +phy_info.plu_rx_speed << ','
            << +phy_info.plu_tx_speed << ','
            << +phy_info.sd_valid << ','
            << +phy_info.ib_cfg_delay_timeout << ','
            << +phy_info.irisc_status << ','
            << phy_info.stamping_reason << ','
            << +phy_info.kr_frame_lock_tuning_failure_events_count << ','
            << +phy_info.kr_full_tuning_failure_count << ','
            << +phy_info.ib_debug_indication << ','
            << +phy_info.pm_debug_indication << ','
            << +phy_info.pd_cat_val << ','
            << +phy_info.pd_catastrophic_enum << ','
            << +phy_info.hst_cat_val << ','
            << +phy_info.hst_catastrophic_enum << ','
            << +phy_info.an_cat_val << ','
            << +phy_info.an_catastrophic_enum << ','
            << +phy_info.pm_cat_val << ','
            << +phy_info.pm_catastrophic_enum << ','
            << +phy_info.pd_debug_indication << ','
            << +phy_info.psi_collision1 << ','
            << +phy_info.psi_collision2 << ','
            << +phy_info.hst_mismatch_reason << ','
            << +phy_info.fp_signal_detect_count << ','
            << +phy_info.pd_count << ','
            << +phy_info.nlpn_debug_ind_mask << ','
            << +phy_info.phy2mod_speed_req << ','
            << +phy_info.eeprom_prsnt << ','
            << +phy_info.no_dme_mod << ','
            << +phy_info.one_pll_mod << ','
            << +phy_info.phy2mod_ack_lanes << ','
            << +phy_info.phy2mod_deactivate_lanes << ','
            << +phy_info.port_xmit_pkts_inc_s1 << ','
            << +phy_info.port_xmit_pkts_inc_s0 << ','
            << +phy_info.cnt_rx_frame_received_ok_s1 << ','
            << +phy_info.cnt_rx_frame_received_ok_s0 << ','
            << +phy_info.macsec_en_plt1_s1 << ','
            << +phy_info.macsec_en_plt1_s0 << ','
            << +phy_info.macsec_en_plt0_s1 << ','
            << +phy_info.macsec_en_plt0_s0 << ','
            << +phy_info.reconciliation_mux_plt1 << ','
            << +phy_info.reconciliation_mux_plt0 << ','
            << +phy_info.tx_bypass_mux_plt1 << ','
            << +phy_info.tx_bypass_mux_plt0 << ','
            << +phy_info.rx_bypass_mux_plt1 << ','
            << +phy_info.rx_bypass_mux_plt0 << ','
            << +phy_info.plr_rtt_hdr_threshold << ','
            << +phy_info.plr_rtt_ndr_threshold << ','
            << +phy_info.plr_rtt_xdr_threshold;
    #undef ARRAY_TO_32_BIT
    #undef ARRAY_TO_64_BIT
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* PEMI Register functions ***
//*****************************************
int PEMIRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer) {
    IBDIAGNET_ENTER;

    pemi_reg *reg = (pemi_reg *) data_to_unpack;
    pemi_reg_unpack(reg, unpacked_buffer);

    const AccRegKeyPort *key = dynamic_cast<const AccRegKeyPort*>(p_key);

    if (key) {
        IBPort *port = this->m_phy_diag->GetPort(key->node_guid, key->port_num);

        if (port && port->p_phy_data) {
            PHYPortData *phy_data = dynamic_cast<PHYPortData*>(port->p_phy_data);

            if (phy_data)
                if (phy_data->pemi_group_cap_mask == std::numeric_limits<PHYPortData::pemi_group_cap_mask_t>::max())
                    phy_data->pemi_group_cap_mask = reg->group_cap_mask;
        }
    }

    if (reg->status)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NOT_SUPPORTED);

    pemi_RAW_page_data raw = reg->page_data.raw;
    unpack_data_func(&reg->page_data, raw.data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PEMIRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    const AccRegKeyPort *key = dynamic_cast<const AccRegKeyPort*>(p_key);

    if (key) {
        IBPort *port = this->m_phy_diag->GetPort(key->node_guid, key->port_num);

        if (port && port->p_phy_data) {
            PHYPortData *phy_data = dynamic_cast<PHYPortData *>(port->p_phy_data);

            if (phy_data) // bit is coded in low 4 bits of page select
                if (phy_data->pemi_group_cap_mask & (1 << (this->m_page_select & 0x0F))) {

                    struct pemi_reg pemi;
                    CLEAR_STRUCT(pemi);

                    pemi.lp_msb             = this->m_lp_msb;
                    pemi.pnat               = this->m_pnat;
                    pemi.local_port         = ((AccRegKeyPort *)p_key)->port_num;
                    pemi.page_select        = this->m_page_select;
                    pemi.module_info_ext    = this->m_module_info_ext;

                    pemi_reg_pack(&pemi, data);

                    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
                }
        }
    }

    IBDIAG_RETURN(IBDIAG_ERR_CODE_DISABLED);
}

PEMI_Module_Samples_Register::PEMI_Module_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x00,
                   (const unpack_data_func_t)pemi_Module_Status_Samples_unpack,
                   ACC_REG_PEMI_MODULE_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_MODULE_S_NAME,
                   ACC_REG_PEMI_MODULE_S_FIELDS_NUM,
                   ACC_REG_PEMI_MODULE_S_HEADER)
{
}

void PEMI_Module_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl              << '#'
            << handler_header    << ','
            << "temperature"     << ','
            << "voltage"         << ','

            << "rx_power_lane0"  << ','
            << "rx_power_lane1"  << ','
            << "rx_power_lane2"  << ','
            << "rx_power_lane3"  << ','
            << "rx_power_lane4"  << ','
            << "rx_power_lane5"  << ','
            << "rx_power_lane6"  << ','
            << "rx_power_lane7"  << ','

            << "tx_power_lane0"  << ','
            << "tx_power_lane1"  << ','
            << "tx_power_lane2"  << ','
            << "tx_power_lane3"  << ','
            << "tx_power_lane4"  << ','
            << "tx_power_lane5"  << ','
            << "tx_power_lane6"  << ','
            << "tx_power_lane7"  << ','

            << "tx_bias_lane0"   << ','
            << "tx_bias_lane1"   << ','
            << "tx_bias_lane2"   << ','
            << "tx_bias_lane3"   << ','
            << "tx_bias_lane4"   << ','
            << "tx_bias_lane5"   << ','
            << "tx_bias_lane6"   << ','
            << "tx_bias_lane7"   << ','

            << "module_st"       << ','

            << "dp_st_lane0"     << ','
            << "dp_st_lane1"     << ','
            << "dp_st_lane2"     << ','
            << "dp_st_lane3"     << ','
            << "dp_st_lane4"     << ','
            << "dp_st_lane5"     << ','
            << "dp_st_lane6"     << ','
            << "dp_st_lane7";
#endif
    IBDIAGNET_RETURN_VOID;
}


void PEMI_Module_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Module_Status_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Module_Status_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(temperature)
        << __FMT__(voltage)

        << __FMT__(rx_power_lane0)
        << __FMT__(rx_power_lane1)
        << __FMT__(rx_power_lane2)
        << __FMT__(rx_power_lane3)
        << __FMT__(rx_power_lane4)
        << __FMT__(rx_power_lane5)
        << __FMT__(rx_power_lane6)
        << __FMT__(rx_power_lane7)

        << __FMT__(tx_power_lane0)
        << __FMT__(tx_power_lane1)
        << __FMT__(tx_power_lane2)
        << __FMT__(tx_power_lane3)
        << __FMT__(tx_power_lane4)
        << __FMT__(tx_power_lane5)
        << __FMT__(tx_power_lane6)
        << __FMT__(tx_power_lane7)

        << __FMT__(tx_bias_lane0)
        << __FMT__(tx_bias_lane1)
        << __FMT__(tx_bias_lane2)
        << __FMT__(tx_bias_lane3)
        << __FMT__(tx_bias_lane4)
        << __FMT__(tx_bias_lane5)
        << __FMT__(tx_bias_lane6)
        << __FMT__(tx_bias_lane7)

        << __FMT__(module_st)

        << __FMT__(dp_st_lane0)
        << __FMT__(dp_st_lane1)
        << __FMT__(dp_st_lane2)
        << __FMT__(dp_st_lane3)
        << __FMT__(dp_st_lane4)
        << __FMT__(dp_st_lane5)
        << __FMT__(dp_st_lane6)
        << __FMT_LAST__(dp_st_lane7);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_Module_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                      export_data_phy_node_t *p_export_data_phy_node,
                                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_module_status_samples =
                                &areg.regs.pemi.page_data.pemi_Module_Status_Samples;

    IBDIAGNET_RETURN_VOID;
}

PEMI_SNR_Samples_Register::PEMI_SNR_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x01,
                   (const unpack_data_func_t)pemi_SNR_Samples_unpack,
                   ACC_REG_PEMI_SNR_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_SNR_S_NAME,
                   ACC_REG_PEMI_SNR_S_FIELDS_NUM,
                   ACC_REG_PEMI_SNR_S_HEADER)
{
}

void PEMI_SNR_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "snr_media_lane0"  << ','
            << "snr_media_lane1"  << ','
            << "snr_media_lane2"  << ','
            << "snr_media_lane3"  << ','
            << "snr_media_lane4"  << ','
            << "snr_media_lane5"  << ','
            << "snr_media_lane6"  << ','
            << "snr_media_lane7"  << ','
            << "snr_host_lane0"   << ','
            << "snr_host_lane1"   << ','
            << "snr_host_lane2"   << ','
            << "snr_host_lane3"   << ','
            << "snr_host_lane4"   << ','
            << "snr_host_lane5"   << ','
            << "snr_host_lane6"   << ','
            << "snr_host_lane7";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PEMI_SNR_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_SNR_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_SNR_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(snr_media_lane0)
        << __FMT__(snr_media_lane1)
        << __FMT__(snr_media_lane2)
        << __FMT__(snr_media_lane3)
        << __FMT__(snr_media_lane4)
        << __FMT__(snr_media_lane5)
        << __FMT__(snr_media_lane6)
        << __FMT__(snr_media_lane7)

        << __FMT__(snr_host_lane0)
        << __FMT__(snr_host_lane1)
        << __FMT__(snr_host_lane2)
        << __FMT__(snr_host_lane3)
        << __FMT__(snr_host_lane4)
        << __FMT__(snr_host_lane5)
        << __FMT__(snr_host_lane6)
        << __FMT_LAST__(snr_host_lane7);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_SNR_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_snr_samples = &areg.regs.pemi.page_data.pemi_SNR_Samples;

    IBDIAGNET_RETURN_VOID;
}


PEMI_Laser_Samples_Register::PEMI_Laser_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x02,
                   (const unpack_data_func_t)pemi_Laser_Monitors_Samples_unpack,
                   ACC_REG_PEMI_LASER_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_LASER_S_NAME,
                   ACC_REG_PEMI_LASER_S_FIELDS_NUM,
                   ACC_REG_PEMI_LASER_S_HEADER)
{
}

void PEMI_Laser_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                           << '#'
            << handler_header                 << ','
            << "laser_age"                    << ','
            << "tec_current"                  << ','
            << "laser_frequency_error_lane0"  << ','
            << "laser_frequency_error_lane1"  << ','
            << "laser_frequency_error_lane2"  << ','
            << "laser_frequency_error_lane3"  << ','
            << "laser_frequency_error_lane4"  << ','
            << "laser_frequency_error_lane5"  << ','
            << "laser_frequency_error_lane6"  << ','
            << "laser_frequency_error_lane7"  << ','
            << "laser_temperature_lane0"      << ','
            << "laser_temperature_lane1"      << ','
            << "laser_temperature_lane2"      << ','
            << "laser_temperature_lane3"      << ','
            << "laser_temperature_lane4"      << ','
            << "laser_temperature_lane5"      << ','
            << "laser_temperature_lane6"      << ','
            << "laser_temperature_lane7";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PEMI_Laser_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Laser_Monitors_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Laser_Monitors_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(laser_age)
        << __FMT__(tec_current)

        << __FMT__(laser_frequency_error_lane0)
        << __FMT__(laser_frequency_error_lane1)
        << __FMT__(laser_frequency_error_lane2)
        << __FMT__(laser_frequency_error_lane3)
        << __FMT__(laser_frequency_error_lane4)
        << __FMT__(laser_frequency_error_lane5)
        << __FMT__(laser_frequency_error_lane6)
        << __FMT__(laser_frequency_error_lane7)

        << __FMT__(laser_temperature_lane0)
        << __FMT__(laser_temperature_lane1)
        << __FMT__(laser_temperature_lane2)
        << __FMT__(laser_temperature_lane3)
        << __FMT__(laser_temperature_lane4)
        << __FMT__(laser_temperature_lane5)
        << __FMT__(laser_temperature_lane6)
        << __FMT_LAST__(laser_temperature_lane7);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_Laser_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                     export_data_phy_node_t *p_export_data_phy_node,
                                                     struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_laser_monitors_samples =
                                &areg.regs.pemi.page_data.pemi_Laser_Monitors_Samples;

    IBDIAGNET_RETURN_VOID;
}

PEMI_PAM4_Samples_Register::PEMI_PAM4_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x03,
                   (const unpack_data_func_t)pemi_PAM4_Level_Transition_Samples_unpack,
                   ACC_REG_PEMI_PAM4_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_PAM4_S_NAME,
                   ACC_REG_PEMI_PAM4_S_FIELDS_NUM,
                   ACC_REG_PEMI_PAM4_S_HEADER)
{
}

void PEMI_PAM4_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                                    <<  '#'
            << handler_header                          << ','
            << "pam4_level_transition_media_lane0"     <<  ','
            << "pam4_level_transition_media_lane1"     <<  ','
            << "pam4_level_transition_media_lane2"     <<  ','
            << "pam4_level_transition_media_lane3"     <<  ','
            << "pam4_level_transition_media_lane4"     <<  ','
            << "pam4_level_transition_media_lane5"     <<  ','
            << "pam4_level_transition_media_lane6"     <<  ','
            << "pam4_level_transition_media_lane7"     <<  ','

            << "pam4_level_transition_host_lane0"      <<  ','
            << "pam4_level_transition_host_lane1"      <<  ','
            << "pam4_level_transition_host_lane2"      <<  ','
            << "pam4_level_transition_host_lane3"      <<  ','
            << "pam4_level_transition_host_lane4"      <<  ','
            << "pam4_level_transition_host_lane5"      <<  ','
            << "pam4_level_transition_host_lane6"      <<  ','
            << "pam4_level_transition_host_lane7";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PAM4_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(pam4_level_transition_media_lane0)
        << __FMT__(pam4_level_transition_media_lane1)
        << __FMT__(pam4_level_transition_media_lane2)
        << __FMT__(pam4_level_transition_media_lane3)
        << __FMT__(pam4_level_transition_media_lane4)
        << __FMT__(pam4_level_transition_media_lane5)
        << __FMT__(pam4_level_transition_media_lane6)
        << __FMT__(pam4_level_transition_media_lane7)

        << __FMT__(pam4_level_transition_host_lane0)
        << __FMT__(pam4_level_transition_host_lane1)
        << __FMT__(pam4_level_transition_host_lane2)
        << __FMT__(pam4_level_transition_host_lane3)
        << __FMT__(pam4_level_transition_host_lane4)
        << __FMT__(pam4_level_transition_host_lane5)
        << __FMT__(pam4_level_transition_host_lane6)
        << __FMT_LAST__(pam4_level_transition_host_lane7);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PAM4_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                    export_data_phy_node_t *p_export_data_phy_node,
                                                    struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_pam4_level_transition_samples =
                                    &areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Samples;

    IBDIAGNET_RETURN_VOID;
}

PEMI_PRE_FEC_BER_Samples_Register::PEMI_PRE_FEC_BER_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x04,
                   (const unpack_data_func_t)pemi_Pre_FEC_BER_Samples_unpack,
                   ACC_REG_PEMI_BER_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_BER_S_NAME,
                   ACC_REG_PEMI_BER_S_FIELDS_NUM,
                   ACC_REG_PEMI_BER_S_HEADER)
{
}

void PEMI_PRE_FEC_BER_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                     << '#'
            << handler_header           << ','
            << "pre_fec_ber_min_media"  << ','
            << "pre_fec_ber_min_host"   << ','
            << "pre_fec_ber_max_media"  << ','
            << "pre_fec_ber_max_host"   << ','
            << "pre_fec_ber_avg_media"  << ','
            << "pre_fec_ber_avg_host"   << ','
            << "pre_fec_ber_val_media"  << ','
            << "pre_fec_ber_val_host";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PEMI_PRE_FEC_BER_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(pre_fec_ber_min_media)
        << __FMT__(pre_fec_ber_min_host)
        << __FMT__(pre_fec_ber_max_media)
        << __FMT__(pre_fec_ber_max_host)
        << __FMT__(pre_fec_ber_avg_media)
        << __FMT__(pre_fec_ber_avg_host)
        << __FMT__(pre_fec_ber_val_media)
        << __FMT_LAST__(pre_fec_ber_val_host);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PRE_FEC_BER_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                           export_data_phy_node_t *p_export_data_phy_node,
                                                           struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_pre_fec_ber_samples =
                                    &areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Samples;

    IBDIAGNET_RETURN_VOID;
}


PEMI_FERC_Samples_Register::PEMI_FERC_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x05,
                   (const unpack_data_func_t)pemi_FERC_Samples_unpack,
                   ACC_REG_PEMI_FERC_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_FERC_S_NAME,
                   ACC_REG_PEMI_FERC_S_FIELDS_NUM,
                   ACC_REG_PEMI_FERC_S_HEADER)
{
}

void PEMI_FERC_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                   << '#'
            << handler_header         << ','
            << "ferc_min_media"       << ','
            << "ferc_min_host"        << ','
            << "ferc_max_media"       << ','
            << "ferc_max_host"        << ','
            << "ferc_avg_media"       << ','
            << "ferc_avg_host"        << ','
            << "ferc_val_media"       << ','
            << "ferc_val_host";

#endif

    IBDIAGNET_RETURN_VOID;
}

void PEMI_FERC_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_FERC_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_FERC_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(ferc_min_media)
        << __FMT__(ferc_min_host)
        << __FMT__(ferc_max_media)
        << __FMT__(ferc_max_host)
        << __FMT__(ferc_avg_media)
        << __FMT__(ferc_avg_host)
        << __FMT__(ferc_val_media)
        << __FMT_LAST__(ferc_val_host);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_FERC_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                    export_data_phy_node_t *p_export_data_phy_node,
                                                    struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_ferc_samples = &areg.regs.pemi.page_data.pemi_FERC_Samples;

    IBDIAGNET_RETURN_VOID;
}

PEMI_Module_Properties_Register::PEMI_Module_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x10,
                   (const unpack_data_func_t)pemi_Module_Status_Properties_unpack,
                   ACC_REG_PEMI_MODULE_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_MODULE_P_NAME,
                   ACC_REG_PEMI_MODULE_P_FIELDS_NUM,
                   ACC_REG_PEMI_MODULE_P_HEADER)
{
}

void PEMI_Module_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                   << '#'
            << handler_header         << ','
            << "module_status_cap"     << ','
            << "temp_high_alarm"       << ','
            << "temp_low_alarm"        << ','
            << "temp_high_warning"     << ','
            << "temp_low_warning"      << ','
            << "voltage_high_alarm"    << ','
            << "voltage_low_alarm"     << ','
            << "voltage_high_warning"  << ','
            << "voltage_low_warning"   << ','
            << "rx_power_high_alarm"   << ','
            << "rx_power_low_alarm"    << ','
            << "rx_power_high_warning" << ','
            << "rx_power_low_warning"  << ','
            << "tx_power_high_alarm"   << ','
            << "tx_power_low_alarm"    << ','
            << "tx_power_high_warning" << ','
            << "tx_power_low_warning"  << ','
            << "tx_bias_high_alarm"    << ','
            << "tx_bias_low_alarm"     << ','
            << "tx_bias_high_warning"  << ','
            << "tx_bias_low_warning";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PEMI_Module_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Module_Status_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Module_Status_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT__(module_status_cap)
        << __FMT__(temp_high_alarm)
        << __FMT__(temp_low_alarm)
        << __FMT__(temp_high_warning)
        << __FMT__(temp_low_warning)
        << __FMT__(voltage_high_alarm)
        << __FMT__(voltage_low_alarm)
        << __FMT__(voltage_high_warning)
        << __FMT__(voltage_low_warning)
        << __FMT__(rx_power_high_alarm)
        << __FMT__(rx_power_low_alarm)
        << __FMT__(rx_power_high_warning)
        << __FMT__(rx_power_low_warning)
        << __FMT__(tx_power_high_alarm)
        << __FMT__(tx_power_low_alarm)
        << __FMT__(tx_power_high_warning)
        << __FMT__(tx_power_low_warning)
        << __FMT__(tx_bias_high_alarm)
        << __FMT__(tx_bias_low_alarm)
        << __FMT__(tx_bias_high_warning)
        << __FMT_LAST__(tx_bias_low_warning);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_Module_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                         export_data_phy_node_t *p_export_data_phy_node,
                                                         struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_module_status_properties =
                                        &areg.regs.pemi.page_data.pemi_Module_Status_Properties;

    IBDIAGNET_RETURN_VOID;
}

PEMI_SNR_Properties_Register::PEMI_SNR_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x11,
                   (const unpack_data_func_t)pemi_SNR_Properties_unpack,
                   ACC_REG_PEMI_SNR_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_SNR_P_NAME,
                   ACC_REG_PEMI_SNR_P_FIELDS_NUM,
                   ACC_REG_PEMI_SNR_P_HEADER)
{
}

void PEMI_SNR_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ','
            << "snr_cap"                 << ','
            << "snr_media_low_alarm"     << ','
            << "snr_media_low_warning"   << ','
            << "snr_host_low_alarm"      << ','
            << "snr_host_low_warning";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PEMI_SNR_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_SNR_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_SNR_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT__(snr_cap)
        << __FMT__(snr_media_low_alarm)
        << __FMT__(snr_media_low_warning)
        << __FMT__(snr_host_low_alarm)
        << __FMT_LAST__(snr_host_low_warning);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_SNR_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                      export_data_phy_node_t *p_export_data_phy_node,
                                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_snr_properties = &areg.regs.pemi.page_data.pemi_SNR_Properties;

    IBDIAGNET_RETURN_VOID;
}


PEMI_Laser_Properties_Register::PEMI_Laser_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x12,
                   (const unpack_data_func_t)pemi_Laser_Monitors_Properties_unpack,
                   ACC_REG_PEMI_LASER_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_LASER_P_NAME,
                   ACC_REG_PEMI_LASER_P_FIELDS_NUM,
                   ACC_REG_PEMI_LASER_P_HEADER)
{
}

void PEMI_Laser_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                                 << '#'
            << handler_header                       << ','
            << "laser_monitor_cap"                  << ','
            << "laser_age_high_alarm"               << ','
            << "laser_age_high_warning"             << ','
            << "tec_current_high_alarm"             << ','
            << "tec_current_low_alarm"              << ','
            << "tec_current_high_warning"           << ','
            << "tec_current_low_warning"            << ','
            << "laser_frequency_error_high_alarm"   << ','
            << "laser_frequency_error_high_warning" << ','
            << "laser_temp_high_alarm"              << ','
            << "laser_temp_low_alarm"               << ','
            << "laser_temp_high_warning"            << ','
            << "laser_temp_low_warning";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PEMI_Laser_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Laser_Monitors_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Laser_Monitors_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT__(laser_monitor_cap)
        << __FMT__(laser_age_high_alarm)
        << __FMT__(laser_age_high_warning)
        << __FMT__(tec_current_high_alarm)
        << __FMT__(tec_current_low_alarm)
        << __FMT__(tec_current_high_warning)
        << __FMT__(tec_current_low_warning)
        << __FMT__(laser_frequency_error_high_alarm)
        << __FMT__(laser_frequency_error_high_warning)
        << __FMT__(laser_temp_high_alarm)
        << __FMT__(laser_temp_low_alarm)
        << __FMT__(laser_temp_high_warning)
        << __FMT_LAST__(laser_temp_low_warning);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_Laser_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                        export_data_phy_node_t *p_export_data_phy_node,
                                                        struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_laser_monitors_properties =
                        &areg.regs.pemi.page_data.pemi_Laser_Monitors_Properties;

    IBDIAGNET_RETURN_VOID;
}

PEMI_PAM4_Properties_Register::PEMI_PAM4_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x13,
                   (const unpack_data_func_t)pemi_PAM4_Level_Transition_Properties_unpack,
                   ACC_REG_PEMI_PAM4_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_PAM4_P_NAME,
                   ACC_REG_PEMI_PAM4_P_FIELDS_NUM,
                   ACC_REG_PEMI_PAM4_P_HEADER)
{
}

void PEMI_PAM4_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
     IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                       << '#'
            <<  handler_header                            << ','
            << "pam4_level_transition_cap"                << ','
            << "pam4_level_transition_media_low_alarm"    << ','
            << "pam4_level_transition_media_low_warning"  << ','
            << "pam4_level_transition_host_low_alarm"     << ','
            << "pam4_level_transition_host_low_warning";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PAM4_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT__(pam4_level_transition_cap)
        << __FMT__(pam4_level_transition_media_low_alarm)
        << __FMT__(pam4_level_transition_media_low_warning)
        << __FMT__(pam4_level_transition_host_low_alarm)
        << __FMT_LAST__(pam4_level_transition_host_low_warning);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PAM4_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                       export_data_phy_node_t *p_export_data_phy_node,
                                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_pam4_level_transition_properties =
                            &areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Properties;

    IBDIAGNET_RETURN_VOID;
}

PEMI_PRE_FEC_BER_Properties_Register::PEMI_PRE_FEC_BER_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x14,
                   (const unpack_data_func_t)pemi_Pre_FEC_BER_Properties_unpack,
                   ACC_REG_PEMI_BER_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_BER_P_NAME,
                   ACC_REG_PEMI_BER_P_FIELDS_NUM,
                   ACC_REG_PEMI_BER_P_HEADER)
{
}

void PEMI_PRE_FEC_BER_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                   << '#'
            << handler_header                         << ','
            << "pre_fec_ber_cap"                      << ','
            << "pre_fec_ber_min_high_media_alarm"     << ','
            << "pre_fec_ber_min_high_media_warning"   << ','
            << "pre_fec_ber_min_high_host_alarm"      << ','
            << "pre_fec_ber_min_high_host_warning"    << ','
            << "pre_fec_ber_max_high_media_alarm"     << ','
            << "pre_fec_ber_max_high_media_warning"   << ','
            << "pre_fec_ber_max_high_host_alarm"      << ','
            << "pre_fec_ber_max_high_host_warning"    << ','
            << "pre_fec_ber_avg_high_media_alarm"     << ','
            << "pre_fec_ber_avg_high_media_warning"   << ','
            << "pre_fec_ber_avg_high_host_alarm"      << ','
            << "pre_fec_ber_avg_high_host_warning";
#endif
    IBDIAGNET_RETURN_VOID;
}
void PEMI_PRE_FEC_BER_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT__(pre_fec_ber_cap)
        << __FMT__(pre_fec_ber_min_high_media_alarm)
        << __FMT__(pre_fec_ber_min_high_media_warning)
        << __FMT__(pre_fec_ber_min_high_host_alarm)
        << __FMT__(pre_fec_ber_min_high_host_warning)
        << __FMT__(pre_fec_ber_max_high_media_alarm)
        << __FMT__(pre_fec_ber_max_high_media_warning)
        << __FMT__(pre_fec_ber_max_high_host_alarm)
        << __FMT__(pre_fec_ber_max_high_host_warning)
        << __FMT__(pre_fec_ber_avg_high_media_alarm)
        << __FMT__(pre_fec_ber_avg_high_media_warning)
        << __FMT__(pre_fec_ber_avg_high_host_alarm)
        << __FMT_LAST__(pre_fec_ber_avg_high_host_warning);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PRE_FEC_BER_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                              export_data_phy_node_t *p_export_data_phy_node,
                                                              struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_pre_fec_ber_properties =
                                        &areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Properties;

    IBDIAGNET_RETURN_VOID;
}

PEMI_FERC_Properties_Register::PEMI_FERC_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x15,
                   (const unpack_data_func_t)pemi_FERC_Properties_unpack,
                   ACC_REG_PEMI_FERC_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_FERC_P_NAME,
                   ACC_REG_PEMI_FERC_P_FIELDS_NUM,
                   ACC_REG_PEMI_FERC_P_HEADER)
{
}

void PEMI_FERC_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl            << '#'
            << handler_header  << ','
            << "ferc_cap";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PEMI_FERC_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_FERC_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_FERC_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT_LAST__(ferc_cap);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_FERC_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                       export_data_phy_node_t *p_export_data_phy_node,
                                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_ferc_properties =
                            &areg.regs.pemi.page_data.pemi_FERC_Properties;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***     MPCNT Register functions      ***
//*****************************************

int MPCNTRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer) {
    IBDIAGNET_ENTER;

    mpcnt_reg *reg = (mpcnt_reg *) data_to_unpack;
    mpcnt_reg_unpack(reg, unpacked_buffer);

    mpcnt_RAW_page_data raw = reg->page_data.raw;
    unpack_data_func(&reg->page_data, raw.data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int MPCNTRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    const AccRegKeyDPN* key = dynamic_cast<const AccRegKeyDPN*>(p_key);

    if (!key)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    struct mpcnt_reg mpcnt;
    CLEAR_STRUCT(mpcnt);

    mpcnt.depth         = key->depth;
    mpcnt.pcie_index    = key->pci_idx;
    mpcnt.node          = key->pci_node;
    mpcnt.grp           = this->m_grp;
    mpcnt.clr           = this->m_clr;

    mpcnt_reg_pack(&mpcnt, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int MPCNTRegister::BuildDB(AccRegHandler* p_handler,
                          list_p_fabric_general_err &phy_errors,
                          ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    p_handler->SetHeader(PCI_DPN_HEADER);

    // go over all the mpein MADs and send mpir accordingly
    for (map_akey_areg::iterator mpein_iter = m_mpein_map.begin();
         mpein_iter != m_mpein_map.end(); mpein_iter++) {

        AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)mpein_iter->first;

        // get node
        IBNode *p_curr_node =
                m_phy_diag->GetFabric()->getNodeByGuid(p_dpnkey->node_guid);
        if (!p_curr_node) {
            m_phy_diag->SetLastError("DB error - found null node in "
                                     "NodeByName map for key = " U64H_FMT,
                                                p_dpnkey->node_guid);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        AccRegVia_t acc_reg_via = Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        // get port
        IBPort *p_curr_port = NULL;
        for (u_int32_t port_num = 1; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
            p_curr_port = p_curr_node->getPort((phys_port_t)port_num);

            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;
            break; // use the first valid port.
        }

        if (!p_curr_port) {
            m_phy_diag->SetLastError("DB error - failed to find valid port for node %s",
                                     p_curr_node->getName().c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        AccessRegister acc_reg;
        CLEAR_STRUCT(acc_reg);

        AccRegKeyDPN *p_dpnkey_new = new AccRegKeyDPN(p_dpnkey->p_port,
                                                      p_dpnkey->depth,
                                                      p_dpnkey->pci_idx,
                                                      p_dpnkey->pci_node);

        rc = p_handler->SendAccReg(acc_reg_via,
                                   p_curr_node,
                                   0, // port 0
                                   p_curr_port->base_lid,
                                   acc_reg,
                                   p_dpnkey_new,
                                   p_progress_bar);

        if (rc == IBDIAG_ERR_CODE_DB_ERR)
            IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(rc);
}


MPCNT_PCIe_Counters_Register::MPCNT_PCIe_Counters_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map)
    : MPCNTRegister(phy_diag,
                    mpein_map,
                    0x00,
                    (const unpack_data_func_t)pcie_perf_counters_unpack,
                    ACC_REG_MPCNT_PCI_CNT_INTERNAL_SECTION_NAME,
                    ACC_REG_MPCNT_PCI_CNT_NAME,
                    ACC_REG_MPCNT_PCI_CNT_FIELDS_NUM,
                    ACC_REG_MPCNT_PCI_CNT_HEADER)
{
}


void MPCNT_PCIe_Counters_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        +areg.regs.mpcnt.page_data.pcie_perf_counters.x << ','
    #define __FMT_LAST__( x )   +areg.regs.mpcnt.page_data.pcie_perf_counters.x << std::endl

    sstream
        << std::dec
        << __FMT__(life_time_counter_high)
        << __FMT__(life_time_counter_low)
        << __FMT__(rx_errors)
        << __FMT__(tx_errors)
        << __FMT__(l0_to_recovery_eieos)
        << __FMT__(l0_to_recovery_ts)
        << __FMT__(l0_to_recovery_framing)
        << __FMT__(l0_to_recovery_retrain)
        << __FMT__(crc_error_dllp)
        << __FMT__(crc_error_tlp)
        << __FMT__(tx_overflow_buffer_pkt)
        << __FMT__(outbound_stalled_reads)
        << __FMT__(outbound_stalled_writes)
        << __FMT__(outbound_stalled_reads_events)
        << __FMT__(outbound_stalled_writes_events)
        << __FMT__(tx_overflow_buffer_marked_pkt)
        << __FMT__(effective_ber_magnitude)
        << __FMT__(effective_ber_coef)
        << __FMT__(time_since_last_clear_high)
        << __FMT_LAST__(time_since_last_clear_low);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

MPCNT_PCIe_Lane_Counters_Register::MPCNT_PCIe_Lane_Counters_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map)
    : MPCNTRegister(phy_diag,
                    mpein_map,
                    0x01,
                    (const unpack_data_func_t)pcie_lanes_counters_unpack,
                    ACC_REG_MPCNT_PCI_LCNT_INTERNAL_SECTION_NAME,
                    ACC_REG_MPCNT_PCI_LCNT_NAME,
                    ACC_REG_MPCNT_PCI_LCNT_FIELDS_NUM,
                    ACC_REG_MPCNT_PCI_LCNT_HEADER)
{
}


void MPCNT_PCIe_Lane_Counters_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.mpcnt.page_data.pcie_lanes_counters.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.mpcnt.page_data.pcie_lanes_counters.x << std::endl

    sstream
        << std::hex
        << __FMT__(life_time_counter_high)
        << __FMT__(life_time_counter_low)
        << __FMT__(error_counter_lane0)
        << __FMT__(error_counter_lane1)
        << __FMT__(error_counter_lane2)
        << __FMT__(error_counter_lane3)
        << __FMT__(error_counter_lane4)
        << __FMT__(error_counter_lane5)
        << __FMT__(error_counter_lane6)
        << __FMT__(error_counter_lane7)
        << __FMT__(error_counter_lane8)
        << __FMT__(error_counter_lane9)
        << __FMT__(error_counter_lane10)
        << __FMT__(error_counter_lane11)
        << __FMT__(error_counter_lane12)
        << __FMT__(error_counter_lane13)
        << __FMT__(error_counter_lane14)
        << __FMT_LAST__(error_counter_lane15);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

MPCNT_PCIe_Timers_Register::MPCNT_PCIe_Timers_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map)
    : MPCNTRegister(phy_diag,
                    mpein_map,
                    0x02,
                    (const unpack_data_func_t)pcie_lanes_counters_unpack,
                    ACC_REG_MPCNT_PCI_TIMERS_INTERNAL_SECTION_NAME,
                    ACC_REG_MPCNT_PCI_TIMERS_NAME,
                    ACC_REG_MPCNT_PCI_TIMERS_FIELDS_NUM,
                    ACC_REG_MPCNT_PCI_TIMERS_HEADER)
{
}


void MPCNT_PCIe_Timers_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.mpcnt.page_data.pcie_timers_states.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.mpcnt.page_data.pcie_timers_states.x << std::endl

    sstream
        << std::hex
        << __FMT__(life_time_counter_high)
        << __FMT__(life_time_counter_low)
        << __FMT__(time_to_boot_image_start)
        << __FMT__(time_to_link_image)
        << __FMT__(calibration_time)
        << __FMT__(time_to_first_perst)
        << __FMT__(time_to_detect_state)
        << __FMT__(time_to_l0)
        << __FMT__(time_to_crs_en)
        << __FMT__(time_to_plastic_image_start)
        << __FMT__(time_to_iron_image_start)
        << __FMT__(perst_handler)
        << __FMT__(times_in_l1)
        << __FMT__(times_in_l23)
        << __FMT__(dl_down)
        << __FMT__(config_cycle1usec)
        << __FMT__(config_cycle2to7usec)
        << __FMT__(config_cycle8to15usec)
        << __FMT__(config_cycle16to63usec)
        << __FMT__(config_cycle64usec)
        << __FMT__(correctable_err_msg_sent)
        << __FMT__(non_fatal_err_msg_sent)
        << __FMT_LAST__(fatal_err_msg_sent);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

MPCNT_PCIe_Phys_Layer_Register::MPCNT_PCIe_Phys_Layer_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map)
    : MPCNTRegister(phy_diag,
                    mpein_map,
                    0x03,
                    (const unpack_data_func_t)pcie_lanes_counters_unpack,
                    ACC_REG_MPCNT_PCI_PHYS_INTERNAL_SECTION_NAME,
                    ACC_REG_MPCNT_PCI_PHYS_NAME,
                    ACC_REG_MPCNT_PCI_PHYS_FIELDS_NUM,
                    ACC_REG_MPCNT_PCI_PHYS_HEADER)
{
}

void MPCNT_PCIe_Phys_Layer_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "eqlz_timeout_ph3" << ','
            << "eqlz_timeout_ph2" << ','
            << "eqlz_timeout_ph1" << ','
            << "eqlz_timeout_ph0";
#endif
    IBDIAGNET_RETURN_VOID;
}


void MPCNT_PCIe_Phys_Layer_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.mpcnt.page_data.pcie_physical_layer.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.mpcnt.page_data.pcie_physical_layer.x << std::endl

    sstream
        << std::hex
        << __FMT__(eqlz_timeout_ph3)
        << __FMT__(eqlz_timeout_ph2)
        << __FMT__(eqlz_timeout_ph1)
        << __FMT_LAST__(eqlz_timeout_ph0);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}





//*****************************************
//***     PPBMP Register functions      ***
//*****************************************
PPBMPRegister::PPBMPRegister(PhyDiag* phy_diag) : Register(phy_diag,
                                                           ACCESS_REGISTER_ID_PPBMP,
                                                           (const unpack_data_func_t)ppbmp_reg_unpack,
                                                           ACC_REG_PPBMP_INTERNAL_SECTION_NAME,
                                                           ACC_REG_PPBMP_NAME,
                                                           ACC_REG_PPBMP_FIELDS_NUM,
                                                           NOT_SUPPORT_PPBMP,
                                                           ACC_REG_PPBMP_FIELDS_HEADER,
                                                           SUPPORT_SW_CA,
                                                           true,
                                                           false,
                                                           VIA_SMP,
                                                           VIA_GMP),
                                                  m_pnat(ACC_REG_PNAT_IB_PORT)
{



}

int PPBMPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct ppbmp_reg ppbmp;
    CLEAR_STRUCT(ppbmp);

    ppbmp.local_port = ((AccRegKeyPortIndex *)p_key)->port_num;
    ppbmp.pnat = this->m_pnat;
    ppbmp.monitor_group = ((AccRegKeyPortIndex *)p_key)->index;

    ppbmp_reg_pack(&ppbmp, data);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPBMPRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct ppbmp_reg &ppbmp = areg.regs.ppbmp;

    double normal_th = ppbmp.monitor_parameters.normal_th_mantissa *
            (1 / pow(10, ppbmp.monitor_parameters.normal_th_exp));

    double warning_th = ppbmp.monitor_parameters.warning_th_mantissa *
            (1 / pow(10, ppbmp.monitor_parameters.warning_th_exp));

    double alarm_th = ppbmp.monitor_parameters.alarm_th_mantissa *
            (1 / pow(10, ppbmp.monitor_parameters.alarm_th_exp));


    ios_base::fmtflags f(sstream.flags());

    sstream << scientific
            << normal_th             << ','
            << warning_th            << ','
            << alarm_th              << ',';

    sstream.flags(f);

    sstream << +ppbmp.time_window    << ','
            << +ppbmp.sampling_rate  << endl;

    IBDIAGNET_RETURN_VOID;
}

void PPBMPRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl             << '#'
            << handler_header   << ","
            << "scientific"     << ','
            << "normal_th"      << ','
            << "warning_th"     << ','
            << "alarm_th"       << ','
            << "time_window"    << ','
            << "sampling_rate";
#endif
    IBDIAGNET_RETURN_VOID;
}




//*****************************************
//***     MRCS Register functions      ***
//*****************************************
MRCSRegister::MRCSRegister(PhyDiag* phy_diag) : Register(phy_diag,
                                                         ACCESS_REGISTER_ID_MRCS,
                                                         (const unpack_data_func_t)mrcs_reg_unpack,
                                                         ACC_REG_MRCS_INTERNAL_SECTION_NAME,
                                                         ACC_REG_MRCS_NAME,
                                                         ACC_REG_MRCS_FIELDS_NUM,
                                                         NOT_SUPPORT_MRCS,
                                                         ACC_REG_MRCS_FIELDS_HEADER,
                                                         SUPPORT_SW_CA,
                                                         true,
                                                         false,
                                                         VIA_SMP,
                                                         VIA_GMP)
{
}

int MRCSRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->serdes_version != SerDes_7nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void MRCSRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct mrcs_reg &mrcs = areg.regs.mrcs;

    sstream << +mrcs.ref_clock_stat[0].measured_freq    << ','
            << +mrcs.ref_clock_stat[0].min_freq         << ','
            << +mrcs.ref_clock_stat[0].max_delta_freq   << ','
            << +mrcs.ref_clock_stat[0].max_freq         << ','
            << +mrcs.ref_clock_stat[1].measured_freq    << ','
            << +mrcs.ref_clock_stat[1].min_freq         << ','
            << +mrcs.ref_clock_stat[1].max_delta_freq   << ','
            << +mrcs.ref_clock_stat[1].max_freq         << endl;

    IBDIAGNET_RETURN_VOID;
}

void MRCSRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                       << '#'
            << handler_header             << ","
            << "mrcs[0].measured_freq"    << ','
            << "mrcs[0].min_freq"         << ','
            << "mrcs[0].max_delta_freq"   << ','
            << "mrcs[0].max_freq"         << ','
            << "mrcs[1].measured_freq"    << ','
            << "mrcs[1].min_freq"         << ','
            << "mrcs[1].max_delta_freq"   << ','
            << "mrcs[1].max_freq";
#endif
    IBDIAGNET_RETURN_VOID;
}

//**************************************************
//***   Access Register Handler Implementation   ***
//**************************************************
AccRegHandler::AccRegHandler(Register *p_register, string hh)
    : phy_errors(&p_register->GetPhyDiag()->GetPhyErrors()),
          can_send_mads_by_lid(false),
          clbck_error_state(IBDIAG_SUCCESS_CODE),
          p_reg(p_register),
          handler_header(hh),
          data_map(keycomp)
{
}

AccRegHandler::~AccRegHandler()
{
    IBDIAGNET_ENTER;

    if (p_reg)
        delete p_reg;

    for (map_akey_areg::iterator it = this->data_map.begin(); it != this->data_map.end(); ++it) {
         delete(it->first);
    }
    this->data_map.clear();

    IBDIAGNET_RETURN_VOID;
}

int AccRegHandler::PCIAddressCollectGetClbck(const clbck_data_t &clbck_data,
                                  int rec_status,
                                  void *p_attribute_data)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    AccRegVia_t acc_reg_via = (AccRegVia_t)(uint64_t)clbck_data.m_data4;
    // invoke handler callback in order to handle return status and insert reply to the map
    if (acc_reg_via == VIA_SMP)
        rc = AccRegHandler::SMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);

    if (acc_reg_via == VIA_GMP)
        rc = AccRegHandler::GMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);

    if (rc)
        IBDIAGNET_RETURN(rc)

    //in the case of success we are sure to get already unpacked and saved data
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    AccRegKey *p_key = (AccRegKey *)clbck_data.m_data2;

    std::map<AccRegKey *, struct acc_reg_data>::iterator it = this->data_map.find(p_key);
    if (it == this->data_map.end())
       IBDIAGNET_RETURN(1);

    const acc_reg_data& reg_data = it->second;
    p_node->addPCIAddress(PCI_Index(reg_data.regs.mpir.pcie_index,
                                    reg_data.regs.mpir.depth,
                                    reg_data.regs.mpir.node),
                          PCI_Address(reg_data.regs.mpir.bus,
                                      reg_data.regs.mpir.local_port,
                                      reg_data.regs.mpir.device),
                          reg_data.regs.mpir.sdm);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int AccRegHandler::SMPAccessRegisterHandlerGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAGNET_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (clbck_error_state)
    {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }
    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    if (rec_status &&
        IsSupportAccRegFailed(p_node,
                              NOT_SUPPORT_SMP_ACCESS_REGISTER,
                              this->p_reg->GetNotSupportedBit()))
    {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    if (rec_status) {
        switch(rec_status) {
            case IBIS_MAD_STATUS_UNSUP_METHOD_ATTR: {
                SET_NSB(p_node, this->p_reg->GetNotSupportedBit());

                stringstream ss;
                ss << "The firmware of this device does not support SMP register ID: 0x"
                   << std::hex << std::setw(4) << std::setfill('0') << this->p_reg->GetRegisterID()
                   << " [err=0x" << std::hex << std::setw(4) << std::setfill('0') << rec_status << "]";

                this->phy_errors->push_back(new FabricErrPhyNodeNotSupportCap(p_node, ss.str()));

                break;
            }

            case IBIS_MAD_STATUS_TIMEOUT: {
                SET_NSB(p_node, NOT_SUPPORT_SMP_ACCESS_REGISTER);
                this->phy_errors->push_back(new FabricErrPhyNodeNotRespond(p_node, "SMPAccessRegister [timeout]"));

                break;
            }

            default: {
                stringstream ss;
                ss << "SMPAccessRegister [err=0x"
                   << std::hex << std::setw(4) << std::setfill('0') << rec_status << "]";

                this->phy_errors->push_back(new FabricErrPhyNodeNotRespond(p_node, ss.str()));

                break;
            }
        }

        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    struct SMP_AccessRegister *p_access_reg =
        (struct SMP_AccessRegister *)p_attribute_data;

    //already reported that this node doesn't support
    if (p_access_reg->status &&
        IsSupportAccRegFailed(p_node,
                              NOT_SUPPORT_SMP_ACCESS_REGISTER,
                              this->p_reg->GetNotSupportedBit()))
    {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    //check if the access register status, is it supported, etc.
    if (p_access_reg->status) {
        SET_NSB(p_node, this->p_reg->GetNotSupportedBit());

        this->phy_errors->push_back(new FabricNodeErrPhyRetrieveGeneral(p_node ,p_access_reg->status));

        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    AccRegKey *p_key = (AccRegKey *)clbck_data.m_data2;

    struct acc_reg_data register_unpack_data;
    CLEAR_STRUCT(register_unpack_data);

    if (this->p_reg->UnpackData(p_key, &register_unpack_data.data, p_access_reg->reg.data)) {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    if (!p_reg->IsValidPakcet(register_unpack_data)) {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    // insert the data into a map
    std::pair<std::map<AccRegKey *, struct acc_reg_data>::iterator,bool> ret;
    ret = this->data_map.insert(pair <AccRegKey *,
                                struct acc_reg_data>(p_key,
                                                     register_unpack_data));

    if ((!ret.second) || clbck_error_state) {
        this->p_reg->GetPhyDiag()->SetLastError("Failed to add %s data for node=%s," \
                " err=%s",
                (CHECK_NAME_RETRIEVE(this->
                                     GetPReg()->
                                     GetSectionName())).c_str(),
                p_node->getName().c_str(), this->p_reg->GetPhyDiag()->GetLastError());

        delete p_key;
        IBDIAGNET_RETURN(1);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


int AccRegHandler::GMPAccessRegisterHandlerGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAGNET_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (clbck_error_state) {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    if (rec_status && IsSupportAccRegFailed(p_node,
                                            NOT_SUPPORT_GMP_ACCESS_REGISTER,
                                            this->p_reg->GetNotSupportedBit()))
    {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    if (rec_status) {
        switch(rec_status) {
            case IBIS_MAD_STATUS_UNSUP_METHOD_ATTR:
            case MAD_STATUS_UNSUP_REGISTER_GMP: {
                SET_NSB(p_node, this->p_reg->GetNotSupportedBit());

                stringstream ss;
                ss << "The firmware of this device does not support GMP register ID: 0x"
                   << std::hex << std::setw(4) << std::setfill('0') << this->p_reg->GetRegisterID()
                   << " [err=0x" << std::hex << std::setw(4) << std::setfill('0') << rec_status << "]";

                this->phy_errors->push_back(new FabricErrPhyNodeNotSupportCap(p_node, ss.str()));

                break;
            }

            case IBIS_MAD_STATUS_TIMEOUT: {
                SET_NSB(p_node,  NOT_SUPPORT_GMP_ACCESS_REGISTER);
                this->phy_errors->push_back(new FabricErrPhyNodeNotRespond(p_node, "GMPAccessRegister [timeout]"));

                break;
            }

            default: {
                stringstream ss;
                ss << "GMPAccessRegister [err=0x"
                   << std::hex << std::setw(4) << std::setfill('0') << rec_status << "]";

                this->phy_errors->push_back(new FabricErrPhyNodeNotRespond(p_node, ss.str()));

                break;
            }
        }

        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    struct GMP_AccessRegister *p_access_reg = (struct GMP_AccessRegister *)p_attribute_data;
    AccRegKey *p_key = (AccRegKey *)clbck_data.m_data2;

    struct acc_reg_data register_unpack_data;
    CLEAR_STRUCT(register_unpack_data);

    if (this->p_reg->UnpackData(p_key, &register_unpack_data.data, p_access_reg->reg.data)) {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    // insert the data into the map
    pair<map<AccRegKey*, struct acc_reg_data>::iterator,bool> ret;
    ret = data_map.insert(pair <AccRegKey*, struct acc_reg_data>(p_key, register_unpack_data));

    if ((!ret.second) || clbck_error_state) {
        this->p_reg->GetPhyDiag()->SetLastError("Failed to add %s data for node=%s, err=%s",
                    (CHECK_NAME_RETRIEVE(GetPReg()->GetSectionName())).c_str(),
                    p_node->getName().c_str(),
                    this->p_reg->GetPhyDiag()->GetLastError());

        delete p_key;
        IBDIAGNET_RETURN(1);
    }

    // call virtual func of register to handle data
    // beside insert it to the map in the handler
    this->p_reg->HandleData(p_node, p_key, register_unpack_data);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void AccRegHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                               export_data_phy_node_t *p_export_data_phy_node)
{
}

void AccRegHandler::ExportDataImplNode(export_data_phy_node_t *p_export_data_phy_node)
{
    if (!p_export_data_phy_node)
        return;

    for (map_akey_areg::iterator nI = this->data_map.begin(); nI != this->data_map.end(); ++nI) {
        AccRegKey *p_key = nI->first;
        if (!p_key) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            IBDIAGNET_RETURN_VOID;
        }

        if (p_key->node_guid != p_export_data_phy_node->node_guid)
            continue;

        this->p_reg->ExportRegisterData(NULL, p_export_data_phy_node, nI->second, nI->first);
    }
}

template<typename KeyType>
void AccRegHandler::ExportDataImplPort(export_data_phy_port_t *p_export_data_phy_port)
{
    if (!p_export_data_phy_port)
        return;

    for (map_akey_areg::iterator nI = this->data_map.begin(); nI != this->data_map.end(); ++nI) {
        KeyType *p_key = (KeyType *)nI->first;
        if (!p_key) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            IBDIAGNET_RETURN_VOID;
        }

        if (p_key->port_guid != p_export_data_phy_port->port_guid ||
                p_key->port_num != p_export_data_phy_port->port_num)
            continue;

        this->p_reg->ExportRegisterData(p_export_data_phy_port, NULL, nI->second, nI->first);
    }
}

int AccRegGroupHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    map_pnode_to_max_group_num_t pnode_to_max_group_num;

    ProgressBarNodes progress_bar;

    for (map_str_pnode::iterator nI = this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.begin();
            nI != this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByName map for key = %s",
                                                    (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        //send group 0, to discover how many groups there are.
        AccRegKeyGroup *p_plkey = new AccRegKeyGroup(p_curr_node->guid_get(), 0);

        AccessRegister acc_reg;
        CLEAR_STRUCT(acc_reg);

        rc = this->SendAccReg(acc_reg_via,
                              p_curr_node,
                              0, // port 0
                              lid,
                              acc_reg,
                              p_plkey,
                              &progress_bar);

        if (rc == IBDIAG_ERR_CODE_DB_ERR || clbck_error_state)
            goto exit;
    }

    // need to get all the mads for group 0
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();

    // build the pnode_to_max_group_num map
    for (map_akey_areg::iterator nI = this->data_map.begin(); nI != this->data_map.end(); ++nI) {

        AccRegKey *p_aregkey = nI->first;
        if (!p_aregkey) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        map_guid_pnode::iterator ngI = this->p_reg->GetPhyDiag()->GetFabric()->
                NodeByGuid.find(((AccRegKeyGroup *)p_aregkey)->node_guid);

        if (ngI == this->p_reg->GetPhyDiag()->GetFabric()->NodeByGuid.end())
            goto exit;

        IBNode *p_curr_node = ngI->second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByGuid map for key = " U64H_FMT,
                                                    ((AccRegKeyGroup *)p_aregkey)->node_guid);
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        struct acc_reg_data areg = nI->second;
        struct ppll_reg *p_ppll = (struct ppll_reg *)&areg.data;
        pnode_to_max_group_num.insert(pair < IBNode *, u_int8_t >(p_curr_node, p_ppll->num_pll_groups));
    }

    // go over all the nodes and send for the rest of the group other then 0
    for (map_pnode_to_max_group_num_t::iterator nI = pnode_to_max_group_num.begin();
            nI != pnode_to_max_group_num.end(); ++nI) {

        IBNode *p_curr_node = nI->first;
        u_int8_t max_gr = nI->second;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        //starting from group 1, we already sent group 0
        for (u_int8_t gr = 1; gr < max_gr; gr++) {

            AccRegKeyGroup *p_plkey = new AccRegKeyGroup(p_curr_node->guid_get(), gr);

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            rc = this->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  0, // port 0
                                  lid,
                                  acc_reg,
                                  p_plkey,
                                  &progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR || clbck_error_state)
                goto exit;
        }
    }

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void AccRegGroupHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    this->ExportDataImplNode(p_export_data_phy_node);

    IBDIAGNET_RETURN_VOID;
}

int AccRegNodeSensorHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    for (map_akey_areg::iterator areg_iter = p_acc_reg_handler->data_map.begin();
            areg_iter != p_acc_reg_handler->data_map.end(); ++areg_iter) {

        AccRegKeyNode *p_nkey = (AccRegKeyNode*)areg_iter->first;
        IBNode *p_curr_node = p_reg->GetPhyDiag()->GetFabric()->NodeByGuid[p_nkey->node_guid];

        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in" \
                                                    " NodeByGuid map for key = " U64H_FMT,
                                                    p_nkey->node_guid);
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        list_uint8 sensors;
        rc = p_acc_reg_handler->GetPReg()->AvailableSensors(areg_iter->second, sensors);

        if (rc)
            goto exit;

        // go over all sensors of this node:
        for (list_uint8::iterator sensors_iter = sensors.begin();
                sensors_iter != sensors.end(); ++sensors_iter) {

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            AccRegKeyNodeSensor *p_ns_key =
                    new AccRegKeyNodeSensor(p_nkey->node_guid, *sensors_iter);

            rc = this->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  0, // port 0
                                  lid,
                                  acc_reg,
                                  p_ns_key,
                                  &progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR || clbck_error_state)
                goto exit;
        }
    }

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    IBDIAGNET_RETURN(rc);
}

void AccRegNodeSensorHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                         export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    this->ExportDataImplNode(p_export_data_phy_node);

    IBDIAGNET_RETURN_VOID;
}


int AccRegTypeGroupHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    for (map_str_pnode::iterator nI = p_reg->GetPhyDiag()->GetFabric()->NodeByName.begin();
         nI != p_reg->GetPhyDiag()->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByName map for key = %s",
                                                    nI->first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        for (int tp = 0; tp < max_type; tp++) {
            //send group 0, to discover how many groups there are.

            //Skip PCIe Group
            if(tp == 1)
                continue;

            AccRegKeyTypeGroup *p_gpkey = new AccRegKeyTypeGroup(
                    p_curr_node->guid_get(), tp, 0, 0, max_indices);

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            rc = this->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  0, // port 0
                                  lid,
                                  acc_reg,
                                  p_gpkey,
                                  &progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR || clbck_error_state)
                goto exit;
        }
    }

    // need to get all the mads for group 0
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();

    //go over all the nodes and send for the rest of the group other then 0
    for (map_akey_areg::iterator aregI = data_map.begin(); aregI != data_map.end(); ++aregI) {

        AccRegKeyTypeGroup* p_aregkey = (AccRegKeyTypeGroup*)aregI->first;
        if (!p_aregkey) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            goto exit;
        }

        // make sure that we are on group 0
        if (p_aregkey->group_num != 0)
            continue;

        map_guid_pnode::iterator ngI = this->p_reg->GetPhyDiag()->GetFabric()->
            NodeByGuid.find(p_aregkey->node_guid);
        if (ngI == this->p_reg->GetPhyDiag()->GetFabric()->NodeByGuid.end()) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - GUID " U64H_FMT
                                                    "not found in NodeByGuid map",
                                                    p_aregkey->node_guid);
            goto exit;
        }

        IBNode *p_curr_node = ngI->second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByGuid map for key = " U64H_FMT,
                                                    p_aregkey->node_guid);
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        u_int8_t opamp_gr_type = p_aregkey->group_type;

        struct acc_reg_data areg = aregI->second;
        u_int8_t max_opamp_gr;
        u_int16_t max_num_idx;

        this->p_reg->GetGroupData(areg, max_opamp_gr, max_num_idx);

        //start from group 1, we already sent group 0
        for (u_int8_t gr = 1; gr <= max_opamp_gr; gr++) {
            for (u_int16_t start_idx = 0; start_idx < max_num_idx; start_idx += max_indices) {

                u_int16_t idx_to_access =(((max_num_idx - start_idx) < max_indices)?
                        (max_num_idx % max_indices) : max_indices);

                AccRegKeyTypeGroup *p_gpkey = new AccRegKeyTypeGroup(
                        p_curr_node->guid_get(),
                        opamp_gr_type,
                        gr,
                        start_idx,
                        idx_to_access);

                AccessRegister acc_reg;
                CLEAR_STRUCT(acc_reg);

                rc = this->SendAccReg(acc_reg_via,
                                      p_curr_node,
                                      0, // port 0
                                      lid,
                                      acc_reg,
                                      p_gpkey,
                                      &progress_bar);

                if (rc == IBDIAG_ERR_CODE_DB_ERR || clbck_error_state)
                    goto exit;
            }
        }
    }

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}


int AccRegSpecificHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;
    rc = this->p_reg->BuildDB(this, phy_errors, &progress_bar);

    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void AccRegSpecificHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    for (map_akey_areg::iterator nI = this->data_map.begin(); nI != this->data_map.end(); ++nI) {
        AccRegKey *p_key = nI->first;
        if (!p_key) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            IBDIAGNET_RETURN_VOID;
        }

        if (this->p_reg->CheckRegisterKey(p_export_data_phy_port, p_export_data_phy_node, p_key))
            this->p_reg->ExportRegisterData(p_export_data_phy_port, p_export_data_phy_node, nI->second, p_key);
    }

    IBDIAGNET_RETURN_VOID;
}


int Register::PackDataSMP(AccRegKey *p_key, SMP_AccessRegister *acc_reg, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    acc_reg->register_id = this->GetRegisterID();
    acc_reg->len_reg = SMP_ACC_REG_TLV_LEN;

    IBDIAGNET_RETURN(PackData(p_key, acc_reg->reg.data, p_node));
}

int Register::PackDataGMP(AccRegKey *p_key, GMP_AccessRegister *acc_reg, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    IBDIAGNET_RETURN(PackData(p_key, acc_reg->reg.data, p_node));
}

bool Register::IsEnabledByPCAM(const IBNode* p_node) const {

    if (this->m_register_id >= 0x5000 || this->m_register_id < 0x507F)
        if (p_node) {
            PHYNodeData* data = dynamic_cast<PHYNodeData*>(p_node->p_phy_data);
            if (data && data->pcam)
                return data->pcam->port_access_reg_cap_mask[this->m_register_id - 0x5000];
        }

    return true;
}

int Register::HandleNodeNotSupportAccReg(PhyDiag *p_phy_diag,
                                         IBNode* p_node,
                                         uint64_t not_supported_bit)
{
    IBDIAGNET_ENTER;

    // check if warning was already set
    if (NSB_BITSET(p_node)[not_supported_bit])
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

    SET_NSB(p_node, not_supported_bit);
    stringstream err_str;
    err_str << "This device does not support "
            << (not_supported_bit == NOT_SUPPORT_SMP_ACCESS_REGISTER ? "SMP" : "GMP")
            << " access register MAD capability";

    p_phy_diag->GetPhyErrors().push_back(new FabricErrPhyNodeNotSupportCap(p_node, err_str.str()));

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

AccRegVia_t Register::Validation(IBNode *p_node, int &rc)
{
    IBDIAGNET_ENTER;

    bool smp_sup = m_phy_diag->GetCapabilityModule()->IsSupportedSMPCapability(
            p_node, EnSMPCapIsAccessRegisterSupported) &&
                    !(NSB_BITSET(p_node)[NOT_SUPPORT_SMP_ACCESS_REGISTER]);

    bool gmp_sup = m_phy_diag->GetCapabilityModule()->IsSupportedGMPCapability(
            p_node, EnGMPCapIsAccessRegisterSupported) &&
                    !(NSB_BITSET(p_node)[NOT_SUPPORT_GMP_ACCESS_REGISTER]);

    bool smp_able = (GetRegisterType() == VIA_SMP) && (p_node->serdes_version < SerDes_7nm);

    if (GetRegisterPriority() == VIA_GMP) {
        if (gmp_sup)
            IBDIAGNET_RETURN(VIA_GMP);
        if (smp_sup && smp_able)
            IBDIAGNET_RETURN(VIA_SMP);
    } else { // VIA_SMP
        if (smp_sup && smp_able)
            IBDIAGNET_RETURN(VIA_SMP);
        if (gmp_sup)
            IBDIAGNET_RETURN(VIA_GMP);
    }

    rc = HandleNodeNotSupportAccReg(m_phy_diag, p_node, NOT_SUPPORT_GMP_ACCESS_REGISTER);

    if (!smp_sup)
        rc = HandleNodeNotSupportAccReg(m_phy_diag, p_node, NOT_SUPPORT_SMP_ACCESS_REGISTER);

    IBDIAGNET_RETURN(NOT_SUP_ACC_REG);
}

int AccRegHandler::SendSMPReg(IBNode *p_curr_node,
                              phys_port_t port_num,
                              SMP_AccessRegister *smp_reg,
                              AccRegKey *p_key,
                              ProgressBar *p_progress_bar,
                              clbck_data_t *call_back_data)
{
    IBDIAGNET_ENTER;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = this;
    clbck_data.m_data1 = p_curr_node;
    clbck_data.m_data2 = p_key;

    clbck_data_t *p_clbck_data;
    if (call_back_data) {
        call_back_data->m_p_progress_bar = p_progress_bar;
        p_clbck_data = call_back_data;
    } else {
        clbck_data.m_handle_data_func = &forwardClbck<AccRegHandler,
                &AccRegHandler::SMPAccessRegisterHandlerGetClbck>;
        clbck_data.m_p_progress_bar = p_progress_bar;
        p_clbck_data = &clbck_data;
    }

    // handle direct route
    direct_route_t *p_curr_direct_route = NULL;
    p_curr_direct_route = this->p_reg->GetPhyDiag()->GetIBDiag()->GetDR(p_curr_node);

    if (!p_curr_direct_route) {
        this->p_reg->GetPhyDiag()->SetLastError("DB error - can't find direct route "
                                                "to node=%s (node guid: " U64H_FMT ")",
                                                p_curr_node->getName().c_str(),
                                                p_curr_node->guid_get());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    int rc = p_reg->PackDataSMP(p_key, smp_reg, p_curr_node);

    if (!rc)
        rc = this->p_reg->GetPhyDiag()->SMPAccRegGetByDirect(p_curr_direct_route,
                                                             port_num,
                                                             smp_reg,
                                                             p_clbck_data);

    IBDIAGNET_RETURN(rc);
}

int AccRegHandler::SendGMPReg(IBNode *p_curr_node,
                              uint16_t lid,
                              GMP_AccessRegister *gmp_reg,
                              AccRegKey *p_key,
                              ProgressBar *p_progress_bar,
                              clbck_data_t *call_back_data)
{
    IBDIAGNET_ENTER;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = this;
    clbck_data.m_data1 = p_curr_node;
    clbck_data.m_data2 = p_key;

    clbck_data_t *p_clbck_data;
    if (call_back_data) {
        call_back_data->m_p_progress_bar = p_progress_bar;
        p_clbck_data = call_back_data;
    } else {
        clbck_data.m_handle_data_func = &forwardClbck<AccRegHandler,
                &AccRegHandler::GMPAccessRegisterHandlerGetClbck>;
        clbck_data.m_p_progress_bar = p_progress_bar;
        p_clbck_data = &clbck_data;
    }

    int rc = p_reg->PackDataGMP(p_key, gmp_reg, p_curr_node);

    if (!rc)
        rc = this->p_reg->GetPhyDiag()->GMPAccRegGet(lid,
                                                     p_reg->GetRegisterID(),
                                                     gmp_reg,
                                                     p_clbck_data);

    IBDIAGNET_RETURN(rc);
}

int AccRegHandler::SendAccReg(AccRegVia_t acc_reg_via,
                              IBNode *p_curr_node,
                              phys_port_t port_num,
                              uint16_t lid,
                              AccessRegister &acc_reg,
                              AccRegKey *p_key,
                              ProgressBar *p_progress_bar,
                              clbck_data_t *call_back_data)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_ERR_CODE_DISABLED;

    if (this->p_reg->IsEnabledByPCAM(p_curr_node))
    {
        switch (acc_reg_via) {
            case VIA_SMP: {
                rc = this->SendSMPReg(p_curr_node,
                                      port_num,
                                      &acc_reg.smp_reg,
                                      p_key,
                                      p_progress_bar,
                                      call_back_data);

                break;
            }

            case VIA_GMP: {
                rc = this->SendGMPReg(p_curr_node,
                                      lid,
                                      &acc_reg.gmp_reg,
                                      p_key,
                                      p_progress_bar,
                                      call_back_data);
                break;
            }

            default:
                cerr << "Invalid Access Register type!" << endl;
        }
    }

    if (rc)
        delete p_key;

    IBDIAGNET_RETURN(rc == IBDIAG_ERR_CODE_DISABLED ? IBDIAG_SUCCESS_CODE : rc);
}

int AccRegPortLaneHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    for (map_str_pnode::iterator nI = p_reg->GetPhyDiag()->GetFabric()->NodeByName.begin();
         nI != p_reg->GetPhyDiag()->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByName map for key = %s",
                                                    nI->first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        // in case of SLRIP register - loop in lane is not relevant
        // if serdes version is 7nm
        if (p_reg->GetRegisterID() == ACCESS_REGISTER_ID_SLRIP) {
            if (p_curr_node->serdes_version >= SerDes_7nm)
                max_idx_in_lane = 1;
            else
                max_idx_in_lane = SLRIP_IB_SEL_NUM;
        }

        // go over all ports of this node:
        for (phys_port_t i = 1; i < p_curr_node->numPorts + 1; ++i) {

            IBPort *p_curr_port = p_curr_node->getPort(i);

            // port not exist
            if (!p_curr_port)
                continue;

            // skip ports that are not in sub fabric.
            if (!p_curr_port->getInSubFabric())
                continue;

            // validate disconnected
            if (p_curr_port->get_internal_state() <= IB_PORT_STATE_DOWN &&
                    (!p_reg->GetRetriveDisconnected() ||
                        p_curr_node->type == IB_CA_NODE))
                continue;

            // special port not supported
            if (p_curr_port->isSpecialPort())
                continue;

            for (u_int32_t ln = 0; ln < LANE_NUM; ln++) {
                for (u_int8_t lane_idx = 0; lane_idx < max_idx_in_lane; lane_idx++) {

                    AccessRegister acc_reg;
                    CLEAR_STRUCT(acc_reg);

                    AccRegKeyPortLane *p_plkey = new AccRegKeyPortLane(
                            p_curr_node->guid_get(),
                            p_curr_port->guid_get(),
                            i,
                            ln,
                            lane_idx);

                    rc = this->SendAccReg(acc_reg_via,
                                          p_curr_node,
                                          i,
                                          p_curr_port->base_lid,
                                          acc_reg,
                                          p_plkey,
                                          &progress_bar);

                    if (rc == IBDIAG_ERR_CODE_DB_ERR)
                        goto exit;

                } // idx_in_lane
            } // lane
        } // port
    } // node

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void AccRegPortLaneHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    this->ExportDataImplPort<AccRegKeyPortLane>(p_export_data_phy_port);

    IBDIAGNET_RETURN_VOID;
}

int AccRegNodeHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    for (map_str_pnode::iterator nI = this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.begin();
            nI != this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByName map for key = %s",
                                                    nI->first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        AccessRegister acc_reg;
        CLEAR_STRUCT(acc_reg);

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        AccRegKeyNode *p_nkey = new AccRegKeyNode(p_curr_node->guid_get());

        rc = this->SendAccReg(acc_reg_via,
                              p_curr_node,
                              0, // port 0
                              lid,
                              acc_reg,
                              p_nkey,
                              &progress_bar);

        if (rc == IBDIAG_ERR_CODE_DB_ERR)
            goto exit;

    } // nodes loop

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void AccRegNodeHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                   export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    this->ExportDataImplNode(p_export_data_phy_node);

    IBDIAGNET_RETURN_VOID;
}

AccRegKey* AccRegPortHandler::CreateKey(uint64_t ng, uint64_t pg, uint8_t pn)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(new AccRegKeyPort(ng, pg, pn));
}

int AccRegPortHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    for (map_str_pnode::iterator nI = this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.begin();
         nI != this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByName map for key = %s",
                                                    (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        // go over all ports of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {

            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);

            // port not exist
            if (!p_curr_port)
                continue;

            // skip ports that are not in sub fabric.
            if (!p_curr_port->getInSubFabric())
                continue;

            // validate disconnected
            if (p_curr_port->get_internal_state() <= IB_PORT_STATE_DOWN &&
                    (!p_reg->GetRetriveDisconnected() ||
                        p_curr_node->type == IB_CA_NODE))
                continue;

            // special port not supported
            if (p_curr_port->isSpecialPort())
                continue;

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            AccRegKey *p_pkey = CreateKey(p_curr_node->guid_get(),
                                          p_curr_port->guid_get(),
                                          i);

            rc = this->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  i,
                                  p_curr_port->base_lid,
                                  acc_reg,
                                  p_pkey,
                                  &progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR)
                goto exit;

        } // ports
    } // nodes

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void AccRegPortHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                   export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    this->ExportDataImplPort<AccRegKeyPort>(p_export_data_phy_port);

    IBDIAGNET_RETURN_VOID;
}


// ** AccRegPortIndexHandler **
AccRegKey* AccRegPortIndexHandler::CreateKey(uint64_t ng, uint64_t pg, uint8_t pn)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(new AccRegKeyPortIndex(ng, pg, pn, this->cur_index));
}

int AccRegPortIndexHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    for (uint64_t idx : indexes) {
        cur_index = idx;
        rc = AccRegPortHandler::BuildDB(phy_errors);
    }

    IBDIAGNET_RETURN(rc);
}


void AccRegHandler::DumpCSV(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;
    stringstream sstream;

    // Skiped. At least PPLLRegister::DumpRegisterData and SLTPRegister::DumpRegisterData
    // write warnings to the ibdiagnet2.log file
    (void) csv_out.DumpStart(p_reg->GetSectionName().c_str());

    p_reg->DumpRegisterHeader(sstream, handler_header);
    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    for (map_akey_areg::iterator nI = this->data_map.begin();
         nI != this->data_map.end(); ++nI) {
        AccRegKey * p_aregkey = (*nI).first;
        if (!p_aregkey)
        {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            IBDIAGNET_RETURN_VOID;
        }

        sstream.str("");
        p_aregkey->DumpKeyData(sstream);
        struct acc_reg_data areg = (*nI).second;
        p_reg->DumpRegisterData(areg, sstream, *p_aregkey);
        csv_out.WriteBuf(sstream.str());
    }
    csv_out.DumpEnd(p_reg->GetSectionName().c_str());
    IBDIAGNET_RETURN_VOID;
}
